/*      /\
 *     /  \		(C) Copyright 2016 Parliament Hill Computers Ltd.
 *     \  /		All rights reserved.
 *      \/
 *       .		Author: Alain Williams, February 2016
 *       .
 *        .		SCCS: @(#)CharInfo.c	1.3 07/13/17 18:51:56
 *          .
 *
 * Display information about a (unicode) character.
 * This is a simple developer test program, not intended for end user use.
 *
 * **** Look for the string FIXME and do so ****
 *
 * Processes startup files: /etc/xx.rc and $HOME/.xx.rc files
 *
 * Processes short and long options.
 * An option name cannot start with a with an open square bracket, see: RcFileSections
 *
 * There are different functions that do much the same thing, eg: Log() and LogMsg()
 * You prob only want one of them.
 *
 * Exit codes:
 * 0	OK
 * 10	Argument/option error - see Die_argerror
 * 11	Memory allocation error - see Die_memalloc
 */

#include <stdio.h>
#include <sys/types.h>
#include <stdlib.h>
#include <stdarg.h>

// If distributing far need configure that generates this:
// #ifdef HAVE_STRING_H
#include <string.h>
//#endif

#include <time.h>
#include <errno.h>
#include <pwd.h>

#define _GNU_SOURCE
#include <getopt.h>

// stolen from estruct.h
typedef	long	UniChar;		// Typedef & the compiler helps
#define CHAR	unsigned char		// Legacy, do not use
#define BYTE	unsigned char		// For underlying storage
typedef	long		LineLn;		// Line length (bytes or chars)
typedef	long		BufLns;		// # lines in a buffer
typedef	long		BufChrs;	// # chars or bytes in a buffer


#include "../utf8.h"
#include <wchar.h>

#define	CONFIG_MAX_LINE_LENGTH	1024		// Longest line acceptable in a configuration file
#define	RC_NAME		"CharInfo"		// Name used in finding config files: /etc/fred.rc & $HOME/.fred.rc
#define	RC_FILE_HOME	"/." RC_NAME ".rc"	// .rc file in home directory
#define	RC_FILE_ETC	"/etc/" RC_NAME ".rc"	// .rc file in /etc

char const* progname;	/* What this program is called - name of binary */
int	exit_c;		/* Exit status to return to O/S */

// Exit codes with special meanings
#define	Die_argerror	10	// Argument error
#define	Die_memalloc	11	// Memory allocation error


int	niceProg;	/* Nice this program */
int	Verbose;
int	Quiet;
char*	Username;	/* Logged in username */
char*	HomeDirectory;	/* User's $HOME */
int	CheckValid = 1;


UniChar	Character;	// The character that we show information about

/* **************** Total fudge ****************
 * We link in utf8.o this uses a few symbols elsewhere in microemacs
 * Provide dummies so that the linker works.
 * These are used in the following functions that MUST NOT be called:
 * utf8_errmsg() utf8_backToStart() utf8_CurrCharInfo() utf8_modecheck()
 */
char*	curwp;
char*	curbp;
void	mlwrite(void)
{
	;
}
int	lineMoreBytes(void)
{
	return 1;
}


//char*	a_arg;		// Arg to -a
//char*	na_arg;		// Arg to --na
//int	oa_opt;		// --oa given
//char*	oa_arg;		// Arg to --oa
//int	IntArg;		// Arg to --integer
//int	IntArg_Given;

/* Usage message.
 */
char*	usagem[] = {
	"This program displays information about a unicode character.",
	"",
	"Usage: %s [options]",
	"Options:",
//	"-a arg	Options that needs arg",
	"-c char",
	"	The literal character",
	"-u xxxx",
	"	Unicode point for the character, a hexadecimal number",
	"-h, --help",
	"	Help",
	"--no-check",
	"	Don't check code for validity - use with care",
//	"-i n, --integer=n",
//	"	Option that take a positive integer as an argument",
//	"--na arg, --na=arg",	// May want to add: -n n
//	"	Option that must have string argument arg",
//	"--oa, --oa=arg",
//	"	Option that may have optional argument arg",
//	"--nice	Nice the program",
	"-q, --quiet",
	"	Quiet",
	"-v, --verbose",
	"	Verbose, multiple times increases verbosity",
	"Initialisation files: " RC_FILE_ETC " and $HOME" RC_FILE_HOME,
	"Author: Alain D D Williams <addw@phcomp.co.uk> February 2016, licenced GPL 2",
	(char*)NULL	/* Must end the list */
};

// **************** Command line processing ****************
extern	int	optind, opterr, optopt;
extern	char*	optarg;
// GNU extensions, useful but not portable:
// -W opt	If OptString contains 'W;' treat as --opt  This is special use of -W which is allowed by POSIX
// x::		x takes an optional argument
// At start '-' means stop processing at first no option. Standard Unix getopt() does this anyway
//		GNU getopt() looks beyond unless this character is there
// ':' means return ':' if there is a missing argument.

#ifdef __gnu_linux__	// Attempt to be portable - prob fails
#define	GETOPT_MINUS	"-"
#else
#define	GETOPT_MINUS	""
#endif

//char const*	GetOptStr = GETOPT_MINUS ":a:hi:qv";		// Options (could dynamically build from long_options[] -- too much effort!)
char const*	GetOptStr = GETOPT_MINUS ":c:hqu:v";		// Options (could dynamically build from long_options[] -- too much effort!)
// 2nd elem: 0 no arg, 1 mandatory arg, 2 optional arg. 3rd elem always set NULL
char const*	CmdLineOnly = "h";			// Command line only options
char const*	NotCmdLine  = "";			// Option not on command line

// Constants for long arguments that do not have a corresponding short argument - make > 255
#define	OPT_na		256	// --na
#define	OPT_oa		257	// --oa
#define	OPT_nice	258	// --nice
#define	OPT_nocheck	259	// --no-check

int	shortOptionOrVal;		// Used below & in read_opts() and OptName()
// Keep in alphabetic order:
const	struct option long_options[] = {
	{"no-check",	0, &shortOptionOrVal, OPT_nocheck},
	{"help",	0, &shortOptionOrVal, 'h'},
//	{"integer",	1, &shortOptionOrVal, 'i'},		// Takes integer argument
//	{"na",		1, &shortOptionOrVal, OPT_na},		// Needs Argument
//	{"nice",	0, &shortOptionOrVal, OPT_nice},	// Nice program
//	{"oa",		2, &shortOptionOrVal, OPT_oa},		// Optional Argument
	{"quiet",	0, &shortOptionOrVal, 'q'},
	{"verbose",	0, &shortOptionOrVal, 'v'},
	{(char*)NULL,	0, (int*)NULL, 0}
};

char const* const	LongCmdLineOnly[] = {	// Array of long options only available on the command line
	"help",
	(char*)NULL
};

char const* const	LongNotCmdLine[] = {	// Array of long options not available on the command line
	(char*)NULL
};

// Array of zero or more sections that will be read in .rc files.
// The form of a section is (lines will be space trimmed at both ends):
//    [Section-Name]
// The section ends with the next line that starts with an open square bracket
//
// If the array is empty - then all lines in the .rc files are read,
// although lines starting with an open square bracket are then ignored.
char const* const	RcFileSections[] = {
//	"[template]",	// FIXME
	(char*)NULL
};
// If the config file has sections it might be desirable to have an initial, unnamed,
// section that will be treated as applying to everything. If you want to do this
// set the constant below to 1.
// Another way of doing this may be to have a section named [*] - or something.
#define	INITIAL_DEFAULT_SECTION		0

/* Functions defined in this module, in alphabetic order */
void			Die(int ec, char* Mes, ...);
void			Error(char* Mes, ...);
void			GetUserName(void);
void			Log(char* Mes, ...);
void			Note(char* Mes, ...);
const char*		OptName(int c, char const* const argv[]);
BYTE*			our_utf8_encode(UniChar uc);
void			ProcConfFile(char* ConfFile);
void			process(char const* CmdArg);
void			ReadOptions(int argc, char const* const argv[]);
void			read_opts(int argc, char const* const argv[]);
void			read_opts(int argc, char const* const argv[]);
char const* const*	strInArray(char const* const* strarray, const char* find);
char*			strmake(char const* Arg, char const* where);
void			usage(int exitc);

main(int argc, char const* const argv[])
{
	char*	str;

	progname = argv[0];

	GetUserName();	// Need HomeDirectory

	ReadOptions(argc, argv);

	// Print stuff out about Character
	printf("Character is U+%x = ", Character);
	str = our_utf8_encode(Character);
	while(*str)
		printf("%#2.2x ", *str++ & 0xff);
	printf("\n");

	printf("Width=%d\n", mk_wcwidth(Character));

	return(exit_c);
}

// **** Argument processing ****

/* Read options/arguments from various places:
 * RC_FILE_ETC
 * $HOME RC_FILE_HOME
 * command line
 *
 * If something print a message and exit
 */
void	ReadOptions(int argc, char const* const argv[])
{
	ProcConfFile(RC_FILE_ETC);

	// If we know the home directory:
	if(*HomeDirectory) {
		char*	home;
		char*	homerc = RC_FILE_HOME;

		if( !(home = (char*)malloc(strlen(HomeDirectory) + strlen(homerc) + 1)))
			Die(Die_memalloc, "Failed to allocate string for $HOME" RC_FILE_HOME);
		strcat(strcpy(home, HomeDirectory), homerc);
		ProcConfFile(home);
		free(home);
	}

	// Command line:
	read_opts(argc, argv);
}

/* Interpret '-' arguments -options.
 * If there is an error in the arguments, print the options list and exit.
 */
void	read_opts(int argc, char const* const argv[])
{
	register int	c;
	register int	error;
	int		i;

	error = 0;

	opterr = 1;	// Stop getopt_long printing error messages - we do it

	// Command line arguments:
	while((c = getopt_long(argc, (char* const*)argv, GetOptStr, long_options, &i)) != -1) {
		// This is a bit convoluted. Some short and long options are equivalent.
		// The simple way of managing this would be to set 'flag' (3rd member in each long_options)
		// to (char*)NULL - this would work.
		// However: when there is an error it is nice to be able to tell the user showing the exact
		// option that they used, ie not the long one if they gave the short one.

		// The only way to distinguish which was used is to set the 'flag', then the value of 'val'
		// is stored there. getopt_long() then returns 0 when a long option is recognised.
		// In this program 'flag' always points to shortOptionOrVal.

		// If it is a long only option the value will be > 255

		// See also OptName() complication about this

		const char*	opt_as_typed = OptName(c, argv);

		// printf("optname='%s' optopt=%c c=%d = '%c'\n", opt_as_typed, optopt, c, c);

		if(c == 0)	// Get the short (-x) value - this is what we test for
			c = shortOptionOrVal;

		if((c <= 255 && strchr(NotCmdLine, c)) || (c > 255 && strInArray(LongNotCmdLine, long_options[i].name))) {
			fprintf(stderr, "%s option '%s' is not available on the command line\n", progname, opt_as_typed);
			error = 1;
		} else
			error |= ProcArg(c, opt_as_typed, "command line");
	}

	if(error)
		usage(2);
}

/* Read and set options from a config (.rc) file.
 * It is not an error if the file does not exist.
 *
 * If there is an error, stop the program at the end of the file.
 */
void	ProcConfFile(char* ConfFile)
{
	int	error = 0;
	FILE*	cf;
	int	LineNo = 0;
	int	InSection = INITIAL_DEFAULT_SECTION;
	char	buf[CONFIG_MAX_LINE_LENGTH + 1];
	char	Errm[200];

	if(!(cf = fopen(ConfFile, "r")))
		return;

	while(fgets(buf, sizeof(buf), cf)) {
		char*	cp;

		LineNo++;
		snprintf(Errm, sizeof(Errm), "config file '%s' line %d", ConfFile, LineNo);

		if(cp = strchr(buf, '\n'))
			*cp = '\0';

		// Be nice to idiots who edit files on a MS Windows machine
		if((cp = strrchr(buf, '\r')) && cp[1] == '\0')
			*cp = '\0';

		// Trim: forwards & back
		cp = strchr(buf, '\0') - 1;
		while(cp >= buf && (*cp == ' ' || *cp == '\t'))
			*cp-- = '\0';
		cp = buf;
		while(*cp == ' ' || *cp == '\t')
			cp++;

		// Skip empty lines & comments
		if(*cp == '\0' || *cp == '#')
			continue;

		// If sections are defined, see if we are in/out of one
		if(RcFileSections[0]) {

			// Is this the start of a new section ?
			if(*cp == '[') {
				InSection = 0;

				// See if this is one of the ones listed.
				// If so note that we have found
				if(strInArray(RcFileSections, cp))
					InSection = 1;

				// Section lines are not option lines,
				// so skip this regardless of the above test result.
				continue;
			}

			// If we were not in a good/recognised section - abandon, get another line
			if( !InSection)
				continue;
		}

		error |= ProcConfLine(cp, Errm);

	}

	fclose(cf);

	if(error)
		usage(2);
}

/* Process a line out of the conf file - cp
 * Errm is a string saying something about where the error was found (eg line #)
 * Cope with options of the form '-c' and '--option' and 'option'
 *
 * Return true if error.
 *
 * Simplified & borrowed from SeekSpeed.c
 */
int	ProcConfLine(char* cp, char const* Errm)
{
	char*	os;
	int	opt;
	int	OptHasArg;
	char*	optStr;
	char	endCh;
	char	optBuf[4];
	struct option const*	op;

	if(*cp == '-' && cp[1] != '-') {
		// Single letter option, eg: -x
			
		optBuf[0] = '-';	// For error message
		optBuf[1] = cp[1];
		optBuf[2] = '\0';
		optStr = optBuf;

		// Search the option string:
		if(*++cp == '\0' || *cp == ':' || !(os = strchr(GetOptStr, *cp))) {
			fprintf(stderr, "%s unrecognised option: '%s'\n", Errm, optStr);
			return(1);
		}

		if(strchr(CmdLineOnly, *cp)) {
			fprintf(stderr, "%s option only allowed on command line: '%s'\n", Errm, optStr);
			return(1);
		}

		OptHasArg = os[1] == ':';
		opt = *cp++;

		endCh = ' ';	// Set to anything but '=' so that eating of spaces will happen
	} else {
		// Multiple letter option names
		char*	optionName;

		optStr = cp;
		// It can be '--something' or just 'something'; treat them as the same
		if(*cp == '-' && cp[1] == '-') {
			cp += 2;

			// Expect a word after the --
			if(*cp == '\0' || *cp == ' ' || *cp == '\t') {
				fprintf(stderr, "%s found '--' not followed by an option\n", Errm);
				return(1);
			}
		}
		// Check if it is --something
		optionName = cp;

		// Find the end of the option & NUL terminate
		while(*++cp != '\0' && *cp != ' ' && *cp != '\t' && *cp != '=')
				;
		endCh = *cp;
		*cp = '\0';

		// Look it up
		for(op = long_options; ; op++) {
			if( !op->name) {
				fprintf(stderr, "%s unrecognised option: '%s'\n", Errm, optStr);
				return(1);
			}

			if( !strcmp(op->name, optionName))
				break;		// Found it!
		}

		// What do we know about it ?
		opt = op->val;			// Value to switch on
		OptHasArg = op->has_arg == 1;	// Must this have an argument ? Optional args not handled

		// Skip the '\0' to next char on the line - unless we already reached end of line
		if(endCh != '\0')
			cp++;

		// If the ending character was white space, span foward on white space
		// eat any '=' that we may find. This is to allow something like:
		// option-name  =  option-value
		if(endCh == ' ' || endCh == '\t') {
			while(*cp == ' ' || *cp == '\t')
				cp++;
			if(*cp == '=')
				cp++;
			// Any spaces after the '=' are eaten below
		}

		// Some are for the command line only:
		if(strInArray(LongCmdLineOnly, optStr + 2)) {
			fprintf(stderr, "%s option only allowed on the command line: '%s'\n", Errm, optStr);
			return(1);
		}
	}

	// See if there is anything else on the line, span spaces.
	while(*cp == ' ' || *cp == '\t')
		cp++;

	if(OptHasArg) {
		if( !*cp) {
			fprintf(stderr, "%s option needs an argument: '%s'\n", Errm, optStr);
			return(1);
		}

		// If the argument starts and ends with a quote character, remove it:
		if(*cp == '\'' || *cp == '"') {
			char*	endch = cp + strlen(cp) - 1;

			if(*endch == *cp && endch != cp) {
				cp++;
				*endch-- = '\0';
			}
		}

		// The argument is the remaining characters on the line.
		optarg = cp;

	} else if( *cp != '\0') {
		fprintf(stderr, "%s option should not have argument: '%s%c%s'\n", Errm, optStr, endCh, cp);
		return(1);
	}

	return(ProcArg(opt, optStr, Errm));
}

/* Process an argument, be that from the command line of a .rc file:
 * c	The argument letter or something like OPT_example_long_option
 * opt	The string that makes the option, eg '-x' '--help' 'help' --- used in error messages
 * where Where the argument came from - for error messages
 *
 * **** WARNING ****
 * If the option has an argument it will be in global: optarg - beware, this might be over written,
 * so if the string value is to be kept a copy must be made.
 *
 * Return true if error
 * This (& elsewhere) can leak memory - but small amounts - not important.
 */
int	ProcArg(int c, char const* opt, char const* where)
{
	int	error = 0;
	long	tl;	/* temp long */
	char*	te;	/* temp end pointer */

	// printf("ProcArg(int %c (%d), opt '%s', char const* '%s') optarg='%s'\n", c, c, opt, where, optarg);

	switch(c) {
	case 'c':	/* Literal character */
		error |= ProcCarg(optarg);
		break;
	case 'h':	/* Help */
		usage(0);
		/*NOTREACHED*/
	case OPT_nice:	/* Nice outself */
		niceProg = 1;
		break;
	case 'u':	/* Utf8 as hex number */
		error |= ProcUarg(optarg);
		break;
	case 'v':	/* Verbose */
		Verbose++;
		break;
	case 'q':	/* Quiet */
		Quiet = 1;
		break;
	case OPT_nocheck: /* Check valid code */
		CheckValid = 0;
		printf("OPT_nocheck\n");
		break;
	case ':':
		Error("Argument missing to option '%s'", opt);
		error = 1;
		break;
	case '?':
		Error("Unknown option '%s'", opt);
		error = 1;
		break;
	default:
		Error("Internal consistency error - getopt returned %d (%#x)", c, c);
		error = 1;
		break;
	}

	return(error);
}

/* Output a usage message and end with exit status exitc.
 */
void	usage(int exitc)
{
	register char**	mes;
	
	for(mes = usagem; *mes != (char*)NULL; mes++) {
		(void) fprintf(stderr, *mes, progname);
		fputc('\n', stderr);
	}
	fprintf(stderr, "Version: %s\n", "1.3 07/13/17");

	exit(exitc);
	/*NOTREACHED*/
}

/* Return a string to a representation of the option with code c
 * The option may be a short or long one. The long name is returned by preference.
 * The string is in a static area.
 * If c is 0 then the option will have been a long one, shortOptionOrVal
 * This is only used when processing command line options.
 */
const char*	OptName(int c, char const* const argv[])
{
	static	char	opt[100] = "?";
	struct option*	op;

	// Dig through options to convert c to the long name.
	// Try this first, options have a short and long form, if we are here
	// we don't know which was given since the value for the long form is the short one.
	// Give the long one in preference since it is more meaningful.
	if(c == 0) {
		for(op = (struct option*)long_options; op->name; op++)
			if(op->val == shortOptionOrVal) {
				snprintf(opt, sizeof(opt), "--%s", op->name);
				return(opt);
			}

		// Should not get here
		snprintf(opt, sizeof(opt), "Unknown opt --%d (program error)", shortOptionOrVal);
		return(opt);
	}

	// printf("optind=%d '%s'\n", optind, argv[optind - 1]);

	// Unknown option, the character that is not known is in optopt.
	// Ditto if an argument is missing
	if(c == '?' || c == ':') {
		// Check the current option being processed. If it starts -- return that
		// Otherwise we get the single character option name - which will confuse
		if(argv[optind - 1][0] == '-' && argv[optind - 1][1] == '-')
			return(argv[optind - 1]);

		c = optopt;
	}

	// It was a short option
	if(c <= 255) {
		// Try to print nicely
		if(iscntrl(c))
			snprintf(opt, sizeof(opt), "-^%c", c | 0x40);
		else
			snprintf(opt, sizeof(opt), "-%c", c);
	} else {
		// Should not happen
		snprintf(opt, sizeof(opt), "--bad_long=%d", c);
	}

	return(opt);
}


// **** Logging functions ****

// Print fprintf type string and exit ec.
void	Die(int ec, char* Mes, ...)
{
	va_list ap;

	va_start(ap, Mes);
	fprintf(stderr, "%s: ", progname);
	vfprintf(stderr, Mes, ap);
	fprintf(stderr, "\n");
	va_end(ap);

	exit(ec);
}

// Print fprintf type string
void	Error(char* Mes, ...)
{
	va_list ap;

	va_start(ap, Mes);
	fprintf(stderr, "%s: ", progname);
	vfprintf(stderr, Mes, ap);
	fprintf(stderr, "\n");
	va_end(ap);
}

// Log a message to stderr, prefix with the time, stamp every day.
// Only log if Verbose.
// Arguments as printf()
void	logMsg(char* fmt, ...)
{
	va_list 		ap;
	static	struct tm	LastTM;	/* Time we were last here */
        static	struct tm*	NowTM;	/* Time now */
	time_t			now;

	if( ! Verbose)
                return;

	now = time((time_t*)NULL);
	NowTM = localtime(&now);

        /* Print date & tty name if changed since we were last here */
        if(NowTM->tm_yday != LastTM.tm_yday) {
		fprintf(stderr, "%4.4d/%2.2d/%2.2d %2.2d:%2.2d:%2.2d %s user %s\n",
			NowTM->tm_year + 1900, NowTM->tm_mon + 1, NowTM->tm_mday,
			NowTM->tm_hour, NowTM->tm_min, NowTM->tm_sec,
			progname, Username);
	}

	/* Always show current time */
	fprintf(stderr, "%2.2d:%2.2d:%2.2d ", NowTM->tm_hour, NowTM->tm_min, NowTM->tm_sec);

	va_start(ap, fmt);
	vfprintf(stderr, fmt, ap);
	va_end(ap);
	fprintf(stderr, "\n");

	fflush(stderr);	/* May not write for a long time */

	memcpy((void*)&LastTM, (void*)NowTM, sizeof(LastTM));	/* Set last time */
}

// Log something - unless quiet
void	Log(char* Mes, ...)
{
	va_list ap;

	if(Quiet)
		return;

	va_start(ap, Mes);
	fprintf(stderr, "%s: ", progname);
	vfprintf(stderr, Mes, ap);
	fprintf(stderr, "\n");
	va_end(ap);
}

// Log something - only if Verbose
void	Note(char* Mes, ...)
{
	va_list ap;

	if( ! Verbose)
		return;

	va_start(ap, Mes);
	fprintf(stderr, "%s: ", progname);
	vfprintf(stderr, Mes, ap);
	fprintf(stderr, "\n");
	va_end(ap);
}

// **** Misc Utility ****

/* Find a string in an array of strings.
 * strarray - NULL terminated array - haystack
 * find     - what to look for
 * Return: if found the address of member in strarray, otherwise NULL
 */
char const* const*	strInArray(char const* const* strarray, const char* find)
{
	for(; *strarray; strarray++)
		if( !strcmp(*strarray, find))
			return(strarray);

	return((char const *const *)NULL);
}

/* Return an allocated copy of the argument string Arg.
 * If the allocation fails, kill the program with where as part of the error message.
 */
char*	strmake(char const* Arg, char const* where)
{
	char*	s;

	if(s = strdup(Arg))
		return(s);

	Die(Die_memalloc, "%s out of memory in string duplicate", where);
}

/* Put something in Username - if not already there.
 * Also get $HOME into HomeDirectory -- "" if not known
 */
void	GetUserName(void)
{
	uid_t			uid;
	struct passwd*		pwuid;
	char*			home;
	char*			errm = "Program initialisation - GetUserName";

	if(Username)		/* Already done ? */
		return;

	/* Get username (for logging) */
	uid = getuid();	
	if((pwuid = getpwuid(uid)) && pwuid->pw_name)
		Username = strmake(pwuid->pw_name, errm);
	else {
		char	Buf[40];
		snprintf(Buf, sizeof(Buf) - 1, "uid %d", (int)uid);
		Username = strmake(Buf, errm);
	}

#if	1
	if( !(HomeDirectory = getenv("HOME")))
		HomeDirectory = "";
#else
	// An alternative way of doing it, not recommended by getpwent() man page
	// See also: https://developer.gnome.org/glib/stable/glib-Miscellaneous-Utility-Functions.html#g-get-home-dir
	// Get ~ ($HOME) for defaults file
	home = (pwuid && pwuid->pw_dir) ? pwuid->pw_dir : "";
	HomeDirectory = strmake(home, errm);
#endif
}

// **** Real work starts here ****

/* Process the -c option
 * Decode the character and assign to Character.
 * Return true if error.
 */
int	ProcCarg(char* opt)
{
	int	len = CHAR_LEN(*opt);

	if(len != strlen(opt)) {
		Error("The argument to -c is a character that should be %d bytes long, but %d were found", len, strlen(opt));
		return 1;
	}

	if( !utf8_check(opt)) {
		Error("The argument to -c is not a valid UTF8 character");
		return 1;
	}
	
	Character = utf8_decode(opt);

	return 0;
}

/* Process the -u option
 * Decode the character and assign to Character.
 * Return true if error.
 */
int	ProcUarg(char* opt)
{
	long	uc;	// Unicode char
	char	gash;	// Ensure nothing follows unicode #

	if(sscanf(opt, "%lx%c", &Character, &gash) != 1) {
		Error("Could not interpret as unicode character (hex) 'U+%s'", opt);
		return 1;
	}

	// Is this legal ?
	if(CheckValid && !utf8_isvalid(Character)) {
		Error("Error in U+%s: %s", opt, utf8_perror());
		return 1;
	}

	return 0;
}


// Do something with those pesky command line arguments
void	process(char const* CmdArg)
{
	printf("Process command line argument: %s\n", CmdArg);
}

// Stolen from random.c
/* Reverse overlapping copy.
 * Ie use this when a byte is to be inserted in the middle of a string.
 */ 
void	memrcpy(char* to, char* from, int len)
{
	to += len;
	from += len;

	while(--len >= 0)
		*--to = *--from;
}

/* Overlapped memory copyer
 * memcpy isn't guaranteed to work with overlapped copies.
 */ 
void	memocpy(char* to, char* from, int len)
{
	while(--len >= 0)
		*to++ = *from++;
}

// Stolen from utf8.c
// Difference is not checking for validity
/* Encode character uc as a UTF8 string.
 * Return the string - in a static buffer.
 * Return NULL on error, utf8_errno will be set.
 *
 * Note that this will be NUL terminated, and if the input was U+0
 * the string returned will be the empty string - followed by a NUL byte.
 */
BYTE*	our_utf8_encode(UniChar uc)
{
	static	BYTE	ret[UTF8_MAX_BYTES + 1];		// What we return
	BYTE*		s = ret;

#if	0
	// Throw up hands if not a valid code point:
	if( !utf8_isvalid(uc))
		return((BYTE*) NULL);
#endif

	if(uc <= UTF8_MAX_CP1)
		*s++ = (char) uc;
	else   if(uc <= UTF8_MAX_CP2) {
		*s++ = (char)( uc >>  6 | UTF8_B1_BITS2);
		*s++ = (char)((uc       & UTF8_Bn_VMASK) | UTF8_Bn_TPVALUE);
	} else if(uc <= UTF8_MAX_CP3) {
		*s++ = (char)( uc >> 12 | UTF8_B1_BITS3);
		*s++ = (char)((uc >>  6 & UTF8_Bn_VMASK) | UTF8_Bn_TPVALUE);
		*s++ = (char)((uc       & UTF8_Bn_VMASK) | UTF8_Bn_TPVALUE);
	} else {
		*s++ = (char)( uc >> 18 | UTF8_B1_BITS4);
		*s++ = (char)((uc >> 12 & UTF8_Bn_VMASK) | UTF8_Bn_TPVALUE);
		*s++ = (char)((uc >>  6 & UTF8_Bn_VMASK) | UTF8_Bn_TPVALUE);
		*s++ = (char)((uc       & UTF8_Bn_VMASK) | UTF8_Bn_TPVALUE);
	}

	*s = '\0';

	return(ret);
}



/* end */
