/*
 * ADDW SCCS: @(#)basic.c	1.80 07/13/20 14:44:28
 */

/* 
 * The routines in this file move the cursor around on the screen. They
 * compute a new value for the cursor, then adjust ".". The display code
 * always updates the cursor location, so only moves between lines, or
 * functions that adjust the top line in the window and invalidate the
 * framing, are hard.
 */
#include	"estruct.h"
#include	"utf8.h"
#include	"edef.h"

#if	SHOWMODREL
char	basicvers[] = "@(#)basic.c	1.80 07/13/20 14:44:28";
#endif

/* Warning that the end of buffer/narrow has been reached.
 */
void	sayEndOf(void)
{
	mlwrite("%W""%BEnd of %s", (curbp->b_abotline == curbp->b_linep) ? "buffer" : "narrow");
}

/* Warning that the top of buffer/narrow has been reached.
 */
void	sayTopOf(void)
{
	mlwrite("%W""%BTop of %s", (curbp->b_btopline == curbp->b_linep) ? "buffer" : "narrow");
}

/* Move the cursor to the beginning of the current line.
 *
 * Any count is ignored.
 *
 * Command:  beginning-of-line
 * Bound to: C-a   home
 *
 * Return TRUE
 */
int	gotobol(int f, Cnt n)
{
	curwp->w_doto = 0;
	curwp->w_partoff = 0;

	return(TRUE);
}

/* Move the cursor backwards by "n" characters.
 *
 * n must be >= 0.
 *
 * Error if you try and move out of the buffer. Set the flag if the line
 * pointer for dot changes.
 *
 * Return TRUE on success
 *	  FALSE on fail (hit end of buffer)
 */
int	moveBack(LineLn n)
{
	register LINE*	lp;

	while(--n >= 0) {
		if(curwp->w_doto == 0 && curwp->w_partoff == 0) {
			// Start of line, move back a line
			if(SonLine(lp = lback(curwp->w_dotp)))
				return(FALSE);

			curwp->w_dotp  = lp;
			curwp->w_doto  = lused(lp);
			curwp->w_flag |= WFMOVE;

			// Hex option - allow cursor to rest on 2nd hex digit
			if(curwp->w_opts & WO_HEX)
				curwp->w_partoff = 1;
		} else {
			// Move back within the line
			if(curwp->w_opts & WO_HEX) {
				// Hex option, move back over part chars
				if(--curwp->w_partoff < 0) {
					curwp->w_partoff = 1;
					curwp->w_doto--;
				}
			} else {
				// Move back a character, which might be more than 1 byte in UTF8 mode
				curwp->w_doto--;
				if(lisUtf8(curwp->w_dotp))
					UTF8_BACK_FIND_START_NX(lgetc(curwp->w_dotp, curwp->w_doto), curwp->w_doto);
			}
		}
	}

	return(TRUE);
}

/* Move the cursor backwards by "n" characters.
 *
 * Error if you try and move out of the buffer. Set the flag if the line
 * pointer for dot changes.
 *
 * Say warning if fail.
 *
 * Default count is 1. Must be >= 0
 *
 * Command:  backward-character
 * Bound to: C-b   left-arrow
 *
 * Return TRUE on success
 *	  FALSE on fail (hit end of buffer)
 */
int	backchar(int f, Cnt n)
{
	if(moveBack(n))
		return(TRUE);

	sayTopOf();

	return(FALSE);
}

/*
 * Move the cursor to the end of the current line. Trivial. No errors.
 * Note that this will result at the cursor being 'over' the virtual
 * newline at the line end.
 *
 * The count is ignored.
 *
 * Command:  end-of-line
 * Bound to: C-e   end
 *
 * Return TRUE
 */
/* ARGSUSED */
int	gotoeol(int f, Cnt n)
{
	curwp->w_doto  = lused(curwp->w_dotp);

	/* Hex option - allow cursor to rest on 2nd hex digit
	 * But in NONL mode that does not exist at theend of file.
	 */
	if((curwp->w_opts & WO_HEX) && !((curbp->b_mode & MDNONL) && EofLine(lforw(curwp->w_dotp))))
		curwp->w_partoff = 1;

	return(TRUE);
}

/* Move the cursor forewards by "n" characters.
 *
 * n must be >= 0
 *
 * Error if you try and move off the end of the buffer. Set the flag if the
 * line pointer for dot changes. It is benign (but a waste of time) to call
 * this multiple times at EOF.
 *
 * Return TRUE on success
 *	  FALSE on fail (hit end of buffer)
 */
int	moveForw(LineLn n)
{
	while(--n >= 0) {
		if(curwp->w_doto == lused(curwp->w_dotp)) {
			// End of line, move forwards a line

			/* NONL mode, do not move past the end of the last line */
			if((curbp->b_mode & MDNONL) && EofLine(lforw(curwp->w_dotp)))
				return(FALSE);

			if(curwp->w_opts & WO_HEX) {
				// Hex mode - allow cursor to rest on 2nd hex digit
				if(++curwp->w_partoff <= 1)
					continue;

				curwp->w_partoff = 0;
			}

			if(EonLine(curwp->w_dotp))
				return(FALSE);

			curwp->w_dotp  = lforw(curwp->w_dotp);
			curwp->w_doto  = 0;
			curwp->w_flag |= WFMOVE;
		} else {
			// The line after the narrow might be not empty
			// But can't got there
			if(EonLine(curwp->w_dotp))
				return(FALSE);

			// More text on the line
			if(curwp->w_opts & WO_HEX) {
				// Hex option: move foward a nybble first, then a byte
				if(++curwp->w_partoff > 1) {
					curwp->w_partoff = 0;
					curwp->w_doto++;
				}
			} else {
				if(lisUtf8(curwp->w_dotp))
					curwp->w_doto += CHAR_LEN(lgetc(curwp->w_dotp, curwp->w_doto));
				else
					curwp->w_doto++;
			}
		}
	}

	return(TRUE);
}

/* Move the cursor forwards by "n" characters.
 *
 * Error if you try and move off the end of the buffer. Set the flag if the
 * line pointer for dot changes. It is benign (but a waste of time) to call
 * this multiple times at EOF.
 *
 * Say warning if fail.
 *
 * Default count is 1. Must be >= 0
 *
 * Command:  forward-character
 * Bound to: C-f   right-arrow
 *
 * Return TRUE on success
 *	  FALSE on fail (hit end of buffer)
 */
int	forwchar(int f, Cnt n)
{
	if(moveForw(n))
		return(TRUE);

	sayEndOf();

	return(FALSE);
}

/* Implement forwCharSeveral() & backCharSeveral()
 *
 * n	The number of times to move the window width fraction
 * mv	The function that moves up/down, must be forwchar or backchar
 *
 * This is complicated by characters that do not have width 1,
 * these being Unicode double width and more importantly tabs.
 *
 * The whole point of this is to be able to move "." rapidly by
 * some fraction of the window width, not over a number of characters.
 *
 * Return TRUE if the movement(s) work
 *	  FAIL if hit start/end of line.
 */
int	moveHorizSeveral(Cnt n, int(*mv)(int f, Cnt n))
{
	int	wf;	// several fraction of window width

	// How much to move by. At least one character position
	if((wf = (int)((float)curwp->w_ncolumns / WINDOW_SEVERAL_FRACTION + 0.5F)) < 1)
		wf = 1;

	// Remember which line we start on
	LINE*	startLine = curwp->w_dotp;

	while(--n >= 0) {
		// Where do we want to get to ?
		LineLn	pos = getdcol(0);
		LineLn	getTo = pos + (mv == backchar ? -wf : wf);
		pnmark0();

		// Move one position at a time until we get to the right column:
		for( ; mv == backchar ? (pos > getTo) : (pos < getTo); pos = getdcol(0))
			if((*mv)(1, 1) == FALSE || curwp->w_dotp != startLine) {
				retmark0(FALSE);
				return(FALSE);
			}
	}

	return(TRUE);
}

/* Move "." several characters to the right.
 * Do not move to another line.
 * Move by 1 / WINDOW_SEVERAL_FRACTION
 *
 * Command name: forward-character-several
 *
 * Move argument * the fraction.
 *
 * This will stop when it reaches end of line.
 *
 * Default count 1. Must be >= 0
 *
 * Command:  forward-character-several
 * Bound to: S-right-arrow
 *
 * Return TRUE if the movement(s) work
 *	  FAIL if hit start/end of line.
 */
int	forwCharSeveral(int f, Cnt n)
{
	return(moveHorizSeveral(n, forwchar));
}

/* Move "." several characters to the left.
 * Do not move to another line.
 * Move by 1 / WINDOW_SEVERAL_FRACTION
 *
 * Command name: backard-character-several
 *
 * Move argument * the fraction.
 *
 * This will stop when it reaches start of line.
 *
 * Default count 1. Must be >= 0
 *
 * Command:  backward-character-several
 * Bound to: S-left-arrow
 *
 * Return TRUE if the movement(s) work
 *	  FAIL if hit start/end of line.
 */
int	backCharSeveral(int f, Cnt n)
{
	return(moveHorizSeveral(n, backchar));
}

/* Move "." to line lnum.
 *
 * Numeric mark 1 is placed at "." before the move.
 *
 * A number < 0 means count up from the bottom of the buffer.
 *
 * Return:
 *	TRUE  if this works.
 *	FALSE buffer does not have lnum lines.
 */
int	GotoLineNumber(BufLns lnum)
{
	LINE*	lp;
	BufLns	ln = lnum;

	if(lnum < 1)
		lnum = curbp->b_numLines + lnum + 1;

	if(lnum > curbp->b_numLines + 1) {
		mlwrite("%EBuffer only %D lines long", (long)curbp->b_numLines);
		return(FALSE);
	}

	// Move forward that number of lines
	for(lp = curbp->b_linep; --ln >= 0; lp = lforw(lp))
		;

	/* Is it within any narrow ?
	 * We could have done this test when we were moving forwards
	 * just now - keep it simple; only really bad on huge buffers.
	 */
	if( ! isInNarrow(lp, "Line %D not in narrow", (long)lnum))
		return(FALSE);

	pnmark1();

	/* first, we go to the start of the buffer */
        curwp->w_dotp  = lp;
        curwp->w_doto  = 0;
	curwp->w_partoff = 0;

	return(TRUE);
}

/* Move "." to a display column number on the current line.
 *
 * A number < 0 means count up from the end of the line.
 *
 * Return:
 *	TRUE  if this works.
 *	FALSE current line does not have col columns.
 */
int	GotoColNumber(LineLn col)
{
	LineLn	llength;

	// Length of line - display positions:
	llength = getccolAtCol(FALSE, lused(curwp->w_dotp));

	LineLn	ocol = col;	// Original col

	if(col < 0)
		col = llength + col;

	if(col < 0 || col > llength) {
		mlwrite("%ELine width %D, GOTO position %D", (long)llength, (long)ocol);
		return(FALSE);
	}

	curwp->w_doto = 0;

	while(getdcol(0) < col)	// Display positions, not character positions
		if(moveForw(1) == FALSE)
			return(FALSE);

	return(TRUE);
}

/* Move to a particular line.
 *
 * If the count is < 1 count from the bottom of the buffer.
 * A count of 0 will move "." to the line after the last line.
 *
 * Default count unset - in which case prompt for a line number.
 *
 * Command:  goto-line
 * Bound to: M-g
 *
 * Return TRUE on success
 *	  FALSE on fail (hit end of buffer, or line does not exist)
 *	  ABORT C-g
 */
int	gotoline(int f, Cnt n)
{
	register int status;	/* status return */

	/* get an argument if one doesnt exist */
	if(f == FALSE) {
		if((status = mlprompt("Line to GOTO: ", MLINTEGER | MLNOSPACE, bmlPromptHistp)) != TRUE) {
			mlwrite(MsgAborted);
			return(status);
		}
		n = atoCnt(ltext(bmlPromptHistp->b_dotp));
	}

	return(GotoLineNumber((BufLns)n));
}

/* Move to a column on the current line.
 *
 * The count n is the column number, if < 0 cound from end of line.
 * Columns counted from 0.
 *
 * The column is a display column not a character count or byte offset.
 *
 * If the count is not set: prompt for the line number.
 *
 * Command:  goto-column
 * Bound to: M-G
 *
 * Return TRUE  on success
 *	  FALSE if n is greater than the end of the current line
 *	        where "." will be left at the end of line.
 *	  ABORT C-g
 */
int	gotocolumn(int f, Cnt n)
{
	int	status;	// Status return

	/* get an argument if one doesnt exist */
	if(f == FALSE) {
		if((status = mlprompt("Column to GOTO: ", MLINTEGER | MLNOSPACE, bmlPromptHistp)) != TRUE) {
			mlwrite(MsgAborted);
			return(status);
		}
		n = atoCnt(ltext(bmlPromptHistp->b_dotp));
	}

	return(GotoColNumber((LineLn) n));
}

/* If at the top of the buffer, goto the bottom, else goto the top.
 *
 * This now moves to the top of the narrow.
 *
 * Mark 1 is set to "." before movement.
 *
 * The count is ignored.
 *
 * Command:  top-bottom-switch
 * Bound to: C-\
 *
 * Return TRUE
 */
int	topbot(int f, Cnt n)
{
	if(SonLine(lback(curwp->w_dotp)))
		return(gotoeob(f, n));
	else
		return(gotobob(f, n));
}

/* Goto the beginning of the buffer.
 *
 * Massive adjustment of dot. This is considered to be hard motion; it really
 * isn't if the original value of dot is the same as the new value of dot.
 *
 * If the buffer is narrowed: move to the first line in the narrow
 *
 * Mark 1 is set to "." before movement.
 *
 * The count is ignored.
 *
 * Command:  beginning-of-file
 * Bound to: M-<
 *
 * Return: TRUE on success - always
 */
/* ARGSUSED */
int	gotobob(int f, Cnt n)
{
	pnmark1();

	curwp->w_dotp  = lforw(curbp->b_btopline);
	curwp->w_doto  = 0;
	curwp->w_partoff = 0;
	curwp->w_flag |= WFHARD;

	return(TRUE);
}

/* Move to the end of the buffer.
 *
 * Dot is always put at the end of the file (ZJ). The standard screen code does
 * most of the hard parts of update.
 *
 * If the buffer is narrowed: move to the last line in the narrow
 *
 * Mark 1 is set to "." before movement.
 *
 * The count is ignored.
 *
 * Command:  end-of-file
 * Bound to: M->
 *
 * Return TRUE
 */
/* ARGSUSED */
int	gotoeob(int f, Cnt n)
{
	pnmark1();

        curwp->w_dotp  = curbp->b_abotline;
        curwp->w_doto  = 0;

	// NONL mode - cannot go beyond last line in buffer (not narrow)
	if((curbp->b_mode & MDNONL) && curwp->w_dotp == curbp->b_linep)
		moveBack(1);

	curwp->w_partoff = 0;
        curwp->w_flag |= WFHARD;

        return(TRUE);
}

/* Move up or down n lines.
 *
 * A few places internally use forwline() to move up/down the buffer.
 *
 * Say warning if fail.
 *
 * forwline() now assumes that the line count is not negative.
 * This checks the sign of the count and moves up/down as appropriate.
 *
 * Return:
 *	TRUE  if movement is successful - does not hit end of buffer.
 *	FALSE if the count is zero. Should not happen.
 */
int	signedForwline(Cnt n)
{
	if(n == 0)
		return(FALSE);

	if(n > 0)
		return(forwline(TRUE, n));

	return(backline(TRUE, -n));
}

/* Do the work of forwLineSeveral() backLineSeveral()
 *
 * n	Count of how many times to do it
 * mv	Function to move
 *
 * Return TRUE if the movement(s) work
 *	  FAIL if hit top/bottom of the buffer.
 */
int	moveVertSeveral(Cnt n, int(*mv)(int f, Cnt n))
{
	int	wf;	// several fraction of window height

	if((wf = (int)((float)curwp->w_ntrows / WINDOW_SEVERAL_FRACTION + 0.5F)) < 1)
		wf = 1;

	return((*mv)(1, n * wf));

}

/* Move "." several next lines - ie down.
 * Move by 1 / WINDOW_SEVERAL_FRACTION
 *
 * Move argument * the fraction.
 *
 * This will stop when it reaches bottom of file. Maybe it should stop
 * when it cannot move another fraction.
 *
 * Default count 1. Must be >= 0
 *
 * Command:  next-line-several
 * Bound to: S-down-arrow
 *
 * Return TRUE if the movement(s) work
 *	  FAIL if hit top/bottom of the buffer.
 */
int	forwLineSeveral(int f, Cnt n)
{
	return(moveVertSeveral(n, forwline));
}

/* Move "." several back lines - ie up.
 * Move by 1 / WINDOW_SEVERAL_FRACTION
 *
 * Command name: previous-line-several
 *
 * Move argument * the fraction.
 *
 * This will stop when it reaches top of file. Maybe it should stop
 * when it cannot move another fraction.
 *
 * Default count 1. Must be >= 0
 *
 * Command:  previous-line-several
 * Bound to: S-up-arrow
 *
 * Return TRUE if the movement(s) work
 *	  FAIL if hit top/bottom of the buffer.
 */
int	backLineSeveral(int f, Cnt n)
{
	return(moveVertSeveral(n, backline));
}

/* Move down by nlines lines.
 *
 * nlines must be >= 0
 *
 * The last command controls how the goal column is set.
 *
 * Stops at bottom of narrow.
 *
 * Return TRUE if moved
 *	  FALSE when did not - at end of buffer
 */
int	moveDown(BufLns nlines)
{
        register LINE*	dlp;

	// Find line columns. Do before EoB test to set line move in thisflag
	curgoalSet();

	/* And move the point down.
	 * In NONL mode do not move below the last line.
	 */
	dlp = curwp->w_dotp;
	while(nlines > 0 && !EonLine(dlp) &&
	      !((curbp->b_mode & MDNONL) && EonLine(lforw(dlp)))) {
		dlp = lforw(dlp);
		--nlines;
	}

	/* reseting the current position */
	curwp->w_dotp  = dlp;
	curwp->w_doto  = EonLine(dlp) ? 0 : getgoal(dlp);
//	curwp->w_partoff = 0;	// Maybe improve later
	curwp->w_flag |= WFMOVE;

	if(nlines == 0)
		return(TRUE);

	sayEndOf();

	return(FALSE);
}

/* Move forward by n lines.
 *
 * The last command controls how the goal column is set.
 *
 * Say warning if fail.
 *
 * Stops at bottom of narrow.
 *
 * Default count 1. Must be >= 0
 *
 * Command:  next-line
 * Bound to: C-n   down-arrow
 *
 * Return TRUE  if moved down all n lines
 *	  FALSE when did not move all n lines - "." at end of buffer/narrow
 */
int	forwline(int f, Cnt n)
{
	return(moveDown(n));
}

/* Move up nlines lines.
 *
 * This function is like "forwline", but goes backwards. The scheme is exactly
 * the same. Figure out the new line and call "movedot" to perform the motion.
 *
 * Stops at top of narrow.
 *
 * nlines must be >= 0
 *
 * Return TRUE if moved up nlines
 *	  FALSE when did not - leave "." at start of buffer/narrow
 *		"." might have moved up some lines, just not all nlines
 *		The 'Top of ...' message will be generated
 */
int	moveUp(BufLns nlines)
{
	register LINE*	dlp;

	// Find line columns. Do before EoB test to set line move in thisflag
	curgoalSet();

	/* and move the point up */
	dlp = curwp->w_dotp;
	while(nlines > 0 && !SonLine(lback(dlp))) {
		dlp = lback(dlp);
		--nlines;
	}

	/* reseting the current position */
	curwp->w_dotp  = dlp;
	curwp->w_doto  = getgoal(dlp);
//	curwp->w_partoff = 0;	// Maybe improve later
	curwp->w_flag |= WFMOVE;

	if(nlines == 0)
		return(TRUE);

	sayTopOf();

	return(FALSE);
}


/* Move up n lines.
 *
 * This function is like "forwline", but goes backwards. The scheme is exactly
 * the same. Figure out the new line and call "movedot" to perform the motion.
 *
 * Say warning if fail.
 *
 * Stops at top of narrow.
 *
 * Default count 1. Must be >= 0
 *
 * Command:  previous-line
 * Bound to: C-p   up-arrow
 *
 * Return TRUE  if moved up n lines
 *	  FALSE when did not move all n lines - "." at start of buffer/narrow
 */
int	backline(int f, Cnt n)
{
	return(moveUp(n));
}

/* Go back to the beginning of the count paragraphs.
 *
 * Here we look for a <NL><NL> or <NL><TAB> or <NL><SPACE>
 * or <NL><DOT> combination to delimit the beginning of a paragraph.
 *
 * If "." does not contain alphanumeric first move up until we are on
 * a line that does; this is deemed to be in a paragraph. Then search
 * for the start of the paragraph.
 *
 * In modes: INDENT, C or PHP lines starting with the same width of non
 * alphanumerics are seen as being in the same paragraph. This 'looks through'
 * things like comment lead-ins at start of line.
 * However allow the first  paragraph to be indented and subsequent ones start
 * at column 0.
 *
 * Need to unnecessarily initialise indent to stop the compiler wingeing.
 *
 * Works with UTF-8 when inword() does.
 *
 * Default count is 1. Must be >= 0
 *
 * Command:  previous-paragraph
 * Bound to: M-p
 *
 * Return TRUE if moved, even if not all requested n paragraphs
 *	  FALSE when did not - at end of buffer
 */
#if	WORDPRO
int	gotobop(int f, Cnt n)
{
	UniChar	ch;
	int	movedl = 0;
	int	doIndent = curbp->b_mode & (MDINDEN | MDPHP | MDCMOD);
	LineLn	indent = 0;		// Line indent if doing indent
	int	moved = FALSE;

	curwp->w_flag |= WFMOVE;	// force screen update - we will prob move

	while(--n >= 0) {	/* for each one asked for */
		LINE*	startLINE = curwp->w_dotp;
		LineLn	startOFF  = curwp->w_doto;

		/* first scan back until we are in a word */
		if(moveBack(1) == FALSE)
			goto hitTop;

		while(!inword())
			if(moveBack(1) == FALSE)
				goto hitTop;

		curwp->w_doto = 0;	/* and go to the B-O-Line */

		/* In indent mode, record the level of indent.
		 * This is so with which we have something to compare lines that
		 * we move to.
		 *
		 * If the level of indent is -1 this means that there were not alpha
		 * characters on the line ... but there could have been non-white space
		 * this is really important as it could mean a '*' or '//' that
		 * should be skipped over to find the previous paragraph of a comment
		 * (also very important to fill-paragraph).
		 * So: skip back over these lines.
		 */
		if(doIndent) {
			while((indent = GetWordCol(1)) == -1 && 
			      !EofLine(lback(curwp->w_dotp)))
				curwp->w_dotp = lback(curwp->w_dotp);
		}

		/* Scan back until we hit a <NL><NL> or <NL><TAB>
		 * or a <NL><SPACE> or <NL><DOT>
		 * If Indent: move over lines that have the same level of indent.
		 */
		while( ! SonLine(lback(curwp->w_dotp)) &&
		      lused(curwp->w_dotp) != 0 &&
		      (ch = lgetc(curwp->w_dotp, 0)) != '.') {
			if(doIndent) {
				/* Indent mode: stop when see a different level of indent
				 * If the indent is 0, peek at the line before, if it has
				 * the same indent as the one just moved to then it is
				 * part of paragraph that the one just moved to - ie not
				 * part of the paragraph that we have been looking at.
				 */
				LineLn	newIndent = GetWordCol(1);
				if(newIndent != indent) {
					// But first return to that line of previous indent:
					if(indent == 0 && ! SonLine(lback(curwp->w_dotp))) {
						// Peek at previous:
						curwp->w_dotp = lback(curwp->w_dotp);
						if(GetWordCol(1) == newIndent)
							curwp->w_dotp = lforw(curwp->w_dotp);
					}

					// Move back to the start of the paragraph
					curwp->w_dotp = lforw(curwp->w_dotp);
					break;
				}
			} else {
				// Non indent - stop when see white space
				if(ch == CharTab || ch == ' ')
					break;
			}

			// Go back a line
			curwp->w_dotp = lback(curwp->w_dotp);
			movedl = 1;
		}

		/* If we found a line starting '.' move to the next line.
		 * However if we started on this line just move to the BOL and
		 * prevent the movement forward over the '.'.
		 */
		if(lused(curwp->w_dotp) > 0 &&
		   lgetc(curwp->w_dotp, 0) == '.') {
			if(!movedl) {
				curwp->w_doto = 0;
				if( !CheckCanMoveUp(FALSE))
					return(moved);

				continue;
			}
			else
				curwp->w_dotp = lforw(curwp->w_dotp);
		}

		/* And then forward until we are in a word */
		while(!inword() && moveForw(1))
			;

		// If we did not move: stop
		if(startLINE == curwp->w_dotp && startOFF == curwp->w_doto)
			goto hitTop;

		moved = TRUE;
	}

	return(moved);

hitTop:	// Hit the top of the buffer
	sayTopOf();

	return(moved);
}

/* Go forward to the end of the count paragraphs.
 *
 * if we aren't in a paragraph get into one first (this allows us to move
 * forward paragraphs easily). Here we look for a <NL><NL> or <NL><TAB> or
 * <NL><SPACE> or <NL><DOT> combination to delimit the end of a paragraph.
 *
 * In modes: INDENT, C or PHP lines starting with the same width of non
 * alphanumerics are seen as being in the same paragraph. This 'looks through'
 * things like comment lead-ins at start of line.
 * However allow the first  paragraph to be indented and subsequent ones start
 * at column 0.
 *
 * Also in C/PHP we span over '*' like after this comment, and also '//'
 * as in a block of that style comment.
 *
 * Need to unnecessarily initialise indent to stop the compiler wingeing.
 *
 * Works with UTF-8 when inword() does.
 *
 * Default count is 1. Must be >= 0
 *
 * Command:  next-paragraph
 * Bound to: M-n
 *
 * Return TRUE if moved, even if not all requested n paragraphs
 *	  FALSE when did not - at end of buffer
 */
int	gotoeop(int f, Cnt n)
{
	int	ch;
	int	doIndent = curbp->b_mode & (MDINDEN | MDPHP | MDCMOD);
	LineLn	indent = 0;		// Line indent if doing indent
	int	moved = FALSE;

	curwp->w_flag |= WFMOVE;	// force screen update - we will prob move

	while(--n >= 0) {	/* for each one asked for */
		int	lines = 0;
		LINE*	startLINE = curwp->w_dotp;
		LineLn	startOFF  = curwp->w_doto;

		/* first scan forward until we are in a word */
		if(moveForw(1) == FALSE)
			goto hitBot;

		while(!inword())
			if(moveForw(1) == FALSE)
				goto hitBot;

		/* In indent mode, record the level of indent.
		 * This is so with which we have something to compare lines that
		 * we move to.
		 *
		 * If the level of indent is -1 this means that there were not alpha
		 * characters on the line ... but there could have been non-white space
		 * this is really important as it could mean a '*' or '//' that
		 * should be skipped over to find the next paragraph of a comment
		 * (also very important to fill-paragraph).
		 * So: skip forward over these lines.
		 */
		if(doIndent) {
			while((indent = GetWordCol(1)) == -1 && 
			      !EonLine(lforw(curwp->w_dotp)))
				curwp->w_dotp = lforw(curwp->w_dotp);
		}

		/* Lines starting '.' - just move to the end of line,
		 * that is deemed an end of paragraph.
		 * NB: Line after narrow can be not empty
		 */
		if(lused(curwp->w_dotp) != 0 && lgetc(curwp->w_dotp, 0) == '.' &&
				!EonLine(curwp->w_dotp)) {
			curwp->w_doto = lused(curwp->w_dotp);	// Virtual \n at line end
		} else {
			curwp->w_doto = 0;	/* and go to the B-O-Line */

			/* Move to the next line if not at end of narrow */
			if( ! EonLine(curwp->w_dotp)) {
				curwp->w_dotp = lforw(curwp->w_dotp);
				lines++;
				if(EonLine(curwp->w_dotp))
					goto hitBot;
			}

			/* And scan forword until we hit a <NL><NL> or <NL><TAB>
			 * or a <NL><SPACE> or <NL><DOT>
			 * If Indent: move over lines that have the same level of indent.
			 */
			while( ! EonLine(curwp->w_dotp) &&
			      lused(curwp->w_dotp) != 0 &&
			      (ch = lgetc(curwp->w_dotp, 0)) != '.') {

				if(doIndent) {
					// Indent mode: stop when see a different level of indent
					LineLn	newIndent = GetWordCol(1);

					if(newIndent != indent) {
						/* OK to accept a new indent if it is 0,
						 * below one line indented by more.
						 * Ie have an indented first line and the rest is start of line
						 */
						if(newIndent != 0)
							break;
						if(lines > 1)
							break;

						// Rest of para - no indent
						indent = 0;
					}
				} else {
					// Non indent - stop when see white space
					if(ch == CharTab || ch == ' ')
						break;
				}

				// Go forwards a line
				curwp->w_dotp = lforw(curwp->w_dotp);
				lines++;
			}
		}

		/* And then backward until we are in a word */
		while(moveBack(1) && !inword())
			;

		/* Which took us back one too far */
		if(moveForw(1) == FALSE)
			goto hitBot;

		/* If we did not move then restart the next-paragraph starting
		 * on the next line.
		 * This can happen on a line starting '.' when we just end up
		 * moving back & forth over the non word characters at line end.
		 */
		if(startLINE == curwp->w_dotp && startOFF == curwp->w_doto) {
			if(EonLine(lforw(startLINE)))
				goto hitBot;

			// Next line, reset count as the last loop did nothing:
			curwp->w_dotp = lforw(startLINE);
			curwp->w_doto = 0;
			n++;
		} else {
			moved = TRUE;
		}
	}

	return(moved);

hitBot:	// Hit the bottom of the buffer
	sayEndOf();

	return(moved);
}
#endif

/* Check that there is something above that we can move to.
 * 
 * If not display an error and return FALSE.
 *
 * If page - looking to see if we can move the top of page up, else check moving the current line.
 */
int	CheckCanMoveUp(int page)
{
	LINE*	lp = page ? curwp->w_topLinep : curwp->w_dotp;

	/* if we are on the first line as we start....fail the command */
	if(SonLine(lback(lp))) {
		sayTopOf();
		return(FALSE);
	}

	return(TRUE);
}



/* Given a pointer to a LINE, and the current cursor goal column, return the
 * best choice for the offset.
 *
 * The offset is returned. Used by forwline(), backline(), forwpage() and
 * backpage().
 *
 * Also try to restore the amount of window scroll left.
 */
LineLn	getgoal(const LINE* dlp)
{
	/* If changing (== restoring) a left scroll note that the window needs hard
	 * refresh. This is expensive so only do it if we need to.
	 * Although we have requested an offset we might not get it if the line
	 * is not long enough to show it.
	 */
	if(curwp->w_leftoffset != curwp->w_goalLeftoffset) {
		curwp->w_flag  |= WFMODE|WFHARD;
		curwp->w_leftoffset = curwp->w_goalLeftoffset;
	}

	return getdoto(curwp, dlp, curwp->w_goalCol, 
			curwp->w_opts & WO_HEX ? DisCol_HEX : (curwp->w_opts & WO_META_ATTRIB ? DisCol_MetaChar : DisCol_Char));
}

/* If the last command was not a line move, reset the goal column and window
 * scroll offset.
 *
 * This is so that we can try to restore them on a subsequent line
 * that the cursor rests on.
 *
 * This is only used by line/page up/down.
 *
 * These functions call curgoalSet() to start with and then getgoal() to assign to w_doto.
 * NB: lastflag & thisflag.
 */
void	curgoalSet(void)
{
	if((lastflag & CFCPCN) == 0) {
		curwp->w_goalCol = getdcol(0);
		curwp->w_goalLeftoffset = curwp->w_leftoffset;
	}

	/* Flag this command as a line move */
	thisflag |= CFCPCN;
}

/* Scroll forward by n pages.
 *
 * Now tries to preserve column # & line number from top of page.
 * If we hit the bottom of the buffer, ensure that the last line is visible/
 *
 * Ensure that "." moves down a multiple of whole page lengths or not at all.
 *
 * This used to have the compile time config CVMVAS
 *
 * Default count 1. Must be >= 0
 *
 * Command:  next-page
 * Bound to: C-v   page-down
 *
 * Return TRUE if moved, even if not n times (where n > 1)
 *	  FALSE when did not - at end of buffer
 */
int	forwpage(int f, Cnt n)
{
	LINE*	lp = curwp->w_dotp;		// Line that the cursor is on
	int	didSomething = !n;

	// pl - number of lines to move down for a page, aka page length
	int	pl = curwp->w_ntrows - PageContextLines;
	if(pl < 1)
		pl = 1;

	/* Convert from pages to lines.
	 * Note if move forward X pages, we move (X - 2) * n, you might expect X * n - 2
	 * but that would mean once going forward 4 pages, then back 1 page 4 times
	 * would not return to the same place.
	 *	n *= curwp->w_ntrows - PageContextLines; // Convert from pages to lines
	 */

	// Find line columns. Do before EoB test to set line move in thisflag
	curgoalSet();

	while(--n >= 0) {
		// Do a page at a time so get error if try to move too many pages

		int	pageLen;

		if(curwp->w_dotp == curbp->b_abotline)
			break;

		/* "." is important, not top of page.
		 * If "." cannot move down a page, we do not move the top line.
		 * However if the bottom of the buffer is not visible
		 * then move w_topLine down until it is.
		 *
		 * Can "." be moved down a page ?
		 */
		for(pageLen = pl, lp = curwp->w_dotp;
		    lforw(lp) != curbp->b_abotline && --pageLen >= 0; )
			lp = lforw(lp);

		if(pageLen <= 0) {
			/* Yes it can.
			 * So move the top line down a page as well.
			 */
			curwp->w_dotp = lp;
			for(pageLen = pl; --pageLen >= 0; ) {
				curwp->w_topLinep = lforw(curwp->w_topLinep);
				curwp->w_toplinenum++;
			}
			didSomething = 1;
		} else {
			/* Hit the bottom when trying to move "." down a page.
			 * Set w_topLine so that the last line is visible.
			 * Do this by counting up from the bottom line.
			 *
			 * If we see the current top then stop, the bottom is already
			 * being shown so don't change anything.
			 */
			LINE*	bot = lback(curbp->b_abotline);
			BufLns	blno = curbp->b_numLines;

			while(--pl >= 0) {
				bot = lback(bot);
				blno--;
				if(bot == curwp->w_topLinep)
					break;
			}

			if(bot != curwp->w_topLinep) {	// Top moved ?
				didSomething = 1;
				curwp->w_topLinep = bot;
				curwp->w_toplinenum = blno;
			}

			/* There is one special case: "." could go to b_abotline.
			 * This is not a hardly-ever, consider: end-of-file,
			 * several previous-page then several next-page.
			 *
			 * Our first scan down will tell us if this is the case:
			 */
			if(lforw(lp) == curbp->b_abotline && pageLen == 1) {
				didSomething = 1;
				curwp->w_dotp = curbp->b_abotline;
			}

			// Can't go further
			break;
		}
	}

	if( !didSomething) {
		sayEndOf();
		return(FALSE);
	}

	// Where in the line do we place "." ?
	if(curwp->w_dotp == curbp->b_abotline)
		curwp->w_doto = 0;	// Nothing to see here
	else
		curwp->w_doto = getgoal(lp);

	curwp->w_flag |= WFHARD | WFNWID;

	return(TRUE);
}

/* Scroll backwards by n pages.
 *
 * The overlap between the two windows is PageContextLines ($page-context and $overlap).
 * Default 2 - the value is from the ITS EMACS manual.
 *
 * We do a hard update for exactly the same reason.
 *
 * Now tries to preserve column # & line number from top of page (lnftop).
 * When the top of page hits the top of buffer; lnftop might increase (ie "."
 * moves down the page) so that "." stays on the same line. This means that
 * X page ups followed by X page downs will result in "." getting back to
 * the same starting line.
 *
 * This used to have the compile time config CVMVAS
 *
 * Default count 1. Must be >= 0
 *
 * Command:  previous-page
 * Bound to: M-v   page-up
 *
 * Return TRUE if moved, even if not n times (where n > 1)
 *	  FALSE when did not - at start of buffer
 */
int	backpage(int f, Cnt n)
{
	LINE*	lp;
	int	didSomething = 0;

	// pl - number of lines to move down for a page, aka page length
	int	pl = curwp->w_ntrows - PageContextLines;
	if(pl < 1)
		pl = 1;

	/* Convert from pages to lines.
	 * Note if move forward X pages, we move (X - 2) * n, you might expect X * n - 2
	 * but that would mean once going forward 4 pages, then back 1 page 4 times
	 * would not return to the same place.
	 *	n *= curwp->w_ntrows - PageContextLines; // Convert from pages to lines
	 */

	// Find line columns. Do before EoB test to set line move in thisflag
	curgoalSet();

	while(--n >= 0) {
		// Do a page at a time so get error if try to move too many pages

		int	pageLen;

		/* Move top of page up by a pagefull
		 * At the top of buffer might not be able to move a full page
		 */
		for(pageLen = pl; --pageLen >= 0 && !SonLine((lp = lback(curwp->w_topLinep))); ) {
			curwp->w_topLinep = lp;
			curwp->w_toplinenum--;
			didSomething = 1;
		}

		/* Now move "." up by a page full.
		 * It might be possible to move "." up a page even if the top of page
		 * does not move much/at-all - this can happen near the top of buffer.
		 */
		for(pageLen = pl, lp = curwp->w_dotp; --pageLen >= 0 && !SonLine(lback(lp)); )
			lp = lback(lp);

		/* If lp was moved by a pagefull - set "." to lp
		 * If not then "." is visible anyway - so leave it (but a bit
		 * higher up in the window)
		 */
		if(pageLen == -1) {
			curwp->w_dotp = lp;
			didSomething = 1;
		}
	}

	if(!didSomething) {
		sayTopOf();
		return(FALSE);
	}

	curwp->w_doto  = getgoal(curwp->w_dotp);
	curwp->w_flag |= WFHARD | WFNWID;

	return(TRUE);
}

/* Set the mark in the current window to the value of "." in the window.
 *
 * No errors are possible.
 *
 * This counts as a line change from the display point of view since we need
 * to show the mark with a different colour.
 *
 * Say that the mark has been set unless the count it set.
 *
 * This sets the part byte offset (think HEX). This is of very little real consequence
 * but it means that the user can return to the part byte that they were on.
 * Copy/delete/... ignores it.
 *
 * Default count unset.
 *
 * Command:  set-mark
 * Bound to: C-SPC   M-SPC   M-.
 *
 * Return TRUE
 */
/*ARGSUSED*/
int	setmark(int f, Cnt n)
{
	curwp->w_markp = curwp->w_dotp;
	curwp->w_marko = curwp->w_doto;
	curwp->w_markpartoff = curwp->w_partoff;
        curwp->w_flag |= WFEDIT;

	if(f == FALSE)
		mlwrite("%IMark set");

	return(TRUE);
}

/* Swap the values of "." and "mark" in the current window.
 *
 * This is pretty easy, bacause all of the hard work gets done by the standard
 * routine that moves the mark about. The only possible error is "no mark".
 *
 * The count is ignored.
 *
 * Command:  exchange-point-and-mark
 * Bound to: C-x C-x
 *
 * Return TRUE if mark & "." swapped
 *	  FALSE if the buffer does not contain a mark
 */
/*ARGSUSED*/
int	swapmark(int f, Cnt n)
{
	register LINE*	odotp;
	LineLn		odoto;
	char		opart;	// Preserve part offset for mark - hex mode

	if(curwp->w_markp == NULL) {
		mlwrite("%WNo mark in this window");
		return(FALSE);
	}

	pnmark1();

	odotp = curwp->w_dotp;
	odoto = curwp->w_doto;
	opart = curwp->w_partoff;
	curwp->w_dotp  = curwp->w_markp;
	curwp->w_doto  = curwp->w_marko;
	curwp->w_partoff  = curwp->w_markpartoff;
	curwp->w_markp = odotp;
	curwp->w_marko = odoto;
	curwp->w_markpartoff = opart;
	curwp->w_flag |= WFHARD | WFNWID;

	return(TRUE);
}

/* end */
