/*
 * ADDW SCCS: @(#)bind.c	1.122 10/14/20 16:22:19
 */

/*	This file is for functions having to do with key bindings,
	descriptions, help commands and startup file.

	written 11-feb-86 by Daniel Lawrence
								*/

#include	"estruct.h"
#include	"utf8.h"
#include	"edef.h"
#include	"efunc.h"

#if	MSDOS && (TURBO || LATTICE || AZTEC || MSC)
#include	"epath.h"
#endif

#if	SHOWMODREL
char	bindvers[] = "@(#)bind.c	1.122 10/14/20 16:22:19";
#endif

/* In the listing of bindings there are 3 columns, the command name, the
 * key sequence and a description. The name is padded with spaces.
 * 18 as some bindings descriptions can be quite long.
 */
#define	BINDING_LIST_NAME_LEN	30
#define	BINDING_LIST_KSEQ_LEN	18	// Space for key sequence


/* Initialise the hash table that indexes CmdNames
 * This must be called early.
 */
void	hashInitBinds(void)
{
	hashInit(&hashCmdNames);
}

/* Describe the command for a certain key.
 *
 * If given an argument this will display the help page for the
 * command that the key is bound to.
 *
 * Default count is unset.
 *
 * Command:  describe-key
 * Bound to: C-x ?
 *
 * Returns:
 *	TRUE  success
 *	FALSE
 *	    The key is not something that is bound
 *	    Cannot open the help window
 */
int	deskey(int f, Cnt n)
{
	register UniChar c;		/* key to describe */
	register BYTE*	ptr;		/* string pointer to scan output strings */
	NBIND*		np;		/* Pointer to binding */
	char*		desc;		/* Description */
	BYTE		outseq[NSTRING];/* output buffer for command sequence */
	BYTE		keyprompt[] = ": describe-key ";

	/* Prompt the user to type us a key to describe */
	mlwrite(keyprompt);

	/* Get the command sequence to describe
	 * change it to something we can print as well.
	 * Want the cursor in message window.
	 */
	CursorInMsg = 1;
	if((c = getckey(FALSE)) == -1)
		return(FALSE);

	cmdstr(c, &outseq[0]);	// Make printable

	/* Find the right function description & get it's name.
	 * &DummyUnboundKey if not bound
	 */
	np = getbind(c, TRUE);
	ptr = np->n_name;
	if((desc = np->n_desc) == NULL)
		desc = "";

	if(f) {
		BUFFER*	bp;
		if( !(bp = bfindCheck(HelpBname, BFNCREATE|BFNACTIVE)))
			return(FALSE);

		// If it is not displayed split the current window:
		if(bp->b_nwnd == 0 && splitwind(FALSE, 1) == FALSE)
			return(FALSE);

		return(helpCmdShow(np->n_name));
	}

	mlwrite("%s%s %s %s", keyprompt, outseq, ptr, desc);

	return(TRUE);
}

/* Find a procedure's entry in ProcNames.
 *
 * name		The name of the procedure
 * create	TRUE if to create an entry if it could not be found.
 *
 * If an entry is created a malloced copy of name is used.
 *
 * Beware: the result returned should be used quickly, the address might change
 * after a bit -- after bind-procedure-to-key
 *
 * Return:
 *	A pointer to the entry
 *	NULL if not found or could not create (out of memory)
 */
NBIND*	procFind(BYTE* name, int create)
{
	NBIND*	probe;
	BYTE*	nname;		// Copy of name

	// Linear search is prob good enough, I doubt that there will be many
	for(probe = ProcNames; probe->n_name != NULL; probe++) {
		if(strcmp(name, probe->n_name) == 0)
			return(probe);
	}

	if(create == 0)
		return((NBIND*)NULL);

	int	entries = (int)(probe - ProcNames);

	// Make the table bigger & allocate the name
	if((nname = strmake(name)) == NULL ||
	   (probe = realloc(ProcNames, (size_t)(entries + 2) * sizeof(NBIND))) == NULL) {
		if(nname)
			free(nname);
		return((NBIND*)NULL);
	}
	if(ProcNames != probe) {
		/* It has been moved -- relocate entries in keytab
		 * otherwise some of the k_ubp entries will point at rubbish.
		 * ProcNames still points at where the table was.
		 */
		NBIND*	kend = ProcNames + entries;	// Last entry in old table
		KEYTAB*	ktp;
		for(ktp = keytab; ktp->k_u.k_ubp != NULL; ++ktp)
			if(ktp->k_u.k_ubp >= ProcNames && ktp->k_u.k_ubp <= kend)
				ktp->k_u.k_ubp = probe + (ktp->k_u.k_ubp - ProcNames);

		/* Adding new entries to ProcNames is not the cleverest ...
		 * but it doesn't happen very often, not worth the effort/code-size
		 * to do it better.
		 */

		ProcNames = probe;	// Note new location
	}

	// Find the entry lexically greater than name, or end of the array
	while(probe->n_name != NULL && strcmp(name, probe->n_name) > 0)
		probe++;

	// Move everything down -- which might just be the last NUL entry
	memmove(probe + 1, probe, (size_t)(entries + 1 - (probe - ProcNames)) * sizeof(*probe));

	// We'll use the old last entry for the new entry
	memset(probe, 0, sizeof(NBIND));	// Clear just-added/new-NULL entry

	// Fill in what we know, the rest will be zeros as was last NULL
	probe->n_name = nname;

	return(probe);
}

/* Bind an entry in a name binding table to c.
 *
 * bindp	The binding table entry
 * c		The keycode to bind it to
 *
 * This knows if it is a procedure or command as n_func is NULL for procedures.
 *
 * The name binding table could either be ProcNames or CmdNames.
 *
 * Return:
 *	TRUE  - bound
 *	FALSE - binding table full.
 */
int	bindNbToKey(NBIND* bindp, KeyCode c)
{
	BYTE	outseq[80];	/* output buffer for keystroke sequence */
	KEYTAB*	ktp;		/* pointer into the command table */

	/* Search the table to see if it exists
	 * or we hit the end of the table
	 */
	for(ktp = keytab; ktp->k_code != c && ktp->k_u.k_ubp != NULL; ++ktp) {
		/* if we run out of binding room, bitch */
		if(ktp >= keytab + NBINDS - 1) {
			mlwrite("%EBinding table FULL!");
			return(FALSE);
		}
	}

	// Change the key sequence to something we can print
	cmdstr(c, &outseq[0]);

	// Tell the user what we are about to do:
	mlwrite("%s %s bound to key: %s", (bindp->n_func ? "command" : "procedure"), bindp->n_name, outseq);

	hashKeyRebuild = TRUE;	// Changed it, so rebuild it

	// Set new/replaced entry:
	ktp->k_u.k_ubp = bindp;
	ktp->k_code = c;

	return(TRUE);
}

/* Bind a procedure (macro) to a key.
 *
 * The procedure does not need to exist (yet). When the bound key
 * is later used the procedure must exist, or be auto loaded.
 *
 * A new entry in ProcNames will be created if needs be.
 * Copy any comment in b_fnameClean - if it exists
 *
 * If the count is set and > 0 : the existance of the procedure will
 * be checked for. Auto loading will be tried.
 *
 * Default count is unset.
 *
 * Command:  bind-procedure-to-key
 * Bound to: M-K
 *
 * Returns:
 *	TRUE  success
 *	FALSE
 *	   Key got is not a command key
 *	   Binding table is full or out of memory - cannot extend ProcNames.
 */
int	bindProcToKey(int f, Cnt n)
{
	KeyCode	c;	/* command key to bind */
	NBIND*	bindp;
	int	s;

	if((s = mlprompt("Name of procedure to bind: ", MLXBUF | MLU8LEN, bmlBufHistp)) != TRUE)
		return(s);

	/* get the command sequence to bind */
	c = getckey(FALSE);

	if(c == -1)
		return(FALSE);

	if((bindp = procFind(ltext(bmlBufHistp->b_dotp), TRUE)) == (NBIND*)NULL) {
		mlwrite(MsgOutOfMemoryAlloc, "ProcNames");
		return(FALSE);
	}

	return(bindNbToKey(bindp, c));
}

/* bindtokey:	add a new key to the key binding table
 *
 * command arguments [IGNORED]
 *
 * Default count is unset.
 *
 * Command:  bind-to-key
 * Bound to: M-k
 *
 * Returns:
 *	TRUE  success
 *	FALSE
 *	   Key got is not a command key
 *	   Binding table is full
 */
/* ARGSUSED */
int	bindtokey(int f, Cnt n)
{
	KeyCode		c;	/* command key to bind */
				/* ptr to the requested function to bind to */
#ifdef	BIND_TO_PREFIX_KEY
	KEYTAB*		ktp;	/* pointer into the command table */
	//UniChar	(*kfunc)(int* flagset, int* count, UniChar ch, int* displayit, int* reprocess);
	int	(*kfunc)();	/* ptr to the requested function to bind to */
	register int found;	/* matched command flag */
#endif
	NBIND*	bindp;

	bindp = getname(": bind-to-key enter command: ");
	if(bindp == NULL) {
		mlwrite("%ENo such function '%l'", bmlEmCmdHistp->b_dotp);
		return(FALSE);
	}

	mlwrite("Bind command %l to key: ", bmlEmCmdHistp->b_dotp);

	/* get the command sequence to bind */
#ifdef	BIND_TO_PREFIX_KEY
	kfunc = bindp->n_func;
	c = getckey((kfunc == metaPrefix) || (kfunc == ctlx) ||
		    (kfunc == ctlu) || (kfunc == ctrlg));
#else
	c = getckey(FALSE);
#endif
	if(c == -1)
		return(FALSE);

#ifdef	BIND_TO_PREFIX_KEY
	// Don't have that
	/* If the function is a prefix key we need to do some more work.
	 * Why anyone would want to do this I don't know.
	 */
	if(kfunc == metaPrefix || kfunc == ctlx ||
	   kfunc == ctlu || kfunc == ctrlg) {

		found = FALSE;

		/* Find the function, & reset the mapping to that.
		 * This only allows one binding for a prefix key.
		 * Fair enough I suppose.
		 * NB: Prefix table isn't linked to name table.
		 */
		for(ktp = preftab; ktp->k_u.k_ufp != NULL; ktp++)
			if(ktp->k_u.k_ufp == kfunc) {
				ktp->k_code = c;
				found = TRUE;
				break;
			}

		/* Didn't find it in the prefix table, have a look
		 * in the general binding table, and unbind it.
		 */
		if(found == FALSE)
			for(ktp = keytab; ktp->k_u.k_ubp != NULL; ++ktp)
				if(ktp->k_u.k_ubp->n_func == kfunc)
					unbindchar(ktp->k_code);

		/* reset the appropriate global prefix variable */
		if(kfunc == ctrlg)
			abortc = c;
	}
#endif

	return(bindNbToKey(bindp, c));
}

/* unbindkey:	delete a key from the key binding table
 *
 * The count is ignored.
 *
 * Command:  unbind-key
 * Bound to: M-C-k
 *
 * Returns:
 *	TRUE  success
 *	FALSE not a valid key
 *	      key not bound
 *
 * ARGSUSED
 */
int	unbindkey(int f, Cnt n)
{
	KeyCode	c;		/* command key to unbind */
	BYTE outseq[80];	/* output buffer for keystroke sequence */

	/* prompt the user to type in a key to unbind */
	mlwrite(": unbind-key ");

	/* get the command sequence to unbind */
	c = getckey(FALSE);		/* get a command sequence */

	if(c == -1)
		return(FALSE);

	/* change it to something we can print as well */
	cmdstr(c, &outseq[0]);

        /* if it isn't bound, bitch */
	if(unbindchar(c) == FALSE) {
		mlwrite("%EKey %s not bound", outseq);
		return(FALSE);
	}

	mlwrite("%IKey %s unbound", outseq);

	return(TRUE);
}


/* Unbind a key.
 *
 * c:	command key to unbind
 *
 * Return:
 *	TRUE  key unbound
 *	FALSE the key was not bound
 */
int	unbindchar(KeyCode c)
{
	hashNx		index;
	register KEYTAB *ktp;	/* pointer into the command table */
	register KEYTAB *sktp;	/* saved pointer into the command table */

#if 1
	if((index = hashLookup(&hashKeyBindings, &c)) == HASH_MISSING)
		return(FALSE);

	ktp = keytab + index;
#else
	/* search the table to see if the key exists */
	for(ktp = keytab; ; ktp++) {
	        /* if it isn't bound, bitch */
		if(ktp->k_u.k_ubp == NULL)
			return(FALSE);

		if(ktp->k_code == c)
			break;
	}
#endif

	/* save the pointer and scan to the end of the table */
	sktp = ktp;
	while(ktp->k_u.k_ubp != NULL)
		++ktp;

	/* If that wasn't the last one in the table, copy the last one
	 * over what we are unbinding.
	 */
	if(ktp != sktp) {
		--ktp;		/* backup to the last legit entry */

		/* copy the last entry to the current one */
		sktp->k_code = ktp->k_code;
		sktp->k_u.k_ubp = ktp->k_u.k_ubp;
	}

	/* null out the last one */
	ktp->k_code = 0;
	ktp->k_u.k_ubp = NULL;
	hashKeyRebuild = TRUE;

	return(TRUE);
}

/* Describe bindings.
 *
 * bring up a fake buffer and list the key bindings
 * into it with view mode
 *
 * The count is ignored.
 *
 * Command:  describe-bindings
 * Bound to: unbound
 *
 * Return: TODO
 */
int	desbind(int f, Cnt n)
#if	APROP
{
	return(buildlist(""));
}

/* Apropos (List functions that match a substring).
 *
 * The count is ignored.
 *
 * Command:  apropos
 * Bound to: M-a
 *
 * Return: TODO
 *
 * Return TRUE  if we find it.
 *	  FALSE cannot find/create/clear [Help] buffer
 *		split window
 *		out of memory
 *	  ABORT - C-g
 */
int	apro(int f, Cnt n)
{
	int status;		/* status return */

	status = mlprompt("Apropos [keyword help]. Search string (* for all): ",
			MLNOSPACE | MLEMACSCOMMAND, bmlHelpSearch);
	if(status != TRUE)
		return(status);

	if(lgetc(bmlHelpSearch->b_dotp, 0) == '*' && lused(bmlHelpSearch->b_dotp) == 1)
		lsetused(bmlHelpSearch->b_dotp, 0);

	return(buildlist(ltext(bmlHelpSearch->b_dotp)));
}


/* Build a binding list (limited or full)
 *
 * mstring	match string if a partial list, else the empty string
 *
 * Return TRUE  on success
 *	  FALSE cannot find/create/clear [Help] buffer
 *		split window
 *		out of memory
 */
int	buildlist(const char* mstring)
#endif
{
	KEYTAB*	ktp;		/* pointer into the command table */
	NBIND*	nptr;		/* pointer into the name binding table */
	BUFFER*	bp;		/* buffer to put binding list into */
	size_t	cpos;		/* current position to use in outseq */
	int	firstBind;	/* If true then first binding for a command */
	size_t	MaxNameLen;	// How long for names
	BYTE	outseq[300];	/* output buffer for keystroke sequence */
	char**	hp;		/* Probe into header */
	char*	header[] = {	/* Info at the top of the buffer */
		"Notation for input keys includes:",
		"C-a     - Control a. Press and hold Control, press a, release both",
		"M-a     - Meta a. Press and release ESC and then a. Press and hold Alt press a might also work",
		"S-s     - Shift and s. For special keys, eg S-home is Shift and Home",
//		"A-s     - Alt and s. Press and release ESC and then s will also work",
		"M-S-s   - Meta and Shift s",
		"C-x a   - C-x and then a",
//		"C-x M-a - C-x and then M-a",		// Doesn't happen
		"FN10    - Function key 10",
		"SPC     - Space",
		"CIRC    - Circumflex (^)",
		"Move the cursor on to a command name, press enter to see help about it",
		"",
		NULL
	};

	// Find length of longest command name
	for(MaxNameLen = 0, nptr = CmdNames; nptr->n_name != NULL; ++nptr) {
		size_t	len = strlen(nptr->n_name);
		if(len > MaxNameLen)
			MaxNameLen = len;
	}

	// Look down list of procedures - see if any have longer name:
	for(nptr = ProcNames; nptr->n_name != NULL; ++nptr) {
		size_t	len = strlen(nptr->n_name);
		if(len > MaxNameLen)
			MaxNameLen = len;
	}

	// Minimum length
	if(MaxNameLen < BINDING_LIST_NAME_LEN)
		MaxNameLen = BINDING_LIST_NAME_LEN;

	MaxNameLen += 2;	// Provide some separation space


	/* Get a buffer for it. Name for the buffer is [Help] (which it is
	 * part of - to not create another window if help on a command is asked for).
	 * The 'file name' is set to show that it the binding list.
	 * The 'file name' is important, it is tested in helpLink()
	 */
	if((bp = bfindCheck(HelpBname, BFNCREATE|BFNACTIVE)) == (BUFFER*)NULL)
		return(FALSE);

	setVisibleFname(bp, AprFname);
	if(bclear(bp) == FALSE) {
		mlwrite("%ECannot display binding list");
		return(FALSE);
	}
	bp->b_flag |= BHELP | BFINVS;	// Get help bindings, make invisible

	/* let us know this is in progress */
	mlwrite("%IBuilding binding list");

	// If the buffer is visible, don't change anything

	if( ! bp->b_nwnd) {
		// If the buffer is not visible, connect to a window

		/* Split the current window to make room for the binding list */
		if(splitwind(FALSE, 1) == FALSE)
			return(FALSE);

		DisconnectBuffer(curwp);	/* disconnect the current buffer */

		/* connect the current window to this buffer */
		curbp = bp;	/* make this buffer current in current window */

		ConnectBuffer(bp, curwp, 0);
	}
	bp->b_mode = 0;		/* no modes active in binding list */

	// Output buffer header:
	for(hp = header; *hp; hp++)
		if(addline(*hp, bp, 1) != TRUE)
			return(FALSE);

	/* Build the contents of this window, inserting it line by line.
	 * Just to make things interesting we are working down 2 tables:
	 * CmdNames	Emacs commands - static
	 * ProcNames	Procedures - dynamically built
	 *
	 * The entry to work on (bindp) is the one that sorts first.
	 */
	NBIND*	procp = ProcNames;
		nptr  = CmdNames;

	// sindex() needs TEXT* these days, construct enough for it
	TEXT	mtext;
	mtext.t_text = (BYTE*)mstring;
	mtext.t_used = (LineLn)strlen(mstring);

	while(1) {
		NBIND*	bindp;		// The NBIND that we work on

		// Work out which one to look at:
		if(nptr->n_name == NULL) {
			if(procp->n_name == NULL)
				break;
			bindp = procp++;
		} else {
			if(procp->n_name == NULL) {
				bindp = nptr++;
			} else {
				bindp = strcmp(procp->n_name, nptr->n_name) > 0 ?
					nptr++ : procp++;
			}
		}

#if	APROP
		/* if we are executing an apropos command..... */
		if(mstring[0] != '\0') {
                    /* and current command name or description doesn't include the search string
		     * Beware description might be missing
		     * A bit of work is needed to construct a TEXT* for sindex()
		     */
		    TEXT	ntext;
		    ntext.t_text = (BYTE*)bindp->n_name;
		    ntext.t_used = (LineLn)strlen(bindp->n_name);
		    if(sindex(&ntext, &mtext) == 0) {
			if(bindp->n_desc == NULL) {
				continue;
			} else {
				ntext.t_text = (BYTE*)bindp->n_desc;
				ntext.t_used = (LineLn)strlen(bindp->n_desc);
				if(sindex(&ntext, &mtext) == 0)
					continue;
			}
		    }
		}
#endif

		/* This will end up pointing to the synonym - if there is one.
		 * This only works if there is no more than 1 synonym.
		 * Remember that at startup k_ubp it set to point to one entry in CmdNames
		 */
		NBIND*	synBind = bindp;

		// If this has synonyms - find & show them:
		if(bindp->n_flag & N_SYNONYM) {
			/* This search cannot be done by hash. We would be looking
			 * for a non-unique key - hashing needs unique keys.
			 */
			NBIND*	nbp;
			for(nbp = CmdNames; nbp->n_name != NULL; nbp++) {
				// Same functions ?
				if(nbp->n_func != bindp->n_func || nbp == bindp)
					continue;

				synBind = nbp;
			}
		}

		/* add in the command name */
		strcpy(outseq, bindp->n_name);
		cpos = strlen(outseq);

		firstBind = 1;

		/* search down any keys bound to this */
		for(ktp = keytab; ktp->k_u.k_ubp != NULL; ++ktp) {
			// Show binding if this or the synonym
			if(ktp->k_u.k_ubp == bindp || ktp->k_u.k_ubp == synBind) {
				/* padd out some spaces */
				while(cpos < MaxNameLen)
					outseq[cpos++] = ' ';

				/* add in the command sequence */
				cmdstr(ktp->k_code, &outseq[cpos]);

				if(firstBind) {
					// First time append the description

					// Length in terms of (possibly multibyte) chars:
					size_t	chrs = utf8_strlen(outseq);

					// Length in terms of bytes:
					cpos = strlen(outseq);

					/* Difference in the 2 lengths is how many more
					 * we must pad to make the descriptions line up.
					 */
					chrs -= cpos;
					while(cpos + chrs < MaxNameLen + BINDING_LIST_KSEQ_LEN)
						outseq[cpos++] = ' ';
					
					// If no description (macro) NUL terminate
					strcpy(outseq + cpos, (bindp->n_desc != NULL ? bindp->n_desc : ""));
					firstBind = 0;
				}

				/* and add it as a line into the buffer */
				if(addline(outseq, bp, 1) != TRUE)
					return(FALSE);

				cpos = 0;	/* and clear the line */
			}
		}

		/* if no key was bound, we need to dump it anyway */
		if(cpos > 0) {
			// Add the description
			while(cpos < MaxNameLen + BINDING_LIST_KSEQ_LEN)
				outseq[cpos++] = ' ';
			if(bindp->n_desc != NULL)
				strcpy(outseq + cpos, bindp->n_desc);

			if(addline(outseq, bp, 1) != TRUE)
				return(FALSE);
		}

		// If synonym: say so
		if(synBind != bindp) {
			sprintf(outseq, "  Synonym %s", synBind->n_name);
			if(addline(outseq, bp, 1) != TRUE)
				return(FALSE);
		}
	}

	bp->b_mode |= MDVIEW;		/* put this buffer view mode */
	bp->b_flag &= ~BFCHG;		/* don't flag this as a change */
	utf8_bufmodecheck(bp);		/* Set UTF-8 encoding */
	bp->b_dotp = lforw(bp->b_linep);/* Set "." to first line */

	ReconnectWindowsToBuffer(bp);	// Ensure that LINE pointers, etc, are valid

	upmode(NULL, WFMODE);

	mlerase();

	// Start the help stack - this is an entry point, so zap anything there
	bclear(bmlHelpStack);
	helpPushStack('a', mstring);

	return(TRUE);
}

/* Put the bindings for command cmd into binds.
 *
 * If there is more than one binding, separate the bindings by 3 spaces.
 *
 * Put all the function keys first, then other bindings.
 * To do this we pass over keytab twice.
 * This has the effect that if only one binding is shown it will
 * be the function key (if any).
 *
 * Here: function key includes things like arrow, home, etc.
 * The KeyCode for these will have the SPEC bit set.
 *
 * Return TRUE   - worked, look in binds. Empty string means not bound
 *	  FALSE  - memory error or something
 *	  FAILED - unknown command
 */
int	getCmdBindings(const char* cmd, TEXT* binds)
{
	NBIND*	nptr;		// Binding table entry
	KEYTAB*	ktp;		// Pointer into the command table
	char*	bindSep = "";	// Put before bindings in bind
	int	pass;		// Which pass over keytab

	tsetempty(binds);

	if((nptr = fncmatch(cmd)) == (NBIND*)NULL)
		return(FAILED);

	for(pass = 0; pass < 2; pass++)
		for(ktp = keytab; ktp->k_u.k_ubp != NULL; ++ktp) {
			if(ktp->k_u.k_ubp == nptr) {
				BYTE	bind[30];

				// First pass: just the SPEC keys
				// Next  pass: not  the SPEC keys
				if(pass == 0 ? (ktp->k_code & SPEC) == 0 : (ktp->k_code & SPEC) != 0)
					continue;

				cmdstr(ktp->k_code, bind);
				tcats(binds, bindSep);
				tcats(binds, bind);
				bindSep = "   ";
			}
		}


	return(TRUE);
}

#if	0
// No longer used, sindex() used instead, as it is case insensitive.
#if	APROP
/* This is essentially strstr()
 *
 * Does source include sub?
 *
 * source	string to search in
 * sub		substring to look for
 */
int	strinc(BYTE* source, BYTE* sub)
{
	BYTE*	sp;	/* ptr into source */
	BYTE*	nxtsp;	/* next ptr into source */
	BYTE*	tp;	/* ptr into substring */

	/* for each character in the source string */
	sp = source;
	while(*sp) {
		tp = sub;
		nxtsp = sp;

		/* is the substring here? */
		while(*tp) {
			if(*nxtsp++ != *tp)
				break;
			else
				tp++;
		}

		/* yes, return a success */
		if(*tp == 0)
			return(TRUE);

		/* no, onward */
		sp++;
	}
	return(FALSE);
}
#endif
#endif

/* Print to stdout a list of all Emacs commands from the CmdNames table.
 *
 * If bindings show the command name & its bindings.
 */
void	ListCommands(int bindings)
{
	NBIND*	np;
	int	len = 0;
	int	i;

	// Find the longest so that we can align the command descriptions:
	for(np = CmdNames; np->n_name; np++)
		if((i = (int)strlen(np->n_name)) > len)
			len = i;
		
	for(np = CmdNames; np->n_name; np++) {
		char*	desc = np->n_desc;

		if(desc == NULL)	// Should not happen
			desc = "";

		// Show bindings instead of command description ?
		if(bindings) {
			getCmdBindings(np->n_name, playText);
			desc = ttext(playText);
		}

		printf("%-*s  %s\n", len, np->n_name, desc);
	}
}

/* Get a command key sequence from the keyboard.
 *
 * mflag	going for a meta sequence? Just get the key
 *		else have OR'd in CNTRL, META, CTLX, SPEC, INTL.
 *
 * If, in a macro, an INT is the value: return this - whatever nonsense it makes.
 *
 * This can return -1 if there was some failure in interpretting the key.
 */
KeyCode	getckey(int mflag)
{
	KeyCode		c;		/* character fetched */
	VAL*		token;
	int		dummy1 = 0;	/* dummy var # 1 */
	Cnt		dummy2 = 0;	/* dummy var # 2 */

	/* check to see if we are executing a command line */
	if(clexec) {
		token = valGet(V_STR);
		macarg(&token);	/* get the next token */

		// Map INTs to keycode:
		if(vtype(token) == V_INT)
			c = (KeyCode)vint(token);
		else
			c = stock(vtext(valPrint(token)));

		valRelease(token);

		return(c);
	}

	/* or the normal way */
	CursorInMsg = 1;

	// Get raw key - don't want refresh key exfiltrated
	c = tgetc(MGC_MAP | MGC_NO_REFRESH, 0);

	if( ! mflag)
		c = getcmd(c, &dummy1, &dummy2);

	return(c);
}

/* Execute the startup file
 *
 * sfname	name of startup file (NULL if default search is wanted)
 *
 * This will only be called once -guaranteed.
 * The startup file(s) run are malloced & stored in StartupFiles:
 * - fineFile() returns a pointer to a dynamic area, so malloc/copy
 * - paths displayed with status info - makelist()
 *
 * Return:
 *	TRUE  all found files executed without error
 *	FALSE execution of a file failed
 */
int	startup(const BYTE* sfname)
{
	char**	np = StartupFiles;
	int	res = TRUE;

	/* look up the startup file */
	if(sfname != NULL) {
		if(findFile(sfname, fT_Macro))
			*np++ = strmake_t(ttext(foundFiles));
	} else {
		if(findFile("", fT_StartGlob))
			*np++ = strmake_t(ttext(foundFiles));
		if(findFile("", fT_StartUser))
			*np++ = strmake_t(ttext(foundFiles));
	}

	*np = NULL;

	// Execute the files that we found.
	// Note if one fails
	for(np = StartupFiles; *np; np++)
		if( !dofile(*np))
			res = FALSE;

	return(res);
}

#if	MSDOS && (TURBO || LATTICE || AZTEC || MSC)
/* Now only used in spawn.c execprog() and not on Unix systems.
 *
 * Look up the existance of a file along the normal or PATH
 * environment variable. Look first in the HOME directory if
 * asked and possible
 *
 * fname		base file name to search for
 * hflag		Look in the HOME environment variable first?
 *
 * Return a file name, on error return NULL.
 * playText3 is used.
 */
BYTE*	flook(BYTE* fname, int hflag)
{
	BYTE*	path;	/* environmental PATH variable */
	int	i;	/* index */

	// If there is a '/' in the name, see if it gives a path:
	if(strchr(fname, '/') && ffropen(fname) == FIOSUC) {
		ffclose();
		return(fname);
	}

#if	DOTINPATH
	/* always try the current directory first
	 * However it is a security risk in that it will find files
	 * that depend on PWD and might be malicious. Thus configurable.
	 */
	if(ffropen(fname) == FIOSUC) {
		ffclose();
		return(fname);
	}
#endif

#if	ENVFUNC

	if(hflag && home[0] != '\0') {
		/* build home dir file spec */
		printToText(&ltxt(playText3), "%s/%s", home, fname);

		/* and try it out */
		if(ffropen(ttext(playText3)) == FIOSUC) {
			ffclose();
			return(ttext(playText3));
		}
	}
#endif

#if	ENVFUNC
	/* get the PATH variable */
	path = getenv("PATH");
	if(path != NULL)
		while(*path) {

			/* build next possible file spec */
			tsetused(playText3, 0);
#if	ST520 && MWC
			while(*path && (*path != PATHCHR) && (*path != ','))
#else
			while(*path && (*path != PATHCHR))
#endif
				tcatc(playText3, *path++);	// *sp++ = *path++;

			/* add a terminating dir separator if we need it */
			if(tused(playText3) != 0)
#if	ST520
				tcatc(playText3, '\\');
#else
				tcatc(playText3, '/');
#endif

			tcats(playText3, fname);

			/* and try it out */
			if(ffropen(ttext(playText3)) == FIOSUC) {
				ffclose();
				return(ttext(playText3));
			}

#if	ST520 && MWC
			if((*path == PATHCHR) || (*path == ','))
#else
			if(*path == PATHCHR)
#endif
				++path;
		}
#endif

	/* look it up via the old table method */
	for(i = PNAME_DSTART; i < (int)NPNAMES; i++) {
		if( !pathname[i][0])	// Ignore emtpy
			continue;

		/* Append a '/' to the directory path - avoid there not being one,
		 * this has happened with a RPM build where the config was not quite right.
		 * Will get '//' on occasion - harmless.
		 */
		printToText(playText3, "%s/%s", pathname[i], fname);

		/* and try it out */
		if(ffropen(ttext(playText3)) == FIOSUC) {
			ffclose();
			return(ttext(playText3));
		}
	}

	return(NULL);	/* no such luck */
}
#endif	// MSDOS && (TURBO || LATTICE || AZTEC || MSC)

/* Print info about command files to buffer.
 *
 * This is quite important, otherwise it is hard for the user to work
 * out where these files should be put.
 * 
 * Return
 *	TRUE if all OK
 *	FALSE - out of memory
 */
int	WriteCmdFilesToBuffer(BUFFER* blp)
{
	int	ret = 0;

	ret |= WriteSearchPathsToBuffer(blp);

	// Startup files that were found:
	if(StartupFiles[0]) {
		char**	sf;

		ret |= !addline("", blp, 1);
		ret |= !addline("Startup file(s) run:", blp, 1);

		for(sf = StartupFiles; *sf; sf++) {
			if( !**sf)	// Ignore empty
				continue;

			printToText(playText, "    %s", *sf);
			ret |= !addLINE(playText, blp, 0);
		}
	}

	return( !ret);
}

/* change a key command to a string we can print out
 *
 * c	sequence to translate
 * seq	destination string for sequence
 *
 * A break with prior notation happened in 2017.
 * ^a used to be the preferred notation for control, this changed to C-a
 * Both are still recognised on input (see stock()) but C- is now shown on output.
 *
 * Also the CTLX prefix is now shown with a space after it.
 * Do this with any other prefixes that are invented in the future.
 *
 * Commented out idea of showing things like M-left-arrow as A-left-arrow.
 *
 * How others show control:
 * https://en.wikipedia.org/wiki/Control_key
 */
void	cmdstr(KeyCode c, BYTE* seq)
{
	char*	ptr;
	int	is_ctrl = 0;
	int	is_pref = 0;		// Has a M- or C- prefix when displayed
//	BYTE*	oseq = seq;

	/* Say if meta.
	 * Do this before the others as it makes more sense to the user if they
	 * have pressed ESC to get the meta modification.
	 */
	if(c & META) {
		*seq++ = 'M';
		*seq++ = '-';
		is_pref = 1;
	}

	/* Say if shifted */
	if(c & SHIFT) {
		*seq++ = 'S';
		*seq++ = '-';
	}

	/* C-x is a prefix
	 * Put a space after it to make it clearer to understand
	 */
	if(c & CTLX) {
		*seq++ = 'C';
		*seq++ = '-';
#if	MODFOLD == 0
		*seq++ = 'x';
#else
		*seq++ = 'X';
#endif
		*seq++ = ' ';
		is_pref = 1;
	}

	/* Apply control sequence if needed.
	 */
	if(((c & CNTRL) | ((c & ~(META | CTLX | 0x1f)) == 0))) {
		*seq++ = 'C';
		*seq++ = '-';
		is_ctrl = 1;
	}

	/* apply SPEC sequence if needed */
	if((c & SPEC) && (ptr = TTkeyname(c & ~(META | CTLX)))) {
//		if(is_meta)
//			*oseq = 'A';
	} else {
		BindSpecChrs*	bsc;

		c = (KeyCode)((KeyCode_u)c & ~K_CODE_MASK);	/* strip the prefixes */

		ptr = NULL;		// This not being NULL is how we know if we found a special

		/* Recognise specials like SPC
		 * This doesn't work properly for ESC since it is held as '['|META
		 * Fix when redo bindings.
		 */
		if(is_ctrl && c == '\0')
			c = ' ';
		for(bsc = BindSpecials; bsc->bs_rep != NULL; bsc++) {
			if(c == bsc->bs_char && (!is_pref || (bsc->bs_flag & BS_prefCTL) == 0)) {
				ptr = bsc->bs_rep;
				if(is_ctrl && (bsc->bs_flag & BS_ctlOK) == 0)
					seq -= 2;	// Step back over "C-" Prevent "C-TAB"
			}
		}

		if(ptr == NULL) {
			if(is_ctrl)
				c = ChrFromCtrl(c);
#if	MODFOLD == 0
			// Show control like ^a rather than ^A
			if(is_ctrl && isupper(c))
				c = tolower(c);
#endif

			// Check that the character is valid:
			if( !utf8_isvalid(c))
				ptr = utf8_perror();	// Whoops!
			else
				ptr = utf8_encode(c);

		}
	}

	/* and output the final sequence */
	strcpy(seq, ptr);
}

/* This function looks a key binding up a the binding table.
 *
 * c	   key to find what is bound to it
 * helpOK  If the buffer is in help mode, reverse those bindings as well
 *
 * This never returns NULL, what it returns might not be in the binding
 * table 'CmdNames'.
 */
NBIND*	getbind(KeyCode c, int helpOK)
{
	if(helpOK && (curbp->b_flag & BHELP)) {
		/* Some keys are interpretted differently.
		 * Temp fix until buffer specific key bindings has been done.
		 */
		switch(c) {
		  case RETCHAR:	// Follow a help link
			return(HelpBufferBinds + HELP_BINDS_LINK_FOLLOW);
		  case CharDelete:
		  case '\b':
			return(HelpBufferBinds + HELP_BINDS_PREVIOUS_PAGE);
		  case CharTab:
			return(HelpBufferBinds + HELP_BINDS_NEXT_PLACE);
		  case UTC_BAKTAB:
			return(HelpBufferBinds + HELP_BINDS_PREVIOUS_PLACE);
		}

		// Drop through to normal use
	}

	/* If bindings have been changed - rebuild the hash table
	 *
	 * Maintain stats on how full the binding table is.
	 */
	if(hashKeyRebuild) {
		KEYTAB*	kp;
		int	nrbs = 0;

		hashEmpty(&hashKeyBindings);
		hashInit(&hashKeyBindings);

		ST_keyBindings = hashCmdNames.hs_used;

		hashKeyRebuild = FALSE;

		// Rebuild the table looked at by tgetc() to decide if to do a screen refresh
		for(kp = keytab; kp->k_u.k_ufp != NULL &&
			nrbs < (int)(sizeof(RefreshBinds) / sizeof(RefreshBinds[0])); kp++) {
			if(kp->k_u.k_ubp->n_func == refreshScreen)
				RefreshBinds[nrbs++] = kp->k_code;
		}
		NumRefreshBinds = nrbs;
	}

	hashNx		index;

	if((index = hashLookup(&hashKeyBindings, &c)) != HASH_MISSING)
		return(keytab[index].k_u.k_ubp);

	/* Self inserting */
        if((c >= ' ' && c <= CharMax))
		return(&DummySelfInsert);


	/* No such binding */
	return(&DummyUnboundKey);
}

/* getfname:	This function takes a ptr to function and gets the name
 *		associated with it
 *
 * func		ptr to the requested function to bind to
 *
 * This is not, currently, used.
 */
BYTE*	getfname(int (*const func)(int f, Cnt n))
{
	NBIND*	nptr;	/* pointer into the name binding table */

	/* skim through the table, looking for a match */
	for(nptr = CmdNames; nptr->n_name != NULL; nptr++)
		if(nptr->n_func == func)
			return(nptr->n_name);

	return((char*)NULL);
}

/* Find Emacs command cname in the CmdNames table.
 *
 * cmdName	name to attempt to match
 *
 * Return any match or NULL if none
 */
NBIND*	fncmatch(const BYTE* cmdName)
{
	hashNx		index;

	if((index = hashLookup(&hashCmdNames, &cmdName)) == HASH_MISSING)
		return((NBIND*) NULL);

	return(CmdNames + index);
}

/* stock:	String key name TO Command Key
 *
 * keyname	name of key to translate to Command key form
 *
 * Eg given	'left-arrow'	return	UTC_LEFT
 * 		'M-x'		return	META | 'x'
 * 	-1 on error
 *
 * So: This understands the prefixes:
 * * M-		META
 * * A-		META	only on terminal specials
 * * S-		SHIFT	only on terminal specials
 * * ^		CNTRL
 * * C-		CNTRL
 *
 * The A- prefix exists since it might be more natural for A-delete
 *
 * If CTLX is seen then another character must be given, eg: ^xa ^x^x ie start again.
 * This may be given as:
 * ^x or ^X or C-x	CTLX
 *
 * At some stage the last point will be generalised so that an arbitrary long
 * sequence may be specified, eg: M-aSPC  M-a^xq All but the last would have
 * to be prefix points - ie not defined as a terminal somewhere else, thus:
 * M-a and M-aq at the same time would be illegal.
 *
 * The ordering is important. Partly because of ambiguity, but also
 * reflects that M- can either be ALT modifier or an ESC prefix.
 * At some stage the parsing might be longer & more order dependent, eg: M-AM-q
 *
 * Order is: M- S- C-
 * Cannot have reptitions in the same sequence. M- and A- count as the same as do ^ and C-
 *
 * The following are recognised specially:
 * SPC	' '
 * DEL	7Fx (hex 0x7F)
 * ESC	1Bx (hex 0x1B)
 *
 * After the prefixes may be exactly one key, eg 'a' 'A' 'left-arrow'.
 *
 * A terminal special is something like left-arrow.
 *
 * BUG: this gets it wrong when there is a ^x
 * * Type: ^x S-M-DEL which is shown as S-M-^xDEL
 * * Type: M-^x and it shows correctly as M-^x
 * Fix this when get arbitrary sequences right.
 * Cannot have M-^xS-end
 */
KeyCode	stock(const BYTE* keyname)
{
	KeyCode		c;		/* key sequence to return */
	KeyCode		fn;
	char*		NeedSpec = NULL;
	BindSpecChrs*	bsc;
	UniChar		key;	// Something at *keyname

	/* parse it up */
	c = 0;

	/* first, the META prefix */
	if(*keyname == 'M' && *(keyname+1) == '-') {
		c = META;
		keyname += 2;
	}

	if(*keyname == 'A' && *(keyname+1) == '-') {
		if(c & META) {		// Can't have M-A-
			mlerr("Bad command key string, cannot have M-A-", NULL);
			return(-1);
		}
		c = META;
		NeedSpec = "A-";
		keyname += 2;
	}

	if(*keyname == 'S' && *(keyname+1) == '-') {
		c |= SHIFT;
		NeedSpec = "S-";
		keyname += 2;
	}

	/* control-x as well... as ^x or ^X */
	if(*keyname == '^' && (*(keyname+1) == 'X' || *(keyname+1) == 'x')) {
		c |= CTLX;
		keyname += 2;
	} else {
		/* Control-x as C-x
		 * Simpler this way to make a separate check.
		 * Deal with a non prefix control X below.
		 */
		if(*keyname == 'C' && *(keyname+1) == '-' && (*(keyname+2) == 'X' || *(keyname+2) == 'x')) {
			c |= CTLX;
			keyname += 3;
		}
	}

	/* A control char?
	 * Recognise this as either: ^a or C-a
	 */
	if((*keyname == '^' && *(keyname+1) != 0) ||
	   (*keyname == 'C' && *(keyname+1) == '-' && *(keyname+2) != 0)) {
		c |= CNTRL;
		if(*keyname == 'C')
			++keyname;
		++keyname;
	}

	/* Terminal special key */
	if((fn = TTspecKey(keyname)) != -1)
		return(c | fn);

	if(NeedSpec) {
		// Complain that a special was not given
		char	mb[100];
		sprintf(mb, "Need terminal special (eg left-arrow) in bind string after %s", NeedSpec);
		mlerr(mb, NULL);
		return(-1);
	}

	// Recognise specials like ESC:
	for(bsc = BindSpecials; bsc->bs_rep != NULL; bsc++) {
		if( !strcmp(keyname, bsc->bs_rep)) {
			if((c & CNTRL) && (bsc->bs_flag & BS_ctlOK) == 0) {
				// Things like C-ESC are not allowed
				char	mb[100];
				sprintf(mb, "Input key sequence C-%s is not allowed", keyname);
				mlerr(mb, NULL);
				return(-1);
			}

			if(c & CNTRL)
				return(ChrToCtrl(bsc->bs_char));

			return(c | bsc->bs_char);
		}
	}

	/* Look at the next character in input. To find out what is meant by
	 * character look at the UTF-8 mode of the buffer.
	 * For now assume that it always is. TODO
	 */
	key = utf8_decode(keyname);

	// Check that there is at most white space after the key. TODO

	/* This code assumes ASCII - unlikely to be anything else.
	 * Unicode is AKA ASCII on steroids.
	 */
	if(key < 32) {
		if(c & CNTRL) {		// Something like ^xC-x^a given
			mlerr("Bad command key string, cannot have ^ or C- on control character", NULL);
			return(-1);
		}
	} else if(c & CNTRL) {
		// This is Control something, but NOT C-arrow or similar
		// Ie ordinary characters
		if(key == '?') {
			// C-? is DEL
			// But this we do not regard as a control key internally
			key = CharDelete;
		} else if(key < CharDelete) {
			/* Printable ASCII, including space, except '?'.
			 * Make it a control character, zap the top 3 bits,
			 * but because we keep them internally as CNTRL | 'A'
			 * OR back in 0x40.
			 * This normalises it so that ^a and ^A both work.
			 */
			key = ChrToCtrl(key);
		} else {
			/* C-DEL and greater (U+80 and more) do not have a defined
			 * control sequence - just complain.
			 */
			mlerr("Control of a character > U+80 is meaningless", NULL);
			return(-1);
		}
	}

#if MODFOLD	// Fold lower to upper
	/* make sure we are not lower case (not with function keys or controls)*/
	if(key >= 'a' && key <= 'z' && !(c & (SPEC | CNTRL)))
		key -= 32;
#endif

	/* the final sequence... */
	c |= key;
	return(c & ~(CNTRL));
}


/* String key name to binding name....
 *
 * skey		name of key to get binding for
 */
BYTE*	transbind(const BYTE* skey)
{
	BYTE*		bindname;
	KeyCode		cmd;

	// Beware: skey could fail to parse to a command:
	bindname = "ERROR";

	if((cmd = stock(skey)) != -1)
		bindname = getbind(cmd, TRUE)->n_name;

	return(bindname);
}

/* This must be called on startup.
 * It copies NBIND pointer into the bindings table.
 * The point is that we the static bindings are done by linking a function
 * name to the keys, we want the keys linked to the function description.
 * Brute force to save some brainpower.
 *
 * Also set up a pointer to the entries for wrapword & nullproc (the latter
 * isn't used).
 */
void	LinkBindings(void)
{
	NBIND*	bp;
	KEYTAB*	kp;

	for(kp = keytab; kp->k_u.k_ufp != NULL; kp++) {
		for(bp = CmdNames; bp->n_name != NULL; bp++) {
			if(bp->n_func == kp->k_u.k_ufp) {
				kp->k_u.k_ubp = bp;
				break;
			}
		}
		if(bp->n_name == NULL)	// Programmer error
			Die("keytab contains entry %#lx missing from CmdNames table\n", (long)kp->k_code);
	}

	for(bp = CmdNames; bp->n_name != NULL; bp++) {
		if(bp->n_func == wrapword)
			BindWrapword = bp;
		if(bp->n_func == nullproc)
			BindNullproc = bp;
	}
}

/* end */
