/*
 * ADDW SCCS: @(#)buffer.c	1.238 10/12/20 22:44:30 - second
 */

/*
 * Buffer management.
 * Some of the functions are internal,
 * and some are actually attached to user
 * keys. Like everyone else, they set hints
 * for the display system.
 */
#include	"estruct.h"
#include	"utf8.h"
#include	"edef.h"

#if	SHOWMODREL
char	buffvers[] = "@(#)buffer.c	1.238 10/12/20 22:44:30";
#endif

/* Offset of the various fields in each line in the buffer list */
#define	GM_ACTIVE	0
#define	GM_CHANGED	1
#define	GM_TRUNC	2
#define	GM_NARROW	3
#define	GM_MODES	5
#define	GM_SIZE		21
#define	GM_LINES	31
#define	GM_MARKS	41	// 2 digits max
#define	GM_BUFFER	44
#if	UNIX
#define	GM_MODE		62
#define	GM_FILE		67
#else
#define	GM_FILE		61
#endif

// Window information layout, modes, etc.
#define	WO_OPTS		2	// Start of window information
#define	WO_SIZE_TO_NAME	21	// Positions from Size to the buffer name
#define	WO_BUFFER_NAME	38	// Where the buffer name starts

/* Attach a buffer to a window.
 *
 * The values of dot and mark come from the buffer.
 *
 * The count is ignored. There is not a default
 *
 * Command:  select-buffer
 * Bound to: C-x b
 *
 * Return:
 *	TRUE  on success
 *	others ?? TODO
 */
int	usebuffer(int f, Cnt n)
{
	register int	s;
	BYTE		obufn[BUFFER_NAME_SIZE];	/* Current buffer name */

	strcpy(obufn, curbp->b_bname);

	if((s = swbufferPrompt(f, n, NULL)))
		strcpy(lbufn, obufn);	// Remember where we came from.

	/* We just created the buffer, check if it should be in some
	 * suffix dependent mode.
	 * Doing it here is good for setting the mode on new buffers, but
	 * not if it is an existing buffer where the mode was deleted.
	 */
	if(curbp->b_numLines == 0)	// New buffer ?
		CheckCMode(curbp);

	return(s);
}


/* Prompt for a buffer, check it and switch to it.
 *
 * If split is not NULL call the function with args: f, n
 * The purpose of split is to make a window in which to
 * display the buffer.
 * This will split top/bottom or left/right
 *
 * bmlBufHistp exists because of this function.
 * I tried to use bmlPromptHistp but it just was corrupted with the prompt
 * and buffer name. Pass buffer history off as a feature!
 *
 * The count is passed to the split function, if any.
 *
 * Return:
 *	TRUE  on success
 *	FALSE xx TODO
 *	      split window
 *	      create the buffer bfindCheck()
 *	      swbuffer()
 *	ABORT C-g
 */
int	swbufferPrompt(int f, Cnt n, int(*split)(int f, Cnt n))
{
	register BUFFER*bp;
	register int	s;
	BYTE		prompt[30 + BUFFER_NAME_SIZE];

	if(lbufn[0] != '\0') {
		sprintf(prompt, "Use buffer %s: ", lbufn);
		s = mlprompt(prompt, MLXBUF | MLU8LEN, bmlBufHistp);

		switch(s) {
		  case TRUE:
			break;
		  case FALSE:
			lcpys(bmlBufHistp->b_dotp, lbufn);
			break;
		  case ABORT:
			return(s);
		}

		/* this is only here to look pretty */
		mlwrite("Use buffer: %l", bmlBufHistp->b_dotp);
	} else
		if((s = mlprompt("Use buffer: ", MLXBUF | MLU8LEN, bmlBufHistp)) != TRUE)
			return(s);

	if((bp = bfindCheck(ltext(bmlBufHistp->b_dotp), BFNCREATE|BFNACTIVE)) == NULL)
		return(FALSE);

	if(split && (*split)(f, n) == FALSE)
		return(FALSE);

	return(swbuffer(bp));
}

/* Find a buffer into another window. This is an alternative to mapping
 * onto splitwindvert().
 *
 * The count is passed to splitwindvert. If it is not set: -1 is assumed
 * A count of:
 *	1 forces "." into the upper window
 *	not 1 forces "." to the lower window
 *
 * Command:  split-and-find-buffer
 * Bound to: C-x 2
 *					Tony Bamford 31/8/86 afb@root.co.uk
 * Return TRUE on success.
 *	others ? TODO
 */
int	owinbuf(int f, Cnt n)
{
	if(f == FALSE) {
		f = TRUE;
		n = -1;
	}

	return(swbufferPrompt(f, n, splitwindvert));
}


/* Switch to the next n'th buffer in the buffer list.
 *
 * Default count is 1. Must not be < 0
 *
 * Command:  next-buffer
 * Bound to: C-x x   FN7
 *
 * Return: ??? TODO
 */
int	nextbuffer(int f, Cnt n)
{
	register BUFFER *bp;	/* Eligible buffer to switch to*/
	register BUFFER *bbp;	/* Eligible buffer to switch to*/

	bp = bbp = curbp;
	while(--n >= 0) {
		/* advance to the next buffer */
		bp = bbp->b_bufp;

		/* cycle through the buffers to find an eligable one */
		while(bp == NULL || bp->b_flag & BFINVS) {
			if(bp == NULL)
				bp = bheadp;
			else
				bp = bp->b_bufp;

			/* don't get caught in an infinite loop! */
			if(bp == bbp)
				return(FALSE);

		}

		bbp = bp;
	}

	return(swbuffer(bp));
}

/* curbp has not been read in - do so.
 *
 * This could be because the buffer is there because the file is
 * on the command line - but has not been looked at yet.
 * The read was deferred when Emacs started.
 *
 * The code was ripped from swbuffer() so that it can be used in
 * zotbuf() when we are killing the last already-read-in buffer.
 *
 * If BLINE_IN_MARKO try to get to the line # in b_toplinenum, -ve means
 * count from bottom of the file. In this situation b_doto specifies the column number.
 *
 * Mark as active then OR in gmode flags, check for CMODE
 *
 * curwp must refer to this buffer.
 *
 * Return TRUE is read in without problem.
 */
int	readInactive(void)
{
	int	retVal;		// Return value

	// Might be
	BufLns	b_topline = curbp->b_toplinenum;
	LineLn	b_doto = curbp->b_doto;

	/* Add the file to the file history buffer - if not there
	 * This is because swbuffer() is how files from the command
	 * line are read in.
	 */
	addHistoryLine(bmlFileHistp, curbp->b_fname, AHL_NO_DUP);

	/* Read it in and activate it.
	 * This might fail, but continue as much as we can.
	 */
	retVal = readin(curbp->b_fname, TRUE, FALSE, TRUE);

	curbp->b_dotp = lforw(curbp->b_linep);
	curbp->b_topLinep = curbp->b_dotp;
	curbp->b_toplinenum = 1;	// Might have had command line -g n
	curbp->b_doto = 0;

	curbp->b_active = TRUE;
	curbp->b_mode |= gmode;		// Only now do we bring in gmode
	CheckCMode(curbp);
	chkmode(curbp);			// Check mode consistency

	// Try to get to the line # in b_topline ?
	if(curbp->b_flag & BLINE_IN_MARKO) {
		GotoLineNumber(b_topline);

		/* Now move to any requested column.
		 * Use goto-column as it deals with edge cases, etc.
		 * But it works on "." in curwp - so need to save/restore that
		 * Beware: BCOL_IN_DOTO is the same as BHELP, but help only happens later
		 */
		if(curbp->b_flag & BCOL_IN_DOTO)
			GotoColNumber(b_doto);

		curbp->b_flag &= ~(BLINE_IN_MARKO | BCOL_IN_DOTO);
	}


#if	UNIX
	if(f_funnyfile)
		curbp->b_flag |= BFWEIRD;/* Remember if its weird*/
#endif

	return(retVal);
}

/* Make buffer BP current, read a file in if needs be.
 * 
 * If BLINE_IN_MARKO try to get to the line # in b_toplinenum, -ve means
 * count from bottom of the file. In this situation b_doto specifies the column number.
 */
int	swbuffer(BUFFER* bp)
{
	int	retVal = TRUE;	// Return value

	DisconnectBuffer(curwp);

	curbp = bp;				// Switch

	// Connect curwp to the buffer, ie set w_dotp, etc:
	ConnectBuffer(bp, curwp, 0);

	if(curbp->b_active != TRUE)		// Buffer not active yet
		retVal = readInactive();

	return(retVal);
}

/* Kill many buffers.
 *
 * The count is ignored.
 *
 * Command:  delete-many-buffers   kill-many-buffers
 * Bound to: unbound
 *
 * Return:
 *	TRUE  many buffers killed
 *	ABORT User replied C-g to a prompt
 */
int	killmany(int f, Cnt n)
{
	BUFFER*	bp;
	int	status;
	int	done;		/* true if we've done something		*/

	for(done = FALSE, bp = bheadp; bp != (BUFFER*) NULL; bp = bp->b_bufp) {

		/* dont kill a special buffer	*/
		if(bp->b_flag&BFINVS)
			continue;

		done = TRUE;

		if((status = mlyesno("Kill%s buffer %s [y/n]? ",
				(bp->b_flag & BFCHG ? " (modified)" : ""),
				bp->b_bname)) == FALSE)
			continue;

		if(status == ABORT)
			return(status);

		bp->b_flag &= ~BFCHG;		/* dont muck about	*/
		zotbuf(bp, TRUE);
	}

	if(!done)
		mlwrite("%WNo suitable buffers to kill");
	else
		mlerase();

	return(TRUE);
}

/* Dispose of a buffer, by name.
 * 
 * If the buffer is currently displayed more than once, just delete the current window.
 *
 * Ask for the name. Look it up (don't get too upset if it isn't there at all!).
 *
 * Clear the buffer (ask if the buffer has been changed).
 * Then free the header line and the buffer header.
 *
 * The default count is unset.
 * If the count is set to non zero: do not display errors.
 *
 * Command:  delete-buffer   kill-buffer
 * Bound to: C-x k
 *
 * Return:
 *	TRUE  Buffer killed
 *	      Buffer does not exist - ie name entered was not of a buffer
 *	FALSE invisible buffer - not killed
 *	ABORT user answered C-g
 *	zotbuf() return TODO
 */
int	killbuffer(int f, Cnt n)
{
	register BUFFER* bp;
	register int	s;
	BYTE		prompt[MAXCOL];	/* prompt we put out		*/
	int		silent = 0;	/* no messages if true		*/

	if(f && n)
		silent = 1;

	// If curbp is an invisible buffer do not offer as default to kill - as it cannot be killed
	if(curbp->b_flag & BFINVS)
		strcpy(prompt, "Kill buffer: ");
	else
		sprintf(prompt, "Kill buffer %s: ", curbp->b_bname);

	switch((s = mlprompt(prompt, MLXBUF | MLU8LEN, bmlBufHistp))) {
	  case TRUE:
		break;
	  case FALSE:
		if(curbp->b_flag & BFINVS)
			return(s);
		lcpys(bmlBufHistp->b_dotp, curbp->b_bname);
		break;
	  case ABORT:
		return(s);
	}

	// Find the buffer that is to be killed:
	if((bp = bfind(ltext(bmlBufHistp->b_dotp), BFNNONE, 0)) == NULL) {
		/* Easy if unknown.	*/
		if(!silent)
			mlwrite("%W""%l: no such buffer", bmlBufHistp->b_dotp);
		return(TRUE);
	}

	if(bp->b_flag & BFINVS) {
		/* dont kill a special buffer	*/
		if(!silent)
			mlwrite("%ECannot kill special buffer %s", bp->b_bname);
		return(FALSE);
	}

	/* No need to say that we killed the buffer - zotbuf() does that.
	 * More interestingly zotbuf() might have said something about
	 * the new buffer displayed.
	 */
	s = zotbuf(bp, silent);

	return(s);
}

/* Kill the buffer pointed to by bp.
 *
 * Say what we have done - unless silent
 *
 * Return ?? TODO
 */
int	zotbuf(BUFFER* bp, int silent)
{
	BUFFER*	bp1;
	BUFFER*	bp2;
	int	s;
	Window*	wp;
	char*	bname;

	if(bp->b_exec) {
		mlwrite("%ECannot delete an executing buffer");
		return(FALSE);
	}

	/* Before we delete the window - check that the user is content
	 * with the buffer being destroyed.
	 */
	if((s = bDiscardOK(bp)) != TRUE)
		return(s);

	/* If the buffer we want to kill is displayed on the screen, make
	 * it the current window and then remove it from the screen. Once
	 * this has been done, we can then go about zapping it.
	 */
	if(bp->b_nwnd > 0) {
		Window*	oCurwp = curwp;

		for(wp = wheadp; wp != (Window*) NULL; ) {
			/* If there is only one window - do not attempt
			 * to delete it. To do so could result in an infinite
			 * loop of killing and recreating buffer "main".
			 */
			if(wheadp->w_wndp == NULL)
				break;

			if(wp->w_bufp == bp) {
				curwp = wp;
				if(curwp == oCurwp)
					oCurwp = NULL;

				curbp = curwp->w_bufp;
				if(delwind(FALSE, FALSE) != TRUE)
					return(FALSE);

				wp = wheadp;	// Restart the scan
				continue;
			}

			wp = wp->w_wndp;	// Next window
		}

		/* Reset curwp to what it was.
		 * If we zapped curwp - point it at the head window
		 */
		if(oCurwp == NULL)
			oCurwp = wheadp;
		curwp = oCurwp;
	}
	bname = bp->b_bname;

	if(strcmp(lbufn, bname) == 0)		/* Zap last buffer name */
		lbufn[0] = '\0';		/* if its this buffer	*/

	/* We earlier asked the question about it being OK to discard changes.
	 * bclear() will call bDiscardOK() again if the buffer is marked changed.
	 * Prevent that.
	 */
	bp->b_flag &= ~BFCHG;

	if((s=bclear(bp)) != TRUE) {		/* Blow text away.	*/
		curbp = curwp->w_bufp;
		return(s);
	}

	lunsetMark(bp->b_linep);		/* Unmark the line	*/
	FreeLine(bp->b_linep);			/* Release header line. */
	bp1 = NULL;				/* Find the header.	*/
	bp2 = bheadp;
	while(bp2 != bp) {
		bp1 = bp2;
		bp2 = bp2->b_bufp;
	}
	bp2 = bp2->b_bufp;			/* Next one in chain.	*/
	if(bp1 == NULL)				/* Unlink it.		*/
		bheadp = bp2;
	else
		bp1->b_bufp = bp2;
	if(!silent)
		mlwrite("%Ibuffer %s killed", bname);	/* say it went ok	*/

	free(bp->b_fname);
	free(bp->b_fnameClean);
	free(bp->b_bname);

	FreeText(bp->b_key);

	free(bp);				/* Release buffer block */

	/* This is only going to happen if we kill the last buffer being displayed.
	 * Sometimes this is caught elsewhere.
	 * Don't check for bfind() [malloc()] failure - we will have plenty of memory.
	 */
	if(wheadp->w_wndp == (Window*)NULL && curwp->w_bufp == bp) {

		// Search to find a visible buffer and display that
		curbp = (BUFFER*)NULL;
		for(bp = bheadp; bp; bp = bp->b_bufp)
			if((bp->b_flag & BFINVS) == 0) {
				// Got one
				curbp = bp;
				break;
			}

		// If we did not find anything - create a buffer "main"
		if(curbp == NULL)
			curbp = bfind(mainBname, BFNCREATE|BFNACTIVE, 0);

		// Display the buffer found or created:
		ConnectBuffer(curbp, curwp, 0);

		// If not read it - do so now:
		if(curbp->b_active != TRUE)
			readInactive();
	}

	curbp = curwp->w_bufp;

	return(TRUE);
}

/* Disconnect the buffer from window wp.
 *
 * This decrememts the how-many-windows count &
 * this is the last that displays it: save to the buffer the
 * position of ".", marks, etc.
 *
 * Note that wp->w_bufp must then be assigned to another buffer
 * unless the window is being destroyed. This is not done here.
 *
 * If the buffer becomes no longer visible - consider auto save.
 */
void	DisconnectBuffer(Window* wp)
{
	BUFFER*	bp = wp->w_bufp;
	NMARK*	nm;

	if(--bp->b_nwnd == 0) {
		/* We are disconnecting the last window that displays the buffer
		 * Save state so that when we next display the window
		 */
		SaveWindowPosToBuffer(bp, wp);
		bp->b_winopts = wp->w_opts;
		CheckRenewVisible();	/* Do we now need to do auto checkage() ? */

		// Auto save now that the buffer is no longer displayed ?
		if((curbp->b_mode & MDASAVE) != 0)
			filesave(FALSE, 0);
	} else {
		// Check for now unused numeric marks
		for(nm = wp->w_nmarks ; nm < wp->w_nmarks + NUM_NUM_MARKS; nm++)
			if(nm->n_pline != (LINE*)NULL) {
				// Unset it: even if it is numerically marked twice,
				// soon it won't be.
				LINE*	line = nm->n_pline;
				nm->n_pline = (LINE*)NULL;
				l_checkNmark(line);
			}
	}
}

/* Save the window wp's "." and mark to buffer bp.
 *
 * This could be on close of last buffer on a window
 * or that some window independent buffer work is needed.
 */
void	SaveWindowPosToBuffer(BUFFER* bp, Window* wp)
{
	bp->b_topLinep   = wp->w_topLinep;
	bp->b_toplinenum = wp->w_toplinenum;
	bp->b_dotp       = wp->w_dotp;
	bp->b_doto       = wp->w_doto;
	bp->b_markp      = wp->w_markp;
	bp->b_marko      = wp->w_marko;
	bp->b_nmarkOff   = wp->w_nmarkOff;
	memcpy((void*)bp->b_nmarks, (void*)wp->w_nmarks, sizeof(bp->b_nmarks));

	/* This is a good time to check that we are not breaking the UTF8
	 * requirement for "." and marks.
	 * Especially here where wp might have the HEX option which allows
	 * that to be broken.
	 */
	utf8_ensurePosStart(bp, 0);
}

/* Record the positions of "." and mark at *pos, also the alpha marks and the
 * numeric marks.
 *
 * Only notice them if they are between starts and endl.
 * endl may be curbp->b_linep - end of buffer.
 * starts must NOT be curbp->b_linep.
 *
 * These days everything that it needs has a copy in the buffer. So start by
 * saving the finding a window that displays it and saving that window window
 * to the buffer.
 *
 * Looks in current buffer - curbp.
 */
void	RecordDotAndMark(BufPos* pos, const LINE* startl, const LINE* endl)
{
	AmBp*	amp;		/* Next Alpha mark struct */
	AmBp*	nmp;		/* Next Numeric mark struct */
	int	started = 0;
	LINE*	lp;
	BufLns	numline = 1;	/* probe line # */
	Window*	foundWp = NULL;	/* Window that has it */
	Window*	wp;

	// Clear out anything in pos:
	memset(pos, 0, sizeof(*pos));
	amp = pos->bp_am;
	nmp = pos->bp_nm;

	// Try and find a window that is displaying the buffer.
	// Start with curwp, choose this by preference
	if(curwp->w_bufp == curbp)
		foundWp = curwp;

	// Scan for other windows if just to trash w_lastdismarkp
	for(wp = wheadp; wp; wp = wp->w_wndp)
		if(wp->w_bufp == curbp) {
			wp->w_lastdismarkp = NULL;	// Will soon be invalid

			if( !foundWp)
				foundWp = wp;
		}

	// Note the position in the window that we (first) found
	if(foundWp)
		SaveWindowPosToBuffer(curbp, foundWp);

	// Set offset to -1 if at end of line:
	pos->bp_doto = curbp->b_doto;
	pos->bp_marko = curbp->b_marko;
	if(pos->bp_doto == lused(curbp->b_dotp))
		pos->bp_doto = -1;
	if(curbp->b_markp && pos->bp_marko == lused(curbp->b_markp))
		pos->bp_marko = -1;

	// UTF8: recalculate the offsets in terms of characters, not bytes:
	if(curbp->b_mode & MDUTF8) {
		if(pos->bp_doto != -1)
			pos->bp_doto = (LineLn)utf8_chrnlen(ltext(curbp->b_dotp), (size_t)pos->bp_doto);
		if(curbp->b_markp && pos->bp_marko != -1)
			pos->bp_marko = (LineLn)utf8_chrnlen(ltext(curbp->b_markp), (size_t)pos->bp_marko);
	}

	/* Find line number in buffer of mark and "."
	 * Don't worry about TopLine, we already have it if there is a window.
	 * If no window we don't care.
	 *
	 * Only 'notice' "." and mark if they are in the range of lines startl .. lback(endl)
	 *
	 * Have to start at top of buffer to get line # right, don't bother to
	 * look beyond endl ... unless it is the last line in the buffer - in which
	 * case process this as numline == -1 and then stop.
	 */
	for(lp = lforw(curbp->b_linep); ; lp = lforw(lp), numline++) {

		// Stop if hit end of lines ... but continue if end of buffer
		if(lp == endl && endl != curbp->b_linep)
			break;

		if(lp == startl)
			started = 1;

		if( ! started) {
			// Stop potential infinite loop:
			if(lp == curbp->b_linep)
				break;

			// Ignore stuff before the start
			continue;
		}

		// Hit end of buffer ? Set that as line # -1
		// Don't worry - the loop cannot continue after this
		if(lp == curbp->b_linep)
			numline = -1;

		if(lp == curbp->b_topLinep)
			pos->bp_topl = numline;
		if(lp == curbp->b_dotp)
			pos->bp_dotl = numline;
		if(lp == curbp->b_markp)
			pos->bp_markl = numline;

		if(lisNUL(lp))		// NUL flag ?
			pos->bp_NUL = 1;

		if(lmarked(lp) || lp == endl) {
			/* The line has at least one alpha mark on it.
			 * Scan them and record in bp_am.
			 */
			AMARK*	ap;		// Alpha mark pointer

			for(ap = curbp->b_amark; ap; ap = ap->m_next) {
				if(ap->m_pline != lp)	// Not this line ?
					continue;

				amp->abp_line = numline;
				amp->abp_name = ap->m_name;
				amp->abp_off = ap->m_oline;

				if(amp->abp_off == lused(lp))
					amp->abp_off = -1;
				else
					if(curbp->b_mode & MDUTF8)
						amp->abp_off = (LineLn)utf8_chrnlen(ltext(lp), (size_t)amp->abp_off);

				amp++;
			}

		}

		if(lisNmarked(lp)) {
			/* Numeric mark somewhere.
			 * Store in bp_nm
			 */
			NMARK*	wmp;		// Numeric mark pointer in buffer

			for(wmp = curbp->b_nmarks; wmp < curbp->b_nmarks + NUM_NUM_MARKS; wmp++) {
				if(wmp->n_pline != lp)
					continue;

				nmp->abp_line = numline;
				nmp->abp_off = wmp->n_oline;
				/* Name is the number
				 * BEWARE: this number is the position in the w_nmarks array,
				 * it is not what the user things as the numeric mark number.
				 * As there are more than 10 numeric marks - the last 2 might
				 * be named ':' & ';'.
				 */
				nmp->abp_name = (BYTE)(wmp - curbp->b_nmarks);

				if(nmp->abp_off == lused(lp))
					nmp->abp_off = -1;
				else
					if(curbp->b_mode & MDUTF8)
						nmp->abp_off = (LineLn)utf8_chrnlen(ltext(lp), (size_t)nmp->abp_off);

				// Next recorded numeric mark
				nmp++;

				// Remove the reference and check for more. Needed ?
				// wmp->n_pline = (LINE*)NULL;
				// l_checkNmark(lp);
			}
		}

		// Just looked at line after the last line in the buffer ?
		if(lp == curbp->b_linep)
			break;
	}
}

/* Now patch things up so that it displays as much as it did before the
 * refresh/filter/... - as far as we can.
 * Go to the line number that we were on, restore top of buffer line and possibly the mark.
 *
 * curbp->b_numLines is fixed.
 *
 * This works on curbp - but maybe not curwp - ie the buffer might not be displayed
 * in a window - but if it is fix the window as well.
 *
 * It is also possible that the region (startl .. endl) does not contain "." or
 * the mark - in which case do not try to fix them.
 *
 * It is also possible that the original region (saved by RecordDotAndMark())
 * did contain "." and/or mark but that the new one does not (it is not as long
 * as the original one). In that case "." will be left at endl and mark at startl.
 *
 * WindowsSync() should prob be called after this.
 *
 * If one line had the L_NUL flag set, then set that on all lines.
 * A line by line record of L_NUL is not kept.
 */
void	RestoreDotAndMark(BufPos* pos, const LINE* startl, const LINE* endl)
{
	AmBp*	amp = pos->bp_am;	/* Next Alpha mark struct */
	AmBp*	nmp = pos->bp_nm;	/* Next Numeric mark struct */
	int	started = 0;		/* Seen startl and not endl */
	LINE*	lp;
	BufLns	numline = 1;		/* probe line # */

	/* Find by line number: the top row of the screen, "." and any mark.
	 *
	 * Once we have reset something to a new LINE* clear the line number
	 * to 0 - this implies that it was something that we were not restoring
	 * (well, now no longer restoring).
	 * This is important: because anything for which we have a line number it is
	 * because it is either:
	 * -1		   end of buffer - so set to endl
	 * something else  the buffer is shorter - so set to endl
	 *
	 * Take care to update all the windows that display the buffer.
	 */
	for(lp = lforw(curbp->b_linep); ; lp = lforw(lp), numline++) {
		// Need a separate one as we set to -1 at buffer end
		BufLns	LineNumber = numline;

		// Stop if hit end of lines ... but continue if end of buffer
		if(lp == endl && endl != curbp->b_linep)
			break;

		if(lp == startl)
			started = 1;

		if( ! started) {
			// Stop potential infinite loop:
			if(lp == curbp->b_linep)
				break;

			// Ignore stuff before the start
			continue;
		}

		if(lp == curbp->b_linep)
			LineNumber = -1;

		if(pos->bp_topl == LineNumber) {
			// Found top displayed line number
			curbp->b_topLinep = lp;
			curbp->b_toplinenum = LineNumber;
			pos->bp_topl = 0;
		}

		// NUL flag on one of the lines ?
		// Don't complain if it does not get set (out of memory)
		if(pos->bp_NUL)
			lsetNUL(lp);

		if(pos->bp_dotl == LineNumber) {
			// Found the position of ".".
			// A complication is that the new line could be shorter than
			// the old one.
			// Also: we count in characters - that depends on UTF8 mode
			curbp->b_dotp = lp;
			if((curbp->b_doto = pos->bp_doto) > lused(lp))
				curbp->b_doto = lused(lp);

			if(curbp->b_doto == -1) {
				curbp->b_doto = lused(lp);
			} else {
				if(curbp->b_mode & MDUTF8) {
					size_t	bl = (size_t)curbp->b_doto;
					curbp->b_doto = (LineLn)utf8_maxlen(ltext(lp), (size_t)lused(lp), &bl);
				}
			}
			pos->bp_dotl = 0;
		}

		if(pos->bp_markl == LineNumber) {
			// Found the position of the mark.
			// A complication is that the new line could be shorter than
			// the old one.
			// Also: we count in characters - that depends on UTF8 mode

			curbp->b_markp = lp;
			if((curbp->b_marko = pos->bp_marko) > lused(lp))
				curbp->b_marko = lused(lp);
			lsetMark(lp);

			if(curbp->b_marko == -1) {
				curbp->b_marko = lused(lp);
			} else {
				if(curbp->b_mode & MDUTF8) {
					size_t	bl = (size_t)curbp->b_marko;
					curbp->b_marko = (LineLn)utf8_maxlen(ltext(lp), (size_t)lused(lp), &bl);
				}
			}
			pos->bp_markl = 0;
		}

		/* Alpha marks ?
		 * There could be more than one on the line.
		 * Once up on a time the marks were left there (pointing to NULL)
		 * by readin(), etc, but now they can be blown away completely - so
		 * might need to be remade.
		 */
		while(amp->abp_line == LineNumber) {
			AMARK*	ap;

			for(ap = curbp->b_amark; ap; ap = ap->m_next)
				if(ap->m_name == amp->abp_name)	// This mark ?
					break;

			if( !ap) {
				// Not found, make a new one
				if((ap = (AMARK*) malloc(sizeof(AMARK))) == (AMARK*) NULL)
					break;	// Can't report it ... do what we can, the user will soon know

				// Stuff at the front - easiest
				ap->m_next = curbp->b_amark;
				curbp->b_amark = ap;
				ap->m_name = amp->abp_name;
			}

			// Set the mark on this line.
			ap->m_pline = lp;
			lsetMark(lp);
			if((ap->m_oline = amp->abp_off) > lused(lp))
				ap->m_oline = lused(lp);

			if(ap->m_oline == -1) {
				ap->m_oline = lused(lp);
			} else {
				if(curbp->b_mode & MDUTF8) {
					size_t	bl = (size_t)ap->m_oline;
					ap->m_oline = (LineLn)utf8_maxlen(ltext(lp), (size_t)lused(lp), &bl);
				}
			}

			amp->abp_line = 0;
			amp++;
		}

		/* Numeric marks on the line ?
		 * The name is the binary number of the numeric mark.
		 */
		while(nmp->abp_line == LineNumber) {
			NMARK*	np = &curbp->b_nmarks[nmp->abp_name];

			np->n_pline = lp;
			lsetNMark(lp);
			if((np->n_oline = nmp->abp_off) > lused(lp))
				np->n_oline = lused(lp);

			if(np->n_oline == -1) {
				np->n_oline = lused(lp);
			} else {
				if(curbp->b_mode & MDUTF8) {
					size_t	bl = (size_t)np->n_oline;
					np->n_oline = (LineLn)utf8_maxlen(ltext(lp), (size_t)lused(lp), &bl);
				}
			}

			nmp->abp_line = 0;
			nmp++;
		}

		// Reached the end of region. Could just break, but if we
		// do then numline would be wrong when setting b_numLines.
		if(lp == endl)
			started = 0;

		// Stop at end of buffer:
		if(lp == curbp->b_linep)
			break;
	}

	// No mark, set to start if there was one in the original region:
	if(pos->bp_markl != 0) {
		curbp->b_markp = (LINE*)endl;
		curbp->b_marko = 0;
	}
	if(pos->bp_topl != 0) {
		curbp->b_topLinep = (LINE*)endl;
		curbp->b_toplinenum = numline;
	}
	if(pos->bp_dotl != 0) {
		curbp->b_dotp = (LINE*)endl;
		curbp->b_doto = 0;
	}

	/* If there were fewer lines than those that were replaced it is
	 * possible that there is an alpha mark LINE* that points to something
	 * that has been zapped. Look for those and set to start of what inserted.
	 */
	while(amp->abp_line != 0) {
		AMARK*	ap;

		for(ap = curbp->b_amark; ap; ap = ap->m_next) {
			if(ap->m_name != amp->abp_name)	// Not this mark ?
				continue;

			// Clear this alpha mark
			ap->m_pline = (LINE*)NULL;
		}

		amp++;
	}

	// Clear numeric marks that were in the range but we have not seen:
	while(nmp->abp_line != 0) {
		NMARK*  np = &curbp->b_nmarks[nmp->abp_name];

		// If restoring an internal use one that we did not find
		// set to the start of the first line in the buffer:
		if(nmp->abp_name > '9') {
			np->n_pline = lforw(curbp->b_dotp);
			np->n_oline = 0;
		} else {
			np->n_pline = (LINE*)NULL;	// Lost, just trash it
		}

		nmp++;
	}

	ReconnectWindowsToBuffer(curbp);

	curbp->b_numLines = numline - 1;	// Fix # lines in buffer
}

/* Connect a buffer to a window.
 *
 * Increment the how-many-windows count & set the window position
 * to the buffer position; window offset to 0. Window hard/forced redisplay.
 *
 * If reConnect, the buffer is already connected, reset position to display.
 *	In which case bp b_dotp/b_markp, etc, must be valid.
 */
void	ConnectBuffer(BUFFER* bp, Window* wp, int reConnect)
{
	if( !reConnect) {
		/* If a window is already displaying the buffer then
		 * the LINE pointers in bp will be out of date (they are
		 * only written to when the last window on the buffer is closed).
		 * So: find a window (not wp) and save its state to bp.
		 */
		if(bp->b_nwnd > 0) {
			Window*	ws;
			for(ws = wmessp; ws != NULL; ws = ws->w_wndp)
				if(ws != wp && ws->w_bufp == bp) {
					SaveWindowPosToBuffer(bp, ws);
					break;
				}
		}

		bp->b_nwnd++;		/* mark us as more in use */
		CheckRenewVisible();	/* Do we now need to do auto checkage() ? */
		wp->w_opts = bp->b_winopts;
	}

	wp->w_bufp = bp;

	wp->w_lasttop = 1;
	wp->w_topLinep = bp->b_topLinep;
	wp->w_toplinenum = bp->b_toplinenum;
	wp->w_lastnumLines = bp->b_numLines;

	wp->w_dotp = bp->b_dotp;
	wp->w_doto = bp->b_doto;

	wp->w_markp = bp->b_markp;
	wp->w_marko = bp->b_marko;

	wp->w_nmarkOff = bp->b_nmarkOff;	// Numeric marks
	memcpy((void*)wp->w_nmarks, (void*)bp->b_nmarks, sizeof(wp->w_nmarks));

	wp->w_partoff = wp->w_markpartoff = wp->w_lastdismarkpo = 0;	// Start afresh
	wp->w_lastdismarkp = NULL;
	wp->w_leftoffset = 0;
	wp->w_flag = WFMODE | WFHARD | WFNWID;

	/* If the window top line is the buffer header line then the window
	 * display will be empty. Force a reframe so that the user sees something.
	 */
	if(wp->w_topLinep == bp->b_linep) {
		wp->w_force = 0;
		wp->w_flag |= WFFORCE;
	}
}

/* Find every window that displays bp and refresh the connection.
 *
 * Good idea to do this after the buffer has been trashed and read in - so that
 * Window LINE pointers may no longer be valid.
 *
 * If bp is not displayed in any window this does nothing - ie it is safe.
 */
void	ReconnectWindowsToBuffer(BUFFER* bp)
{
	Window*	wp;		// Scanning pointer to windows

	for(wp = wmessp; wp; wp = wp->w_wndp)
		if(wp->w_bufp == bp)
			ConnectBuffer(bp, wp, 1);
}

/* Reestablish the connection of curwp & curbp.
 *
 * If curbp is displayed - reconnect
 * If curbp is not displayed - connect afresh
 *
 * This situation can arise when curbp is read in not using
 * a direct requested method like find-file.
 *
 * This can result in curwp changing, so the mode line needs updating.
 * So cause all mode lines to be checked.
 */
void	EnsureCurbpConnectedToCurwp(void)
{
	Window*	wp = onscreen(curbp);

	if(wp != (Window*)NULL) {
		// It could be displayed more than once:
		ReconnectWindowsToBuffer(curbp);
		curwp = wp;
	} else {
		DisconnectBuffer(curwp);
		ConnectBuffer(curbp, curwp, 0);
	}

	upmode((BUFFER*)NULL, WFMODE);
}

/* Change the name of the current buffer.
 *
 * The current window's mode line is marked for renewal - on success
 *
 * Return TRUE if this works
 *	  FALSE on error - when a message will have been written
 */
int	setBufferName(const BYTE* str)
{
	/* Buffer names cannot contain a '/',
	 * to allow this would cause problems with bquicksave().
	 */
	if(strchr(str, '/')) {
		mlwrite("%EA buffer name cannot contain a '/'");
		return(FALSE);
	}

	/* and check for duplicates */
	if(bfind(str, BFNNONE, 0) != NULL) {
		mlwrite("%EBuffer '%s' already exists", str);
		return(FALSE);
	}

	utf8_strfmake(&curbp->b_bname, str);	/* copy buffer name to structure */
	curwp->w_flag |= WFMODE;	/* make mode line replot */

	return(TRUE);
}

/* Rename the current buffer.
 *
 * The count is ignored.
 *
 * Command:  name-buffer
 * Bound to: M-C-n
 *
 * Return TRUE  if this works
 *	  FALSE on error - when a message will have been written
 *	  ABORT C-g
 */
int	namebuffer(int f, Cnt n)
{
	int	ret;

	/* prompt for and get the new buffer name */
	if(mlprompt("New buffer name: ", MLXBUF | MLU8LEN, bmlBufHistp) != TRUE)
		return(FALSE);

	ret = setBufferName(ltext(bmlBufHistp->b_dotp));
	mlerase();

	return(ret);
}

/* List all of the active buffers.
 *
 * First update the special buffer that holds the list ([List]).
 *
 * Next make sure at least 1 window is displaying the buffer list, splitting
 * the screen if this is what it takes.
 *
 * Lastly, repaint all of the windows that are displaying the list.
 *
 * A numeric argument forces it to list invisable buffers as well.
 * The actual value of n is not important - f is what is looked at.
 *
 * If the buffer [List] is currently already displayed, try to preserve which bit
 * is displayed - ie run this again to refresh what the user is looking at.
 *
 * A count forces it to list invisable buffers as well.
 * The actual value of n is not important - f is what is looked at.
 * The default count is not set.
 * 
 * Command:  list-buffers
 * Bound to: C-x C-b
 *
 * Return TRUE if all goes well
 *        FALSE on error
 *	  ABORT if prompt is aborted
 */
/* ARGSUSED */
int	listbuffers(int f, Cnt n)
{
	int	s;
	BUFFER*	saveCurbp = curbp;
	Window*	saveCurwp = curwp;
	BufPos	Pos;		// Where to store buffer position
	int	empty;		// True if the buffer is newly filled, ie empty at start

	empty = blistp->b_linep == lforw(blistp->b_linep);

	/* If there is something there - remember place of ".", etc.
	 * Remember this for use below. It might not be displayed now but it will
	 * be later - can't restore to a position that we did not save.
	 */
	if( !empty) {
		curbp = blistp;
		RecordDotAndMark(&Pos, lforw(blistp->b_linep), blistp->b_linep);
		curbp = saveCurbp;
	}

	// Need to create & connect [List] early so that it shows correctly in the list of windows
	blistp->b_flag &= ~BFCHG;		/* Don't complain!	*/
	if((s = bclear(blistp)) != TRUE)	/* Blow old text away	*/
		return(s);

	if( !showlist(blistp))
		return(FALSE);

	/* showlist() will, sometimes, change curwp.
	 * That means that $curcol, etc, are for the new window.
	 * Reset so that the correct (current when command invoked) window is used.
	 * If the current window is on [List] then these are 0 as the buffer is cleared.
	 */
	curwp = saveCurwp;

	if((s = makelist(f)) != TRUE)
		return(s);

	/* That will have put "." at the bottom of the buffer.
	 * Don't want that if this is the first time this is done
	 * or if the buffer was not being displayed.
	 */
	if(empty)
		blistp->b_dotp = lforw(blistp->b_linep);

	// Make the window dot the first line of the buffer
	// Indeed all windows open on [List]
	ReconnectWindowsToBuffer(blistp);

	// Restore to a saved position ?
	if( !empty) {
		curbp = blistp;
		RestoreDotAndMark(&Pos, lforw(blistp->b_linep), blistp->b_linep);
		WindowsSync();
	}

	// Ensure that what was the current window remains the current one:
	curwp = saveCurwp;
	curbp = saveCurbp;

	// Ensure that new geography has correct edges
	upmode(NULL, WFMODE);

	return(TRUE);
}

/* Create a window that displays buffer dispbp.
 *
 * Return TRUE
 *        FALSE on error
 */
int	showlist(BUFFER* dispbp)
{
	Window*		wp;

	if(dispbp->b_nwnd == 0) {	/* Not on screen yet */
		if((wp = wpopup()) == NULL)
			return(FALSE);

		DisconnectBuffer(wp);
		ConnectBuffer(dispbp, wp, 0);
	}

	ReconnectWindowsToBuffer(dispbp);

	return(TRUE);
}

/* Clear something to all spaces & NUL terminate.
 */
void	ClearOut(BYTE* line, int size)
{
	memset(line, ' ', (size_t)size);
	line[size - 1] = '\0';
}

/* Put the mode letters at line - taken from modecode[].
 * '.' if the mode isn't set.
 *
 * mode is buffer's b_mode
 * flag is buffer's b_flag
 *
 * RENEW shown as 'r' if BRENEWAPPEND - (just read new lines).
 */
void	DispModes(BYTE* line, int mode, int flag)
{
	register int	i;

	for(i = ModeBuf.md_mode_cnt - 1; i >= 0; i--, line++) {
		*line = (BYTE)((mode & (1 << i)) ? modecode[i] : '.');

		// Beware: only if the mode is set:
		if(((1 << i) & mode) == MDRENEW && (flag & BRENEWAPPEND) != 0)
			*line = 'r';
	}
}

/*
 * This routine rebuilds the
 * text in the buffer [List]
 * that holds the buffer list. It is called
 * by the list buffers command.
 *
 * Iflag is TRUE if invisible buffers are to be listed
 *
 * Return TRUE if everything works
 *        FALSE if there  * is an error (if there is no memory).
 */
int	makelist(int iflag)
{
	register BYTE*	cp2;
	register BUFFER*bp;
	register LINE*	lp;
	register int	i;
	int		add_ok = 0;
	BufChrs		nbytes;		/* # of bytes in current buffer */
	BufLns		nlines;		/* # of lines in current buffer */
	BufChrs		nunused;	/* # unused bytes in current buffer's LINEs */
	BufChrs		thisAdj[3];	/* Lines/bytes this buffer when traverse */
	ST_tab*		stt;		/* Traverse STatistics[] */
	LINE*		blistLine = NULL; // Line that has size, etc, for the buffer that we are writing to
	Window*		wp;		/* Window pointer */
	AMARK*		amp;		/* To count alpha marks */
	BYTE		b[20];		/* Scratch */

	ST_maxBytes = 0;		/* Max # bytes in any buffer - ie largest */
	ST_maxLines = 0;		/* Max # lines in any buffer - ie longest */
	ST_longestLine = 0;		/* Longest line in any buffer */

	// Unecessary code to stop the compiler from wingeing
	thisAdj[0] = thisAdj[1] = thisAdj[2] = 0;

	// Unset all filenames
	setFnames(blistp, "");

	// Ensure that the scratch line is big enough (actually a bit more than the minimum)
	// It may be expanded more later, but I want enough to get the basics in
	// Use the playText as insStrLine() makes it longer as needed and cleans to UTF8
	if( !textMoreBytes(playText, GM_FILE + BUFFER_NAME_SIZE))
		return(FALSE);

	// αM does not display properly on some displays, unfortunately, so use AM

	if(
#if	UNIX
	   addline("ACTN    MODES              Size    Lines AM Buffer            Mode File",
							blistp, 1) == FALSE ||
	   addline("----    -----              ----    ----- -- ------            ---- ----",
							blistp, 1) == FALSE)
#else
	   addline("ACTN    MODES              Size    Lines AM Buffer           File",
							blistp, 1) == FALSE ||
	   addline("----    -----              ----    ----- -- ------           ----",
							blistp, 1) == FALSE)
#endif
		return(FALSE);

	/* Build line to report global mode settings */
	ClearOut(ttext(playText), GM_FILE + BUFFER_NAME_SIZE);

	/* Output the mode codes */
	DispModes(tgetcaddr(playText, GM_MODES), gmode, 0);

	strcpy(tgetcaddr(playText, GM_BUFFER), "Global Modes");

	if(addline(ttext(playText), blistp, 1) == FALSE)
		return(FALSE);

	/* output the list of buffers
	 * While we are at it, update buffer statistics.
	 */
	ST_buffers = ST_buffersInvis = ST_buffersActive = ST_lines = ST_bytes =
		ST_unusedBytes = ST_mostUnused = 0;

	// Update all inode meta data in case something changed recently
	checkage(0);

        for(bp = bheadp ; bp != (BUFFER*) NULL; bp = bp->b_bufp) {
		ST_buffers++;

		/* Make all spaces so that we don't need to insert them as
		 * we go along.
		 */
		ClearOut(ttext(playText), GM_FILE + BUFFER_NAME_SIZE);

		/* Invisible ?
		 * Continue even if is so that we get the tallies
		 */
                if((bp->b_flag & BFINVS) != 0) {
			ST_buffersInvis++;
			tputc(playText, GM_ACTIVE, 'I');
		} else {
			/* Output status of ACTIVE flag (has the file been read in?
			 * Do not count invisible buffers as active.
			 */
			if(bp->b_active == TRUE) {	/* "@" if active regular file (else file type) */
				tputc(playText, GM_ACTIVE, (CHAR)(bp->b_flag & BFROMSAVE ? 'S' :
					(bp->b_lsType == '-' ? '@' : bp->b_lsType)));
				ST_buffersActive++;
			}

			/* Overwrite that with X if it is out of date.
			 * Never true if not running under UNIX
			 */
			if((bp->b_flag & BFOLD))
				tputc(playText, GM_ACTIVE, bp->b_lsType == '.' ? 'x' : 'X');

			// Show if it has auto RENEWed:
			if((bp->b_flag & BRENEWED))
				tputc(playText, GM_ACTIVE, 'R');
		}

		/* output status of changed flag */
                if(bp->b_flag & BFCHG)		/* "*" if changed */
			tputc(playText, GM_CHANGED, '*');

		/* report if the file is truncated */
                if(bp->b_flag & BFTRUNC)
			tputc(playText, GM_TRUNC, '#');	/* "#" if truncated */

		if(bp->b_btopline != bp->b_linep || bp->b_abotline != bp->b_linep)
			tputc(playText, GM_NARROW, '<');	/* "<" if narrowed */

		/* output the mode codes */
		DispModes(tgetcaddr(playText, GM_MODES), bp->b_mode, bp->b_flag);

		nunused = nlines = nbytes = 0L;		/* Count lines/bytes in buf */
		lp = lforw(bp->b_linep);
		while(lp != bp->b_linep) {
			LineLn	llen = (long)lused(lp);

			// Allocation size of l_text is granular
			LineLn	unused = lsize(lp) - llen;
			nunused += unused;
			if(unused > ST_mostUnused)
				ST_mostUnused = unused;

			llen++;			// +1 for \n at end

			nbytes += llen;
			if(llen > ST_longestLine)
				ST_longestLine = llen;
			lp = lforw(lp);
			nlines++;
		}

		sprintf(b, "%10ld", (long)nbytes);	/* 10 - large enough for 2^31 */
		ST_lines += nlines;
		ST_bytes += nbytes;
		ST_unusedBytes += nunused;
		strinsstr(tgetcaddr(playText, GM_SIZE), b);

		sprintf(b, "%9ld", (long)nlines);	/* Unlikely, but could be too small */
		strinsstr(tgetcaddr(playText, GM_LINES), b);

		/* A new max ? */
		if(nlines > ST_maxLines)
			ST_maxLines = nlines;
		if(nbytes > ST_maxBytes)
			ST_maxBytes = nbytes;

		// Count & show # α marks:
		for(amp = bp->b_amark, i = 0; amp; amp = amp->m_next)
			if(amp->m_pline)
				i++;
		sprintf(b, "%2d", i);
		strinsstr(tgetcaddr(playText, GM_MARKS), b);

		// Note that if this contained non ASCII characters then the number of
		// characters could be more than the number of bytes. This means that all
		// the fields then do not line up.
		strinsstr(tgetcaddr(playText, GM_BUFFER), bp->b_bname);

#if	UNIX
		// Where to write the mode
		cp2 = tgetcaddr(playText, GM_MODE) + strlen(bp->b_bname) - utf8_strlen(bp->b_bname);

		/* Mode in octal, in width 4 in case setuid */
		sprintf(b, "%4.3o", filemode(bp->b_fname));
		strinsstr(cp2, b);

		cp2 += GM_FILE - GM_MODE;	/* Where to write the file name */
#else
		// Where to write the file path (name):
		cp2 = tgetcaddr(playText, GM_FILE) + strlen(bp->b_bname) - utf8_strlen(bp->b_bname);
#endif

		// Copy in the name, do not over run the buffer or move part of a UTF8 character
		// Copy the NUL byte else the line will have lots of trailing space.
		tsetused(playText, cp2 - ttext(playText));
		if(insStrText(playText, tused(playText), bp->b_fnameClean, (LineLn)strlen(bp->b_fnameClean) + 1, 1) < 0)
			return(FALSE);

		// Only generate the line if it is not invisible or iflag is set
		if((bp->b_flag&BFINVS) == 0 || iflag) {
			if(addline(ttext(playText), blistp, 0) == FALSE)
				return(FALSE);

			/* If that is the buffer that we are writing to, note the
			 * line that has just been added.
			 * This is so that at the end of everything we can update the
			 * buffer size - currently it is small, we will have only counted
			 * what there is in there now.
			 */
			if(bp == blistp)
				blistLine = lback(blistp->b_linep);
		}

		/* Note what we tallied so that we can fix later */
		if(bp == blistp) {
			thisAdj[0] = nlines;
			thisAdj[1] = nbytes;
			thisAdj[2] = nunused;
		}
	}

	// Output something that explains what the modes are
	add_ok |= !addline("", blistp, 1);

	add_ok |= !addline("ACTN Flags", blistp, 1);
	add_ok |= !addline("----------", blistp, 1);
	add_ok |= !addline("@    Active - regular file read in or created", blistp, 1);
	add_ok |= !addline(".    Active - file not read in or created", blistp, 1);
	add_ok |= !addline("!    Insufficient permission on file (/ means directory: perms or exist)", blistp, 1);
	add_ok |= !addline("X    File on disk modified since read in", blistp, 1);
	add_ok |= !addline("x    File removed since read in", blistp, 1);
	add_ok |= !addline("R    Buffer auto RENEWed as file modified", blistp, 1);
	add_ok |= !addline("S    Active - " EMACS_SAVE_SUFFIX " file read in", blistp, 1);
	add_ok |= !addline("I    Invisible/Internal editor special buffer", blistp, 1);
	add_ok |= !addline("d    Unreadable file type. Like ls: one of d, c, b, p, s"
#ifdef	S_IFDOOR	// Only solaris
		", D"
#endif
		, blistp, 1);
	add_ok |= !addline(" *   Changed since last saved to file", blistp, 1);
	add_ok |= !addline("  #  Truncated", blistp, 1);
	add_ok |= !addline("   < Narrowed", blistp, 1);
	add_ok |= !addline("", blistp, 1);

	// Show/explain what the modes mean:
	ClearOut(ttext(playText), GM_FILE + BUFFER_NAME_SIZE);
	strcpy(tgetcaddr(playText, GM_MODES), "Buffer Modes");
	add_ok |= !addline(ttext(playText), blistp, 1);

	/* Produce them with the least significant bits to the right
	 * This helps to make sense of the values in $cmode & $gmode.
	 */
	for(i = ModeBuf.md_mode_cnt - 1; i >= 0; i--) {
		BYTE*	lnp = tgetcaddr(playText, GM_MODES);
		ClearOut(ttext(playText), GM_FILE + BUFFER_NAME_SIZE);
		strncpy(lnp, "....................", (unsigned)ModeBuf.md_mode_cnt);
		lnp[ModeBuf.md_mode_cnt - i - 1] = (BYTE)modecode[i];
		cp2 = ModeBuf.md_modes[i].mn_name;
		lnp += ModeBuf.md_mode_cnt + 1;
		strinsstr(lnp, cp2);
		lnp += 7;
		strcpy(lnp, ModeBuf.md_modes[i].mn_explan);
		add_ok |= !addline(ttext(playText), blistp, 1);
	}

	add_ok |= !addline("", blistp, 1);
	add_ok |= !addline("AM = Alpha Marks", blistp, 1);

	/* Information about windows:
	 */
	add_ok |= !addline("", blistp, 1);
	add_ok |= !addline("Windows", blistp, 1);
	add_ok |= !addline("-------", blistp, 1);
	add_ok |= !addline("           Origin    Left    Size", blistp, 1);
	add_ok |= !addline("   Opts    row col   shift  row col   Buffer", blistp, 1);
	add_ok |= !addline("  -------  --- ---   -----  --- ---   ------", blistp, 1);

	// Display default window options:
	ClearOut(ttext(playText), GM_FILE + BUFFER_NAME_SIZE);
	for(i = 0; i < WinOpt.md_mode_cnt; i++)
		tputc(playText, i + WO_OPTS, (BYTE)((DefaultWindowOpts & (1 << i)) ? winOptsExplan[i].mn_name[0] : '.'));
	sprintf(tgetcaddr(playText, WO_BUFFER_NAME - 10), "%3d %3d   Global Options", AllocLines, AllocColumns);
	add_ok |= !addline(ttext(playText), blistp, 1);

	for(wp = wmessp; wp; wp = wp->w_wndp) {
		BYTE*	lb = tgetcaddr(playText, WO_OPTS);
		ClearOut(ttext(playText), GM_FILE + BUFFER_NAME_SIZE);

		// Display window options:
		for(i = 0; i < WinOpt.md_mode_cnt; i++)
			*lb++ = (BYTE)((wp->w_opts & (1 << i)) ? winOptsExplan[i].mn_name[0] : '.');

		lb = tgetcaddr(playText, WO_OPTS + WinOpt.md_mode_cnt + 2);
		// Rows includes the mode line, cols does not include any LH/RH separator
		sprintf(lb, "%3d %3ld %7ld  %3d %3d   ", wp->w_toprow, (long)wp->w_leftcol, (long)wp->w_leftoffset,
			wp->w_ntrows + ((wp->w_attrib & WATT_MODELINE) != 0),
			wp->w_ncolumns);
		lb += WO_SIZE_TO_NAME;
		strcat(lb, wp->w_bufp->b_bname);
		add_ok |= !addline(ttext(playText), blistp, 1);
	}


	// Explain window options
	add_ok |= !addline("", blistp, 1);
	add_ok |= !addline("  Window Options", blistp, 1);

	for(i = 0; i < WinOpt.md_mode_cnt; i++) {
		BYTE*	lb = tgetcaddr(playText, WO_OPTS);
		ClearOut(ttext(playText), GM_FILE + BUFFER_NAME_SIZE);
		strncpy(lb, "....................", (unsigned)WinOpt.md_mode_cnt);
		lb[i] = (BYTE)(WinOpt.md_modes[i].mn_name[0]);
		cp2 = WinOpt.md_modes[i].mn_name;
		strinsstr(tgetcaddr(playText, WinOpt.md_mode_cnt + 4), cp2);
		strcpy(tgetcaddr(playText, WinOpt.md_mode_cnt + 20), WinOpt.md_modes[i].mn_explan);
		add_ok |= !addline(ttext(playText), blistp, 1);
	}

	add_ok |= ! WriteEnvVariablesToBuffer(blistp);

	add_ok |= ! WriteUserVariablesToBuffer(blistp);

	add_ok |= ! WriteCmdFilesToBuffer(blistp);

	add_ok |= ! WriteInputKeyMappingToBuffer(blistp);

	add_ok |= ! WriteStatisticsToBuffer(blistp);

	/* Count lines/bytes in blistp
	 * The point is that we had not yet written to it when we counted it.
	 */
	nunused = nlines = nbytes = 0L;
	lp = lforw(blistp->b_linep);
	while(lp != blistp->b_linep) {
		LineLn	llen = (long)lused(lp);

		// Allocation size of l_text is granular
		LineLn	unused = lsize(lp) - llen;
		ST_unusedBytes += unused;
		if(unused > ST_mostUnused)
			ST_mostUnused = unused;

		llen++;			// +1 for \n at end

		nbytes += llen;
		if(llen > ST_longestLine)
			ST_longestLine = llen;
		lp = lforw(lp);
		nlines++;
	}
	/* A new max ? */
	if(nlines > ST_maxLines)
		ST_maxLines = nlines;
	if(nbytes > ST_maxBytes)
		ST_maxBytes = nbytes;

	/* When we wrote the number of bytes in this buffer earlier - we did
	 * not count the number of bytes written since.
	 * Patch that up, this is tacky.
	 */
	if(blistLine) {
		sprintf(b, "%10ld", (long)nbytes);
		strinsstr(lgetcaddr(blistLine, GM_SIZE), b);
		sprintf(b, "%9ld", (long)nlines);
		strinsstr(lgetcaddr(blistLine, GM_LINES), b);
	}

	/* Now fix up the summary written by WriteStatisticsToBuffer()
	 * I did try to optimise by doing a few, but it was tacky/fragile.
	 * The extra CPU is negligible - done infrequently.
	 */
	ST_lines	+= nlines  - thisAdj[0];
	ST_bytes	+= nbytes  - thisAdj[1];
	ST_unusedBytes	+= nunused - thisAdj[2];
	// Rough %age wasted space in LINEs
	ST_unusedPercent = (ST_val)((double)ST_unusedBytes * 100.0 / ((double)(ST_unusedBytes + ST_bytes - ST_lines)) + 0.5);

	/* This wasn't done in the first place, do it now.
	 * This value doesn't include any malloc overhead, but is better than nothing.
	 * This can easily exceed the bytes allocated for buffer text: ST_bytes
	 */
	ST_linesBytes = ST_lines * (ST_val)sizeof(LINE);

	for(stt = STatistics, lp = ST_FirstLine; stt->st_value; stt++, lp = lforw(lp)) {
		sprintf(b, "%10ld", *stt->st_value);
		strinsstr(lgetcaddr(lp, 0), b);
	}

	return(add_ok ? FALSE : TRUE);				/* All done */
}

/* Print statistics gathered to buffer.
 *
 * This may involve updating some before printing.
 *
 * Return TRUE  on success.
 *	  FALSE if ran out of memory.
 */
int	WriteStatisticsToBuffer(BUFFER* buffer)
{
	ST_tab*	stt;
	BYTE	line[300];
	int	ret = 0;

	/* Wall clock run time */
	ST_wallElapsed = time((time_t*)NULL) - ST_wallStart;

#if	UNIX && defined(RUSAGE_SELF)
	// System time & CPU time
	getrusage(RUSAGE_SELF, &ST_rusage);	// Ignore error
	ST_utimeMs = ST_rusage.ru_utime.tv_sec * 1000 + ST_rusage.ru_utime.tv_usec / 1000;
	ST_stimeMs = ST_rusage.ru_stime.tv_sec * 1000 + ST_rusage.ru_stime.tv_usec / 1000;
#endif

	ret |= !addline("", buffer, 1);
	ret |= !addline("Statistics", buffer, 1);
	ret |= !addline("----------", buffer, 1);

	for(stt = STatistics; stt->st_value; stt++) {
		sprintf(line, "%10ld  %s", *stt->st_value, stt->st_descr);
		ret |= !addline(line, buffer, 1);

		/* For stats patching - get first line */
		if(stt == STatistics)
			ST_FirstLine = lback(blistp->b_linep);
	}

	return( !ret);
}

/* Append a line to a buffer and give it some contents.
 *
 * str	NUL terminated string contents
 * bp	The buffer to append to
 * u8clean TRUE iftext is good UTF8 and does not need
 *	to be checked. This is to save time by not having to check
 *	everything in the text.
 *
 * The line added is available as: lback(bp->b_linep)
 *
 * Return TRUE  on success.
 *	  FALSE if ran out of memory.
 */
int	addline(const BYTE* str, BUFFER* bp, int u8clean)
{
	TEXT	text;

	return(addLINE(tfakes(&text, str), bp, u8clean));
}

/* Append a line to a buffer and give it some contents.
 *
 * line	LINE contents to append
 * bp	The buffer to append to
 * u8clean TRUE iftext is good UTF8 and does not need
 *	to be checked. This is to save time by not having to check
 *	everything in the text.
 *
 * This essentially does the same thing as addline() but takes a
 * LINE* rather than a string. The point is that a LINE can contain the
 * NUL byte, whereas a NUL byte terminates a string.
 *
 * Return TRUE on success.
 *	  FALSE if ran out of memory.
 */
int	addLINE(const TEXT* text, BUFFER* bp, int u8clean)
{
	LINE*	lp;

	// Buffer not in UTF8 mode - no need to check anyway
	if( !(bp->b_mode & MDUTF8))
		u8clean = 1;

	// The chances are that text is OK, so try to mark it clean
	if( !u8clean && utf8_chkbuf(ttext(text), (size_t)tused(text)))
		u8clean = 1;

	if((lp = lalloc(tused(text) + (u8clean ? 0 : UTF8_REPLACEMENT_LENGTH - 1))) == NULL)
		return(FALSE);

	/* What was that about ?
	 * If not u8clean then we might need to insert a UTF8_REPLACEMENT
	 * Ensure that we have enough room to do that without realloc().
	 */

	/* The line used length will have been set to zero.
	 * This is what insStrLine() expects.
	 */

	/* Insert the string at the start of the newly created line.
	 */
	if(insStrLine(lp, 0, ttext(text), tused(text), u8clean) == -1) {
		FreeLine(lp);
		return(FALSE);
	}

	// Just in case what was inserted contained UTF8 characters, cause
	// any code that looks at the line to take extra care on it.
	if(bp->b_mode & MDUTF8)
		lsetUtf8(lp);

	lsetforw(lback(bp->b_linep), lp);	/* Hook onto the end    */
	lsetback(lp, lback(bp->b_linep));
	lsetback(bp->b_linep, lp);
	lsetforw(lp, bp->b_linep);

	if(bp->b_mode & MDUTF8)		/* If the buffer is UTF8 able */
		lsetUtf8(lp);		/* set the flag on the line - without looking at what we put there */

	bp->b_numLines++;

	return(TRUE);
}

/* Trim the current buffer so that it has at most n lines.
 *
 * If n is positive: remove lines from the bottom of the buffer.
 * If n is negative: remove lines from the top    of the buffer.
 *
 * This MUST be given a count.
 * If it is not given a count it will fail and do nothing.
 *
 * DO NOT optimise the count test away with the N_DEFAULT_1 flag.
 * The purpose of demanding a count is to try to prevent this
 * command being run accidentally.
 *
 * Command:  trim-buffer
 * Bound to: unbound
 *
 * Return: TRUE  Lines might have been deleted
 *	   FALSE command was not called with a count
 */
int	buffertrim(int f, Cnt n)
{
	if(!f) {
		mlwrite("%ETrim-buffer must be given a count. Are you really sure that you want to do this ?");
		return(FALSE);
	}

	if(trimbuffer(curbp, (BufLns)n))
		lchange(WFHARD);

	return(TRUE);
}

/* Trim a buffer so that it has no more than lines lines in it.
 *
 * Any excess are thrown away. The discarded content is not saved anywhere.
 *
 * Count lines from the top of buffer, ie throw away lines at the bottom of the buffer.
 * If lines is negative count from the bottom, ie throw away lines at the top of the buffer.
 *
 * "." and the marks are left as clost as possible to where they were.
 * If the buffer is being viewed then alphanumeric marks may get set
 * to where they were before the buffer was viewed in a window.
 *
 * Returns the number of lines removed.
 */
BufLns	trimbuffer(BUFFER* bp, BufLns lines)
{
	LINE*	head = bp->b_linep;
	LINE*	lp = head;
	LINE*	nxtlp;
	BufLns	ret = 0;

	// From top means count from the top
	int	fromTop = lines > 0 ? 1 : 0;

	if(lines < 0)
		lines = -lines;

	// First line top/bottom:
	lp = fromTop ? lforw(lp) : lback(lp);

	// Find the point at which we start trimming:
	while(--lines >= 0) {
		nxtlp = fromTop ? lforw(lp) : lback(lp);
		// Nothing left - ie buffer not that long ?
		if(nxtlp == head)
			return(ret);

		lp = nxtlp;
	}

	// Remove from that point until we meet the head
	while(lp != head) {
		nxtlp = fromTop ? lforw(lp) : lback(lp);
		lfree(bp, lp);
		bp->b_numLines--;
		lp = nxtlp;
		ret++;
	}

	/* This is an edge case, but important.
	 * If the history buffer is being viewed then "." should stay on the same line.
	 * However the buffer's version will be elsewhere - probably the last line, (it
	 * is wherever it was when last synched with a window).
	 * So: reset it to where the window shows it.
	 * Also copy over the mark & numeric marks.
	 *
	 * We must do this as ReconnectWindowsToBuffer() sets window's "." to buffer's "."
	 *
	 * None of this happens 99.9% of the time as history buffers are rarely viewed.
	 */
	if(bp->b_nwnd != 0) {
		Window* wp = onscreen(bp);
		if(wp)
			SaveWindowPosToBuffer(bp, wp);
	}

	/* This will destroy the framing (which is recalculated anyway), but if
	 * this is not done then line numbers display may/will be wrong.
	 */
	bp->b_toplinenum = 1;
	bp->b_topLinep = lforw(bp->b_linep);

	// Only really needed if it is displayed in a window:
	ReconnectWindowsToBuffer(bp);

	return(ret);
}

/* Append a string to a history buffer, then trim the buffer
 * to $message-history lines (SaveMlHist) - but more if [Message History].
 *
* bp		Buffer to append to
 * str		NUL terminated string
 * flags	Bits that can be ORed:
 *		AHL_NO_DUP
 *			do not add if line is the same as the last line in the buffer.
 *			Useful for things like search where the same string will be looked for time and again.
 *		AHL_ADD_ZLEN
 *			Add even if zero length
 *
 * If the line (string) is empty (zero length) do not add it.
 * Only want interesting things in history.
 *
 * The newly added line will become the history buffer's ".", ditto for any
 * windows unless it is displayed in the current window.
 *
 * Return:
 *	FALSE on error, prob out of memory
 *	True on success, TRUE being one of:
 *		1 - line added
 *		2 - a duplicate was found, thus no line added
 *		3 - zero length line. Well, not added but not an error here
 */
int	addHistoryLine(BUFFER* bp, const BYTE* str, int flags)
{
	// This works by making a LINE and adding that
	TEXT	text;

	return(addHistoryText(bp, tfakes(&text, str), flags));
}

/* This essentially does the same thing as addHistoryLine() but takes a
 * TEXT* rather than a string. The point is that a TEXT can contain the
 * NUL byte, whereas a NUL byte terminates a string.
 *
 * The newly added line will become the history buffer's "." - unless
 * the history buffer is the current window.
 *
 * bp		Buffer to append to
 * text		TEXT containing the message
 * flags	Bits that can be ORed:
 *		AHL_NO_DUP
 *			do not add if line is the same as the last line in the buffer.
 *			Useful for things like search where the same string will be looked for time and again.
 *		AHL_ADD_ZLEN
 *			Add even if zero length
 *
 * Return:
 *	FALSE on error, prob out of memory
 *	True on success, TRUE being one of:
 *		1 - line added
 *		2 - a duplicate was found, thus no line added
 *		3 - zero length line. Well, not added but not an error here
 */
int	addHistoryText(BUFFER* bp, const TEXT* text, int flags)
{
	LINE*	nl;	// The new line

	int	trimLines = SaveMlHist;	// How many lines to trim the buffer to
	if(bp == bmlHistp)
		trimLines *= 10;	// Many more for the message buffer

	if((flags & AHL_ADD_ZLEN) == 0 && tused(text) == 0)
		return(3);

	if((flags & AHL_NO_DUP) && ! textcmp(text, &ltxt(lback(bp->b_linep))))
		return(2);	// Duplicate found

	if( !addLINE(text, bp, 0))	// Can't be sure where the text/line is coming from
		return(FALSE);

	trimbuffer(bp, -trimLines);

	// Find the line, set the NUL flag:
	nl = lback(bp->b_linep);
	if( !lsetNUL(nl))
		return(FALSE);

	/* If the buffer is displayed in a window, ensure that the window is updated */
	if(bp->b_nwnd != 0)
		upmode(bp, WFHARD);

	return(1);
}

/* Are any buffers changed - ie modified compared to the file on disk ?
 *
 * Look through the list of buffers. Return the number of changed buffers,
 * which is TRUE if there are any changed buffers.
 *
 * Buffers that hold magic internal stuff are not considered; who cares if the
 * list of buffer names is hacked ?
 *
 * Return the number of buffers that have been changed.
 * This will be 0 if no buffers have been changed.
 */
int	anycb(void)
{
	register BUFFER* bp;
	register int	 changed = 0;

	bp = bheadp;
	while(bp != NULL) {
		if((bp->b_flag & BFINVS) == 0 && (bp->b_flag & BFCHG) != 0)
			changed++;

		bp = bp->b_bufp;
	}

	return(changed);
}

#if	UNIX
/* Clear a buffer's UNIX inode information to the state of a file that does not
 * exist.
 */
void	ClearInodeInfo(BUFFER* bp)
{
	bp->b_lsType = '.';
	bp->b_mtime = (Time_T) -1;
	bp->b_stdev = (dev_t) -1;
	bp->b_stino = (ino_t) -1;
}
#endif

/* Fix buffer's line count, "." and topLine.
 *
 * Only needed when lunstitch() used outside lfree() or ldelnewline().
 * You should NOT want to do this. Only currently used in help routines.
 */
void	bLinesFix(BUFFER* bp)
{
	LINE*	line = lforw(bp->b_linep);

	bp->b_topLinep = bp->b_dotp = line;
	bp->b_doto = 0;
	bp->b_toplinenum = 1;

	bp->b_numLines = 0;
	for( ; line != bp->b_linep; line = lforw(line))
		bp->b_numLines++;
}

/* Find a buffer by name.
 *
 * bname	Buffer name
 * cflag	Flags that can be ORed:
 *   BFNNONE	Nothing special
 *   BFNCREATE	If the buffer is not found then create it
 *   BFNACTIVE	Mark a newly created buffer as active
 *   BFNVIEW	Make the buffer read only
 * bflag	Bxx values for the buffer's b_flag
 *
 * The buffer is marked b_active and b_mode set to gmode.
 *
 * Return a pointer to the BUFFER structure associated with it.
 * 	NULL if not found & !cflag.
 *	     if out of memory
 */
BUFFER*	bfind(const BYTE* bname, int cflag, int bflag)
{
	register BUFFER*bp;
	register BUFFER*sb;	/* buffer to insert after */
	register LINE*	lp;

	/* Find existing one - if it exists */
	for(bp = bheadp; bp != NULL; bp = bp->b_bufp)
		if(strcmp(bname, bp->b_bname) == 0)
			return(bp);

	if((cflag & BFNCREATE) != 0) {
		if((bp = (BUFFER*)calloc(1, sizeof(BUFFER))) == NULL)
			return(NULL);

		if((lp = lalloc(0)) == NULL) {
			free((void*) bp);
			return(NULL);
		}

		// 1 as will lsetNUL()
		if((bp->b_key = talloc(1)) == NULL) {
			FreeLine(lp);
			free((void*)bp);
			return(NULL);
		}

		/* Find the place in the list to insert this buffer.
		 */
		if(bheadp == NULL
#if	ALPHORD
		|| strcmp(bheadp->b_bname, bname) > 0
#endif
							) {
			/* insert at the begining */
			bp->b_bufp = bheadp;
			bheadp = bp;
		} else {
			sb = bheadp;
			while(sb->b_bufp != NULL) {
				/*
				 * If not ordering alphabetically then just
				 * put the new buffer at the end of the list.
				 */
#if	ALPHORD
				if(strcmp(sb->b_bufp->b_bname, bname) > 0)
					break;
#endif
				sb = sb->b_bufp;
			}

			/* and insert it */
			bp->b_bufp = sb->b_bufp;
			sb->b_bufp = bp;
		}

		/* and set up the other buffer fields */
		bp->b_topLinep = lp;
		bp->b_toplinenum = 1;
		bp->b_active = (cflag & BFNACTIVE) != 0;
		bp->b_dotp  = lp;
		// Comment out as calloc() NULs it
		// bp->b_doto  = 0;
		// bp->b_markp = NULL;
		// bp->b_marko = 0;
		// bp->b_numLines = 0;
		bp->b_flag  = (short)(bflag | BFIMPLICIT);	/* Nowt there yet */
		bp->b_mode  = gmode;		// Checked below
		if((cflag & BFNVIEW) != 0)
			bp->b_mode |= MDVIEW;

		// New buffer, new (default) set of window options:
		bp->b_winopts = DefaultWindowOpts;
		bp->b_tabsize = Tabsize;

		// bp->b_nwnd  = 0;
		// bp->b_exec  = 0;
		bp->b_linep    = lp;
		bp->b_btopline = lp;
		bp->b_abotline = lp;
		// bp->b_amark = (AMARK*)NULL;
		// bp->b_nmarkOff = 0;
		// memset((void*)bp->b_nmarks, '\0', sizeof(bp->b_nmarks));

#if	UNIX
		ClearInodeInfo(bp);
#endif
		// bp->b_fname = bp->b_fnameClean = NULL;
		setFnames(bp, "");
		bp->b_bname = utf8_strmake_t(bname);
#if	CRYPT
		tsetNUL(bp->b_key);
		tsetused(bp->b_key, 0);
#endif
		lsetforw(lp, lp);
		lsetback(lp, lp);

		chkmode(bp);		// Mode consistency checking
	}

	return(bp);
}

/* As bfind() but if there is a failure a message is generated.
 *
 * bname	Buffer name
 * cflag	Bits as bfind()
 *
 * Return
 *	the buffer pointer
 *	NULL if not found & !cflag
 *	     out of memory
 */
BUFFER*	bfindCheck(BYTE* bname, int cflag)
{
	BUFFER*	bp;

	if((bp = bfind(bname, cflag, 0)) == (BUFFER*)NULL)
		mlwrite((cflag ? "%ECannot create buffer %s" : "%ECannot find buffer %s"),
			bname);

	return(bp);
}

/* Find buffer, read in if not there, do not leave curbp
 *
 * bname	The name of the buffer to create
 * fname	The name of the file to read
 *		If this is NULL use b_fname - only with existing buffer
 * 
 * Either bname or fname can be NULL - not both
 * Use bname - which should give a fname
 *
 * The buffer will be marked VIEW
 * 
 * Return
 *	Pointer to the buffer
 *	NULL - cannot allocate buffer or cannot read file
 */
BUFFER*	bfindFile(BYTE* bname, BYTE* fname)
{
	int	create = BFNVIEW;
	BUFFER*	bp;

	if(bname == NULL) {
		bname = makename(fname);
		create = BFNCREATE|BFNVIEW;	// No to be marked active
	}

	if((bp = bfindCheck(bname, create)) == NULL)
		return((BUFFER*)NULL);

	/* That might have create a new buffer or found an existing one.
	 * An existing buffer might not have been read in (be active).
	 */
	if(bp->b_active != TRUE) {
		if(fname == NULL)
			fname = bp->b_fname;

		BUFFER*	savecur = curbp;
		curbp = bp;
		//do I need a full connectbuffer ?
		int	res = readin(fname, FALSE, FALSE, TRUE);
		curbp = savecur;

		if(res == FALSE) {
			bp = NULL;
			mlwrite("%ECannot read %s as %e", fname, errno);
		}
	}

	return(bp);
}


/* Is it OK to discard a buffer's contents ?
 *
 * Buffer bp might have some unsaved changes.
 * If the buffer is invisible - the contents are regarded as disposible.
 *
 * Return TRUE if it is OK to discard the buffer contents.
 *		Check: returns TRUE on invisible buffer ? TODO
 *        FALSE on error
 *	  ABORT if prompt is aborted
 */
int	bDiscardOK(const BUFFER* bp)
{
	if((bp->b_flag & BFINVS) ||		/* Scratch buffer ?	*/
	   (bp->b_flag & BFCHG) == 0)		/* Something changed ?	*/
		return(TRUE);

	// Check the file name, mught be empty when clean one is not (command)
	return(mlyesno((bp->b_fname[0] ?
		    "%h%s: Discard changes (file %s) [y/n]? " :
		    "%h%s: Discard changes [y/n]? "),
			bp->b_bname, bp->b_fnameClean));
}

/* Blow away all marks in a buffer.
 * The mark, alpha marks, numeric marks.
 *
 * This does not clear away numeric marks that may be stored in a Window.
 *
 * Do this when zapping a buffer or clearing out so can put something new there.
 */
void	bclearMarks(BUFFER* bp)
{
	bp->b_markp      = NULL;		/* Invalidate "mark"	*/
	bp->b_marko      = 0;

	// Alpha marks
	while(bp->b_amark) {
		AMARK*	next = bp->b_amark->m_next;
		free(bp->b_amark);
		bp->b_amark = next;
	}

	// Numeric marks
	memset((void*)bp->b_nmarks, '\0', sizeof(bp->b_nmarks));
}

/* This routine blows away all of the text in a buffer.
 *
 * If the buffer is marked as changed then we ask if it is ok to blow it away;
 * this is to save the user the grief of losing text. The window chain is
 * nearly always wrong if this gets called; the caller must arrange for the
 * updates that are required.
 *
 * Return TRUE if everything looks good.
 *        FALSE on error
 *	  ABORT if prompt is aborted
 */
int	bclear(BUFFER* bp)
{
	register LINE*	lp;
	register int	s;

	if((s = bDiscardOK(bp)) != TRUE)
		return(s);

	// Do before lfree()s - makes it easier/quicker
	bp->b_topLinep   = bp->b_linep;
	bp->b_btopline   = bp->b_linep;		// Widen the buffer
	bp->b_abotline   = bp->b_linep;
	bp->b_toplinenum = 1;
	bp->b_dotp       = bp->b_linep;		/* Fix "."		*/
	bp->b_doto       = 0;
	bp->b_numLines   = 0;
	bclearMarks(bp);

	bp->b_flag &= ~(BFCHG | BRENEWED);	/* Not changed		*/
	while((lp = lforw(bp->b_linep)) != bp->b_linep)
		lfree(bp, lp);

	ClearProcInfo(bp);

	return(TRUE);
}

/* Unmark the current buffer's changed flag
 *
 * The count is ignored. There is not a default count.
 *
 * Command:  unmark-buffer
 * Bound to: M-~
 *
 * Return: TRUE always, this cannot fail.
 */
/*ARGSUSED*/
int	unmark(int f, Cnt n)
{
	curbp->b_flag &= ~(BFCHG | BFROMSAVE);
	curwp->w_flag |= WFMODE;

	upmode(NULL, WFMODE);

	return(TRUE);
}

/* To be called when the file name is changed.
 * 
 * This updates both b_fname and the UTF8 clean b_fnameClean.
 * It is hoped that both copies of the file name are the same, if they are
 * different it is because the name has something very strange in it.
 *
 * Returns the name - a version that is valid; it is quite possible
 * that the original string gets trashed (ie freed).
 * This is why on the 2nd line below use b_fname and not fname.
 */
CHAR*	setFnames(BUFFER* bp, const CHAR* fname)
{
	strfmake(&bp->b_fname, fname);
	utf8_strfmake(&bp->b_fnameClean, bp->b_fname);

	return(bp->b_fname);
}

/* Set a file name to be shown for the buffer - but only for show.
 *
 * The point is that some buffers (eg Help, command) have a name
 * to hint the user what the contents are, but we do not want to use
 * this as a file name to read/write files.
 *
 * So: set b_fnameClean but not b_fname.
 */
void	setVisibleFname(BUFFER* bp, const CHAR* fname)
{
	setFnames(bp, fname);
	bp->b_fname[0] = '\0';
}

/* Manipulation of DOS buffer mode
 *
 * Look to see if every line in the buffer has a trailing \r (RETCHAR)
 *
 * Return a RETC_??? value, one of: RETC_NONE RETC_ALL RETC_SOME RETC_EMPTY
 */
int	IsDosBuffer(const BUFFER* bp)
{
	LINE*	lp;
	int	some = 0;
	int	someNot = 0;

	for(lp = lforw(bp->b_linep); lp != bp->b_linep; lp = lforw(lp))
		if(lused(lp) && lgetc(lp, lused(lp) - 1) == RETCHAR)
			some = RETC_WITH;
		else
			someNot = RETC_WITHOUT;

	return(some | someNot);
}

/* Remove a \r (RETCHAR) from the end of every line in the buffer.
 * If the last character is not \r don't touch that line.
 *
 * Removal is quite simple: shorten the used length by 1.
 * Then check/fix cursor & mark positions that might have been after the \r.
 * This could be called with a non displayed buffer [think: auto renew],
 * so fix in buffer as well as windows.
 *
 * This will mark windows that contain buffers that change as needing
 * refreshing; it will not mark buffers as changed [think: buffer just
 * read in & \r auto removed].
 *
 * Return one of: RETC_NONE RETC_ALL RETC_SOME RETC_EMPTY
 */
int	RemoveRetchar(BUFFER* bp)
{
	LINE*	lp;
	int	some = 0;	// Some lines changed
	int	someNot = 0;	// Some lines not changed

	for(lp = lforw(bp->b_linep); lp != bp->b_linep; lp = lforw(lp))
		if(lused(lp) && lgetc(lp, lused(lp) - 1) == RETCHAR) {
			lsetused(lp, lused(lp) - 1);
			some = RETC_WITH;
		} else
			someNot = RETC_WITHOUT;

	/* It is possible that the cursor or a mark was on a virtual newline.
	 * That has moved left once page because we have just zapped the RETCHAR.
	 * If that is the case we must move it so that it is within the line.
	 *
	 * Beware: it is possible that a \r was not removed from some particular
	 * line and that that line is empty (lused() == 0), do not make
	 * w_doto (or something) negative.
	 *
	 * This is a lot of code for something that is probably infrequent.
	 */
	if(some) {
		Window*	wp;
		AMARK*	amp;

		// Windows on this buffer:
		for(wp = wheadp; wp != NULL; wp = wp->w_wndp) {
			int	changed = 0;

			if(wp->w_bufp != bp)
				continue;

			if(wp->w_doto >= lused(wp->w_dotp) && wp->w_doto > 0) {
				wp->w_doto--;
				changed = 1;
			}

			if(wp->w_markp && wp->w_marko >= lused(wp->w_markp) && wp->w_marko > 0) {
				wp->w_marko--;
				changed = 1;
			}

			if(changed)
				wp->w_flag |= WFHARD;
		}

		// Buffer lines
		if(bp->b_doto >= lused(bp->b_dotp) && bp->b_doto > 0)
			bp->b_doto--;

		if(bp->b_markp && bp->b_marko >= lused(bp->b_markp) && bp->b_marko > 0)
			bp->b_marko--;

		// Alpha marks:
		for(amp = bp->b_amark; amp; amp = amp->m_next)
			if(amp->m_pline && amp->m_oline >= lused(amp->m_pline) && amp->m_oline > 0)
				amp->m_oline--;
	}

	return(some | someNot);
}

/* Add a \r onto the end of every line in the buffer.
 * If there is already one there, do not add another.
 *
 * This might need extending the allocation of the line.
 *
 * Note: RemoveRetchar() then AddRetchar() might not yeild the starting
 * buffer: if \r is not there, then it cannot be removed. If there is already
 * one there then another will not be added.
 * This probably does what most people want.
 *
 * See comments in RemoveRetchar().
 *
 * Return a RETC_??? value.
 */
int	AddRetchar(BUFFER* bp)
{
	LINE*	lp;
	int	some = 0;	// Some lines changed
	int	someNot = 0;	// Some lines not changed

	for(lp = lforw(bp->b_linep); lp != bp->b_linep; lp = lforw(lp))
		// Make the line bigger if needs be
		if( !lused(lp) || lgetc(lp, lused(lp) - 1) != RETCHAR) {
			if(lused(lp) == lsize(lp) && !lineMoreBytes(lp, lused(lp) + 1))
				return(RETC_ERROR);

			lputc(lp, lused(lp), RETCHAR);
			lsetused(lp, lused(lp) + 1);	// Test just above that it can take it
			some = RETC_WITH;
		} else
			someNot = RETC_WITHOUT;

	/* If the cursor was at the end of line now it will not be, it will
	 * be on a \r. This is prob not what the user expects, so move it back
	 * to the end of line.
	 * This could be the wrong thing is someNot, but we have no idea
	 * which lines were not done - too complicated to worry about for a
	 * case that is probably rare. Follow the principle of least expected
	 * user surprise.
	 *
	 * This is a lot of code for something that is probably infrequent.
	 */
	if(some) {
		Window*	wp;
		AMARK*	amp;

		// Windows on this buffer:
		for(wp = wheadp; wp != NULL; wp = wp->w_wndp) {
			int	changed = 0;

			if(wp->w_bufp != bp)
				continue;

			if(wp->w_doto == lused(wp->w_dotp) - 1) {
				wp->w_doto++;
				changed = 1;
			}

			if(wp->w_markp && wp->w_marko == lused(wp->w_markp) - 1) {
				wp->w_marko++;
				changed = 1;
			}

			if(changed)
				wp->w_flag |= WFHARD;
		}

		// Buffer lines
		if(bp->b_doto == lused(bp->b_dotp) - 1)
			bp->b_doto++;

		if(bp->b_markp && bp->b_marko == lused(bp->b_markp) - 1)
			bp->b_marko++;

		// Alpha marks:
		for(amp = bp->b_amark; amp; amp = amp->m_next)
			if(amp->m_pline && amp->m_oline == lused(amp->m_pline) - 1)
				amp->m_oline++;
	}

	return(some | someNot);
}

/* Check that a buffer should have a DOS mode change.
 *
 * Called before un/setting DOS mode.
 * bp	  the BUFFER to check/set.
 * set 	  true if setting the mode
 * global true if setting the global flag - else the current buffer
 *
 * Always allow setting of the global flag
 *
 * If given a non zero count allow but do not transform the buffer.
 * This makes it easy to read a DOS formatted file and write it back
 * without those pesky \r on every line.
 *
 * Return true if it is OK to do the action.
 */
int	dos_modeset(BUFFER* bp, int set, int global, int f, Cnt n)
{
	int	status;

	// No prob setting a global 'desire' to dos
	if(global)
		return(TRUE);

	// If a non zero argument always allow:
	if(f && n)
		return(TRUE);

/* Now allow DOS mode in VIEW mode. Although it technically changes the buffer,
 * the change is predictable and easy to undo. It changes the buffer, not the
 * file - the user will not be able to write back, if they do the effect of DOS
 * mode is undone.
 */
// 	if(bp->b_mode & MDVIEW) {
// 		mlwrite("%WCannot change in read only buffer");
// 		return(FALSE);
// 	}

	if(set) {
		status = RemoveRetchar(bp);
	} else {
		if((status = AddRetchar(bp)) == RETC_ERROR)
			mlwrite("%EBuffer not completely converted");
	}

	/* Did that change the buffer ?
	 * WFMODE set in calling routine.
	 */
	if(status == RETC_ERROR || status == RETC_ALL || status == RETC_SOME)
		bp->b_flag |= BFCHG;

	// Always say 'yes' to the change, even if no lines were changed.
	// Only 'no' on error
	return(status != RETC_ERROR);
}

/* Set/unset the RENEW mode on the current buffer - curbp.
 * 
 * Called before un/setting RENEW mode.
 * bp	  the BUFFER to check/set.
 * set 	  true if setting the mode
 * global true if setting the global flag - else the current buffer
 *
 * Do not allow setting of the flag if the buffer has been modified.
 *
 * If RENEW mode is set, VIEW mode is also set.
 *
 * This clears BRENEWED - our record that the buffer has been auto renewed.
 * This is done regardless of setting/unsetting RENEW mode.
 *
 * Unlike other modeset functions this actually un/sets the mode, this
 * is because CheckRenewVisible() is called.
 *
 * If the count is positive a renew will assume that the file has been
 * appended to and just read the new lines; ie it will not reread the
 * entire file.
 *
 * Return true as it is OK to do the action.
 */
int	renew_modeset(BUFFER* bp, int set, int global, int f, Cnt n)
{
	if(global)		// We only worry about specific buffers.
		return(TRUE);

	bp->b_flag &= ~(BRENEWED);	// Forget that we have auto-renewed it

	if(set) {
		// Not on a modified buffer:
		if(bp->b_flag & BFCHG) {
			mlwrite("%ECannot set RENEW mode on a modified buffer");
			return(FALSE);
		}
		bp->b_mode |=  MDRENEW | MDVIEW;
		if(f && n > 0)
			bp->b_flag |=  BRENEWAPPEND;
		else
			bp->b_flag &= ~BRENEWAPPEND;

		// Setting RENEW, read in if it is out of date:
		if(set) {
			char*	act = "renewed";
			switch(CheckBufferRenew(bp)) {
			  case 1: // Unlikely, but maybe possible:
				act = "out of date";
				// DROP THROUGH
			  case 2:
				mlwrite("%IBuffer %s %s", bp->b_bname, act);

			  // case 0 - not out of date - say now't
			}
		}

	} else {
		// Can always unset
		bp->b_mode &= ~MDRENEW;
	}

	CheckRenewVisible();

	return(TRUE);
}

/* Check if any visible buffers needs renewing.
 *
 * If one is in RENEW mode - set RenewBufferIsVisible TRUE.
 */
void	CheckRenewVisible(void)
{
	BUFFER*	bp;

	for(bp = bheadp; bp != (BUFFER*) NULL; bp = bp->b_bufp)
		if((bp->b_mode & MDRENEW) != 0 && bp->b_nwnd != 0) {
			RenewBufferIsVisible = 1;
			return;
		}

	RenewBufferIsVisible = 0;
}


/* Set/unset C mode on the current buffer - curbp.
 *
 * Called before un/setting C mode.
 * bp	  the BUFFER to check/set.
 * set 	  true if setting the mode
 * global true if setting the global flag - else the current buffer
 *
 * When setting this unsets PHP & INDENT modes
 * The tab size is set to default.
 *
 * Return true as it is OK to do the action.
 */
int	c_modeset(BUFFER* bp, int set, int global, int f, Cnt n)
{
	int*		modep;		/* Which mode flags we are changing */

	modep = global ? &gmode : &bp->b_mode;

	if(set) {
		*modep &= ~(MDINDEN | MDPHP);
		bp->b_tabsize = Tabsize;
	}

	return(TRUE);
}

/* Set/unset PHP mode on the current buffer - curbp.
 *
 * Called before un/setting PHP mode.
 * bp	  the BUFFER to check/set.
 * set 	  true if setting the mode
 * global true if setting the global flag - else the current buffer
 *
 * When setting this unsets C & INDENT modes & sets TABEXPAND
 * The tab size is set to PhpTabsize
 *
 * Return true as it is OK to do the action.
 */
int	php_modeset(BUFFER* bp, int set, int global, int f, Cnt n)
{
	int*		modep;		/* Which mode flags we are changing */

	modep = global ? &gmode : &bp->b_mode;

	if(set) {
		*modep &= ~(MDINDEN | MDCMOD);
		*modep |= MDTABEXP;
		bp->b_tabsize = PhpTabsize;
	} else {
		bp->b_tabsize = Tabsize;	// Revert to default
	}

	return(TRUE);
}

/* Set/unset INDENT mode on the current buffer - curbp.
 *
 * Called before un/setting INDENT mode.
 * bp	  the BUFFER to check/set.
 * set 	  true if setting the mode
 * global true if setting the global flag - else the current buffer
 *
 * When setting this unsets C & PHP modes
 *
 * Return true as it is OK to do the action.
 */
int	indent_modeset(BUFFER* bp, int set, int global, int f, Cnt n)
{
	int*		modep;		/* Which mode flags we are changing */

	modep = global ? &gmode : &bp->b_mode;

	if(set)
		*modep &= ~(MDPHP | MDCMOD);

	return(TRUE);
}

/* end */
