/*      /\
 *     /  \		(C) Copyright 2018 Parliament Hill Computers Ltd.
 *     \  /		All rights reserved.
 *      \/
 *       .		Author: Alain Williams, July 2018.
 *       .
 *        .		SCCS: @(#)charconv.c	1.6 09/22/20 22:44:11
 *          .
 *
 * Code that is to do with case convertion of characters, eg to
 * upper case, lower case, title case.
 *
 * There is code elsewhere that does similar jobs. Eventually try to
 * consolidate here - for reasons as consistency as much as code size
 */

#include	"estruct.h"
#include        "utf8.h"
#include        "edef.h"

#if	SHOWMODREL
char	charcvers[] = "@(#)charconv.c	1.6 09/22/20 22:44:11";
#endif

/* Copy from from to to converting alphabetics to lower case.
 * TODO Unicode.
 *
 * to which is modified, so the return value can be used as success boolen.
 *
 * Return to.
 *	NULL on out of memory.
 *	NULL is returned if textOOM - which might have been set earlier.
 */
TEXT*	textToLower(TEXT* to, const TEXT* from)
{
	int	len = (int)tused(from);
	int	i;
	BYTE	c;

	tsetempty(to);

	// Presize so not needed, possibly more than once, in lcatm()
	if( !textMoreBytes(to, len))
		return(NULL);

	for(i = 0; i < len; i++) {
		c = tgetch(from, i);
		// Convert all to lower
		if(Isupper(c))
			c = (BYTE)tolower(c);
		tcatm(to, &c, 1);
	}

	if(textOOM)
		return((TEXT*)NULL);

	return(to);
}

/* Copy from from to to converting alphabetics to upper case.
 * TODO Unicode.
 *
 * to which is modified, so the return value can be used as success boolen.
 *
 * Return to.
 *	NULL on out of memory.
 *	NULL is returned if textOOM - which might have been set earlier.
 */
TEXT*	textToUpper(TEXT* to, const TEXT* from)
{
	int	len = (int)tused(from);
	int	i;
	BYTE	c;

	tsetempty(to);

	// Presize so not needed, possibly more than once, in lcatm()
	if( !textMoreBytes(to, len))
		return(NULL);

	for(i = 0; i < len; i++) {
		c = tgetch(from, i);
		// Convert all to UPPER
		if(Islower(c))
			c = (BYTE)toupper(c);
		tcatm(to, &c, 1);
	}

	if(textOOM)
		return((TEXT*)NULL);

	return(to);
}

/* Copy from from to to converting start of word alphabetics to title case.
 * Start of word is: start of text or after a space or tab.
 * TODO Unicode.
 *
 * to which is modified, so the return value can be used as success boolen.
 *
 * Return to.
 *	NULL on out of memory.
 *	NULL is returned if textOOM - which might have been set earlier.
 */
TEXT*	textToTitle(TEXT* to, const TEXT* from)
{
	int	len = (int)tused(from);
	int	first = 1;		// Next alpha is first of word
	int	i;
	BYTE	c;

	tsetempty(to);

	// Presize so not needed, possibly more than once, in lcatm()
	if( !textMoreBytes(to, len))
		return(NULL);

	for(i = 0; i < len; i++) {
		c = tgetch(from, i);

		// Convert all first-alpha-of-word to upper
		if(c == ' ' || c == '\t') {
			first = 1;
		} else {
			if(first && Isletter(c)) {
				if(Islower(c))
					c = (BYTE)toupper(c);
				first = 0;	// Not first letter any more
			}
		}

		tcatm(to, &c, 1);
	}

	if(textOOM)
		return((TEXT*)NULL);

	return(to);
}



/* end */
