/*
 * ADDW SCCS: @(#)crypt.c	1.17 08/06/20 00:09:15 - second
 */

/*	Crypt:	Encryption routines for MicroEMACS
		written by Dana Hoggatt and Daniel Lawrence
*/

#include	"estruct.h"
#include	"utf8.h"
#include	"edef.h"

#if	SHOWMODREL
char	cryptvers[] = "@(#)crypt.c	1.17 08/06/20 00:09:15";
#endif

#if	CRYPT

/* Below is the code from Emacs 5, ifdefed with EMACS5_CRYPT
 *
 * This is described in a comment dated 2-dec-88:
 *    fixed bug in crypt() algorith!!!! This works properly now.
 * No indication as to what the problem was.
 * The Emacs 5 code has not been tested here.
 *
 * This is included as someone might know better.
 *
 * However: a limitation with this code is that it only encrypts US-ASCII printable
 * characters. This gives crackers clues because they can see the newlines and
 * tabs (although tabs might not be so much of a problem). It also leave anything
 * U+7F & above in clear - ie all multibyte Unicode characters; this is bad.
 * This routine needs to be replaced - possibly with compatability flags
 * so that 'old' files can be decrypted.
 */
#define	EMACS5_CRYPT	1		// Use code from Emacs 5

static	int	mod95(int val);

/* Reset encryption key of current buffer.
 *
 * If f (any non-zero value of n) the key will be verified, ie the user asked
 * to retype to confirm and the two values compared.
 *
 * If this is run from a macro the key will be taken from the macro buffer
 * (hopefully kept in a variable, not as a literal string) and the verification
 * will not be done.
 *
 * If run from a macro and the count is set (any value including zero) then the
 * user will be prompted for the key, ie it will not be read from the macro
 * buffer. The test for verification (count being non-zero) is done as above.
 *
 * Any verification is always interactive.
 *
 * No default count.
 *
 * Command:  set-encryption-key
 * Bound to: M-e
 *
 * Return: TODO
 */
int	crysetkey(int f, Cnt n)
{
	register int status;	/* return status */
	char*	prompt = "Encryption String: ";
	int	intFlag = 0;	// Prompt must be interactive

	// Should the prompt for the key always be interactive:
	if(clexec == FALSE || (f && n != 0))
		intFlag = MLINTERACTIVE;

	while(1) {
		/* Get the string to use as an encrytion string.
		 * This is left in playCrypt:
		 */
		if((status = mlprompt(prompt, MLSTRING | MLNOECHO | intFlag, bmlPromptHistp)) != TRUE)
			return(status);

		// Copy to where it is used below:
		tcpyt(playCrypt2, playCrypt);

		if(!f || n == 0)
			break;		// Verification not wanted

		// Verification always interactive:
		if((status = mlprompt("Verify Encryption String: ", MLSTRING | MLNOECHO | MLINTERACTIVE, bmlPromptHistp)) != TRUE)
			return(status);

		if(strcmp(ttext(playCrypt), ttext(playCrypt2)) == 0)
			break;

		prompt = "Encryption Strings differ, try again: ";
	}

	/* and encrypt it */
	cryptBuffer((char *)NULL, 0);
	cryptBuffer(ttext(playCrypt2), tused(playCrypt2));

	/* and save it off */
	tcpyt(curbp->b_key, playCrypt2);

	tsetused(playCrypt, 0);
	tsetused(playCrypt2, 0);

	mlerase();		/* clear it off the bottom line */
	return(TRUE);
}

/**********
 *
 *	crypt - in place encryption/decryption of a buffer
 *
 *	(C) Copyright 1986, Dana L. Hoggatt
 *	1216, Beck Lane, Lafayette, IN
 *
 *	When consulting directly with the author of this routine, 
 *	please refer to this routine as the "DLH-POLY-86-B CIPHER".  
 *
 *	This routine was written for Dan Lawrence, for use in V3.8 of
 *	MicroEMACS, a public domain text/program editor.  
 *
 *	I kept the following goals in mind when preparing this function:
 *
 *	    1.	All printable characters were to be encrypted back
 *		into the printable range, control characters and
 *		high-bit characters were to remain unaffected.	this
 *		way, encrypted would still be just as cheap to 
 *		transmit down a 7-bit data path as they were before.
 *
 *	    2.	The encryption had to be portable.  The encrypted 
 *		file from one computer should be able to be decrypted 
 *		on another computer.
 *
 *	    3.	The encryption had to be inexpensive, both in terms
 *		of speed and space.
 *
 *	    4.	The system needed to be secure against all but the
 *		most determined of attackers.
 *
 *	For encryption of a block of data, one calls crypt passing 
 *	a pointer to the data block and its length. The data block is 
 *	encrypted in place, that is, the encrypted output overwrites 
 *	the input.  Decryption is totally isomorphic, and is performed 
 *	in the same manner by the same routine.  
 *
 *	Before using this routine for encrypting data, you are expected 
 *	to specify an encryption key.  This key is an arbitrary string,
 *	to be supplied by the user.  To set the key takes two calls to 
 *	cryptBuffer().  First, you call 
 *
 *		cryptBuffer(NULL, vector)
 *
 *	This resets all internal control information.  Typically (and 
 *	specifically in the case on MICRO-emacs) you would use a "vector" 
 *	of 0.  Other values can be used to customize your editor to be 
 *	"incompatable" with the normally distributed version.  For 
 *	this purpose, the best results will be obtained by avoiding
 *	multiples of 95.
 *
 *	Then, you "encrypt" your password by calling 
 *
 *		cryptBuffer(pass, strlen(pass))
 *
 *	where "pass" is your password string.  cryptBuffer() will destroy 
 *	the original copy of the password (it becomes encrypted), 
 *	which is good.	You do not want someone on a multiuser system 
 *	to peruse your memory space and bump into your password.  
 *	Still, it is a better idea to erase the password buffer to 
 *	defeat memory perusal by a more technical snooper.  
 *
 *	For the interest of cryptologists, at the heart of this 
 *	function is a Beaufort Cipher.	The cipher alphabet is the 
 *	range of printable characters (' ' to '~'), all "control" 
 *	and "high-bit" characters are left unaltered.
 *
 *	The key is a variant autokey, derived from a wieghted sum 
 *	of all the previous clear text and cipher text.  A counter 
 *	is used as salt to obiterate any simple cyclic behavior 
 *	from the clear text, and key feedback is used to assure 
 *	that the entire message is based on the original key, 
 *	preventing attacks on the last part of the message as if 
 *	it were a pure autokey system.
 *
 *	Overall security of encrypted data depends upon three 
 *	factors:  the fundamental cryptographic system must be 
 *	difficult to compromise; exhaustive searching of the key 
 *	space must be computationally expensive; keys and plaintext 
 *	must remain out of sight.  This system satisfies this set
 *	of conditions to within the degree desired for MicroEMACS.
 *
 *	Though direct methods of attack (against systems such as 
 *	this) do exist, they are not well known and will consume 
 *	considerable amounts of computing time.  An exhaustive
 *	search requires over a billion investigations, on average.
 *
 *	The choice, entry, storage, manipulation, alteration, 
 *	protection and security of the keys themselves are the 
 *	responsiblity of the user.
 *
 * ADDW notes:
 *
 * https://en.wikipedia.org/wiki/Beaufort_cipher
 *
 * https://www.dcode.fr/beaufort-cipher  has something that will break it.
 * Beaufort Cipher is a variant of the Vigenere Cipher: rather than add the key
 * to the plaintext, Beaufort subtracts the plaintext from the key.
 *
 * http://practicalcryptography.com/ciphers/beaufort-cipher/
 * http://home.earthlink.net/~fpcorr/Beaufort/BeaufortDecoder.html
 *
 * When the comments below talk about characters they really mean bytes.
 * This code was written in the pre-UTF-8 days when: byte == character.
 *
 * This code does not touch anything >= U+7F.
 *
 * bptr		buffer of characters to be encrypted
 * len		number of characters in the buffer
 *
 **********/
void	  cryptBuffer(char* bptr, LineLn len)
{
	register int cc;	/* current character being considered */

	static long key = 0;	/* 29 bit encipherment key */
	static int salt = 0;	/* salt to spice up key with */

	if (!bptr) {		/* is there anything here to encrypt? */
		key = len;	/* set the new key */
		salt = (int)len;/* set the new salt */
		return;
	}
	while (len--) { 	/* for every character in the buffer */

		cc = *bptr;	/* get a character out of the buffer */

		/* only encipher printable characters */
		if ((cc >= ' ') && (cc <= '~')) {

#if	EMACS5_CRYPT
/*	We feed the upper few bits of the key back into itself. This assures us
	that the starting key affects the entire message. We go through some
	effort here to prevent the key from exceeding 29 bits.	This is so the
	aritmetic caluclation in a later statement, which impliments our
	autokey, won't overflow, making the key go negative. Machine behavior
	in these cases does not tend to be portable.	*/

			key = (key & 0x1FFFFFFFl) ^ ((key >> 29) & 0x3l);
#else
/**  If the upper bit (bit 29) is set, feed it back into the key.  This 
	assures us that the starting key affects the entire message.  **/

			key &= 0x1FFFFFFFL;	/* strip off overflow */
			if (key & 0x10000000L) {
				key ^= 0x0040A001L;	/* feedback */
			}
#endif

/**  Down-bias the character, perform a Beaufort encipherment, and 
	up-bias the character again.  We want key to be positive 
	so that the left shift here will be more portable and the 
	mod95() faster	 **/

			cc = mod95((int)(key % 95) - (cc - ' ')) + ' ';

/**  the salt will spice up the key a little bit, helping to obscure 
	any patterns in the clear text, particularly when all the 
	characters (or long sequences of them) are the same.  We do 
	not want the salt to go negative, or it will affect the key 
	too radically.	It is always a good idea to chop off cyclics 
	to prime values.  **/

			if (++salt >= 20857) {	/* prime modulus */
				salt = 0;
			}

/**  our autokey (a special case of the running key) is being 
	generated by a wieghted checksum of clear text, cipher 
	text, and salt.   **/

#if	EMACS5_CRYPT
			key = key + key + (cc ^ *bptr) + salt;
#else
			key = key + key + cc + *bptr + salt;
#endif
		}
		*bptr++ = (char)cc;	/* put character back into buffer */
	}
	return;
}

static	int	mod95(register int val)
{
	/*  The mathematical MOD does not match the computer MOD  */

	/*  Yes, what I do here may look strange, but it gets the
		job done, and portably at that.  */

	while (val >= 9500)
		val -= 9500;
	while (val >= 950)
		val -= 950;
	while (val >= 95)
		val -= 95;
	while (val < 0)
		val += 95;
	return (val);
}
#else
void	nocrypt(void)
{
}
#endif

/* end */
