/*
 * ADDW SCCS: @(#)efunc.h	1.99 09/16/20 16:02:04
 */

/*	EFUNC.H:	MicroEMACS function declarations and names

		This file list all the C code functions used by MicroEMACS
	and the names to use to bind keys to them. To add functions,
	declare it here in both the extern function list and the name
	binding table.

*/

/* Many functions are invoked with the arguments f and n
 *
 * n is a repeat count, it can be negative - set N_FAIL_NEG_ARG to prevent this
 *
 * f is a flag. This sometimes gets set to True - I don't know when.
 */


/*	Name to function binding table
 *
 *	This table gives the names of all the bindable functions
 *	end their C function address. These are used for the bind-to-key
 *	function.
 * NB the 3rd field is left uninitialised as zero unless it is used.
 *
 * WARNING: If a function is NOT in here but is in keytab[] you will get a crash as the keytab
 * union will be left as a function pointer (k_ufp) rather than a NBIND pointer (k_ubp).
 *
 * Note: NextCommand() assumes that the command names will fit in the buffer passed to it,
 * which seems to mean are less that NSTRING.
 *
 * Keep the entries here sorted by command name. The user will see the ordering if they hit next/prev
 * when entering a command name.
 * Sorting: not as obvious as you might think. Use LANG=C this means ASCII sorting,
 * so that 'i-shell' sorts before 'incremental-search' rather than after it.
 *
 * The command names will never contain an upper case character.
 * This means that user named procedures should contain one to avoid
 * name space clashes.
 */

DCLXTRN(char	ExecNumMacro[], "Execute the contents of a numbered buffer");

// There is not (yet) a self insert command, but this is wanted
// for error reporting and the describe-key command.
EXTERN	NBIND	DummySelfInsert
#ifdef	maindef
     = {
	"self-insert",	SelfInsert,	NULL,	N_MODIFY | N_DEFAULT_1 | N_FAIL_NEG_ARG,
		"The character is entered into the buffer"
     }
#endif
     ;

// Used when recording last few commands executed
EXTERN	NBIND	DummyUnboundKey
#ifdef	maindef
     = {
	"unbound-key",	UnboundKey,	NULL,	0,
		"Unbound key"
     }
#endif
     ;


// Indexes into HelpBufferBinds[]
#define	HELP_BINDS_LINK_FOLLOW		0
#define	HELP_BINDS_PREVIOUS_PAGE	1
#define	HELP_BINDS_NEXT_PLACE		2
#define	HELP_BINDS_PREVIOUS_PLACE	3
EXTERN	NBIND	HelpBufferBinds[]
#ifdef	maindef
     = {
	{"help-link-follow",		helpLink,		NULL,		0,
		"Follow link in [Help] buffer"},
	{"help-previous-page",		helpPrevious,		NULL,		0,
		"Return to previous [Help] page"},
	{"help-next-place",		helpNextPlace,		NULL,		0,
		"Find next place in [Help]"},
	{"help-previous-place",		helpPreviousPlace,	NULL,		0,
		"Find previous place in [Help]"},
	{NULL,				NULL,			NULL,		0,
		NULL}
     }
#endif
     ;

// NULL terminated list of procedure names that can be bound to
EXTERN	NBIND*	ProcNames;

EXTERN	NBIND	CmdNames[]
#ifdef	maindef
     = {
	{"abort-command",		ctrlg,		NULL,		N_AI_BEFORE,
		"Abort current action"},
	{"add-global-mode",		setgMode,	NULL,		N_AI_BEFORE,
		"Interactive add global mode (default for new buffers)"},
	{"add-mode",			setMode,	NULL,		N_AI_BEFORE,
		"Interactive add buffer mode"},
	{"add-window-option",		AddWindowOpt,	NULL,		0,
		"Prompt for and add window specific modes"},
	{"append-file",			fileappend,	NULL,		N_MODIFY | N_RESTRICTED,
		"Ask for a file name, and write the contents of the current buffer to the end of that file"},
#if	APROP
	{"apropos",			apro,		NULL,		N_AI_BEFORE,
		"Brief help on commands chosen by keyword entered"},
#endif
	{"backward-character",		backchar,	forwchar,	N_DEFAULT_1,
		"Move left a character"},
	{"backward-character-several",	backCharSeveral, forwCharSeveral, N_AI_BEFORE | N_DEFAULT_1,
		"Move left several characters"},
	{"begin-macro",			ctlxlp,		NULL,		N_AI_BEFORE,
		"Start recording a keyboard macro"},
	{"beginning-of-file",		gotobob,	NULL,		N_AI_BEFORE | N_TRACK,
		"Move to the start of the buffer"},
	{"beginning-of-line",		gotobol,	NULL,		N_AI_BEFORE,
		"Move to start of the current line"},
	{"bind-to-key",			bindtokey,	NULL,		N_AI_BEFORE | N_RESTRICTED,
		"Set that a terminal keystroke will run an Emacs command"},
	{"bind-procedure-to-key",		bindProcToKey,	NULL,		N_AI_BEFORE | N_RESTRICTED,
		"Set that a terminal keystroke will run an Emacs macro"},
	{"buffer-position",		showcpos,	NULL,		N_AI_BEFORE,
		"Show position in the buffer and info about the character at dot"},
	{"case-region-lower",		lowerregion,	NULL,		N_MODIFY | N_NOT_HEX,
		"Change the alphabetic characters in the region to lower case"},
	{"case-region-upper",		upperregion,	NULL,		N_MODIFY | N_NOT_HEX,
		"Change the alphabetic characters in the region to upper case"},
	{"case-word-capitalize",	capword,	NULL,		N_MODIFY | N_DEFAULT_1 | N_FAIL_NEG_ARG | N_NOT_HEX,
		"Change the first letter of the next count words to upper case"},
	{"case-word-lower",		lowerword,	NULL,		N_MODIFY | N_DEFAULT_1 | N_FAIL_NEG_ARG | N_NOT_HEX,
		"Change the next count words to lower case"},
	{"case-word-upper",		upperword,	NULL,		N_MODIFY | N_DEFAULT_1 | N_FAIL_NEG_ARG | N_NOT_HEX,
		"Change the next count words to upper case"},
	{"change-char-case",		invert,		NULL,		N_MODIFY | N_DEFAULT_1 | N_FAIL_NEG_ARG | N_NOT_HEX,
		"Swap upper/lower case for count next characters"},
	{"change-directory",		changedir,	NULL,		N_RESTRICTED,
		"Change the current directory"},
	{"change-file-name",		filename,	NULL,		N_AI_BEFORE | N_RESTRICTED,
		"Change the file name associated with the current buffer"},
	{"change-screen-size",		newsize,	NULL,		N_AI_BEFORE,
		"Set screen size; not on Unix"},
	{"change-screen-width", 	newwidth,	NULL,		N_AI_BEFORE,
		"Resize and repaint the screen, use current terminal size; not on Unix"},
	{"clear-and-redraw",		refreshScreen,	NULL,		N_AI_BEFORE,
		"Refresh the screen"},
	{"clear-message-line",		clrmes,		NULL,		N_AI_BEFORE,
		"Clear the message line"},
	{"comment-region",		commentRegion,	NULL,		N_AI_BEFORE | N_MODIFY | N_NOT_HEX,
		"Comment out lines in the region"},
	{"copy-region",			copyregion,	NULL,		N_AI_BEFORE,
		"Copy all of the characters in the region to the kill buffer - to be yanked later"},
#if	WORDPRO
	{"count-words",			wordcount,	NULL,		N_AI_BEFORE | N_NOT_HEX,
		"Count the # of words in the marked region"},
#endif
	{"cycle-background-colour",	cyclegbcolor,	NULL,		0,
		"Change the global background colour"},
	{"cycle-foreground-colour",	cyclegfcolor,	NULL,		0,
		"Change the global foreground colour"},
	{"delete-blank-lines",		deblank,	NULL,		N_AI_BEFORE | N_MODIFY | N_NOT_HEX,
		"Delete blank lines around dot"},
	{"delete-buffer",		killbuffer,	NULL,		N_AI_BEFORE | N_RESTRICTED | N_SYNONYM,
		"Prompt for and close a buffer"},
	{"delete-global-mode",		delgMode,	NULL,		N_AI_BEFORE,
		"Interactive delete global mode (default for new buffers)"},
	{"delete-many-buffers",		killmany,	NULL,		N_AI_BEFORE | N_RESTRICTED | N_SYNONYM,
		"Remove all buffers, prompts yes/no for each"},
	{"delete-mode",			delMode,	NULL,		N_AI_BEFORE,
		"Interactive delete buffer mode"},
	{"delete-next-character",	forwdel,	backdel,	N_MODIFY | N_DEFAULT_1 | N_FAIL_NEG_ARG,
		"Delete count characters"},
	{"delete-next-word",		delfword,	delbword,	N_MODIFY | N_DEFAULT_1 | N_FAIL_NEG_ARG | N_NOT_HEX,
		"Delete count words forwards"},
	{"delete-other-windows",	onlywind,	NULL,		N_AI_BEFORE,
		"Make the current window the only one, to fill the screen"},
	{"delete-previous-character",	backdel,	forwdel,	N_MODIFY | N_DEFAULT_1 | N_FAIL_NEG_ARG,
		"Delete count characters before dot"},
	{"delete-previous-word",	delbword,	delfword,	N_MODIFY | N_DEFAULT_1 | N_FAIL_NEG_ARG | N_NOT_HEX,
		"Delete count words before dot"},
	{"delete-window",		delwind,	NULL,		N_AI_BEFORE,
		"Delete the current window"},
	{"delete-window-option",	DelWindowOpt,	NULL,		0,
		"Prompt for and delete window specific modes"},
	{"describe-bindings",		desbind,	NULL,		N_AI_BEFORE | N_NOT_HEX,
		"List the key bindings into a buffer. Help"},
	{"describe-key",		deskey,		NULL,		N_AI_BEFORE,
		"Prompt for a key, describe the command. Help"},
#if	AEDIT
	{"detab-line",			detab,		NULL,		N_AI_BEFORE | N_DEFAULT_1 | N_MODIFY | N_NOT_HEX,
		"Change tabs to spaces for count lines"},
	{"detab-region",		detabRegion,	NULL,		N_AI_BEFORE | N_MODIFY | N_NOT_HEX,
		"Change tabs to spaces in the current region"},
#endif
	{"display",			dispvar,	NULL,		N_RESTRICTED,
		"Show an environment or user variable"},
	{"end-macro",			ctlxrp,		NULL,		N_AI_BEFORE,
		"Stop recording a keyboard macro"},
	{"end-of-file",			gotoeob,	NULL,		N_AI_BEFORE | N_TRACK,
		"Move to the end of the buffer"},
	{"end-of-line",			gotoeol,	NULL,		0,
		"Move to the end of the current line"},
#if	AEDIT
	{"entab-line",			entab,		NULL,		N_AI_BEFORE | N_DEFAULT_1 | N_MODIFY | N_NOT_HEX,
		"Change spaces to tabs for count lines"},
	{"entab-region",		entabRegion,	NULL,		N_AI_BEFORE | N_MODIFY | N_NOT_HEX,
		"Change spaces to tabs in the current region"},
#endif
	{"exchange-point-and-mark",	swapmark,	NULL,		N_AI_BEFORE,
		"Swap the position of \".\" and \"mark\" in the current window"},
	{"execute-buffer",		execbuf,	NULL,		N_AI_BEFORE | N_DEFAULT_1 | N_FAIL_NEG_ARG | N_RESTRICTED | N_SYNONYM,
		"Execute the contents of a buffer of commands"},
	{"execute-command-line",	execcmd,	NULL,		N_AI_BEFORE | N_RESTRICTED,
		"Prompt for and execute an Emacs command"},
	{"execute-file",		execfile,	NULL,		N_AI_BEFORE | N_RESTRICTED,
		"Prompt for a file, execute its contents count times"},
	{"execute-macro",		ctlxe,		NULL,		N_AI_BEFORE | N_DEFAULT_1 | N_FAIL_NEG_ARG,
		"Execute a keyboard macro count times"},
	{"execute-macro-1",		cbuf1,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-2",		cbuf2,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-3",		cbuf3,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-4",		cbuf4,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-5",		cbuf5,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-6",		cbuf6,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-7",		cbuf7,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-8",		cbuf8,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-9",		cbuf9,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-10",		cbuf10,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-11",		cbuf11,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-12",		cbuf12,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-13",		cbuf13,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-14",		cbuf14,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-15",		cbuf15,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-16",		cbuf16,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-17",		cbuf17,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-18",		cbuf18,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-19",		cbuf19,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-20",		cbuf20,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-21",		cbuf21,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-22",		cbuf22,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-23",		cbuf23,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-24",		cbuf24,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-25",		cbuf25,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-26",		cbuf26,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-27",		cbuf27,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-28",		cbuf28,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-29",		cbuf29,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-30",		cbuf30,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-31",		cbuf31,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-32",		cbuf32,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-33",		cbuf33,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-34",		cbuf34,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-35",		cbuf35,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-36",		cbuf36,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-37",		cbuf37,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-38",		cbuf38,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-39",		cbuf39,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-macro-40",		cbuf40,		NULL,		N_AI_BEFORE,
		ExecNumMacro},
	{"execute-named-command",	namedcmd,	NULL,		N_RESTRICTED,
		"Prompt for and run an editor command"},
#if	PROC
	{"execute-procedure",		execproc,	NULL,		N_RESTRICTED | N_SYNONYM,
		"Execute a procedure"},
#endif
	{"execute-program",		spawn,		NULL,		N_AI_BEFORE | N_RESTRICTED | N_SYNONYM,
		"Prompt for and run a shell program"},
	{"exit-emacs",			quit,		NULL,		N_AI_BEFORE,
		"Leave Emacs, prompt if non saved buffers"},
#if	WORDPRO
	{"fill-paragraph",		fillpara,	NULL,		N_AI_BEFORE | N_MODIFY | N_DEFAULT_1 | N_FAIL_NEG_ARG | N_NOT_HEX,
		"Tidy a paragraph, wrap long lines"},
#endif
	{"filter-buffer",		filterBuffer,	NULL,		N_AI_BEFORE | N_MODIFY | N_NOT_HEX | N_RESTRICTED | N_NOT_EXEC,
		"Write buffer to prompted-for shell stdin, replace contents with it's stdout"},
	{"filter-region",		filterRegion,	NULL,		N_AI_BEFORE | N_MODIFY | N_NOT_HEX | N_RESTRICTED | N_NOT_EXEC,
		"Write region to prompted-for shell stdin, replace contents with it's stdout"},
	{"find-c-error",		FindCError,	NULL,		N_NOT_HEX,
		"Find unmatched parenthesis in C program"},
#if	CTAGS	// Unfortunately the sorting desire means that the #if happens twice
	{"find-current-tag",		findcurtag,	NULL,		N_AI_BEFORE | N_NOT_HEX,
		"Select name under cursor, find in ctags file, read file, move to the function/..."},
#endif
	{"find-file",			filefind,	NULL,		N_AI_BEFORE | N_RESTRICTED,
		"Prompt and read a file into a new buffer"},
#if	CTAGS
	{"find-tag",			findtag,	NULL,		N_AI_BEFORE | N_NOT_HEX,
		"Prompt for name, find in ctags file, read file, move to the function/..."},
#endif
	{"forget-file-name",		forgetFilename,	NULL,		N_AI_BEFORE | N_RESTRICTED,
		"Remove file name associated with the current buffer"},
	{"forward-character",		forwchar,	backchar,	N_AI_BEFORE | N_DEFAULT_1,
		"Move right a character"},
	{"forward-character-several",	forwCharSeveral, backCharSeveral, N_AI_BEFORE | N_DEFAULT_1,
		"Move right several characters"},
	{"get-and-goto-character",	getcscan,	NULL,		N_AI_BEFORE,
		"Prompt for scan character, search/find to it on the current line"},
	{"global",			global_var,	NULL,		0,
		"Declare a variable like %fred"},
	{"goto-character",		cscan,		NULL,		N_AI_BEFORE,
		"Search/find on the current line to previously scanned character"},
	{"goto-column",			gotocolumn,	NULL,		N_AI_BEFORE | N_TRACK,
		"Move to column number"},
	{"goto-line",			gotoline,	NULL,		N_AI_BEFORE | N_TRACK,
		"Move to line number count, prompt if not set"},
#if	CFENCE
	{"goto-matching-fence", 	getfence,	NULL,		N_AI_BEFORE | N_NOT_HEX,
		"Move to pair of {}[]()<> that is at dot"},
#endif
	{"grow-window",			enlargewind,	shrinkwind,	N_AI_BEFORE | N_DEFAULT_1,
		"Make current window count lines taller"},
	{"handle-tab",			tab,		NULL,		N_FAIL_NEG_ARG | N_NOT_HEX,
		"Set to count spaces width an inserted tab is expanded to"},
	{"help",			help,		NULL,		N_AI_BEFORE,
		"Read help file into a buffer"},
	{"hunt-backward",		backhunt,	forwhunt,	N_AI_BEFORE | N_DEFAULT_1 | N_NOT_HEX,
		"Find backwards for the string prompted and moved to with search-forward or search-reverse"},
	{"hunt-forward",		forwhunt,	backhunt,	N_AI_BEFORE | N_DEFAULT_1 | N_NOT_HEX,
		"Find forwards for the string prompted and moved to with search-forward or search-reverse"},
	{"i-shell",			spawncli,	NULL,		N_AI_BEFORE | N_RESTRICTED,
		"Run interactive sub shell (command line interpreter)"},
#if	ISRCH
	{"incremental-search",		fisearch,	NULL,		N_AI_BEFORE | N_DEFAULT_1 | N_NOT_HEX,
		"Find forwards, search string: character at a time"},
#endif
	{"indent-line",			indentLines,	NULL,		N_AI_BEFORE | N_MODIFY | N_DEFAULT_1 | N_FAIL_NEG_ARG | N_NOT_HEX,
		"Increase indent by width of count tabs"},
	{"indent-region",		indentRegion,	NULL,		N_AI_BEFORE | N_MODIFY | N_DEFAULT_1 | N_FAIL_NEG_ARG | N_NOT_HEX,
		"Increase indent of lines in the region by width of count tabs"},
	{"initialise-kill-buffer",	initkill,	NULL,		N_AI_BEFORE,
		"Empty the kill buffer (yank ring)"},
	{"insert-command",		inscmd,		NULL,		N_AI_BEFORE | N_MODIFY | N_NOT_HEX | N_RESTRICTED,
		"Insert the output of a one line command into the current buffer"},
	{"insert-file",			insfile,	NULL,		N_AI_BEFORE | N_MODIFY | N_NOT_HEX | N_RESTRICTED,
		"Prompt for file path, insert into current buffer"},
	{"insert-space",		insspace,	NULL,		N_MODIFY | N_DEFAULT_1 | N_FAIL_NEG_ARG | N_NOT_HEX,
		"Insert count spaces"},
	{"insert-string",		istring,	NULL,		N_MODIFY | N_DEFAULT_1 | N_FAIL_NEG_ARG | N_NOT_HEX,
		"Prompt for string, insert into buffer"},
	{"insert-tab",			instab,		NULL,		N_MODIFY | N_DEFAULT_1 | N_FAIL_NEG_ARG | N_NOT_HEX,
		"Insert a tab character, maybe transforming"},
	{"insert-unicode",		insUnicode,	NULL,		N_MODIFY | N_DEFAULT_1 | N_FAIL_NEG_ARG,
		"Prompt for hex value of Unicode character to insert"},
	{"kill-buffer",			killbuffer,	NULL,		N_AI_BEFORE | N_RESTRICTED | N_SYNONYM,
		"Remove current buffer, prompts yes/no"},
	{"kill-many-buffers",		killmany,	NULL,		N_AI_BEFORE | N_RESTRICTED | N_SYNONYM,
		"Remove all buffers, prompts yes/no for each"},
#if	WORDPRO
	{"kill-paragraph",		killpara,	NULL,		N_AI_BEFORE | N_DEFAULT_1 | N_MODIFY | N_FAIL_NEG_ARG | N_NOT_HEX,
		"Delete count paragraphs"},
#endif
	{"kill-region",			killregion,	NULL,		N_AI_BEFORE | N_MODIFY,
		"Delete text in the region (between dot and mark)"},
	{"kill-to-end-of-line",		killtext,	NULL,		N_AI_BEFORE | N_MODIFY | N_NOT_HEX,
		"Delete text from dot to the end of the line"},
#if	FLABEL
	{"label-function-key",		fnclabel,	NULL,		0,
		"Label a function key"},
#endif
	{"list-buffers",		listbuffers,	NULL,		N_AI_BEFORE,
		"Show active buffers & much more Emacs status information"},
	{"local",			local_var,	NULL,		0,
		"Declare a local variable"},
	{"map-key",			mapkey,		NULL,		N_RESTRICTED,
		"Read in a key mapping for all future keyboard input"},
#if	0
	{"meta-prefix",			metaPrefix,	NULL,		0,
		"When Meta (escape) is pressed, further command prompting"},	// Useful ?
#endif
	{"meta-quote",			meta_quote,	NULL,		N_MODIFY | N_DEFAULT_1,
		"Prompt & insert character, OR in top bit (0x80)"},
	{"move-to-window-down",		MoveToWindowDown, MoveToWindowUp, 0,
		"Move to window below"},
	{"move-to-window-left",		MoveToWindowLeft, MoveToWindowRight, 0,
		"Move to window to the left"},
	{"move-to-window-right",	MoveToWindowRight, MoveToWindowLeft, 0,
		"Move to window to the right"},
	{"move-to-window-up",		MoveToWindowUp,	MoveToWindowDown, 0,
		"Move to window above "},
	{"move-window-down",		mvdnwind,	mvupwind,	N_DEFAULT_1 | N_TRACK,
		"Move window down by count lines"},
	{"move-window-left",		moveLeftWind,	moveRightWind,	N_DEFAULT_1,
		"Move window left by count columns"},
	{"move-window-right",		moveRightWind,	moveLeftWind,	N_DEFAULT_1,
		"Move window right by count columns"},
	{"move-window-up",		mvupwind,	mvdnwind,	N_DEFAULT_1 | N_TRACK,
		"Move window up by count lines"},
	{"move-window-several-down",	moveDownWinSev,	moveUpWinSev,	N_DEFAULT_1 | N_TRACK,
		"Move window down by several lines"},
	{"move-window-several-left",	moveLeftWinSev, moveRightWinSev, N_DEFAULT_1,
		"Move window left by several columns"},
	{"move-window-several-right",	moveRightWinSev, moveLeftWinSev, N_DEFAULT_1,
		"Move window right by several columns"},
	{"move-window-several-up",	moveUpWinSev,	moveDownWinSev,	N_DEFAULT_1 | N_TRACK,
		"Move window up by several lines"},
	{"name-buffer",			namebuffer,	NULL,		N_AI_BEFORE | N_RESTRICTED | N_NOT_EXEC,
		"Rename the current buffer"},
	{"narrow-to-region",		narrow,		NULL,		0,
		"Changes only on part of buffer"},
	{"newline",			newline,	NULL,		N_AI_AFTER | N_MODIFY | N_DEFAULT_1 | N_FAIL_NEG_ARG | N_NOT_HEX,
		"Insert a line"},
	{"newline-and-indent",		newlineAndIndent, NULL,		N_AI_AFTER | N_MODIFY | N_DEFAULT_1 | N_FAIL_NEG_ARG | N_NOT_HEX,
		"Insert line then tab/space/... to match current line indent"},
	{"newline-and-maybe-indent",	newlineMaybeIndent, NULL,	N_AI_AFTER | N_MODIFY | N_DEFAULT_1 | N_FAIL_NEG_ARG | N_NOT_HEX,
		"Insert line then unless INDENT mode insert tab/space/... to match current line indent"},
	{"next-buffer",			nextbuffer,	NULL,		N_DEFAULT_1 | N_FAIL_NEG_ARG | N_FAIL_ZERO_ARG,
		"Switch to the next buffer in the buffer list"},
	{"next-line",			forwline,	backline,	N_AI_BEFORE | N_TRACK | N_DEFAULT_1,
		"Move down a line"},
	{"next-line-several",		forwLineSeveral, backLineSeveral, N_AI_BEFORE | N_TRACK | N_DEFAULT_1,
		"Move down several lines"},
	{"next-page",			forwpage,	backpage,	N_AI_BEFORE | N_TRACK | N_DEFAULT_1,
		"Move forward by count pages"},
#if	WORDPRO
	{"next-paragraph",		gotoeop,	gotobop,	N_AI_BEFORE | N_TRACK | N_DEFAULT_1 | N_NOT_HEX,
		"Move forward to the end of the current paragraph"},
#endif
	{"next-window",			nextwind,	NULL,		0,
		"Change to window after the current one"},
	{"next-word",			forwword,	backword,	N_AI_BEFORE | N_DEFAULT_1 | N_NOT_HEX,
		"Move forward by count words"},
	{"nop",				nullproc,	NULL,		0,
		"No-operation, does nothing"},
	{"open-line",			openline,	NULL,		N_MODIFY | N_DEFAULT_1 | N_FAIL_NEG_ARG,
		"Insert line, cursor remains at opened end of old line"},
	{"overwrite-string",		ovstring,	NULL,		N_AI_BEFORE | N_DEFAULT_1 | N_FAIL_NEG_ARG | N_MODIFY | N_NOT_HEX ,
		"Prompt for string, use to replace buffer text"},
	{"pipe-command",		pipecmd,	NULL,		N_NOT_HEX | N_RESTRICTED,
		"Run shell command, save output in a buffer"},
	{"pipe-to-command",		PipeToCommand,	NULL,		N_RESTRICTED,
		"Write buffer/narrow to shell command"},
	{"place-alpha-mark",		pamark,		NULL,		0,
		"Mark dot, name it with a letter of alphabet"},
	{"previous-line",		backline,	forwline,	N_AI_BEFORE | N_TRACK | N_DEFAULT_1,
		"Move up a line"},
	{"previous-line-several",	backLineSeveral, forwLineSeveral, N_AI_BEFORE | N_TRACK | N_DEFAULT_1,
		"Move up several lines"},
	{"previous-page",		backpage,	forwpage,	N_AI_BEFORE | N_TRACK | N_DEFAULT_1,
		"Move up a page in the buffer"},
#if	WORDPRO
	{"previous-paragraph",		gotobop,	gotoeop,	N_AI_BEFORE | N_TRACK | N_DEFAULT_1 | N_NOT_HEX,
		"Move back to the beginning of the current paragraph"},
#endif
	{"previous-window",		prevwind,	NULL,		0,
		"Change to window before current one"},
	{"previous-word",		backword,	forwword,	N_AI_BEFORE | N_DEFAULT_1 | N_NOT_HEX,
		"Move back by count words"},
	{"query-replace-string",	qreplace,	NULL,		N_AI_BEFORE | N_MODIFY | N_FAIL_NEG_ARG | N_NOT_HEX,
		"Search/find and replace, prompt/confirm each change"},
	{"quick-exit",			quickexit,	NULL,		N_AI_BEFORE,
		"Save all buffers and quit"},
	{"quote-character",		quote,		NULL,		N_MODIFY | N_DEFAULT_1 | N_NOT_HEX,
		"Prompt & insert character, don't check for it being a command key"},
	{"read-file",			fileread,	NULL,		N_AI_BEFORE | N_MODIFY | N_NOT_HEX | N_RESTRICTED,
		"Read a file into the current buffer"},
	{"redraw-display",		reposition,	NULL,		N_TRACK,
		"Redisplay such that dot is on 'count' line in the window"},
	{"renew-buffer",		renew,		NULL,		N_RESTRICTED,
		"Kills the named buffer, then reread the corresponding file"},
	{"replace-string",		sreplace,	NULL,		N_AI_BEFORE | N_MODIFY | N_FAIL_NEG_ARG | N_NOT_HEX,
		"Search/find and replace, don't query each change"},
	{"resize-window",		resize,		NULL,		0,
		"Change # lines of the current window to count lines"},
	{"restore-window",		restwnd,	NULL,		0,
		"Return to saved window"},
	{"return-alpha-mark",		retamark,	NULL,		0,
		"Return to an alphabetic (named) or numeric mark"},
#if	ISRCH
	{"reverse-incremental-search",	risearch,	NULL,		N_AI_BEFORE | N_DEFAULT_1 | N_NOT_HEX,
		"Find backwards, search string: character at a time"},
#endif
	{"reyank",			reyank,		NULL,		N_AI_BEFORE | N_MODIFY | N_DEFAULT_1,
		"Remove text just yanked, replace with previous text in kill ring"},
#if	PROC
	{"run",				execproc,	NULL,		N_RESTRICTED | N_SYNONYM,
		"Execute a procedure"},
#endif
	{"save-file",			filesave,	NULL,		N_AI_BEFORE | N_MODIFY,
		"Save the contents of the current buffer to its associated file"},
	{"save-many-buffers",		savemany,	NULL,		N_AI_BEFORE,
		"Save modified buffers, prompt each for confirmation"},
	{"save-window",			savewnd,	NULL,		N_AI_BEFORE,
		"Make note of id of current window"},
	{"scroll-next-down",		scrnxtdw,	scrnxtup,	N_DEFAULT_1,
		"Scroll the next window down (forward) a page"},
	{"scroll-next-up",		scrnxtup,	scrnxtdw,	N_DEFAULT_1,
		"Scroll the next window up (back) a page"},
	{"search-forward",		forwsearch,	backsearch,	N_DEFAULT_1 | N_NOT_HEX,
		"Find looking down the buffer"},
	{"search-reverse",		backsearch,	forwsearch,	N_DEFAULT_1 | N_NOT_HEX,
		"Find looking up the buffer"},
	{"select-buffer",		usebuffer,	NULL,		0,
		"Choose a buffer to display in current window"},
	{"set", 			setvar,		NULL,		0,
		"Set a variable"},
#if	CRYPT
	{"set-encryption-key",		crysetkey,	NULL,		0,
		"Set encryption key of current buffer"},
#endif
	{"set-fill-column",		setfillcol,	NULL,		N_FAIL_NEG_ARG | N_NOT_HEX,
		"Set fill column (line wrap) to count or $fillcol"},
	{"set-mark",			setmark,	NULL,		0,
		"Set the mark in the current window at dot"},
	{"shell-command",		spawn,		NULL,		N_AI_BEFORE | N_RESTRICTED | N_SYNONYM,
		"Run a one-liner shell command"},
	{"show-marks",		showMarks,	NULL,		0,
		"Show alpha and numeric marks"},
#if	SHOWMODREL
	{"show-module-release",		showmods,	NULL,		0,
		"Print out the version strings for each module"},
#else
	{"show-module-release",		ctrlg,	NULL,		0,
		"Print out the version strings for each module (not configured)"},
#endif
	{"shrink-window",		shrinkwind,	enlargewind,	N_AI_BEFORE | N_DEFAULT_1,
		"Make window have fewer lines"},
	{"split-and-find-buffer",	owinbuf,	NULL,		0,
		"Split window top/bottom, select a buffer"},
	{"split-and-find-file",		owinfil,	NULL,		N_RESTRICTED,
		"Split window top/bottom, read a file into other window"},
	{"split-current-window",	splitwindvert,	NULL,		N_SYNONYM,
		"Split the current window top/bottom"},
	{"split-horizontally-and-find-buffer", splitwindhorizbuf, NULL, N_RESTRICTED,
		"Split window left/right, select a buffer"},
	{"split-horizontally-and-find-file",	oHorizWinfil,	NULL,	N_RESTRICTED,
		"Split window left/right, read a file into other window"},
	{"split-window",		splitwind,	NULL,		0,
		"Split the current window top/bottom or left/right"},
	{"split-window-horizontally",	splitwindhoriz,	NULL,		0,
		"Split current window left/right"},
	{"split-window-vertically",	splitwindvert,	NULL,		N_SYNONYM,
		"Split the current window top/bottom"},
	{"store-macro",			storemac,	NULL,		0,
		"Set up a macro buffer and flag to store all executed command lines there"},
#if	PROC
	// This command name is also tested for in exec.c dobuf()
	{"store-procedure",		storeproc,	NULL,		0,
		"Set up a procedure buffer and flag to store all executed command lines there"},
#endif
#ifdef	SIGTSTP				/* USG as well as BSD these days */
	{"suspend-emacs",		bktoshell,	NULL,		N_RESTRICTED,
		"Suspend back to the shell - job control"},
#endif
#if	TEST_HOOK
	{"test-emacs",			test_emacs, NULL,		0,
		"Emacs development test hook"},
#endif
	{"toggle-insert-mode",		ToggleInsertMode, NULL,		0,
		"Flip or set current buffer's insert/overwrite mode"},
	{"top-bottom-switch",		topbot,		NULL,		N_AI_BEFORE | N_TRACK,
		"If at the top of the buffer, goto the bottom, else goto the top"},
	{"transpose-characters",	twiddle,	NULL,		N_AI_BEFORE | N_MODIFY,
		"Twiddle: swap the two characters on either side of dot"},
	{"trim-buffer",			buffertrim,	NULL,		N_MODIFY | N_NOT_HEX | N_NOT_EXEC,
		"Trim the current buffer so that it has at most count lines"},
#if	AEDIT
	{"trim-line",			trim,		NULL,		N_AI_BEFORE | N_MODIFY | N_DEFAULT_1 | N_NOT_HEX,
		"Trim trailing white space from the point to line end"},
	{"trim-region",			trimRegion,	NULL,		N_AI_BEFORE | N_MODIFY | N_NOT_HEX,
		"Trim trailing white space from lines in the region"},
#endif
	{"unbind-key",			unbindkey,	NULL,		N_RESTRICTED,
		"Unset that a key will run an Emacs command"},
	{"undo",			undo,		NULL,		N_RESTRICTED | N_MODIFY,
		"Undo change to buffer - not yet implemented"},
	{"unindent-line",		unindentLines,	NULL,		N_AI_BEFORE | N_MODIFY | N_DEFAULT_1 | N_FAIL_NEG_ARG | N_NOT_HEX,
		"Decrease indent by width of count tabs"},
	{"unindent-region",		unidentRegion,	NULL,		N_AI_BEFORE | N_MODIFY | N_DEFAULT_1 | N_FAIL_NEG_ARG | N_NOT_HEX,
		"Decrease indent of lines in the region by width of count tabs"},
	{"unmark-buffer",		unmark,		NULL,		0,
		"Forget that the current buffer has been changed/modified"},
	{"update-screen",		upscreen,	NULL,		0,
		"Force a screen update"},
	{"view-file",			viewfile,	NULL,		N_AI_BEFORE | N_RESTRICTED,
		"Open a file in read-only mode"},
	{"widen",			widen,		NULL,		N_SYNONYM,
		"Changes on all of buffer"},
	{"widen-from-narrow",		widen,		NULL,		N_SYNONYM,
		"Changes on all of buffer"},
	{"window-bottom-edge-move",	WindowMoveBottom, NULL,		N_DEFAULT_1,
		"Move the bottom edge of the window by count - which may be -ve"},
	{"window-left-edge-move",	WindowMoveLeft, NULL,		N_DEFAULT_1,
		"Move the left edge of the window by count - which may be -ve"},
	{"window-menu",			windowManipulate, NULL,		0,
		"Interactive window size/position/... manipulation"},
	{"window-right-edge-move",	WindowMoveRight, NULL,		N_DEFAULT_1,
		"Move the right edge of the window by count - which may be -ve"},
	{"window-top-edge-move",	WindowMoveTop,	NULL,		N_DEFAULT_1,
		"Move the top edge of the window by count - which may be -ve"},
	{"wrap-word",			wrapword,	NULL,		N_AI_BEFORE | N_MODIFY | N_NOT_HEX,
		"Insert line break before current word"},	// ??
	{"write-file",			filewrite,	NULL,		N_AI_BEFORE | N_RESTRICTED,
		"Ask for a file name, and write the contents of the current buffer to that file"},
	{"write-region",		regionwrite,	NULL,		N_AI_BEFORE | N_RESTRICTED,
		"Ask for a file name, and write region to that file"},
	{"write-message",		writemsg,	NULL,		0,
		"Write a string on the message line"},
	{"yank",			yank,		NULL,		N_MODIFY | N_DEFAULT_1 | N_FAIL_NEG_ARG,
		"Insert text back from the kill buffer"},
	{NULL,				NULL,		NULL,		0,	NULL}
   }

#endif	// maindef
;

/* end */
