/*
 * ADDW SCCS: @(#)estruct.h	2.331 10/15/20 10:22:01
 */

/*	ESTRUCT:	Structure and preprocessor defines for
 *			MicroEMACS 3.9
 *
 *			written by Dave G. Conroy
 *			modified by Steve Wilhite, George Jones
 *			greatly modified by Daniel Lawrence
 *			Yet more hacks: Alain D D Williams
 *
 * Almost all data structures have to be declared here otherwise you get errors
 * if a function prototype mentions it in edef.h
 *
 * There is a lot of definitions for environments that will, probably, never
 * be seen again. Just leave them there, if only as a reminder of times past.
 */

/* For a long time this has only been compiled on machines where sizeof(int) >= 4.
 * This matters where there may be a line > 64K.
 * Some convertion of int -> long for offsets has been done, but prob not complete.
 */


#ifdef	LATTICE
#undef	LATTICE		// don't use their definitions...use ours
#endif
#ifdef	MSDOS
#undef	MSDOS
#endif
#ifdef	CPM
#undef	CPM
#endif
#ifdef	AMIGA
#undef	AMIGA
#endif
#ifdef	CTRLZ
#undef	CTRLZ
#endif

/*	Version definition moved to evers.h	*/

/*	Machine/OS definitions			*/

#define	AMIGA			0	// AmigaDOS
#define	ST520			0	// ST520, TOS
#define	MEGAMAX			0	// Whatever one of those was
#define	MSDOS			0	// MS-DOS
#ifndef	V7
#define	V7			0	// V7 UN*X or Coherent or BSD4.2
#endif
#ifndef	 BSD_UNIX
#define	 BSD_UNIX		0	// UNIX BSD 4.2	and ULTRIX
#endif
#ifndef	USG
#define	USG			1	// UNIX system V
#endif
#define	VMS			0	// VAX/VMS
#define	CPM			0	// CP/M-86
#define	CPM68K			0	// CPM something
#define	FINDER			0	// Macintosh OS

#if	AMIGA
#define	OPERATING_SYSTEM	"AMIGADOS"
#endif
#if	VMS
#define	OPERATING_SYSTEM	"VMS"
#endif
#if	MSDOS
#define	OPERATING_SYSTEM	"MSDOS"
#endif

#if	V7 || BSD_UNIX || USG
#define	UNIX			1	// Running some flavour of UNIX
#define	OPERATING_SYSTEM	"UNIX"
#endif

#if	USG && BSD_UNIX
%ERROR					// make the compiler barf
You cannot define both USG (System V) and  BSD_UNIX, you will have to define one
or the other
#endif

#if	USG && V7
%ERROR					// make the compiler barf
You cannot define both V7 (Version 7) and USG (System V), you will have to
define one or the other
#endif

#ifndef	OPERATING_SYSTEM
#define	OPERATING_SYSTEM	"unknown"
#endif

/*	Important definitions			*/

// We now have modern compilers and environments that provides <stdarg.h>
// #define	VA_ARGS	0	/* set this if you have varargs			*/
// #define	STD_ARG	1	/* set this if you have stdarg			*/

/*	Compiler definitions			*/
#define	MWC			0	// Marc Williams C
#define	MWC86			0	// Marc Williams C
#define	LATTICE			0	// Lattice 2.14 thruough 3.0 compilers
#define	AZTEC			0	// Aztec C 3.20e
#define	MSC			0	// MicroSoft C compile version 3 & 4
#define	TURBO			0	// Turbo C/MSDOS
#define	HAVE_CTYPE		1	// <ctype.h> is available
#define	_GNU_SOURCE		1	// Get some GNU definitions in system header files

/* Use this to tell the compiler that the function does not return.
 * This should be on the predeclaration, not the function definition itself.
 */
#ifdef	__GNUC__
#define	__NOT_RETURN(funcDef)	funcDef __attribute__ ((noreturn))
#else
#define	__NOT_RETURN(funcDef)	funcDef
#endif	// __GNUC__

/* Tell the compiler that the code does not go any further.
 * Put this on a line that will not be executed.
 * Makes the code a tiny bit larger, but stops compiler warnings.
 */
#define	__NOT_REACHED		exit(0);

// If you want to ignore the return value of a function
// Do not put a ';' after this, not needed & some compilers complain
#define	IGNORE(M)		if(1==((int)M)){;}

// Make a word into a string so that it can be concatenated by the compiler
#define	STRINGIFY(word)		#word
// Number #define -> string so that it can be concatenated by the compiler
#define	STR_DEFINE(x)		STRINGIFY(x)


/* Debugging options
 * Beware: RAMSIZE makes the assumption that the size of the allocated memory is
 * kept in the word before the pointer. This is not always true.
 */

#define	RAMSIZE			0	// dynamic RAM memory usage tracking
#define	RAMSHOW			0	// auto dynamic RAM reporting

/*	Special keyboard definitions		*/
 
#define	WANGPC			0	// WangPC - mostly escape sequences
#define	VT100			0	// Handle VT100 style keypad
 
/*	Terminal Output definitions		*/

#define	EGA			0	// IBM-PC EGA
#define	ANSI			0	// ANSI escape sequences
#define	HP150			0	// HP150 screen driver
#define	HP110			0	// HP110 screen driver
#define	VMSVT			0	// various VMS terminal entries
#define	VT52			0	// VT52 terminal (Zenith).
#define	RAINBOW			0	// Use Rainbow fast video.
#define	TERMCAP			1	// Use TERMCAP or curses
#define	TERMCAP_USE_TERM_H	1	// Use <term.h>
#define	CURSES			1	// Use curses (aka new TERMCAP)
#define	NEEDSYSIOCTL		1	// Need <sys/ioctl.h>
#define	FIXSG			0	// Fix stand-out glitch
#define	IBMPC			0	// IBM-PC specific driver
#define	DG10			0	// Data General system/10
#define	TIPC			0	// TI Profesional PC driver
#define	Z309			0	// Zenith 100 PC family	driver
#define	MAC			0	// Macintosh
#define	ATARI			0	// Atari 520/1040ST screen

#define	DISPWIDE		1	// Display wide characters on the screen
#define	LEGACY_DISPLAY		0	// Use reverse and not colours, also some ancient term quirks - maybe should be run-time flag


/*	Configuration options	*/

// #define CVMVAS		1	// Arguments to page forward/back in pages
#define	CLRMSG			0	// Space clears the message line with no insert
#define	PRSVMSG			1	// Try to preserve mesage
#define	CFENCE			1	// Fence matching in CMODE
#define	RDFENCE			1	// If fence off screen, screen is redrawn
#define	TYPEAH			1	// Type ahead causes update to be skipped
#define	DEBUGM			1	// $debug triggers macro debugging
#define	VISMAC			0	// Update display during keyboard macros
#define	CTRLZ			0	// Add a ^Z at end of files under MSDOS only
#define	ADDCR			0	// Add \r to end of lines written to file
#define	ALPHORD			0	// Have buffers ordered alphabetically
#define	NBRACE			1	// New style brace matching command
#define	SHOWMODREL		0	// Implement the show-module-release command

#define	REVSTA			1	// Status line appears in reverse video
// COLOUR is now assumed to be always true.
// Much conditional compilation to allow it to be taken out has been removed
#define	COLOR			1	// Color commands and windows
#define	FLABEL			0	// Labels on function keys

#define	LOCKING			0	// File locking module
#define	FILOCK			0	// file locking under unix BSD 4.2
#define	BACKUP			1	// Automatically make a backup file
#define	BUPEVS			0	// Backup file every save (or with first save only)
#define	ISRCH			1	// Incremental searches like ITS EMACS
#define	WORDPRO			1	// Advanced word processing features
#define	CBMODE			0	// Use CBREAK mode rather than raw mode	(BSD)
#define	APROP			1	// Add code for Apropos command
#define	CRYPT			1	// File encryption enabled?
#define	MAGIC			1	// Include regular expression matching
#define	AEDIT			1	// Advanced editing options: en/detabbing
#define	PROC			1	// Named procedures
#define	SYNC			1	// Sync file (not machine) after writes (UNIX)
#define	WINSIZE			1	// UNIX Window size change supported
#define	SigType			void	// What the signal() function returns
#define	SIGSETJMP		1	// Use sigsetjmp() rather than setjmp()
#define	DOTINPATH		0	// "." (pwd) is in command file open path
#define	SYSRAND			1	// Use system supplied rand() not own ernd()
#define	MODFOLD			0	// Fold M- and ^X alpha commands to upper case
#define	ARG_IS_LONG		1	// Argument n to commands is type long not int

// Some that might be defined on the compile line
#ifndef	TEST_HOOK
#define	TEST_HOOK		0	// Enable arbitrary test code and command test-emacs
#endif

#ifndef	CHECK_EVERY
#define	CHECK_EVERY		0	// Enable CheckEveryCommand()
#endif
#if	CHECK_EVERY
// A macro to call this at a suitable point:
#define	cHECKeVERYcOMMAND()	CheckEveryCommand()
#else
#define	cHECKeVERYcOMMAND()
#endif

#ifndef	CLEAN
#define	CLEAN			0	// de-alloc memory on exit
#endif

#ifndef	TERMDEBUG
#define	TERMDEBUG		0	// -K option, input & termcap debug/trace
#endif

#ifndef	CALLED
#define	CALLED			0	// is Emacs a called subroutine? or stand alone
#endif

#ifndef	TRACEBUFFER
#define	TRACEBUFFER		0	// Create & write to a trace/debug buffer
#endif


#define	ASCII			1	// Always using ASCII char sequences for now
#define	EBCDIC			0	// later IBM mainframe versions will use EBCDIC

#define	CONFIG_MAX_UNICODE	0	// See utf8.h - non standard

// Include here so that config options above will take effect:
#include <stdio.h>

#if	UNIX
// Get some POSIX includes

#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <time.h>
#include <signal.h>
#include <setjmp.h>
#include <stdarg.h>

#include <limits.h>	// Actually: part of the C standard
#include <stddef.h>

typedef	time_t	Time_T;	// Time: since Epoch, granularity 1 second
#else

// A good starting default:
typedef	long	Time_T;	// Time: since Epoch, granularity 1 second
#endif


// Was dependent on #if	IBMPC == 0, now just set off
#define	TONY_BINDINGS	0	// sorry about this folks

/*	System dependant library redefinitions, structures and includes */

#if	TURBO
#include	<dos.h>
#include	<mem.h>
#undef	peek
#undef	poke
#define		peek(a,b,c,d)	movedata(a,b,FP_SEG(c),FP_OFF(c),d)
#define		poke(a,b,c,d)	movedata(FP_SEG(c),FP_OFF(c),a,b,d)
#endif	// TURBO

#if	VMS
#define	atoi	xatoi
#define	abs	xabs
#define	getname	xgetname
#endif	// VMS

#if	LATTICE
#define	unsigned
#endif

#if	AZTEC
#undef	fputc
#undef	fgetc
#if	MSDOS
#define	fgetc	a1getc
#else
#define	fgetc	agetc
#endif	// MSDOS
#define	fputc	aputc
#define	int86	sysint
#define	intdos(a, b)	sysint(33, a, b)
#define	inp	inportb
#define	outp	outportb

struct XREG {
	int ax,bx,cx,dx,si,di;
};

struct HREG {
	char al,ah,bl,bh,cl,ch,dl,dh;
};

union REGS {
	struct XREG	x;
	struct HREG	h;
};

struct SREGS {
	unsigned cs, ss, ds, es;
};
#endif	// AZTEC

#if	MSDOS && MWC
#include	<dos.h>
#define	int86(a, b, c)	intcall(b, c, a)
#define	inp	in

struct XREG {
	int	ax,bx,cx,dx,si,di,ds,es,flags;
};

struct HREG {
	char	al,ah,bl,bh,cl,ch,dl,dh;
	int	ds,es,flags;
};

union REGS {
	struct XREG	x;
	struct HREG	h;
};
#endif	// MSDOS && MWC

#if	MSDOS && MSC
#include	<dos.h>
#include	<memory.h>
#define	peek(a,b,c,d)	movedata(a,b,FP_SEG(c),FP_OFF(c),d)
#define	poke(a,b,c,d)	movedata(FP_SEG(c),FP_OFF(c),a,b,d)
#define	movmem(a, b, c)		memcpy(b, a, c)
#endif	// MSDOS && MSC

#if	MSDOS && LATTICE
#undef	CPM
#undef	LATTICE
#include	<dos.h>
#undef	CPM
#endif	// MSDOS && LATTICE

#if	VMS
#define	unlink(a)	delete(a)
#endif	// VMS

/*	Define some ability flags */

#if	IBMPC || Z309
#define	MEMMAP	1
#else
#define	MEMMAP	0
#endif	// IBMPC || Z309

#if	MSDOS || UNIX || (ST520 && MWC)
#define	ENVFUNC	1
#else
#define	ENVFUNC	0
#endif	// MSDOS || UNIX || (ST520 && MWC)

/* At some points it is assumed that we are running on a 32 bit machine or bigger,
 * eg flags arg to mlGetKey() MGC_?? will not fit in 16 bits.
 */
#if	INT_MAX < 32768
#error	"16 bit ints not supported"
#endif

/*	Emacs global flag bit definitions (for gflags)	*/

#define	GFREAD	1

#ifdef	__FreeBSD__
// Type of 3rd argument to setbuffer()
typedef	int	SETBsize_t;
#else
typedef	size_t	SETBsize_t;
#endif	// __FreeBSD__

/*	Internal constants	*/

/* Beware there is some pretty cavalier copying around w/out checking buffer
 * sizes. These have increased over the years - a lot of it was origianlly
 * tied to max screen width (which is now variable & can be large).
 * A lot of big buffers existed on the stack - many of these have been removed.
 *
 * NBUFN is just large enough for [Message History].
 */

#define	NBINDS			280	// max # keys that can be bound
#define	NFILEN			4096	// # of bytes, file name - beware macros have a smaller limit
#define	NBUFN			17	// # of characters, buffer name - b_bname (NB char != byte)
#define	NLINE			256	// # of bytes, input line
#define	NSTRING			256	// # of bytes, string buffers
#define	NPAT			256	// # of bytes, pattern
#define	TERM_HUGE		10000	// Impossible cursor positions
#define	NLOCKS			100	// max # of file locks active
#define	NCOLORS			8	// number of supported colors
#define	NBLOCK			16	// line block chunk size. *must* be power of 2
#define	NVSIZE			21	// max #chars in a var name
#define	MAXCOL			160	// size of various prompt buffers
#define	CMDMAXLEN		60	// Max length of a command name
#define	MAX_MACRO_NUM		40	// Max macro number - start count at 1
#define	MILLION			1000000	// A useful number to have
#define	CRASHMSGLINES		10	// How many message buffer lines to print
#define	CMDSKEPT		10	// How many commands executed to keep info on
#define	WINDOW_SEVERAL_FRACTION	4	// Fraction of window height/width moved by move-window-several-xx
#define	MAX_COMMENT_LEN		20	// Max length of start-of-line comment string deduced from buffer start
#define	TAGSIZE			200	// How big a tag can be
#define	MAX_QUESTION		100	// How many times to ask a question before giving up


// Size of buffer that holds a buffer name
#define	BUFFER_NAME_SIZE		((NBUFN + 2) * UTF8_MAX_BYTES + 1)

/* Colours internally use numbers - give them names.
 * The values are those used by termcap - ANSI values.
 * Other systems will need to map them.
 *
 * British spelling neatly avoids a clash with constants in <ncurses.h>
 * All the real colour values are non negative. But COLOUR_SYS_DEF is allowed. See tcap.c
 */
#define	COLOUR_MIN		-1
#define	COLOUR_SYS_DEF		-1	// Try to set this and end up with Original - system default
#define	COLOUR_BLACK		0
#define	COLOUR_RED		1
#define	COLOUR_GREEN		2
#define	COLOUR_YELLOW		3
#define	COLOUR_BLUE		4
#define	COLOUR_MAGENTA		5
#define	COLOUR_CYAN		6
#define	COLOUR_WHITE		7
#define	COLOUR_MAX		7

/* These are the values returned by the functions that check and
 * manipulate the DOS buffer mode - ie removal/adding of \r (RETCHAR) from
 * line ends on read/write.
 * Carefully chosen values to simplify return value calculation in functions like IsDosBuffer()
 */
#define	RETC_WITH		1				// Some with \r or converted
#define	RETC_WITHOUT		2				// Some without \r or not converted
#define	RETC_EMPTY		0				// Buffer empty
#define	RETC_ALL		RETC_WITH			// All lines affected
#define	RETC_NONE		RETC_WITHOUT			// None removed/added
#define	RETC_SOME		(RETC_WITH | RETC_WITHOUT)	// Some, but not all, lines affected
#define	RETC_ERROR		4				// Some error - out of memory on adding \r



/* The following used in k_code in KEYTAB, stored in a KeyCode
 * These must all fall out of the UniChar range - max 0x10FFFF
 * This is within what a long can hold.
 * It might be within what an int can hold, see UniChar typedef,
 * update there if something larger than the current largest is defined.
 *
 * CNTRL is to be used with keyboard specials such as arrow & function keys.
 * A control character (0x00-> 0x1f) is stored exactly as its value.
 *
 * If compiling for the wimsy of Unicode characters greater than U+10FFFF
 * then these constants might need to be long long. This needs a compiler
 * that implements C99. See also utf8.h and the UniChar definition below.
 */
#if	CONFIG_MAX_UNICODE >= 0x01000000
#define	CNTRL			0x0100000000LL	// Control flag, or'ed in
#define	META			0x0200000000LL	// Meta flag, or'ed in
#define	CTLX			0x0400000000LL	// ^X flag, or'ed in
#define	SPEC			0x0800000000LL	// special key (function keys)
#define	INTL			0x1000000000LL	// Internal - not from a key
#define	SHIFT			0x2000000000LL	// Shifted special, eg home
#define	K_CODE_STATUS		0x7000000000LL	// Special status type things
#define	K_CODE_MASK		0xFF00000000LL	// Mask to remove the above flags
#else
#define	CNTRL			0x01000000	// Control flag, or'ed in
#define	META			0x02000000	// Meta flag, or'ed in
#define	CTLX			0x04000000	// ^X flag, or'ed in
#define	SPEC			0x08000000	// special key (function keys)
#define	INTL			0x10000000	// Internal - not from a key
#define	SHIFT			0x20000000	// Shifted special, eg home
#define	K_CODE_MASK		0xFF000000	// Mask to remove the above flags
#define	K_CODE_STATUS		0x70000000	// Special status type things
#endif

// Special keycode values - do not correspond to a real key
#define	KEY_TIMEOUT		(K_CODE_STATUS | 1)	// Timeout on terminal read
#define	KEY_NONE		(K_CODE_STATUS | 2)	// No key. in the context of TTungetc()
#define	KEY_END_OF_FILE		(K_CODE_STATUS | 3)	// End of file - when term not initialised

/* Special internal values for function keys, used with UNIX termcap
 * decoding of the keys.
 * The encoding isn't random, the key will be printed out as FN? where ?
 * is the letter following, so let's make it something vaguely mnemonic.
 */
#define	UTC_UP			(SPEC | 'U')	// Up arrow
#define	UTC_DOWN		(SPEC | 'D')	// Down arrow
#define	UTC_LEFT		(SPEC | 'L')	// Left arrow
#define	UTC_RIGHT		(SPEC | 'R')	// Right arrow
#define	UTC_PGUP		(SPEC | 'B')	// Back a page
#define	UTC_PGDOWN		(SPEC | 'F')	// Forward page
#define	UTC_HOME		(SPEC | 'S')	// Start of line
#define	UTC_END			(SPEC | 'E')	// End of line
#define	UTC_DEL			(SPEC | 'd')	// Delete forwards
#define	UTC_INS			(SPEC | 'I')	// Toggle insert mode

/* Keypad in non NumLock mode
 * Encoding is the digit engraved, and 'e' for enter
 */
#define	UTC_KP_ENT		(SPEC | 'e')	// Keypad Enter
#define	UTC_KP_UL		(SPEC | '7')	// Keypad upper left
#define	UTC_KP_UR		(SPEC | '9')	// Keypad upper right
#define	UTC_KP_CEN		(SPEC | '5')	// Keypad center
#define	UTC_KP_LL		(SPEC | '1')	// Keypad lower left
#define	UTC_KP_LR		(SPEC | '3')	// Keypad lower right

// Allocate from 1 up:
#define	UTC_BAKTAB		(SPEC | 1  )	// Back tab

/* Encoding of Unix Termcap Function keys.
 * Legacy names like UTC_F1 now gone.
 *
 * If the bottom 8 bits are zero, then the function key number is
 * 8 bits above that.
 */
#define	UTC_FN_KEY(n)		(SPEC | ((n) << 8))


#ifdef	FALSE
#undef	FALSE
#endif
#ifdef	TRUE
#undef	TRUE
#endif

#define	FALSE			0	// False, no, bad, etc.
#define	TRUE			1	// True, yes, good, etc.
#define	ABORT			2	// Death, ^G, abort, etc.
#define	FAILED			3	// Not-quite fatal false return

#define	STOP			0	// keyboard macro: not in use
#define	PLAY			1	//		   playing
#define	RECORD			2	//		   recording

/*	Macro directive definitions
 *
 * Values below must be reflected in ordering in dname[] & dname_len[]
 *
 * BEWARE: These are numbered starting from 1, this is reflected where
 * they are used - see dobuf().
 * The reason is that a non-zero number is cached in a line Note
 */

#define	DNOTHING		0	// Not a !something
#define	DIF			1
#define	DELSE			2
#define	DENDIF			3
#define	DGOTO			4
#define	DRETURN			5
#define	DENDM			6
#define	DWHILE			7
#define	DENDWHILE		8
#define	DBREAK			9
#define	DFORCE			10
#define	DCONTINUE		11
#define	DELSEIF			12

#define	NUMDIRS			12

/*
 * PTBEG, PTEND, FORWARD, and REVERSE are all toggle-able values for
 * the scan routines.
 * Beware: scanner() does exclusive-or on these values.
 */
#define	PTBEG			0	// Leave the point at the beginning on search
#define	PTEND			1	// Leave the point at the end on search
#define	FORWARD			0	// forward direction
#define	REVERSE			1	// backwards direction

#define	FIOSUC			0	// File I/O, success.
#define	FIOFNF			1	// File I/O, file not found.
#define	FIOEOF			2	// File I/O, end of file.
#define	FIOERR			3	// File I/O, error.
#define	FIOMEM			4	// File I/O, out of memory
#define	FIOFUN			5	// File I/O, eod of file/bad line -- no \n at end of last line

/* Two variables: thisflag & lastflag
 * These are used when the action of a command depends on the previous command.
 * CFCPCN	Last was movement up or down. Preserve desired column even on shorter lines.
 *		This is kept in curwp->w_goalCol
 *		Also used by the likes of move-window-right
 * CFMARGIN0	No margin on adjusting w_leftoffset.
 *		The point is that we let "." get to the window edge to avoid left/right shift,
 *		but will do the shift if needed to keep "." visible.
 *		This is implied by CFCPCN - which is more extreme as "." can be off screen.
 * CFDELETE	Multiple characters have been deleted - killed, eg: kill-to-end-of-line, kill-region
 *		Contrast to single characters, eg: delete-next-character without a count.
 * CFYANK	Controls walking through the kill ring with yank & reyank.
 */
#define	CFCPCN			0x0001	// Last command was Line/Page Up/Down, move-window*left/right
#define	CFMARGIN0		0x0002	// Left/right scroll margin is 0
#define	CFDELETE		0x0004	// Last command was a kill
#define	CFYANK			0x0008	// Last command was yank

#if	UNIX
#define	PATHCHR			':'
#define	PATHSTR			":"
#define	CTAGS			1	// Tag files for source code
#else
#define	PATHCHR			';'
#define	PATHSTR			";"
#define	CTAGS			0	// Tag files for source code
#endif


/* How many bytes needed to convert an integer to a string.
 * These days: enough for a long. This is very generous.
 */
#define	INTWIDTH		(sizeof(long) * 3)

/* Min # lines for a window - to be workable:
 * 1 line text
 * 1 line mode
 * Min # lines for the screen - to be workable:
 * 1 window (ie 2 lines)
 * 1 line command/prompt line
 *
 * The minimum # columns calculation is much more nebulous. Things like:
 * * minimum mode line info
 * * line number + some text
 * * what we scroll left/right by - this just works
 * Since we don't do line numbers (yet): set this to 1
 */
#define	MIN_WINDOW_TEXT_LINES	1			// Min # lines text for a buffer window
#define	MIN_WINDOW_LINES	(MIN_WINDOW_TEXT_LINES + 1)	// Min # lines for a buffer window - to be workable
#define	MIN_WINDOW_COLS		1			// Min # cols for a buffer window - to be workable
#define	WINDOW_SCROLL_RATIO	10			// Scroll when cursor gets this part of the edge
#define	MIN_SCREEN_LINES	(MIN_WINDOW_LINES + 1)	// +1 for the message line
#define	MIN_SCREEN_COLS		MIN_WINDOW_COLS		// Min # cols for the screen - to be workable
#define	MIN_WINDOW_SHOW_LINE_END 6			// Show EOL marker if window at least this wide

/* The size of screen that SHALL be used is: AllocLines, AllocColumns
 * This will respect the minimum sizes above - even if so doing creates
 * a mess on tiny screens.
 */

#define	LINES_EMPTY_BUFFER_END	3			// # lines to show under "." after end-of-file & similar

/* Minimum and maximum sizes of a tab.
 * The current size is in Tabsize and $tabwidth
 * This really ought to be buffer specific.
 */
#define	DEF_TAB			8	// Detault tab size
#define	PHP_TAB			4	// PHP tab size - really the indent factor
#define	TAB_MINIMUM		1	// Can't get smaller
#define	TAB_MAXIMUM		300	// Arbitrary



/* There is a difference between characters & bytes.
 * A byte is the underlying storage used in files and also l_text in LINE.
 * A character is a Unicode point, a number U+0 .. U+10FFFF,
 * it is stored in one or more bytes.
 *
 * Earlier versions used the same for both. Worse: a straight 'char' was
 * used which, potentially, being signed is limited to U+0 .. U+7F without
 * problems needing to having to go '& 0x7F' and the such.
 *
 * Note also that UniChar can be assigned UNISTR_END as a non-value/end-of-string.
 * Having a type 'UniChar' is also partly documentation, better than 'long',
 * also the compiler helps.
 *
 * There is also a difference between characters and key codes.
 * A key code is what is read from the terminal. In addition to Unicode
 * characters is can also describe things like arrow & function keys.
 *
 * A key code will have something ORed in, in mask K_CODE_MASK.
 * Eg CNTRL, META, CTLX, SPEC, INTL. This also allows a command sequence (eg M-B)
 * to be stored in one key code value.
 *
 * Save on storage requirements if we can; maybe use 32 bit int rather than
 * 64 bit long. Need 64 bits if we handle non standard maximum Unicode character.
 *
 * UNICHAR_LMOD and KEYCODE_LMOD are the length modifiers needed when printing
 * with printf() and scanf() to generate %x or %lx.
 * Alternatively, when printing, cast to long and use %lX or %ld.
 *
 * KeyCode_u is there to suppress warnings from the compiler with -Wconversion.
 * These happen with bit bashing operations which are done on unsigned quantities.
 * Do not used except in casts in bit bashing operations.
 */
#if	CONFIG_MAX_UNICODE >= 0x01000000	// See above where CNTRL is defined
typedef	long long	KeyCode;		// Storage for command key from terminal
typedef	unsigned long long KeyCode_u;		// Bit bashing keycode type
#define	KEYCODE_LMOD	"ll"
#else
#if	INT_MAX > SHIFT
typedef	int		KeyCode;		// Storage for command key from terminal
typedef	unsigned int	KeyCode_u;		// Bit bashing keycode type
#define	KEYCODE_LMOD	""
#else
typedef	long		KeyCode;		// Storage for command key from terminal
typedef	unsigned long	KeyCode_u;		// Bit bashing keycode type
#define	KEYCODE_LMOD	"l"
#endif
#endif	// CONFIG_MAX_UNICODE >= 0x01000000

#if	CONFIG_MAX_UNICODE > INT_MAX		// Is an int not big enough to take a Unicode ?
typedef	long		UniChar;		// Storage for Unicode code points
#define	UNICHAR_LMOD	"l"
#else
typedef	int		UniChar;		// Storage for Unicode code points
#define	UNICHAR_LMOD	""
#endif

#define	UNISTR_END	(-1)			// String terminator for vtputUniStr()

#define	CHAR		unsigned char		// Legacy, do not use
#define	BYTE		unsigned char		// For underlying storage

/* Storage unit for a screen usage map cell.
 * Maps are used in window.c as a way of finding out which window is in a particular direction.
 * Typedef on the very off chance that we need to increase the max value beyond 255 (UCHAR_MAX).
 */
typedef	BYTE	MapCell;
#define	MAX_WINDOW_MAP	UCHAR_MAX		// Max value in a MapCell

/* Options to getstring(), mlreply() & mlprompt() ... Values ORed together.
 * The lower bits make up a type, this is as some have special lookup mechanisms,
 * eg file or buffer modes. Only give one type, types are not bitwise distinct.
 * These lengths may imply a maximum length of input string for mlprompt().
 * Update MLmaxLengths[] if another type is invented.
 */
#define	MLNOOPT			0x0000		// No options
#define	MLTYPEMASK		0x000F		// Type/use of input string
//				0x0000		// Option must not be this value else VAL* might change type
#define	MLXFILE			0x0001		// Entering a filename
#define	MLXBUF			0x0002		// Entering a buffer name
#define	MLSHELLCOMM		0x0003		// Entering a shell command
#define	MLXMODE			0x0004		// Entering a mode
#define	MLTAG			0x0005		// Choosing a C tag
#define	MLWOPTION		0x0006		// Entering a window option
#define	MLEMACSCOMMAND		0x0007		// Entering an Emacs command
#define	MLEMACSVAR		0x0008		// Entering a variable name
#define	MLINTEGER		0x0009		// Entering an integer (long)
#define	MLSTRING		0x000A		// Non specific string

// The following can be ORed with one of the above:
#define	MLUPPER			0x0010		// Convert chars to uppercase
#define	MLLOWER			0x0020		// Convert chars to lowercase
#define	MLNOSPACE		0x0040		// No spaces allowed
#define	MLNOECHO		0x0080		// Dont echo the string
#define	MLDEFISLAST		0x0100		// Default is last result
//#define	MLDEFINBUF	0x0100		// Default is in o/p buffer
#define	MLLN0END		0x0200		// \n at pos 0 is end input
#define	MLNOMAP			0x0400		// Suppress tty key mapping
#define	MLEOLESC		0x0800		// Use ESC not \n as end of line
#define	MLU8LEN			0x1000		// Length is UTF-8 chars, not bytes
#define	MLINTERACTIVE		0x2000		// Always prompt even if clexec
#define	MLEMPTYOK		0x4000		// An empty/no value is acceptable
#define	MLSELECTONLY		0x8000		// Only select from choices, to text entry


// Values of bfind() cflag - OR together. Also bfindCheck()
#define	BFNNONE			0x00		// Nothing special
#define	BFNCREATE		0x01		// If the buffer is not found then create it
#define	BFNACTIVE		0x02		// Mark a newly created buffer as active
#define	BFNVIEW			0x04		// Make the buffer read only

/*	Possible values of cmode					*/

#define	CMDOFF			0		// Turned off
#define	CMDON			1		// Turned on
#define	CMDEXT			2		// Extended CMODE

/*	Macro argument token types					*/

#define	TKBAD			-1		// Invalid/does not exist
#define	TKNUL			0		// End-of-string
#define	TKARG			1		// Interactive argument
#define	TKBUF			2		// Buffer argument
#define	TKVAR			3		// User variables
#define	TKENV			4		// Environment variables
#define	TKFUN			5		// Function....
#define	TKDIR			6		// Directive
#define	TKLBL			7		// Line label
#define	TKLIT			8		// Numeric literal
#define	TKSTR			9		// Quoted string literal
#define	TKCMD			10		// Command name
#define	TKCHR			11		// Single char read ahead
#define	TKARRAY			12		// Array
#define	TKHASH			13		// Hash
#define	TKCOMMA			14		// Array/hash separator

/*	Internal defined functions					*/
#ifdef	abs
#undef	abs
#endif

/* Note that the names of the macros have a capital letter, this is because
 * they combine the isascii().
 *
 * BEWARE: when/if extending to non-ascii letters. Isletter() is used
 * to restrict to ASCII letters. So may need an ASCII restricting version
 * and any Unicode code-point versions.
 */
#if	HAVE_CTYPE
// Character classification is much easier with <ctype.h>
#include <ctype.h>

#define	Isletter(c)	(isascii(c) && isalpha(c))
#define	Islower(c)	(isascii(c) && islower(c))
#define	Isupper(c)	(isascii(c) && isupper(c))
#define	Isdigit(c)	(isascii(c) && isdigit(c))
#define	Isxdigit(c)	(isascii(c) && isxdigit(c))
#define	Isalnum(c)	(isascii(c) && isalnum(c))
#define	Iscntrl(c)	(isascii(c) && iscntrl(c))

/* Toggle the case of a letter, this is more difficult/wordy.
 * Assume that it is an alpha.
 */
#define	CHCASE(c)	(Isupper(c) ? tolower(c) : toupper(c))

#if	ASCII
/* Display of controls.
 * Have a printable character make it the control character.
 * Take the control character and make it printable.
 * For the purposes of this: DEL is deemed C-?
 * ASCII only - not Unicode
 */
#define	ChrToCtrl(c)		((c) & 0x1F)	// Convert printable to control
#define	ChrFromCtrl(c)		((c) | 0x40)	// Make a control character something that can be printed

// Fails if ever we use EBCDIC
#endif	// ASCII

#else	// HAVE_CTYPE
/* DIFCASE represents the integer difference between upper
 *  and lower case letters.  It is an xor-able value, which is
 *  fortunate, since the relative positions of upper to lower
 * case letters is the opposite of ascii in ebcdic.
 */

// **************** BUG this does not work if c is Unicode
#define	toupper(c)	((c) ^ DIFCASE)
#define	tolower(c)	((c) ^ DIFCASE)
#define	CHCASE(c)	((c) ^ DIFCASE)	// Toggle the case of a letter.

#if	ASCII
#define	DIFCASE		0x20
#define	Isletter(c)	(('a' <= (c) && 'z' >= (c)) || ('A' <= (c) && 'Z' >= (c)))
#define	Islower(c)	(('a' <= (c) && 'z' >= (c)))
#define	Isupper(c)	(('A' <= (c) && 'Z' >= (c)))
#define	Isdigit(c)	((c) >= '0' && (c) <= '9')
#define	Isxdigit(c)	(Isdigit(c) && ((c) >= 'a' && (c) <= 'f') || ((c) >= 'A' && (c) <= 'F'))
#define	Iscntrl(c)	((c) >= '\0' && (c) < ' ')
#define	Iscntrl(c)	(((c) & ~0x1F) == 0)

#endif

#if	EBCDIC
#define	DIFCASE		0x40
#define	Isletter(c)	(('a' <= (c) && 'i' >= (c)) || ('j' <= (c) && 'r' >= (c)) || ('s' <= (c) && 'z' >= (c)) || ('A' <= (c) && 'I' >= (c)) || ('J' <= (c) && 'R' >= (c)) || ('S' <= (c) && 'Z' >= (c)))
#define	Islower(c)	(('a' <= (c) && 'i' >= (c)) || ('j' <= (c) && 'r' >= (c)) || ('s' <= (c) && 'z' >= (c)))
#define	Isupper(c)	(('A' <= (c) && 'I' >= (c)) || ('J' <= (c) && 'R' >= (c)) || ('S' <= (c) && 'Z' >= (c)))
#define	Isdigit(c)	((c) >= '0' && (c) <= '9')
#define	Isxdigit(c)	(Isdigit(c) && ((c) >= 'a' && (c) <= 'f') || ((c) >= 'A' && (c) <= 'F'))
#define	Iscntrl(c)	((c) >= '\0' && (c) < ' ')
#endif

#define	Isalnum(c)	(Isletter(c) && Isdigit(c))

#endif	// HAVE_CTYPE

// Preferably use utf8_isvalid():
#define	IsChar(c)	(((CHAR)(c)) == (c))	// Fit in a byte ?


/* The second argument to named commands is the count which is integral.
 * How big is that: int or long ?
 * Define CNT_MAX - maximum Cnt value
 */
#if	ARG_IS_LONG
typedef	long		Cnt;
#define	atoCnt(str)	atol(str)
#define	CNT_MAX		LONG_MAX
#else
typedef	int		Cnt;
#define	atoCnt(str)	atoi(str)
#define	CNT_MAX		INT_MAX
#endif


// Debugging for terminfo/termcap, sequences generated & what is read from the terminal
// -K command line option. Some need more -K to generate something.
// When used these SHOULD NOT have a ; after them.
#if	TERMDEBUG
#define	TERMTRACE(x)     if(TermDebFD) fprintf x;
#define	TERMOUTTRACE(x)  if(TermDebFD && TermDebug > 1) fprintf x;
#define	TERMOUTTRACE2(x) if(TermDebFD && TermDebug > 2) fprintf x;
#define	TERMOUTTRACE3(x) if(TermDebFD && TermDebug > 3) fprintf x;
#define	TERMOUTTRACE4(x) if(TermDebFD && TermDebug > 4) fprintf x;
#define	TERMOUTTRACE5(x) if(TermDebFD && TermDebug > 5) fprintf x;
#else
#define	TERMTRACE(x)     ;
#define	TERMOUTTRACE(x)  ;
#define	TERMOUTTRACE2(x) ;
#define	TERMOUTTRACE3(x) ;
#define	TERMOUTTRACE4(x) ;
#define	TERMOUTTRACE5(x) ;
#endif

/* What is stored in the video/screen buffers.
 * See display.c for a discussion about this.
 * That is also the only place where these should be used.
 */
typedef	unsigned char	VidAttr;	// Attributes: standout, colour, ...

#if	DISPWIDE
typedef	UniChar	VideoC;			// Unicode characters - possibly >= U+100
#else
typedef	char	VideoC;			// ASCII characters
#endif

/* Some characters that always have special meaning
 */
#define	CharTab			'\t'	// Tab character
#define	CharEscape		'\33'	// Escape character - 0x1B
#define	CharDelete		'\177'	// Delete character - 0x7F
#define	CharRevSearch		'\x12'	// ^r Reverse Search
#define RETCHAR			'\r'	// Carriage return
#define BEL			'\a'	// BEL (alarm) character
#define	BindNameSpace		"SPC"	// How ' ' can be named in key bindings
#define	BindNameDelete		"DEL"	// How 0x7F (^?) can be named in key bindings
#define	BindNameEscape		"ESC"	// How 0x1B (\033) can be named in key bindings
#define	BindNameCircumflex	"CIRC" // How circumflex ('^') can be named in key bindings
#define	BindNameTab		"TAB"	// How tab ('\t') can be named in key bindings
#define	BindNameReturn		"RET"	// How return ('\r') can be named in key bindings

/* DEL is not obvious: don't show as a literal delete character
 * Could have shown as ^? or delete.
 * delete could be an escape sequence that termcap names
 * Beware: what this key sends can be different in visual mode
 * or the keyboard initialised as can the backspace key.
 */

/* A structure that maps charactes with a special representation in binding.
 * Terminated by bs_rep == NULL.
 */
typedef	struct	{
	char*	bs_rep;		// The special representation, eg DEL
	BYTE	bs_char;	// The character, eg \177
	char	bs_flag;
#define	BS_ctlOK	0x01	// C-bs_rep is allowable
#define	BS_prefCTL	0x02	// Show control after prefix, not the bs_rep
				// eg C-x C-m not C-x RET
} BindSpecChrs;

/* CountMode values to getdiscol() & getdoto() */
#define	DisCol_Char	0	// Count normal characters, meta/ctrl as M- & ^
#define	DisCol_MetaChar	1	// Count normal chars, meta/ctrl as attributes underline/standout
#define	DisCol_WP	2	// Count like a word processor
#define	DisCol_HEX	3	// Count display is in hex

/* Other characters:
 */
#define	CharMax		UTF8_MAX_CODE_POINT	// Topmost char that we can represent

// Flags for writeout() - can be ORed together
#define	WOUT_NONE	0x00	// Nothing special - just write the buffer
#define	WOUT_APPEND	0x01	// Append to the file
#define	WOUT_REGION	0x02	// Write the region


/*	Dynamic RAM tracking and reporting redefinitions	*/

#if	RAMSIZE
#define	malloc	allocate
#define	free	release
#endif

/* De-allocate memory always on exit (if the operating system or
 * main program can not)
 */
#if	CLEAN && !CALLED
#define	exit(a)	cexit(a)
#endif



/* How long things are.
 * Special types partly for documentation, but also allows an easy change
 * if needs be. Its use is limited ATM, converting from long slowly.
 * These types cannot be unsigned as a -ve value is, on occasion, used to
 * mean counting from bottom/end.
 *
 * BEWARE: Where any of these is used with a printf() it should be cast
 * to long and converted with %ld.
 * Although these are long, the type might be a compile time int instead.
 * Especially with LineLn - which would save 8 bytes per LINE.
 *
 * Type LineLn is what stores a line length.
 *
 * Really need a #define string to get printf to convert properly.
 *
 * # bytes & # chars in a buffer or line length might not be the same thing,
 * but we can use the same type to store either.
 *
 * There is an assumption that any of these can be cast to size_t without loss
 * of any value that will be used.
 * This happens when one of these is passed to a function that takes size_t
 * as an argument, eg: bytecopy(). The fact that the IEEE defines the
 * length copied as size_t suggests that this is safe.
 */
typedef	long		LineLn;		// Line length (bytes or chars)
typedef	long		BufLns;		// # lines in a buffer
typedef	long		BufChrs;	// # chars or bytes in a buffer



/* A type in which we store text.
 *
 * This is used in LINEs in BUFFERs and also as free standing TEXT variables.
 *
 * This type can contain the NUL byte as length is by l_used.
 * The end of line is not stored as a byte; it is implied.
 *
 * t_flags does contain 2 values used in LINE: L_NMARK & L_MARKED
 *
 * L_UTF8 is set if a multibyte character has been seen in the text.
 * This might be wrong if all mb characters have been deleted, the correctness
 * of this flag is not checked after every deletion. It is set on every
 * insertion.
 *
 * t_notes is for non-TEXT/LINE flags.
 * Set to 0 when the TEXT is created.
 *
 * TEXT can have a trailing NUL byte in t_text. This will be preserved but does
 * not count as part of the t_used size. This is to make it easy for TEXT to be
 * used as, eg, a file name to fopen().
 *
 * L_NUL must be 1 - see textMoreBytes() - sorry!
 */
typedef struct	{			// TEXT struct
	BYTE*		t_text;		// A bunch of characters.
	LineLn		t_size;		// Allocated size
	LineLn		t_used;		// Used size
	short		t_notes;	// Non specified value held in the line
	BYTE		t_flags;
#define	L_NUL		0x01		// There is a NUL at line end
#define	L_NMARK		0x02		// There is a window numeric mark on the line
#define	L_MARKED	0x04		// Line is alpha marked if this set
#define	L_UTF8		0x08		// Line may contain UTF8 chars
} TEXT;

/* All text in a buffer is kept in circularly linked lists of LINE structures.
 *
 * These begin at the header line (which is the empty line beyond the end of the
 * buffer). This line is pointed to by the BUFFER.b_linep.
 *
 * l_text.t_flags L_MARKED/L_NMARK is set if there is a alpha/numeric mark on
 * the LINE - in its buffer. Use lmarked()/lisNmarked().
 *
 * Beware: l_text.t_note will not be preserved on buffer refresh or if two
 * LINEs are joined. Useful in history buffers.
 *
 * l_text is a separate defined struct as things like playLine & variables do
 * not need the next/prev pointers; it also helps with correct argument usage.
 *
 * Putting TEXT first in LINE results in smaller object size, think: lgetc() macro
 * use of &ltxt(lp).
 */
typedef struct	line {							// struct line
	TEXT		l_text;						// Text of the LINE
	struct line*	l_fp;						// Link to the next line
	struct line*	l_bp;						// Link to the previous line
} LINE;

// Accessor macros/in-line-functions -- made it easy when the structures changed:
#define	ltxt(lp)		((lp)->l_text)				// TEXT within a LINE*

// Lots of pairs TEXT/LINE:
#define	tgetch(tp, n)		((tp)->t_text[(n)])			// Get byte at offset n in the TEXT
#define	lgetc(lp, n)		(tgetch(&ltxt(lp), (n)))		// Get byte at offset n in the LINE

#define	tgetcaddr(tp, n)	((tp)->t_text + (n))			// Address of current char
#define	lgetcaddr(lp, n)	(tgetcaddr(&ltxt(lp), (n)))		// Address of current char

#define	tputc(tp, n, c)		((tp)->t_text[(n)]=(c))			// Set a byte at offset n in the TEXT
#define	lputc(lp, n, c)		(tputc((&ltxt(lp)), (n), (c)))		// Set a byte at offset n in the LINE

#define tused(tp)		((tp)->t_used)				// Bytes in use
#define lused(lp)		(tused(&ltxt(lp)))			// Bytes in use

#define tsize(tp)		((tp)->t_size)				// Allocated size
#define lsize(lp)		(tsize(&ltxt(lp)))			// Allocated size

#define ttext(tp)		((tp)->t_text)				// Address of BYTE* in TEXT*
#define ltext(lp)		(ttext(&ltxt(lp)))			// Address of BYTE* in LINE*

#define	tisUtf8(tp)		((tp)->t_flags & L_UTF8)		// True if UTF8 flag set
#define	lisUtf8(lp)		(tisUtf8(&ltxt(lp)))			// True if UTF8 flag set

#define	tisNUL(tp)		((tp)->t_flags & L_NUL)			// Is 1 or 0 by dint of value of L_NUL
#define	lisNUL(lp)		(tisNUL(&ltxt(lp)))			// Is 1 or 0 by dint of value of L_NUL

#define	tnotes(tp)		((tp)->t_notes)				// Notes in TEXT
#define	lnotes(lp)		(tnotes(&ltxt(lp)))			// Notes in LINE

#define	tbend(tp)		(tgetcaddr((tp), tused(tp)))		// Address of byte just Beyond/after text end
#define	lbend(lp)		(lgetcaddr((lp), lused(lp)))		// Address of byte just Beyond/after line end


// LINE only macros:
#define	lforw(lp)		((lp)->l_fp)				// Next line
#define	lback(lp)		((lp)->l_bp)				// Previous line
#define lmarked(lp)		(ltxt(lp).t_flags & L_MARKED)		// Line has an alpha mark
#define lisNmarked(lp)		(ltxt(lp).t_flags & L_NMARK)		// Line has a numeric mark

static	inline	BYTE	lgetc_t(LINE* lp, LineLn n) {			// Like lgetc(), but returns '\n' if too far in the line
	if(n < lused(lp))
		return(lgetc(lp, n));
	return((BYTE)'\n');
}

/* A macro does not work well with if/while.
 * Modern compilers optimise out any overhead of inline functions.
 */
static	inline	void	tsetUtf8(TEXT* tp)		{ (tp)->t_flags |=		L_UTF8;	 }	// Set UTF8 flag
static	inline	void	tunsetUtf8(TEXT* tp)		{ (tp)->t_flags &=	 (BYTE)~L_UTF8;	 }	// Unset UTF8 flag

static	inline	void	lsetUtf8(LINE* lp)		{ tsetUtf8(&ltxt(lp)); }			// Set UTF8 flag
static	inline	void	lunsetUtf8(LINE* lp)		{ tunsetUtf8(&ltxt(lp)); }			// Unset UTF8 flag

static	inline	void	lsetMark(LINE* lp)		{ ltxt(lp).t_flags |=		L_MARKED; }	// set marked flag
static	inline	void	lunsetMark(LINE* lp)		{ ltxt(lp).t_flags &=	 (BYTE)~L_MARKED; }	// Unset marked flag
static	inline	void	lsetNMark(LINE* lp)		{ ltxt(lp).t_flags |=		L_NMARK;  }	// set numeric marked flag
static	inline	void	lunsetNMark(LINE* lp)		{ ltxt(lp).t_flags &=	 (BYTE)~L_NMARK;  }	// Unset numeric marked flag

static	inline	void	lsetforw(LINE* lp, LINE* f)	{ (lp)->l_fp = (f); }				// Set forward line
static	inline	void	lsetback(LINE* lp, LINE* b)	{ (lp)->l_bp = (b); }				// Set backward line

static	inline	void	tsetsize(TEXT* tp, LineLn s)	{ (tp)->t_size = (s); }				// Set size
static	inline	void	lsetsize(LINE* lp, LineLn s)	{ tsetsize(&ltxt(lp), (s)); }			// Set size

// This could increment t_size but does not. The byte may be 'reclaimed' when lineMoreBytes() is next used
static	inline	void	tunsetNUL(TEXT* tp)		{ (tp)->t_flags &=	   (BYTE)~(L_NUL); }	// Unset L_NUL flag
static	inline	void	lunsetNUL(LINE* lp)		{ tunsetNUL(&ltxt(lp)); }			// Unset L_NUL flag

// See also function lsetNUL()
static	inline	void	tsetNotes(TEXT* tp, short nv)	{ (tp)->t_notes = (nv); }			// Set TEXT's notes
static	inline	void	lsetNotes(LINE* lp, short nv)	{ tsetNotes(&ltxt(lp), (nv)); }			// Set LINE's notes


/* Set a LINE's used length.
 * Also replaces a NUL that might have got lost.
 * BEWARE: this does NOT reallocate or check the need,
 * you must be sure that the l_size is big enough.
 * This does not check for a negative length - must not have.
 */
static	inline	void	tsetused(TEXT* tp, LineLn u)	{
	tp->t_used = u;
	if(tisNUL(tp))
		tputc(tp, u, '\0');
}
#define	lsetused(lp, u)					tsetused(&ltxt(lp), (u))

/* Special case of clearing out - set empty.
 * Always put the NUL byte, even if not needed. There will always be room.
 */
static	inline	void	tsetempty(TEXT* tp)	{
	tp->t_used = 0;
	tputc(tp, 0, '\0');
}


// Temp shims when introducing TEXT alongside LINE
#define	lcatm(l, s, len)	tcatm(&ltxt(l), (s), (len))
#define	linsm(l, o, s, len)	tinsm(&ltxt(l), (o), (s), (len))
#define	linsl(l, o, s)		tinst(&ltxt(l), (o), &ltxt(s))
#define	linst(l, o, s)		tinst(&ltxt(l), (o), (s))
#define	tinsl(l, o, s)		tinst((l),      (o), &ltxt(s))
#define	linss(l, o, s)		tinss(&ltxt(l), (o), (s))
#define	lcpym(l, s, len)	tcpym(&ltxt(l), (s), (len))
#define	lcpyl(d, s)		tcpyt(&ltxt(d), &ltxt(s))
#define	lcpyt(d, s)		tcpyt(&ltxt(d), (s))
#define	tcpyl(d, s)		tcpyt((d),      &ltxt(s))
#define	lcatl(d, s)		tcatt(&ltxt(d), &ltxt(s))
#define	lcatt(d, s)		tcatt(&ltxt(d), (s))
#define	tcatl(d, s)		tcatt((d),      &ltxt(s))
#define	lcpys(d, s)		tcpys(&ltxt(d), (s))
#define	lcpyc(d, c)		tcpyc(&ltxt(d), (c))

#define	lcats(d, s)		tcats(&ltxt(d), (s))
#define	lcatc(d, c)		tcatc(&ltxt(d), (c))
#define	linsc(d, o, c)		tinsc(&ltxt(d), (o), (c))
#define	ldeln(l, o, len)	tdeln(&ltxt(l), (o), (len))
#define	ltrim(l)		ttrim(&ltxt(l))
#define	linecmp(l1, l2)		textcmp(&ltxt(l1), &ltxt(l2))

#define	lineMoreBytes(l, n)	textMoreBytes(&ltxt(l), (n))
#define	lsetNUL(l)		tsetNUL(&ltxt(l))

#define	insStrLine(l, o, t, len, c)	insStrText(&ltxt(l), (o), (t), (len), (c))


/* Last (empty) line.
 * Eg EofLine(curwp->w_dotp).
 * Last line that contains something: EofLine(curwp->w_dotp->l_fp).
 *
 * Also end of narrow.
 * Also start of narrow.
 */
#define	EofLineB(bp, lp)	((lp) == (bp)->b_linep)		// Last in buffer
#define	SonLineB(bp, lp)	((lp) == (bp)->b_btopline)	// First in narrow
#define	EonLineB(bp, lp)	((lp) == (bp)->b_abotline)	// Last in narrow

#define	EofLine(lp)		EofLineB(curbp, (lp))		// Last in current buffer
#define	SonLine(lp)		SonLineB(curbp, (lp))		// First in current narrow
#define	EonLine(lp)		EonLineB(curbp, (lp))		// Last in current narrow

/* Array of values.
 *
 * a_alloc	Number of elements allocated
 * a_used	Number of elements in use
 * a_elems	Vector of a_alloc VAL*
 *		When a_alloc is 0 this might be NULL
 *
 * When a_alloc is > a_used the non used elements will have a valid VAL* that
 * might be V_UNDEF.
 *
 * Index from 0.
 */
typedef	struct	{
	struct	VAL**	a_elems;
	int		a_alloc;	// Allocated size of a_elems
	int		a_used;		// How many a_elems used
} Array;


/* A hash is an array of pairs: key, value
 *
 * h_size	number of hash elements in h_elems
 *		A hash element is (key, value), so the number of
 *		VAL*s in h_elems is 2 * h_size
 * h_used	number of hash elements in h_elems that are used
 *		Same comment as h_size about number in h_elems
 *
 * Maybe later use the hashing functions for fast lookup
 */
typedef	struct	{
	struct	VAL**	h_elems;
	int		h_alloc;	// Allocated size of h_elems
	int		h_used;		// How many h_elems used
} Hash;


/* Values - could be in a variable or expression value.
 * There is over lap with ENVVAR (evar.h)
 *
 * This VAL for prompting, eg mlreply(), as when in a macro arguments can be
 * more than just a string. The interesting ones are arrays & hashs.
 *
 * If V_BOOL: still stored in v_int, but will convert TRUE/FALSE
 * ON/OFF "1"/"0" to 1/0 when getting an argument.
 * Parsing as a number is OK
 *
 * Need to have anonymous variables, this is a value that
 * could be part of an expression or returned from a function.
 *
 * Where do I keep variable name ? Its scope ?
 * Below isn't a variable, it is a value
 *
 * To avoid constand alloc()/free() unused VALs are kept in a
 * NULL terminated chain, linked by v_next.
 *
 * Update V_TypeNames[] if the types are changed.
 * V_ANY should not be found in a VAL - it is used to say that any
 * type is acceptable - eg in funcs[]; also missing argument will
 * not be complained about, eg: !return.
 */
typedef	struct	VAL	{
	unsigned int	v_flags;
#define	V_UNDEF		0x000000	// Value is undefined
#define	V_INT		0x000001	// Value is v_int
#define	V_STR		0x000002	// Value is v_text
#define	V_CHAR		0x000003	// Value is v_int but is UniChar
#define	V_BOOL		0x000004	// Value is v_int
#define	V_ARRAY		0x000005	// Value is v_array
#define	V_HASH		0x000006	// Value is v_hash
#define	V_ERROR		0x00000e	// No value - error in generation
#define	V_ANY		0x00000f	// Value can be any of the above - determined run time
#define	V_TYPES		0x00000f	// Type mask -- fits in a nybble
#define	V_READONLY	0x000010	// It is a constant
#define	V_NOFREE	0x000020	// Do not free - it is not allocated
#define	V_BEEN_HERE	0x000040	// This node already been inspected
#define	V_REFCNTMASK	0xffff00	// Reference count bits
#define	V_REFCNTNMASK	0x0000ff	// Reference count NOT bits
#define	V_REFSHIFT	8		// Bits to shift to get reference count out of v_flags
	union	{	// VALUE
		TEXT		v_text;
		long		v_int;
		Array		v_array;
		Hash		v_hash;
		struct VAL*	v_next;		// Next unused VAL
	} v_v;
} VAL;

#define vtext(vp)		ttext(&(vp)->v_v.v_text)	// Address of text (BYTE*) in VAL*
#define vTEXT(vp)		(&(vp)->v_v.v_text)		// Address of TEXT in VALue
#define	vint(vp)		((vp)->v_v.v_int)		// Get INT value
#define	vSetInt(vp, i)		((vp)->v_v.v_int = (i))		// Set INT value
#define	vtype(vp)		((vp)->v_flags & V_TYPES)	// Type of the VALue
#define	vtypename(vp)		(V_TypeNames[vtype(vp)])	// Type of the VALue as text
#define	vSetType(vp, t)		((vp)->v_flags = (unsigned)((t) | (int)((unsigned)(vp)->v_flags & (unsigned)~V_TYPES)))	// Set type of the VALue

static	inline	void	vRefSet(VAL* vp, int refcnt)		// Set the reference count for VAL vp
{
	vp->v_flags = (unsigned)((int)(vp->v_flags & V_REFCNTNMASK) | (int)(refcnt << V_REFSHIFT));
}

#define	vRefGet(vp)		((int)((vp)->v_flags >> V_REFSHIFT))	// Reference count for VAL vp

// AKA valGet(V_UNDEF) but uses the static one to avoid malloc() overhead
extern	VAL	undefVAL;
static	inline	VAL*	valGetUndef(void)
{
	vRefSet(&undefVAL, ((int)vRefGet(&undefVAL) + 1));
	return(&undefVAL);
}


/* Formal parameters to procedures are stored as a linked list
 * of argument names using the following simple structure.
 * pa_name includes the '%'.
 * The '%' is replaced with '[' if the argument is an array - this
 * is ONLY for this structure this is invisible to the user.
 */
typedef struct PARG {
	struct PARG*	pa_next;		// ptr to next linked argument
	char		pa_name[NVSIZE + 1];	// name of the argument
} PARG;


/* The !while directive in the execution language needs to
 * stack references to pending whiles. These are stored linked
 * to each currently open procedure via a linked list of
 * the following structure
 *
 * For w_type BTCONTINUE w_end points to the !while statement
 */
typedef struct WHBLOCK {
	struct WHBLOCK* w_next;	// next while
	LINE*	w_begin;	// ptr to !while/!break/!continue statement
	LINE*	w_end;		// ptr to the !endwhile statement
	int	w_type;		// block type
	int	w_index;	// Index into w_array
	VAL*	w_array;
} WHBLOCK;

// Values for w_type (same as Dxxx values makes it easier):
#define	BTWHILE			DWHILE
#define	BTBREAK			DBREAK
#define	BTCONTINUE		DCONTINUE

/* List of goto *lablels built up by dobuf()
 * This makes !gotos much faster than searching the buffer
 */
typedef	struct	golist	{
	struct	golist*	gl_next;	// Chain
	BYTE*		gl_name;	// Name - does not include the '*'
	LINE*		gl_line;	// The LINE* that it was found on
} golist;


/* Procedure Information - linked from the BUFFER.
 * The idea is that this is obtained by scanning once, not every dobuf()
 *
 * Reference (pointer) from BUFFER - most buffers will not need it.
 */
typedef	struct	{
	PARG*	pri_args;		// ptr to the first argument
	WHBLOCK* pri_whlist;		// ptr to !while list
	golist*	pri_gotos;		// List of labels found in the buffer
	int	pri_numargs;		// Number of arguments to procedure
	int	pri_numlocals;		// Number of local variables
	int	pri_scanned;		// Buffer has been scanned for !while/!goto etc
} Prinfo;

// This is about what is stores in line notes used in dobuf()
#define	MAC_NOTE_DIRMASK	0x000f		// Stored values for things like !if
#define	MAC_NOTE_HELP		0x0010		// The line is in a help block
#define	MAC_NOTE_WHILE_ARRAY	0x0020		// The line !while var array



/* We need to know things about lines that are displayed in a window.
 * Something to keep them in. One per line, referenced from w_lineFlags.
 */
typedef	BYTE	Lflags;

#define	LF_StandOut	0x01		// A cell/character is displayed standout or underline

/* Numeric marks.
 * Every window has 12 of them. They are placed automatically when 'great movement' takes
 * place. Numbers 1 -> 9 are rotated. Mark 0 is placed often and may become mark 1.
 * Implemented as an array in the Window & BUFFER structs.
 *
 * 10 are visible to the user, numbered 0 .. 9
 *  2 are available internally (indexes 10, 11) to be used when a LINE* is
 *    needed but update() (or tgetc()) is called - which might result in the
 *    buffer being re-read and all LINEs trashed. These will be made safe by
 *    RestoreDotAndMark()
 * Remember that any LINE referenced by a numeric mark must have L_NMARK set.
 */
typedef	struct {
	LINE*	n_pline;	// Pointer to line marked line - might be NULL if deallocated
	LineLn	n_oline;	// Offset within line - undefined if n_pline is NULL
} NMARK;

/* There is a window structure allocated for every active display window. The
 * windows are kept in a big list, in top to bottom, left to right screen order, with
 * the listhead at "wheadp". Each window contains its own values of dot and mark.
 * The flag field contains some bits that are set by commands to guide
 * redisplay. Although this is a bit of a compromise in terms of decoupling,
 * the full blown redisplay is just too expensive to run for every input
 * character.
 *
 * The full list starts with wmessp, its w_wndp is wheadp. The point is that wmessp
 * (message window) must not participate in many of the things that other widows
 * do. So simplest to have a separate head - which is the head that most of the
 * code was originally written for.
 *
 * NB: w_doto can be set to 'index' the character beyond the end of what is
 * in w_dotp. This is legal in that it is referring to the virtual newline
 * at the end. This means that, especially in multibyte mode, the current
 * character must not be looked either before going back one or checking
 * the length.
 *
 * w_ntrows is the # rows of text, the buffer's mode line will be below that.
 *
 * w_ncolumns must be > BYTE as screens can be wider than 255
 * w_leftcol	  is the leftmost column of the window
 * w_lefttextcol  is the leftmost column of the window that contains buffer text
 * w_numwidth	  is the number of digits in displayed line number + 1 for the space
 * The difference between w_lefttextcol & w_leftcol might be a displayed line number.
 * w_numwidth == w_lefttextcol - w_leftcol; but is present as it is used a lot.
 *
 * The # columns taken on the screen is: w_ncolumns + ornamentation.
 * The # text columns taken on the screen is: w_ncolumns - w_numwidth
 *
 * Ornamentation is whatever is displayed on the left to visually separate
 * this buffer from any buffer to the left/right on the screen. It is present on the
 * right hand side of every window except the one(s) on the very right of the screen.
 *
 * w_leftcol & w_lefttextcol are measured from LH of the screen (counting from 0).
 *
 * The screen is not continually updated, w_toplinenum is what is logically the
 * line number of top line in the buffer that is displayed, count from 1.
 *
 * w_lastnumLines is used to determine if the bottom line displayed has scrolled,
 * it may differ from b_numLines until the screen has been redrawn.
 *
 * To keep the cursor within what is displayed the whole window will scroll left,
 * how far is recorded in w_leftoffset. If the window has not scrolled this is 0.
 * (This is different from earlier versions that scrolled just one line.)
 *
 * Columns in this context is for a terminal, ie fixed width. Some characters are
 * double width (eg U+9605) and these will occupy 2 columns.
 *
 * w_curcol is for private use in display.c It is the current column being
 * written. 0 based column in the entire line. To get the virtual screen
 * column: subtract w_leftoffset.
 * Part of the reason is to break the link with the terminal LH edge (think:
 * left/right windows).
 *
 * w_disrow and w_discol are relative to the screen origin. They are where the
 * cursor was last displayed. With HEX option w_discol is of the first nybble, which
 * is why w_lastcurcol exists - it is where the last cursor was, including +1 for 2nd nybble.
 *
 * w_cursattr is the attributes (eg colour) cursor position before it is
 * highlightened to PEN_CURSOR. Ie this is for windows other than the current.
 *
 * w_lastdismarkp and w_lastdismarko are just used for comparing, the values
 * will not otherwise be used.
 *
 * w_goalCol & w_goalLeftoffset are what we would like to have; however a vertical
 * movement has taken the cursor to a line that is too short or the user has explicitly
 * set w_leftoffset (move-window-* command), so the cursor is shown somewhere to the
 * left of the goal.
 * Try to put it on the goal when we find a line long enough. Get a similar
 * thing when the current line has a tab, in which the cursor is a bit to the
 * right of the goal.
 *
 * w_lineFlags is a vector for all window lines, ie includes the mode line
 * (even if the window does not have/display one).
 *
 * w_force: reframe() sets ".": > 0 lines from top; < 0 lines from bottom; 0 - center.
 *
 * w_nmarks holds the numeric marks - see NMARK as to why there are 12.
 *
 * Invent a static NULL buffer that we can set w_bufp to - deleted windows that can't be given away; readonly (MDVIEW, BFREAD ?) invisible (BFINVS)
 *
 * Below, these are, in some places, referred to as 'units':
 * S	means that it is in screen units, ie relative to screen origin.
 * L	means that it is in terms of the line - which includes any space for line number
 * W	means relative to the origin of the window
 * B	means byte offset within a LINE
 *
 * The partoffsets have a value of 0 or 1.
 * S is short - for lines more than needed at time of writing, but will get > 255 at some point,
 * columns already there.
 */
typedef struct  Window {
	struct	Window* w_wndp;		// Next window
	struct	BUFFER* w_bufp;		// Buffer displayed in window

	LINE*	w_topLinep;		// Top line displayed in the window, often not first in buffer

	LINE*	w_dotp;			// Line containing "."
	LineLn	w_doto;			// Byte offset for "." B

	LINE*	w_markp;		// Line struct for current mark
	LineLn	w_marko;		// Byte offset for current mark B

	LINE*	w_lastdismarkp;		// Line struct and offset of
	LineLn	w_lastdismarko;		// mark when last displayed B

	LINE*	w_modeline;		// Modeline text
	Lflags*	w_lineFlags;		// Vector of per displayed line flags

	BufLns	w_toplinenum;		// Number of top line of window - starts 1
	BufLns	w_lasttop;		// Last displayed top line #
	BufLns	w_lastnumLines;		// When Last displayed: # lines in buffer
	LineLn	w_leftoffset;		// Left scroll for window L
	LineLn	w_curcol;		// Current output column L
	LineLn	w_goalCol;		// Ideally the cursor here L (CFCPCN)
	LineLn	w_goalLeftoffset;	// Ideal w_leftoffset L

#define	NUM_NUM_MARKS	12		// How many numeric marks. 10 + 2 for keeps
	NMARK	w_nmarks[NUM_NUM_MARKS];// Numeric marks, qv w_nmarkOff

	int	w_leftcol;		// Left most column of window S
	int	w_lefttextcol;		// Left most text column S
	int	w_numwidth;		// Width of any displayed line numbers
	int	w_ncolumns;		// window width - w/out ornamentation
	int	w_disrow;		// Cursor position S
	int	w_discol;		// Cursor position S
	int	w_lastcurcol;		// Column the cursor was last in S
	int	w_lastcurlen;		// Length of last painted cursor
	int	w_toprow;		// Origin 0 top row of window S
	int	w_ntrows;		// # of rows of text in window

	short	w_opts;			// Window specific option flags

	VidAttr	w_cursattr;		// Attributes of cursor position

	char	w_partoff;		// Part offset within a character at w_doto - think hex
	char	w_markpartoff;		// Part offset for w_marko
	char	w_lastdismarkpo;	// Part offset for w_lastdismarko

	char	w_force;		// If != 0, force row of dot to line w_force
	char	w_flag;			// Flags on what needs displaying

	char	w_attrib;		// Attributes of a window

	char	w_nmarkOff;		// Offset of mark 0 in w_nmarks

#if	LEGACY_DISPLAY
	char	w_fcolor;		// current forground colour
	char	w_bcolor;		// current background colour
#endif
} Window;

/* w_flag can contain things to do with how the window needs to be redisplayed.
 * Beware, some are not tested explicitly but their setting makes
 * a difference. Eg WFHARD.
 * This is because they are not all used as just bit fields, the value
 * of the whole byte (maybe with something masked out) can be important.
 * Nasty.
 */
#define	WFFORCE			0x01	// Window needs forced reframe
#define	WFMOVE			0x02	// Movement from line to line
#define	WFEDIT			0x04	// Editing within a line
#define	WFHARD			0x08	// Better to a full display
#define	WFMODE			0x10	// Update mode line & window edges
#define	WFNWID			0x20	// Need to check line # width requirement

/* w_attrib can contain the following.
 * Generally it records non user settable things about the window.
 *
 * If WATT_MODETEXT, WATT_MODELINE must be set.
 *
 * With WATT_BORDER: the modeline counts as one of the sides
 */
#define	WATT_MODELINE		0x01	// The window has a mode line
#define	WATT_MODESTATIC		0x02	// Modeline text is static, not generated by modeline()
#define	WATT_BORDER		0x04	// Every window side has a border
#define	WATT_NOSPLIT		0x08	// The window cannot be split

/* Bits in w_opts.
 * Do not have it bit 'holes', this has to ascend as does winOptsExplan[].
 */
#define	WO_HEX			0x0001	// Display bytes in hex
#define	WO_COLOUR_TABS		0x0002	// Show where tabs are by colouring them
#define	WO_NUMBER		0x0004	// Display line numbers
#define	WO_META_ATTRIB		0x0008	// Display meta & control chars with underline & standout modes
#define	WO_TRACK		0x0010	// Track option - same movement in tracked windows
#define	WO_SHOW_LINE_END	0x0020	// Show where end of line is with a $
#define	WO_POSITION_SHOW	0x0040	// Show col#/line# byte-position% in buffer

/* Alphabetic marks
 */
typedef	struct	AMARK {
	struct	AMARK*	m_next;	// pointer to next mark in list
	LINE*	m_pline;	// pointer to line marked line - might be NULL if deallocated
	LineLn	m_oline;	// offset within line
	BYTE	m_name;		// alphabetic name of the mark
} AMARK;


/* Text is kept in buffers. A buffer header, described below, exists for every
 * buffer in the system. The buffers are kept in a big list, so that commands
 * that search for a buffer by name can find the buffer header. There is a
 * safe store for the dot and mark in the header, but this is only valid if
 * the buffer is not being displayed (that is, if "b_nwnd" is 0).
 *
 * The text for the buffer is kept in a circularly linked list of lines, with a
 * pointer to the header line in "b_linep".
 * The first line in the buffer is lforw(b_linep). The last lback(b_linep).
 * "." is allowed to rest on an empty b_linep.
 * b_linep is not immutable: appending a line to the buffer will change it, see lnewline().
 *
 * The term 'the narrow' means the lines between lforw(b_btopline) & lback(b_abotline).
 * This will be all the lines in the buffer unless it has been narrowed.
 * Most commands operate in the narrow, but a few always use the entire buffer.
 *
 * Buffers may be "Inactive" which means the files associated with them
 * have not been read in yet. These get read in at "use buffer" time.
 *
 * b_fname is what is used to open/close - ie give to the operating system.
 * b_fnameClean is what is displayed by the editor (eg by list-buffers),
 * this has been made UTF8 clean (which b_fname prob was anyway).
 * b_fnameClean may hold other things, eg:
 * * a description of a procedure that the buffer might hold
 * * command buffer - command that was run
 * b_bname might be truncated, it is UTF8 clean.
 * All 3 are NUL terminated and are never NULL - may contain an empty malloced string..
 * b_fname is used as a proxy for the buffer being associated with a file,
 * see setVisibleFname().
 *
 * Some members are only valid when the buffer is not being displayed in a window:
 * b_topLinep b_dotp b_doto b_markp b_marko b_winopts b_toplinenum b_nmarks b_nmarkOff
 * These are stored in the buffer to restore to a window.
 *
 * b_nmarks holds the numeric marks - see NMARK as to why there are 12.
 *
 * b_lsType is the file type as reported by 'ls', one of '?/.!-sbcbpD'.
 * Permission error: '/' - directory (or missing parent directory),
 * '!' - cannot access file, not due to '/' (dir issue).
 * '.' file does not exist
 * '?' initial unknown value.
 */
typedef struct  BUFFER {
	struct	BUFFER* b_bufp;		// Link to next struct BUFFER
	BYTE*	b_fname;		// File name
	BYTE*	b_fnameClean;		// File name - UTF8 clean copy
	BYTE*	b_bname;		// Buffer name

	LINE*	b_linep;		// Link to the header LINE
	BufLns	b_numLines;		// Number of lines in the buffer

	LINE*	b_btopline;		// Line before first line for most commands
	LINE*	b_abotline;		// Line after  last  line for most commands

	LINE*	b_topLinep;		// Top line displayed in the window
	BufLns	b_toplinenum;		// Number of top line of window - starts 1

	LINE*	b_dotp;			// Link to "." LINE structure
	LineLn	b_doto;			// Offset of "." in above LINE

	LINE*	b_markp;		// LINE struct of current mark
	LineLn	b_marko;		// Offset of "." in mark
	AMARK*	b_amark;		// Pointer to mark list

	NMARK	b_nmarks[NUM_NUM_MARKS];// Numeric marks, qv b_nmarkOff

	int	b_mode;			// editor mode of this buffer
	int	b_flag;			// Flags
	short	b_winopts;		// w_opts of last window to display it
	MapCell	b_nwnd;			// Count of windows on buffer
	char	b_active;		// window activated flag

	short	b_tabsize;		// How wide a tab this buffer
	char	b_nmarkOff;		// Offset of mark 0 in b_nmarks

	Prinfo*	b_procinfo;		// Information about a procedure
	BYTE	b_exec; 		// how many active executions

#if	UNIX
	char	b_lsType;		// File type as ls first column
	Time_T	b_mtime;		// modification time of file
	off_t	b_fileEnd;		// Position of EOF when last read
	dev_t	b_stdev;		// st_dev for the file
	ino_t	b_stino;		// inode number of the file
#endif

#if	BACKUP && !BUPEVS
//	int	b_bupflg;		// Backup file at next save?
#endif

#if	CRYPT
	TEXT*	b_key;			// Current encrypted key
#endif
} BUFFER;

/* b_flag values */
#define	BFINVS			0x0001	// Internal invisible buffer
#define	BFCHG			0x0002	// Changed since last write
#define	BFREAD			0x0004	// Buffer to stay in read mode
#define	BFWEIRD			0x0008	// Buffer is sort of writable - see f_funnyfile

#define	BFOLD			0x0010	// Buffer is out of date
#define	BFTRUNC			0x0020	// buffer was truncated when read
#define	BFIMPLICIT		0x0040	// Implicit \n at end of last line
#define	BSYNC_CURSOR		0x0080	// Sync cursor of all windows on buffer
#define	BLINE_IN_MARKO		0x0100	// b_marko contains a desired line # on readin
#define	BCOL_IN_DOTO		0x0200	// b_doto contains a desired column # on readin
#define	BHELP			0x0200	// Help buffer
					// BEWARE: last 2 are the same bit - should not be an issue
#define	BRENEWAPPEND		0x0400	// File is being appended to, only RENEW/read new lines
#define	BRENEWED		0x0800	// Buffer has been read in as in RENEW mode
#define	BRENEWEDAGAIN		0x1000	// Flipped on every renew
#define	BRENEWEDAGAIN1		0x2000	// Flipped on every 2nd renew (aka 2 bit counter)
#define	BIMMUTABLE		0x4000	// Buffer cannot have VIEW removed
#define	BFROMSAVE		0x8000	// Buffer read from .emacssave

/* mode flags (b_mode)
 * Do not have it bit 'holes', this has to ascend as does modeNameExplan[].
 */
#define	MDWRAP			0x0001	// word wrap
#define	MDCMOD			0x0002	// C indentation and fence match
#define	MDEXACT			0x0004	// Exact matching for searches
#define	MDVIEW			0x0008	// read-only buffer
#define	MDOVER			0x0010	// overwrite mode
#define	MDMAGIC			0x0020	// regular expresions in search
#define	MDCRYPT			0x0040	// encrytion mode active
#define	MDASAVE			0x0080	// auto-save mode
#define	MDINDEN			0x0100	// indent mode
#define	MDRENEW			0x0200	// renew buffer mode
#define	MDNONL			0x0400	// No \n at EOF
#define	MDPHP			0x0800	// PHP indentation and fence
#define	MDUTF8			0x1000	// Multibyte characters
#define	MDDOS			0x2000	// DOS line endings, \r assumed
#define	MDTABEXP		0x4000	// Expand input tabs to spaces


/* The starting position of a region, and the size of the region in
 * characters, is kept in a region structure.  Used by the region commands.
 *
 * Origin means nearest the start of the buffer.
 * Ending means the one furthest from the start of the buffer.
 * r_lines will be 1 if mark & "." are on the same line, 2 if on adjacent lines.
 */
typedef struct  {
	LINE*	r_linep;		// Origin LINE address.
	LineLn	r_offset;		// Origin LINE offset.
	LINE*	r_endLinep;		// Ending LINE address.
	LineLn	r_endOffset;		// Ending LINE offset.
	BufChrs	r_size;			// Length in characters - not bytes
	BufLns	r_lines;		// # Lines in the region
} REGION;

/* Structure to store the "." and mark positions in a buffer.
 * This is by position, not using structures - so that it can be
 * restored later once the buffer is refreshed - or something.
 *
 * The first line # is 1.
 * A line # of 0 means that the line was not seen.
 * A line # of -1 means the end of buffer - b_linep, in this case the line offset is ignored.
 * A line offset of -1 means the position of the end of line.
 * The offset is not defined unless the line # > 0.
 *
 * The alpha marks are kept in the order that they are found in
 * the buffer - makes replacing them easier/faster.
 * Same for numeric marks.
 *
 * For alpha marks abp_name is 'a' .. 'z'
 * For num   marks abp_name is '0' .. '9' and ';' & ':' for the internal use ones.
 *
 * The purpose of abp_flags is to allow things like L_NUL to be restored.
 *
 * First a structure for alpha marks - also used for numeric marks.
 */
typedef	struct	{			// Alpha mark Buffer Position
	BufLns	abp_line;		// Line # of the mark
	LineLn	abp_off;		// Offset of the mark in the line
	BYTE	abp_name;		// Alpha mark's name
} AmBp;
typedef	struct	{
	BufLns	bp_dotl;		// Line # of "."
	LineLn	bp_doto;		// Offset of "." in the line
	BufLns	bp_markl;		// Line # of mark
	LineLn	bp_marko;		// Offset of mark in the line
	BufLns	bp_topl;		// Line # at top of screen
	AmBp	bp_am[27];		// 27 = 26 + 1 NULL to end
	AmBp	bp_nm[NUM_NUM_MARKS + 1];// Numeric marks (+ 1 NULL)
	BYTE	bp_NUL;			// At least one line had the L_NUL flag set
} BufPos;


/* Structure to hold user variables and their definitions.
 * This is used for both the global and the local symbol tables.
 */
typedef struct UVAR {
	char	u_name[NVSIZE + 1];	// name of user variable
	VAL*	u_value;		// value
} UVAR;

// Scope values to findvar():
#define	VT_NONE			0	// Don't declare it if not found
#define	VT_LOCAL		1	// Local to the current procedure
#define	VT_GLOBAL		2	// Global to all procedures
#define	VT_PROBE		8	// ORed with the above - could create a variable

/* A UTABLE is a user variable table.... containing some header
 * information and an array of user variable names and definitions.
 * They are held together in a linked list, the last member of
 * the list being the global user variable table.
 */
typedef struct UTABLE {
	struct UTABLE*	ut_next;	// Ptr to next user variable table
	BUFFER*		ut_bufp;	// Ptr to buffer holding procedure
					// associated with this symbol table
	int		ut_size;	// Max number of variables in table
	UVAR		ut_uv[1];	// List of variable names/definitions
					// in this variable table
} UTABLE;

/* When Emacs' command interpetor needs to get a variable's name,
 * rather than it's value, it is passed back as a VDESC variable
 * description structure. The v_num field is a index into the
 * appropriate variable table.
 * v_type hold values : TKVAR, TKENV, TKBAD (= not legal).
 * v_num  If v_type is:
 *	  TKBAD not defined
 *	  TKENV	index in envars[]
 *	  TKVAR index into ut_uv the table referenced by v_ut
 * v_ut	  Which table the variable was found in. This will either
 *	  be the global variable table or the variables for the current
 *	  procedure.
 *	  If the variable was not found & not VT_PROBE this will be NULL
 */
typedef struct VDESC {
	int	v_type;		// Type of variable
	int	v_num;		// Ordinal pointer to variable in list
	UTABLE*	v_ut;		// Ptr to appropriate user table if user var
} VDESC;




/* The editor communicates with the display using a high level interface. A
 * "TERM" structure holds useful variables, and indirect pointers to routines
 * that do useful operations. The low level get and put routines are here too.
 * This lets a terminal, in addition to having non standard commands, have
 * funny get and put character code too. The calls might get changed to
 * "termp->t_field" style in the future, to make it possible to run more than
 * one terminal type.
 *
 * Note that t_mrow and t_nrow were one less than the actual number of lines
 * on the screen - the last line being the message line.
 * This was stupid and caused confusion, fixed: this comment left for a bit
 * in case the fix was not perfect.
 *
 * Difference between t_mrow & t_nrow and t_mcol & t_ncol.
 * t_mxxx is the max physically supported, t_nxxx is what Emacs uses.
 * t_nxxx <= t_mxxx. Only where is can be changed, not under Unix. See WINSIZE.
 * Use what is in AllocLines & AllocColumns, anything else and there is
 * potential for memory corruption - there might be a potential time
 * lag catching up with t_nxxx - think screen size change when redrawing.
 * While you are at it: look at ml_bsize.
 *
 * t_eeop now does home, terminfo does this, others might not.
 */
typedef struct  {
	short	t_mrow;			// max number of rows allowable
	short	t_nrow;			// current number of rows used
	short	t_mcol;			// max Number of columns possible
	short	t_ncol;			// current number of columns used
	short	t_margin;		// min margin for extended lines UNUSED
	short	t_scrsiz;		// size of scroll region " UNUSED
	int	t_pause;		// # times thru update to pause UNUSED
	int	(*t_open)(void);	// Open terminal at the start.
	void	(*t_close)(void);	// Close terminal at end.
	void	(*t_kopen)(int ksOnly);	// Open keyboard
	void	(*t_kclose)(void);	// Close keyboard
	KeyCode	(*t_getchar)(long timeout); // Get character from keyboard
	int	(*t_putchar)(int c);	// Put character to display.
	int	(*t_flush)(void);	// Flush output buffers.
	void	(*t_move)(int r, int c);// Move the cursor, origin 0.
	void	(*t_eeol)(void);	// Erase to end of line.
	void	(*t_eeop)(void);	// Erase the page & go home
	void	(*t_beep)(void);	// Beep.
	void	(*t_rev)(int s);	// set reverse video state
	void	(*t_ul)(int s);		// set underline video state
	int	(*t_rez)(const BYTE* res); // change screen resolution
	void	(*t_setfor)(int fc);	// set forground colour
	void	(*t_setback)(int bc);	// set background colour
	char*	(*t_keyname)(KeyCode c);// Return name of function key
	int	(*t_what)(int what);	// Does the terminal do 'what'
	void	(*t_scroll)(int topLine, int rlines, int moveBy); // Scroll up/down
	KeyCode	(*t_special)(const BYTE* seq);// Decode sequences (FN keys)
	KeyCode	(*t_ungetc)(KeyCode c);	// Push a character back to keyboard
	void	(*t_fixmargin)(void);	// Fix ttcol when hit right margin
	void	(*t_inschars)(int n);	// Insert n characters
	void	(*t_delchars)(int n);	// Delete n characters
	void	(*t_enterins)(void);	// Enter insert mode
	void	(*t_exitins)(void);	// Exit insert mode
} TERM;

/* TEMPORARY macros for terminal I/O  (to be placed in a machine
 * dependant place later).
 */
#define	TTopen		(*term.t_open)
#define	TTclose		(*term.t_close)
#define	TTkopen		(*term.t_kopen)
#define	TTkclose	(*term.t_kclose)
#define	TTgetc		(*term.t_getchar)
#define	TTungetc	(*term.t_ungetc)
#define	TTputc		(*term.t_putchar)
#define	TTflush		(*term.t_flush)
#define	TTmove		(*term.t_move)
#define	TTeeol		(*term.t_eeol)
#define	TTeeop		(*term.t_eeop)
#define	TTbeep		(*term.t_beep)
#define	TTrev		(*term.t_rev)
#define	TTul		(*term.t_ul)
#define	TTrez		(*term.t_rez)
#define	TTforg		(*term.t_setfor)
#define	TTbacg		(*term.t_setback)
#define	TTkeyname	(*term.t_keyname)
#define	TTdoes		(*term.t_what)
#define	TTscroll	(*term.t_scroll)
#define	TTspecKey	(*term.t_special)
#define	TTrightMargin	(*term.t_fixmargin)
#define	TTinschars	(*term.t_inschars)
#define	TTdelchars	(*term.t_delchars)
#define	TTenterins	(*term.t_enterins)
#define	TTexitins	(*term.t_exitins)

/* Can/will the terminal do things ?
 * Argument to TTdoes().
 * Returns true if it can.
 */
#define	TT_DOES_BLANK_UL	1	// Terminal will underline blanks
#define	TT_DOES_SCROLL		2	// Terminal will scroll an area
#define	TT_DOES_RIGHT_MARGIN	3	// A char written at right margin leaves it there
#define	TT_DOES_INSERT_CHARS	4	// Terminal does hardware insert & delete characters


/* Because MSDOS is stupid you sometimes get problems if the file is
 * opened/closed with the keypad set. Let's have a macro for this & keep
 * the rest of the code looking clean.
 */
#if	MSDOS
#define	TTkopenM	(*term.t_kopen)()
#define	TTkcloseM	(*term.t_kclose)()
#else
#define	TTkopenM
#define	TTkcloseM
#endif


/* Input (keyboard) key mapping.
 * This is done after multi byte interpretation, ie escapes to function keys
 * and UTF8 to UniChar.
 * Note that this linked list is looked at with keymap and is kept in
 * order of chm_from.
 * chm_from might be a Basic Latin if chm_to is outside of that (0 .. 127).
 */
typedef	struct	CharMap	{
	struct CharMap*	chm_next;	// Next in NULL terminated list
	UniChar		chm_from;	// What we map from
	UniChar		chm_to;		// What we map to - ie replace from with
} CharMap;

/* Flags to variously: mlGetKey() tgetc()
 */
#define	MGC_NONE		0x0000000	// Nothing in particular
#define	MGC_TIMOUT_MASK		0x0000FFF	// Timeout here (seconds) - else $prefdlay (ddelay) (1-4095 - 1h 8m)
#define	MGC_MAP			0x0001000	// Perform input key mapping as with: map-key
#define	MGC_LEAVEP		0x0002000	// Leave the prompt shown after mlGetKey() returns
#define	MGC_NO_CURSOR		0x0004000	// Do not touch CursorInMsg, cursor stays where it is
#define	MGC_RET_TIMEO		0x0008000	// If there is a timeout, return KEY_TIMEOUT
#define	MGC_APSTR_NOC		0x0010000	// Timeout append str is coloured & cursor does not follow
#define	MGC_NO_REFRESH		0x0020000	// Do not test input for screen refresh key
#define	MGC_NO_KMACRO		0x0040000	// Do not supply input from kbdmode == PLAY
#define	MGC_PROMPTING		0x0080000	// In a prompt - do not disturb
#define	MGC_DELAY_PROMPT	0x0100000	// Delay before writing the prompt/message
#define	MGC_MACARG		0x0200000	// If in macro return the first character from the next argument.



/* The bottom line of the screen is used for two things:
 * * Messages: which are displayed & overwritten
 * * Prompts: which must not be touched when in use - especially by messages
 *
 * However bottom line can be used, then temporarily, something else be shown
 * so we need to be able to preserve & restore state. See msgPush() & msgPop().
 * Resetting to 'reasonable' values on pop didn't always work, eg position of ".".
 * So save all that. See more comments near msgPush().
 *
 * The message window is only one line high, the most important thing is the
 * offset of ".". The line will be in a history buffer.
 *
 * The names of the struct members reflect that of a Window.
 */
typedef	struct	{
	LineLn	ms_doto;		// Offset of "." in the line
	LineLn	ms_leftoffset;		// Any left scrolling
	VidAttr	ms_cursattr;		// Cursor attribused
	int	ms_pallet;		// Pallet for message line
	char	ms_force;		// Reframing
} msgState;


/* Structure for the name binding table.
 * The reverse function (if defined) must take the same flags as n_func.
 *
 * If n_func is NULL then this describes a procedure named n_name. To run it
 * will be in a buffer called [n_name], but the NBIND entry can exist before
 * the buffer is created.
 *
 * Notes:
 * N_DEFAULT_1		The default count is 1, what this does is to set the flag.
 *			Thus: if this not set and a count is not given, arg n will be 1
 *			and arg f will be FALSE.
 *			This may be relied on, ie f does not always need to be tested.
 * N_SYNONYM		A command can only have one synonym.
 *			Limitation in buildlist(). The error is benign.
 *
 * NULL terminated array, test n_name for end of array.
 */
typedef struct {
	char*	n_name;			// Name of function
	int	(*n_func)(int, Cnt);	// function n_name is bound to
	int	(*n_rfunc)(int, Cnt);	// Reverse function - may be NULL
	short	n_flag;			// Info about the routine
#define	N_MODIFY	0x0001		// Changes a buffer - somehow
#define	N_AI_BEFORE	0x0002		// Clear AutoIndented before call
#define	N_AI_AFTER	0x0004		// Clear AutoIndented after call
#define	N_TRACK		0x0008		// Track action needed
#define	N_AI_KEEP_AI	0x0010		// Preserve value of JustAutoIndented
#define	N_DEFAULT_1	0x0020		// Default arg is 1 (if unset)
#define	N_FAIL_NEG_ARG	0x0040		// Fail if the argument is -ve
#define	N_FAIL_ZERO_ARG	0x0080		// Fail if the argument is zero
#define	N_NOT_HEX	0x0100		// Fail if the window has hex option
#define	N_RESTRICTED	0x0200		// Must not run in restricted mode
#define	N_SYNONYM	0x0400		// Command has a synonym
#define	N_NOT_EXEC	0x0800		// Not if current buffer is being executed
	char*	n_desc;			// Short description of function - may be NULL if procedure
} NBIND;

/* Structure for prefix binding table.
 */
typedef	struct	{
	KeyCode	pt_code;		// Key code
					// Function to handle it
	KeyCode	(*pt_fun)(KeyCode ch, int* displayit, int* reprocess);
} PRETAB;

/* Structure for the table of initial key bindings.
 * We initialise k_ufp with the static bindings, and look up the function
 * to get k_ubp at the start of the program.
 */
typedef struct  {
	KeyCode		k_code;		// Key code
	union {
		int	(*k_ufp)(int, Cnt);	// Routine to handle it
		NBIND*	k_ubp;
	} k_u;
} KEYTAB;

/* Reverse map table, used to permit what the user sees as normal use of
 * function keys (ie using their bindings). We do this by matching the function
 * pointer.
 * End of table when rm_code is 0.
 */
typedef	struct	{
	KeyCode	rm_code;		// INTL code to generate
	int	(*rm_fp)(int f, Cnt n);	// Function it maps
} RMTAB;

/* Holds information about the last commands executed.
 * This is so that on a crash an indication about what was happening
 * can be printed.
 *
 * lc_status is set to -1 just before the command is run, allows detection
 * of it not completing.
 */
typedef	struct	{
	NBIND*	lc_cmd;			// What command
	int	lc_status;		// Status code returned
	int	lc_flag;		// Flag to say if count is valid
	Cnt	lc_arg;			// Command argument/count
} LASTCMD;


/* The editor used to hold deleted text chunks in the KILL buffer ...
 * now it is stored as LINEs in buffer '[Kill Buffer]', this simplifies
 * the code and allows insertion of kill text at places other than
 * the end of the chunk. The buffer is also editable.
 * The use of LINEs is more efficient, as kill blocks had a fixed 256 byte
 * buffer - that you used several of if needs be.
 *
 * The max # lines in the [Kill Buffer] is SaveMlHist.
 *
 * Note that a line is a contiguous kill, which could be several ordinary
 * buffer lines long.
 */

/* The following are values for the 2nd argument to ldelete(),
 * what to do with the characters that are deleted (killed):
 */
#define	KL_DISCARD		0	// Throw them away
#define	KL_KILL_APPEND		1	// Put at the end   of the kill buffer text
#define	KL_KILL_START		2	// Put at the start of the kill buffer text

// Flags to addHistoryLine() & addHistoryText() - can be ORed
#define	AHL_NONE		0x00	// Nothing
#define	AHL_NO_DUP		0x01	// Do not add if same as last line
#define	AHL_ADD_ZLEN		0x02	// Add even if zero length


/* These are all the sorts of things that we can look for
 * with findFile().
 * Other definitions are in findFile.c
 */
typedef enum {
	fT_Save,		// Save files
	fT_Help,		// Help .help files
	fT_HelpIndex,		// Help .index files
	fT_HelpTmpl,		// Help .template files
	fT_StartGlob,		// Global (ie for everyone) start file
	fT_StartUser,		// A user's own start file
	fT_Macro,		// Macro files
	fT_MacroKeys,		// Index of keys in macro files
} findTypes;



/* This is to make mnexpand() usable both with buffer modes and window options.
 */
typedef	struct	{
	char*		mn_name;	// Mode name
	char*		mn_explan;	// Mode explanation
} ModeName;
typedef	struct	{
	ModeName*	md_modes;	// Array of mode names & explanations
	char*		md_err;		// Error string/message
	int		md_mode_cnt;	// Number of modes
} ModeDef;


/* Used to initialise history buffers.
 * A NULL terminated array of them.
 */
typedef	struct	{
	BUFFER**	hl_bufp;	// Address of buffer pointer
	char*		hl_name;	// Buffer name
} HistList;


/* Statistics are displayed as value and name.
 * Have a NULL terminated array of these to iterate over.
 *
 * Statistics are stored in type ST_val.
 * long is 64 bit on LP64 (most *nix boxes) - might need long long on
 * LLP64 machine (MS Windows).
 */
typedef	long	ST_val;
typedef	struct	{
	ST_val*	st_value;		// Value to display
	char*	st_descr;		// Description
} ST_tab;



/* Stuff to help with substitution in a help file,
 * eg **var:$language** or **binding:backward-character**
 *
 * The 2 cases are very close, the struct summarises the differences
 */
struct	sv_parm	{
	int	(*sv_getval)(const char* name, TEXT* val); // Function to get value
	char*	sv_noval;				// Default 'no value' string
	int	sv_maxlen;				// Max length of name
	int	sv_onlyOne;				// Only show one binding
};



// /* Now for the CMLIST structure. It is used to remember previously typed
//  * commands (ie the ones passed to spawn() etc etc). It is a linked list
//  * with a field for a pointer to the text of the command which was typed.
//  *
//  * This is now no longer used, command line history now in bmlCmdHistp.
//  */
// 
// #define	NPREVCMDS		20	// number of previous commands remembered
// 
// typedef	struct	CMLIST {
// 	struct	CMLIST*	cm_next;// link to next CMLIST, NULL if last
// 	BYTE*		cm_text;// the actual text of the command
// } CMLIST;
// 
// #define	CMDBUFLEN	256	// Length of our command buffer

/* Incremental search defines.
 * IS_...  are tokens directly recognised.
 * ISI_... are function keys (SPEC bit set), we reverse map the binding, check
 * 	what function it is bound to, derive a meaning which gives an ISI_ code
 *	(INTL bit set).
 */
#if	ISRCH

#define	IS_VMSQUOTE		0x16		// VMS quote character C-v
#define	IS_VMSFORW		0x18		// VMS search forward  C-x
#define	IS_QUIT			CharEscape	// Exit the search
#define	ISI_REVERSE		(INTL | 0x0001)	// Search backward
#define	ISI_FORWARD		(INTL | 0x0002)	// Search forward
#define	ISI_RUBOUT		(INTL | 0x0003)	// Delete previous character
#define	ISI_ABORT		(INTL | 0x0004)	// Abort
#define	ISI_QUOTE		(INTL | 0x0005)	// Quote next character
#define	ISI_UPLINE		(INTL | 0x0006)	// Up/Previous line
#define	ISI_DOWNLINE		(INTL | 0x0007)	// Down/next line
#define	ISI_MOVE_DOWN		(INTL | 0x0008)	// move-window-down
#define	ISI_MOVE_UP		(INTL | 0x0009)	// move-window-up
#define	ISI_MOVE_LEFT		(INTL | 0x000A)	// move-window-left
#define	ISI_MOVE_RIGHT		(INTL | 0x000B)	// move-window-right
#define	ISI_REFRESH_SCREEN	(INTL | 0x000C)	// Repaint after possible corruption

/* IS_QUIT is no longer used, the variable metac is used instead */

#endif

/* These encode search options and are put in [Srch History] - l_notes
 * The idea is that a saved search is not just a string but also its options.
 *
 * Some of these options can be set elsewhere than EditSearchOpts()/CheckCaselessSearch()
 * so there is a can-it-be-done flag
 */
#define	SO_EQI_flag	0x01	// There is a value for case insensitve search
#define	SO_EQI_val	0x02	// Do search insensitively
#define	SO_SMW_flag	0x04	// There is a value for space matches white space
#define	SO_SMW_val	0x08	// Space matches white space
#define	SO_WORD_val	0x10	// Match away from word

/* Case of a word
 * This is used in search & replace
 */
typedef	enum {
	Lc,	// All lower case
	Uc,	// All upper case
	Tc,	// Title case: first upper, rest lower
	Oc,	// Other case: first lower a subsequent upper; first upper & subsequent lower & upper
	Np	// Not present or unknown
} Case;

/* This describes text used in search & replace.
 * Used for: search string, replacement string & string matched (by search string).
 *
 * Beware: 2 different definitions of 'word'
 * * wd_case  - a word starts with an alphabetic character
 * * newlines - a word is something white space separated (' ', TAB & NL)
 *
 * Transformation means that there is nothing in the string that
 * would stop the replacement string being munged to match the
 * string found.
 */
typedef	struct	{
	TEXT*	wd_line;	// Text from the line scanned
	BYTE*	wd_nlAddr;	// Address of first \n
	int	wd_newlines;	// # newlines seen
	int	wd_nlWord;	// word number that the \n is before
	int	wd_noUpper;	// The string does not contain an upper case character
	Case	wd_case[2];	// [0] first word, [1] subsequent words
} wDesc;

/* What eq() returns.
 * The point is that there are 2 match values, one of which says that
 * the next subject character be tried against the same pattern. Thus a
 * space might match several spaces/tabs.
 *
 * You may assume that EQ_DIFFER is a C boolean false, the other 2 are true.
 */

#define	EQ_DIFFER		0	// They differ
#define	EQ_MATCH		1	// The two characters match
#define	EQ_MORE			2	// They match, try again with the next subject character


#if	MAGIC

/* Defines for the metacharacters in the regular expressions in search
 * routines.
 *
 * The following are values in MC.mc_type & RMC.mc_type:
 */

#define	MCNIL			0	// Like the '\0' for strings.
#define	LITCHAR			1	// A non special (literal) character
#define	ANY			2	// '.'
#define	CCL			3	// Character Class
#define	NCCL			4	// Negated Character Class
#define	BOL			5	// Beginning Of Line
#define	EOL			6	// End Of Line
#define	DITTO			7	// & in replacement
#define	CLOSURE			256	// An or-able value.
#define	MASKCL			(CLOSURE - 1)

#define	MC_ANY			'.'	// 'Any' character (except newline).
#define	MC_CCL			'['	// Character class.
#define	MC_NCCL			'^'	// Negate character class.
#define	MC_RCCL			'-'	// Range in character class.
#define	MC_ECCL			']'	// End of character class.
#define	MC_BOL			'^'	// Beginning of line.
#define	MC_EOL			'$'	// End of line.
#define	MC_CLOSURE		'*'	// Closure - does not extend past newline.
#define	MC_DITTO		'&'	// Use matched string in replacement.
#define	MC_ESC			'\\'	// Escape - suppress meta-meaning.

#define	BIT(n)		(1 << (n))	// An integer with one bit set.

/* HICHAR - 1 is the largest character we will deal with.
 * HIBYTE represents the number of bytes in the bitmap.
 * This is used in character classes.
 *
 * This does not cope with most UTF-8
 */

#define	HICHAR			256
#define	HIBYTE			(HICHAR >> 3)

/* Typedefs that define the bitmap type for searching (BITMAP),
 * the meta-character structure for MAGIC mode searching (MC),
 * and the meta-character structure for MAGIC mode replacment (RMC).
 *
 * In non-UTF-8 mode store a literal character in lchar[0]
 *
 * lchar can store 6 bytes, UTF-8 only needs 4, but if CONFIG_MAX_UNICODE
 * is set high (not normal) it could be up to 6.
 * This is not an issue as BITMAP is 32 bytes.
 * Can't use UTF8_MAX_BYTES as utf8.h is included after here.
 */
typedef char*	BITMAP;

typedef	struct {
	short int	mc_type;	// One of: MCNILL ... DITTO (above)
	union {
		BYTE	lchar[6];	// When mc_type == LITCHAR: a literal char in a magic search
		BITMAP	cclmap;		// When mc_type == CCL or NCCL
	} u;
} MC;

typedef	struct {
	short	int	mcr_type;	// One of: MCNILL ... DITTO (above)
	BYTE*		rstr;		// Replacement string
} RMC;

#endif	// MAGIC

/* Direction indicators, used by ModeMatch() mnexpand() & similar
 */
#define	FORW			1	// List expand/choose next value
#define	BACKW			(-1)	// List expand/choose previous value

// Flag values returned by printToText():
#define	PTL_PALLET_MASK		0x007	// Mask for MsgWindowPallet - PAL_something
#define	PTL_MSG_PALLET		0x008	// Set MsgWindowPallet
#define	PTL_NO_HISTORY		0x010	// Seen %h - do not write to history
#define	PTL_NO_MSGLINE		0x020	// Seen %# - do not write to message line
#define	PTL_ERROR		0x040	// Seen %E - the message was an error - write to history
#define	PTL_FLUSH		0x080	// Seen %F - Flush the message window
#define	PTL_IMPORTANT		0x100	// Seen %W/%E - important message
#define	PTL_BELL		0x200	// Seen %B - ring the bell
#define	PTL_APPEND		0x400	// Seen %A - appended to the last line
#define	PTL_DEBUG_FILE		0x800	// Seen %! - append to file DEBUG - Only if TERMDEBUG is #defined


// Different pallets are used in different situations
// I can't think of a better word, pallet is used to mean other things; but this is how I use it here.
// Note that the values must fit within PTL_PALLET_MASK.
#define	PAL_WINDOW		0		// A normal window editing area
#define	PAL_MODE		1		// Mode line under a window
#define	PAL_MSG_INFO		2		// Message line: information message
#define	PAL_MSG_WARN		3		// Message line: warning message
#define	PAL_MSG_ERROR		4		// Message line: error message
#define	PALLETS			5		// Count of pallets

// Inks (pens) used in different situations.
// Note that pens have width as well as colour (think something like 읒)
// The first group here is found in every pallet, index into inkWell[]:
#define	INK_PLAIN		0		// Normal text
#define	INK_WRAP		1		// Last/first char showing more to right/left
#define	INK_WIDTH_NARROW	2		// Char with width less than 1
#define	INK_WIDTH_WIDE		3		// Char with width more than 1
#define	INK_WIDE_EDGE		4		// Char with width more than 1 but replaced at edge with width 1 replacement
#define	INK_EOL_MARKER		5		// End of line marker (left/right scrolling)
#define	INKS			6		// Count of pens above

/* The following are only in normal windows, editing area, index into pens[].
 * Beware: some of these can be used on wide characters. The appropriate width
 * MUST be set, copying from the current attribute if needs be with va_get_width().
 */
#define	PEN_MARK		(PALLETS * INKS + 0)	// Mark
#define	PEN_CURSOR		(PALLETS * INKS + 1)	// Cursor active window
#define	PEN_NAW_CURSOR		(PALLETS * INKS + 2)	// Cursor in non active windows
#define	PEN_CURSOR_ALT		(PALLETS * INKS + 3)	// Alternate colour, when over character in standout attrib
#define	PEN_LINE_NUMBERS	(PALLETS * INKS + 4)	// Line numbers
#define	PEN_TAB			(PALLETS * INKS + 5)	// Visible tabs

// Borders (not mode line):
#define	PEN_CURR_BORDER		(PALLETS * INKS + 6)	// Current window
#define	PEN_BORDER		(PALLETS * INKS + 7)	// Non current window

#define	PEN_EOL_POSITION	(PALLETS * INKS + 8)	// End of line position

#define	PEN_BLANK_SCREEN	(PALLETS * INKS + 9)	// Notional blank screen; window independent; hmmm

#define	PEN_MSGL_COLOUR		(PALLETS * INKS + 10)	// Colouring for appended parts of the message line



// Hashing
#define	HASH_EMPTY	SHRT_MAX	// Mark a hash slot as empty
#define	HASH_MISSING	HASH_EMPTY	// Return value to mean entry not found

/* Hash lookups result in an index into some table.
 * This is the type of the index.
 * As small as possible, but must be more than 255.
 *
 * A value of HASH_EMPTY means that the slot is not used
 */
typedef	short	hashNx;

/* This describes a struct so that something within it can be hashed.
 * The structure will be used in an array that is terminated by a
 * pointer within it being NULL.
 *
 * See hash.c
 */
typedef	struct	hashSt	{
	void*	hs_table;	// The array that contains the structs to hash
	hashNx*	hs_index;	// Indicies into table - hashed to find index
	hashNx	(*hs_hash)(struct hashSt*, const void*); // Generate a hash value
	int	(*hs_cmp)(const void*, const void*); // compare
	int	hs_keyOff;	// Offset of unique key to hash in struct
	int	hs_ptrOff;	// Offset of pointer in struct - if NULL means end
	int	hs_size;	// Sizeof the struct type
	int	hs_used;	// Entries used in hs_index
	int	hs_slots;	// Available slots in hs_index
} hashSt;


/* Include the error list if compiling under UNIX.
 */
#if	UNIX
#include <errno.h>
#ifndef	LINUX5
extern	int	errno;
// No longer needed, use strerror()
// extern	char*	sys_errlist[];
#endif
#endif

/* And now for some nasty, nitty gritty stuff ....
 */
#if	 BSD_UNIX
extern	char*	index();
extern	char*	rindex();
#define	strchr	index
#define	strrchr	rindex
#else
#ifndef	LINUX5
extern	char*	strchr();
extern	char*	strrchr();
#define	index	strchr
#define	rindex	strrchr
#endif
#endif


/* end */
