/*
 * ADDW SCCS: @(#)eval.c	1.216 10/14/20 22:26:05
 */

/*	EVAL.C:	Expresion evaluation functions for
		MicroEMACS

	written 1986 by Daniel Lawrence				*/

#include	"estruct.h"
#include	"utf8.h"
#include	"edef.h"
#include	"evar.h"
#include	"evers.h"

#if	SHOWMODREL
char	evalvers[] = "@(#)eval.c	1.216 10/14/20 22:26:05";
#endif

// Describe environment variables to the hash code
hashSt	hashEnv = {
	envars,
	NULL,
	hashStr,
	hashStrcmp,
	offsetof(ENVVAR, ev_name),
	offsetof(ENVVAR, ev_name),
	sizeof(ENVVAR),
	0,
	0
};

// Describe functions to the hash code
hashSt	hashFunc = {
	funcs,
	NULL,
	hashStr,
	hashStrcmp,
	offsetof(UFUNC, f_name),
	offsetof(UFUNC, f_name),
	sizeof(UFUNC),
	0,
	0
};

/* initialize the user variable list
 * and environment variable hash table.
 */
void	varinit(void)
{
	/* allocate the global user variable table */
	uv_global = uv_head =
		(UTABLE*)malloc(sizeof(UTABLE) + MAXVARS * sizeof(UVAR));

	/* and set up its fields */
	uv_head->ut_next = (UTABLE*)NULL;
	uv_head->ut_size = MAXVARS;
	uv_head->ut_bufp = (BUFFER*)NULL;
	uv_init(uv_head);

	hashInit(&hashEnv);
	hashInit(&hashFunc);
}


/* Initialize the entries in one user variable table
 *
 * Maybe u_value should be assigned NULL.

 * This is called on startup and at the start of every procedure and uv_clean()
 * at its end. To avoid the overhead of valGet()/valRelease() use the static
 * undefVAL.
 */
void	uv_init(UTABLE* ut)
{
	int 	i;

	for(i = 0; i < ut->ut_size; i++) {
		ut->ut_uv[i].u_name[0] = '\0';
		ut->ut_uv[i].u_value = valGetUndef();
	}
}

/* Discard the contents of a user variable table.
 *
 * BEWARE: this will leave u_value pointing at freed strings,
 * so you should prob free the table quickly.
 */
void	uv_clean(UTABLE* ut)
{
	int	i;

	/* now clear the entries in this one */
	for(i = 0; i < ut->ut_size; i++)
		valRelease(ut->ut_uv[i].u_value);
}

/* discard and clear all user variable tables
 */
void	varclean(UTABLE* ut)
{
	/* first clean all the ones under this one */
	if(ut->ut_next != (UTABLE*)NULL)
		varclean(ut->ut_next);

	/* clear the contents of this table */
	uv_clean(ut);

	/* and then deallocate the this table itself */
	free(ut);
}

/* evaluate a function
 * name:	name of function to evaluate
 *
 * The name can either be the long name or truncated to 3 chars.
 * The long name use is preferred: lookup faster, more meaningful name
 *
 * Return a VALue that is the result of evaluating the function
 */
VAL*	gtfun(const BYTE* fname)
{
	register int fnum;		/* index to function to eval */
	char*	tsp;			/* temporary string pointer */
	BUFFER*	bp;			/* scratch buffer pointer */
	VAL*	args[3];		// Values of arguments here
	VAL*	result = (VAL*)NULL;	// What to return
	int	nargs;			/* How many arguments */
	long	ia[3];			// Arguments as a number - if that is what they are
	int	i;
	long	l;
	int	incr;
	int	SomeError = 0;		// If TRUE something has gone wrong
	int	resISa1 = 0;		// TRUE if result is in same variable same as arg1
	Array*	ap;			// Used in array manipulation
	Hash*	hp;			// Used in hash manipulation
	VAL**	vpp;
	VAL*	needle;
	BYTE*	bytep;

	// Look up the name using the hash code
	if((fnum = hashLookup(&hashFunc, &fname)) == HASH_MISSING) {
		// If that does not find it truncate to 3 chars max & iterate search

		BYTE	funcName[4];	// Copy of fname that we can mangle

		/* look the function up in the function table
		 * only first 3 chars significant
		 * and let it be upper or lower case
		 */
		memcpy(funcName, fname, sizeof(funcName) - 1);
		funcName[sizeof(funcName) - 1] = '\0';
		mklower(funcName);

		for(fnum = 0; fnum < (int)NFUNCS; fnum++)
			if(strncmp(funcName, funcs[fnum].f_name, (sizeof(funcName) - 1)) == 0)
				break;

		/* return errorm on a bad reference */
		if(fnum == NFUNCS) {
			mlwrite("%#%EUnknown function &%s", fname);
			return((VAL*)&errorVAL);
		}
	}

	// So that we know, at return, what has been given a value
	args[0] = args[1] = args[2] = (VAL*)NULL;

	nargs = 0;
	if(U1t(funcs[fnum].f_type) != V_UNDEF)
		nargs = 1;
	if(U2t(funcs[fnum].f_type) != V_UNDEF)
		nargs = 2;
	if(U3t(funcs[fnum].f_type) != V_UNDEF)
		nargs = 3;

	// Get values for the arguments:
	for(i = 0; i < nargs; i++) {
		// Almost always start str, might get changed
		args[i] = valGet(Unt(funcs[fnum].f_type, i + 1));

		if(macarg(&args[i]) != TRUE &&
		   !(Unt(funcs[fnum].f_type, i + 1) == V_ANY && vtype(args[i]) == V_ERROR)) {
			mlwrite("%#%ECannot get arg%d to function &%s", i + 1, funcs[fnum].f_name);
			mlwrite("%#Expected %s got %s", V_TypeNames[Unt(funcs[fnum].f_type, i + 1)], vtypename(args[i]));
			// mlwrite("%#execText='%t'/%d", execText, execPos);
			SomeError = TRUE;
			goto error;
		}

		/* Do we check the first argument for UTF8 goodness ?
		 * This string could have come from anywhere, it can't be trusted.
		 *
		 * Check it before unleashing on the code below. Note that we
		 * don't test everything that deals with strings just those where
		 * bad encoding could cause problems.
		 *
		 * In practice we will prob end up testing the same good strings
		 * over & again - can't do much else - they are not tagged.
		 */
		if(i == 0 &&
		   (funcs[fnum].f_type & UU8STR) != 0 &&
		   vtype(args[i]) == V_STR &&	// Maybe permanent check
		   !utf8_check(vtext(args[0]))) {
			mlwrite("%#Badly UTF8 encoded argument to function &%s", funcs[fnum].f_name);
			SomeError = TRUE;
			goto error;
		}
		// mlwrite("%#arg%d is %s '%v'", i + 1, vtypename(args[i]), args[i]);

		// Copy INT types to convenience var:
		if(vtype(args[i]) == V_INT || vtype(args[i]) == V_BOOL || vtype(args[i]) == V_CHAR)
			ia[i] = vint(args[i]);
	}

	/* Does the result use arg1 or allocate another VALue
	 * Important: test that args[0] is only referenced once, else cannot
	 * just overwrite the value.
	 */
	if(nargs > 0 &&
	    vRefGet(args[0]) == 1 &&
	    (funcs[fnum].f_type & UUA1RES) != 0 &&
	    (resISa1 = U1t(funcs[fnum].f_type) == vtype(args[0])))
		result = args[0];
	else
		result = valGet(Urt(funcs[fnum].f_type));


	/* And now evaluate it! */
	switch(fnum) {
		case UFADD:	/* add two numbers together */
			ia[0] += ia[1];
			break;

		case UFSUB:	/* subtraction */
			ia[0] -= ia[1];
			break;

		case UFTIMES:	/* multiplication */
			ia[0] *= ia[1];
			break;

		case UFDIV:	/* division */
		case UFMOD:	/* mod */
			/* Want to avoid/trap SIGFPE - division by 0
			 * The obvious way does not work. Trapping the signal is easy, set
			 * a flag, re-arm signal & return: the kernel restarts the instruction
			 * and we get another SIGFPE, ... -> infinite loop.
			 * [sig]longjmp() is also problematic with SIGFPE.
			 * http://www.open-std.org/jtc1/sc22/wg14/www/docs/n1318.htm
			 * https://lkml.org/lkml/1996/4/27/27
			 * https://stackoverflow.com/questions/14233464/can-a-c-program-continue-execution-after-a-signal-is-handled/30169485
			 * http://technopark02.blogspot.com/2005/10/handling-sigfpe.html
			 *
			 * However we only do integer division (for now) so can only have
			 * a problem if the divisor is zero; so test for that.
			 */
			if(ia[1] == 0) {
				// Whoops! Say what happened, cause an error:
				mlwrite("%#%EDivision by zero: &%s %D %D", funcs[fnum].f_name, ia[0], ia[1]);
				SomeError = TRUE;
				break;
			}
			ia[0] = fnum == UFDIV ? (ia[0] / ia[1]) : (ia[0] % ia[1]);
			break;

		case UFNEG:	/* negate */
			ia[0] = -ia[0];
			break;

		case UFCAT:	/* concatenate string */
			if( !resISa1)
				tcpyt(vTEXT(result), vTEXT(args[0]));
			tcatt(vTEXT(result), vTEXT(args[1]));

			break;

		case UFLEFT:	/* left string(string, len) */
		    {		// Return len leftmost characters from string
			size_t want = (size_t)ia[1];				// Wanted number of characters
			size_t by_in = (size_t)tused(vTEXT(args[0]));		// Bytes in input string - that we extract from
			want = utf8_maxlen(vtext(args[0]), by_in, &want);	// # bytes in len characters
			tcpym(vTEXT(result), vtext(args[0]), (LineLn)want);
			break;
		    }

		case UFRIGHT:	/* right string(string, pos) */
		    {	// Return pos rightmost characters from string
			// If string is not that long - return what there is
			size_t want = (size_t)ia[1];				// Wanted number of characters
			size_t by_in = (size_t)tused(vTEXT(args[0]));		// Bytes in input string - that we extract from
			size_t ch_in = utf8_chrnlen(vtext(args[0]), by_in);	// # chars in input

			if(ch_in <= want) {					// Need all of input ?
				tcpyt(vTEXT(result), vTEXT(args[0]));
				break;
			}

			size_t skip = ch_in - want;	// Chars to skip at start
			// Find how many bytes to skip
			skip = utf8_maxlen(vtext(args[0]), by_in, &skip);
			tcpym(vTEXT(result), tgetcaddr(vTEXT(args[0]), skip), (LineLn)(by_in - skip));

			break;
		    }

		case UFMID:	/* mid string(string, pos, len) */
		    {	// Starting at pos chars in string return len chars
			// return(strncpy(result, &arg1[ia2-1], ia3));
			size_t start = (size_t)ia[1];				// Starting offset - characters
			size_t want = (size_t)ia[2];				// Wanted number of characters
			size_t by_in = (size_t)tused(vTEXT(args[0]));;		// Bytes in input string - that we extract from
			size_t ch_in = utf8_chrnlen(vtext(args[0]), by_in);	// # chars in input

			if(ch_in <= start) {
				tsetempty(vTEXT(result));
				break;
			}

			// Get how many bytes in the start characters:
			size_t st_by = utf8_maxlen(vtext(args[0]), by_in, &start);

			// Get how many bytes in want characters at offset start:
			size_t copy = utf8_maxlen(vtext(args[0]) + st_by, by_in - st_by, &want);

			// Copy
			tcpym(vTEXT(result), vtext(args[0]) + st_by, (LineLn)copy);
			break;
		    }

		case UFNOT:	/* logical not */
			ia[0] = ! ia[0];
			break;

		case UFEQUAL:	/* numeric equality check */
			ia[0] = ia[0] == ia[1];
			break;

		case UFLESS:	/* numeric less than */
			ia[0] = ia[0] < ia[1];
			break;

		case UFGREATER:	/* numeric greater than */
			ia[0] = ia[0] > ia[1];
			break;

		case UFLEQUAL:	/* numeric less or equal */
			ia[0] = ia[0] <= ia[1];
			break;

		case UFGEQUAL:	/* numeric greater or equal */
			ia[0] = ia[0] >= ia[1];
			break;

		case UFNEQUAL:	/* numeric not equal */
			ia[0] = ia[0] != ia[1];
			break;

		case UFIN:	/* number in array */
			// vpp  - pointer to array of values
			// incr - how much to increment vpp to get next value
			// i    - number of values
			if(vtype(args[1]) == V_ARRAY) {
				ap = &args[1]->v_v.v_array;
				vpp = ap->a_elems;
				i = ap->a_used;
				incr = 1;
			} else if(vtype(args[1]) == V_HASH) {
				hp = &args[1]->v_v.v_hash;
				vpp = hp->h_elems + 1;
				i = hp->h_used;
				incr = 2;
			} else {
				mlwrite("%ESecond argument to &%s must be array or hash. Given %s", funcs[UFNUMBER].f_name, vtypename(args[1]));
				SomeError = TRUE;
				break;
			}
			l = ia[0];
			ia[0] = FALSE;

			while(--i >= 0) {
				if((needle = valToInt(*vpp)) != (VAL*)NULL) {
					long	n = vint(needle);
					valRelease(needle);
					if(n == l) {
						ia[0] = TRUE;
						break;
					}
				}
				vpp += incr;	// Next element
			}
			break;

		case UFSLEQUAL:	/* string less or equality check */
			ia[0] = textcmp(vTEXT(args[0]), vTEXT(args[1])) <= 0;
			break;

		case UFSGEQUAL:	/* string greater or equality check */
			ia[0] = textcmp(vTEXT(args[0]), vTEXT(args[1])) >= 0;
			break;

		case UFSNEQUAL:	/* string not equality check */
			ia[0] = textcmp(vTEXT(args[0]), vTEXT(args[1])) != 0;
			break;

		case UFSEQUAL:	/* string logical equality check */
			ia[0] = textcmp(vTEXT(args[0]), vTEXT(args[1])) == 0;
			break;

		case UFSLESS:	/* string logical less than */
			ia[0] = textcmp(vTEXT(args[0]), vTEXT(args[1])) < 0;
			break;

		case UFSGREAT:	/* string logical greater than */
			ia[0] = textcmp(vTEXT(args[0]), vTEXT(args[1])) > 0;
			break;

		case UFSIN:	/* string in array or hash */
			// vpp  - pointer to array of values
			// incr - how much to increment vpp to get next value
			// i    - number of values
			if(vtype(args[1]) == V_ARRAY) {
				ap = &args[1]->v_v.v_array;
				vpp = ap->a_elems;
				i = ap->a_used;
				incr = 1;
			} else if(vtype(args[1]) == V_HASH) {
				hp = &args[1]->v_v.v_hash;
				vpp = hp->h_elems + 1;
				i = hp->h_used;
				incr = 2;
			} else {
				mlwrite("%ESecond argument to &%s must be array or hash. Given %s", funcs[UFNUMBER].f_name, vtypename(args[1]));
				SomeError = TRUE;
				break;
			}
			ia[0] = FALSE;

			// args[0] might not be STR, eg INT - get STR representation
			needle = valStrClone(valPrint(args[0]));

			while(--i >= 0) {
				if(textcmp(vTEXT(valPrint(*vpp)), vTEXT(needle)) == 0) {
					ia[0] = TRUE;
					break;
				}
				vpp += incr;	// Next element
			}
			valRelease(needle);
			break;

		case UFIND:	/* evaluate indirect value */
			tcpyt(vTEXT(result), vTEXT(getval(vTEXT(args[0]))));
			break;

		case UFAND:	/* logical and */
			ia[0] = ia[0] && ia[1];
			break;

		case UFOR:	/* logical or */
			ia[0] = ia[0] || ia[1];
			break;

		case UFLENGTH:	/* string length - characters, not bytes */
			ia[0] = (long)utf8_chrnlen(vtext(args[0]), (size_t)tused(vTEXT(args[0])));
			break;

		case UFUPPER:	/* uppercase string */
				// Stops on NUL, overwrites input - so don't copy to result
			if( !resISa1)
				tcpyt(vTEXT(result), vTEXT(args[0]));
			mkupper(vtext(result));
			break;

		case UFLOWER:	/* lower case string */
			if( !resISa1)
				tcpyt(vTEXT(result), vTEXT(args[0]));
			mklower(vtext(result));
			break;

		case UFSPLIT:	// substrings from input separated by string
			bytep = vtext(args[1]);
			l = tused(vTEXT(args[1]));
			while(tused(vTEXT(args[0])) > 0 && l > 0) {
				BYTE*	sub = memmem(bytep, (size_t)l, vtext(args[0]), (size_t)tused(vTEXT(args[0])));
				if(sub == NULL)	// Notfound
					break;

				VAL*	ss = valGet(V_STR);
				tcpym(vTEXT(ss), bytep, (LineLn)(sub - bytep));
				valArrayPush(result, ss);
				valRelease(ss);
				l -= sub - bytep + tused(vTEXT(args[0]));
				bytep = sub + tused(vTEXT(args[0]));
			}

			/* = is important since it catches the case where there is a
			 * needle string at the very end of the haystack: want an empty
			 * array element.
			 */
			if(l >= 0) {
				VAL*	ss = valGet(V_STR);
				tcpym(vTEXT(ss), bytep, (LineLn)(tgetcaddr(vTEXT(args[1]), tused(vTEXT(args[1]))) - bytep));
				valArrayPush(result, ss);
				valRelease(ss);
			}
			break;

		case UFTRUTH:	/* lower case string */
			/* Truth of the universe logical test */
			ia[0] = ia[0] == 42;
			break;

		case UFASCII:	/* char to integer conversion */
			ia[0] = utf8_decode(vtext(args[0]));
			break;

		case UFCHR:	/* integer to char conversion */
			// UTF8 code works just nicely with ASCII as well
			tcpyc(vTEXT(result), (UniChar)ia[0]);
			break;

		case UFGTKEY:	/* get 1 character */
			tcpyc(vTEXT(result), (UniChar)tgetc(MGC_MAP, 0));
			break;

		case UFRND:	/* get a random number */
#if	SYSRAND
			// If not seeded - make something and seed
			// Not random to crypto standards
			if( !RandSeeded) {
				seed = (int)time(NULL);
#if	UNIX
				seed ^= getpid();
#endif
				srand((unsigned int)seed);
				RandSeeded = 1;
			}
			ia[0] = (rand() % labs(ia[0])) + 1;
			break;
#else
			ia[0] = (ernd() % labs(ia[0])) + 1;
			break;
#endif

		case UFABS:	/* absolute value of a number */
			ia[0] = labs(ia[0]);
			break;

		case UFSINDEX:	/* find the index of one string in another */
			ia[0] = sindex(vTEXT(args[0]), vTEXT(args[1]));
			break;

		case UFENV:	/* retrieve a system environment var */
#if	ENVFUNC
			tsp = getenv(vtext(args[0]));
			if(tsp == NULL)
				tsp = "";
			tcpys(vTEXT(result), tsp);
			break;
#else
			tsetempty(vTEXT(result));
			break;
#endif

		case UFBIND:	/* lookup what function name is bound to a key */
			tcpys(vTEXT(result), transbind(vtext(args[0])));
			break;

		case UFEXIST:	/* check if a file exists */
			ia[0] = fexist(vtext(args[0]));
			break;

		case UFFIND:	/* look for a file on the path... */
			/* Beware foundFiles could contain more than one file
			 * name - only copy the first one.
			 */
			if( !findFile(vtext(args[0]), fT_Macro))
				tsetempty(vTEXT(result));
			else
				tcpys(vTEXT(result), ttext(foundFiles));
			break;

		case UFSTAT:	// get information about a file
			if( ! macFnStat(result, args[0]))
				SomeError = TRUE;
			break;

 		case UFBAND:	/* bitwise and   9-10-87  jwm */
			ia[0] &= ia[1];
			break;

 		case UFBOR:	/* bitwise or    9-10-87  jwm */
			ia[0] |= ia[1];
			break;

 		case UFBXOR:	/* bitwise xor   9-10-87  jwm */
			ia[0] ^= ia[1];
			break;

		case UFBNOT:	/* bitwise not */
			ia[0] = ~ia[0];
			break;

		case UFXLATE:	/* XLATE character string translation */
			if(xlat(vTEXT(result), vTEXT(args[0]), vTEXT(args[1]), vTEXT(args[2])) != TRUE)
				SomeError = TRUE;
			break;

		case UFCALL:	/* call a procedure, return what is in !return */
			/* find it, return ERROR if it does not exist */
			if((bp = findProcedure(vtext(args[0]))) == (BUFFER*)NULL) {
				SomeError = TRUE;
				break;
			}

			/* Execute it, return the value in result.
			 * The type can be any type. See DRETURN in dobuf().
			 */
			ia[0] = dobuf(bp, &result);
			break;

		case UFDOCMD:	/* call a procedure, return TRUE/FALSE/... */
			// The argument is the command name. Do not mention it in funcs[]
			// as docmd() gets it directly from execText

			// The idea is to test for a command succeeding, prob in a !if or !while.
			// Do not want this failure to set $early-log - so preserve/restore it
			i = EarlyLog;

			ia[0] = docmd();

			EarlyLog = i;
			break;

		case UFGTCMD:	/* get 1 Emacs command */
			{	int	dummy1;
				Cnt	dummy2;
				BYTE	cmdseq[20];

				cmdstr(getcmd(tgetc(MGC_MAP | MGC_NO_REFRESH, 0), &dummy1, &dummy2), cmdseq);
				tcpys(vTEXT(result), cmdseq);
				break;
			}

		case UFISNUM:	/* is the arg a number?
				 * Differ from standard me: empty string is false.
				 */
			if(vtext(args[0])[0] == '\0')
				ia[0] = 0;
			else
				ia[0] = is_num(vtext(args[0]));
			break;

		case UFNUMBER:	/* Convert to number
				 * Decimal integers & also 0xnnn & 0nnn
				 */
			{	char*	endptr;
				long	num = strtol(vtext(args[0]), &endptr, 0);
				if(vtext(args[0])[0] == '\0' || *endptr != '\0') {
					mlwrite("%E&%s bad number '%v'", funcs[UFNUMBER].f_name, args[0]);
					SomeError = TRUE;
					break;
				}
				vint(result) = num;
			}

			break;

		case UFREVERSE:	/* reverse */
			// No overflow possible as result is bigger than NUL terminated argument
			// result in arg0
			if( !resISa1)
				tcpyt(vTEXT(result), vTEXT(args[0]));
			utf8_memrev(vtext(result), (size_t)tused(vTEXT(args[0])));

			break;

		case UFTRIM:	/* Trim whitespace off the end of a string
				 */
			trimstr(vTEXT(result), vTEXT(args[0]));
			break;

		case UFTYPE:	// Type of argument
			tcpys(vTEXT(result), vtypename(args[0]));
			break;

		case UFCOUNT:	// Count elements in an array
			ia[0] = args[0]->v_v.v_array.a_used;
			break;

		case UFACAT:	// catenate two arrays
			i = args[0]->v_v.v_array.a_used + args[1]->v_v.v_array.a_used;
			valArraySize(result, i);
			for(i = 0; i < args[0]->v_v.v_array.a_used; i++)
				valArrayPush(result, args[0]->v_v.v_array.a_elems[i]);
			for(i = 0; i < args[1]->v_v.v_array.a_used; i++)
				valArrayPush(result, args[1]->v_v.v_array.a_elems[i]);
			break;

		case UFPOP:	// pop from end of an array
			ap = &args[0]->v_v.v_array;
			if(ap->a_used < 1) {
				mlwrite("%EArray empty, cannot pop");
				SomeError = TRUE;
				break;
			}
			i = --ap->a_used;	// How many left
			valCopy(&result, ap->a_elems[i]);
			valRelease(ap->a_elems[i]);
			// There is still something there - sort of
			ap->a_elems[i] = valGetUndef();
			break;

		case UFPUSH:	// push onto end of an array
			valArrayPush(args[0], args[1]);
			ia[0] = args[0]->v_v.v_array.a_used;
			break;

		case UFSHIFT:	// pop from start of an array
			ap = &args[0]->v_v.v_array;
			if(ap->a_used < 1) {
				mlwrite("%EArray empty, cannot shift");
				SomeError = TRUE;
				break;
			}

			// Copy the 0'th element to the result
			valCopy(&result, ap->a_elems[0]);
			valRelease(ap->a_elems[0]);

			i = --ap->a_used;	// How many left
			if(i > 0)	// Move what is left down
				memmove(&ap->a_elems[0], &ap->a_elems[1],
					(size_t)(i * (int)sizeof(VAL*)));

			/* a_elems[i] has been copied down and is a duplicate a_elems[i-1]
			 * so risks being released twice. So put something where that
			 * does not matter.
			 */
			ap->a_elems[i] = valGetUndef();

			break;

		case UFUNSHIFT:	// push onto start of an array
			ap = &args[0]->v_v.v_array;
			if(ap->a_alloc <= ap->a_used)	// Need to make bigger ?
				valArraySize(args[0], ap->a_used);

			// Release the UNDEF that will be overwritten by the shift below
			valRelease(ap->a_elems[ap->a_used]);

			// Shift everything up one to make room for the value to be pushed
			if(ap->a_used > 0) {
				memmove(&ap->a_elems[1], &ap->a_elems[0],
					(size_t)((ap->a_used) * (int)sizeof(ap->a_elems[0])));

				/* elems[0] & elems[1] are the same, splat elems[0] with UNDEF
				 * otherwise elems[1] gets lost when elems[0] is released
				 * in the valCopy()
				 */
				ap->a_elems[0] = valGetUndef();
			}

			// Assign the element to be pushed to the start of the array
			valCopy(&ap->a_elems[0], args[1]);

			// Result is how many elements now in the array
			ia[0] = ++ap->a_used;

			break;

		case UFKEYS:	// return array of keys in hash
			hp = &args[0]->v_v.v_hash;
			valArraySize(result, hp->h_used);
			for(i = 0; i < hp->h_used; i++)
				valArrayPush(result, hp->h_elems[i * 2]);

			break;

		case UFHCAT:	// catenate two hashes
			i = args[0]->v_v.v_hash.h_used + args[1]->v_v.v_hash.h_used;
			valHashSize(result, i);

			// Note that [1] is inserted first so that if there are duplicate keys the values in [0] are in result
			for(i = 0; i < args[1]->v_v.v_hash.h_used; i++)
				valHashAdd(result, args[1]->v_v.v_hash.h_elems[i * 2], args[1]->v_v.v_hash.h_elems[i * 2 + 1]);
			for(i = 0; i < args[0]->v_v.v_hash.h_used; i++)
				valHashAdd(result, args[0]->v_v.v_hash.h_elems[i * 2], args[0]->v_v.v_hash.h_elems[i * 2 + 1]);

			break;

		case UFKEY_EXIST: // does key exist in a hash
			ia[0] = valHashFind(args[1], args[0], FALSE) != (VAL**)NULL;
			break;

		default:	/* never should get here */
			Die("Error gtfun('%s') fnum=%d", fname, fnum);
	}

error:
	// Free args allocated - no longer used VARs
	// Not arg0 if the return result is there - unless error
	if(nargs > 0 && ( !resISa1 || SomeError))
		valRelease(args[0]);
	if(args[1] != (VAL*)NULL)
		valRelease(args[1]);
	if(args[2] != (VAL*)NULL)
		valRelease(args[2]);

	// Error ?
	if(SomeError)
		return((VAL*)&errorVAL);

	// Returning an INT - copy in the value
	if(Urt(funcs[fnum].f_type) == V_INT || Urt(funcs[fnum].f_type) == V_BOOL ||
	   Urt(funcs[fnum].f_type) == V_CHAR) {
		if(vtype(result) != Urt(funcs[fnum].f_type) || vRefGet(result) != 1) {
			valRelease(result);
			result = valGet(Urt(funcs[fnum].f_type));
		}

		vint(result) = ia[0];
	}

	return(result);
}

/* Find out stuff about filename
 *
 * result	hash to create with results
 *
 * Return:
 *	TRUE  it worked
 *	FALSE something went wrong - cannot stat() the file
 */
int	macFnStat(VAL* result, VAL* filename)
{
	char	ftype[2] = { 0, 0 };

	if(CacheStat(vtext(filename)) == -1)
		return(FALSE);

	switch(statbuf.st_mode & S_IFMT) {
	  case S_IFREG:
		ftype[0] = 'f';
		break;
	  case S_IFDIR:
		ftype[0] = 'd';
		break;
#ifdef	S_IFIFO
	  case S_IFIFO:
		ftype[0] = 'p';
		break;
#endif
	  case S_IFCHR:
		ftype[0] = 'c';
		break;
	  case S_IFBLK:
		ftype[0] = 'b';
		break;
#ifdef	S_IFSOCK
	  case S_IFSOCK:
		ftype[0] = 's';
		break;
#endif
#ifdef	S_IFDOOR
	  case S_IFDOOR:
		ftype[0] = 'D';
		break;
#endif
	  default:
		ftype[0] = '?';
		break;
	}

	valHashAddKs(result, "dev",     valIntToVAL((long)statbuf.st_dev));
	valHashAddKs(result, "ino",     valIntToVAL((long)statbuf.st_ino));
	valHashAddKs(result, "mode",    valIntToVAL((long)statbuf.st_mode));
	valHashAddKs(result, "nlink",   valIntToVAL((long)statbuf.st_nlink));
	valHashAddKs(result, "uid",     valIntToVAL((long)statbuf.st_uid));
	valHashAddKs(result, "gid",     valIntToVAL((long)statbuf.st_gid));
	valHashAddKs(result, "rdev",    valIntToVAL((long)statbuf.st_rdev));
	valHashAddKs(result, "size",    valIntToVAL((long)statbuf.st_size));
	valHashAddKs(result, "blksize", valIntToVAL((long)statbuf.st_blksize));
	valHashAddKs(result, "blocks",  valIntToVAL((long)statbuf.st_blocks));
	valHashAddKs(result, "atime",   valIntToVAL((long)statbuf.st_atime));
	valHashAddKs(result, "mtime",   valIntToVAL((long)statbuf.st_mtime));
	valHashAddKs(result, "ctime",   valIntToVAL((long)statbuf.st_ctime));
	valHashAddKs(result, "slink",   valStrToVAL(isSymlink(vtext(filename)) ? "l" : "f"));
	valHashAddKs(result, "type",    valStrToVAL(ftype));

	return(TRUE);
}

/*
 * num_to_onoff		if value contains a number then return "on"
 *			if value contains zero then return "off"
 */
BYTE*	num_to_onoff(const BYTE* value)
{
	if(atoi(value))
		return(StrOn);
	else
		return(StrOff);
}

/*
 * onoff_to_num		convert a text string (value) of "on" or "true" to 1
 *			convert a text string (value) of "off" or "false" to 0
 *			accept BOOL or INT as valid but dont convert if numeric
 *			Other types - fail
 *
 * NB: value is overwritten to contain '1' or '0' if input is valid.
 * No length checking as we put a string length 1 where there was
 * something longer.
 *
 * Return TRUE if we converted or found acceptable input: on/true/off/false/0/1.
 *	  FALSE - not a valid boolean or val is not type V_STR.
 */
int	onoff_to_num(VAL** val)
{
	BYTE*	value;

	if(vtype(*val) == V_BOOL || vtype(*val) == V_INT)
		return(TRUE);

	if(vtype(*val) != V_STR)
		return(FALSE);

	// Make it our own:
	if(vRefGet(*val) > 1) {
		VAL*	nv = valGet(V_STR);
		tcpyt(vTEXT(nv), vTEXT(*val));
		valRelease(*val);
		*val = nv;
	}

	value = ttext(&(*val)->v_v.v_text);
	if(cstrcmp(value, StrOn) == 0 || cstrcmp(value, truem) == 0) {
		strcpy(value, "1");
		return(TRUE);
	}

	if(cstrcmp(value, StrOff) == 0 || cstrcmp(value, falsem) == 0) {
		strcpy(value, "0");
		return(TRUE);
	}

	if((*value == '1' || *value == '0') && value[1] == '\0')
		return(TRUE);

	return(FALSE);
}


/* Look up & return a user var's value - as VAL.
 * If not found return NULL.
 *
 * This used to check that the name was not too long and if so truncate and
 * find that name. This truncation is no longer done. Legality checks are done
 * in findvar() - which is used user variables are set.
 *
 * vname	name of user variable to fetch - without leading %
 */
VAL*	gtusrVAL(const BYTE* vname)
{
	register int	vnum;	/* ordinal number of user var */
	register UTABLE*ut;	/* ptr to the current variable table */

	/* scan through each user variable table starting with the
	 * most local and going to the global table
	 */
	ut = uv_head;
	while (ut) {

		/* scan this table looking for the user var name */
		for (vnum = 0; vnum < ut->ut_size; vnum++) {

			/* out of entries? */
			if (ut->ut_uv[vnum].u_name[0] == 0)
				goto next_ut;

			/* is this the one? */
			if (strcmp(vname, ut->ut_uv[vnum].u_name) == 0) {
				/* return its value..... */
				return(ut->ut_uv[vnum].u_value);
			}
		}

next_ut:	ut = ut->ut_next;
	}

	/* We have run off the end - not found */
	return((VAL*)NULL);
}

/* Look up & return a user var's value - a string.
 * If not found return NULL.
 *
 * vname	name of user variable to fetch - without leading %
 *
 * BEWARE: the returned value must be used quickly, it might be in playValStr.
 * This will be if vname does NOT have a string value.
 */
BYTE*	gtusr(const BYTE* vname)
{
	VAL*	vp = gtusrVAL(vname);

	if(vp == (VAL*)NULL)
		return((BYTE*)NULL);

	return(vtext(vtype(vp) == V_STR ? vp : valPrint(vp)));
}

/* Look up & return a user var's value - a VALue.
 *
 * If not found print an error message & return an error value.
 *
 * If gtusrCkReport is FALSE do not print the message, but still return an
 * error. This is about a !while which has an optional second argument (an
 * array) in which case the first argument can legally be an uninitialised
 * variable. Yes: this ought to be passed down as an argument ... but there are
 * many levels.
 *
 * vname	name of user variable to fetch - without leading %
 */
VAL*	gtusrCk(const BYTE* vname)
{
	VAL*	value = gtusrVAL(vname);

	if(value != (VAL*)NULL)
		return(value);
	
	if(gtusrCkReport)
		mlwrite("%#%EMissing value to var '%%%s'", vname);

	return((VAL*)&errorVAL);
}

/* Find an environment variable with name vname.
 * vname does not start with '%'.
 *
 * Return a pointer to the entry in envars.
 *	NULL if it is not found.
 */
ENVVAR*	envFind(const BYTE* vname)
{
	hashNx		index;

	if((index = hashLookup(&hashEnv, &vname)) == HASH_MISSING)
		return((ENVVAR*) NULL);

	return(envars + index);
}

/* vname	name of environment variable to retrieve - not including the $
 *
 * BEWARE: could return to a static location, copy soon
 */
BYTE*	gtenv(const BYTE* vname)
{
	register int	vnum;	/* ordinal number of var refrenced */
	ENVVAR*		varp;
	long		l_value = 0;	// Init to quieten warning about uninitialised use of this
	char*		s_value = NULL;	// Init to quieten warning about uninitialised use of this
	int		gotValue = 0;
	static	char	s_result[INTWIDTH + 1];


	if((varp = envFind(vname)) == (ENVVAR*)NULL) {
		mlwrite("%#%ECannot find envvar '$%s'", vname);
		return(errorm);
	}
	vnum = (int)(varp - envars);

	// Can we get the value by using a pointer ?
	switch(varp->ev_type & EV_TYPEMASK) {
	  case EV_INT:
	  case EV_BOOL:
		if(varp->v.i == NULL)
			break;
		l_value = *varp->v.i;
		gotValue = 1;
		break;
	  case EV_STRING:
		if(varp->v.s == NULL)
			break;
		s_value = *varp->v.s;
		gotValue = 1;
		break;
	  case EV_LONG:
		if(varp->v.l == NULL)
			break;
		l_value = *varp->v.l;
		gotValue = 1;
		break;
	  case EV_TEXT:
		if(varp->v.t == NULL)
			break;
		s_value = ttext(*varp->v.t);
		gotValue = 1;
		break;
	}

	/* otherwise, fetch the appropriate value */
	if( !gotValue)
		switch(vnum) {
		  case EVPAGELEN:	// number of lines used by editor
			l_value = AllocLines;
			break;

		  case EVCURCOL:	// current column pos of cursor
			l_value = getdcol(FALSE);
			break;

		  case EVCURCHARPOS:	// current character position of cursor
			if(lisUtf8(curwp->w_dotp))
				l_value = (long)utf8_chrnlen(ltext(curwp->w_dotp), (size_t)curwp->w_doto);
			else
				l_value = (long)curwp->w_doto;
			break;

		  case EVCURLINE:	// curent line in file
			l_value = getcline();
			break;

		  case EVBUFLINES: 	// Number of lines in buffer
			l_value = curbp->b_numLines;
			break;

		  case EVRAM:		// ram in use by malloc
			l_value = envram / 1024L;
			break;

		  case EVCURWIDTH:	// current screen width
			l_value = AllocColumns;
			break;

		  case EVCBUFNAME:	// current buffer name
			s_value = curbp->b_bname;
			break;

		  case EVCFNAME:	// current file name
			s_value = curbp->b_fname;
			break;

		  case EVSRES:		// current screen resolution
			s_value = sres;
			break;

		  case EVPALETTE:	// current palette string
			s_value = palstr;
			break;

		  case EVACOUNT:	// # of chars until next auto-save
			l_value = gacount;
			break;

		  case EVCURCHAR:	// current character under the cursor
			/* Character under the cursor
			 * If UTF8 take care to return the Unicode value, not just the
			 * value of the first byte.
			 */
			l_value = lused(curwp->w_dotp) == curwp->w_doto ?
				'\n' :
				lisUtf8(curwp->w_dotp) ?
					utf8_decode(lgetcaddr(curwp->w_dotp, curwp->w_doto)) :
					lgetc(curwp->w_dotp, curwp->w_doto);
			break;

		  case EVSEED:		// current random number seed
			l_value = seed;
			break;

		  case EVWLINE:		// # of lines in current window
			l_value = curwp->w_ntrows;
			break;

		  case EVCWLINE:	// current screen line in window
			l_value = getwpos();
			break;

		  case EVTARGET:	// target for line moves
			saveflag = lastflag;
			l_value = curwp->w_goalCol;
			break;

		  case EVSEARCH:	// search pattern
			s_value = ttext(searchPat_e);
			break;

		  case EVKILL:		// kill buffer (read only)
			s_value = getkill();
			break;

		  case EVCMODE:		// mode of current buffer
			l_value = curbp->b_mode;
			break;

		  case EVPENDING:	// type ahead pending flag
#if	TYPEAH
			l_value = typahead(TRUE);
#else
			l_value = 0;
#endif
			break;

		  case EVLWIDTH:	// width of current line - in screen terms
			;		// Ie tabs variable width, Unicode chars 1 or 2
			LineLn	sav_doto = curwp->w_doto;
			curwp->w_doto = lused(curwp->w_dotp);
			l_value = (long)getdcol(FALSE);
			curwp->w_doto = sav_doto;
			break;

		  case EVLINE:		// text of current line
			s_value = getctext();
			break;

		  case EVTABWIDTH:	// width of displayed tabs
		  case EVTAB:		// width of displayed tabs
			l_value = curbp->b_tabsize;
			break;

		  case EVTERSE:		// be terse on output of modes on modeline
			l_value = beterse;		// num_to_onoff(beterse);
			break;

		  case EVAUTOC:		// turn on CMODE on editing .c and .h file
			l_value = (cmode == CMDON);	// num_to_onoff(cmode == CMDON);
			break;

		  case EVAUTOEC:	// extended CMODE stuff
			l_value = (cmode == CMDEXT);	// num_to_onoff(cmode == CMDEXT);
			break;

		  case EVMETA_ATTRIB: /* Meta & control characters displayed using attributes */
			l_value = (DefaultWindowOpts & WO_META_ATTRIB) != 0;
			break;

		  case EVCOLOURISE_TAB:	/* Show tabs in colour */
			l_value = (DefaultWindowOpts & WO_COLOUR_TABS) != 0;
			break;

		  case EVCURWIND:	// current window number
			l_value = WindowNumber(curwp) + 1;
			break;

		  case EVFCOL:		// first displayed column number, window left shift
			l_value = curwp->w_leftoffset;
			break;

		  case EVLASTMESG:	// last message written to command line
			s_value = getlinestr(lback(bmlHistp->b_linep));
			break;

		  case EVTIME:		// current date and time
			{
				time_t	tim = time(NULL);
				char*	s;
				s_value = ctime(&tim);
				if((s = strchr(s_value, '\n')))
					*s = '\0';	// Kill ending newline
			}
			break;

		  case EVMARKCOL:	// current column pos of mark
			// current column pos of mark
			// Save ".", move to mark, measure, move back
			if(curwp->w_markp) {
				swapmark(FALSE, 0);
				l_value = getdcol(FALSE);
				swapmark(FALSE, 0);
			} else
				l_value = -1;
			break;

		  case EVMARKLINE:	// current line in file of mark
			// current line in file of mark
			// Save ".", move to mark, measure, move back
			if(curwp->w_markp) {
				swapmark(FALSE, 0);
				l_value = getcline();
				swapmark(FALSE, 0);
			} else
				l_value = -1;
			break;

		  case EVTF8_ERROR:	// Description of $utf8-errno
			s_value = utf8_perror();
			break;

		  case EVSECONDS:	// Time since the epoch, seconds
			l_value = (long)time((time_t*)NULL);
			break;

		  case EVPID:		// Emacs process ID
			l_value = (long)getpid();
			break;

		  case EVSCROLL:	// Will the screen scroll up/down ?
			l_value = (long)(TTdoes(TT_DOES_SCROLL) && ttDoScroll);
			break;

		  case EVSCROLLLR:	// Will the screen scroll left/right ?
			l_value = (long)(TTdoes(TT_DOES_INSERT_CHARS) && ttDoScrollLr);
			break;

		  default:
			Die("%EBug: gtenv(%s)", vname);
			/* NOT REACHED */
		}

	// So we should have varp at the ENVVAR for the variable and
	// the value in l_value or s_value.
	switch(varp->ev_type & EV_TYPEMASK) {
	  case EV_INT:
	  case EV_LONG:
		if(varp->ev_type & EV_HEXINT) {
			strcpy(s_result, "0x");
			mlputli(l_value, 16, s_result + 2);
		} else {
			mlputli(l_value, 10, s_result);
		}
		return(s_result);
	  case EV_BOOL:
		if(varp->ev_type & EV_ONOFF)
			return(l_value ? StrOn : StrOff);
		return(l_value ? truem : falsem);
	  case EV_STRING:
	  case EV_TEXT:
		return(s_value);
	  case EV_CHAR:
		s_result[0] = (char)l_value;
		s_result[1] = '\0';
		return(s_result);
	}

	__NOT_REACHED
}

/* return some of the contents of the kill buffer
 */
BYTE*	getkill(void)
{
	int	size;			/* max number of chars to return */
	LINE*	lp;
	static BYTE value[NSTRING];	/* temp buffer for value */

	if(bmlKill->b_numLines <= 0)
		/* no kill ....just a null string */
		value[0] = 0;
	else {
		/* copy in the contents... */
		lp = lback(bmlKill->b_linep);
		size = (int)lused(lp);
		if(size >= NSTRING)
			size = NSTRING - 1;
		strncpy(value, ltext(lp), (size_t)size);
		value[size] = '\0';
	}

	/* and return the constructed value */
	return(value);
}

/* Trim whitespace off the end of a string
 *
 * res	Where the result is put
 * str	String to trim
 */
void	trimstr(TEXT* res, TEXT* str)
{
	tsetempty(res);
	if(tused(str) == 0)
		return;

	BYTE*	end = tgetcaddr(str, tused(str));
	while(end >= tgetcaddr(str, 0)) {
		if(end[-1] != ' ' && end[-1] != '\t')
			break;
		end--;
	}

	tcpym(res, tgetcaddr(str, 0), (LineLn)(end - tgetcaddr(str, 0)));
}

/* Find the next Emacs environment variable - eg: $french-spacing
 *
 * pos		Preserved position pointer. Set NULL on keystroke that is not for/back-ward search.
 * dirn		direction to go, +ve forward, -ve back
 * prompt	Prompt at the start of the line
 *
 * *pos is set to where we are in the envars table.
 * If this is NULL find an initial position by looking for the first
 * command that starts with the text in buf.
 * 'first' is first from start or end depending on dirn.
 *
 * Normally return no-character UNISTR_END, but on error *could* return the next one read
 *
 * We know that wmessp->w_dotp is good to hold at least the prompt.
 * NextVarName() does this.
 *
 * buf will start with a '$'
 *
 * Return the next keystroke or UNISTR_END
 */
KeyCode	NextEnvVarName(void** pos, int dirn, const BYTE* prompt)
{
	ENVVAR*	epos = *pos;		// Get state
	static	ENVVAR*	lastVar;	// Last entry in environment variable table envars
	int	promptLen = (int)strlen(prompt) + 1;	// Includes the '$'
	LINE*	dotp = wmessp->w_dotp;

	// Going backwards, end not known, find it
	if(dirn < 0 && !lastVar) {
		for(lastVar = envars; lastVar->ev_name; lastVar++)
			;

		// Back up to the last one
		lastVar--;
	}

	// No state, find something
	if( !epos) {
		// Find where to start searching

		epos = (dirn > 0) ? envars : lastVar;	// start/end of the array

		/* Now search until we find something that matches what we have in dotp
		 * Description of strncmp() does not say what happens if compared length is 0,
		 * so deal with that specially by not moving from where we are.
		 * Since envars does not contain the '$' - one less.
		 */
		if(lused(dotp) > promptLen) {
			while(epos) {
				if(strncmp(lgetcaddr(dotp, promptLen), epos->ev_name, (size_t)lused(dotp) - (size_t)promptLen) == 0)
					break;
				// No match, move to next. Beware the ends
				if(dirn > 0) {
					// Forwards
					if( !(++epos)->ev_name)
						epos = (ENVVAR*)NULL;
				} else {
					// Backwards
					if(--epos < envars)
						epos = (ENVVAR*)NULL;
				}
			}

			// What did we find ? Did we run off the end before a match ?
			if( !epos)
				return(mlerr("Environment variable not found", (char*)NULL));

		}

	} else {
		// There is a previous match, just go next/prev.
		// No match, move to next. Beware the ends; wrap around
		if(dirn > 0) {
			// Forwards
			if( !(++epos)->ev_name)
				epos = envars;
		} else {
			// Backwards
			if(--epos < envars)
				epos = lastVar;
		}
	}

	// Put what we found in dotp and set length
	lsetused(dotp, promptLen);
	insStrLine(dotp, promptLen, epos->ev_name, (LineLn)strlen(epos->ev_name), 1);
	wmessp->w_doto = lused(dotp);

	*pos = epos;		// Preserve state

	return(UNISTR_END);
}

/* Comments as for NextEnvVarName().
 *
 * We know that wmessp->w_dotp is good to hold at least the prompt.
 * NextVarName() does this.
 *
 * buf will start with a '%'.
 *
 * Return the next keystroke or UNISTR_END
 */
KeyCode	NextUserVarName(void** pos, int dirn, const BYTE* prompt)
{
	static	int	nvars;		// Number of user variables
	static	int	cvar;		// The number of the user var last returned
	static	UTABLE*	vut;		// Ther variable table
	int	promptLen = (int)strlen(prompt) + 1;	// Includes the '%'
	LINE*	dotp = wmessp->w_dotp;

	// First time ?
	// This prompt anyway, the number of vars is likely to change between prompts
	if(!*pos) {
		// Count the number of user vars that we have at the moment:
		for(nvars = 0, vut = uv_head; vut; vut = vut->ut_next) {
			// Skip until we find the next in the list
			if(vut->ut_next)
				continue;
			while(vut->ut_uv[nvars].u_name[0])
				if(++nvars > vut->ut_size) {
					// That will have found 1 more than the number 
					nvars--;
					break;
				}

			break;
		}

		// Set the current one to one beyond the end
		cvar = dirn > 0 ? -1 : nvars + 1;

		*pos = &cvar;
	}

	if(nvars == 0)
		return(mlerr("No user variables currently exist", (char*)NULL));

	// Next in the direction requested
	cvar += dirn > 0 ? 1 : -1;

	// Wrap round at ends:
	if(cvar < 0)
		cvar = nvars - 1;
	if(cvar >= nvars)
		cvar = 0;

	// Copy in the cvar'th varible:
	lsetused(dotp, promptLen);
	insStrLine(dotp, promptLen, vut->ut_uv[cvar].u_name, (LineLn)strlen(vut->ut_uv[cvar].u_name), 1);
	wmessp->w_doto = lused(dotp);

	return(UNISTR_END);
}

/* Display an environment or user variable
 *
 * Command:  display
 * Bound to: C-x g
 *
 * The count is ignored.
 *
 * Return TRUE  on OK
 *	  FALSE Nothing entered at prompt
 *	  ABORT User aborts the prompt - C-g
 *	  getVarValue()
 */
int	dispvar(int f, Cnt n)
{
	register int status;	/* status return */

	/* first get the variable to display */
	status = mlprompt("Variable to display: ", MLEMACSVAR, bmlVarHistp);

	if(status != TRUE)
		return(status);

	if((status = getVarValue(ltext(bmlVarHistp->b_dotp), playText)) != TRUE) {
		mlwrite("%E""%l - is not set", bmlVarHistp->b_dotp);
		return(status);
	}

	mlwrite("%l = %t", bmlVarHistp->b_dotp, playText);

	return(status);
}

/* Set a variable.
 *
 * Prompt for a value for the variable.
 *
 * If the count is set, the count value is assigned to the variable.
 * The count is not set by default.
 *
 * Examples:
 *     set %name fred
 *     set %a[3] "name 3"
 *     set %a[&add %num 2] &cat "number " %val
 *     set %h{foo} "bar"
 *
 * Command:  set
 * Bound to: C-x v
 *
 * Return:
 *	TRUE  Success
 *	FALSE bad variable name
 *	      Nothing entered at prompt
 *	ABORT User aborts the prompt - C-g
 *	svar()
 */
int	setvar(int f, Cnt n)
{
	register int status;		// status return
	VDESC	vd;			// variable num/type
	TEXT*	var;			// name of variable to fetch
	BYTE	vn[NVSIZE + 1];		// Variable name
	VAL*	vp;
	VAL**	varp;			// Pointer to user variable

	/* First get the name of the variable to set.
	 * Must NOT use mlreply() as when running a macro it will
	 * grab the variable name, evaluate it & return the current value.
	 */
	if(clexec == FALSE) {
		status = mlprompt("Variable to set: ", MLEMACSVAR, bmlVarHistp);

		if(status != TRUE)
			return(status);
		var = &ltxt(bmlVarHistp->b_dotp);

	}  else {	/* macro line argument */
		/* grab token and skip it */
		var = playVar;
		token(var);
	}

	/* check the legality and find the variable.
	 * This will create user variables - even though we do not know if we have a good value.
	 */
	/* if its not legal....bitch */
	if((status = findvar(var, &vd, NVSIZE + 1, VT_GLOBAL | VT_PROBE)) == FALSE) {	// vd.v_type == TKBAD) {
		mlwrite(BadVar, var);
		return(FALSE);
	}

	// User or environment variable ?
	varp = vd.v_type == TKVAR ? &vd.v_ut->ut_uv[vd.v_num].u_value : NULL;

	/* We have the variable name but somewhere that might be
	 * overwritten. One particular case is where the value involves
	 * &call of a procedure within which a variable is set.
	 * Keep a copy of the name so that we findvar() the right one later.
	 */
	strncpy(vn, ttext(var), NVSIZE);
	vn[NVSIZE] = '\0';

	/* If the %var exists & name is immediately followed by '[' or '{' then
	 * get hold of a VAL** that we can assign to later.
	 */
	if(varp != NULL &&
	   (tgetch(execText, execPos) == '[' || tgetch(execText, execPos) == '{'))
		if((varp = parseStructVar(*varp, TRUE)) == NULL)
			return(FALSE);

	// Get the value for that variable, VAL vp
	if(f == TRUE) {
		vp = valGet(V_INT);
		vSetInt(vp, n);
	} else {
		/* This is ripped out of mlprompt()
		 * The point is that we want vp to end up typed and (possibly)
		 * the result of an &function or %variable.
		 * Scalar types will convert - array/hashes won't
		 */
		vp = valGet(V_ANY);

		int	ret = nextarg("Value: ", &vp, MLSTRING, bmlValHistp);

		if(ret != TRUE) {
			valRelease(vp);
			return(ret);
		}

		// Add as a new line to the history
		// Beware type in valbuf might not be string
		if(clexec == FALSE) {
			if(!addHistoryText(bmlValHistp,
				vTEXT(vtype(vp) == V_STR ? vp: valPrint(vp)), AHL_ADD_ZLEN)) {
				TTbeep();
				valRelease(vp);
				return(FAILED);
			}

			// In case the history buffer is in the current window:
			bmlValHistp->b_dotp = lback(bmlValHistp->b_linep);
			bmlValHistp->b_doto = lused(bmlValHistp->b_dotp);
		}
	}

	// This time create the variable, first restore its name:
	tcpys(var, vn);

	if(varp == NULL)
		// Environment variable
		status = svar(&vd, vp);	// Find & create if nesc
	else	// Quicker copy
		valCopy(varp, vp);

	valRelease(vp);

#if	DEBUGM
	/* if $debug == TRUE, every assignment will echo a statment to
	 * that effect here.
	 */
	
	if(macbug && (strcmp(ttext(var), "%track") != 0)) {
		/* write out the debug line */
		mlforce("(((%d:%s <- %v)))", status, ttext(var), vp);
		update(TRUE);

		/* and get the keystroke to hold the output ...
		 * but not if running execute-command-line - see execcmd()
		 */
		if(in_execcmd == FALSE && get1key() == abortc) {
			mlforce("%IMacro aborted");
			status = FALSE;
		}
	}
#endif

	/* and return it */
	return(status);
}

/* Create a variable
 * varType will have the value VT_GLOBAL or VT_LOCAL
 *
 * Does the work of global_var() & local_var()
 *
 * Return TRUE  OK
 *	  FALSE Error - a message will have been generated
 *	        Nothing entered at prompt
 *	  ABORT User aborts the prompt - C-g
 */
int	createVariable(int varType)
{
	int	status;		/* status return */
	VDESC	vd;		/* variable num/type */
	TEXT*	var;		/* name of variable to fetch */
	VAL*	vp;		// Play value

	/* First get the name of the variable to set.
	 * Must NOT use mlreply() as when running a macro it will
	 * grab the variable name, evaluate it & return the current value.
	 */
	if(clexec == FALSE) {
		status = mlprompt((varType == VT_GLOBAL ?
				"Global variable to declare: " :
				"Local variable to declare: "),
				MLEMACSVAR, bmlVarHistp);
		if(status != TRUE)
			return(status);
		var = &ltxt(bmlVarHistp->b_dotp);

	} else {	/* macro line argument */
		/* grab token and skip it */
		var = playVar;
		token(var);
	}

	/* check the legality and find the var
	 * if its not legal....bitch
	 */
	if(findvar(var, &vd, NVSIZE + 1, varType) == FALSE) {	// vd.v_type == TKBAD) {
		mlwrite(BadVar, var);
		return(FALSE);
	}

	// Prompt for a value - does not have to be one,
	// important for macros where this is just a declaration without assignment
	if((status = mlprompt("Value: ", MLEMPTYOK, bmlValHistp)) != TRUE &&
	   status != FALSE)
		return(FALSE);

	// And set the appropriate value - if given
	if(status == TRUE) {
		vp = valGet(V_STR);
		tcpyl(vTEXT(vp), bmlValHistp->b_dotp);
	} else
		vp = valGetUndef();

	status = svar(&vd, vp);

	valRelease(vp);

	/* if $debug == TRUE, every assignment will echo a statment to
	   that effect here. */
	
	if(macbug && (strcmp(ttext(var), "%track") != 0)) {
		/* write out the debug line */
		mlforce((varType == VT_GLOBAL ?
				"(((Globally declare %t)))" :
				"(((Locally declare %t)))"),
			var);
		update(TRUE);

		/* and get the keystroke to hold the output */
		if(get1key() == abortc) {
			mlforce("%IMacro aborted");
			status = FALSE;
		}
	}

	/* and return it */
	return(status);
}

/* Declare a global variable.
 *
 * The count is ignored. A default value is not set.
 *
 * Command:  global
 * Bound to: unbound
 *
 * Return TRUE if OK
 *	  FALSE if error - a message will have been generated
 */
int	global_var(int f, Cnt n)
{
	return(createVariable(VT_GLOBAL));
}

/* Declare a local variable.
 *
 * The count is ignored. A default value is not set.
 *
 * Command:  local
 * Bound to: unbound
 *
 * Return TRUE if OK
 *	  FALSE if error - a message will have been generated
 */
int	local_var(int f, Cnt n)
{
	return(createVariable(VT_LOCAL));
}

/* Set %something to value, create it if needs be.
 *
 * var	 Name of the variable - starting with %
 * value String value - other types not supported
 *
 * Return TRUE  if that worked.
 *	  FALSE bad variable
 */
int	setUserVar(BYTE* var, const BYTE* value)
{
	VDESC vd;		/* variable num/type */

	/* check the legality and find the var */
	tcpys(playVar, var);
	
	/* if its not legal....bitch */
	if(findvar(playVar, &vd, NVSIZE + 1, VT_LOCAL) == FALSE) {
		mlwrite(BadVar, playVar);
		return(FALSE);
	}

	/* and set the appropriate value */
 	return(svar(&vd, valStrToVAL(value)));
}

/* Find a variable, determine it's type and name
 * The variable is created if not found.
 *
 * var		name of var to get, first char is: $, %, &
 * vd		structure to hold type and ptr - this is written to
 * size		size of var array, used for when var is '&ind' followed by var name
 * scope	intended scope of any created user variables (%something)
 * 		VT_NONE		Don't create it
 *		VT_LOCAL	Create local
 *		VT_GLOBAL	Create global
 *
 *		This no more:
 *		VT_PROBE	ORed in with VT_GLOBAL.
 *				The will not create a variable, but checks that
 *				it looks as if we could.
 *
 * Change: it always creates the var if it does not exist AND a type is given,
 * a type being VT_LOCAL or VT_GLOBAL.
 * The only time this causes a problem is if evaluating the value (command:
 * set) fails, but that just leaves a variable with value UNDEF
 *
 * Return:
 *	TRUE  Something worked, look in vd
 *	FALSE Did not work, vd is bad, an error message will have been generated
 */
int	findvar(TEXT* var, VDESC* vd, int size, int scope)
{
	int	vnum = 0;	/* subscript in variable arrays: 0 to keep the compiler happy for when v_type = TKBAD */
	int	vtype;		/* type to return */
	UTABLE*	vut;		/* user var table to search */
	ENVVAR*	varp;

//	scope &= ~(VT_PROBE);	// Remove probe bit

fvar:	vtype = TKBAD;
	vut = uv_head;

	switch(tgetch(var, 0)) {

		case '$': /* check for legal environment var */
			{
			char*	vname = tgetcaddr(var, 1);	// Name to look for

			if((varp = envFind(vname)) != (ENVVAR*)NULL) {
				vtype = TKENV;
				vnum = (int)(varp - envars);
				break;
			}

			// Not found
			mlwrite("%EVariable does not exist: %s", var);
			return(FALSE);
			}

		case '%':
			/* Is the name legal ?
			 * Note first char is more restricted. NVSIZE does include the '%'
			 */
			if(strchr(VarNameChar1, tgetch(var, 1)) == NULL ||
			   strspn(tgetcaddr(var, 1), VarNameChars) != strlen(tgetcaddr(var, 1)) ||
			   tused(var) > NVSIZE) {
				mlwrite("%EIllegal variable name '%t'", var);
				return(FALSE);
			}

			/* check for existing legal user variable */
			while(vut) {
				for(vnum = 0; vnum < vut->ut_size; vnum++)
					if(strcmp(tgetcaddr(var, 1),
					    vut->ut_uv[vnum].u_name) == 0) {
						vtype = TKVAR;
						goto retvar;
					}

				if(scope == VT_LOCAL)
					break;

				vut = vut->ut_next;	// Look at the global table
			}

			/* if we should not define/create one.... */
			if(scope == VT_NONE)
				break;

			/* scope it as requested */
			if(scope == VT_LOCAL)
				vut = uv_head;
			else
				vut = uv_global;

			/* create a new variable */
			for(vnum = 0; ; vnum++) {
				// Table full:
				if(vnum >= vut->ut_size) {
					mlwrite("%ECannot add variable '%t', table full", var);
					return(FALSE);
				}

				if(vut->ut_uv[vnum].u_name[0] == '\0') {
					vtype = TKVAR;
					// Free variable slot, copy in the name:
					strncpy(vut->ut_uv[vnum].u_name, tgetcaddr(var, 1), NVSIZE);
					vut->ut_uv[vnum].u_name[NVSIZE] = '\0';
					vut->ut_uv[vnum].u_value = valGetUndef();
					break;
				}
			}
			break;

		case '&':	/* indirect operator? */
			if(strncmp(tgetcaddr(var, 1), "ind", 3) == 0) {
				/* grab token, and eval it */
				execPos += 3;
				token(var);
				tcpyt(var, vTEXT(getval(var)));

				goto fvar;
			}
	}

	// If we get here and vtype is still TKBAD then we have not been able to find/create the variable

	/* return the results */
retvar:	vd->v_num = vnum;
	vd->v_type = vtype;
	vd->v_ut = vut;
	return(vtype != TKBAD ? TRUE : FALSE);
}

/* Set a variable
 *
 * var		variable to set
 * value	value to set to
 *
 * Return
 *	TRUE  if things went well
 *	FALSE TODO
 */
int	svar(VDESC* var, const VAL* value)
{
	register int	vnum;	/* ordinal number of var refrenced */
	register int	vtype;	/* type of variable to set */
	register UTABLE *vut;	/* user table pointer */
	register int	status;	/* status return */
	register UniChar c;	/* translated character */
	ENVVAR*		varp;
	int		setValue;
	long		l_value = 0;// Init to quieten warning about uninitialised use of this
	char*		play;
	VAL*		playVAL;

	/* simplify the vd structure (we are gonna look at it a lot) */
	vnum = var->v_num;
	vtype = var->v_type;
	vut = var->v_ut;

	/* and set the appropriate value */
	status = TRUE;

	switch(vtype) {
	case TKVAR: /* set a user variable */
		// Copy the value - changing the type if needs be
		valCopy(&(vut->ut_uv[vnum].u_value), (VAL*)value);

		/* setting a variable to error stops macro execution */
// 		if(strcmp(value, errorm) == 0)	// FIXME TODO
// 			status = FALSE;

		break;

	case TKENV: /* set an environment variable */
		varp = envars + vnum;

		// Readonly ?
		if(varp->ev_type & EV_READONLY) {
			mlwrite("%EYou cannot change the value of $%s", varp->ev_name);
			return(FALSE);
		}

		status = FALSE;	// Assume this while early checking

		/* This is nasty.
		 * value can be one of several types, we want to convert to a EV_TYPEMASK.
		 * There is not a one-to-one mapping.
		 * Create an acceptable type for EV_TYPEMASK and use valCopyValue() to
		 * do the conversion - which might fail with an error message.
		 * More checking is then needed.
		 *
		 * First generate a VAL type from an EV_ one - this is the type
		 * that we can valCopyValue() to generate:
		 */
		playVAL = valGet((int[]){V_INT, V_BOOL, V_STR, V_INT, V_STR, V_STR}[varp->ev_type & EV_TYPEMASK]);
		// This will succeed in cases where it should not:
		if(valCopyValue(&playVAL, (VAL*)value) != TRUE)
			goto EnvOff;

		// Parse & error check value
		// What is the type wanted ?
		switch(varp->ev_type & EV_TYPEMASK) {
		  case EV_INT:
		  case EV_LONG:
			l_value = vint(playVAL);
			if((varp->ev_type & EV_MIN) && l_value < varp->ev_min) {
				mlwrite("%EMinimum value for $%s is %D", varp->ev_name, varp->ev_min);
				goto EnvOff;
			}
			if((varp->ev_type & EV_MAX) && l_value > varp->ev_max) {
				mlwrite("%EMaximum value $%s is %D", varp->ev_name, varp->ev_max);
				goto EnvOff;
			}
			break;

		  case EV_BOOL:
			l_value = vint(playVAL);
			break;

		  case EV_STRING:
		  case EV_TEXT:
			play = vtext(valPrint(playVAL));
			if((varp->ev_type & EV_UTF8) && !utf8_check(play)) {
				mlwrite("%EValue for $%s is not valid UTF-8 '%s'", varp->ev_name, value);
				goto EnvOff;
			}

			// Length is either in terms of bytes or characters
			size_t	len = (varp->ev_type & EV_UTF8) ? utf8_strlen(play) : strlen(play);

			if((varp->ev_type & EV_MIN) && (long)len < varp->ev_min) {
				mlwrite("%EString to $%s is too short (minimum %D)", varp->ev_name, varp->ev_min);
				goto EnvOff;
			}

			if((varp->ev_type & EV_MAX) && (long)len > varp->ev_max) {
				mlwrite("%EString to $%s is too long (maximum %D)", varp->ev_name, varp->ev_max);
				goto EnvOff;
			}
			break;

		  case EV_CHAR:
			if(tused(vTEXT(playVAL)) > 1 || *vtext(playVAL) > UTF8_MAX_CP1) {
				mlwrite("%EValue for $%s is not ASCII '%v'", varp->ev_name, value);
				goto EnvOff;
			}
			break;
		}

		status = TRUE;	/* by default */
		setValue = 0;	// Did not set value in generic code below

		switch(varp->ev_type & EV_TYPEMASK) {
		  case EV_INT:
		  case EV_BOOL:
			if(varp->v.i == NULL)
				break;
			*varp->v.i = (int)l_value;
			setValue = 1;
			break;
		  case EV_STRING:
			if(varp->v.s == NULL)
				break;
			utf8_strfmake((BYTE**)varp->v.s, vtext(playVAL));
			setValue = 1;
			break;
		  case EV_LONG:
			if(varp->v.l == NULL)
				break;
			*varp->v.l = l_value;
			setValue = 1;
			break;
		  case EV_TEXT:
			if(varp->v.t == NULL)
				break;
			status = tcpyt(*varp->v.t, vTEXT(playVAL));
			setValue = 1;
			break;
		}

		// If the generic code could not be used:
		if( !setValue)
			switch(vnum) {
			  case EVPAGELEN:	// number of lines used by editor
				status = newsize(TRUE, (int)l_value);
				break;

			  case EVCURCOL:	// current column pos of cursor
				status = setccol((int)l_value);
				if( !status)
					mlwrite("%#Line is less that %D long", l_value);
				break;

			  case EVCURCHARPOS:	// current character position of cursor
				curwp->w_doto = 0;
				while(curwp->w_doto < lused(curwp->w_dotp) && --l_value >= 0)
					moveForw(1);
				break;

			  case EVCURLINE:	// curent line in file
				status = gotoline(TRUE, (int)l_value);
				break;

			  case EVRAM:		// ram in use by malloc
				break;

			  case EVCURWIDTH:	// current screen width
				status = newwidth(TRUE, (int)l_value);
				break;

			  case EVCBUFNAME:	// current buffer name
				setBufferName(vtext(playVAL));
				break;

			  case EVCFNAME:	// current file name
				setFnames(blistp, vtext(playVAL));
				curwp->w_flag |= WFMODE;
				break;

			  case EVSRES:		// current screen resolution
				status = TTrez(bytecopy(sres, vtext(playVAL), sizeof(sres) - 1));
				break;

			  case EVPALETTE:	// current palette string
				strncpy(palstr, vtext(playVAL), 48);
				spal(palstr);
				break;

			  case EVCURCHAR:	// current character under the cursor
				ldelete(1L, KL_DISCARD);	/* delete 1 char */
				c = (UniChar)l_value;
				if(c == '\n')
					lnewline(0);
				else
					linsert(1, c);
				moveBack(1);
				break;

			  case EVSEED:		// current random number seed
				seed = (int)l_value;
#if	SYSRAND
				srand((unsigned int)seed);	// Set the seed
				RandSeeded = 1;
#endif
				break;

			  case EVWLINE:		// # of lines in current window
				status = resize(TRUE, (int)l_value);
				break;

			  case EVCWLINE:	// current screen line in window
				if(l_value > curwp->w_ntrows) {
					mlwrite("%#Window is %d lines long, cannot go to line %D",
						curwp->w_ntrows, l_value);
					status = FALSE;
					break;
				}

				l_value -= getwpos();	// Now is how many lines to move by

				// Return TRUE on move zero lines
				status = l_value ? signedForwline((int)l_value) : TRUE;
				break;

			  case EVTARGET:	// target for line moves
				curwp->w_goalCol = (int)l_value;
				thisflag = saveflag;
				break;

			  case EVSEARCH:	// search pattern
				tcpys(searchPat_e, vtext(playVAL));
#if	MAGIC
				mcclear();
#endif
				break;

			  case EVCMODE:		// mode of current buffer
				if( !checkNewModes(curbp, (int)l_value))
					status = FALSE;
				curwp->w_flag |= WFMODE;
				break;

			  case EVLINE:		// text of current line
				putctext(vtext(playVAL));
				break;

			  case EVTABWIDTH:	// width of displayed tabs
			  case EVTAB:		// width of displayed tabs
				// Min/max values are already checked
				curbp->b_tabsize = Tabsize = (short)l_value;
				upmode(curbp, WFHARD);
				sgarbf = TRUE;
				break;

			  case EVTERSE:		// be terse on output of modes on modeline
				beterse = (int)l_value;
				upmode(NULL, WFMODE);
				break;

			  case EVAUTOC:		// turn on CMODE on editing .c and .h file
				cmode = l_value ? CMDON : CMDOFF;
				setcmode(cmode);
				break;

			  case EVAUTOEC:	// extended CMODE stuff
				cmode = l_value ? CMDEXT : CMDOFF;
				setcmode(cmode);
				break;

			  case EVMETA_ATTRIB: /* Meta & control characters displayed using attributes */
				if(l_value)
					DefaultWindowOpts |=   WO_META_ATTRIB;
				else
					DefaultWindowOpts &= ~(WO_META_ATTRIB);
				break;

			  case EVCOLOURISE_TAB:	/* Show tabs in colour */
				if(l_value)
					DefaultWindowOpts |=   WO_COLOUR_TABS;
				else
					DefaultWindowOpts &= ~(WO_COLOUR_TABS);
				// Refresh everything, a somewhat heavy handed way
				updgar();
				ReDraw();
				break;

			  case EVMARKCOL:	// current column pos of mark
				// Save ".", move to mark, move back
				if( ! curwp->w_markp) {
					// Set mark to "." if not set
					curwp->w_markp = curwp->w_dotp;
					curwp->w_marko = curwp->w_doto;
					curwp->w_markpartoff = curwp->w_partoff;
				}
				swapmark(FALSE, 0);
				status = setccol((int)l_value);
				swapmark(FALSE, 0);
				if( !status)
					mlwrite("%#Line is less that %D long", l_value);
				break;

			  case EVMARKLINE:	// current line in file of mark
				// Save ".", move to mark, move back
				if(!curwp->w_markp) {
					// Set mark to "." if not set
					curwp->w_markp = curwp->w_dotp;
					curwp->w_marko = curwp->w_doto;
					curwp->w_markpartoff = curwp->w_partoff;
				}

				swapmark(FALSE, 0);
				status = gotoline(TRUE, (int)l_value);
				swapmark(FALSE, 0);
				break;

			  case EVSCROLL:	// Will the screen scroll up/down ?
				ttDoScroll = (int)l_value;
				break;

			  case EVSCROLLLR:	// Will the screen scroll left/right ?
				ttDoScrollLr = (int)l_value;
				break;

			}
		EnvOff:	// End of setting environment variables
			// Freeing playVAL is important
			valRelease(playVAL);

		break;
	}

	return(status);
}

#if	UNIX == 0
/* atoi:	ascii string to integer......This is too
 *		inconsistant to use the system's
 *
 * Also atol()
 * Do atoi() with atol() and cast the return value.
 */
long	atol(BYTE* st)
{
	int result;	/* resulting number */
	int sign;	/* sign of resulting number */
	BYTE c;		/* current char being examined */

	result = 0;
	sign = 1;
	/* skip preceding whitespace */
	while(*st == ' ' || *st == '\t')
		++st;

	/* check for sign */
	if(*st == '-') {
		sign = -1;
		++st;
	}
	if(*st == '+')
		++st;

	/* scan digits, build value */
	while((c = *st++))
		if(Isdigit(c))
			result = result * 10 + c - '0';
		else
			return(0);

	return(result * sign);
}

int	atoi(BYTE* st)
{
	return((int)atol(st));
}

#endif	/* UNIX == 0 */

/*	IntToAsc:	integer to ascii string.......... This is too
 *			inconsistant to use the system's
 *	
 *	The name had to be changed from the nicer "itoa" because of conflict
 *	with the system library itoa.
 *
 *	i	integer (well, long) to translate to a string
 *
 * The return value is in a static area - copy/use quickly before overwritten.
 */
BYTE*	IntToAsc(long i)
{
	static BYTE result[INTWIDTH+1];	/* resulting string */

	mlputli(i, 10, result);

	return(result);
}

/* find the type of a passed token by looking at the start of it
 *
 * token	token to analyze
 *
 * Return a TK??? value
 */
int	gettyp(const TEXT* token)
{
	register BYTE c;	/* first char in token */

	/* grab the first char (this is all we need) */
	c = tgetch(token, 0);

	/* a numeric literal? */
	if(Isdigit(c) || ((c == '-' || c == '+') && Isdigit(tgetch(token, 1))))
		return(TKLIT);

	switch(c) {
		case '\0':	return(TKNUL);	/* no NULs!!! */
		case '"':	return(TKSTR);	/* quoted string literal */

		case '!':	return(TKDIR);	/* directive */
		case '@':	return(TKARG);	/* interactive argument */
		case '#':	return(TKBUF);	/* buffer argument */
		case '$':	return(TKENV);	/* environment variable */
		case '%':	return(TKVAR);	/* user variable */
		case '&':	return(TKFUN);	/* function */
		case '*':	return(TKLBL);	/* line label */
		case '[':	return(TKARRAY);/* array start */
		case '{':	return(TKHASH);	/* hash start */
		case ',':	return(TKCOMMA);/* separator - array/hash */

		default:	return(TKCMD);	/* (Emacs) command name */
	}
}

/* Get the value of user or environment variable vname to vvalue.
 *
 * Return TRUE   - worked, look in vvalue
 *	  FALSE  - memory error or something
 *	  FAILED - unknown variable
 */
int	getVarValue(const char* vname, TEXT* vvalue)
{
	tsetempty(vvalue);

	if(vname[0] == '$') {
		if( !envFind(vname + 1))
			return(FAILED);
		tcpys(vvalue, gtenv(vname + 1));
		return(TRUE);
	} else {
		BYTE*	uval = gtusr(vname + 1);
		if(uval == (BYTE*)NULL)
			return(FAILED);

		tcpys(vvalue, uval);
		return(TRUE);
	}
}

/* Parse a structured variable, return a pointer to VALue.
 *
 * Use this for things like:
 *	%array[ index ]
 *	%hash{ key }
 *
 * value	The variable so far
 *		This MUST be an array or hash, error if not
 * left		This is TRUE if the value is to be assigned to,
 *		ie a 'left hand' part of an assignment.
 *
 * Input will be taken from execText, execPos - where the next character
 * is either '[' or '{'.
 * THIS MUST NOT BE CALLED UNLESS THIS IS TRUE.
 *
 * Return:
 *	Pointer to a VAL to be used as a value or assigned to
 *		this will be an element inside an array or hash
 *	NULL - something has gone wrong
 */
VAL**	parseStructVar(VAL* value, int left)
{
	VAL**	valp = &value;		// OK as [ or { must be next

	while(1) {
		// Array index next ?
		if(tgetch(execText, execPos) == '[') {
			execPos++;	// Skip the [

			// Get an index to an array - a number

			if(vtype(*valp) != V_ARRAY) {
				mlwrite("%EFound '[' but not after array, found %s", vtypename(*valp));
				return(NULL);
			}

			// Get an INT - the index
			VAL*	iv = valGet(V_INT);

			if(nextarg("Array index: ", &iv, MLNOOPT, bmlValHistp) != TRUE) {
				valRelease(iv);
				return(NULL);
			}

			long	index = vint(iv);
			valRelease(iv);

			// No negative indicies
			// Index must be in bounds when not assigning
			if(index < 0 || ( !left && (*valp)->v_v.v_array.a_used <= index)) {
				mlwrite("%EArray index %d out of bounds, count %d", index, (*valp)->v_v.v_array.a_used);
				return(NULL);
			}

			// Grow the array on assignment - if needs be
			if(left && index >= (*valp)->v_v.v_array.a_used) {
				// This might mean extending the array:
				if(index >= (*valp)->v_v.v_array.a_alloc)
					valArraySize(*valp, (int)index);

				(*valp)->v_v.v_array.a_used = (int)(index + 1);
			}

			valp = &((*valp)->v_v.v_array.a_elems[index]);

			// Skip space, eat the ']'
			while(tgetch(execText, execPos) == ' ' || tgetch(execText, execPos) == '\t')
				execPos++;

			if(tgetch(execText, execPos) != ']')
				mlwrite("%EExpecting ']' after array index");

			execPos++;

			continue;
		}

		// Is a hash key next ?
		if(tgetch(execText, execPos) == '{') {
			execPos++;	// Skip the {

			if(vtype(*valp) != V_HASH) {
				mlwrite("%EFound '{' but not after hash, found %s", vtypename(*valp));
				return(NULL);
			}

			// Get a STR - the key
			VAL*	key = valGet(V_STR);

			if(nextarg("Hash key: ", &key, MLNOOPT, bmlValHistp) != TRUE) {
				valRelease(key);
				return(NULL);
			}

			valp = valHashFind(*valp, key, left);
			if(valp == NULL) {
				mlwrite("%EHash does not contain key '%v'", key);

				valRelease(key);
				return(NULL);
			}

			// Skip space, eat the '}'
			while(tgetch(execText, execPos) == ' ' || tgetch(execText, execPos) == '\t')
				execPos++;

			if(tgetch(execText, execPos) != '}')
				mlwrite("%EExpecting '}' after hash key");
			execPos++;

			valRelease(key);

			continue;
		}

		break;	// Not [ or {
	}

	return(valp);
}

/* Find the value of a tokn
 *
 * tokn	tokn to evaluate
 *
 * ** BEWARE ** This can be recursive so there is plenty of opportunity
 * to overwrite things that are in use!
 *
 * The value returned will often be "." in [Eval Stack] - not always
 * The 2 places where this recurses - advance & back "." to make this happen.
 * If "." is b_linep on entry: create a new line.
 *
 * Return a pointer to the next value (VAL).
 *	Could be any type.
 *	Could be to errorVAL - this should be taken as an error.
 *	The value is allocated and must be released by the caller but must be regarded read only.
 */
VAL*	getval(TEXT* tokn)
{
	register int status;	/* error return */
	register BUFFER* bp;	/* temp buffer pointer */
	LineLn		blen;	/* length of buffer argument */
	register int distmp;	/* temporary discmd flag */
	VAL*		value;	// Somewhere for recursive getval() result
	VAL*		key;	// Hash key
	LINE*		leval;	// Last line in eval stack
	VAL*		val;
	TEXT		tokn_1;	// *tokn but ignoring the first character

	/* tokn_1 is for where in a few places we want the text referenced by
	 * tokn - but not the first character. So we create a dummy text that
	 * represents that. Do it where we need it, up top here.
	 * The reason is that it works even if the literal contains NUL
	 */

	// If at end of Eval Stack buffer - allocate a new line:
	if(bmlEvalStack->b_dotp == bmlEvalStack->b_linep) {
		if((leval = BufAppendLine(bmlEvalStack)) == NULL ||
		    lsetNUL(leval) == FALSE) {
			mlwrite("%#%t", errorLine);
			return((VAL*)&errorVAL);
		}
		bmlEvalStack->b_dotp = leval;
	}

	switch(gettyp(tokn)) {
		case TKNUL:	/* End of string - should not happen */
				tsetempty(&ltxt(bmlEvalStack->b_dotp));
				return(playValUndef);

		case TKARG:	/* interactive argument */
				bmlEvalStack->b_dotp = lforw(bmlEvalStack->b_dotp);
				memcpy(&tokn_1, tokn, sizeof(TEXT));
				tokn_1.t_text++;
				tokn_1.t_used--;
				value = getval(&tokn_1);
				bmlEvalStack->b_dotp = lback(bmlEvalStack->b_dotp);
				distmp = discmd;	/* echo it always! */
				discmd = TRUE;

				status = mlprompt(ttext(vTEXT(value)), MLSTRING | MLINTERACTIVE, bmlValHistp);
				valRelease(value);
				discmd = distmp;
				if(status == ABORT) {
					mlwrite("%EUser interactive abort");
					return((VAL*)&errorVAL);
				}

				val = valGet(V_STR);
				tcpyl(vTEXT(val), bmlValHistp->b_dotp);
				return(val);

// tokn not returned
		case TKCHR:	/* Interative single char read */
				vint(playValChr) = ectoc(get1key());
				lcpyc(bmlEvalStack->b_dotp, (UniChar)vint(playValChr));
				return(playValChr);

		case TKBUF:	/* buffer contents fetch */
				/* grab the right buffer */
				bmlEvalStack->b_dotp = lforw(bmlEvalStack->b_dotp);
				memcpy(&tokn_1, tokn, sizeof(TEXT));
				tokn_1.t_text++;
				tokn_1.t_used--;
				value = getval(&tokn_1);
				bmlEvalStack->b_dotp = lback(bmlEvalStack->b_dotp);
				if((bp = bfindCheck(ttext(vTEXT(value)), BFNNONE)) == (BUFFER*)NULL) {
					valRelease(value);
					return((VAL*)&errorVAL);
				}

				Window*	wp = NULL;		// Window showing bp
				LINE**	dotpp	= &bp->b_dotp;	// Address of "." LINE & offset,
				LineLn*	dotop	= &bp->b_doto;	// can be in wp or bp

				/* If the buffer is displayed, get the window
				 * vars instead of the buffer vars.
				 * Choose the current window by preference, otherwise
				 * any window that displays it.
				 */
				if(bp->b_nwnd > 0) {
					wp = curwp->w_bufp == bp ? curwp : onscreen(bp);

					dotpp = &wp->w_dotp;
					dotop = &wp->w_doto;
				}

				/* make sure we are not at the end */
				if(*dotpp == bp->b_linep) {
					mlwrite("%EEnd of buffer '%v' when getting line", value);
					valRelease(value);
					return((VAL*)&errorVAL);
				}

				// Grab the line
				blen = (LineLn)(lused(*dotpp) - *dotop);
				if(blen > NSTRING)
					blen = NSTRING;
				lcpym(bmlEvalStack->b_dotp, lgetcaddr(*dotpp, *dotop), blen);

				// Step "." down one line
				*dotpp = lforw(*dotpp);
				*dotop = 0;

				/* If displayed buffer ensure "." is shown to move
				 * Set WFHARD as wp might not be curwp ... without the flag
				 * the position of "." might not be moved on the screen - nasty!
				 */
				if(wp)
					wp->w_flag |= WFMOVE | WFHARD;

				/* and return the spoils */
				val = valGet(V_STR);
				tcpyl(vTEXT(val), bmlEvalStack->b_dotp);
				return(val);

		case TKVAR:	/* user variable */
				// Do NOT release value - it is a u_value
				// Maybe incr reference & don't need to copy - speed up
				if((value = gtusrCk(tgetcaddr(tokn, 1))) == (VAL*)NULL)
					return((VAL*)&errorVAL);

				if(tgetch(execText, execPos) == '[' || tgetch(execText, execPos) == '{') {
					VAL**	varp = parseStructVar(value, FALSE);
					if(varp == NULL)	// Whoops!
						return((VAL*)&errorVAL);
					return(valNewRef(*varp));
				}

				return(valNewRef(value));

		case TKENV:	/* environment variable */
				lcpys(bmlEvalStack->b_dotp, gtenv(tgetcaddr(tokn, 1)));
				val = valGet(V_STR);
				tcpyl(vTEXT(val), bmlEvalStack->b_dotp);
				return(val);

		case TKFUN:	/* Function
				 * Evaluate the tokn string, get a VALue back.
				 */
				value = gtfun(tgetcaddr(tokn, 1));
				lcpyt(bmlEvalStack->b_dotp, vTEXT(valPrint(value)));

				val = valNewRef(value);
 				valRelease(value);	// This is our responsibility
				return(val);

		case TKCOMMA:	/* separator */
				mlwrite("%EUnexpected '%t'", tokn);
				return((VAL*)&errorVAL);

		case TKDIR:	/* directive */
				mlwrite("%EUnexpected '%t'", tokn);
				return((VAL*)&errorVAL);

		case TKLBL:	/* line label */
				mlwrite("%EUnexpected '%t'", tokn);
				return((VAL*)&errorVAL);

		case TKLIT:	/* numeric literal */
				lcpyt(bmlEvalStack->b_dotp, tokn);
				val = valGet(V_INT);

				// This is nasty - but it means that val is INT
				// and that the tokn
				splatSetText(ttext(tokn));
				splatSetLen(strlen(ttext(tokn)));
				valCopyValue(&val, &splatStrVAL);

				return(val);

		case TKSTR:	/* quoted string literal */
				memcpy(&tokn_1, tokn, sizeof(TEXT));
				tokn_1.t_text++;
				tokn_1.t_used--;
				lcpyt(bmlEvalStack->b_dotp, &tokn_1);

				val = valGet(V_STR);
				tcpyl(vTEXT(val), bmlEvalStack->b_dotp);
				return(val);

		case TKCMD:	/* Emacs command */
				lcpyt(bmlEvalStack->b_dotp, tokn);
				val = valGet(V_STR);
				tcpyl(vTEXT(val), bmlEvalStack->b_dotp);
				return(val);

		case TKARRAY:	// Array
				val = valGet(V_ARRAY);

				while(1) {
					while(tgetch(execText, execPos) == ' ' || tgetch(execText, execPos) == '\t')
						execPos++;

					if(execPos >= tused(execText))
						break;

					if(tgetch(execText, execPos) == ']') {
						execPos++;	// Eat the ']' - assumes nothing after it
						break;
					}
					value = valGet(V_ANY);

					// Get a value and push to the end of the array
					nextarg("Array element: ", &value, MLNOOPT, bmlValHistp);
					valArrayPush(val, value);
					valRelease(value);

					while(tgetch(execText, execPos) == ' ' || tgetch(execText, execPos) == '\t')
						execPos++;

					// Optional comma to separate elements
					if(tgetch(execText, execPos) == ',')
						execPos++;
				}

				return(val);

		case TKHASH:	// Hash
				val = valGet(V_HASH);
				key = (VAL*)NULL;
				value = (VAL*)NULL;

				while(1) {
					while(tgetch(execText, execPos) == ' ' || tgetch(execText, execPos) == '\t')
						execPos++;

					if(execPos >= tused(execText))
						break;
					if(tgetch(execText, execPos) == '}') {
						execPos++;	// Eat the '}'
						break;
					}

					if(tgetch(execText, execPos) == '=' && tgetch(execText, execPos + 1) == '>') {
						mlwrite("%EFound '=>' when expecting hash key");
						valRelease(val);
						return((VAL*)&errorVAL);
					}

					if(key == NULL)
						key = valGet(V_STR);

					// Get the key
					nextarg("Hash key: ", &key, MLSTRING, bmlValHistp);

					// Span white
					while(tgetch(execText, execPos) == ' ' || tgetch(execText, execPos) == '\t')
						execPos++;

					if(tgetch(execText, execPos) != '=' || tgetch(execText, execPos + 1) != '>') {
						mlwrite("%EDid not find '=>' after hash key near '%s'", tgetcaddr(execText, execPos));
						valRelease(val);
						valRelease(key);
						if(value)
							valRelease(value);
						return((VAL*)&errorVAL);
					}
					execPos += 2;	// Skip =>

					// Always get a new value - ANY, otherwise get may type problems
					value = valGet(V_ANY);

					nextarg("Hash value: ", &value, MLNOOPT, bmlValHistp);
					valHashAdd(val, key, value);

					// And release for next iteration
					valRelease(value);
					value = (VAL*)NULL;

					while(tgetch(execText, execPos) == ' ' || tgetch(execText, execPos) == '\t')
						execPos++;

					// Optional comma to separate elements
					if(tgetch(execText, execPos) == ',')
						execPos++;
				}

				if(key)
					valRelease(key);
				if(value)
					valRelease(value);

				return(val);

	}

	__NOT_REACHED
}

/* Convert a string to a numeric logical
 *
 * Case insensitive compare.
 *
 * TRUE or val[0]==T	-> TRUE
 * FALSE or val[0]==F	-> FALSE
 * numeric != 0		-> TRUE
 *
 * There is no check that val converts to a number.
 *
 * val		value to check for stol
 *
 * Return the numeric value.
 */
int	stol(const BYTE* val)
{
	BYTE	lowbuf[20];

	strncpy(lowbuf, val, sizeof(lowbuf) - 1);
	lowbuf[sizeof(lowbuf) - 1] = '\0';
	mklower(lowbuf);
	if(strcmp(lowbuf, StrOff) == 0)
		return(FALSE);
	if(lowbuf[0] == 'f')
		return(FALSE);
	if(strcmp(lowbuf, StrOn) == 0)
		return(TRUE);
	if(lowbuf[0] == 't')
		return(TRUE);

	/* check for numeric truth (!= 0) */
	return((atol(val) != 0));
}

/* Numeric logical to string logical
 *
 * val		value to translate
 *
 * Return "TRUE" or "FALSE"
 */
BYTE*	ltos(int val)
{
	if(val)
		return(truem);
	else
		return(falsem);
}

/* Numeric logical to yes/no
 *
 * val		value to translate
 *
 * Return "yes" or "no"
 */
BYTE*	ltoyn(int val)
{
	if(val)
		return("yes");
	else
		return("no");
}

/* Make a string upper case
 *
 * str		string to upper case
 *
 * Return str that will have been changed.
 */
BYTE*	mkupper(BYTE* str)
{
	BYTE* sp;

	sp = str;
	while(*sp) {
		if(Islower(*sp))
			*sp = (BYTE)toupper(*sp);
		++sp;
	}
	return(str);
}

/* Make a string lower case
 *
 * str		string to lower case
 *
 * Return str that will have been changed.
 */
BYTE*	mklower(BYTE* str)
{
	BYTE* sp;

	sp = str;
	while(*sp) {
		if(Isupper(*sp))
			*sp = (BYTE)tolower(*sp);
		++sp;
	}
	return(str);
}

#if	UNIX == 0
/* take the absolute value of an integer
 */
int	abs(int x)
{
	return(x < 0 ? -x : x);
}
#endif	/* UNIX == 0 */

/*	is_num: ascii string is integer......This is too
 *	inconsistant to use the system's
 */
int	is_num(const char* st)
{
	int	period_flag;	/* have we seen a period yet? */

	/* skip preceding whitespace */
	while (*st == ' ' || *st == '\t')
		++st;

	/* check for sign */
	if((*st == '-') || (*st == '+'))
		++st;

	/* scan digits */
	period_flag = FALSE;
	while (((*st >= '0') && (*st <= '9')) ||
	       (*st == '.' && period_flag == FALSE)) {
		if(*st == '.')
			period_flag = TRUE;
		st++;
	}

	/* scan rest of line for just white space */
	while (*st) {
		if((*st != '\t') && (*st != ' '))
			return(FALSE);

		st++;
	}

	return(TRUE);
}

#if	SYSRAND == 0
/* Returns a random integer.
 *
 * Current code takem from Microemacs 5.0
 *
 * This function implements the "minimal standard" RNG
 * from the paper "RNGs: Good Ones are Hard to Find" by Park and
 * Miller, CACM, Volume 31, Number 10, October 1988. */
 */
long	ernd(void)
{
	long int a=16807L, m=2147483647L, q=127773L, r=2836L;
	long lo, hi, test;

	hi = seed / q;
	lo = seed % q;
	test = a * lo - r * hi;
	seed = (test > 0) ? test : test + m;

#if	0	// Previous version
	seed = abs(seed * 1721 + 10007);
#endif
	return(seed);
}
#endif

/* find pattern within source - case insensitive if buffer mode EXACT is not set.
 * Space only matches space, never tab (or other Unicode spaces).
 *
 * source	source string to search
 * pattern	string to look for
 *
 * BEWARE: This takes the notion of UTF8 flag from curbp.
 *
 * Return position, start is 1; return 0 for not found
 */
int	sindex(const TEXT* source, const TEXT* pattern)
{
	const BYTE*	sp;	/* ptr to current position to scan */
	const BYTE*	csp;	/* ptr to source string during comparison */
	const BYTE*	cp;	/* ptr to place to check for equality */
	const BYTE*	srcend;	/* Byte just after end of source */
	const BYTE*	patend;	/* Byte just after end of pattern */
	BYTE		lowerp;	/* Lowercased character in pattern */
	int		u8;	/* Work in UTF-8 mode ? */

	eqCompare_i = (curwp->w_bufp->b_mode & MDEXACT) == 0;
	spaceMatchWhite = FALSE;

	u8 = curbp->b_mode & MDUTF8;

	/* scanning through the source string */
	sp = ttext(source);
	srcend = tgetcaddr(source, tused(source));
	patend = tgetcaddr(pattern, tused(pattern));

	while(sp < srcend) {
		/* scan through the pattern */
		cp = ttext(pattern);
		csp = sp;
		while(cp < patend) {
			// Since eq() does not downcase the pattern we must do it
			const BYTE*	patp = csp;
			if(eqCompare_i && Isupper(*csp)) {
				lowerp = (BYTE)tolower(*csp);
				patp = &lowerp;
			}

			if(!eq(cp, patp))
				break;

			if(u8) {
				cp  += CHAR_LEN(*cp);
				csp += CHAR_LEN(*csp);
			} else {
				++cp;
				++csp;
			}
		}

		/* Was it a match?
		 * Return a character offset, not a byte offset.
		 */
		if(cp == patend)
			return((u8 ? (int)utf8_chrnlen(ttext(source), (size_t)(sp - ttext(source))) :
				     (int)(sp - ttext(source))) + 1);

		++sp;
	}

	/* no match at all.. */
	return(0);
}

/* Filter a string through a translation table
 *
 * result	where to put the result
 * source	string to filter
 * lookup	characters to translate
 * trans	resulting translated characters
 *
 * Differs from standard Microemacs, in that if trans is shorter than
 * lookup, the 'missing' characters will be removed from source; if where
 * there is a match in lookup and no corresponding replacement.
 *
 * Extensively reworked so that it works with UTF8
 *
 * Return:
 *	TRUE  if it all works
 *	FALSE if something is wrong
 */
int	xlat(TEXT* result, const TEXT* source, const TEXT* lookup, const TEXT* trans)
{
	const BYTE*	sp;	/* pointer into source table */

	/* Special check - not checked above */
	if( !utf8_chkbuf(ttext(lookup), (size_t)tused(lookup)) ||
	    !utf8_chkbuf(ttext(trans), (size_t)tused(trans))) {
		mlwrite("%EBadly UTF8 encoded argument 2 or 3 to function &xlat");
		return(FALSE);
	}

	tsetempty(result);

	/* scan source string */
	sp = ttext(source);
	while(sp < ttext(source) + tused(source)) {
		int	len = CHAR_LEN(*sp);	// Length of the current char

		LineLn	lookChar = 0;		// Which char in lookup is being inspected

		const BYTE*	lp;		// Pointer into lookup table

		/* scan lookup table for a match */
		lp = ttext(lookup);
		while(lp < ttext(lookup) + tused(lookup)) {
			/* Compare the character.
			 * Optimise for single byte ASCII characters
			 */
			if(len == 1 ? *sp == *lp : !memcmp(sp, lp, (size_t)len)) {
				// Not necessarily replacing characters encoded in
				// the same number of bytes
				size_t lc = (size_t)lookChar;
				size_t boff = utf8_maxlen(ttext(trans), (size_t)tused(trans), &lc);

				// Is there a character to replace with ?
				// If not, doing nothing here deletes it from the copy
				// KV: if the last char was spanned, must not translate to NUL
				if((LineLn)lc == lookChar && boff < (size_t)tused(trans)) {
					const BYTE*	tc = tgetcaddr(trans, boff);
					tcatm(result, tc, CHAR_LEN(*tc));
				}
				goto xnext;
			}

			lp += CHAR_LEN(*lp);
			lookChar++;
		}

		/* no match, copy in the source char untranslated */
		tcatm(result, sp, len);

xnext:		sp += len;
	}

	return(TRUE);
}

/* THIS IS A MASSIVELY TACKY BIT OF CODE. What it does is, if "value" is
 * true (non zero), it looks through all the buffers and sets CMODE if
 * the corresponding filenames end in a .c or .h suffix.
 *
 * This is necessary because in main(), the startup() routine is invoked
 * after the first file has been read in. Thus if the emacsrc file specifies
 * that CMODE should be automatically set, it wont be for the first file.
 *
 *						afb@phcomp.co.uk
 *						29 July 1988
 * See also readin().
 */
void	setcmode(int value)
{
	BUFFER*	bp;
	
	if(!value)		/* nothing to do if turning auto-c off */
		return;

	for(bp = bheadp; bp != (BUFFER*) NULL; bp = bp->b_bufp)
		CheckCMode(bp);
}

/* Show what environment variables are set to by putting current values
 * onto the end of buffer.
 * The buffer is UTF8 encoded.
 *
 * Return:
 *	TRUE  if all is OK
 *	FALSE out of memory
 */
int	WriteEnvVariablesToBuffer(BUFFER* buffer)
{
	ENVVAR*	ep;
	int	longestValue;	// Value
	int	longestDesc;	// Description
	int	i;
	BYTE*	value;
	BYTE	line[300];

	addline("", buffer, 1);
	addline("Environment Variables", buffer, 1);
	addline("---------------------", buffer, 1);
	addline("", buffer, 1);

	// Find length of longest value
	// Note: it could be UTF8
	for(ep = envars, longestValue = longestDesc = 0; ep->ev_name; ep++) {
		value = gtenv(ep->ev_name);

		i = (int)strlen(value);

		// If the buffer is UTF8 & the value is valid UTF8 do a UTF8 string length
		if((buffer->b_mode & MDUTF8) && utf8_check(value))
			i = (int)utf8_chrnlen(value, (size_t)i);

		if(i > longestValue)
			longestValue = i;
		if((i = (int)strlen(ep->ev_desc)) > longestDesc)
			longestDesc = i;
	}

	// Cap longest displayed value
	if(longestValue > WRITE_VARIABLE_LONGEST_VALUE) {
		sprintf(line, "Longest variable value length is %d, truncated to %d", longestValue, WRITE_VARIABLE_LONGEST_VALUE);
		if(addline(line, buffer, 1) == FALSE)
			return(FALSE);
		longestValue = WRITE_VARIABLE_LONGEST_VALUE;
	}
	else {
		// Add an empty line else the # lines can change which makes what is displayed jump
		addline("", buffer, 1);
	}

	// Use the play line as insStrLine() makes it longer as needed and cleans to UTF8
	if( !textMoreBytes(playText, longestDesc + longestValue * UTF8_MAX_BYTES + 10))
		return(FALSE);

	for(ep = envars; ep->ev_name; ep++) {
		// This is a bit of a fiddle since we are handling bytes, but
		// the value might be a UTF8 string - so the lengths are confusing.

		size_t	chars = (size_t)longestValue;

		/* Where write description - make descriptions line up
		 * in spite of values being a different number of bytes but
		 * the same width in characters
		 */
		int	j;

		BYTE*	ev = gtenv(ep->ev_name);
		tsetused(playText, 0);

		printToText(playText, "$%s", ep->ev_name);

		// Pad to just more than longest name:
		while(tused(playText) < NVSIZE + 2)
			tcatc(playText, ' ');

		i = (int)utf8_maxlen(ev, strlen(ev), &chars);
		j = i + longestValue - (int)chars;

		tcatm(playText, ev, i);

		// Pad to length of longest name + longest var (that we show) + some padding:
		while(tused(playText) < NVSIZE + 4 + j + 4)
			tcatc(playText, ' ');

		// Copy in description:
		tcats(playText, ep->ev_desc);

		if(addLINE(playText, buffer, 0) == FALSE)
			return(FALSE);
	}

	return(TRUE);
}

/* Show what user variables are set to by putting current values
 * onto the end of buffer.
 *
 * Return:
 *	TRUE  if all is OK
 *	FALSE out of memory
 */
int	WriteUserVariablesToBuffer(BUFFER* buffer)
{
	int	vnum;
	UVAR*	uvp;

	addline("", buffer, 1);
	addline("Global User Variables", buffer, 1);
	addline("---------------------", buffer, 1);
	addline("", buffer, 1);

	// Use the play line as insStrLine() makes it longer as needed and cleans to UTF8
	if( !textMoreBytes(playText, NVSIZE + 10))
		return(FALSE);

	for(vnum = 0, uvp = uv_global->ut_uv; vnum < uv_global->ut_size; vnum++, uvp++) {
		VAL*	v;
		long	l;
		char	elems[30];

		if(uvp->u_name[0] == '\0')
			return(TRUE);

		ClearOut(ttext(playText), NVSIZE + LONGEST_TYPE_NAME + 10);

		tputc(playText, 0, '%');
		strinsstr(tgetcaddr(playText, 1), uvp->u_name);
		strinsstr(tgetcaddr(playText, NVSIZE + 2), vtypename(uvp->u_value));

		v = valPrint(uvp->u_value);	// Might not be a string

		// Array/Hash show how many elements:
		l = -1;
		if(vtype(uvp->u_value) == V_ARRAY)
			l = uvp->u_value->v_v.v_array.a_used;
		if(vtype(uvp->u_value) == V_HASH)
			l = uvp->u_value->v_v.v_hash.h_used;
		if(l != -1) {
			sprintf(elems, "%ld elements", l);
			tcpys(vTEXT(v), elems);
		}

		if(insStrText(playText, NVSIZE + LONGEST_TYPE_NAME + 4, vtext(v), tused(vTEXT(v)), 0) < 0)
			return(FALSE);
		tsetused(playText, NVSIZE + LONGEST_TYPE_NAME + 4 + tused(vTEXT(v)));	// OK - size set above

		if(addLINE(playText, buffer, 0) == FALSE)
			return(FALSE);
	}

	return(TRUE);
}

/* end */
