/*
 * ADDW SCCS: @(#)evar.h	1.105 10/14/20 22:25:59
 */

/*	EVAR.H: Environment and user variable definitions
		for MicroEMACS

		written 1986 by Daniel Lawrence
*/

/*	current user variables (This structure will probably change)	*/

#define MAXVARS 	100

UVAR	uv[MAXVARS];	/* user variables */


/* Environment variables.
 * These can be set in the startup file like: set $fillcol 123
 *
 * The value should be in here ... but is set/obtained via huge switch statements.
 */
#define	EV_INT		0x0000
#define	EV_BOOL		0x0001	// Implemented as int, but values 0, 1
#define	EV_STRING	0x0002	// Compile time char*
#define	EV_LONG		0x0003
#define	EV_TEXT		0x0004
#define	EV_CHAR		0x0005	// ASCII char, read in as string length 1
#define	EV_TYPEMASK	0x000F

#define	EV_READONLY	0x0010	// Cannot assign
#define	EV_MIN		0x0020	// Inspect ev_min
#define	EV_MAX		0x0040	// Inspect ev_max
#define	EV_ONOFF	0x0100	// Display bool as on/off - else TRUE/FALSE
#define	EV_HEXINT	0x0200	// Display int as hex
#define	EV_UTF8		0x0400	// EV_STRING or EV_TEXT must be valid UTF-8
				// This also means that EV_MIN & EV_MAX are UTF-8 characters, not bytes
// Convenience types. The EV_UTF8 is part of the definition
#define	EV_USTRING	(EV_STRING | EV_UTF8)
#define	EV_UTEXT	(EV_TEXT   | EV_UTF8)

/* When WriteEnvVariablesToBuffer() & WriteUserVariablesToBuffer() write values,
 * this is the longest value that will be displayed - before truncation.
 */
#define	WRITE_VARIABLE_LONGEST_VALUE	40	// Characters, not bytes

typedef	struct	ENVVAR	{
	char*	ev_name;		// Variable name
	char*	ev_desc;		// Description
	union	{
		char**	s;		// EV_STRING value is pointer to:
					// - if READONLY can be to compile time value
					// - else a malloced value
		int*	i;		// EV_INT EV_BOOL
		long*	l;		// EV_LONG
		TEXT**	t;		// EV_TEXT
	} v;
	int	ev_type;		// What sort of value it is. EV_xxx above
	long	ev_min;			// Minimum value
	long	ev_max;			// Maximum value, for strings the max length
} ENVVAR;

/* The environment variables.
 * Use enum as this gives automatic numbering - as each enum is really a number.
 * This makes it easier to add/remove variables, in particular reorderd
 * to put vaguely related things together.
 *
 * The names are all in caps - this reflects that they were all CPP #defines at one point.
 */

typedef	enum {

	EVFILLCOL,			// current fill column
	EVAUTO_WIDTH,			// Reset fillcol when window size changes
	EVFRENCH_SPACING,		// One space (not two) at sentence end on fill-paragraph

	EVLASTKEY,			// last keyboard char struck
	EVEOF,				// end of file
	EVCURCHAR,			// current character under the cursor
	EVLWIDTH,			// width of current line
	EVLINE,				// text of current line

	EVPAGELEN,			// number of lines used by editor
	EVCURCOL,			// current column position of cursor: "."
	EVCURCHARPOS,			// current character position of cursor: "."
	EVTARGET,			// target for line moves
	EVCURLINE,			// curent line in file
	EVBUFLINES,			// Number of lines in buffer
	EVMARKCOL,			// current column pos of mark
	EVMARKLINE,			// current line in file of mark
	EVCURWIND,			// current window number
	EVNUMWIND,			// number of windows displayed
	EVFCOL,				// first displayed column number, window left shift
	EVPREFCOLHORIZ,			// Preferably split window horizontally if wider

	EVWINDOWMARGIN,			// try to keep cursor away from window edges
	EVCONTEXT_LINES,		// number of context lines when page up/down
	EVOVERLAP,			// number of context lines when page up/down
	EVJUMP,				// number of lines to scroll at top/bottom of window
	EVJUMPNOSCROLL,			// $jump when a window cannot scroll
	EVWLINE,			// # of lines in current window

	EVCWLINE,			// current screen line in window
	EVSCROLL,			// True if the terminal can scroll up/down
	EVSCROLLLR,			// True if the terminal can scroll left/right
	EVTERM,				// Terminal type name
	EVSRES,				// current screen resolution
	EVCURWIDTH,			// current screen width
	EVPALETTE,			// current palette string
	EVTERMOPENED,			// Value of TTopened - true if terminal initialised

	EVCBUFNAME,			// current buffer name
	EVCFNAME,			// current file name - not necessarily good UTF-8
	EVLASTFILESLINK,		// last opened file was symlink
	EVLASTFILELINKS,		// number of links on the last file opened

	EVRAM,				// ram in use by malloc
	EVFLICKER,			// flicker supression
	EVDEBUG ,			// macro debugging
	EVDISCMD,			// display commands on command line

	EVSTATUS,			// returns the status of the last command
	EVRVAL,				// child process return value
	EVCMDFLAG,			// Flag to current command/macro
	EVCMDCOUNT,			// Count to current command/macro
	EVFILTERCANINTR,		// Buffer filter commands can be sent intr/quit

	EVWRITE,			// write buffers at various times
	EVASAVE,			// # of chars between auto-saves
	EVACOUNT,			// # of chars until next auto-save

	EVVERSION,			// current version number
	EVPROGNAME,			// returns current prog name
	EVBUILD_DATE,			// When Emacs was compiled
	EVOS,				// name of operating system

	EVPID,				// Emacs process ID
	EVSECONDS,			// Time since the epoch, seconds
	EVTIME,				// current date and time

	EVSEED,				// current random number seed
	EVDISINP,			// display command line input characters

	EVSEARCH,			// search pattern
	EVREPLACE,			// replacement pattern
	EVREPLACESAME,			// Prompt to query replacement even when replacement string is same as matched string
	EVSPACEWHITE,			// a space matches white space
	EVMATCH,			// last matched magic pattern

	EVKILL,				// kill buffer (read only)

	EVCMODE,			// mode of current buffer
	EVGMODE,			// global modes

	EVTPAUSE,			// length to pause for paren matching
	EVPREFDELAY,			// prefix delay interval
	EVPENDING,			// type ahead pending flag
	EVTTYTIMEOUT,			// Timeout for escape sequence characters & multi byte chars - microseconds
	EVMAXQUESTION,			// How many times to ask a question before giving up
	EVRENEW,			// Seconds: checks/autoread of RENEW mode buffers
	EVRENEWBELL,			// Ring the bell on buffer RENEW

	EVGFLAGS,			// global internal Emacs flags

	EVRELINK,			// dont ever, ever, ever unlink files
	EVSYNC,				// sync machine after writes

	EVTABWIDTH,			// width of displayed tabs
	EVTAB,				// width of displayed tabs - synonym EVTABWIDTH

	EVTERSE,			// be terse on output of modes on modeline
	EVCCOUNT,			// key count to clear message line
	EVEARLYLOG,			// Log to terminal before initialisation

	EVAUTOC,			// Check suffixes for C & PHP modes
	EVSUFFIX_C,			// Suffixes that enable C mode
	EVSUFFIX_PHP,			// Suffixes that enable PHP mode
	EVAUTOEC,			// extended CMODE stuff
	EVDOUBLESTAR,			// C Comments are of the double ** style
	EVAUTODOS,			// Automatic detection of DOS format input

	EVNOFLASH,			// Don't use visible bell (if available)
	EVNOBELL,			// Don't ring the bell, keep quiet

	EVFLOWCONTROL,			// Enable xon/xoff flow control with the terminal
	EVCATCHSUSPEND,			// Read/catch the ^z, don't let UNIX do it

	EVBGCOLOUR,			// gbcolor - global background colour
	EVFGCOLOUR,			// gfcolor - global foreground colour
	EVMETA_ATTRIB,			// display meta & control chars with underline & standout modes
	EVCOLOURISE_TAB,		// show tabs by colourising them

	EVMESSAGEHIST,			// number of lines in [Message History], etc
	EVLASTMESG,			// last message written to command line

	EVTF8_REPLACE,			// utf8-replacement for bad characters on setting UTF-8 mode
	EVTF8_ERRNO,			// error code on failing to set UTF-8 mode
	EVTF8_ERROR,			// $utf8-errno error as a string
	EVTF8_ERRLEN,			// number of bytes in error on failing to set UTF-8 mode

	EVHOME,				// User's home directory
	EVPWD,				// Current working directory
	EVLANGUAGE,			// User's language
	EVTERRITORY,			// Regional variant of $language 

	EVEMACSSAVE,			// .emacssave suffix files are prob crash save

	EVNULL				// Dummy entry at the end of table

} envNums;

/* List of recognized environment variables.
 * NULL terminated list.
 *
 * Static else clang winges that is it not previously declared extern!
 *
 * Beware: if KeyCode is not int - may get prob, unlikely, see estruct.h
 *
 * Remember that the max size (ie length) of a var name is NVSIZE.
 */
static	ENVVAR	envars[] = {
    [EVFILLCOL] =	{ "fillcol",		"Current fill column",				.v.i = &fillcol,	EV_INT,				0,			0 },
    [EVPREFCOLHORIZ] =	{ "pref-col-horiz",	"Preferably split window horizontally if wider", .v.i = &PrefSplitCols,	EV_INT | EV_MIN,		10,			0 },
    [EVPAGELEN] =	{ "pagelen",		"Number of lines used by editor",		.v.i = NULL,		EV_INT | EV_MIN,		MIN_SCREEN_LINES,	0 },
    [EVCURCOL] =	{ "curcol",		"Current column position of cursor",		.v.i = NULL,		EV_INT | EV_MIN,		0,			0 },
    [EVCURCHARPOS] =	{ "curcharpos",		"Current character position of cursor",		.v.i = NULL,		EV_INT | EV_MIN,		0,			0 },
    [EVCURLINE] =	{ "curline",		"Current line number in buffer",		.v.i = NULL,		EV_INT | EV_MIN,		0,			0 },
    [EVBUFLINES] =	{ "buffer-lines",	"Number of lines in buffer",			.v.i = NULL,		EV_INT | EV_READONLY,		0,			0 },
    [EVRAM] =		{ "ram",		"Ram in use by malloc",				.v.i = NULL,		EV_INT | EV_READONLY,		0,			0 },
    [EVFLICKER] =	{ "flicker",		"Flicker supression",				.v.i = &flickcode,	EV_BOOL,			0,			0 },
    [EVCURWIDTH] =	{ "curwidth",		"Current screen width",				.v.i = NULL,		EV_INT | EV_MIN,		MIN_SCREEN_COLS,	0 },
    [EVCBUFNAME] =	{ "cbufname",		"Current buffer name",				.v.s = NULL,		EV_USTRING | EV_MIN | EV_MAX,	1,			NBUFN },
    [EVCFNAME] =	{ "cfname",		"Current file name",				.v.s = NULL,		EV_STRING,			0,			0 },
    [EVLASTFILESLINK] =	{ "file-symlink",	"Last opened file was symbolic link",		.v.i = &isSlink,	EV_BOOL | EV_READONLY,		0,			0 },
    [EVLASTFILELINKS] =	{ "file-numlinks",	"Number of links on the last file opened",	.v.i = &numLinks,	EV_INT | EV_READONLY,		0,			0 },
    [EVSRES] =		{ "sres", 		"Current screen resolution",			.v.s = NULL,		EV_STRING | EV_MAX,		0,			(sizeof(sres) - 1) },
    [EVDEBUG ] =	{ "debug",		"Macro debugging",				.v.i = &macbug,		EV_BOOL,			0,			0 },
    [EVSTATUS] =	{ "status",		"Status of the last command",			.v.i = &cmdstatus,	EV_INT,				0,			0 },
    [EVPALETTE] =	{ "palette",		"Current palette string",			.v.s = NULL,		EV_STRING | EV_MAX,		0,			sizeof(palstr) },
    [EVASAVE] =		{ "asave",		"# of chars between auto-saves",		.v.i = &gasave,		EV_INT | EV_MIN,		1,			0 },
    [EVACOUNT] =	{ "acount",		"# of chars until next auto-save",		.v.i = &gacount,	EV_INT | EV_MIN,		1,			0 },
    [EVLASTKEY] =	{ "lastkey",		"Last keyboard char struck",			.v.i = &lastkey,	EV_INT | EV_MIN | EV_MAX,	'\0',			UTF8_MAX_CODE_POINT },
    [EVEOF] =		{ "eof",		"End of file",					.v.i = &termEOF,	EV_BOOL | EV_READONLY,		0,			0 },
    [EVCURCHAR] =	{ "curchar",		"Current character under the cursor",		.v.i = NULL,		EV_INT | EV_MIN | EV_MAX,	'\0',			UTF8_MAX_CODE_POINT },
    [EVDISCMD] =	{ "discmd",		"Display commands on command line",		.v.i = &discmd,		EV_BOOL,			0,			0 },
    [EVVERSION] =	{ "version",		"Current version number",			.v.s = &Version,	EV_STRING | EV_READONLY,	0,			0 },
    [EVPROGNAME] =	{ "progname",		"Current prog name - as invoked",		.v.s = (char**)&progname, EV_STRING | EV_READONLY,	0,			0 },
    [EVSEED] =		{ "seed",		"Current random number seed",			.v.i = NULL,		EV_INT,				0,			0 },
    [EVDISINP] =	{ "disinp",		"Display command line input characters",	.v.i = &disinp,		EV_BOOL,			0,			0 },
    [EVWLINE] =		{ "wline",		"# of lines in current window",			.v.i = NULL,		EV_INT | EV_MIN,		MIN_WINDOW_LINES,	0 },
    [EVCWLINE] =	{ "cwline",		"Current screen line in window",		.v.i = NULL,		EV_INT | EV_MIN,		0,			0 },
    [EVTARGET] =	{ "target",		"Target for line moves",			.v.i = NULL,		EV_INT | EV_MIN,		0,			0 },
    [EVSEARCH] =	{ "search",		"Search pattern",				.v.t = NULL,		EV_TEXT | EV_MAX,		0,			NPAT },
    [EVREPLACE] =	{ "replace",		"Replacement pattern",				.v.t = &searchRep,	EV_TEXT | EV_MAX,		0,			NPAT },
    [EVREPLACESAME] =	{ "search-replace-same","Ask even when replacement is same as matched",	.v.i = &AskReplaceSame,	EV_BOOL,			0,			0 },
    [EVSPACEWHITE] =	{ "space-match-white",	"Space matches white space",			.v.i = &spaceMatchWhite,EV_BOOL,			0,			0 },
    [EVMATCH] =		{ "match",		"Last matched magic pattern",			.v.t = &patmatch,	EV_STRING | EV_READONLY,	0,			0 },
    [EVKILL] =		{ "kill",		"Kill buffer (read only)",			.v.s = NULL,		EV_STRING | EV_READONLY,	0,			0 },
    [EVCMODE] =		{ "cmode",		"Mode of current buffer",			.v.i = NULL,		EV_INT | EV_HEXINT,		0,			0 },
    [EVGMODE] =		{ "gmode",		"Global modes",					.v.i = &gmode,		EV_INT | EV_HEXINT,		0,			0 },
    [EVTPAUSE] =	{ "tpause",		"Seconds to pause for paren matching",		.v.i = &tpause,		EV_INT | EV_MIN,		0,			0 },	/* tpause was t_pause in term struct */
    [EVPENDING] =	{ "pending",		"Type ahead pending flag",			.v.i = NULL,		EV_BOOL | EV_READONLY,		0,			0 },
    [EVLWIDTH] =	{ "lwidth",		"Width of current line",			.v.i = NULL,		EV_INT | EV_READONLY,		0,			0 },
    [EVLINE] =		{ "line",		"Text of current line",				.v.s = NULL,		EV_STRING | EV_READONLY,	0,			0 },
    [EVGFLAGS] =	{ "gflags",		"Global internal Emacs flags",			.v.i = &gflags,		EV_INT | EV_HEXINT,		0,			0 },
    [EVRVAL] =		{ "rval",		"Child process exit status (return value)",	.v.i = &rval,		EV_INT | EV_READONLY,		0,			0 },
    [EVRELINK] =	{ "relink",		"Don't ever, ever, ever unlink files",		.v.i = &relinking,	EV_BOOL,			0,			0 },
    [EVPREFDELAY] =	{ "prefdlay",		"Seconds before command prefix (eg ^X) is shown", .v.i = (int*)&ddelay,	EV_INT | EV_MIN,		0,			0 },
    [EVTABWIDTH] =	{ "tabwidth",		"Width of displayed tabs",			.v.i = NULL,		EV_INT | EV_MIN,		TAB_MINIMUM,		TAB_MAXIMUM },
    [EVTAB] =		{ "tab",		"Width of displayed tabs",			.v.i = NULL,		EV_INT | EV_MIN,		TAB_MINIMUM,		TAB_MAXIMUM },
    [EVTERSE] =		{ "terse",		"Be terse on output of modes on modeline",	.v.i = NULL,		EV_BOOL,			0,			0 },
    [EVEARLYLOG] =	{ "early-log",		"Log to terminal before initialisation",	.v.i = &EarlyLog,	EV_BOOL,			0,			0 },
    [EVCCOUNT] =	{ "ccount",		"Key count to clear message line",		.v.i = &ccount,		EV_INT | EV_MIN,		0,			0 },
    [EVAUTOC] =		{ "auto-c",		"Check suffixes for C & PHP modes",		.v.i = NULL,		EV_BOOL | EV_ONOFF,		0,			0 },
    [EVAUTOEC] =	{ "auto-ec",		"Extended CMODE stuff",				.v.i = NULL,		EV_BOOL | EV_ONOFF,		0,			0 },
    [EVSUFFIX_C] =	{ "suffixes-c",		"Suffixes that enable C mode",			.v.s = &suffixesC,	EV_STRING,			0,			0 },
    [EVSUFFIX_PHP] = 	{ "suffixes-php",	"Suffixes that enable PHP mode",		.v.s = &suffixesPHP,	EV_STRING,			0,			0 },
    [EVSYNC] =		{ "sync",		"Flush file to disk after write",		.v.i = &syncmc,		EV_BOOL | EV_ONOFF,		0,			0 },
    [EVWRITE] =		{ "write",		"Write buffers at various times",		.v.i = &wbuffs,		EV_BOOL | EV_ONOFF,		0,			0 },
    [EVTTYTIMEOUT] =	{ "ttytimeout",		"Timeout for escape sequence characters & multi byte chars - microseconds", .v.l = &ttytimeout, EV_LONG | EV_MIN, 0,		0 },
    [EVNOFLASH] =	{ "noflash",		"Don't use visible bell (if available)",	.v.i = &noflash,	EV_BOOL | EV_ONOFF,		0,			0 },
    [EVNOBELL] =	{ "nobell",		"Don't ring the bell",				.v.i = &nobell,		EV_BOOL | EV_ONOFF,		0,			0 },
    [EVDOUBLESTAR] =	{ "doublestar",		"C Comments are of the double ** style",	.v.i = &doublestar,	EV_BOOL,			0,			0 },
    [EVCATCHSUSPEND] =	{ "catchsuspend",	"Read/catch the ^z, don't let UNIX do it",	.v.i = &catchsuspend,	EV_BOOL,			0,			0 },
    [EVAUTO_WIDTH] =	{ "auto-width",		"Reset $fillcol when window size changes",	.v.i = &auto_width,	EV_BOOL,			0,			0 },
    [EVBGCOLOUR] =	{ "bg-colour",		"Global background colour",			.v.i = &gbcolor,	EV_INT | EV_MIN | EV_MAX,	COLOUR_MIN,		COLOUR_MAX },
    [EVFGCOLOUR] =	{ "fg-colour",		"Global foreground colour",			.v.i = &gfcolor,	EV_INT | EV_MIN | EV_MAX,	COLOUR_MIN,		COLOUR_MAX },
    [EVMETA_ATTRIB] =	{ "meta-attrib",	"Display meta & control chars with underline & standout modes", .v.i = NULL, EV_BOOL,			0,			0 },
    [EVCOLOURISE_TAB] =	{ "colourise-tab",	"Show tabs by colourising them",		.v.i = NULL,		EV_BOOL,			0,			0 },
    [EVWINDOWMARGIN] =	{ "window-move-margin",	"Try to keep cursor away from window edges",	.v.i = &WindowMargin,	EV_BOOL,			0,			0 },
    [EVMESSAGEHIST] =	{ "message-history",	"Number of lines in [Message History], etc",	.v.i = &SaveMlHist,	EV_INT | EV_MIN,		0,			0 },
    [EVCURWIND] =	{ "curwind",		"Current window number",			.v.i = NULL,		EV_INT | EV_READONLY,		0,			0 },
    [EVNUMWIND] =	{ "numwind",		"Number of windows displayed",			.v.i = &NumberOfWindows, EV_INT | EV_READONLY,		0,			0 },
    [EVFCOL] =		{ "fcol",		"First displayed column number, window left shift", .v.i = NULL,	EV_INT | EV_READONLY,		0,			0 },
    [EVLASTMESG] =	{ "lastmesg",		"Last message written to command line",		.v.s = NULL,		EV_STRING | EV_READONLY,	0,			0 },
    [EVOS] =		{ "os",			"Name of operating system",			.v.s = &os,		EV_STRING | EV_READONLY,	0,			0 },
    [EVTIME] =		{ "time",		"Current date and time",			.v.s = NULL,		EV_STRING | EV_READONLY,	0,			0 },
    [EVMARKCOL] =	{ "markcol",		"Current column position of mark",		.v.i = NULL,		EV_INT | EV_MIN,		0,			0 },
    [EVMARKLINE] =	{ "markline",		"Current line in mark",				.v.i = NULL,		EV_INT | EV_MIN,		0,			0 },
    [EVTF8_REPLACE] =	{ "utf8-replacement",	"For bad characters on setting UTF-8 mode",	.v.s = (char**)&utf8_modeReplace, EV_USTRING | EV_MIN,	1,			0 },
    [EVTF8_ERRNO] =	{ "utf8-errno",		"Error code on failing to set UTF-8 mode",	.v.i = &utf8_errno,	EV_INT | EV_READONLY,		0,			0 },
    [EVTF8_ERROR] =	{ "utf8-error",		"Text of $utf8-errno",				.v.s = NULL,		EV_STRING | EV_READONLY,	0,			0 },
    [EVTF8_ERRLEN] =	{ "utf8-errlen",	"Number of bytes in error on failing to set UTF-8 mode", .v.i = &utf8_errLen, EV_LONG | EV_READONLY,	0,			0 },
    [EVCONTEXT_LINES] =	{ "page-context",	"Number of context lines when page up/down",	.v.i = &PageContextLines, EV_INT | EV_MIN,		0,			0 },
    [EVOVERLAP] =	{ "overlap",		"Number of context lines when page up/down",	.v.i = &PageContextLines, EV_INT | EV_MIN,		0,			0 },
    [EVJUMP] =		{ "jump",		"Lines to scroll at top/bottom of window",	.v.i = &ScrollJumpBy,	EV_INT,				0,			0 },
    [EVJUMPNOSCROLL] =	{ "jump-no-scroll",	"$jump when window cannot scroll",		.v.i = &NoScrollJumpBy,	EV_INT,				0,			0 },
    [EVSCROLL] = 	{ "scroll",		"Will the terminal scroll up/down",		.v.i = NULL,		EV_BOOL,			0,			0 },
    [EVSCROLLLR] = 	{ "scroll-lr",		"Will the terminal scroll left/right",		.v.i = NULL,		EV_BOOL,			0,			0 },
    [EVTERM] =		{ "TERM",		"Terminal type name",				.v.s = &TermName,	EV_STRING | EV_MIN,		1,			0 },
    [EVTERMOPENED] =	{ "term-opened",	"Terminal has been opened, not initialising",	.v.i = &TTopened,	EV_INT | EV_READONLY,		0,			0 },
    [EVHOME] =		{ "HOME",		"Home directory",				.v.s = &home,		EV_STRING | EV_READONLY,	0,			0 },
    [EVPWD] =		{ "pwd",		"Current working directory",			.v.t = &cwDir,		EV_TEXT | EV_READONLY,		0,			0 },
    [EVAUTODOS] =	{ "autodos",		"On file read check for DOS format, set DOS mode", .v.i = &AutoDos,	EV_BOOL | EV_ONOFF,		0,			0 },
    [EVFILTERCANINTR] =	{ "filter-can-intr",	"Buffer filter commands can be sent intr/quit",	.v.i = &FilterCanIntr,	EV_BOOL,			0,			0 },
    [EVBUILD_DATE] =	{ "build-date",		"When Emacs was compiled",			.v.s = &CompileTime,	EV_STRING | EV_READONLY,	0,			0 },
    [EVFRENCH_SPACING] ={ "french-spacing",	"One space (not two) at sentence end on fill-paragraph", .v.i = &FrenchSpacing, EV_BOOL,		0,			0 },
    [EVSECONDS] =	{ "time-secs",		"Time, seconds since the epoch",		.v.l = NULL,		EV_LONG | EV_READONLY,		0,			0 },
    [EVPID] =		{ "pid",		"Emacs process ID",				.v.i = NULL,		EV_INT | EV_READONLY,		0,			0 },
    [EVRENEW] =		{ "renew-timeout",	"Seconds: checks/autoread of RENEW mode buffers", .v.i = &RenewTimeout,	EV_INT | EV_MIN,		1,			0 },
    [EVRENEWBELL] =	{ "bell-on-renew",	"Ring the bell on buffer RENEW",		.v.i = &RenewBell,	EV_BOOL | EV_MIN,		0,			0 },
    [EVLANGUAGE] =	{ "language",		"User's language",				.v.s = &language,	EV_STRING | EV_MIN,		2,			0 },
    [EVTERRITORY] =	{ "territory",		"Regional variant of $language",		.v.s = &territory,	EV_STRING,			0,			0 },
    [EVEMACSSAVE] =	{ "notice-emacssave",	EMACS_SAVE_SUFFIX " suffix files are prob crash save",	.v.i = &NoticeEmacsSave, EV_BOOL,		0,			0 },
    [EVFLOWCONTROL] =	{ "enable-flow-control","Terminal Xon/Xoff stop/start C-s/C-q",		.v.i = &xon_xoff,	EV_BOOL,			0,			0 },
    [EVCMDFLAG] =	{ "cmd-flag",		"Flag to current command/macro",		.v.i = &CmdFlag,	EV_BOOL,			0,			0 },
#if	ARG_IS_LONG
    [EVCMDCOUNT] =	{ "cmd-count",		"Count to current command/macro",		.v.l = &CmdCount,	EV_LONG,			0,			0 },
#else
    [EVCMDCOUNT] =	{ "cmd-count",		"Count to current command/macro",		.v.i = &CmdCount,	EV_INT,				0,			0 },
#endif
    [EVMAXQUESTION] = 	{ "max-question",	"How many times to ask a question before giving up", .v.i = &MaxQuestion, EV_INT | EV_MIN,		0,			0 },

    [EVNULL] =		{ (char*)NULL,		(char*)NULL,					.v.i = NULL,		0,				0,			0 },
};



/*	list of recognized user functions	*/

typedef struct UFUNC {
	char*	f_name;	/* name of function */
	int	f_type;	/* Type: return, arg1, arg2, arg3 */
} UFUNC;

/* Values in f_type:
 * Types are V_xx, eg V_STR
 * Arguments numbered: 1, 2, 3 - return 'r'
 */
#define	Ur(type)	((type) <<  0)				// Type of result
#define	U1(type)	((type) <<  4)				// Type of 1st argument
#define	U2(type)	((type) <<  8)				// Type of 2nd argument
#define	U3(type)	((type) << 12)				// Type of 3rd argument

// Get the type of the argument, if V_UNDEF - the argument is not there
#define	Urt(type)	(((type) >>  0) & V_TYPES)		// Get type of result
#define	U1t(type)	(((type) >>  4) & V_TYPES)		// Get type of 1st argument
#define	U2t(type)	(((type) >>  8) & V_TYPES)		// Get type of 2nd argument
#define	U3t(type)	(((type) >> 12) & V_TYPES)		// Get type of 3rd argument
#define	Unt(type, n)	(((type) >> (4 * (n))) & V_TYPES)	// Get type of nth argument (1,2,3)

/* If UUA1RES then Ur() and U1() must be the same types
 */
#define	UU8STR		0x10000		// Check first argument for UTF8 goodness
#define	UUA1RES		0x20000		// Result is in 1st argument - ie changed

/* The point about an argument possibly being boolean is that
 * attempt is made to convert TRUE to 1 and FALSE to 0.
 * This implies that it might be numeric.
 */

/* The functions.
 * Use an enum to get automatic numbering.
 * This makes it easier to add/remove variables.
 *
 * The names are all in caps - this reflects that they were all CPP #defines at one point.
 *
 * Group/order by the sort of operator, eg: numeric ones all together
 */
typedef	enum {

	// Numeric
	UFADD,		// add two numbers together
	UFSUB,		// subtraction
	UFTIMES,	// multiplication
	UFDIV,		// division
	UFMOD,		// mod

	UFNEG,		// negate
	UFNOT,		// logical not
	UFABS,		// absolute value of a number

	UFEQUAL,	// numeric equality check
	UFLESS,		// numeric less than
	UFGREATER,	// numeric greater than
	UFLEQUAL,	// numeric less or equal
	UFGEQUAL,	// numeric greater or equal
	UFNEQUAL,	// numeric not equal
	UFIN,		// number in array/hash

	UFRND,		// get a random number
	UFBAND,		// bitwise and
	UFBOR,		// bitwise or
	UFBXOR,		// bitwise xor
	UFBNOT,		// bitwise not
	UFAND,		// logical and
	UFOR,		// logical or

	// String
	UFCAT,		// concatenate string
	UFLENGTH,	// string length
	UFLEFT,		// left string(string, len)
	UFRIGHT,	// right string(string, pos)
	UFMID,		// mid string(string, pos, len)
	UFSINDEX,	// find the index of one string in another

	UFSEQUAL,	// string logical equality check
	UFSLESS,	// string logical less than
	UFSGREAT,	// string logical greater than
	UFSLEQUAL,	// string less or equality check
	UFSGEQUAL,	// string greater or equality check
	UFSNEQUAL,	// string not equality check
	UFSIN,		// string in array/hash

	UFUPPER,	// uppercase string
	UFLOWER,	// lower case string
	UFREVERSE,	// reverse
	UFTRIM,		// trim whitespace off the end of a string
	UFXLATE,	// XLATE character string translation

	UFSPLIT,	// substrings from input separated by string

	UFISNUM,	// is the arg a number?
	UFNUMBER,	// convert to number: hex, octal, number

	UFASCII,	// char to integer conversion
	UFCHR,		// integer to char conversion

	// Array
	UFCOUNT,	// count elements in an array
	UFPOP,		// pop from end of array
	UFPUSH,		// put to end of array
	UFSHIFT,	// pop from start of array
	UFUNSHIFT,	// put to start of array
	UFACAT,		// catenate two arrays

	// Hash
	UFKEYS,		// return array of keys in hash
	UFHCAT,		// catenate two hashes
	UFKEY_EXIST,	// does key exist in a hash

	// Emacs commands/procedures
	UFCALL,		// call a procedure, return what !return sets
	UFDOCMD,	// call a procedure, return TRUE/FALSE/...
	UFGTCMD,	// get 1 Emacs command
	UFBIND,		// lookup what function name is bound to a key

	UFIND,		// evaluate indirect value

	UFTYPE,		// Type of VALue

	// Operating system stuff
	UFENV,		// retrieve a system environment var
	UFEXIST,	// check if a file exists
	UFFIND,		// look for a file on the path...
	UFSTAT,		// get information about a file
	UFGTKEY,	// get 1 character

	// 11 May 2001
	UFTRUTH,	// Truth of the universe logical test

	UF_ENDMARKER,	// NULL - is end
} funcNums;


/* Names of functions.
 * Order alphabetically.
 */
static	UFUNC funcs[] = {
    [UFABS] =		{ "abs",	Ur(V_INT)  | U1(V_INT)	|			   UUA1RES },	// absolute value of a number
    [UFACAT] = 		{ "acat",	Ur(V_ARRAY)| U1(V_ARRAY)| U2(V_ARRAY)			   },	// catenate two arrays
    [UFADD] =		{ "add",	Ur(V_INT)  | U1(V_INT)	| U2(V_INT)  |		   UUA1RES },	// add two numbers together
    [UFAND] =		{ "and",	Ur(V_BOOL) | U1(V_BOOL) | U2(V_BOOL) |		   UUA1RES },	// logical and
    [UFASCII] =		{ "ascii",	Ur(V_INT)  | U1(V_STR)					   },	// char to integer conversion
    [UFBAND] =		{ "band",	Ur(V_INT)  | U1(V_INT)	| U2(V_INT)  |		   UUA1RES },	// bitwise and	 9-10-87  jwm
    [UFBIND] =		{ "bind",	Ur(V_STR)  | U1(V_STR)					   },	// lookup what function name is bound to a key
    [UFBNOT] =		{ "bnot",	Ur(V_INT)  | U1(V_INT)	|			   UUA1RES },	// bitwise not
    [UFBOR] =		{ "bor",	Ur(V_INT)  | U1(V_INT)	| U2(V_INT)  |		   UUA1RES },	// bitwise or	 9-10-87  jwm
    [UFBXOR] =		{ "bxor",	Ur(V_INT)  | U1(V_INT)	| U2(V_INT)  |		   UUA1RES },	// bitwise xor	 9-10-87  jwm
    [UFCALL] =		{ "call",	Ur(V_ANY)  | U1(V_STR)					   },	// call a procedure, return !return
    [UFCAT] =		{ "cat",	Ur(V_STR)  | U1(V_STR)	| U2(V_STR)  |		   UUA1RES },	// concatenate string
    [UFCOUNT] =		{ "count",	Ur(V_INT)  | U1(V_ARRAY)				   },	// count elements in an array
    [UFCHR] =		{ "chr",	Ur(V_STR)  | U1(V_INT)					   },	// integer to char conversion
    [UFDIV] =		{ "divide",	Ur(V_INT)  | U1(V_INT)	| U2(V_INT)  |		   UUA1RES },	// division
    [UFDOCMD] = 	{ "docmd",	Ur(V_INT)						   },	// call a procedure, return TRUE/FALSE/...
    [UFENV] =		{ "env",	Ur(V_STR)  | U1(V_STR)					   },	// retrieve a system environment var
    [UFEQUAL] =		{ "equal",	Ur(V_BOOL) | U1(V_INT)	| U2(V_INT)  |		   UUA1RES },	// numeric equality check
    [UFEXIST] =		{ "exist",	Ur(V_BOOL) | U1(V_STR)					   },	// check if a file exists
    [UFFIND] =		{ "find",	Ur(V_STR)  | U1(V_STR)					   },	// look for a file on the path...
    [UFGEQUAL] =	{ "gequal",	Ur(V_BOOL) | U1(V_INT)	| U2(V_INT)  |		   UUA1RES },	// numeric greater or equality check
    [UFGREATER] =	{ "greater",	Ur(V_BOOL) | U1(V_INT)	| U2(V_INT)  |		   UUA1RES },	// numeric greater than
    [UFGTCMD] =		{ "gtc",	Ur(V_STR)						   },	// get 1 Emacs command
    [UFGTKEY] =		{ "gtkey",	Ur(V_STR)						   },	// get 1 character/keypress
    [UFHCAT] =		{ "hcat",	Ur(V_HASH) | U1(V_HASH)	| U2(V_HASH)			   },	// catenate two hashes
    [UFIN] =		{ "in",		Ur(V_BOOL) | U1(V_INT)	| U2(V_ANY)			   },	// number in array/hash
    [UFIND] =		{ "indirect",	Ur(V_STR)  | U1(V_STR)					   },	// evaluate indirect value
    [UFISNUM] =		{ "isnum",	Ur(V_BOOL) | U1(V_STR)					   },	// is the arg a number?
    [UFKEY_EXIST] =	{ "key-exist",  Ur(V_BOOL) | U1(V_STR)  | U2(V_HASH)			   },	// does key exist in a hash
    [UFKEYS] =		{ "keys",	Ur(V_ARRAY)| U1(V_HASH)					   },	// return array of keys in hash
    [UFLEFT] =		{ "left",	Ur(V_STR)  | U1(V_STR)	| U2(V_INT)			   },	// left string(string, len)
    [UFLENGTH] =	{ "length",	Ur(V_INT)  | U1(V_STR)					   },	// string length
    [UFLEQUAL] =	{ "lequal",	Ur(V_BOOL) | U1(V_INT)	| U2(V_INT)  |		   UUA1RES },	// numeric less or equality check
    [UFLESS] =		{ "less",	Ur(V_BOOL) | U1(V_INT)	| U2(V_INT)  |		   UUA1RES },	// numeric less than
    [UFLOWER] =		{ "lower",	Ur(V_STR)  | U1(V_STR)	|			   UUA1RES },	// lower case string
    [UFMID] =		{ "mid",	Ur(V_STR)  | U1(V_STR)	| U2(V_INT)  | U3(V_INT)	   },	// mid string(string, pos, len)
    [UFMOD] =		{ "modulo",	Ur(V_INT)  | U1(V_INT)	| U2(V_INT)  |		   UUA1RES },	// mod
    [UFNEG] =		{ "negate",	Ur(V_INT)  | U1(V_INT)	|			   UUA1RES },	// negate
    [UFNEQUAL] =	{ "nequal",	Ur(V_BOOL) | U1(V_INT)	| U2(V_INT)  |		   UUA1RES },	// numeric not equality check
    [UFNOT] =		{ "not",	Ur(V_BOOL) | U1(V_BOOL) |			   UUA1RES },	// logical not
    [UFNUMBER] =	{ "number",	Ur(V_BOOL) | U1(V_STR)					   },	// convert to number: hex, octal, number
    [UFOR] =		{ "or",		Ur(V_BOOL) | U1(V_BOOL) | U2(V_BOOL)  |		   UUA1RES },	// logical or
    [UFPOP] =		{ "pop",	Ur(V_ANY)  | U1(V_ARRAY)				   },	// pop from end of array
    [UFPUSH] =		{ "push",	Ur(V_INT)  | U1(V_ARRAY)| U2(V_ANY)			   },	// put to end of array
    [UFREVERSE] =	{ "reverse",	Ur(V_STR)  | U1(V_STR)	|			   UUA1RES },	// reverse
    [UFRIGHT] =		{ "right",	Ur(V_STR)  | U1(V_STR)	| U2(V_INT)			   },	// right string(string, pos)
    [UFRND] =		{ "rnd",	Ur(V_INT)  | U1(V_INT)	|			   UUA1RES },	// get a random number
    [UFSEQUAL] =	{ "sequal",	Ur(V_BOOL) | U1(V_STR)	| U2(V_STR)			   },	// string logical equality check
    [UFSGEQUAL] =	{ "sgequal",	Ur(V_BOOL) | U1(V_STR)	| U2(V_STR)			   },	// string greater or equality check
    [UFSGREAT] =	{ "sgreater",	Ur(V_BOOL) | U1(V_STR)	| U2(V_STR)			   },	// string logical greater than
    [UFSHIFT] =		{ "shift",	Ur(V_ANY)  | U1(V_ARRAY)				   },	// pop from start of array
    [UFSIN] =		{ "sin",	Ur(V_BOOL) | U1(V_STR)	| U2(V_ANY)			   },	// string in array or hash
    [UFSINDEX] =	{ "sindex",	Ur(V_INT)  | U1(V_STR)	| U2(V_STR)			   },	// find the index of one string in another
    [UFSLESS] =		{ "sless",	Ur(V_BOOL) | U1(V_STR)	| U2(V_STR)			   },	// string logical less than
    [UFSLEQUAL] =	{ "slqual",	Ur(V_BOOL) | U1(V_STR)	| U2(V_STR)			   },	// string less or equality check
    [UFSNEQUAL] =	{ "snequal",	Ur(V_BOOL) | U1(V_STR)	| U2(V_STR)			   },	// string not equality check
    [UFSPLIT] = 	{ "split",	Ur(V_ARRAY)| U1(V_STR)	| U2(V_STR)			   },	// substrings from input separated by string
    [UFSTAT] =		{ "stat",	Ur(V_HASH) | U1(V_STR)					   },	// get information about a file
    [UFSUB] =		{ "subtract",	Ur(V_INT)  | U1(V_INT)	| U2(V_INT)  |		   UUA1RES },	// subtraction
    [UFTIMES] =		{ "times",	Ur(V_INT)  | U1(V_INT)	| U2(V_INT)  |		   UUA1RES },	// multiplication
    [UFTRIM] =		{ "trim",	Ur(V_STR)  | U1(V_STR)					   },	// trim whitespace off the end of a string
    [UFTRUTH] =		{ "true",	Ur(V_BOOL) | U1(V_INT)	|			   UUA1RES },	// Truth of the universe logical test
    [UFTYPE] =		{ "type",	Ur(V_STR)  | U1(V_ANY)					   },	// Type of VALue
    [UFUNSHIFT] =	{ "unshift",	Ur(V_INT)  | U1(V_ARRAY)| U2(V_ANY)			   },	// put to start of array
    [UFUPPER] =		{ "upper",	Ur(V_STR)  | U1(V_STR)	|			   UUA1RES },	// uppercase string
    [UFXLATE] =		{ "xlate",	Ur(V_STR)  | U1(V_STR)	| U2(V_STR)  | U3(V_STR)	   },	// XLATE character string translation
    [UF_ENDMARKER] =	{ (char*)NULL,	0							   }	// end
};

#define	NFUNCS	((sizeof(funcs) / sizeof(UFUNC)) - 1)		// Number of functions

/* end */
