/*
 * This file contains PROGNAME, what you want MicroEMACS to call itself on the
 * display line, and the version number VERSION.
 *
 * These are put in here so that you wont have to grub around in the depths
 * of display.c every time you want to remind yourself that you've hacked
 * Emacs by updating the version number.
 *
 * The definitions have been moved out of estruct.h, because, if you change
 * them and you have written your makefile properly, everything gets recompiled
 * whereas putting them here only makes "make" recompile display.c
 * The numbering has also been moved here from Makefile and the .spec file
 * for much the same reasons.
 *
 * ADDW SCCS: @(#)evers.h	4.20 10/06/20 12:02:03
 *
 * BEWARE: the version number is also kept in phcl-MicroEmacs.spec
 * That should be kept in sync with the number below.
 *
 * SCCS/CSSC is used for the version numbering.
 * Beware it's nomenclature is different from what you might expect.
 * It has 4 levels of version:
 * Release Level Branch Sequence
 *
 * Try to map that onto semantic versioning adding build:
 * Major Minor Patch Build - which are obtained by: R L B S
 * with % round each one.
 *
 * This has been reworked so that it looks (& names) more like what is found
 * in a .spec.
 *
 * So, for a full R.L.B.S of 4.11.1.2
 * VERSION	Is 4.11
 * RELEASE	is 1.2
 *		This will start at 0.0 for a new RELEASE, in which case the
 *		numbers will not appear in VERSION_NAME.
 *
 * This is now redundant, don't need the 'ph':
 * VERSION_NAME	Something like 4.11ph or 4.11.1.2ph
 *		Note how the  last 2 numbers might/might-not be there
 *
 * Not that the VERSION_BUILD could be overridden from the build environment.
 * ** Warning ** because of this be careful how you get the build
 *
 * Extract these numbers into various places:
 *
 * Emacs	VERSION	is shown, eg in Mode lines
 *
 * Makefile	VERSION	generated from:
 *			RELEASE - which is VERSION_MAJOR VERSION_MINOR made easy
 *			not interested in patch or build
 * .spec	Version	generated from:
 *			VERSION_MAJOR VERSION_MINOR
 *		Release	generated from:
 *			VERSION_PATCH VERSION_BUILD
 *			VERSION_BUILD is possibly overridden on the makerpm command line
 *
 * VERSION_DATE		Something like 2020/03/21
 *
 * https://semver.org/
 * https://www.gnu.org/software/cssc/manual/Keyword-Substitution.html#Keyword-Substitution
 *
 * You are guaranteed that the VERSION_??? appear as in the order below.
 * Useful if you are looking at this with a script.
 */

#define	PROGNAME	"Emacs"
#define	VERSION_MAJOR	"4"				// Release - Major
#define	VERSION_MINOR	"20"				// Level - Minor
#define	VERSION_PATCH	"0"				// Branch - Patch

#ifndef	VERSION_BUILD
#define	VERSION_BUILD	"0"				// Sequence - Build
#endif

#define	VERSION		"4.20"				// Release.Level
#define	BRANCH		VERSION_PATCH"."VERSION_BUILD	// Branch.Sequence
#define	VERSION_NAME	RELEASE"."BRANCH"ph"		// Release.Level[.Branch.Sequence]ph

#define	VERSION_DATE	"2020/10/06"				// 20yy/mm/dd
