/*
 * ADDW SCCS: @(#)exec.c	1.141 10/15/20 13:48:39
 */

/*	This file is for functions dealing with execution of
 *	commands, command lines, buffers, files and startup files
 *
 *	written 1986 by Daniel Lawrence.
 */

#include	"estruct.h"
#include	"utf8.h"
#include	"edef.h"

#if	SHOWMODREL
char	execvers[] = "@(#)exec.c	1.141 10/15/20 13:48:39";
#endif

// TODO utf8

/* Find a procedure, auto-load it if needed.
 *
 * If something goes wrong then an error message will be generated.
 *
 * Return a buffer pointer
 *	  NULL if it cannot be found
 */
BUFFER* findProcedure(BYTE* procName)
{
	BUFFER*	bp;
	BUFFER*	cb = curbp;
	int	ret;
	BYTE	bufn[BUFFER_NAME_SIZE];/* name of buffer to execute */

	// mlwrite("%#findProcedure('%s')", procName);

	/* Construct the buffer name
	 * Procedures live in a buffer of the same name surrounded by []
	 */
	sprintf(bufn, "[%s]", procName);

	// The buffer might already exist:
	if((bp = bfind(bufn, BFNNONE, 0)) != NULL)
		return(bp);

	/* Save current context so can restore after dobuf() below.
	 * The dobuf() will leave execText in the [Proc Load] buffer which
	 * will have been zapped.
	 * The context will be the line that triggered the autoload.
	 */
	TEXT*	saveLine = execText;
	LineLn	savePos  = execPos;

	// Find a file that contains the macro
	if(findFile(procName, fT_Macro) == 0) {
		mlwrite("%EFile with command not found '%s'", procName);
		return((BUFFER*)NULL);
	}

	/* Temp buffer to load it.
	 * Do this second as it is less likely to fail than the findFile()
	 */
	if((bp = bfindCheck(ProcLoadBname, BFNCREATE|BFNACTIVE|BFNVIEW)) == NULL) {
		ffclose();
		return((BUFFER*)NULL);
	}

	curbp = bp;
	ret = readFD(FALSE);
	curbp = cb;
	if(ret != TRUE) {
		ffclose();
		return((BUFFER*)NULL);
	}

	// This should create the procedure
	// We don't care about any value returned
	VAL* retval = valGet(V_ANY);
	ret = dobuf(bp, &retval);
	valRelease(retval);

	// Restore context changed by dobuf()
	execText = saveLine;
	execPos  = savePos;

	// make invisible ?
	zotbuf(bp, TRUE);

	if(ret != TRUE) {
		mlwrite("%ERun of buffer %s to create procedure '%s' fails %d", bufn, procName, ret);
		return((BUFFER*)NULL);
	}

	// Hopefully that will have created the procedure buffer to be found now
	return(bfind(bufn, BFNACTIVE|BFNVIEW, 0));
}

/* namedcmd:	execute a named command even if it is not bound.
 *
 * command arguments [passed through to command executed]
 *
 * The difference with execcmd (execute-command-line) is that this just takes
 * a command, execcmd take a command line - ie has arguments.
 *
 * There is not a default argument.
 *
 * Command:  execute-named-command
 * Bound to: M-x
 *
 * Returns:
 *	FALSE the command does not exist
 *	otherwise: the return value of the command
 */
int	namedcmd(int f, Cnt n)
{
	NBIND*	kfunc;			/* ptr to the requested function to bind to */

	/* and now get the function name to execute */
	kfunc = getname(": ");
	if(kfunc == NULL) {
		// Don't complain if empty command entered interactively
		if(clexec == TRUE || lused(bmlEmCmdHistp->b_dotp) != 0)
			mlwrite("%ENo such command '%l'", bmlEmCmdHistp->b_dotp);

		return(FALSE);
	}

	/* clear the command line */
	mlerase();

	return(FuncExec(kfunc, f, n));
}

/* Execute a command line command to be typed in by the user.
 *
 * Flag in_execcmd is a tacky hack to interact with setvar()
 * The problem is that in macro debug mode it stops, gets a keystroke,
 * before continuing - this is to allow the assignment to be seen/read
 * by the user.
 * Without in_execcmd this happens when this execs 'set $debug 1',
 * the keystroke read is not expected by the user, so whatever gets lost
 * and causes annoyance.
 *
 * The count is ignored. There is not a default value.
 *
 * Command:  execute-command-line
 * Bound to: M-C-x
 *
 * Return:
 *	FALSE - nothing entered at prompt
 *	ABORT - user aborts the prompt - C-g
 *	the result of executing the command line
 */
/*ARGSUSED*/
int	execcmd(int f, Cnt n)
{
	register int status;		/* status return */

	/* get the line wanted */
	if((status = mlprompt(": ", MLSTRING, bmlEmCmdHistp)) != TRUE)
		return(status);

	in_execcmd = !clexec;	// true if not running from a macro

	execText = &ltxt(bmlEmCmdHistp->b_dotp);
	execPos = 0;
	status = docmd();
	in_execcmd = FALSE;

	return(status);
}

/* docmd: execute the command line in global execText.
 *
 * This function will be used by execute-command-line and by all source and
 * startup files.
 *
 * You would have thought that the command would have been passed as an
 * argument to this function. This used to happen. The problem is that macro
 * parsing is now done using globals execText & execPos and execPos is changed
 * as the line is 'used' ... (it isn't always zero when this is called) so some
 * means of updating execPos would need to be devised ... simpler - just access
 * the global variables.
 *
 * Lastflag/thisflag is also updated.
 *
 * The format of the command line is:
 *	{# arg} <command-name> {<argument string(s)>}
 */
int	docmd(void)
{
	register int f;		/* default argument flag */
	register Cnt n;		/* numeric repeat value */
	BUFFER*	bp;		/* buffer to execute */
	int	status;		/* return status of function */
	int	oldcle;		/* old contents of clexec flag */
	VAL*	tokn;		/* next token off of command line */
	VAL*	retval;		/* Return value of command */
	NBIND*	fctp;

	// mlwrite("%#docmd('%t'/%D)", execText, execPos);

	/* first set up the default command values */
	f = FALSE;
	n = 1;			// The (unset) count MUST be 1 - see estruct.h
	lastflag = thisflag;
	thisflag = 0;

	tokn = valGet(V_STR);	// Cannot used playValStr as this can end up recursive

	if((status = macarg(&tokn)) != TRUE) {	/* and grab the first token */
		valRelease(tokn);
		return(status);		// Fail
	}

	/* process leadin argument */
	if(vtype(tokn) == V_INT) {
		f = TRUE;
		n = vint(tokn);
	} else
		if(gettyp(vTEXT(tokn)) == TKLIT) {
			f = TRUE;
			n = atoCnt(vtext(tokn));
		}

	// Used up tokn as count to the command/function ?
	if(f) {
		// Get the next token
		if((status = macarg(&tokn)) != TRUE) {
			valRelease(tokn);
			return(status);
		}
	}

	// Looking for commands ... so the tokn must be a string
	if(vtype(tokn) != V_STR) {
		mlwrite("%EExpecting command, found '%v'", tokn);
		valRelease(tokn);
		return(FALSE);
	}

	/* Match the token to see if it is an Emacs command */
	// mlwrite("%#look for cmd '%s'", vtext(tokn));
	if((fctp = fncmatch(vtext(tokn))) != (NBIND*)NULL) {
		/* save the arguments and go execute the command */
		oldcle = clexec;		/* save old clexec flag */
		clexec = TRUE;			/* in cline execution */
		// mlwrite("%#Emacs command: %v %d %d : %s", tokn, f, n, (tgetcaddr(execText, execPos)));
		status = FuncExec(fctp, f, n);	/* call the function */
		cmdstatus = status;		/* save the status */
		clexec = oldcle;		/* restore clexec flag */
		valRelease(tokn);

		return(status);
	}

	// Not an Emacs command, maybe a procedure:
	// mlwrite("%#about to findproc('%v') execText='%t'", tokn, execText);

	if((bp = findProcedure(vtext(tokn))) == NULL) {
		mlwrite("%ENo such command/procedure '%v'", tokn);
		valRelease(tokn);
		return(FALSE);
	}

	// mlwrite("%#found proc '%v' n=%d", tokn, n);
	/* execute the buffer */
	oldcle = clexec;	/* save old clexec flag */
	clexec = TRUE;		/* in cline execution */
	retval = valGet(V_ANY);

	while(--n >= 0)
		if((status = dobuf(bp, &retval)) != TRUE)
			break;

	cmdstatus = status;	/* save the status */
	clexec = oldcle;	/* restore clexec flag */

	valRelease(tokn);
	valRelease(retval);

	return(status);
}

/* token:	chop a token off a string - space separated,
 *		recognise double quoted strings.
 *
 * tok	  destination token TEXT, copied from execText & NUL terminated
 *	  If this is empty then no token was found.
 *
 * The token is obtained from execText at offset execPos. execPos
 * will be updated to be after the token found.
 *
 * The first character in tok is the start of the token, eg: % or &,
 * notably a " if it is a string - but remember that strings do not need
 * to start with a quote.
 *
 * Rewritten as it used to append characters using tcatc()/tcatm() and profiling
 * showed that tcatm() used most of the CPU time. utf8_chkbuf() is a large part
 * of that So try:
 * 1) copy to local buffer and append when the buffer is full - fewer tcatm()
 * 2) copy directly from execText, only flush @ end & when see specials (not often)
 * This function uses a lot of the CPU time when running a macro.
 *
 * TODO utf8 - prompt & init file stuff
 * If the source is longer than size - ensure not to have a part
 * copied multibyte character.
 *
 * Return:
 *	TRUE  something put in tok
 *	FALSE no token available - end of line
 */
int	token(TEXT* tok)
{
	register int quotef;	/* is the current string quoted? */
	register BYTE c;	/* temporary character */
	int	eatSep = FALSE;	// Move past the separator that triggers end of token

	tsetempty(tok);

	/* first scan past any whitespace in the source */
	while(tgetch(execText, execPos) == ' ' || tgetch(execText, execPos) == '\t')
		execPos++;

	if(execPos >= tused(execText))	// Nothing to return
		return(FALSE);

	// The difference between the 2 below is that cstart will change after a ~something
	LineLn	tstart = execPos;	// Start of the token
	LineLn	cstart = execPos;	// Start of component we have not output

	/* scan through the source string */
	quotef = FALSE;
	while(execPos < tused(execText)) {
		/* process special characters */
		if(tgetch(execText, execPos) == '~') {

			// Flush any not copied:
			if(execPos > cstart)
				tcatm(tok, tgetcaddr(execText, cstart), execPos - cstart);

			if(++execPos >= tused(execText)) {
				cstart = execPos;	// Don't copy what is already copied
				break;
			}

			/* Differ from Microemacs 5.0 which has ~r and ~n go to \r ???
			 * A quick survey of macros shows ~n being used but not ~r or ~l
			 */
			switch (tgetch(execText, execPos++)) {
				case 'r':	c = '\r'; break;
				case 'l':
				case 'n':	c = '\n'; break;
				case 't':	c = '\t'; break;
				case 'b':	c = '\b'; break;
				case 'f':	c = '\f'; break;
				case 'e':	c = CharEscape; break;
				case '"':	c = '"';  break;
				default:	c = tgetch(execText, execPos - 1);  break;
			}

			tcatc(tok, c);	// Append this char

			cstart = execPos;

		} else {
			/* check for the end of the token */
			if(quotef) {
				if(tgetch(execText, execPos) == '"') {
					eatSep = TRUE;
					break;
				}
			} else {
				// End token on white space
				if(tgetch(execText, execPos) == ' ' || tgetch(execText, execPos) == '\t')
					break;

				/* Also break when see a {}[],
				 * If this is the first thing seen: make it be the token,
				 * if seen something break before it.
				 * This causes token end - don't move past what follows, might
				 * be white space - that will be eaten later
				 */
				if(tgetch(execText, execPos) == '{' || tgetch(execText, execPos) == '}' ||
				   tgetch(execText, execPos) == '[' || tgetch(execText, execPos) == ']' ||
				   tgetch(execText, execPos) == ',') {
					if(tstart == execPos)	// Start of the token - put the bracket to the token
						execPos++;
					break;
				}
			}

			/* set quote mode if quote found */
			if(tgetch(execText, execPos) == '"')
				quotef = TRUE;

			/* Move to the next char.
			 * Beware: if it is multibyte then they all need to be added together.
			 */
			execPos += CHAR_LEN(tgetch(execText, execPos));
		}
	}

	// Copy the remainder
	// In-lineing what tcatm() does gives about a 12% performance improvement
	if(execPos > cstart)
		tcatm(tok, tgetcaddr(execText, cstart), execPos - cstart);

	/* Move past the token ending character which is "
	 * Don't move past end of LINE
	 */
	if(eatSep && execPos < tused(execText))
		execPos++;

	return(TRUE);
}

/* get a macro line argument
 *
 * tok	buffer to place argument
 *	This is a VAR* since it can be typed
 *
 * For now tok will always be V_STR - MUST always be on entry
 * 
 * Return TRUE  - if got something
 *	  FALSE - nothing available
 *	  ABORT - C-g
 */
int	macarg(VAL** tok)
{
	int savcle;	/* buffer to store original clexec */
	int status;

	savcle = clexec;	/* save execution mode */
	clexec = TRUE;		/* get the argument */

	// Use Prompt Line for history
	status = nextarg("", tok, MLNOOPT, bmlPromptHistp);

	clexec = savcle;	/* restore execution mode */

	return(status);
}

/* nextarg:	get the next argument
 *
 * prompt	prompt to use if we must be interactive
 * tokn		VAL to put token into, is probably a V_STR but might be changed.
 *		This is a VAL as the next argument could be typed, can be any type.
 *		If a type is set in option (MLTYPEMASK) this will be a V_STR and type not converted
 * option	options for getstring
 *		Of interest is MLTYPEMASK which says that the type of tokn must
 *		remain a STR as the value is file/buffer name, option, ...
 * history	the buffer that contains history of this prompt type
 *		Not used when running a macro
 *
 * If not interactive (ie in a macro) evaluate the next token since it could
 * be a variable or some expression.
 *
 * Type juggling is a bit of a pain - want to pass on the type that is obtained
 * by getval() - will become important once arrays/hashes are implemented.
 * Overlap/duplication with what gtfun() does.
 *
 * Return TRUE  - if got something
 *	  FALSE - nothing available
 *	  ABORT - C-g
 */
int	nextarg(const BYTE* prompt, VAL** tokn, int option, BUFFER* history)
{
	int	res;

	/* if we are interactive, go get it! */
	if(clexec == FALSE)
		return(getstring(prompt, tokn, option, history));

	/* grab token and advance past */
	token(playValue);

	// No token, fail:
	if(tused(playValue) == 0) {
		// Unless no value is acceptable:
		if((option & MLEMPTYOK) == 0) {
			if(vtype(*tokn) != V_ANY)
				mlwrite("%EMissing argument, prompt '%s'", prompt);

			valChangeType(tokn, V_ERROR);
		}

		return(FALSE);	// No value
	}

	/* evaluate it
	 * This could return something in buffer, eg if the argument
	 * is "fred" it will return the address of the 'f'.
	 * Beware overlapping copies do not always work.
	 *
	 * **** beware this can be recursive ... so use of global is not a good idea
	 */
	VAL*	val = getval(playValue);

	if(tgetch(execText, execPos) == '[' || tgetch(execText, execPos) == '{') {
		/* Looks like we have an array or hash and have to index/key into it.
		 * The value obtained is the one passed back.
		 */
		VAL**	valp = parseStructVar(val, FALSE);

		if(valp == NULL) {
			mlwrite("%EError indexing into array or keying into a hash");
			return(ABORT);
		}

		valCopy(&val, *valp);
	}

	// Fail if getting a value went wrong
	res = vtype(val) != V_ERROR;

	/* If the type of the result token is ANY then convert it to the
	 * type of the value just obtained - including ERROR.
	 * This will likely be for a !return.
	 */
// No need to change type with copy by ref
// 	if(vtype(*tokn) == V_ANY)
// 		valChangeType(tokn, vtype(val));

	// If we got an error - propagate back upwards
	if(vtype(val) == V_ERROR) {
		valChangeType(tokn, V_ERROR);
	} else {
		// Non error - copy the value
		// If we can just copy: right type or don't care what type
		if(vtype(*tokn) == vtype(val) || vtype(*tokn) == V_ANY) {
			valCopy(tokn, val);
		} else {
			// Types differ.
			// Can the type of tokn be changed ?
			if((option & MLTYPEMASK) != 0) {
				// Convertion is not allowed - tokn will be a string
				// Get a string representation of val and copy.
				// Clone in case valPrint returns playValStr
				VAL*	clone = valStrClone(valPrint(val));
				valCopy(tokn, clone);
				valRelease(clone);
			} else {
				// Convert val to the type of tokn & copy
				res = valCopyValue(tokn, val);
			}
		}
		// mlwrite("%#tokn is %s '%v'", vtypename(tokn), *tokn);
	}

	valRelease(val);

	return(res);
}


/* Set up a macro buffer and flag to store all executed command lines there.
 *
 * n:	Name of buffer to use
 *
 * This command cannot be run interactively, ie must be in a macro file/buffer.
 *
 * Global bstore is assigned the BUFFER created for the macro.
 *
 * There is not a default count, one must be set.
 *
 * Command:  store-macro
 * Bound to: unbound
 *
 * Return:
 *	TRUE  success
 *	FALSE
 *	    Must not be run interactively
 *	    No argument (macro number) given
 *	    Number too big
 *	    Cannot create buffer
 */
int	storemac(int f, Cnt n)
{
	BUFFER*	bp;		/* pointer to macro buffer */
	BYTE	bname[BUFFER_NAME_SIZE];	/* name of buffer to use */

	if(clexec == FALSE) {
		mlwrite("%ECan not be run interactively");
		return(FALSE);
	}

	/* must have a numeric argument to this function */
	if(f == FALSE) {
		mlwrite("%ENo macro number specified");
		return(FALSE);
	}

	/* Range check the macro number.
	 * The number 40 should be parameterised. TODO
	 */
	if(n < 1 || n > MAX_MACRO_NUM) {
		mlwrite("%EMacro number %d out of range. Max %d", n, MAX_MACRO_NUM);
		return(FALSE);
	}

	/* Construct the macro buffer name */
	sprintf(bname, "[Macro %2.2d]", (int)n);

	/* set up the new macro buffer */
	if((bp = bfind(bname, BFNCREATE|BFNACTIVE, BFINVS)) == NULL) {
		mlwrite("%ECan not create macro buffer");
		return(FALSE);
	}

	/* and make sure it is empty */
	bclear(bp);

	// Get a b_procinfo
	if(MkProcInfo(bp) != TRUE) {
		mlwrite(MsgOutOfMemory);
		return(FALSE);
	}

	/* and set the macro store pointers to it */
	mstore = TRUE;
	bstore = bp;

	return(TRUE);
}

#if	PROC
/* Set up a procedure buffer and flag to store all executed command lines there.
 *
 * n:	Name of buffer to use
 *	If this is not set then the name is the argument to this command.
 *
 * This command cannot be run interactively, ie must be in a macro file/buffer.
 *
 * Global bstore is assigned the BUFFER created for the macro.
 *
 * A default count is not set
 *
 * Command:  store-procedure
 * Bound to: unbound
 *
 * Return:
 *	TRUE  success
 *	FALSE
 *	    Must not be run interactively
 *	    Cannot create buffer
 *	    Out of memory
 */
int	storeproc(int f, Cnt n)
{
	register struct BUFFER*	bp;			// pointer to macro buffer
	PARG*		last_arg;			// last macro argument
	PARG*		cur_arg;			// current macro argument
	int		arrayArg = FALSE;		// Procedure has an argument that is an array
	Prinfo*		pi;				// Info for the procedure
	BYTE		bname[BUFFER_NAME_SIZE];	// name of buffer to use
	BYTE		pname[BUFFER_NAME_SIZE];	// name of procedure

	if(clexec == FALSE) {
		mlwrite("%ECan not be run interactively");
		return(FALSE);
	}

	/* a numeric argument means its a numbered macro */
	if(f == TRUE)
		return(storemac(f, n));

	/* get the name of the procedure */
	token(playValue);

	if(tused(playValue) == 0) {
		mlwrite("%EMissing procedure name");
		return(FALSE);
	}

	if(tused(playValue) >= BUFFER_NAME_SIZE - 3) {
		mlwrite("%EProcedure name too long: %t", playValue);
		return(FALSE);
	}

	if(fncmatch(ttext(playValue)) != NULL) {
		mlwrite("%EProcedure cannot have the same name as an Emacs command: %t", playValue); 
		return(FALSE);
	}

	/* construct the macro buffer name */
	sprintf(bname, "[%s]", ttext(playValue));

	strcpy(pname, ttext(playValue));	// We'll need this in a bit

	/* set up the new macro buffer */
	if((bp = bfind(bname, BFNCREATE|BFNACTIVE, BFINVS)) == NULL) {
		mlwrite("%ECan not create buffer for procedure %t", playValue);
		return(FALSE);
	}

	/* and make sure it is empty */
	bclear(bp);

	// Get a b_procinfo
	if(MkProcInfo(bp) != TRUE) {
		mlwrite(MsgOutOfMemory);
		return(FALSE);
	}

	pi = bp->b_procinfo;

	/* Retrieve and store any formal parameters.
	 * # is not a comment here as the store-procedure command is earlier on the line.
	 * The formal parameter names must all start with a '%'
	 */
	last_arg = (PARG *)NULL;
	token(playValue);

	while(execPos <= tused(execText) && tused(playValue) > 0 && tgetch(playValue, 0) == '%') {

		if(arrayArg) {
			mlwrite("%ECannot have another parameter after array parameter: %s", cur_arg->pa_name);
			return(FALSE);
		}

		// If there is '[]' after the arg name - it is to be an array
		// The '[]' is not part of the name in playValue
		if(tgetch(execText, execPos) == '['&& tgetch(execText, execPos + 1) == ']') {
			execPos += 2;	// Skip over '[]'
			arrayArg = TRUE;
		}

		if(tused(playValue) > NVSIZE) {
			mlwrite("%EParameter name too long '%t'", playValue);
			return(FALSE);
		}

		/* allocate an argument */
		cur_arg = (PARG *)malloc(sizeof(PARG));
		if (cur_arg == (PARG *)NULL) {
			mlwrite("%ECan not create macro (OOM)");
			return(FALSE);
		}

		/* and add it to the linked list of arguments for this buffer */
		strcpy(cur_arg->pa_name, ttext(playValue));
		if(arrayArg)
			cur_arg->pa_name[0] = '[';	// Replace '%' with '['

		cur_arg->pa_next = (PARG *)NULL;
		if(last_arg == (PARG *)NULL)
			pi->pri_args = cur_arg;
		else
			last_arg->pa_next = cur_arg;

		/* and let the buffer total these */
		pi->pri_numargs++;
		last_arg = cur_arg;

		/* on to the next parameter */
		token(playValue);
        }

	/* A semicolon (start of comment) may also be found here, following it
	 * is a brief description of the procedure.
	 * However use of token() above will have removed the word starting
	 * with ';' and put it in playValue ... but also anything not separated
	 * from it by a space. After that will be the other words in the
	 * comment.
	 */
	if(tgetch(playValue, 0) == ';') {
		tdeln(playValue, 0, 1);		// Zap the ';'

		// Append a space to separate from subsequent words
		tcatc(playValue, ' ');

		// Append any other words
		tcatm(playValue, tgetcaddr(execText, execPos), tused(execText) - execPos);
		ttrim(playValue);

		// Replace the buffer's clean file name with this comment
		free(bp->b_fnameClean);

		bp->b_fnameClean = utf8_strmake(ttext(playValue));

		// Put this comment into any ProcNames struct for this procedure
		NBIND*	nbp;
		if((nbp = procFind(pname, TRUE)) != NULL) {
			if(nbp->n_desc != NULL)
				free(nbp->n_desc);
			nbp->n_desc = strmake(bp->b_fnameClean);
		}
	}
	else	// The message below only shows the first word ... good enough for now
		if(tused(playValue) > 0)
			mlwrite("%WUnexpected text after store-procedure '%t'", playValue);

	/* and set the macro store pointers to it */
	mstore = TRUE;
	bstore = bp;

	return(TRUE);
}

/* Execute a procedure.
 *
 * Execute the procedure count times, once if count unset.
 *
 * Variables $cmd-flag and $cmd-count are set to f and n.
 *
 * Default count unset
 *
 * Command:  execute-procedure   run
 * Bound to: unbound
 *
 * Return
 *	TRUE  Procedure executed and nothing failed
 *	FALSE Procedure does not exist
 *	      Nothing entered at prompt
 *	ABORT User aborts the prompt - C-g
 *	What the procedure returns
 */
int	execproc(int f, Cnt n)
{
	register BUFFER *bp;		/* ptr to buffer to execute */
	register int status;		/* status return */
	BYTE bufn[BUFFER_NAME_SIZE];	/* name of buffer to execute */

	/* find out what buffer the user wants to execute */
	if((status = mlprompt("Execute procedure: ", MLXBUF | MLNOSPACE, bmlProcHistp)) != TRUE)
		return(status);

	/* construct the buffer name */
	sprintf(bufn, "[%s]", ltext(bmlProcHistp->b_dotp));

	/* find the pointer to that buffer */
	if((bp = findProcedure(ltext(bmlProcHistp->b_dotp))) == NULL) {
		mlwrite("%ENo such procedure '%l'", bmlProcHistp->b_dotp);
                return(FALSE);
	}

	// Set before games below:
	CmdFlag = f;
	CmdCount = n;
	VAL* retval = valGet(V_ANY);

	/* And now execute it as asked,
	 * once if f is not set
	 */
	do {
		if((status = dobuf(bp, &retval)) != TRUE)
			return(status);

		if( !f)
			break;

	} while(--CmdCount > 0);

	valRelease(retval);

	return(status);
}
#endif

/* Execute the contents of a buffer of commands.
 *
 * Prompt (or argument) for the name of the buffer to execute.
 *
 * Execute the buffer count times, stops if execution does not return TRUE.
 * Default count 1. Must not be < 0
 *
 * Command:  execute-buffer
 * Bound to: unbound
 *
 * Return:
 *	TRUE  Procedure executed and nothing failed
 *	FALSE Procedure does not exist
 *	      what the execution of the buffer returns (if not TRUE)
 *	      Nothing entered at prompt
 *	ABORT User aborts the prompt - C-g
 */
int	execbuf(int f, Cnt n)
{
	register BUFFER *bp;		/* ptr to buffer to execute */
	register int status;		/* status return */

	/* find out what buffer the user wants to execute */
	if((status = mlprompt("Execute buffer: ", MLXBUF | MLU8LEN, bmlProcHistp)) != TRUE)
		return(status);

	/* find the pointer to that buffer */
	if((bp = bfindFile(ltext(bmlProcHistp->b_dotp), (BYTE*)NULL)) == NULL)
		return(FALSE);

	VAL* retval = valGet(V_ANY);

	/* and now execute it as asked */
	while(--n >= 0)
		if((status = dobuf(bp, &retval)) != TRUE)
			break;

	valRelease(retval);

	return(status);
}

/* Clear out a buffer's b_procinfo.
 *
 * free() everything
 */
void	ClearProcInfo(BUFFER* bp)
{
	Prinfo*	pi;		// What we clear

	pi = bp->b_procinfo;

	if(pi == (Prinfo*)NULL)
		return;

	// Free procedure arguments:
	while(pi->pri_args) {
		PARG*	next = pi->pri_args->pa_next;
		free(pi->pri_args);
		pi->pri_args = next;
	}

	while(pi->pri_whlist != NULL) {
		WHBLOCK* next = pi->pri_whlist->w_next;
		if(pi->pri_whlist->w_array != (VAL*)NULL)
			valRelease(pi->pri_whlist->w_array);
		free(pi->pri_whlist);
		pi->pri_whlist = next;
	}

	// Free references to gotos that we have found
	while(pi->pri_gotos != (golist*)NULL) {
		golist*	next = pi->pri_gotos->gl_next;
		free(pi->pri_gotos->gl_name);
		free(pi->pri_gotos);
		pi->pri_gotos = next;
	}

	free(bp->b_procinfo);
	bp->b_procinfo = (Prinfo*)NULL;
}

/* Ensure that the buffer has a valid b_procinfo.
 *
 * If allocated done with calloc() - so empty.
 *
 * Return TRUE  - OK
 *	  FALSE - failed to allocate
 */
int	MkProcInfo(BUFFER* bp)
{
	if(bp->b_procinfo == (Prinfo*)NULL &&
	  (bp->b_procinfo = (Prinfo*)calloc(1, sizeof(Prinfo))) == NULL)
		return(FALSE);

	return(TRUE);
}

/* Scan a buffer for !while, goto labels, local var declarations, etc.
 * The purpose is to do this once, not every time that a procedure
 * is called or buffer run.
 *
 * Return:
 *	TRUE  - all OK
 *	FALSE - out of memory
 */
int	BuildProcInfo(BUFFER* bp)
{
	int	status = TRUE;	// What we return
	LINE*	lp;		// line being looked at
	WHBLOCK* scanner;	// ptr during scan
	Prinfo*	pi;		// Info that we are building
	int	dirnum;		// Directive index
	char*	abortmsg;	// Message to give to errormsg()
	int	inHelp = 0;	// In a **!help-begin section

	abortmsg = "Error";	// Default - OK if there is a prior error
	scanner = NULL;
	lp = (LINE*)NULL;

	if(MkProcInfo(bp) != TRUE) {
		abortmsg = OutOfMemory;
		goto eabort;
	}

	pi = bp->b_procinfo;

	if(pi->pri_scanned)	// Already done it
		return(TRUE);

	/* scan the buffer to execute, building !while header blocks
	 * and counting local variables
	 */
	for(lp = lforw(bp->b_linep); lp != bp->b_linep; lp = lforw(lp)) {
		/* scan the current line */
		execText = &ltxt(lp);
		execPos = 0;
		// mlwrite("%#scan line '%t'(lp=%p)", execText, lp);

		// NUL terminating makes the code below easier:
		if( !tsetNUL(execText)) {
			abortmsg = OutOfMemory;
			goto eabort;
		}

		// Directive line ?
		if(tgetch(execText, 0) == '*' && tgetch(execText, 1) == '*') {
			/* I should really test to see if the line is:
			 * **#	comment
			 * **!	directive
			 * But no point of doing this as all that matters is
			 * recognising !help-begin & !help-end - the lines are
			 * then not looked at any more - so just goto nxtscan.
			 */
			if(strncmp(tgetcaddr(execText, 2), "!help-begin", 11) == 0) {
				inHelp = 1;
				lsetNotes(lp, MAC_NOTE_HELP);
			}
			if(strncmp(tgetcaddr(execText, 2), "!help-end", 9) == 0)
				inHelp = 0;

			continue;
		}

		// Ignore help text
		if(inHelp) {
			lsetNotes(lp, MAC_NOTE_HELP);
			continue;
		}

		/* Span leading whitespace */
		while(tgetch(execText, execPos) == ' ' || tgetch(execText, execPos) == '\t')
			execPos++;

		/* if theres nothing here, don't bother */
		if(execPos >= tused(execText))
			continue;

		/* Goto label. Ignore the problem of something like: *&add 1 2
		 * also ignore duplicate lables
		 * Put this new label into the list of labels for the buffer
		 * Ignore labels in procedures being built
		 */
		if(tgetch(execText, execPos) == '*' && Isletter(tgetch(execText, execPos + 1))) {
			BYTE	name[NVSIZE + 1];
			int	i;
			golist*	lab;	// New goto label

			if(mstore)	// Don't enable jumping into procedures
				continue;

			execPos++;	// Skip the '*'

			for(i = 0; i < sizeof(name) - 1; i++) {
				if( !Isalnum(tgetch(execText, execPos + i)))
					break;

				name[i] = tgetch(execText, execPos + i);
			}
			name[i] = '\0';

			if((lab = (golist*)malloc(sizeof(golist))) == NULL ||
			    (lab->gl_name = strmake(name)) == NULL) {
				if(lab)
					free(lab);
				abortmsg = OutOfMemory;
				goto eabort;
			}

			lab->gl_line = lp;
			lab->gl_next = pi->pri_gotos;
			pi->pri_gotos = lab;

			continue;
		}

		// Something like !if or !while ?
		dirnum = DNOTHING;	// NUL value
		if(tgetch(execText, execPos) == '!') {
			/* Find out which directive this is.
			 * Beware: the strncmp() compares the initial length and not beyond,
			 * so it will confuse !elseif with !else.
			 */
			execPos++;
			for( ; dirnum < NUMDIRS; dirnum++)
				if(strncmp(tgetcaddr(execText, execPos), dname[dirnum],
					    (size_t)dname_len[dirnum]) == 0 &&
					    !Isletter(tgetch(execText, execPos + dname_len[dirnum])))
					break;

			/* and bitch if it's illegal */
			if(dirnum == NUMDIRS) {
				abortmsg = "Unknown Directive";
				goto eabort;
			}

			/* +1 to convert to the #defined values
			 * Save as a note in the line.
			 */
			dirnum++;
			lsetNotes(lp, (short)dirnum);
		}

		/* if we are already in a stored-procedure */
		if(dirnum == DENDM) {
			mstore = FALSE;
			continue;
		}

		/* stored procedure?
		 * efunc.h also binds "store-procedure" to storeproc()
		 * The purpose of recognising it here is so that we don't
		 * see goto lables within other procedures
		 * I suppose that if mstore I ought to complain as can't nest.
		 */
		if(strncmp(tgetcaddr(execText, execPos), "store-procedure", 15) == 0) {
			mstore = TRUE;
			continue;
		}

		/* local variable declaration? */
		if(strncmp(tgetcaddr(execText, execPos), "local", 5) == 0)
			pi->pri_numlocals++;

		/* if is a !while directive, make a block... */
		if(dirnum == DWHILE) {
			WHBLOCK* whtemp = (WHBLOCK *)malloc(sizeof(WHBLOCK));
			if(whtemp == NULL) {
noram:				abortmsg = "Out of memory during while scan";
				goto eabort;
			}
			whtemp->w_begin = lp;
			whtemp->w_type = BTWHILE;
			whtemp->w_array = (VAL*)NULL;
			whtemp->w_next = scanner;
			scanner = whtemp;
		}

		/* if is a !break or !continue directive, make a block...
		 * The statements are the same, the only difference is
		 * what is the next line of code that is executed.
		 */
		if(dirnum == DBREAK || dirnum == DCONTINUE) {
			WHBLOCK* whtemp;	// Temporary ptr to a WHBLOCK

			if(scanner == NULL) {
				abortmsg = (dirnum == DBREAK ? "!break outside of any !while loop" :
							       "!continue outside of any !while loop");
				goto eabort;
			}
			whtemp = (WHBLOCK *)malloc(sizeof(WHBLOCK));
			if(whtemp == NULL)
				goto noram;
			whtemp->w_begin = lp;
			whtemp->w_type = dirnum;
			whtemp->w_array = (VAL*)NULL;
			whtemp->w_next = scanner;
			scanner = whtemp;
		}

		/* if it is an !endwhile directive, record the spot... */
		if(dirnum == DENDWHILE) {
			WHBLOCK* whtemp;	// Temporary ptr to a WHBLOCK

			if(scanner == NULL) {
				abortmsg = "!endwhile with no preceding !while";
				goto eabort;
			}
			/* move top records from the scanner list to the
			 *  whlist until we have moved all BREAK/CONTINUE records
			 * and one !while record
			 */
			do {

				scanner->w_end = lp;
				whtemp = pi->pri_whlist;
				pi->pri_whlist = scanner;
				scanner = scanner->w_next;
				pi->pri_whlist->w_next = whtemp;
			} while (pi->pri_whlist->w_type == BTBREAK || pi->pri_whlist->w_type == BTCONTINUE);

			// Patch up any !continue so that w_end points to where the !while begins:
			for(whtemp = pi->pri_whlist->w_next; whtemp && whtemp->w_type != BTWHILE; whtemp = whtemp->w_next)
				if(whtemp->w_type == BTCONTINUE)
					whtemp->w_end = pi->pri_whlist->w_begin;
		}
	}

	/* while and endwhile should match! */
	if(scanner != NULL) {
		abortmsg = "!while with no matching !endwhile";
		goto eabort;
	}

	/* Remember we need room for the procedure arguments among the locals.
	 */
	pi->pri_numlocals += pi->pri_numargs;

	pi->pri_scanned = TRUE;

	if(0) {
eabort:		/* Exit the current function with a failure - set status FALSE
		 * Show abortmsg along with buffer & current line info
		 */
		status = FALSE;

		errormesg(abortmsg, bp, lp);
	}

	return(status);
}

/* Execute the contents of the buffer pointed to by the passed bp.
 *
 *	Directives start with a "!" and include:
 *
 *	!endm		End a macro
 *	!if (cond)	conditional execution
 *	!else
 *	!endif
 *	!return [val]	Return (terminating current macro)
 *			If there is a val it is evaluated & returned
 *	!goto <label>	Jump to a label in the current macro
 *	!force		Force macro to continue...even if command fails
 *	!while (cond)	Execute a loop if the condition is true
 *	!endwhile
 *	!break
 *	!continue
 *	
 *	Line Labels begin with a "*" as the first nonblank char, like:
 *
 *	*LBL01
 *
 * The checkage() problem is taken care of - the renew routine will not touch
 * an executing buffer. It is too hard to stop within here.
 * Likewise the buffer cannot be zapped.
 *
 * bp	buffer to execute
 * val	the value that might be returned, should be initialised to V_UNDEF -
 *	which it will be if if nothing is returned. The type might be changed.
 *
 *
 * Save the current execText/execPos and restore on return. This is the
 * equivalent of saving/restoring program counter in a machine level program.
 *
 * This can be recursive via &call
 *
 * Valgrind will report a memory leak at the valGet() below if the buffer being
 * executed calls exit-emacs - ie when this function does not not return, so no
 * change to tidy up.
 *
 * TODO utf8
 *
 * Return TODO
 *	TRUE   - OK
 *	FALSE  - nothing entered at prompt
 *	ABORT  - user aborts the prompt - C-g
 *	FAILED - Out of memory & similar
 */
int	dobuf(BUFFER* bp, VAL** retval)
{
	register int status;	/* status return */
	register LINE* lp;	/* pointer to line to execute */
	LINE*	mp;		/* Macro line storage temp */
	int	dirnum;		/* directive index */
	LineLn	linlen;		/* length of line to execute */
	int	force;		/* force TRUE result? */
	Window*	wp;		/* ptr to windows to scan */
	WHBLOCK* whtemp;	/* temporary ptr to a WHBLOCK */
	VAL*	tokn = NULL;	/* Tokens in here */
	UTABLE*	ut = NULL;	/* new local user variable table */
	int	skipflag = FALSE; /* are we skipping debugging a function? - not yet */
	PARG*	cur_arg;	/* current argument being filled */
	int	cur_index;	/* index into current user table */
	int	whileCont;	// TRUE if continuing a !while with !endwhile or !continue
	char*	abortmsg;	// Message to give to errormsg()
	int	execlevel;	// If 0 run commands

	abortmsg = "Error";	// Default - OK if there is a prior error

	lp = (LINE*)NULL;	// In case of early error (see eabort)

	/* Save globals execText/execPos which store where we are looking/executing
	 * in a TEXT (LINE really).
	 * Needed for UFCALL - which might be part way down a line and will
	 * want to evaluate the rest of the line.
	 * No other calls need this as they are standalone or the next thing is to get the next line.
	 * Restore upon return from here and when assigning to procedure arguments.
	 */
	TEXT* saveExecl = execText;
	LineLn savePos = execPos;

	status = TRUE;

	/* Clear !if level flags/while ptr
	 *
	 * execlevel needs some explanation.
	 * It records nesting of !if and if the current line should be executed.
	 * Simply: if 0 execute the line, non 0 do not.
	 * !else/!elseif look more closely as they need to know if a previous
	 * !if/!elseif has set execute (ie execlevel == 0), if not then they might do so.
	 *   0 - execute, !else/!elseif set non execute and ensure that no subsequent
	 *	 will set execute by setting to 2
	 *   1 - no execute,
	 *       !else set to 0 as no prior !if/!elseif has run
	 *	 !elseif runs condition & sets to 0 if the conidition was TRUE; else leaves it set at 1
	 *
	 * !if - if != 0 then += 2 - this makes the nested !if/!elseif/!else not run
	 * !endif -= 2, not allow negative
	 */
	execlevel = 0;

	if(BuildProcInfo(bp) != TRUE)
		return(FAILED);

	/* Flag that we are executing the buffer */
	bp->b_exec++;

	/* let the first command inherit the flags from the last one..*/
	thisflag = lastflag;

	/* allocate a local user variable table */
	ut = (UTABLE*)malloc(sizeof(UTABLE) + (size_t)bp->b_procinfo->pri_numlocals * sizeof(UVAR));
	if(ut == (UTABLE*)NULL) {
		abortmsg = "Out of memory allocating locals";
		goto eabort;
	}
	ut->ut_next = uv_head;
	ut->ut_size = bp->b_procinfo->pri_numlocals;
	ut->ut_bufp = bp;
	uv_init(ut);
	// Push ut_head after arguments are processed - see later
	// uv_head = ut;

	// This can be reported as a memory leak by valgrind, see above
	tokn = valGet(V_STR);

	/* And evaluate the arguments passed, placing them in the local
	 * variable table.
	 * 
	 * Need to temp restore execText/execPos to values when dobuf() was
	 * called so that we look at the line that invoked this procedure, not
	 * lines within this procedure/buffer.
	 */
	cur_index = 0;
	cur_arg = bp->b_procinfo->pri_args;
	execText = saveExecl;
	execPos = savePos;

	int	argIsArray = FALSE;

	/* Loop over the arguments: create a local variable; assign a value using svar.
	 * Complication is when we find an array: create an empty array variable, push
	 * any/all subsequent values.
	 */
	while(cur_arg != (PARG *)NULL) {
		VAL*	ap = valGet(V_ANY);

		// mlwrite("%#arg '%s' line='%s'", cur_arg->pa_name, tgetcaddr(execText, execPos));

		// Give this argument a name - not the '%':
		strcpy(ut->ut_uv[cur_index].u_name, cur_arg->pa_name + 1);

		// The argument could be an array.
		// We find out as the '%' will have been replaced with '['
		if(cur_arg->pa_name[0] == '[')
			argIsArray = TRUE;

		// The local variable will have been created UNDEF - change to ARRAY if needed
		if(argIsArray)
			valChangeType(&ut->ut_uv[cur_index].u_value, V_ARRAY);

		while(1) {
			/* This is a loop as if the argument is an array we just keep
			 * asking for values until nothing comes back.
			 * Almost always 'asking' means next argument on command line in a macro
			 */
			/* ask for argument names */

			if((status = nextarg("Argument: ", &ap, MLSTRING | (argIsArray ? MLEMPTYOK : 0), bmlValHistp)) != TRUE) {
				valRelease(ap);

				// Nothing left ? OK: if array
				if(argIsArray && status == FALSE)
					goto endArgs;

				abortmsg = "Missing procedure argument";
				goto eabort;
			}

			// Not an array - drop through to setting value
			if( ! argIsArray)
				break;

			// Array: push this value onto the array end
			valArrayPush(ut->ut_uv[cur_index].u_value, ap);
		}

		// Set value to non array variable
		valCopy(&ut->ut_uv[cur_index].u_value, ap);
		valRelease(ap);

		/* on to the next argument */
		cur_arg = cur_arg->pa_next;
		cur_index++;
	}
endArgs:;

	// Make procedure arguments visible:
	uv_head = ut;

	// We might have eaten part of the line that invoked this.
	// Note what we have consumed
	savePos = execPos;

	whileCont = FALSE;

	/* starting at the beginning of the buffer */
	lp = lforw(bp->b_linep);
	while(lp != bp->b_linep) {
		if(IntSignalReceived) {
			abortmsg = "Received INTerrupt signal";
			goto eabort;
		}

		int	savEarlyLog;	// Save $early-log for !force

		execText = &ltxt(lp);
		execPos = 0;

		// mlwrite("%#%s line '%t'", (mstore ? "store" : "run"), execText);

		// Ignore if in a help section
		if((lnotes(lp) & MAC_NOTE_HELP) != 0)
			goto onward;

		/* Trim leading whitespace */
		while(tgetch(execText, execPos) == ' ' || tgetch(execText, execPos) == '\t')
			execPos++;

		/* Dump comments and blank lines */
		if(execPos >= tused(execText) || tgetch(execText, execPos) == '#' || tgetch(execText, execPos) == ';')
			goto onward;

		/* Only do this if we are debugging */
		if(macbug && !mstore && (execlevel == 0))
			if(debug(bp, execText, execPos, &skipflag) == FALSE) {
				abortmsg = "[Macro aborted]";
				goto eabort;
			}

		// Is there a ! directive on this line ?
		dirnum = lnotes(lp) & MAC_NOTE_DIRMASK;

		/* service only the !endm macro here */
		if(dirnum == DENDM) {
			mstore = FALSE;
			bstore = NULL;
			goto onward;
		}

		/* if macro store is on, just salt this away */
		if(mstore) {
			/* allocate the space for the line.
			 * Zero length if help - no need to store that text,
			 * but need to store the line else error reports on
			 * line numbers are wrong.
// Hmmm but we ignore above?
			 */
			linlen = tused(execText) - execPos;
			if((lnotes(lp) & MAC_NOTE_HELP) != 0)
				linlen = 0;
			if((mp = lalloc(linlen)) == NULL) {
				abortmsg = "Out of memory while storing macro";
				goto eabort;
			}

			/* copy the text into the new line */
			if((lnotes(lp) & MAC_NOTE_HELP) == 0)
				lcpys(mp, tgetcaddr(execText, execPos));

			/* attach the line to the end of the buffer */
			lsetforw(lback(bstore->b_linep), mp);
			lsetback(mp, lback(bstore->b_linep));
			lsetback(bstore->b_linep, mp);
			lsetforw(mp, bstore->b_linep);

			// Copy Help/!directive status
			lsetNotes(mp, lnotes(lp));

			bstore->b_numLines++;
			goto onward;
		}

		force = FALSE;

		// Dump labels & directives
		if(tgetch(execText, execPos) == '*')
			goto onward;

		/* now, execute directives */
		if(dirnum != DNOTHING) {
			// mlwrite("%#dir='%s'", dname[dirnum - 1]);

			/* skip past the directive */
			while(execPos < tused(execText) && tgetch(execText, execPos) != ' ' && tgetch(execText, execPos) != '\t')
				execPos++;

			switch(dirnum) {
			case DIF:	/* !if directive */
				if(execlevel == 0) {
					/* grab the value of the logical exp */
					if(macarg(&tokn) != TRUE) {
						abortmsg = "Missing condition after !if";
						goto eabort;
					}

					if(valToBool(tokn) == FALSE)
						execlevel = 1;	// Set not execute until !endif - or maybe !else !elseif

				} else	// Level not 0 - increment by 2
					execlevel += 2;

				goto onward;

			case DELSEIF:	/* !elseif directive */
				if(execlevel == 0) {
					// Currently executing code
					execlevel = 2;		// Stop execution until !endif
				}
				else if(execlevel == 1) {
					/* grab the value of the logical exp */
					if(macarg(&tokn) != TRUE) {
						abortmsg = "Missing condition after !elseif";
						goto eexec;
					}

					if(valToBool(tokn) == TRUE)
						execlevel = 0;
				}
				// Implicit: else execlevel != 0 or 1 -- do nothing
				goto onward;

			case DELSE:	/* !else directive */
				if(execlevel == 1)	// !else to !if/elseif that evaluated FALSE ?
					execlevel = 0;	// Start execution
				else if(execlevel == 0)	// !else to !if/elseif that evaluated TRUE ?
					execlevel = 2;	// Stop execution
				goto onward;

			case DENDIF:	/* !endif directive */
				// Reverse the +1 of our !if
				if((execlevel -= 2) < 0)
					execlevel = 0;
				goto onward;

			case DWHILE:	/* !while directive
					 * This has two forms:
					 * !while expression
					 *	Continue as long as expression is TRUE
					 * !while var array
					 *	Iterate var over every element in array
					 */
				/* grab the value of the logical exp */
				if(execlevel == 0) {
					LineLn	spos = execPos;

					/* This is nasty ...
					 * If we don't know what form the !while takes, assume
					 * that it is the first. If that turns out to be a mistake,
					 * well - we will only do so once.
					 */
					int	maret = TRUE;
					if((lnotes(lp) & MAC_NOTE_WHILE_ARRAY) == 0) {
						gtusrCkReport = FALSE;
						maret = macarg(&tokn);
						gtusrCkReport = TRUE;
					}

					/* skip past white space */
					while(tgetch(execText, execPos) == ' ' || tgetch(execText, execPos) == '\t')
						execPos++;

					// There is still something on the line ... must be the array
					if(execPos < tused(execText)) {
						if((lnotes(lp) & MAC_NOTE_WHILE_ARRAY) == 0) {
							/* Did not know that this was the array form.
							 * tokn will have been converted to V_ERROR, we need a V_STR
							 * Make a note to avoid the mistake next time.
							 * NB don't lose other line notes.
							 */
							valRelease(tokn);
							tokn = valGet(V_STR);
							lsetNotes(lp, MAC_NOTE_WHILE_ARRAY | lnotes(lp));
						}

						/* Iterate over an array.
						 * 
						 * macarg() above got a value - all evaluated
						 * What we need is something that we can assign to - a variable.
						 * Go back to before macarg() and rescan - yes nasty, but until
						 * we did the above we did not know that there was anything after the
						 * variable name.
						 * This is stolen out of setvar().
						 */
						VDESC	vd;	// variable num/type

						execPos = spos;	// Back to before macarg()

						// Get the variable:
						TEXT*	var = playVar;
						if(token(var) == FALSE ||
						   findvar(var, &vd, NVSIZE + 1, VT_GLOBAL) == FALSE ||
						   vd.v_type != TKVAR) {
							mlwrite(BadVar, var);
							abortmsg = "Bad !while";
							goto eabort;
						}
						VAL**	varp = &vd.v_ut->ut_uv[vd.v_num].u_value;


						/* jump down to the endwhile */
						/* find the right while loop */
						whtemp = bp->b_procinfo->pri_whlist;
						while(whtemp) {
							if(whtemp->w_begin == lp)
								break;
							whtemp = whtemp->w_next;
						}

						if(whtemp == NULL) {
							abortmsg = "Internal While loop error";
							goto eabort;
						}

						if(whileCont == TRUE) {
							/* This is not the first time that this !while has executed
							 * Continue and get the next element in the array
							 */
							if(whtemp->w_array == (VAL*)NULL) {
								abortmsg = "Array is NULL in !while";
								goto eabort;
							}

							// End of array ?
							if(++whtemp->w_index >= whtemp->w_array->v_v.v_array.a_used) {
								valRelease(whtemp->w_array);
								whtemp->w_array = (VAL*)NULL;
								// Drop through to !break
							} else {
								// Next element
								valCopy(varp, whtemp->w_array->v_v.v_array.a_elems[whtemp->w_index]);
								goto onward;
							}
						} else {
							// Set up new loop

							if(whtemp->w_array != NULL) {	// Prob left over after goto
								valRelease(whtemp->w_array);
								whtemp->w_array = (VAL*)NULL;
							}

							// Get the array to loop over
							VAL*	whileArray = valGet(V_ARRAY);
							if(macarg(&whileArray) != TRUE) {
								mlwrite("%EExpecting array got %s", vtypename(whileArray));
								valRelease(whileArray);
								abortmsg = "Bad argument to !while";
								goto eabort;
							}

							// Grab the array
							whtemp->w_array = whileArray;

							// Only if it has a first element
							if(whtemp->w_array->v_v.v_array.a_used > 0) {
								// Grab first element
								whtemp->w_index = 0;
								valCopy(varp, whtemp->w_array->v_v.v_array.a_elems[0]);

								goto onward;
							}

							// Array is empty ... drop through
						}

					} else {
						// Is the expression true ?

						// If macarg() failed - oh, ever so long ago:
						if(maret != TRUE)
							goto eexec;

						// Evaluate the single argument as boolean
						if(valToBool(tokn) == TRUE)
							goto onward;
					}
				}
				/* drop down and act just like !break */

			case DBREAK:	/* !break directive */
			case DCONTINUE:	/* !continue directive */
				if(dirnum != DWHILE && execlevel)
					goto onward;

				/* jump down to the endwhile */
				/* find the right while loop */
				whtemp = bp->b_procinfo->pri_whlist;
				while(whtemp) {
					if(whtemp->w_begin == lp)
						break;
					whtemp = whtemp->w_next;
				}

				if(whtemp == NULL) {
					abortmsg = "Internal While loop error";
					goto eabort;
				}

				/* reset the line pointer back.. */
				if(dirnum == DCONTINUE) {
					lp = lback(whtemp->w_end);
					whileCont = TRUE;
					goto onward_while;
				}

				lp = whtemp->w_end;

				goto onward;

			case DENDWHILE:	/* !endwhile directive */
				if(execlevel) {
					goto onward;
				} else {
					/* Find the right while loop.
					 * Check that we have the correct end line.
					 * Cannot do this with !continue, so assume that it is corect,
					 * we'll still get an error if this is a !continue outside a loop.
					 */
					whtemp = bp->b_procinfo->pri_whlist;
					while(whtemp) {
						if(whtemp->w_type == BTWHILE &&
						    whtemp->w_end == lp)
							break;
						whtemp = whtemp->w_next;
					}

					if(whtemp == NULL) {
						abortmsg = "Internal While loop error at !endwhile";
						goto eabort;
					}

					/* reset the line pointer back.. */
					lp = lback(whtemp->w_begin);
					whileCont = TRUE;
					goto onward_while;
				}

			case DGOTO:	/* !goto directive */
				/* .....only if we are currently executing */
				if(execlevel == 0) {
					golist*	gop;	// Goto label probe

					/* grab label to jump to */
					if(macarg(&tokn) != TRUE) {
						abortmsg = "Missing label after !goto";
						goto eabort;
					}

					// Look for it in the list of labels:
					for(gop = bp->b_procinfo->pri_gotos; gop != (golist*)NULL; gop = gop->gl_next)
						if(strcmp(vtext(tokn), gop->gl_name) == 0) {
							lp = gop->gl_line;
							goto onward;
						}

					mlwrite("%ENo such label '%v'", tokn);
					goto eabort;
				}

				goto onward;

			case DRETURN:	/* !return directive */
				if(execlevel == 0) {
					/* If something follows the !return evaluate it.
					 * Put into retval - this will be used by &call to provide a value.
					 *
					 * macarg either succeeds or it does not.
					 * Since *retval is ANY macarg does not print an error if there is nothing
					 * after the RETURN - although *retval ends up as an error ...
					 * but if nothing looks at it ...
					 */
					macarg(retval);
					// mlwrite("%E!return type=%s int=%d @retval=%p val=%v i=%d", vtypename(*retval), vint(*retval), *retval, *retval);

					goto eexec;
				}
				goto onward;

			case DFORCE:	/* !force directive */
				force = TRUE;

			}
		}

		// If on the FALSE side of an !if (or similar) jump to the next line
		if(execlevel)
			goto onward;

		/* Execute the statement.
		 * Note the value of $early-log
		 */
		savEarlyLog = EarlyLog;

		// mlwrite("%#dobuf calling docmd('%t'", execText);

		status = docmd();

		if(force) {
			/* Force the status to always succeed.
			 * An error could cause logging to start in a macro
			 * started from the command line - undo that - reset
			 * $early-log to what it was before the command was run.
			 */
			EarlyLog = savEarlyLog;
			status = TRUE;
		}

		/* check for a command error */
		if(status != TRUE) {
			/* look if buffer is showing */
			wp = wheadp;
			while (wp != NULL) {
				if(wp->w_bufp == bp) {
					/* and point it */
					wp->w_dotp = lp;
					wp->w_doto = 0;
					wp->w_flag |= WFHARD;
				}
				wp = wp->w_wndp;
			}
			/* in any case set the buffer . */
			bp->b_dotp = lp;
			bp->b_doto = 0;

			abortmsg = "Macro Failed";

			goto eabort_ns;
		}

		/* Set that the last statement was not !endwhile or !continue
		 * This is so that a !while knows to reset any array second value
		 */
onward:		whileCont = FALSE;

onward_while:
		/* on to the next line */
		lp = lforw(lp);
		if(skipflag)
			macbug = TRUE;
	}

	if(0) {
eexec:		/* exit the current function - set status TRUE */
		status = TRUE;
	}

	if(0) {
eabort:		/* Exit the current function with a failure - set status FALSE
		 */
		status = FALSE;

eabort_ns:	/* Exit the current function with a failure - leave status as it is
		 * Show abortmsg along with buffer & current line info
		 */
		errormesg(abortmsg, bp, lp);
	}

	execlevel = 0;
	bp->b_exec--;

	/* discard the local user variable table.
	 * If one has been allocated.
	 */
	if(ut != NULL) {
		uv_head = ut->ut_next;
		uv_clean(ut);
		free(ut);
	}

	valRelease(tokn);
	// Restore execText/savePos so that we continue to execute where this was called from
	execText = saveExecl;
	execPos = savePos;

	return(status);
}

/* errormesg:	display a macro execution error along with the buffer and
 *		line currently being executed
 *
 * mesg		error message to display
 * bp		buffer error occured in
 * lp		line error occured in - if NULL show first line
 */
void	errormesg(const char* mesg, const BUFFER* bp, const LINE* lp)
{
	exec_error = TRUE;
	if(lp == (LINE*)NULL)
		lp = lforw(bp->b_linep);

	// Unfortunately in C you cannot go 'const ...'
	mlforce("%E""%F""%s in <%s> at line %D : '%l'", (char*)mesg, (BYTE*)bp->b_bname, getlinenum(bp, lp), lp);
}

/*		Interactive debugger
 *
 * if $debug == TRUE, The interactive debugger is invoked
 * commands are listed out with the ? key
 *
 * bp		buffer to execute
 * eline	TEXT of line to debug
 * skipflag	are we skipping debugging?
 *
 * Return:
 *	TRUE  worked
 *	FALSE user aborted with C-g
 */
int	debug(BUFFER* bp, TEXT* eline, LineLn elineOff, int* skipflag)
{
	register int	oldcmd;			/* original command display flag */
	register int	oldinp;			/* original connamd input flag */
	register int	oldstatus;		/* status of last command */
	register int	c;			/* temp character */
	LineLn		outUsed;		// Debug length used

	tsetempty(trackLine);			// Clear what we track

dbuild:	/* Build the information line to be presented to the user */

	tcpys(outline, "<<<");

	/* display the tracked expression */
	if (tused(trackLine) != 0) {
		oldstatus = cmdstatus;
		execText = trackLine;
		execPos = 0;
		docmd();
		cmdstatus = oldstatus;
		printToText(outline, "%A[=%v]", gtusrCk("track"));
	}

	/* Append:
	 * debug macro name
	 * the line
	 */
	printToText(outline, "%A""%s:%s>>>", bp->b_bname, tgetcaddr(eline, elineOff));
	outUsed = tused(outline);


	/* write out the debug line */
dinput:
	mlforce("%t", outline);

	// If [List] is displayed, update it - so that variables are up to date
	if(blistp->b_nwnd)
		listbuffers(FALSE, 0);

	update(TRUE);

	// Remove extra stuff that might be appended
	tsetused(outline, outUsed);

	/* and get the keystroke */
	c = tgetc(MGC_NONE, 0);

	if(c == CharEscape)
		// META key turns off debugging
		macbug = FALSE;
	else if(c == abortc) {
		// Abort the execution of the macro
		return(FALSE);
	} else switch(c) {

		case '?': /* list commands */
			tcats(outline,
				" <SP>exec, <s>how no assign, <META> stop debug, (C-g)abort, (m)ore");
			goto dinput;

		case 'm': /* more commands listed */
			tcats(outline,
				" (e)val exp, (t)rack exp, (c)ommand line, (x)ecute command");
			goto dinput;

		case 'c': /* execute statement */
		case 'x': /* execute extended command */
			oldcmd = discmd;
			discmd = TRUE;
			oldinp = disinp;
			disinp = TRUE;
			oldstatus = cmdstatus;
			clexec = FALSE;		// Else nextarg() won't prompt
			if(c == 'c')
				execcmd(FALSE, 1);
			else
				namedcmd(FALSE, 1);	// Arg n MUST be 1
			cmdstatus = oldstatus;
			discmd = oldcmd;
			disinp = oldinp;
			goto dbuild;

		case 'e': /* evaluate expresion */
			oldinp = disinp;
			disinp = TRUE;
			// mlprompt() OK as this will always be interactive
			mlprompt("Exp:", MLSTRING, bmlDebug);
			disinp = oldinp;
			oldstatus = cmdstatus;
			tcpys(playText, "set %track ");
			tcatl(playText, bmlDebug->b_dotp);

			execText = playText; // Hmmm overwritable ?
			execPos = 0;
			docmd();

			cmdstatus = oldstatus;
			mlforce(" = [%v]", gtusrCk("track"));
			c = tgetc(MGC_NONE, 0);
			goto dinput;

		case 't': /* track expresion */
			oldinp = disinp;
			disinp = TRUE;
			// mlprompt() OK as this will always be interactive
			mlprompt("Exp:", MLSTRING, bmlDebug);
			disinp = oldinp;
			tcpys(trackLine, "set %track ");
			tcatl(playText, bmlDebug->b_dotp);
			goto dbuild;

		case 's': /* execute a function */
			*skipflag = TRUE;
			macbug = FALSE;
			break;

		case ' ': /* execute a statement */
			break;

		default: /* illegal command */
			TTbeep();
			goto dbuild;
	}
	return(TRUE);
}

/* Execute a series of commands in a file.
 *
 * Default count 1. Must not be < 0
 *
 * Command:  execute-file
 * Bound to: unbound
 *
 * Return:
 *	what dofile() returns
 *	FALSE file not found
 *	      nothing entered at prompt
 *	ABORT user aborts the prompt - C-g
 */
int	execfile(int f, Cnt n)
{
	int	status = TRUE;	/* return status of name query */
	BYTE*	path;

	if((status = mlprompt("File to execute: ", MLXFILE, bmlFileHistp)) != TRUE)
		return(status);

	/* Look up the path for the file */
	if( !findFile(ltext(bmlFileHistp->b_dotp), fT_Macro)) {
		/* if it isn't around */
		mlwrite("%EFile %s cannot be found", ltext(bmlFileHistp->b_dotp));
		return(FALSE);
	}

	path = strmake_t(ttext(foundFiles));

	/* otherwise, execute it */
	while(--n >= 0)
		if((status = dofile(path)) != TRUE)
			break;

	free(path);

	return(status);
}

/*	dofile:	yank a file into a buffer and execute it
 *		if there are no errors, delete the buffer on exit
 *
 * fname	file name to execute
 *
 * Return:
 *	FALSE File not found
 *		what does readin() return TODO
 *	dobuf() return TODO
 */
int	dofile(BYTE* fname)
{
	register BUFFER*bp;	/* buffer to place file to exeute */
	register int	status;	/* results of various calls */

	if((bp = bfindFile((BYTE*)NULL, fname)) == NULL)
		return(FALSE);

	VAL* retval = valGet(V_ANY);

	/* go execute it! */
	status = dobuf(bp, &retval);

	valRelease(retval);

	/* if not displayed, remove the now unneeded buffer and exit */
	if(status == TRUE && bp->b_nwnd == 0)
		zotbuf(bp, TRUE);

	return(status);
}

/* Execute the contents of a numbered buffer
 *
 * bufnum	number of buffer to execute
 *
 * Execute the procedure count times, once if count unset.
 *
 * Variables $cmd-flag and $cmd-count are set to f and n.
 *
 * Default count unset
 *
 * Command:  execute-macro-1, etc
 * Bound to: unbound
 *
 * Return:
 *	TRUE  buffer executed, commands returned TRUE
 *	FALSE buffer not found
 */
int	cbuf(int f, Cnt n, int bufnum)
{
	BUFFER*	bp;		/* ptr to buffer to execute */
	int	status;		/* status return */
	BYTE	bname[BUFFER_NAME_SIZE];	/* name of buffer to use */

	/* Construct the macro buffer name */
	sprintf(bname, "[Macro %2.2d]", bufnum);

	/* find the pointer to that buffer */
	if((bp = bfind(bname, BFNNONE, 0)) == NULL) {
		mlwrite("%EMacro %d not defined", n);
		return(FALSE);
	}

	// Set before games below:
	CmdFlag = f;
	CmdCount = n;
	VAL* retval = valGet(V_ANY);

	/* And now execute it as asked,
	 * once if f is not set
	 */
	do {
		if((status = dobuf(bp, &retval)) != TRUE)
			break;

		if( !f)	// No flag, execute once only
			break;

	} while(--CmdCount > 0);

	valRelease(retval);

	return(status);
}

/* Execute contents of buffer [Macro 01]
 *
 * Default count 1. Must not be < 0
 *
 * Command:  execute-macro-1
 * Bound to: unbound
 *
 * Return:
 *	TRUE  buffer executed, commands returned TRUE
 *	FALSE buffer not found
 */
int	cbuf1(int f, Cnt n)
{
	return(cbuf(f, n, 1));
}

/* Command:  execute-macro-2
 * Bound to: unbound
 */
int	cbuf2(int f, Cnt n)
{
	return(cbuf(f, n, 2));
}

/* Command:  execute-macro-3
 * Bound to: unbound
 */
int	cbuf3(int f, Cnt n)
{
	return(cbuf(f, n, 3));
}

/* Command:  execute-macro-4
 * Bound to: unbound
 */
int	cbuf4(int f, Cnt n)
{
	return(cbuf(f, n, 4));
}

/* Command:  execute-macro-5
 * Bound to: unbound
 */
int	cbuf5(int f, Cnt n)
{
	return(cbuf(f, n, 5));
}

/* Command:  execute-macro-6
 * Bound to: unbound
 */
int	cbuf6(int f, Cnt n)
{
	return(cbuf(f, n, 6));
}

/* Command:  execute-macro-7
 * Bound to: unbound
 */
int	cbuf7(int f, Cnt n)
{
	return(cbuf(f, n, 7));
}

/* Command:  execute-macro-8
 * Bound to: unbound
 */
int	cbuf8(int f, Cnt n)
{
	return(cbuf(f, n, 8));
}

/* Command:  execute-macro-9
 * Bound to: unbound
 */
int	cbuf9(int f, Cnt n)
{
	return(cbuf(f, n, 9));
}

/* Command:  execute-macro-10
 * Bound to: unbound
 */
int	cbuf10(int f, Cnt n)
{
	return(cbuf(f, n, 10));
}

/* Command:  execute-macro-11
 * Bound to: unbound
 */
int	cbuf11(int f, Cnt n)
{
	return(cbuf(f, n, 11));
}

/* Command:  execute-macro-12
 * Bound to: unbound
 */
int	cbuf12(int f, Cnt n)
{
	return(cbuf(f, n, 12));
}

/* Command:  execute-macro-13
 * Bound to: unbound
 */
int	cbuf13(int f, Cnt n)
{
	return(cbuf(f, n, 13));
}

/* Command:  execute-macro-14
 * Bound to: unbound
 */
int	cbuf14(int f, Cnt n)
{
	return(cbuf(f, n, 14));
}

/* Command:  execute-macro-15
 * Bound to: unbound
 */
int	cbuf15(int f, Cnt n)
{
	return(cbuf(f, n, 15));
}

/* Command:  execute-macro-16
 * Bound to: unbound
 */
int	cbuf16(int f, Cnt n)
{
	return(cbuf(f, n, 16));
}

/* Command:  execute-macro-17
 * Bound to: unbound
 */
int	cbuf17(int f, Cnt n)
{
	return(cbuf(f, n, 17));
}

/* Command:  execute-macro-18
 * Bound to: unbound
 */
int	cbuf18(int f, Cnt n)
{
	return(cbuf(f, n, 18));
}

/* Command:  execute-macro-19
 * Bound to: unbound
 */
int	cbuf19(int f, Cnt n)
{
	return(cbuf(f, n, 19));
}

/* Command:  execute-macro-20
 * Bound to: unbound
 */
int	cbuf20(int f, Cnt n)
{
	return(cbuf(f, n, 20));
}

/* Command:  execute-macro-21
 * Bound to: unbound
 */
int	cbuf21(int f, Cnt n)
{
	return(cbuf(f, n, 21));
}

/* Command:  execute-macro-22
 * Bound to: unbound
 */
int	cbuf22(int f, Cnt n)
{
	return(cbuf(f, n, 22));
}

/* Command:  execute-macro-23
 * Bound to: unbound
 */
int	cbuf23(int f, Cnt n)
{
	return(cbuf(f, n, 23));
}

/* Command:  execute-macro-24
 * Bound to: unbound
 */
int	cbuf24(int f, Cnt n)
{
	return(cbuf(f, n, 24));
}

/* Command:  execute-macro-25
 * Bound to: unbound
 */
int	cbuf25(int f, Cnt n)
{
	return(cbuf(f, n, 25));
}

/* Command:  execute-macro-26
 * Bound to: unbound
 */
int	cbuf26(int f, Cnt n)
{
	return(cbuf(f, n, 26));
}

/* Command:  execute-macro-27
 * Bound to: unbound
 */
int	cbuf27(int f, Cnt n)
{
	return(cbuf(f, n, 27));
}

/* Command:  execute-macro-28
 * Bound to: unbound
 */
int	cbuf28(int f, Cnt n)
{
	return(cbuf(f, n, 28));
}

/* Command:  execute-macro-29
 * Bound to: unbound
 */
int	cbuf29(int f, Cnt n)
{
	return(cbuf(f, n, 29));
}

/* Command:  execute-macro-30
 * Bound to: unbound
 */
int	cbuf30(int f, Cnt n)
{
	return(cbuf(f, n, 30));
}

/* Command:  execute-macro-31
 * Bound to: unbound
 */
int	cbuf31(int f, Cnt n)
{
	return(cbuf(f, n, 31));
}

/* Command:  execute-macro-32
 * Bound to: unbound
 */
int	cbuf32(int f, Cnt n)
{
	return(cbuf(f, n, 32));
}

/* Command:  execute-macro-33
 * Bound to: unbound
 */
int	cbuf33(int f, Cnt n)
{
	return(cbuf(f, n, 33));
}

/* Command:  execute-macro-34
 * Bound to: unbound
 */
int	cbuf34(int f, Cnt n)
{
	return(cbuf(f, n, 34));
}

/* Command:  execute-macro-35
 * Bound to: unbound
 */
int	cbuf35(int f, Cnt n)
{
	return(cbuf(f, n, 35));
}

/* Command:  execute-macro-36
 * Bound to: unbound
 */
int	cbuf36(int f, Cnt n)
{
	return(cbuf(f, n, 36));
}

/* Command:  execute-macro-37
 * Bound to: unbound
 */
int	cbuf37(int f, Cnt n)
{
	return(cbuf(f, n, 37));
}

/* Command:  execute-macro-38
 * Bound to: unbound
 */
int	cbuf38(int f, Cnt n)
{
	return(cbuf(f, n, 38));
}

/* Command:  execute-macro-39
 * Bound to: unbound
 */
int	cbuf39(int f, Cnt n)
{
	return(cbuf(f, n, 39));
}

/* Command:  execute-macro-40
 * Bound to: unbound
 */
int	cbuf40(int f, Cnt n)
{
	return(cbuf(f, n, 40));
}

/* end */
