/*
 * ADDW SCCS: @(#)expand.c	1.53 09/22/20 22:44:19
 */

/*
 * EXPAND:	File name expansion under UNIX		afb 25/9/86
 *		buffer name expansion			    25/9/86
 *		previous shell-command expansion	    25/9/86
 *		mode name expansion			    24/6/88
 *		window option expansion (hack on mode exp)  Feb 2016 addw
 *		Emacs command name expansion		    Feb 2016 addw
 *		variable name expansion			    Oct 2017 addw
 *
 * The stuff in here is only ever called from mlreplyt() (at the moment
 * anyway).
 *
 * Perhaps this stuff should go into input.c with mlreplyt(), maybe
 * fnexpand() should go into file.c ? Well, its all here now.
 */

#include	"estruct.h"
#include	"utf8.h"
#include	"edef.h"

#if	SHOWMODREL
char	expndvers[] = "@(#)expand.c	1.53 09/22/20 22:44:19";
#endif

/* Return a pointer to the previous word in a LINE of NUL separated
 * words.
 * buf	points to the current position in the buffer
 * text	is the text containing the strings
 */
const BYTE*	prevword(const BYTE* buf, const TEXT* text)
{
	// Start of line ? go to the end
	if(buf <= ttext(text))
		buf = tgetcaddr(text, tused(text));

	// Back over the NUL ending the previous word
	if(buf > ttext(text))
		--buf;

	// Find the word start
	while(buf > ttext(text) && *--buf != '\0')
		;

	/* We will have ended on a NUL, so step forwards
	 * unless we stopped as we hit start of line
	 */
	if(buf != ttext(text))
		++buf;

	return(buf);
}

#if	UNIX
#include	<pwd.h>
#endif

/*
 * Expand twiddles in fname (csh style)
 *
 * The result will be NUL terminated - although the NUL will not be part of
 * the TEXT length, but there never the less.
 *
 * If an error occurs, return a pointer to an error string. return NULL
 * otherwise.
 *
 * If newlen points to NULL (which it may if we dont care about the new
 * length) then dont assign to it.
 *
 * Prompt tells us if we are interactive. If its NULL, then we're not.
 *
 * fileName might have a prompt at the start, it will have the NUL flag set.
 *
 * Return an error message, or NULL if it works
 *
 * The error message might be in playError
 */
BYTE*	exp_twiddle(TEXT* fileName, int promptLen)
{
#if	UNIX
	BYTE*	p;		// Play
	char*	star;		// Pointer to a '*'
	BYTE*	fname = tgetcaddr(fileName, promptLen);
	int	nameLen = (int)tused(fileName) - promptLen;
	int	slashIn2;	// '/' in 2nd position in fileName
	struct	passwd*	pw_ent;

	if(nameLen < 1 || fname[0] != '~')
		return("File name cannot be expanded");

	// It helps a lot to have a NUL terminated name
	// ensure that it is there
	// It will be removed just before return
	if(tsetNUL(fileName) == FALSE)
		return(OutOfMemory);

	/* This is the easy bit, where the user has typed "~/foo" and all
	 * we need is the home directory from the environment.
	 * Also: '~'  '~*'  '~/ *' (without spaces in that last one) 
	 */
	slashIn2 = fname[1] == '/';
	if(slashIn2 || fname[1] == '*' || nameLen == 1) {

		/* Maybe ought to use getpwuid() */
		if(home[0] == '\0')
			return("HOME not defined");

		// Insert at the start of the string, overwriting the '~/'
		strocpy(fname, fname + (nameLen > 1));
		tsetused(fileName, tused(fileName) - (1/* + (nameLen > 1)*/));	// 1 = '~', & maybe '/'

		if(tinss(fileName, promptLen, home) == FALSE)
			return(OutOfMemory);

		// If $HOME does not end with '/' & there was not one in fileName, insert one:
		if(home[strlen(home) - 1] != '/' &&
		   !slashIn2 &&
		   tinsc(fileName, (LineLn)(promptLen + (int)strlen(home)), '/') == FALSE)
			return(OutOfMemory);

		return(NULL);
	}

	/* This is the hard bit. The expansion is "~user".
	 * Stamp on any '/', but could have:
	 * ~user
	 * ~user/
	 * ~user/path
	 */
	if((p = strchr(fname, '/')) != (BYTE*)NULL)
		*p = '\0';

	/* If we see something like ~user*fred treat it as ~user/fred
	 * The point is that pressing TAB will result in a '*' being appended,
	 * want to stop silly errors 'Unknown user: user*'.
	 */
	if((star = strchr(fname, '*')) != (char*)NULL) {
		if(p)
			*p = '/';
		p = star;
		*p = '\0';
	}

	/*
	 * Find the passwd entry for "user"
	 */
	pw_ent = getpwnam(fname + 1);
	endpwent();

	if(pw_ent == (struct passwd*) NULL) {
		tcpys(playError, "Unknown user: ");
		tcats(playError, fname + 1);

		if(p)	// Repair
			*p = '/';

		return(ttext(playError));
	}

	if(p) {
		*p = '/';

		// Copy remains of the path over ~user/
		strocpy(fname, p);
		tsetused(fileName, tused(fileName) - (p - fname));
	} else {
		// Just a username
		tsetused(fileName, promptLen);
		tcats(fileName, "/");	// Append / for directory
	}

	// Insert at the start of the string:
	if(tinss(fileName, promptLen, pw_ent->pw_dir) == FALSE)
		return(OutOfMemory);

	return(NULL);

#else	/* ! UNIX */
	return(NULL);
#endif
}

/* Return a pointer to shell meta, int characters in "fname",
 * Return NULL if there are none.
 */
const BYTE*	sh_meta(const BYTE* fname)
{
	for(/*void*/; *fname; fname++)
		switch(*fname) {
		  case '?':
		  case '*':
		  case '$':
		  case '[':
		  case ']':
			return(fname);
		}

	return(NULL);
}

/* Expand the filename stored in wmessp->w_dotp" when running under UNIX (I have no
 * idea how to expand filenames under MS-DOS CP/M etc etc).
 *
 * This function is called from only two places in all of the code. This is
 * from mlreplyt() when "MLXFILE" (defined in estruct.h) is or'ed in with
 * "option".
 * 
 * One call is made when the user has just typed return, and metacharacters
 * are expanded "once off" and the first match is returned. This is refered
 * to in the following comments as "non-interactive" mode.
 * 
 * The second call is made whenever the user types a space. Since a space is
 * not a very reasonable character to have in filenames, it is used as an
 * "expansion" character, ie the user has typed in something containing shell
 * metacharacters and would like see what results. This is called "interactive"
 * mode.
 *
 * The routine uses the shell to expand metacharacters in much the same way
 * as "vi" does, this is alot easier and quicker than writing metacharacter
 * expansion code here. Expansions are stored in the buffer "cmd", which is
 * rather large, so as to cope with "very wild" wildcards.
 *
 * In "non-interactive" mode, the first match is copied from the cmd buffer
 * into buf and the function returns. In "interactive" mode, whenever the
 * user hits space, the next match is copied from the cmd buffer into buf.
 *
 * "dirn" is used to distinguish which mode (interactive or non-interactive)
 * we are in. If zero then we are non-interactive. If non zero then
 * interactive, -1 (BACKW) go backwards.
 *
 * dotp is safe in this function as wmessp will not be reread.
 *
 * prompt	Prompt at the start of the line
 *
 * Set errflag if there is some sort of error in expanding things (like
 * there were no expansions for the pattern, or the user didnt get the
 * twiddle right etc etc).
 *
 * Return values:
 * non-interactive mode:	0
 * interactive mode:		first character user typed which wasnt space
 */


KeyCode	fnexpand(const BYTE* prompt, int dirn, int* errflag)
{
#if	UNIX
	int		charsin = 0;	/* count of characters read	 */
	KeyCode		c;		/* character user types		 */
	const BYTE*	pcmd;		/* pointer to cmd		 */
	TEXT*		cmd;		/* buffer to stuff expansions into */
	const BYTE*	tp;		/* Temp BYTE pointer		 */
	TEXT*		base;		/* basename of buf - ie a directory */
	int		offset;		/* offset into buf		 */
	int		promptLen = (int)strlen(prompt);
	TEXT*		dotp = &ltxt(wmessp->w_dotp);
	BYTE*		fname;		/* The name that we are expanding*/
	BYTE		pipeBuf[1024];	/* Read from pipe into here 	 */
	const BYTE*	fileList;	/* List of file names in playText3 */

	*errflag = 0;		// Clear the error state

	// Should never happen:
	if(tused(dotp) <= promptLen)
		return(' ');

	// NUL terminate:
	tsetNUL(dotp);
	fname = tgetcaddr(dotp, promptLen);

	base = playDir;	// Workspace
	tsetused(base, 0);

	if(fname[0] == '~') {
		if((tp = exp_twiddle(dotp, promptLen)) != (BYTE*) NULL) {
			*errflag = 1;
			/*
			 * Since we can expand no further, if non interactive then
			 * return, while if interactive, sit mopping spaces.
			 */
			return(mlerr(tp, (char*) NULL));
		}

		fname = tgetcaddr(dotp, promptLen);	// Could have moved

		if(!sh_meta(fname)) {
			/* 
			 * What we have here is a twiddle with no other
			 * metacharacters. Redisplay the expanded string
			 * and if interactive, get another character.
			 */
			wmessp->w_doto = tused(dotp);
			if(dirn == 0)
				return(0);
			else {
				wmessp->w_flag |= WFHARD;
				return(tgetc(MGC_MAP, 0));
			}
		}
	}

	// Get the file name (after any prompt):
	fname = tgetcaddr(dotp, promptLen);

	if((tp = sh_meta(fname)) == (BYTE*) NULL) {
		/*
		 * There were not any shell metacharacters in the buffer. Make
		 * sure that there is a '/' character at the end if fname is
		 * a directory name.
		 */
		BYTE*	lastc = tgetcaddr(dotp, tused(dotp) - 1);

		if(*lastc != '\0' && isdir(fname) && *lastc != '/')
			tcats(dotp, "/");	// Append '/'

		/*
		 * If interactive, display expansion and mop spaces,
		 * since we cannot expand further.
		 *
		 * If non-interactive, return.
		 */
		if(dirn == 0)
			return(0);

		*errflag = 1;

		return(mlerr("no metacharacters", (char*) NULL));
	}
	offset = (int)(tp - fname);

	/*
	 * See if what we are trying to expand is a directory name followed by
	 * a metacharacter. If so, we can "cd" somewhere before doing the
	 * expansion.
	 */

	/*
	 * There are shell metacharacters in the buffer. Wind "offset" back to
	 * either i) the start of the file name or ii) the last '/' so that we
	 * can find what directory the expansion must be performed in.
	 */
	while(fname[offset] != '/' && offset > 0)
		offset--;

	/*
	 * Here we are, either at the start of the buffer, or we're looking
	 * at the end of a directory name. If the latter, copy the directory
	 * name into "base".
	 */
	if(offset > 0) {
		tcatm(base, fname, offset);
		offset++;		/* move past the '/' */
	}

	/*
	 * The next step is to see if the directory name in "base" is valid.
	 *
	 * If this is not done, the shell will die (horribly) when it tries
	 * to cd to the directory given.
	 */
	if(tused(base) != 0 && access(ttext(base), X_OK) < 0) {
		*errflag = 1;
		return(mlerr(ttext(base), "no directory"));
	}

	cmd = playText2;
	tsetused(cmd, 0);

	/* If there is somewhere to cd to, and we have search permission on
	 * the directory, cd there and expand the wildcard.
	 *
	 * Use -P option to cd - this gets it to follow symbolic links, ie end
	 * up with the same thing that you get when the file is opened.
	 * This is to stop the shell trying to do clever things with ..
	 *
	 * This expansion:
	 * works with: ksh bash dash zsh
	 * does not work with: csh tcsh fish
	 *
	 * Consider using wordexp() - which is a GNU thing only
	 */
	if(tused(base) != 0)
		printToText(cmd, "cd -P '%t';", base);

	/* Protect the user from themselves, quote shell special characters.
	 * Allow them to use:
	 * *	$	for variables
	 * *	*?[	globbing
	 * *	\	for protecting the above & \
	 */
	tp = fname + offset;
	tsetempty(playText4);
	while(*tp) {
		if(strchr(" \t\n|&;()<>{}`'\"~!", *tp) != NULL)
			tcatc(playText4, '\\');
		if(*tp == '\\') {
			tcatc(playText4, '\\');
			if(strchr("$*[?", tp[1]))
				tp++;
		}
		tcatc(playText4, *tp++);
	}

	// Use the shell to glob and printf to NUL separate them:
	printToText(cmd, "%A IFS=; for f in %t; do printf '%%s\\0' \"$f\"; done 2>/dev/null", playText4);

	if(pi_open(ttext(cmd), "r", -1) != FIOSUC) {
		*errflag = 1;
		c = mlerr("pi_open fail", strerror(errno));

		return(c);
	}

	// Read everything from the pipe into playText3
	tsetused(playText3, 0);
	while((charsin = (int)fread(pipeBuf, 1, sizeof(pipeBuf), ffp)) > 0) {
		tcatm(playText3, pipeBuf, charsin);
	}

	if(ferror(ffp)) {
		/*
		 * Eeeek ! Something has gone drastically wrong.
		 */
		*errflag = 1;
		c = mlerr("bad read from pipe", strerror(errno));
		(void) pi_close();

		return(c);
	}

	if(pi_close() == -1) {
		/*
		 * Eeeek ! Something else has gone drastically wrong.
		 */
		*errflag = 1;
		c = mlerr("pi_close error", strerror(errno));

		return(c);
	}

	// If we got nothing back - then say so:
	if(tused(playText3) == 0)
		return(mlerr("no expansions", NULL));

	fileList = ttext(playText3);	// fileList is a convenience variable

	/*
	 * Work out where to start, first element in buffer or last.
	 */
	if(dirn == BACKW)
		/* Going backwards ! */
		pcmd = prevword(fileList, playText3);
	else
		pcmd = fileList;

	/* We sit in a loop, copying each match from "fileList" into "buf",
	 * starting at the offset (which may be the start of the buffer).
	 *
	 * If we are in "interactive mode" and the user hits space, we go
	 * around again, if anything other than a space we return whatever
	 * it was that wasnt a space.
	 *
	 * If in "non-interactive mode" just copy the first match and
	 * return 0.
	 */
	for(;;) {
		int	buflen;

		tsetused(dotp, promptLen + (LineLn)offset);
		tcats(dotp, pcmd);

		fname = tgetcaddr(dotp, promptLen);	// Address can change as l_text moved

		/* Now "buf" contains a valid filename, if it is a directory
		 * name, make sure it ends with a '/' charcter.
		 */

		buflen = (int)strlen(fname);
		if(isdir(fname) && fname[buflen - 1] != '/') {
			tcats(dotp, "/");
			fname = tgetcaddr(dotp, promptLen);
		}

		/* Complain if the pattern didnt expand.
		 */
		if(*pcmd == '\0' && sh_meta(fname) != (BYTE*) NULL) {
			*errflag = 1;
			c = mlerr("no expansions", (char*) NULL);

			return(c);
		}

		/* But if the pattern did expand and we are non-interactive
		 * then that's ok
		 */
		if(dirn == 0) {
			/* non-interactive expansion, return */
			return(0);
		}

		/* find out what the user wants to do */
		wmessp->w_flag |= WFHARD;
		wmessp->w_doto = tused(dotp);
		c = tgetc(MGC_MAP, 0);

		if(c != ' ' && c != '\t' && c != CharRevSearch) {
			return(c);
		}

		dirn = (c == ' ' || c == '\t' ? FORW : BACKW);

		if(dirn == FORW) {
			pcmd = strchr(pcmd, '\0') + 1;
			if(pcmd >= tgetcaddr(playText3, tused(playText3)))
				pcmd = ttext(playText3);
		} else {
			pcmd = prevword(pcmd, playText3);
		}
	}

	/* NOT REACHED */

#else
	*errflag = 0;
	return(mlerr("file expansion unimplemented", (char*)NULL));
#endif	/* UNIX */
}


/* Return a pointer to the buffer before "pbuf" in the buffer list.
 */
BUFFER*	prevbuf(const BUFFER* pbuf)
{
	BUFFER*	bp   = bheadp;
	BUFFER*	last = bheadp;	// Initialise to stop warning
	
	for(/*void*/;bp != (BUFFER*) NULL;bp = bp->b_bufp) {
		last = bp;
		if(bp->b_bufp == pbuf)
			return(bp);
	}

	/* Didnt find it, must have been passed bheadp. Return
	 * a pointer to the last buffer in the list, which just
	 * happens to be pointed to by "last".
	 */
	return(last);
}


/* Interactive buffer name expansion. Trivial compared with fnexpand().
 *
 * Text left in wmessp->w_dotp;
 * 
 * We only ever have an interactive mode, since if the user types a buffer
 * name, hits return and it sees it expanded into something else, they'd get
 * pretty annoyed (well I would anyway).
 *
 * Metacharacters are NOT handled. We only look for a buffer name which
 * matches what has been typed in. The general idea is much the same as
 * for fnexpand(), space is the switch character, telling us to look for
 * another match if there is one.
 *
 * dotp is safe in this function as wmessp will not be reread.
 *
 * prompt	current prompt
 * dirn		direction to go, + forward, - backward
 * win		original window - that was being edited
 */
KeyCode	bnexpand(const char* prompt, int dirn, Window* win)
{
	KeyCode	c;
	int	match	 = 0;			/* a match found ?	*/
	int	startlen;			/* start length of buf	*/
	BUFFER*	bp	 = win->w_bufp;		/* index buffer pointer	*/
	BYTE*	eprom    = "    File: ";	/* extra prompt		*/
	BYTE*	nofile	 = "    [no file]";	/* no file prompt	*/
	int	promptLen = (int)strlen(prompt);
	LINE*	dotp = wmessp->w_dotp;

	startlen = (int)lused(dotp) - promptLen;
	
	/* The start position is determined as follows:
	 * if going forward, start at the next buffer in the list,
	 * if going backward, start at the previous buffer in the list.
	 */
	if(dirn == FORW) {
		/*
		 * Going forward.
		 */
		if(bp->b_bufp != (BUFFER*) NULL)
			bp = bp->b_bufp;
		else
			bp = bheadp;
	} else {
		/*
		 * Going backward.
		 */
		bp = prevbuf(bp);
	}

	for(;;) {
		for(/*void*/; bp != (BUFFER*) NULL;
				bp = dirn == FORW ? bp->b_bufp : prevbuf(bp)) {

			if(dirn == BACKW && bp == win->w_bufp &&
						(match == 0 || match == -1))
				/* come back to the start */
				break;

			// Not for UTF8 chars. TODO
			if(strncmp(lgetcaddr(dotp, promptLen), bp->b_bname, (size_t)startlen) != 0)
				continue;

			/* Show up invisible buffers only if we are looking
			 * for something specific.
			 */
			if(bp->b_flag&BFINVS && startlen == 0)
				continue;

			/* We've got a match, copy it into buf.
			 */
			match = 1;

			lsetused(dotp, promptLen);
			lcats(dotp, bp->b_bname);

			/* If the buffer name and filename are not the same
			 * then display the filename. This is very simple minded
			 * ie it is a straight comparison, and not a comparison
			 * of the "basename" of the filename.
			 */
			if(strcmp(lgetcaddr(dotp, promptLen), bp->b_fname) == 0 && bp->b_fname[0] == '\0')
				lcats(dotp, nofile);
			else {
				lcats(dotp, eprom);
				lcats(dotp, bp->b_fname);
			}

			wmessp->w_doto = lused(dotp);
			wmessp->w_flag |= WFHARD;
			if((c = tgetc(MGC_MAP, 0)) != ' ' && c != CharRevSearch) {
				// Remove any text after the buffer name:
				lsetused(dotp, (LineLn)promptLen + (LineLn)strlen(bp->b_bname));
				return(c);
			}
			
			dirn = (c == ' ' ? FORW : BACKW);
			
			/* The user wants another match, restore the line to what
			 * it was before we expanded it. This is easy since
			 * we know the original length of the buffer, and that
			 * anything we copied in was the same up to this
			 * length.
			 */
			lsetused(dotp, startlen + promptLen);
		}

		/* No matches so far, go 'round again if we're not trying to
		 * loop infinitely.
		 */
		if(dirn > 0)
			bp = bheadp;
		else
			bp = prevbuf(bheadp);

		if(match > 0)
			/* We found something, its safe to go around again.
			 */
			continue;

		if(--match < -1)
			/* We've been around once already and found nothing.
			 * Dont go round again or we might be here for a long
			 * while.
			 */
			return(mlerr("no matches", (char*) NULL));
	}
}

/* Buf contains len characters which may be the start of a mode name.
 * Return the index for that mode.
 * If len is < 0, move in direction dirn (FORW|BACKW) from index.
 * Pass index value of -1 if you don't care or want to start off.
 * Return -1 on error.
 *
 * NB: case conversion not done, what is in buf should be in upper case.
 */
int	ModeMatch(const ModeDef* Opts, const BYTE* buf, int len, int dirn, int index)
{
	
	if(len > 0) {
		for(index = 0; index < Opts->md_mode_cnt; index++)
			if(strncmp(buf, Opts->md_modes[index].mn_name, (size_t)len) == 0)
				return(index);

		return(-1);
	}

	dirn == FORW ? index++ : index--;
			
	/* perform wrap around */
	if(index < 0)
		index = Opts->md_mode_cnt - 1;
	else if(index >= Opts->md_mode_cnt)
		index = 0;

	return(index);
}

/* Show buffer modes or window options in the prompt line.
 *
 * explan	An explanation of what this is
 * Opts		The options structure
 * modes	The mode/option bits to show
 */
void	ShowModes(const char* explan, const ModeDef* Opts, int modes)
{
	int	i;

	tcpys(playText, explan);

	for(i = 0; i < Opts->md_mode_cnt; i++)
		if(modes & (1 << i)) {
			tcats(playText, " ");
			tcats(playText, Opts->md_modes[i].mn_name);
		}

	mlwrite("%t", playText);
}

/* mnexpand() a mode expansion routine which works in a similar way to the
 * buffer name expansion above.
 *
 * It stems from a request from Alain Williams (who I work for these days)
 * who said (and I must agree) that it is often difficult to remember mode
 * names and it would be nice to "space round them" as for buffer names.
 *
 * All this routine does is to copy into the buffer successive elements of
 * the modeNameExpan[] or WinOpt[] arrays defined in edef.h
 *
 * dotp is safe in this function as wmessp will not be reread.
 *
 * Opt		Details on what the options are
 * prompt	prompt to display
 * dirn		direction to go, +ve forward, -ve back
 * errflag	set if error on expansion
 *
 * The current text is in message line "."
 *
 * Return the character that stopped the expansion.
 */
KeyCode	mnexpand(const ModeDef* Opts, const BYTE* prompt, int dirn, int* errflag)
{
	KeyCode	c;
	int	i;
	int	promptLen = (int)strlen(prompt);
	LINE*	dotp = wmessp->w_dotp;
	int	startlen = (int)lused(dotp) - promptLen;	// length of initial text
	
	if((i = ModeMatch(Opts, lgetcaddr(dotp, promptLen), startlen, FORW, -1)) < 0) {
		*errflag = -1;
		return(mlerr(Opts->md_err, (char*)NULL));
	}

	for(;;) {
		// Put the current candidate after the prompt:
		lsetused(dotp, promptLen);
		lcats(dotp, Opts->md_modes[i].mn_name);	/* copy it into the buffer */
		wmessp->w_doto = lused(dotp);
		wmessp->w_flag |= WFHARD;

		/* get a character, if not space or ^R then return it */
		if((c = tgetc(MGC_MAP, 0)) != ' ' && c != CharRevSearch)
			return(c);
		
		/* work out new direction */
		dirn = (c == ' ' ? FORW : BACKW);

		/* Take out what we added, we know we can get back to
		 * the original because what we added was extra and taking
		 * it off gets us back to where we were.
		 */
		if((i = ModeMatch(Opts, lgetcaddr(dotp, promptLen), startlen, dirn, i)) < 0) {
			*errflag = -1;
			return(mlerr(Opts->md_err, (char*)NULL));
		}
	}
}

/* Find the next Emacs command - eg: abort-command
 *
 * pos		Preserved position pointer. Set NULL on keystroke that is not for/back-ward search.
 * dirn		direction to go, +ve forward, -ve back
 * prompt	Prompt at the start of the line
 *
 * *pos is set to where we are in the NBIND table CmdNames.
 * If this is NULL find an initial position by looking for the first
 * command that starts with the text in buf.
 * 'first' is first from start or end depending on dirn.
 *
 * Normally return no-character UNISTR_END, but on error *could* return the next one read
 */
KeyCode	NextCommand(void** pos, int dirn, const BYTE* prompt)
{
	NBIND*	np = *pos;		// Get state
	static	NBIND*	nameEnd;	// Last entry in CmdNames
	int	promptLen = (int)strlen(prompt);
	LINE*	dotp = wmessp->w_dotp;

	// Going backwards, end not known, find it
	if(dirn < 0 && !nameEnd) {
		for(nameEnd = CmdNames; nameEnd->n_name; nameEnd++)
			;

		// Back up to the last one
		nameEnd--;
	}

	// No state, find something
	if( !np) {
		// Find where to start searching

		np = (dirn > 0) ? CmdNames : nameEnd;	// start/end of the array

		/* Now search until we find something that matches what we have in buf.
		 * Description of strncmp() does not say what happens if compared length is 0,
		 * so deal with that specially by not moving from where we are.
		 */
		if(lused(dotp) > promptLen) {			while(np) {
				if(strncmp(lgetcaddr(dotp, promptLen), np->n_name, (size_t)lused(dotp) - (size_t)promptLen) == 0)
					break;
				// No match, move to next. Beware the ends
				if(dirn > 0) {
					// Forwards
					if( !(++np)->n_name)
						np = (NBIND*)NULL;
				} else {
					// Backwards
					if(--np < CmdNames)
						np = (NBIND*)NULL;
				}
			}

			// What did we find ? Did we run off the end before a match ?
			if( !np)
				return(mlerr("Emacs command not found", (char*)NULL));

		}

	} else {
		// There is a previous match, just go next/prev.
		// No match, move to next. Beware the ends; wrap around
		if(dirn > 0) {
			// Forwards
			if( !(++np)->n_name)
				np = CmdNames;
		} else {
			// Backwards
			if(--np < CmdNames)
				np = nameEnd;
		}
	}

	// Put what we found in the buffer and set length
	// Put what we found in dotp and set length
	lsetused(dotp, promptLen);
	lcats(dotp, np->n_name);
	wmessp->w_doto = lused(dotp);

	*pos = np;		// Preserve state

	return(UNISTR_END);
}

/* Find the next Emacs variable - eg: $french-spacing %count
 *
 * pos		Preserved position pointer. Set NULL on keystroke that is not for/back-ward search.
 * dirn		direction to go, +ve forward, -ve back
 * prompt	Prompt at the start of the line
 *
 * *pos is set to where we are in the NBIND table CmdNames.
 * If this is NULL find an initial position by looking for the first
 * command that starts with the text in buf.
 * 'first' is first from start or end depending on dirn.
 *
 * dotp is safe in this function as wmessp will not be reread.
 *
 * Normally return no-character UNISTR_END, but on error *could* return the next one read
 */
KeyCode	NextVarName(void** pos, int dirn, const BYTE* prompt)
{
	int	promptLen = (int)strlen(prompt);
	LINE*	dotp = wmessp->w_dotp;

	// This test is also important for NextEnvVarName() & NextUserVarName() assumption
	if(lused(dotp) <= promptLen)
		return(mlerr("Enter '%' or '$' to start a variable name (environment or user)", (char*)NULL));

	// Look at first char - must be '%' or '$'
	if(lgetc(dotp, promptLen) != '%' && lgetc(dotp, promptLen) != '$')
		return(mlerr("Variables start '%' or '$'", (char*)NULL));

	if(lgetc(dotp, promptLen) == '$')
		return(NextEnvVarName(pos, dirn, prompt));

	return(NextUserVarName(pos, dirn, prompt));
}

/* end */
