/*
 * ADDW SCCS: @(#)file.c	2.210 10/14/20 15:21:49 - second
 */

/*	FILE.C:   for MicroEMACS

	The routines in this file handle the reading, writing
	and lookup of disk files.  All of details about the
	reading and writing of the disk are in "fileio.c".

*/

#include	"estruct.h"
#include        "utf8.h"
#include        "edef.h"

#if	SHOWMODREL
char	filevers[] = "@(#)file.c	2.210 10/14/20 15:21:49";
#endif

#if	UNIX
#include <libgen.h>
#include <fcntl.h>
#include <stdlib.h>
#endif


/* Test return code from ffgetline() and return an appropriate error string
 * or the empty string if all is OK.
 *
 * On error if appropriate set the buffer-is-truncated flag BFTRUNC.
 *
 * If incompleteFlag this is allowed to set the incomplete-last-line flag MDNONL
 * on the buffer.
 */
char*	fileErrString(int s, int incompleteFlag)
{
	char*	msg = "";

	switch(s) {
	    case FIOERR:
		msg = "I/O ERROR ";
		curbp->b_flag |= BFTRUNC;
		break;

	    case FIOMEM:
		msg = "OUT OF MEMORY, ";
		curbp->b_flag |= BFTRUNC;
		break;

	    case FIOFUN:
		msg = "Incomplete last line, ";
		if(incompleteFlag)
			curbp->b_mode |= MDNONL;
		break;
	}

	return(msg);
}

/* Generate a message that says how many lines have been read/inserted
 * and what error (if there was an error).
 *
 * The other thing that it does is to make the messages %E or %I,
 * ie what highlighting there is.
 *
 * Also:
 * fileErrString() returns an error string and clears buffer flags
 * utf8_errmsg() will return any error message and
 * if there is an error clear the buffer curbp flag MDUTF8 & clear l_utf8 on all lines
 */
void	ReportRead(const char* act, int s, BufLns lines)
{
	// Get the error string - "" if no error
	char*	ers = fileErrString(s, 1);
	// If no error use %I rather than %E
	char*	mlwFmt = ers[0] ? "%E""%s%s%s %D line%s" : "%I""%s%s%s %D line%s";

	mlwrite(mlwFmt, ers, act, utf8_errmsg(1), lines, ((lines != 1) ? "s" : ""));
}

/* Append to the current message line the status of the last opened file
 * with respect to the number of links and if it is a symbolic link.
 *
 * Say nothing if
 *	numLinks is <= 1
 *	isSlink is FALSE
 *
 * This will only ever say something on a Unix system.
 */
void	sayLinks(void)
{
	if(numLinks > 1)
		mlwrite("%A. File has %d links", numLinks);

	if(isSlink)
		mlwrite("%A. File is a symbolic link");
}

/* Read a file into the current buffer.
 * This is really easy; all you do it find the name of the file, and call the
 * standard "read a file into the current buffer" code.
 *
 * The count is ignored. There is not a default count.
 *
 * Command:  read-file
 * Bound to: C-x C-r
 *
 * Return:
 *	TRUE   File read
 *	FALSE  cannot get a lock on the file
 *	       cannot open the file
 *	       prompt error
 *	ABORT  cannot access the file
 *	       it is a nast file
 *	       prompt error, C-g presse
 *	FAILED
 *	       prompt error, nothing entered
 */
int	fileread(int f, Cnt n)
{
	int	s;

	if((s = mlprompt("Read file: ", MLXFILE, bmlFileHistp)) != TRUE)
		return(s);

	s = readin(ltext(bmlFileHistp->b_dotp), TRUE, FALSE, TRUE);
#if	UNIX
	if(f_funnyfile)
		curbp->b_flag |= BFWEIRD;	/* Remember if its weird*/
#endif
	return(s);
}

/* Insert a file into the current buffer.
 *
 * This is really easy; all you do it find the name of the file, and call the
 * standard "insert a file into the current buffer" code.
 *
 * The count is ignored. There is not a default count.
 *
 * Command:  insert-file
 * Bound to: C-x C-i
 *
 * Return:
 *	TRUE  file inserted
 *	ifile()
 *	FALSE nothing entered at prompt
 *	ABORT C-g pressed
 */
int	insfile(int f, Cnt n)
{
	int	s;

	if((s = mlprompt("Insert file: ", MLXFILE, bmlFileHistp)) != TRUE)
		return(s);

	return(ifile(ltext(bmlFileHistp->b_dotp)));
}

/* Select a file for editing.
 *
 * Look around to see if you can find the fine in another buffer; if you can
 * find it just switch to the buffer. If you cannot find the file, create a new
 * buffer, read in the text, and switch to the new buffer.
 *
 * The count is ignored. There is not a default count.
 *
 * Command:  find-file
 * Bound to: C-x C-f   FN8
 *
 * Return:
 *	TRUE  file opened & read in
 *	FALSE nothing entered at prompt
 *	ABORT C-g pressed
 *	????
 */
int	filefind(int f, Cnt n)
{
	register int	s;		/* status return	    */
	BUFFER*		obufp = curbp;	/* where we are		    */
	BYTE		obufn[BUFFER_NAME_SIZE];	/* Current buffer name */

	strcpy(obufn, curbp->b_bname);

	if((s = mlprompt("Find file: ", MLXFILE, bmlFileHistp)) != TRUE)
		return(s);

	if((s = getfile(ltext(bmlFileHistp->b_dotp), TRUE)) == TRUE)
		/* Remember which buffer we came from.
		 */
		strcpy(lbufn, obufn);

	if(s == ABORT)
		/* It didnt work ! We've got a bit of a catastrophy on our
		 * hands now since getfile half swapped buffers for us. Now
		 * we have so swap back.
		 */
		swbuffer(obufp);

	return(s);
}

/* Find file into other window.
 *
 * This is trivial since all the hard stuff is done by the routines splitwind()
 * and filefind().
 *
 * The count is used to put the "found file" into the upper or lower window.
 * An argument of 1 forces "." into the upper window,
 * an argument of not 1 forces "." to the lower window.
 *
 * The default count is -1.
 *
 * Command:  split-and-find-file
 * Bound to: C-x 4
 *
 * Return:
 *	TRUE  file opened & read in
 *	FALSE nothing entered at prompt
 *	ABORT C-g pressed
 *	???
 */
int	owinfil(int f, Cnt n)
{
	register int s;		/* status return */

	if((s = mlprompt("Find file: ", MLXFILE, bmlFileHistp)) != TRUE)
		return(s);

	if(f == FALSE) {
		f = TRUE;
		n = -1;
	}

	if((s = splitwindvert(f, n)) != TRUE)
		return(s);

	return(getfile(ltext(bmlFileHistp->b_dotp), TRUE));
}

/* Split window horizontally & find file into the new window.
 *
 * This is trivial since all the hard stuff is done by the routines
 * splitwindhoriz() and filefind().
 *
 * The new window is the right hand one, but if the count is set it is the left
 * hand one.
 *
 * Command:  split-horizontally-and-find-file
 * Bound to: C-x 5
 *
 * Return: TODO
 *	FALSE nothing entered at prompt
 *	ABORT C-g pressed
 */
int	oHorizWinfil(int f, Cnt n)
{
	register int	s;			/* status return */
	Window*		oCurwp = curwp;

	if((s = mlprompt("Find file: ", MLXFILE, bmlFileHistp)) != TRUE)
		return(s);

	if((s = splitwindhoriz(FALSE, 0)) != TRUE)
		return(s);

	if(f) {
		curwp = oCurwp;
		curbp = curwp->w_bufp;
	}

	return(getfile(ltext(bmlFileHistp->b_dotp), TRUE));
}

/* Open a file in VIEW mode.
 *
 * The count is ignored.
 *
 * Command:  view-file
 * Bound to: C-x C-v
 *
 * Return: TODO
 *	FALSE nothing entered at prompt
 *	ABORT C-g pressed
 */
int	viewfile(int f, Cnt n)
{
	int	 s;			/* status return			*/
	BUFFER*	 obufp = curbp;		/* where we started			*/

	if((s = mlprompt("View file: ", MLXFILE, bmlFileHistp)) != TRUE)
		return(s);

	if((s = getfile(ltext(bmlFileHistp->b_dotp), FALSE)) != ABORT) {
		/* if we succeed, put it in view mode
		 */
		curwp->w_bufp->b_mode |= MDVIEW;

		/* scan through and update mode lines of all windows */
		upmode(NULL, WFMODE);
	}

	if(s == ABORT)
		/* It didnt work ! We've got a bit of a catastrophy on our
		 * hands now since getfile half swapped buffers for us. Now
		 * we have so swap back.
		 */
		swbuffer(obufp);

	return(s);
}

#if	CRYPT
/* Reset the encryption key if needed.
 * Prompt for a key if one is not known.
 *
 * If verify request the user to verify (ie retype) the key,
 * only done if one has to be entered.
 *
 * Return:
 *     TRUE  - OK to process
 *     FALSE - prompt error
 *     ABORT - prompt error
 *     FAILED- prompt error
 */
int	resetkey(int verify)
{
	register int s; /* return status */

	/* turn off the encryption flag */
	cryptflag = FALSE;

	/* if we are in crypt mode */
	if(curbp->b_mode & MDCRYPT) {
		if(tused(curbp->b_key) == 0 &&
		   (s = crysetkey(verify, 0)) != TRUE)
			return(s);

		/* let others know... */
		cryptflag = TRUE;

		/* and set up the key to be used! */
		/* de-encrypt it */
		cryptBuffer((char *)NULL, 0);
		cryptBuffer(ttext(curbp->b_key), tused(curbp->b_key));

		/* re-encrypt it...seeding it to start */
		cryptBuffer((char *)NULL, 0);
		cryptBuffer(ttext(curbp->b_key), tused(curbp->b_key));
	}

	return(TRUE);
}
#endif

/* Work out if file fname is already in a buffer.
 *
 * *reason	This is set to a description of why the file is
 *		already there: link, symlink, ""
 *		"" means that the user is using the same name - no
 *		need to explain that.
 *
 * Return: NULL     - no don't have the file
 *	   Non NULL - the buffer that has the file
 */
BUFFER*	alreadyHaveFile(BYTE* fname, char** reason)
{
	BUFFER*	bp;

#if	UNIX
	ino_t		inode = 0;	// Initialise to keep the compiler quiet
	dev_t		stdev = 0;
	int		inodeSet = 0;

	if(CacheStat(fname) != -1) {
		inode = (ino_t)statbuf.st_ino;
		stdev = statbuf.st_dev;
		inodeSet = 1;
	}
#endif

#if	MSDOS
	mklower(fname);		/* msdos isn't case sensitive */
#endif
	
	/* If we already have a buffer allocated, just switch to it.
	 * This will read it in if it has not already been read.
	 * This will happen if we are getting a file mentioned on the command line.
	 */
	for(bp = bheadp; bp != (BUFFER*) NULL; bp = bp->b_bufp) {

		if(bp->b_flag & BFINVS)
			continue;

		/* Check if the file name is the same.
		 * It is important here, using systems like DOS,
		 * that the comparison is case independent.
		 */ 
#if	MSDOS
		if(cstrcmp(bp->b_fname, fname) == 0)
#else
		if(strcmp(bp->b_fname, fname) == 0)
#endif
		{
			*reason = "";
			return(bp);	// Already have the file - same name
		}

#if	UNIX
		/* On UNIX boxes the files could be the same if they are linked
		 * This code does not distinguish if it is a hard or symbolic link.
		 */
		if(inodeSet && bp->b_stdev == stdev && bp->b_stino == inode) {
			/* Consider opening "dir/../A" or "../dir/A" it could be
			 * just a different path for "A".
			 * Telling the user that it is a link would be confusing,
			 * unless/especially if "dir" is a symlink.
			 * This could get horribly complicated.
			 *
			 * Really ought to compare directory inode # in each case,
			 * but assume that if the last names in the paths are the
			 * same that the path brings to the same directory.
			 *
			 * If (the last component of) either name is a symbolic
			 * link - assume that that is why we are here.
			 */
			*reason = "";

			char*	fnNew;
			char*	fnBuf;

			if((fnNew = strrchr(fname, '/')) == NULL)
				fnNew = fname;
			else
				fnNew++;
			if((fnBuf = strrchr(bp->b_fname, '/')) == NULL)
				fnBuf = bp->b_fname;
			else
				fnBuf++;

			if(strcmp(fnNew, fnBuf) != 0)
				*reason = " (link)";
			if(isSymlink(fname) || isSymlink(bp->b_fname))
				*reason = " (symlink)";

			return(bp);
		}
#endif
		// Different name & (on UNIX) different device/inode
	}

	return((BUFFER*)NULL);
}

/* Read a file in.
 *
 * fname:	file name to find
 * lockfl	check for file locks - only #if FILOCK
 *
 * This no longer has to worry about reading from a pipe, see readStream()
 *
 * The compiled name of this function (ie as seen by gdb) is _getfile
 * see edef.h
 *
 * Return: TRUE, FALSE, ABORT
 */
int	getfile(BYTE* fname, int lockfl)
{
	register BUFFER*bp;
	register int	s;
	BYTE*		bname;		/* buffer name to put file */
	BUFFER*		prevBuf;
	char*		reason = "";	// Why we already have the file

#if	MSDOS
	mklower(fname);		/* msdos isn't case sensitive */
#endif

	if((bp = alreadyHaveFile(fname, &reason)) != NULL) {
		/* Say that it is an old buffer now,
		 * swbuffer() might read it in and so overwrite message with 'Read xx lines',
		 * that does not matter - we just knew about it but had not yet read it.
		 */
		mlwrite("%s: old buffer%s", bp->b_bname, reason);

		swbuffer(bp);

		/* There used to be code here that fiddled to set w_topLinep so that
		 * "." was 1/2 way down the window. But now ConnectBuffer() will
		 * restore it to what it was or set to top line in buffer --
		 * depending on if the file had been read in.
		 */

		return(TRUE);
	}

	/* New buffer name.
	 */
	bname = makename(fname);
	if((bp = bfindCheck(bname, BFNCREATE|BFNACTIVE)) == NULL)
		return(FALSE);

	DisconnectBuffer(curwp);		/* Undisplay		*/
	prevBuf = curbp;
	curbp = bp;				/* Switch to it		*/
	ConnectBuffer(curbp, curwp, 0);

	/* Read it in */
	if((s = readin(fname, lockfl, FALSE, TRUE)) == ABORT) {
		/* If it didnt work.
		 * Connect back to where we were.
		 * kill the empty buffer
		 */
		DisconnectBuffer(curwp);
		curbp = prevBuf;
		ConnectBuffer(curbp, curwp, 0);

		zotbuf(bp, TRUE);

		return(s);
	}

#if	UNIX
	if(f_funnyfile)
		curbp->b_flag |= BFWEIRD;	/* Remember if its weird */
#endif
	return(s);
}

/* Look for a save file and open it on fpp if it is younger than fname.
 * The user is prompted to check if the want to use the save file.
 *
 * If a save file was found but it is older than fname or the user
 * did not want to use it - the user is asked if they want to remove
 * the save file.
 * Looks at %notice-emacssave
 *
 * This must not be used for when multiple files may be returned.
 * 
 * Return:
 *	FIOSUC if the save file is opened.
 *	FIOFNF if the save file was not found, or find not even attempted
 */
int	OpenSaveFile(const BYTE* fname)
{
	char*		play;
	Time_T		fnameModtime;
	struct	stat	saveStat;	// Stat of save file

	// Do we bother with save files ?
	if( !NoticeEmacsSave)
		return(FIOFNF);

	/* Look at what we will open - modification time.
	 */
	fnameModtime = CacheStat(fname) ? (Time_T)0 : statbuf.st_mtime;		

	// Get last component:
	if((play = strrchr(fname, '/')) != (char*)NULL)
		fname = play + 1;
	if(fname[0] == '\0')
		return(FIOFNF);

	// Try to find possibly saved file
	if(findFile(fname, fT_Save)) {
		char*	old = "old";
		// mlwrite("%#Opened %s ffp=%p", ttext(foundFiles), ffp);

		// Find age of what we found
		fstat(fileno(ffp), &saveStat);

		/* If it is newer than what we were asked to read,
		 * offer to read it in.
		 */
		if(saveStat.st_mtime > fnameModtime) {
			if(mlyesno("Found %s, read this in [y/n]? ", ttext(foundFiles)) == TRUE)
				return(FIOSUC);
			old = "newer";
		}

		// Close unwanted file handle:
		ffclose();

		// Offer to remove an unwanted/old save file:
		if(mlyesno("Found %s %s, remove it [y/n]? ", old, ttext(foundFiles)) == TRUE) {
			/* The messages below will prob not be seen without the sleep.
			 */
			if(unlink(ttext(foundFiles)) == 0)
				mlwrite("%F""%I""%s removed", ttext(foundFiles));
			else
				mlwrite("%F""%E""%s not removed: %e", ttext(foundFiles), errno);
			sleep((unsigned)tpause);
		}
	}

	return(FIOFNF);	// Nothing found
}

/* Read from ffp into the current buffer.
 *
 * The current contents (if any) are left undisturbed.
 * Lines are inserted at the end of the buffer.
 *
 * If AutoDos check buffer for \r: remove trailing \r then set DOS mode
 *
 * Any narrow is widened to the whole buffer.
 *
 * readTail	read the lines that have been appended to the file since
 *		it was last read. \r strip if DOS mode.
 *
 * This cares not if ffp is open on a file or pipe.
 * ffp is closed at EOF.
 *
 * The current buffer is marked active: b_active
 *
 * Return	TRUE on OK
 *		FALSE
 *		ABORT
 */
int	readFD(int readTail)
{
	LINE*	lnew;
	BufLns	nline = 0;		// # lines read in
	int	s;
	char*	dosMsg = (char*)NULL;	// Message if DOS mode

	utf8_errclear();

	widen(FALSE, 0);		// Lose any narrowing

	/* turn off ALL keyboard translation in case we get a dos error */
	TTkcloseM;

#if	UNIX
	if(readTail) {
		// fseek() is highly unlikely to fail, but if it does
		// read the whole file in:
		if(fseeko(ffp, curbp->b_fileEnd, SEEK_SET) == -1) {
			bclear(curbp);
			readTail = FALSE;
		}
	}
#endif

	curbp->b_active = TRUE;

	while((s = ffgetline(&lnew, curbp->b_mode & MDUTF8)) == FIOSUC || s == FIOFUN) {
		LINE*	lbefore = lback(curbp->b_linep);

		// Stitch the line in:
		lsetback(curbp->b_linep, lnew);
		lsetforw(lnew, curbp->b_linep);
		lsetback(lnew, lbefore);
		lsetforw(lbefore, lnew);

		// DOS mode - strip the \r at line end
		// Screw up slighly if there is not a \r - but unlikely
		if(readTail && (curbp->b_mode & MDDOS) && lused(lnew) > 0 && lgetc(lnew, lused(lnew) - 1) == RETCHAR)
			lsetused(lnew, lused(lnew) - 1);

		++nline;

		if(ffutf8err)
			utf8_setmsg(nline);

		if(s == FIOFUN) {
			s = FIOEOF;	// EOF, just missing \n
			break;
		}
	}

	/* Deal with DOS mode.
	 * If $autodos see if we should set DOS mode and do so.
	 * We only do so automatically if all lines end \r.
	 * If DOS mode: remove \r at line ends.
	 * Not when read/update - we will already have removed the \r
	 * Set a message for the user.
	 */
	if( !readTail) {
		if(AutoDos &&  IsDosBuffer(curbp) == RETC_ALL)
			curbp->b_mode |= MDDOS;
		if((curbp->b_mode & MDDOS) && (RemoveRetchar(curbp) & RETC_SOME) != 0)
			dosMsg = "(DOS mode)";
	}

#if	UNIX
	// Record position of EOF
	curbp->b_fileEnd = ftello(ffp);
#endif
	ffclose();			/* Ignore errors */

	TTkopenM;			/* Open the keyboard again */

	// fileErrString() returns an error string and clears buffer flags
	// utf8_errmsg() will return any error message and
	// if there is an error clear the buffer curbp flag MDUTF8 & clear l_utf8 on all lines

	ReportRead("Read", s, nline);
	if(dosMsg != NULL)	// Append to message highlighting DOS mode ?
		mlwrite("%A %s", dosMsg);
	sayLinks();

	curbp->b_numLines = nline;		// How many lines in buffer

	// In case it is already displayed, make sure that all line, etc, pointers
	// are safe.
	// Set "." to the first line in the file.
	curbp->b_topLinep = curbp->b_dotp = lforw(curbp->b_linep);
	ReconnectWindowsToBuffer(curbp);

	// Convert return from FIO??? to TRUE/FALSE
	return(s == FIOEOF ? TRUE : FALSE);
}

/* Read file "fname" into the current buffer, blowing away any text found there.
 *
 * Called by both the read and find commands. Return the final status of the
 * read. Also called by the mainline, to read in a file specified on the
 * command line as an argument.
 *
 * fname	name of file to read - a copy will be made elsewhere
 * lockfl	check for file locks - only #if FILOCK
 * readTail	read the lines that have been appended to the file since
 *		it was last read
 * checkSave	Check for a save file
 *
 * The compiled name of this function (ie as seen by gdb) is _readin
 * see edef.h
 *
 * It is possible that the file has already been opened, check ffp,
 * if so just use it. It will be closed when this function returns.
 *
 * Update curbp->b_lsType if not reading from a pipe.
 * This is currently safe: curbp is correctly set when this is called.
 * If it is anything other than a regular file make the buffer read only.
 *
 * Mark curbp->b_active as TRUE.
 *
 * Return	TRUE   on OK - file read in
 *		FALSE  cannot get a lock on the file
 *		       cannot open the file
 *		       prompt error
 *		ABORT  cannot access the file
 *		       it is a nasty file
 *		       prompt error
 *		FAILED
 *		       prompt error
 */
int	readin(const BYTE* fname, int lockfl, int readTail, int checkSave)
{
	register int	i;
	int		s;

#ifdef	UNIX
	int		dirwrt = 0;	/* if true, directory is writable */
	int		mode = 0;	/* mode of file */
#endif	/* UNIX */

#ifdef	UNIX
	/* Check that we can read it, don't if already open
	 */
	if(ffp == (FILE*)NULL) {
		if(CacheStat(fname) != -1)
			mode = (int) statbuf.st_mode &
				(S_IRWXU | S_IRWXG | S_IRWXO);

		if(isnasty(fname, &curbp->b_lsType)) {
			curbp->b_mode |= MDVIEW;
			return(ABORT);
		}

		if(access(fname, R_OK) < 0 && errno == EACCES) {
			// If did not fail on the directory, set file fail:
			if(curbp->b_lsType != '/')
				curbp->b_lsType = errno == ENOENT ? '.' : '!';
			// Prevent SCCS issue:
			mlwrite("%E""%s: cannot access: %e", fname, errno);
			return(ABORT);
		}
	}
#endif
#if	FILOCK
	if(ffp != (FILE*)NULL && lockfl && (s = lockchk(fname)) == ABORT) {
		s = FALSE;
		goto err;
	}
#endif
#if	CRYPT
	// We don't force the key to be verified, the user will try again if it is wrong
	if((s = resetkey(0)) != TRUE)
		goto err;
#endif

	/* Blow away current contents
	 * Not if reading recent lines appended
	 */
	if(!readTail && (s = bclear(curbp)) != TRUE)
		goto err;

	// If we automatically put into view mode because it was readonly,
	// then remove view mode -- unless globally we are in view mode.
	if(!readTail && (curbp->b_flag & BFREAD) &&
	   (curbp->b_flag & BIMMUTABLE) == 0 && (gmode & MDVIEW) == 0)
		curbp->b_mode &= ~(MDVIEW);

	curbp->b_flag &= ~(BFINVS|BFCHG|BFREAD|BFTRUNC|BFWEIRD|BFIMPLICIT|BRENEWED|BFROMSAVE);
	// Leave readonly mode (maybe from global -v)
	// curbp->b_mode &= ~(MDVIEW);

#if	UNIX
	ClearInodeInfo(curbp);		/* Invalidate mod time, etc */
	f_funnyfile = 0;
#endif

	/* If this is a file - still want to set the buffer name,
	 * which might invalidate fname, so assign new.
	 * If this is a renew then curbp->b_fname == fname (ie same address) but it works.
	 */
	fname = setFnames(curbp, fname);

	/* Set C mode if the file name is right.
	 * See also setcmode().
	 */
	CheckCMode(curbp);

#if 0
	/* Let a user macro get hold of things...if he wants */
	FuncExec(BindNullproc, FALSE, 1);
#endif

	// If the file is already open - this part succeeds
	// If not - read the file in:
	if(ffp == (FILE*)NULL) {
		s = FIOFNF;

		/* Maybe check for a file saved in an abort. If so mark the
		 * buffer modified since the save file contents are different
		 * from the file that we were asked to open.
		 */
		if(checkSave && (s = OpenSaveFile(fname)) == FIOSUC)
			curbp->b_flag |= BFCHG | BFROMSAVE;

		/* turn off ALL keyboard translation in case we get a dos error */
		TTkcloseM;

		if(s == FIOFNF && (s = ffropen(fname)) == FIOERR)	/* Hard file open. */
			return(FALSE);

		if(s == FIOFNF) {			/* File not found. */
#if	UNIX
			BYTE*	p;

			/* See if we can write to the directory. */
			tcpys(playDir, fname);

			/* Get the directory name ...
			 * "." if we can't find a '/'
			 */
			if((p = rindex(ttext(playDir), '/')) != (BYTE*) NULL)
				tsetused(playDir, (LineLn)(p - ttext(playDir) + 1));
			else
				tcpys(playDir, ".");

			if(access(ttext(playDir), W_OK) != -1)
				dirwrt = 1;

			if(!dirwrt) {
				curbp->b_mode |= MDVIEW;
				/*
				 * Set the BFREAD bit so that we know we
				 * are artificially in view mode.
				 */
				curbp->b_flag |= BFREAD;

				if(access(ttext(playDir), F_OK) == 0)
					mlwrite("%E""Directory %t: %e", playDir, (int)EACCES);
				else
					mlwrite("%E""%t: no such directory", playDir);
			} else
				mlwrite("%INew file %s", fname);

#else
			mlwrite("%INew file %s", fname);
#endif

			/* If it doesn't exist, the buffer cannot be old
			 * any more - removes 'X' in window mode line
			 */
			curbp->b_flag &= ~BFOLD;

			// Nothing to read ... skip all the file reading in stuff
			return(FALSE);
		}

		// If the file was prob produced when Emacs crashed - make buffer read only:
		if(NoticeEmacsSave && (i = (int)strlen(fname)) > (int)SaveSuffixLen &&
		   strcmp(fname + i - SaveSuffixLen, SaveSuffix) == 0)
			curbp->b_mode |= MDVIEW;

#if	UNIX
		f_funnyfile = 0;

		if(access(fname, W_OK) < 0) {
			if(dirwrt && relinking) {
				mlwrite("%FReading %s (mode %o)", fname, mode);
				f_funnyfile = 1;
			} else
				mlwrite("%FReading %s (readonly)", fname);
		} else
			mlwrite("%FReading: %s", fname);
#else
		mlwrite("%FReading: %s", fname);
#endif
	}

	// Read the file into the buffer
	s = readFD(readTail);

#if	UNIX
	UpdateMetaData(curbp, fname);
#endif

	if(s != TRUE)			// Fail if error
		return(FALSE);

#if	UNIX
	/* Put the buffer in view mode if file not writable and keep a record
	 * of the fact in BFREAD.
	 *
	 * The first line in the if is a bit complicated. The point is that the
	 * file might not have existed but we read from a save file; ie Emacs
	 * died when it was creating a new file but before it had ever saved it.
	 *
	 * If running setuid root access() will always say that the file is writable,
	 * so if root and none of the write bits are set put it into view mode.
	 */
	if((access(fname, W_OK) != -1 || (curbp->b_flag & BFROMSAVE)) &&
	   (!suidRoot || CacheStat(fname) == -1 ||
	    ((statbuf.st_mode & (S_IWUSR | S_IWGRP | S_IWOTH)) != 0)))
		return(TRUE);

	curbp->b_flag |= BFREAD;
	curbp->b_mode |= MDVIEW;
#endif

	return(TRUE);

err:	// Come here on error when ffp might need closing
	ffclose();
	return(s);
}

/* Read a stream into buffer bname, curbp is set to this buffer.
 *
 * The stream/file is already opened on ffp.
 * Create/find the buffer bname.
 *
 * Unless append is true clear any existing contents.
 *
 * fname will be put as the buffer's b_fname, if not NULL.
 *
 * The steam will be closed as returns.
 *
 * Part of the reason for having this is that it simplifies readin(),
 * not needing arguments like frompipe.
 *
 * Return TRUE if this works
 *	  FALSE if this does not work
 */
int	readStream(const char* bname, const char* fname, int append)
{
	int	s = TRUE;
	BUFFER*	bp;
	BUFFER*	saveBp = curbp;
	BufLns	previousLines;		// Lines in buffer before new read

	if((bp = bfind(bname, BFNCREATE|BFNACTIVE, 0)) == (BUFFER*)NULL)
		goto fail;

	if(!append && !bclear(bp))
		goto fail;

	previousLines = bp->b_numLines;	// Will be 0 if just cleared

	curbp = bp;
	if(fname)
		setFnames(curbp, fname);

	// Read the file into the buffer
	if(readFD(FALSE) != TRUE)
		s = FALSE;

	curbp->b_numLines += previousLines;

	return(s);

fail:
	curbp = saveBp;
	ffclose();
	return(FALSE);
}

/* Check (by looking at the file name) that the file could be a C or PHP source
 * file. If it is set C mode & remove I mode (if set).
 *
 * Ignore inactive buffers. This is partly because this is called when $auto-c
 * is set TRUE - via setcmode().
 *
 * Check the file name associated with the buffer, but use the buffer name if
 * file name is empty
 */
void	CheckCMode(BUFFER* bp)
{
	BYTE*	name;

	if( !bp->b_active)	// Ignore inactive
		return;

	// Name to check: prefer file name, can use buffer name
	name = bp->b_fname;
	if(name[0] == '\0')
		name = bp->b_bname;

	if(hasSuffix(name, suffixesC)) {		// $suffixes-c
		bp->b_mode |= MDCMOD;
		bp->b_mode &= ~(MDINDEN | MDPHP);
	} else	if(hasSuffix(name, suffixesPHP)) {	// $suffixes-php
			bp->b_mode |= MDPHP | MDTABEXP;
			bp->b_tabsize = PhpTabsize;
			bp->b_mode &= ~(MDINDEN | MDCMOD);
		}
}

/* Take a file name, and from it fabricate a buffer name.
 *
 * This routine knows about the syntax of file names on the target system.
 *
 * I suppose that this information could be put in a better place than a line of
 * code.
 *
 * Generate a name, that can be UTF8 - must be if we handle UTF8 file names,
 * choose last component (ie ignore directories);
 * stop if there is a ';' in the name - something about macros, see storeproc().
 * Disabiguate with <x> (x is hex).
 *
 * Return a pointer to a static area, so copy quickly.
 */
BYTE*	makename(BYTE fname[])
{
	register BYTE*	cp1;
	register BYTE*	cp2;
	int		i;
	char		num[20];	/* number to make buffer name unique */
	static	BYTE	bname[BUFFER_NAME_SIZE];	// What is returned

	// Find end of name:
	cp1 = fname;
	while(*cp1 != '\0')
		++cp1;

#if	AMIGA
	while(cp1!=&fname[0] && cp1[-1]!=':' && cp1[-1]!='/')
		--cp1;
#endif
#if	VMS
	while(cp1!=&fname[0] && cp1[-1]!=':' && cp1[-1]!=']')
		--cp1;
#endif
#if	CPM || CPM68K
	while(cp1!=&fname[0] && cp1[-1]!=':')
		--cp1;
#endif
#if	MSDOS
	while(cp1!=&fname[0] && cp1[-1]!=':' && cp1[-1]!='\\'&&cp1[-1]!='/')
		--cp1;
#endif
#if	ST520
	while(cp1!=&fname[0] && cp1[-1]!=':' && cp1[-1]!='\\')
		--cp1;
#endif
#if	FINDER
	while(cp1!=&fname[0] && cp1[-1]!=':' && cp1[-1]!='\\'&&cp1[-1]!='/')
		--cp1;
#endif
#if	UNIX
	/* If fname is like '/etc/' then back away from the trailing '/'.
	 * Otherwise we end up with an empty buffer name.
	 * We could get one of these on the command line, rare & stupid.
	 */
	if(cp1 > fname && cp1[-1] == '/')
		cp1--;

	// Get last directory component, ie ignore any parent directories:
	while(cp1 != &fname[0] && cp1[-1] != '/')
		--cp1;
#endif

	// Copy up to NBUFN characters into bname
	for(cp2 = bname, i = NBUFN; --i >= 0 && *cp1 && *cp1 != ';'; cp1 += CHAR_LEN(*cp1), cp2 += CHAR_LEN(*cp2))
		UTF8_MOVE(cp2, cp1);

	*cp2 = '\0';

	/* Sanitise the buffer name - carefully chosen file names could otherwise
	 * induce an infinite loop.
	 */
	tsetused(playText, 0);
	if(tcpym(playText, bname, (LineLn)strlen(bname) + 1) == FALSE)
		Die(MsgOutOfMemoryAlloc, "buffer name");
	strcpy(bname, ttext(playText));

	/* Now we've made our buffer name, check that it is unique.
	 */
	for(i = 1; bfind(bname, BFNNONE, 0) != (BUFFER*) NULL; i++) {
		/* There is another buffer with this name. Try sticking a
		 * hexadecimal number in angled braces on the end of the
		 * name.
		 */
		size_t	len;

		/* Blat any previous attempt at disambiguation
		 */
		if(i > 1 && (cp1 = strrchr(bname, '<')) != (BYTE*) NULL)
			*cp1 = '\0';

		len = (size_t)sprintf(num, "<%x>", i);

		// If the combination is now too long, shorten the buffer name
		if(utf8_strlen(bname) + len >= NBUFN)
			utf8_strtrunc(bname, NBUFN - len);

		strcat(bname, num);
	}

	return(bname);
}

/* Ask for a file name, and write the contents of the current buffer to that
 * file.
 *
 * Update the remembered file name and clear the buffer changed flag. This
 * handling of file names is different from the earlier versions, and is more
 * compatable with Gosling EMACS than with ITS EMACS.
 *
 * If the file already exists or is not writable Emacs will not do the write.
 * If the count is positive then an attempt to write.
 *
 * The default count is unset.
 *
 * Command:  write-file
 * Bound to: C-x C-w
 *
 * Return: TRUE, FALSE, ABORT
 * TODO
 */
int	filewrite(int f, Cnt n)
{
	int	force = FALSE;

	if(f && n > 0)
		force = TRUE;

	return(bufRegWrite("Write file: ", force, WOUT_NONE));
}

/* Ask for a file name and write the current region to that file.
 * See writefile()
 *
 * If the file already exists or is not writable Emacs will not do the write.
 * If the count is positive then an attempt to write.
 *
 * The default count is unset.
 *
 * Command:  write-region
 * Bound to: C-x W
 *
 * Return: TRUE, FALSE, ABORT
 * TODO
 */
int	regionwrite(int f, Cnt n)
{
	int	force = FALSE;

	// Does a region exist ? Also a few other things
	if(getregion((REGION*)NULL) != TRUE)
		return(FALSE);

	if(f && n > 0)
		force = TRUE;

	return(bufRegWrite("Write region to file: ", force, WOUT_REGION));
}

/* Write a buffer or region to a file.
 *
 * prompt	The prompt for the user
 * force	TRUE if the write is to be forced (if possible) if
 *		the file permissions to not allow a write.
 * flags	This is where you can ask for the region only to be written.
 *		WOUT_xxx
 *
 * Used by filewrite() & regionwrite()
 *
 * Return:
 *	TRUE
 *	FALSE - nothing entered at prompt
 *	ABORT - user aborts the prompt - C-g
 *	fileDoWrite()
 */
int	bufRegWrite(const char* prompt, int force, int flags)
{
	int	s;

	if((s = mlprompt(prompt, MLXFILE, bmlFileHistp)) != TRUE)
		return(s);

	return(fileDoWrite(ltext(bmlFileHistp->b_dotp), force, flags));
}

/* Write the buffer to the end of a file.
 *
 * The file name is prompted for.
 * Create the file if it does not exist.
 *
 * The count is ignored. A default is not set
 *
 * Command:  append-file
 * Bound to: C-x A
 *
 * Return: TRUE, FALSE, ABORT
 *	FALSE - nothing entered at prompt
 *	ABORT - user aborts the prompt - C-g
 * TODO
 */
int	fileappend(int f, Cnt n)
{
	register int	s;

	if((s = mlprompt("Append file: ", MLXFILE, bmlFileHistp)) != TRUE)
		return(s);

	return(fileDoWrite(ltext(bmlFileHistp->b_dotp), 1, WOUT_APPEND));
}

/* Do the work of filewrite()/fileappend()
 *
 * fname	Name of the file to write to
 *		Set the buffer's b_fname to fname to this, unless WOUT_APPEND.
 * force	Write even if no write permission, try create/rename
 * flags	As for writeout(): can specify append or write region
 *		WOUT_xxx
 *		Unless WOUT_APPEND: set b_fname & unset: BFCHG flag (+ others)
 *
 * If WOUT_APPEND then force MUST be 1 - ie ignore that it already exists.
 *
 * Return: TRUE,
 *	FALSE - file exists and not force
 *	ABORT - file is nasty
 */
int	fileDoWrite(const BYTE* fname, int force, int flags)
{
	register int	s;
	BUFFER*		bp = bheadp;

#if	UNIX
	char		c;

	if(isnasty(fname, &c))
		return(ABORT);

	if(force == 0 && access(fname, W_OK) == 0) {
		mlwrite("%W""%s exists, use positive count to overwrite", fname);
		return(FALSE);
	}
#endif
	if((s = writeout(fname, flags)) == TRUE) {
		// Unless append set file name & things:
		if(flags != WOUT_APPEND && flags != WOUT_REGION) {
#if	UNIX
			/* If we couldnt write to a file when we read it in, the
			 * buffer will have been put into VIEW mode. Since we can
			 * now write the file, knock view mode off.
			 */
			if((curbp->b_flag & BFREAD) && (curbp->b_flag & BIMMUTABLE) == 0)
				curbp->b_mode &= ~MDVIEW;

#endif
			setFnames(curbp, fname);
			curbp->b_flag &= ~(BFCHG | BRENEWED | BFROMSAVE);
		}

		upmode(curbp, WFMODE);		  /* Update mode lines.   */
	}
	/* Check to see if the new filename conflicts with the filenames for
	 * any other buffer and produce a warning if so.
	 */
	for(/*void*/;bp != (BUFFER*) NULL;bp = bp->b_bufp)
		if(bp != curbp
#if	UNIX
			&& curbp->b_stino == bp->b_stino
			&& curbp->b_stino != (ino_t)-1
			&& curbp->b_stdev == bp->b_stdev
			&& curbp->b_stdev != (dev_t)-1) {
#else
			&& strcmp(curbp->b_fname, bp->b_fname) == 0) {
#endif
			mlwrite("%WWarning: file %s: multiple buffers '%s' '%s", 
				curbp->b_fnameClean, curbp->b_bname, bp->b_bname);
			break;
		}
	return(s);
}

#if	0
/* Removed, use pipe-to-command in a narrow
 * This allowed piping parts of lines - lose that slight extra flexibility
 *
 * PipeBuffer() moved to spawn.c as PipeToCommand, and command
 * pipe-buffer-to-command renamed pipe-to-command
 */

/* Pipe the region to a shell command.
 * The user is prompted R| to distinguish from piping the buffer.
 *
 * Lines are sent down the pipe with lines ending \n. If the buffer
 * is in DOS mode a \r is not added.
 *
 * The region may consist of part lines.
 * If the start of region is not at the start of a line, the initial
 * characters are not output.
 * If the end of region is not at the start of a line, the previous
 * line will not be output with an ending newline.
 *
 * After the command the user is prompted before Emacs continues.
 * This prompt does not happen if this is called from a macro
 * unless the count is set.
 * 
 * The default count is not set.
 *
 * Command:  pipe-region-to-command
 * Bound to: M-|
 *
 * Return: TRUE
 *	   FALSE
 */
int	PipeRegion(int f, Cnt n)
{
	REGION	reg;	// Current region

	if(getregion(&reg) == FALSE)
		return(FALSE);

	return(regionToPipe(&reg, "R|", f));
}
#endif	// Removed

/* Save the contents of the current buffer in its associatd file.
 * No nothing if nothing has changed (this may be a bug, not a feature).
 * Error if there is no remembered file name for the buffer.
 *
 * This does the work for save-many-buffers
 * This does the work for quick-exit
 * May get called by suspend-emacs
 *
 * The count is ignored. The default count is not set.
 * 
 * Command:  save-file
 * Bound to: C-x C-s   C-x s   FN9
 *
 * Return: TODO
 */
int	filesave(int f, Cnt n)
{
	register int	s;

	if((curbp->b_flag&BFCHG) == 0)		/* Return, no changes.	*/
		return(TRUE);

	if(curbp->b_fname[0] == '\0') {		/* Must have a name.	*/
		mlwrite("%WNo file name");
		return(FALSE);
	}

	/* complain about truncated files */
	if((curbp->b_flag&BFTRUNC) != 0) {
		if(mlyesno("Truncated file. Write it out [y/n]? ") == FALSE) {
			mlwrite(MsgAborted);
			return(FALSE);
		}
	}

	/* Check if a .emacssave was touched since the file was read
	 * in to the buffer. If so as if the write should still happen.
	 *
	 * Have to test here since the .emacssave might have been written by another
	 * instance of Emacs.
	 */
	if(NoticeEmacsSave && curbp->b_mtime != (Time_T)-1 && (curbp->b_flag & BFROMSAVE) == 0 &&
	   findFile(curbp->b_fname, fT_Save)) {
		struct	stat	saveStat;	// Stat of save file
		fstat(fileno(ffp), &saveStat);
		ffclose();

		if(saveStat.st_mtime > curbp->b_mtime) {
			if(mlyesno("A " EMACS_SAVE_SUFFIX
					" was written after this buffer was read. Write it out [y/n] ?") != TRUE) {
				mlwrite(MsgAborted);
				return(FALSE);
			}
		}
		
	}

	if((s = writeout(curbp->b_fname, WOUT_NONE)) == TRUE) {
		curbp->b_flag &= ~(BFCHG | BRENEWED);
		upmode(curbp, WFMODE);		/* Update mode lines.	*/

		// Offer to remove an unwanted/old save file:
		if((curbp->b_flag & BFROMSAVE) && findFile(curbp->b_fname, fT_Save)) {
			/* The above opens it, which we do not want but it
			 * does tell us if it still exists.
			 */
			ffclose();

			if(mlyesno("Saved buffer was read from %s, remove it [y/n]? ", ttext(foundFiles)) == TRUE) {
				/* The messages below will prob not be seen without the sleep.
				 */
				if(unlink(ttext(foundFiles)) == 0) {
					mlwrite("%F""%I""%s removed", ttext(foundFiles));
					curbp->b_flag &= ~(BFROMSAVE);
				} else
					mlwrite("%F""%E""%s not removed: %e", ttext(foundFiles), errno);
			}

			upmode(curbp, WFMODE);		/* Update mode lines.	*/
		}
	}

	return(s);
}

/* Open an emergency save file for buffer bp in directory h+dir.
 * Try to do so using the file name, if not the buffer name.
 *
 * Ie the directory name is the concatenation of h & dir.
 *
 * NB: file name may be empty.
 *
 * If the open succeeds say so to stderr.
 *
 * Return:
 *	>= 0	file descriptor
 *	-1	file open error
 *	-2	name too long
 */
int	openSave(const BUFFER* bp, const char* h, const char* dir)
{
	int	fd;
	char*	name;		// Name to use - does not have '/' in it
	size_t	len = strlen(dir) + strlen(h) + SaveSuffixLen + 3;
	char	fname[255];

	// Get file name:
	if(bp->b_fname[0] == '\0') {
		name = bp->b_bname;	// File name empty, use buffer name
	} else {
		if((name = strrchr(bp->b_fname, '/')))
			name++;
		else
			name = bp->b_fname;
	}

	// Will the file name fit ? If not try the buffer name:
	if(len + strlen(name) > sizeof(fname) &&
	   len + strlen(name = bp->b_bname) > sizeof(fname)) {
		ErrorWriteList("Directory name too long: ", dir, "\n", (char*)NULL);
		return(-2);
	}

	strcat(strcat(strcat(strcat(strcpy(fname, h), dir), "/"), name), SaveSuffix);

	/* We might create the file where some ne'er-do-well could see it.
	 * Protect potential privacy & secrets - read only to the user on create.
	 */
	if((fd = open(fname, O_WRONLY | O_CREAT, S_IRUSR | S_IWUSR)) >= 0)
		ErrorWriteList("Opened: ", fname, "\n", (char*)NULL);

	return(fd);
}

/* Something has gone wrong - quick/dirty buffer save.
 * Save as file_name.emacssave into the current directory,
 * if that cannot be opened/write then save into $HOME.
 *
 * If %_path-save exists, use that as a list of directories
 * to try to save in.
 * A directory name can start with "~" or "$HOME" which will
 * be replaced by the value in home.
 *
 * If file-name makes it too long, try buffer_name.
 * If the file already exists - append to it.
 *
 * Do NOT use stdio - it will try to malloc(), things are bad,
 * we don't know if a malloc() will work. We do not care how
 * slow this might be.
 *
 * Note that initSize is off_t as we do not use stdio, contrast to
 * b_fileEnd which is now also off_t
 *
 * Encrypted buffers are not written.
 * 
 * Write a commentary to stderr.
 *
 * Return TRUE if all went well.
 */
void	bquicksave(BUFFER* bp)
{
	int	fd;			// File descriptor
	BufLns	lines = 0;
	int	Errno = 0;		// Errno value
	LINE*	lp;
	char*	lineEnd;
	off_t	initSize;		// Initial file size
	char	defPath[] = "."PATHSTR"~"; // Default save path - must be writable memory
	char*	path;
	char*	pathKeep;
	char	msg[100];

	ErrorWriteList("Attempting save of buffer: ", bp->b_bname, (char*)NULL);
	if(bp->b_fnameClean[0] != '\0')
		ErrorWriteList(" file: ", bp->b_fnameClean, (char*)NULL);
	ErrorWrite("\n");

	if(bp->b_mode & MDCRYPT) {
		ErrorWrite("Encrypted buffer - not saving\n\n");
		return;
	}

	if((path = gtusr("_path-save")) == (char*)NULL)
		path = defPath;
	pathKeep = path;

	while(1) {
		// If find directory separator - break path:
		char*	sep = strchr(path, PATHCHR);
		if(sep)
			*sep = '\0';

		// Home directory ?
		char*	h = "";
		if(home[0] != '\0') {
			if(path[0] == '~') {
				path++;
				h = home;
			} else if(strncmp(path, "$HOME", 5) == 0) {
				path += 5;
				h = home;
			}
		}

		fd = openSave(bp, h, path);

		// Repair any path separation:
		if(sep)
			*sep = PATHCHR;

		if(fd >= 0)	// We opened something, use it
			break;

		if(sep == (char*)NULL) {
			ErrorWriteList("Cannot write in any of ", pathKeep, "\n", (char*)NULL);
			return;
		}

		path = sep + 1;	// Next directory in path
	}
   
	// Move to file end - if not a new file:
	if((initSize = lseek(fd, 0, SEEK_END)) != -1) {
		sprintf(msg, "Appending to %s file. Current size %ld\n", SaveSuffix, (long)initSize);
		ErrorWrite(msg);
	}

	lineEnd = bp->b_mode & MDDOS ? "\r\n" : "\n";

	// Write the lines out:
	for(lp = lforw(bp->b_linep); lp != bp->b_linep; lp = lforw(lp)) {
		BYTE*	txt = ltext(lp);
		size_t	len = (size_t)lused(lp);
		ssize_t	wrote;

		while(len > 0) {
			if((wrote = write(fd, txt, len)) < 0) {
				Errno = errno;
				goto close;
			}
			len -= (size_t)wrote;
			txt += wrote;
		}

		IGNORE(write(fd, lineEnd, strlen(lineEnd)))
		lines++;
	}

close:	if(close(fd) == -1)
		Errno = errno;

	// Say what happened:
	sprintf(msg, "Buffer written %ld lines", (long)lines);
	ErrorWrite(msg);
	// I'm not happy with this use of strerror(), it might generate another signal
	if(Errno)
		ErrorWriteList(" with error: ", strerror(Errno), (char*)NULL);

	ErrorWrite("\n\n");
}

/* Save many buffers. Similar to killmany().
 *
 * This does the work for exit-emacs
 * Invisible buffers and VIEW mode buffers are ignored, even if modified.
 *
 * If a buffer does have a name unless the count is set in which case this acts
 * like write-file (a name is prompted for).
 *
 * The default count is not set.
 *
 * Command:  save-many-buffers
 * Bound to: unbound
 *
 * Return:
 *	TRUE if at least one buffer was saved
 */
int	savemany(int f, Cnt n)
{
	BUFFER*	bp;
	BUFFER*	cbufp;		/* current buffer 			*/
	int	status;
	int	done;		/* true if we've done something		*/

	cbufp = curbp;

	for(done = FALSE, bp = bheadp; bp != (BUFFER*) NULL; bp = bp->b_bufp) {

		/* Continue if buffer unmodified, in view mode, or no
		 * filename to write to
		 */
		if((bp->b_flag & BFCHG) == 0 || (bp->b_flag & (MDVIEW | BFINVS)))
			continue;

		/* Pass up buffers with no filenames unless we were called
		 * with an argument
		 */
		if(bp->b_fname[0] == '\0' && f == FALSE)
			continue;

		if((status = mlyesno((bp->b_fname[0] ?
					"Save buffer %s (file %s) [y/n]? " :
					"Save buffer %s (no file) [y/n]? "),
					bp->b_bname, bp->b_fnameClean)) == FALSE)
			continue;

		if(status == ABORT)
			return(status);

		if(swbuffer(bp) == TRUE) {
			if(bp->b_fname[0])
				(void) filesave(f, n);
			else {
				(void) filewrite(f, n);
			}
		}

		done = TRUE;
	}

	if(!done)
		mlwrite("%INo buffers needed saving");
	else {
		swbuffer(cbufp);
		mlerase();
	}

	return(done);
}

/* Re-read the associated file into a buffer.
 *
 * Prompt for a buffer name (return defaults to the current buffer), kill the
 * named buffer and then reread the corresponding file.
 *
 * If a non zero count is supplied, renew the buffer even if it is modified.
 *
 * The default count is unset.
 *
 * Command:  renew-buffer
 * Bound to: C-x .
 *
 * Return: TODO   as readin()
 *	FALSE - No file associated with the buffer
 *		Buffer name entered at prompt does not exist
 *		The buffer is being executed
 *	ABORT - User aborts the prompt - C-g
 */
int	renew(int f, Cnt n)
{
	int	s;
	BYTE	prompt[MAXCOL + 1];
	BUFFER*	bp;

	if(curbp->b_fname[0] != '\0')
		sprintf(prompt, "Renew buffer [%s]: ", curbp->b_bname);
	else
		sprintf(prompt, "Renew buffer: ");

	if((s = mlprompt(prompt, MLXBUF | MLU8LEN, bmlBufHistp)) == ABORT)
		return(s);

	// User did not type anything, use curent buffer
	if(s == FALSE)
		lcpys(bmlFileHistp->b_dotp, curbp->b_bname);

	if((bp = bfindCheck(ltext(bmlFileHistp->b_dotp), BFNNONE)) == (BUFFER*)NULL)
		return(FALSE);

	// Do not change an executing buffer
	if(bp->b_exec != 0) {
		mlwrite("%ECannot renew an executing buffer");
		return(FALSE);
	}

	/* Final check to see if there really IS an associated file */
	if(bp->b_fname[0] == '\0') {
		mlwrite("%ENo file associated with buffer: %l", bmlFileHistp->b_dotp);
		return(FALSE);
	}

	// Remove the buffer modification flag if the count is not zero
	if(f && n && (bp->b_flag & BFCHG))
		bp->b_flag &= ~(BFCHG | BRENEWED | BFROMSAVE);

	return(renewbuf(bp, FALSE));
}

/* Renew the buffer bp.
 *
 * The buffer is cleared and the file read back in.
 *
 * readTail	If TRUE: read the new lines that have been appended to the
 *		file since it was last read. Do not read the whole file.
 *		Read from b_fileEnd. If the file is smaller than that assume
 *		that it has been rewritten, refresh the entire buffer.
 *		The file is not kept open, so if it is moved - whatever
 *		new contents will be read in.
 *
 * Keep cursor & mark position if the buffer is displayed in curwp.
 * If it is not curwp then is it displayed in another window (use the first one found) ?
 * If not displayed then the buffer's own positions.
 * That is all a lot of different possibilities.
 *
 * If I wanted to do this perfectly: I would keep separate topLine, "." & mark information
 * for every window that had the buffers displayed - too much work.
 * The chances are that it is only displayed in one window - most of the time anyway.
 *
 * Attempt is made to preserve Alpha and Numeric marks.
 * But at the end of this they will all point to something valid.
 *
 * But BEWARE - this function is a pain as it can be called any time that
 * update() is called - which means any time tgetc() is called. Thus a long
 * running function that has a long lasting LINE* value into a buffer - it
 * could get into trouble as that value ends up becoming invalid if the buffer
 * is zapped & reloaded.
 * If long lasting LINE*s are needed use high numbered Numeric marks - see isearch().
 *
 * Return:
 * 	TRUE   OK - file read in
 *	FALSE  cannot get a lock on the file
 *	       cannot open the file
 *	       prompt error
 *	ABORT  cannot access the file
 *	       it is a nasty file
 *	       prompt error
 *	FAILED
 *	       prompt error
 */
int	renewbuf(BUFFER* bp, int readTail)
{
	BufPos	Pos;		/* Where to store buffer position */
	int	ret;

	/* readin() plays with the current buffer. So temporarily swap curbp to bp.
	 * This will be needed when a buffer has RENEW mode - renewed when it is
	 * not the current buffer.
	 */
	BUFFER*	saveBp = curbp;
	curbp = bp;

	RecordDotAndMark(&Pos, lforw(bp->b_linep), bp->b_linep);

	/* Read the file back in.
	 * bclear() not needed, readin() does it, checking
	 * if the user is OK to re-read a modified buffer.
	 * bclear() junks the buffer contents but leaves it in the buffer
	 * chain, file names, etc.
	 */

	if((ret = readin(bp->b_fname, FALSE, readTail, FALSE)) == TRUE) {
		RestoreDotAndMark(&Pos, lforw(bp->b_linep), bp->b_linep);

		/* Remove out of date flag in mode line */
		bp->b_flag &= ~BFOLD;

#if	UNIX
		if(f_funnyfile)
			bp->b_flag |= BFWEIRD;	/* Remember if its weird*/
#endif
	}

	curbp = saveBp;

	return(ret);
}

/* Check if a buffer needs to be renewed from disk, maybe read it in.
 * This will make the check regardless of the RENEW buffer flag/mode.
 *
 * This does no check if:
 * * there is not an associated file
 * * the file is not a regular file
 * * the buffer is not active
 *
 * If the buffer is out of date:
 * * if the buffer is in RENEW mode it will be read in.
 *   The bell is rung if RenewBell ($bell-on-renew).
 *   The buffer name appended to playText2.
 * * if the buffer is NOT in RENEW mode the buffer name is appended to
 *   playText and the BFOLD flag set.
 *   This allows reporting of out of date buffers
 *
 * A buffer will not be read in (even if RENEW mode) if the ffp is in use.
 * This should not happen, but has in the past. This would result in ffp
 * being opened again & 'closed twice' - and causes a crash.
 * In this case just report it as out of date.
 *
 * An executing buffer (macro) will not be renewed.
 *
 * Return 2 buffer was out of date - was read in
 *	  1 if the buffer was out of date - not read in
 * 	  0 if it was in date or should not be checked (for any reason).
 */
int	CheckBufferRenew(BUFFER* bp)
{
#if	UNIX
	Time_T	modtime;	// mtime before we update/check-again
	ino_t	inode;
	dev_t	stdev;

	// If there is not a filename - skip this one
	if(bp->b_fname[0] == '\0')
		return(0);

	/* If the file has not been looked at - skip
	 * This means that we will not notice changes to files that we have not
	 * yet read it - I can't see that that is a problem.
	 */
	if(bp->b_active == FALSE)
		return(0);

	// Do not change an executing buffer
	if(bp->b_exec != 0)
		return(0);

	// Only check age on regular files:
	if(bp->b_lsType != '-' && bp->b_lsType != '.')
		return(0);

	/* What was the mod time when we last looked, then get new mod time.
	 * Also do this for inode & device (in case it has been replaced).
	 */
	modtime = bp->b_mtime;
	inode = bp->b_stino;
	stdev = bp->b_stdev;

	UpdateMetaData(bp, bp->b_fname);

	/* It did not exist and still does not exist then forget it.
	 * If it has appeared/disappeared then report as changed.
	 */
	if(bp->b_mtime == (Time_T)-1 && modtime == -1)
		return(0);

	// Nothing changed ?
	if(modtime == bp->b_mtime && inode == bp->b_stino &&
	   stdev == bp->b_stdev)
		return(0);

	/* Need to update something on the mode lines
	 * Quite what we will see shortly below.
	 */
	upmode(bp, WFMODE);

	/* Sorry, its out of date. Work out what to do.
	 */
	if((bp->b_mode & MDRENEW) && bp->b_mtime != (Time_T)-1 && ffp == (FILE*)NULL) {

		off_t	size = fileSize(bp->b_fname);

		int	readNew = 0;	// True means read what has been appended

		// Can we give the hint that it is a log file that has grown or something ?
		readNew = ((bp->b_flag & BRENEWAPPEND) && size > bp->b_fileEnd &&
			(bp->b_mode & MDCRYPT) == 0);

		if(renewbuf(bp, readNew) == TRUE && RenewBell) {
			// Ring the bell to say a buffer renewed
			TTbeep();
			mlFlush();
		}
		bp->b_flag |= BRENEWED;		// Set status shown by list-buffers

		// 2 bit counter of how often the buffer has been renewed:
		if((bp->b_flag & BRENEWEDAGAIN) != 0)
			bp->b_flag ^= BRENEWEDAGAIN1;	// Top bit of 2 bit counter - renew status
		bp->b_flag ^= BRENEWEDAGAIN;	// Odd/even renew status

		tcats(playText2, bp->b_bname);
		tcats(playText2, " ");

		return(2);	// Buffer was out of date; was read in
	}

	/* Its not in RENEW mode, add it to the list.
	 * Reset the modify time to what it was when it was last read,
	 * thus the buffer will be re-read when ever it is put into RENEW mode.
	 */
	tcats(playText, bp->b_bname);
	tcats(playText, " ");
	bp->b_mtime = modtime;
	bp->b_flag |= BFOLD;

	return(1);	// Buffer was out of date; not read in
#else
	return(0);
#endif
}

/* This function goes through all the buffers and if it finds an out of date
 * buffer, (where out of date means that the file was read into the buffer at
 * an earlier time than the current modification time) then if it is in renew
 * mode, it automatically reads it back in, or if it is not in renew mode it
 * prints a warning and or's in BFOLD with the flags field of the buffer.
 *
 * If RenewModeOnly only check buffers in RENEW mode that are visible.
 * We then read it in if needed.
 *
 * Returns the number of out-of-date buffers that were not auto read back in.
 */
int	checkage(int RenewModeOnly)
{
#if	UNIX
	BUFFER*	bp;
	int	numOutOfDate = 0;
	int	numReadIn = 0;

	/* It is just about possible that this could be recursively called when
	 * a message in a read route gets flushed.
	 * Stop that happening.
	 */
	static	int	here;	// Is this being executed ?
	if(here)
		return(0);
	here++;

	/* It is important to do this first, so that if any of the subsequent
	 * routines call update(), we wont end up recursing.
	 */
	agetime = FALSE;
	statbufInvalid = 1;	// Things might have changed since last did a stat()

	tcpys(playText,  "Old buffers: ");
	tcpys(playText2, "Renewed: ");

	for(bp = bheadp; bp != (BUFFER*) NULL; bp = bp->b_bufp) {

		/* Only check buffers in RENEW mode ?
		 * This might be called a lot. Its too expensive to check
		 * files that we have not been asked to.
		 */
		if(RenewModeOnly && ((bp->b_mode & MDRENEW) == 0 || bp->b_nwnd == 0))
			continue;

		switch(CheckBufferRenew(bp)) {
		  case 2:
			numReadIn++;
			break;
		  case 1:
			numOutOfDate++;
			break;
		  // case 0 - not out of date
		}
		
	}

	// Say what happened - if anything
	// If the prompt line is busy this might not be seen

	if(numOutOfDate == 0)
		tsetempty(playText);
	if(numReadIn == 0)
		tsetempty(playText2);

	if(numOutOfDate + numReadIn != 0)
		mlwrite("%W""%t%t", playText, playText2);

	here--;

	return(numOutOfDate);
#else
	return(0);
#endif
}


/* This function performs the details of file writing.
 *
 * Uses the file management routines in the "fileio.c" package. The number of
 * lines written is displayed. Sadly, it looks inside a LINE; provide a macro
 * for this. Most of the grief is error checking of some sort.
 *
 * fn	    File name
 * flags    One or more of (can be ORed):
 *	WOUT_NONE	Nothing special - just write the whole buffer
 *	WOUT_APPEND	Append the narrow to the file
 *	WOUT_REGION	Write the region, not the whole buffer
 *			If this is specified a region MUST exist
 *
 * If append: open to write at the end of existing content.
 * Must not try to append if encrypting.
 *
 * Used by the non UNIX version of filter-buffer
 *
 * Return:
 *	TRUE  - it all worked
 *	FALSE - prompt error, cannot open file
 *	ABORT - file modified since read in, prompt error
 *	FAILED- prompt error
 */
int	writeout(const BYTE* fn, int flags)
{
	register int	s;
	register LINE*	lp;
	register LINE*	nlp;
	BufLns		nline;		// # lines written
	int		DosLine = curbp->b_mode & MDDOS;
	REGION		reg;		// Region to write

#if	ST520 || ADDCR
	DosLine = 1;
#endif

#if	UNIX
	int	new_file = access(fn, F_OK);

	/* Check that the file has not been modified since it was read it in.
	 * This is a bit of a problem since the info we require is hidden
	 * in a buffer structure which may not necessarily be the current
	 * buffer. If it is not in the current buffer then don't bother to
	 * look for it.
	 */
	if(strcmp(curbp->b_fname, fn) == 0) {
		/* We are only interested in this file/buffer
		 * but it is easiest to check them all.
		 */
		checkage(0);

		/* File modified since buffer read it ?
		 * If b_lsType == '.' the file no longer exists.
		 */
		if((curbp->b_flag & BFOLD) != 0 &&
		   curbp->b_lsType != '.'	&&	// File removed - don't prompt about 'modified'
		   mlyesno("%s modified, really write [y/n]? ", fn) != TRUE)
			return(ABORT);
	}

#endif
#if	CRYPT
	// Verify the key. It would be a disaster if the user mistyped the key:
	if((s = resetkey(1)) != TRUE)
		return(s);
#endif

	// ffwopen() tells the user what we are doing

	/* turn off ALL keyboard translation in case we get a dos error */
	TTkcloseM;

	if((s = ffwopen(fn, (flags & WOUT_APPEND))) != FIOSUC) {	/* Open writes message	*/
		TTkopenM;
		return(FALSE);
	}

	// Separate Writing message - in case writing takes a long time
	mlwrite("%I""%FWriting%s: %s", ((flags & WOUT_APPEND) ? " append" : ""), fn);

	nline = 0;				/* Number of lines.	*/

	/* Create region consisting of the entire buffer.
	 *
	 * If writing the region - narrow it down to that.
	 * Thus if a region does not exist - we don't fall over.
	 * This default should not happen.
	 */
	reg.r_linep = lforw(curbp->b_linep);
	reg.r_offset = 0;
	reg.r_endLinep = curbp->b_linep;
	reg.r_endOffset = 0;

	// Append: make this the narrow
	if(flags & WOUT_APPEND) {
		reg.r_linep = lforw(curbp->b_btopline);
		reg.r_endLinep = curbp->b_abotline;
	}

	if(flags & WOUT_REGION)
		getregion(&reg);	// A region does exist - right ?

	lp = reg.r_linep;		// First line
	LineLn	stOff = reg.r_offset;	// Where, in the line, we write from

	// Write the region:
	while(lp != reg.r_endLinep) {
		nlp = lforw(lp);

		// Take care to avoid a \n in NONL - if there is nothing on the last line
		if((s = ffputline(lgetcaddr(lp, stOff), (LineLn)(lused(lp) - stOff),
		     (nlp == reg.r_endLinep && reg.r_endOffset == 0 && (curbp->b_mode & MDNONL)),
		     DosLine)) != FIOSUC)
			break;

		++nline;
		lp = nlp;
		stOff = 0;
	}

	// Any last/partial line of region to write ?
	if(s == FIOSUC && reg.r_endOffset > stOff)
		s = ffputline(lgetcaddr(lp, stOff), reg.r_endOffset- stOff,
			(curbp->b_mode & MDNONL), DosLine);

	if(s == FIOSUC) {		       /* No write error.	*/
#if	UNIX
		struct	stat	writeStat;	// Stat of stream just written
		fstat(fileno(ffp), &writeStat);
		numLinks = (int)(writeStat.st_nlink);
#endif

		s = ffclose();
		if(s == FIOSUC) {	       /* No close error.	*/
#if	UNIX
			if(new_file)
				mlwrite("%INew file %s, Wrote %D line%s",
					fn, nline, nline == 1 ? "" : "s");
			else
#endif
				mlwrite("%IWrote%s %D line%s", ((flags & WOUT_APPEND) ? " append" : ""),
						   nline, (nline == 1 ? "" : "s"));

			sayLinks();
		}
	} else					/* Ignore close error	*/
		ffclose();			/* if a write error.	*/
	TTkopenM;

#if	UNIX
	if(flags != WOUT_APPEND && flags != WOUT_REGION) {
		// Remove modified flag and RENEWed flag:
		curbp->b_flag &= ~(BFOLD | BRENEWED);
		statbufInvalid = 1;				// We just wrote to it
		UpdateMetaData(curbp, fn);
	}
#endif
	if(s != FIOSUC)				/* Some sort of error.	*/
		return(FALSE);

#if	UNIX
	if(w_relink == TRUE) {
		mlwrite("%WWarning %s changed owner", fn);
		w_relink = FALSE;
	}

#if	0
	/* This can hang the editor for a long time. See fileio.c - see fsync() in ffclose() */
	if(syncmc == TRUE)
		sync();
#endif
#endif
	return(TRUE);
}

/* Remove the file name associated with the current buffer.
 * Do this by setting it to the empty string.
 *
 * The count is ignored.
 *
 * Command:  forget-file-name
 * Bound to: unbound
 *
 * Return: TRUE - always
 */
int	forgetFilename(int f, Cnt n)
{
	setFnames(curbp, "");
	upmode(curbp, WFMODE);			/* Update mode line */

	return(TRUE);
}

#if	UNIX
/* Update bp's inode information from file fname
 */
void	UpdateMetaData(BUFFER* bp, const char* fname)
{
	if(CacheStat(fname) == -1) {
		ClearInodeInfo(bp);
	} else {
		bp->b_mtime = statbuf.st_mtime;
		bp->b_stdev = statbuf.st_dev;
		bp->b_stino = (ino_t) statbuf.st_ino;
		getFileType(fname, &bp->b_lsType);
	}
}
#endif

/* The command allows the user to modify the file name associated with
 * the current buffer.
 *
 * It is like the "f" command in UNIX "ed". The operation is not simple.
 * 
 * The modification times, inode number and st_dev fields of the new file have
 * to be updated in the buffer structure. This is not to mention the
 * possibility of allowing the user to inadvertantly point the new filename at
 * a file that they might actually want.
 *
 * If the new name is of a file that exists the user is prompted to check.
 * This prompt is not done if the count is non zero.
 *
 * The default count is not set.
 *
 * Command:  change-file-name
 * Bound to: C-x n
 *
 * Return: TODO
 *	FALSE - nothing entered at prompt
 *	ABORT - user aborts the prompt - C-g
 */
int	filename(int f, Cnt n)
{
	register int	s;
	TEXT*		fname = playText;
	TEXT*		dir = playDir;
#if	UNIX
	char		c;
#endif

	if((s = mlprompt("New filename: ", MLXFILE, bmlFileHistp)) == ABORT)
		return(s);

	tcpyl(fname, bmlFileHistp->b_dotp);
#if	UNIX
	/* Abort if error & the error is not that the file does not exist (c = .).
	 * Ie: OK if the file is a regular file or it does not exist (but can
	 * access the directory).
	 */
	if(s != FALSE) {
		char*	msg = getFileType(ttext(fname), &c);

		if(c != '-' && c != '.') {
			if( !msg)
				msg = "missing directory";
			mlwrite("%ECannot use %t as is: %s", fname, msg);
			return(ABORT);
		}
	}

	if(s != FALSE) {
		// Set the name of the directory in dir
		BYTE*	sl;
		tcpyt(dir, fname);
		if((sl = strrchr(ttext(dir), '/')) != (BYTE*)NULL)
			tsetused(dir, (LineLn)(sl - ttext(dir)));
		else
			tcpys(dir, ".");
	}

	if(s != FALSE && access(ttext(fname), F_OK) != -1) {
		/* The file exists. Work out what to do.
		 */
		if(access(ttext(fname), W_OK) == -1 &&
				( access(ttext(dir), W_OK) == -1 ||
				 (access(ttext(dir), W_OK) != -1 && !relinking) )) {
			mlwrite("%E""%t exists and is not writeable", fname);
			return(ABORT);
		}
		if(!(f && n) && mlyesno("File exists - are you sure [y/n]? ") != TRUE)
			return(ABORT);

		/* It does seem to be what the user wants, update the
		 * modification time and inode numbers in the buffer structure.
		 */
		UpdateMetaData(curbp, ttext(fname));

		if(access(ttext(fname), W_OK) == 1 && access(ttext(dir), W_OK) == -1)
			curbp->b_flag |= BFWEIRD;

	} else if(s == TRUE) {
		/* The file does not exist, zap the contents of the
		 * modification time, inode and stdev fields.
		 */
		ClearInodeInfo(curbp);
	}

	curbp->b_flag &= ~BFOLD;		/* buffer is no longer old */

	/* and if the file didnt exist .....
	 */
	if(s != FALSE && access(ttext(dir), W_OK) == -1) {
		mlwrite("%WDirectory %t not writable", dir);
		curbp->b_flag |= BFREAD;
		curbp->b_mode |= MDVIEW;
	} else {
#endif
		if((curbp->b_flag & BIMMUTABLE) == 0) {
			curbp->b_mode &= ~MDVIEW;	/* no longer read only mode */
			curbp->b_flag &= ~BFREAD;
		}
#if	UNIX
	}
#endif

	setFnames(curbp, (s == FALSE ? (BYTE*)"" : ttext(fname)));

	upmode(curbp, WFMODE);			/* Update mode lines.	*/

	return(TRUE);
}

/* Open a file and insert into the current position within the current buffer.
 *
 * Called by insert file command and by insert from pipe command.
 * Return the final status of the read.
 *
 * Increment b_numLines by the number of lines read in.
 *
 * Too much duplication with readin()
 */
int	ifile(BYTE fname[])
{
	int	s;		// Success code

#if	CRYPT
	// Get the Crypt key before opening.
	// We don't force the key to be verified, the user will try again if it is wrong
	if((s = resetkey(0)) != TRUE)
		return(s);
#endif

	// Open the file:
	if(ffropen(fname) == FIOFNF) {		// Open failed
#if	UNIX
		mlwrite("%E""%s: %e", fname, errno);
#else
		mlwrite("%E""%s: No such file", fname);
#endif
		return(FALSE);
	}

	return(istream(fname));
}

/* Insert a stream into the current position within the current buffer.
 * The stream could be a file or a pipe.
 *
 * what	something that can be used to tell the user what is being inserted
 *
 * Any narrow is NOT widened
 *
 * Increment b_numLines by the number of lines read in.
 * 
 * Close the stream.
 */
int	istream(const BYTE* what)
{
	BufLns	nline = 0;	// # lines read in
	LINE*	lnew;		// Line read from input
	BUFFER*	bp;
	int	s = FIOSUC;	// Success code

	bp = curbp;				/* Cheap.			*/

	/* We have changed
	 * Also: some user action to read it in, so remove the auto RENEWed flag
	 * (checkage will set it back).
	 */
	bp->b_flag &= ~(BFINVS | BRENEWED);

	setmark(FALSE, FALSE);		// Set mark before text that we insert

	mlwrite("%F""%IInserting %s", what);

	utf8_errclear();

	while((s = ffgetline(&lnew, bp->b_mode & MDUTF8)) == FIOSUC || s == FIOFUN) {
		bp->b_flag |= BFCHG;	/* We have changed */

		// Trim a \r if reading into a DOS mode buffer:
		if((bp->b_mode & MDDOS) && lused(lnew) && lgetc(lnew, lused(lnew) - 1) == RETCHAR)
			lsetused(lnew, lused(lnew) - 1);

		/* Insert 'X' at ".", result on lines below.
		 * FIOSUC		FIOFUN
		 * 12.34 ->		12.34 ->
		 * 12X			12X34
		 * 34
		 *
		 * .1234 ->		.1234 ->
		 * X			X1234
		 * .1234
		 *
		 * 1234. ->		1234. ->
		 * 1234X		1234X.
		 * .
		 *
		 * . ->			. ->
		 * X			X
		 * .			.
		 *
		 * Inserting at start of line or end of buffer ?
		 * Just stitch the new line in.
		 */
		if(curwp->w_dotp == curbp->b_linep ||
		   (curwp->w_doto == 0 && s != FIOFUN)) {

			LINE*	lbefore = lback(curwp->w_dotp);	// line previous to insert

			// Stitch the line:
			lsetback(curwp->w_dotp, lnew);
			lsetforw(lnew, curwp->w_dotp);
			lsetback(lnew, lbefore);
			lsetforw(lbefore, lnew);

			/* If at end of buffer read text not ended by \n
			 * then the buffer does not have a \n at its end:
			 */
			if(s == FIOFUN && curwp->w_dotp == curbp->b_linep)
				curbp->b_mode |= MDNONL;

			/* If this is the first line inserted then the mark goes
			 * at the start of the line just inserted:
			 */
			if(nline == 0)
				curwp->w_markp = lnew;
		} else {
			// Inserting in the middle or end of line
			if(insStrLine(curwp->w_dotp, curwp->w_doto, ltext(lnew),
					lused(lnew), 0) < 0) {
				// Out of memory - crunch & fail the insert
				FreeLine(lnew);
				s = FIOERR;
				break;
			}

			FixOffsets(curwp->w_dotp, curwp->w_doto, lused(lnew), curbp, 0);

			if(s != FIOFUN)		// Append line if one was read
				lnewline(0);

			FreeLine(lnew);		// Its text has been copied
		}

		++nline;

		if(ffutf8err)
			utf8_setmsg(nline);

		// Line unterminated by \n means success but stop
		if(s == FIOFUN) {
			s = FIOSUC;
			break;
		}
	}

	ffclose();		/* Ignore errors */

	// fileErrString() returns an error string and clears buffer flags
	// utf8_errmsg() will return any error message and
	// if there is an error clear the buffer curbp flag MDUTF8 & clear l_utf8 on all lines

	ReportRead("Inserted", s, nline);

	bp->b_numLines += nline;	// How many more lines in buffer
	adjustTopLineNum(nline);

	upmode(bp, WFHARD | WFMODE);	// Refresh all windows displaying the buffer

	/* Copy window parameters back to the buffer structure */
	SaveWindowPosToBuffer(curbp, curwp);

	if(s == FIOERR)				/* False if error.		*/
		return(FALSE);

	return(TRUE);
}

/* Do stat() on filename & cache the result at statbuf.
 * Set statbufInvalid TRUE to invalidate the cache, force another stat()
 *
 * Return the return code of the last stat(), good for checking errors,
 * errno will be correctly set.
 *
 * BEWARE: where ever you assign member st_ino to something of type ino_t
 * cast the value to ino_t.
 * For some reason on Minix 3.2.1 ino_t is unsigned int but .st_ino is unsigned long long
 */
int	CacheStat(const BYTE* filename)
{
	if(statbufInvalid || strcmp(filename, ttext(statFileName)) != 0) {
		statStatus = stat(filename, &statbuf);
		statErrno = errno;
		tcpys(statFileName, filename);
		statbufInvalid = 0;
	}

	errno = statErrno;

	return(statStatus);
}

/* Inspect file and learn things about the file type.
 * The purpose is to deal with non ordinary/regular files.
 *
 * *ftype will be updated to be a character like ls shows
 *
 * Return:
 * * a string that can be shown to the user to say what sort of file it is.
 * * NULL if the file can't be accessed, when ftype is:
 *   .	A non existing file
 *   /	The file cannot be accessed because of:
 *	- directory permissions
 *	- a parent directory does not exist
 *	- what looks like a directory is not, eg /dev/null/supposed-file
 *   ?	Don't know why
 * Maybe ! cannot access file ?-- which exists
 *
 */
char*	getFileType(const char* file, char* ftype)
{
	BYTE*	msg;		/* Relates to file type */

#if	UNIX
	/* If the file cannot be stat'd or its a regular file then it passes.
	 */
	if(CacheStat(file) != 0) {
		// Flag error:
		/* ENOENT could be because:
		 * * the file does not exist, set '.'
		 * * a directory in the path does not exist, set '/'
		 * EACESS can only be because of insufficent permissions on a
		 *   directory in the path, set '/'
		 * ENOTDIR eg /dev/null/supposed-file, set '/'
		 */
		*ftype = '?';
		if(errno == ENOTDIR || errno == EACCES)
			*ftype = '/';

		/* Might be due to a directory not being accessable or not existing.
		 * If we can see the directory then we know that it is that
		 * the file that does not exist, otherwise it a directory
		 * in the path does not exist.
		 * It is also possible that "." no longer exists!!
		 */
		if(errno == ENOENT) {
			int	statr;
			char*	dir = strrchr(file, '/');
			struct	stat	sb;

			// stat the file's parent directory:
			if(dir != (char*)NULL) {
				*dir = '\0';
				statr = stat(file, &sb);
				*dir = '/';
			} else
				statr = stat(".", &sb);

			if(statr == -1 && errno == ENOENT)
				*ftype = '/';
			else
				*ftype = '.';
		}

		return(NULL);
	}

	switch(statbuf.st_mode & S_IFMT) {
	  case S_IFREG:
		msg = "regular";
		*ftype = '-';
		break;
	  case S_IFDIR:
		msg = "directory";
		*ftype = 'd';
		break;
#ifdef	S_IFIFO
	  case S_IFIFO:
		msg = "FIFO";
		*ftype = 'p';
		break;
#endif
	  case S_IFCHR:
		msg = "character special";
		*ftype = 'c';
		break;
	  case S_IFBLK:
		msg = "block special";
		*ftype = 'b';
		break;
#ifdef	S_IFSOCK
	  case S_IFSOCK:
		msg = "socket";
		*ftype = 's';
		break;
#endif
#ifdef	S_IFDOOR
	  case S_IFDOOR:
		msg = "Solaris door";
		*ftype = 'D';
		break;
#endif
	  default:
		msg = "unknown file type";
		*ftype = '?';
		break;
	}

	return(msg);		/* Relates to file type */
#else
	return "";
#endif	
}

/* Is the file nasty - ie something we really shouldnt be editing.
 *
 * Set *ftype to one of '?/.!-sbcbpD'.
 *
 * One day this will see if the file is executable and ask for confirmation
 * before proceeding.
 *
 * Return:
 *	TRUE  we should not edit it, eg it is a directory or
 *	      a block/character special device.
 *	FALSE the file is not nasty
 *	      if:
 *	      * it is a text file (*ftype = '-') or
 *	      * it does not exist (*ftype = '/' or '.' or '?')
 */
int	isnasty(const BYTE* file, char* ftype)
{
#if	UNIX
	BYTE*	msg;		/* Relates to file type */

	if( !(msg = getFileType(file, ftype)) || *ftype == '-')
		return(FALSE);

	mlwrite("%E""%s is %s", file, msg);

	return(TRUE);
#else
	return(FALSE);
#endif
}

/* Return the size of a file.
 *
 * Return -1 on error.
 */
off_t	fileSize(const BYTE* fname)
{
#if	UNIX
	/*
	 * If the file cannot be stat'd or its a regular file then it passes.
	 */
	if(CacheStat(fname) != 0)
		return(-1);

	return(statbuf.st_size);
#else
	// I am being lazy, I could get it by: fopen(), fseek(to end), fseek()
	return(0);
#endif
}

int	isdir(const BYTE* fn)
{
#if	UNIX
	/*
	 * If the file cannot be stat'd or its a regular file then it passes.
	 */
	if(CacheStat(fn) != 0)
		return(FALSE);

	return(S_ISDIR(statbuf.st_mode));
#else
	return(FALSE);
#endif
}

/* Return the modification time of a file, cast to a long. By casting to a
 * long the nastiness of the differing types of "time_t" are hidden.
 *
 * Return -1 if the file cannot be stat'd or we're not running under UNIX.
 */
Time_T	modtime(const BYTE* fname)
{
#if	UNIX
	if(CacheStat(fname) == 0)
		return(statbuf.st_mtime);
#endif
	return(-1);
}

/* Return the mode of a file.
 * Include the setuid bits.
 */
int	filemode(const char* fn)
{
#if	UNIX
	if(fn == (char*) NULL || fn[0] == '\0')
		return(0);

	if(CacheStat(fn) == -1)
		return(0);

	return(statbuf.st_mode & 07777);
#else
	return(0);
#endif
}

/* Change directory, prompt for it first.
 *
 * The count is ignored. The default count is not set
 *
 * Command:  change-directory
 * Bound to: C-x d
 *
 * Return: TODO
 *	FALSE - nothing entered at prompt
 *	ABORT - user aborts the prompt - C-g
 */
int	changedir(int f, Cnt n)
{
#if	UNIX
	/* Change directory.
	 *
	 * Alain D D Williams, Sept 1986.
	 *
	 * Hacked by Tony Bamford to change relative filenames to absolute
	 * paths, Nov 1986.
	 */
	register int	s;
	TEXT*		nname = playText2;
	BYTE*		dirs[3];
	BUFFER*		bp  = curbp;		/* index buffer pointer		*/

	tcpyt(playText5, cwDir);			// Note where we are now

	printToText(playPrompt, "[cwd %t]  cd ", cwDir);
	if((s = mlprompt(ttext(playPrompt), MLXFILE, bmlFileHistp)) == ABORT || s == FALSE)
		return(s);

	// Try to change to the directory asked for:
	if(chdir(ltext(bmlFileHistp->b_dotp)) == -1) {

		// Whoops - failed:
		if(lgetc(bmlFileHistp->b_dotp, 0) == '/')
			mlwrite("%Echdir(%l): %e", bmlFileHistp->b_dotp, errno);
		else
			mlwrite("%Echdir(%t/%l): %e", cwDir, bmlFileHistp->b_dotp, errno);

		return(FALSE);
	}

	gwd();

	// Change the value of PWD in the environment:
#if	_BSD_SOURCE || _POSIX_C_SOURCE >= 200112L || _XOPEN_SOURCE >= 600
	setenv("PWD", ttext(cwDir), 1);
#else
	tcpys(nname, "PWD=");
	tcatt(nname, cwDir);
	newPWD = strmake_t(ttext(nname));
	putenv(newPWD);
#endif

	// Scan the buffers and rewrite file names to reflect the new relative name:
	for(bp = bheadp; bp != (BUFFER*) NULL; bp = bp->b_bufp) {

		if(bp->b_fname[0] == '\0')
			continue;

		if(bp->b_flag & BFINVS)
			continue;

		dirs[0] = ttext(playText5);
		dirs[1] = bp->b_fname;
		dirs[2] = (BYTE*)NULL;

		catpath(nname, dirs, bp, cwDir);
		setFnames(bp, ttext(nname));
	}

	upmode(NULL, WFMODE);

	// Show what we did, trim any trailing '/' from the name:
	if(lused(bmlFileHistp->b_dotp) > 1 && lgetc(bmlFileHistp->b_dotp, lused(bmlFileHistp->b_dotp) - 1) == '/')
		lsetused(bmlFileHistp->b_dotp, lused(bmlFileHistp->b_dotp) - 1);

	mlwrite("%Icwd %l", bmlFileHistp->b_dotp);

	return(TRUE);
#else
	mlwrite("%ENot available");
	return(FALSE);
#endif
}

#if	UNIX

#if 0	// Use the system provided ones
BYTE*	basename(BYTE* file)
{
	static	BYTE*	basebuf;
	BYTE*		p;

	if(basebuf)
		free((void*)basebuf);

	basebuf = strmake_t(file);

	if((p = strrchr(basebuf, '/')) == (BYTE*) NULL)
		return(basebuf);

	return(++p);
}

BYTE*	dirname(BYTE* file)
{
	static	BYTE*	dirbuf;
	BYTE*		p;

	if(dirbuf)
		free((void*)dirbuf);

	dirbuf = strmake_t(file);

	if((p = strrchr(dirbuf, '/')) == (BYTE*) NULL)
		return(NULL);

	*p = '\0';
	return(dirbuf);
}
#endif

void	gwd(void)
{
	/* Get the current working directory into a static area and put into
	 * TEXT cwDir.
	 *
	 * This routine accounts for the differences between BSD getwd() and
	 * System V getcwd().
	 *
	 * getwd() is deprecated by GNU.
	 *
	 * Return a LINE containing the empty string on error or if we dont
	 * have a routine for the current system (true for VMS, MS-DOS, etc etc
	 * until someone writes them).
	 */
#ifdef	PATH_MAX
#define	DSIZE	PATH_MAX
#else
#define	DSIZE	510
#endif
	char	dir[DSIZE + 2];
	char*	res;

#if	 BSD_UNIX
	res = getwd(dir);

#else	/*  BSD_UNIX */
	res = getcwd(dir, DSIZE);

#endif	/* !  BSD_UNIX */

	if( !res)	// Something went wrong ?
		res = "";
	tcpys(cwDir, res);
}

/* Make the simplified pathname, of a catenation of components
 *
 * name		The name that we create - write to this
 * comps[]	Array of components
 *		[0] Old current directory
 *		[1] File name - relative to [0] or absolute
 * bp		Buffer struct for the file
 * newdir	Path of the directory that we changed to
 *
 * The result is put into name.
 *
 * This always generates absolute paths, it might be nice to try to
 * keep a relative path if that can be done.
 *
 * M C Atkins, University of York (mcvax!minster.york.ac.uk!martin)
 */

#define isend(cp) (*(cp) == '/' || *(cp) == '\0')
#define isdot(cp) (*(cp) == '.' && isend(cp+1))
#define isdotdot(cp) (*(cp) == '.' && isdot(cp+1))

void	catpath(TEXT* name, BYTE* comps[], BUFFER* bp, TEXT* newdir)
{
	if(comps[1][0] == '/') {
		// File path is absolute
		// Just use that
		tcpys(name, comps[1]);
	} else {
		/* File path is relative.
		 * Build an absolute one taking care to remove /../ where we can
		 */
		tsetempty(name);
		if(comps[0] != 0 && comps[0][0] == '/')
			tcatc(name, '/');

		for( ; *comps != NULL; comps++) {
			BYTE*	fcp = *comps;

			while(*fcp) {
				if(*fcp == '/') {
					fcp++;
					continue;
				}
				if(isdot(fcp)) {
					fcp++;
					continue;
				}

				if(isdotdot(fcp)) {
					fcp += 2;	// Move over the ..
					// what does this do with '..fred' ?

					if(tused(name) > 1) {
						LineLn	loff = tused(name) - 1;

						// Back up over & remove the last file name component
						while(loff > 0 && tgetch(name, loff - 1) != '/')
							loff--;

						/* The old code had something about testing name for ../
						 * here - I think that I am missing something.
						 */
						tsetused(name, loff);
						continue;
					}
				}

				// Copy over to the next /
				while(*fcp != '\0' && *fcp != '/')
					tcatc(name, *fcp++);
				tcatc(name, '/');
			}
		}

		if(tused(name) > 1)
			tsetused(name, tused(name) - 1);
	}

	/* See if we can get a relative name.
	 * The new directory is an absolute path
	 * The file name is an absolute path
	 */

	// Keep the current name in case we screw up
	tcpyt(playText, name);

	// Find the lengths of each to the last directory separator
	LineLn	nameDl = (BYTE*)strrchr(ttext(name), '/') - ttext(name);
	LineLn	dirDl = (BYTE*)strrchr(ttext(newdir), '/') - ttext(newdir);

	/* Is the directory path of file longer than new directory path ?
	 * If so then the file might be below the directory
	 */
	if(nameDl >= tused(newdir)) {
		if(memcmp(ttext(name), ttext(newdir), (size_t)tused(newdir)) == 0 &&
		   tused(newdir) != 1 && tgetch(name, tused(newdir)) == '/') {
			// File name is part of directory name
			tcpys(playText4, tgetcaddr(name, tused(newdir) + 1));
			tcpyt(name, playText4);
		}
	} else {
		// Find the length to the last '/' in file name path:
		LineLn	dirlen = (BYTE*)memrchr(ttext(name), '/', (size_t)tused(name)) - ttext(name);

		/* Is the directory path of file shorter than new directory path ?
		 * Check that new directory could be in a directory below
		 * the one containing the file - in which case we can make
		 * relative with some ../
		 */
		if(nameDl <= dirDl &&
		   memcmp(ttext(name), ttext(newdir), (size_t)nameDl) == 0) {
			// Count how many directories below ?
			int	slashes = 0;
			BYTE*	sl = tgetcaddr(newdir, nameDl);

			for( ; (sl = strchr(sl, '/')) != NULL; sl++)
				slashes++;

			// Insert that number of ../ and append the file name:
			tcpys(playText4, tgetcaddr(name, dirlen + 1));
			tsetempty(name);
			while(--slashes >= 0)
				tcats(name, "../");

			tcatt(name, playText4);
		}
	}

	/* If that changed the file path - check that we can see the file AND
	 * that it is the same file - if not then restore the absolute path.
	 * This is to fix some of the silly things that sym links can do.
	 */
	if(textcmp(name, playText) != 0 &&
	   (CacheStat(ttext(name)) == -1 || statbuf.st_ino != bp->b_stino || statbuf.st_dev != bp->b_stdev))
		tcpyt(name, playText);
}
#endif	/* UNIX */

/* end */
