/*
 * ADDW SCCS: @(#)main.c	1.282 10/13/20 11:57:46
 */


/* make global definitions not external */
#define	maindef

#include        "estruct.h"	/* global structures and defines */

#include	<errno.h>
#include	<signal.h>

#include	"evers.h"	/* Version */
#include	"utf8.h"
#include	"edef.h"	/* global definitions */
#include	"efunc.h"	/* function declarations and name table	*/
#include	"ebind.h"	/* default key bindings */

#if	SHOWMODREL
char	mainvers[] = "@(#)main.c	1.282 10/13/20 11:57:46";
#endif

/* for MSDOS, increase the default stack space */

#if	MSDOS && LATTICE
unsigned _stack = 32766;
#endif

#if	ATARI && MWC
long _stksize = 32766L;		/* reset stack size (must be even) */
#endif

#if	MSDOS && AZTEC
int _STKSIZ = 32766/16;		/* stack size in paragraphs */
int _STKRED = 1024;		/* stack checking limit */
int _HEAPSIZ = 4096/16;		/* (in paragraphs) */
int _STKLOW = 0;		/* default is stack above heap (small only) */
#endif

#if	MSDOS && TURBO
extern unsigned _stklen = 32766;
#endif

#if     VMS
#include        <ssdef.h>
#define GOOD    (SS$_NORMAL)
#endif

#ifndef GOOD
#define GOOD    0
#endif

#if	CLRMSG
#undef	PRSVMSG
#define	PRSVMSG		1
#endif

/* Usage message */
static	char*	Explanation[] = {
	"Microemacs editor",
	"Usage: %s [-options] [@macros] [files...]",
	"-a	Process error file",
#if	UNIX
	"-c	Turn off signal catching",
#endif
	"-i var=value or -i var value",
	"	Set an environment variable ($var) or user variable (%%var)",
	"-e	Edit file (can change buffer)",
	"-g nn	Goto line nn in next file, see +nn",
#if	CRYPT
	"-kkey	Encryption key following files - no space before the key",
#endif
#if	TERMDEBUG
	"-K	Trace input - debugging",
#endif
	"--list-bindings",
	"	List Emacs commands & their bindings",
	"--list-commands",
	"	List Emacs commands & short description",
	"--list-keys",
	"	List terminal keys available",
	"-m mod	Set mode mod for following files",
	"-M mod	Unset mode mod",
	"-N	set Number window option - ie show line numbers",
	"-r	Restricted use; no shells, no new files",
	"-s pat	Search for pat",
	"-S	Slow tty output. -SS is slower, up to -SSSS",
#if	UNIX
	"-t tag	Search for tag",
#endif
	"-v	View mode (can't change buffer)",
	"--version",
	"	Display Emacs version",
	"-x --help\n	eXplain, help",
	"--	Rest of arguments are files. No more -options, +nn, @macros",
	"+nn	Goto line nn, +-nn count up from file end, + (no nn) means +-1, +0 line after +-1",
	"+nn.cc +nn,cc +nn:cc",
	"	Goto line nn, column cc",
	"(-e, -v: apply to all following files on command line, unless re-specified)",
	(char*)NULL
};

/* Private functions */
static	void	CatchSignals(void);
static	void	ParseOptions(char** argv);
static	void	__NOT_RETURN(Usage(int));

/* Private variables */
#if	CTAGS
static	char*	tagp;			/* the tag itself */
static	int	tagModes;		/* Modes at the time of the -t option */
#endif

#if	UNIX
static	int	sigcatch = TRUE;	/* catch signals if true */
#endif

static	int 	errflag = FALSE;	/* C error processing? */
static	int	startflag = FALSE;	/* startup executed flag */

/* Do we need to goto a line at start? */
static	BufLns	gline;			/* if so, what line? */
static	LineLn	gcol;			/* and column */

static	int	searchflag = FALSE;	/* Do we need to search at start? */

static	BYTE*	firstBufName;		// Name of first file on command line

static	int	CmdModes;		/* Command line specified modes */

static	int	seenMain;			// Use when a command line file name is "main"


/* This is set if there is an initialisation error
 * This is kludgy, interacts with ShowingImportantMessage
 * It is about not having an early error message not being seen.
 * The message must be written using mlwrite() so that it gets put
 * into the message buffer.
 * There must be a nicer way of doing this.
 */
static	char*	InitialError = NULL;

/*
 * argc		# of arguments
 * argv		argument strings
 */
#if	CALLED == 0	// CALLED == 0 as ctags will index main() rather than emacs()
int	main(int argc, char* argv[])
#else
int	emacs(int argc, char* argv[])
#endif
{
        KeyCode		c;		/* command character - could be Unicode self inserting character */
	BUFFER*		bp;		/* temp buffer pointer */
	int		flag = 42;	/* default flag */
	Cnt		count;		/* numeric repeat count */
	char*		InitialMessage;	/* First thing in the message line */

	ST_wallStart = time((time_t*)NULL);

	InitialMessage = "%IFor help briefly press Escape and then ?";

#if	MSDOS
	arglower(argv);		/* change args to lowercase */
#endif

	/* Get the program name, last component of path */
	if((progname = strrchr(argv[0], '/')) != NULL)
		progname++;
	else
		progname = argv[0];

	home = getenv("HOME");	/* Might be NULL */
	if(!home)
		home = "";

	if((TermName = getenv("TERM")) == NULL)
		TermName = "";
	TermName = strmake_t(TermName);	// Allocate so that it can be changed

	// Init suffixes for C & PHP mode
	suffixesC = strmake_t(".c .h .cc .cpp .cxx");
	suffixesPHP = strmake_t(".php .pl .pm");

#ifdef	LINUX5
	uid_t	suid;		// Saved UID

	getresuid(&uid, &euid, &suid);
	suidRoot = uid == 0 || euid == 0 || suid == 0;
#else
	uid = getuid();
	euid = geteuid();
	suidRoot = uid == 0 || euid == 0;
#endif

#if	CALLED
	eexitflag = FALSE;	/* not time to exit yet */
#endif

	/* initialize the editor */
	Utf8Init();
	mapinit();
	LinkBindings();		/* Initialise table linkage & hash table */
	hashInitBinds();	// Initialise Emacs command name hash
	edinit(mainBname);	/* Buffers, windows */
	varinit();		/* user variables */
	RestInit();		// Get shell name

	ParseOptions(argv);	/* Command line options */

	curwp->w_opts |= DefaultWindowOpts;	/* Option reading was too late for this */

	CatchSignals();

	/* if we are C error parsing... run it! */
	if(errflag && startup("error.cmd") == TRUE)
		startflag = TRUE;

	/* If invoked with no other startup files,
	 * run the system startup file here.
	 */
	if(startflag == FALSE) {
		(void) startup(NULL);

		startflag = TRUE;
#if	LEGACY_DISPLAY
		/* initalize colors to global defaults */
		curwp->w_fcolor = (char) gfcolor;
		curwp->w_bcolor = (char) gbcolor;
#endif
	}

	// --list-bindings ?
	if(ListBindings) {
		ListCommands(TRUE);
		exit(0);
	}

	/* We may need to do some the those immediately below before
	 * some of ParseOptions.
	 */

	/* Allocate screen buffers.
	 *
	 * Set # display rows & columns - we didn't know if when we called edinit().
	 *
	 * The windows will be given something reasonable by FixWindowSizes()
	 * called from vtinit().
	 */
	vtinit(0);		/* Display */

	sanitiseColours();

	/* Command line -t tag option ?
	 * Process it now not when options are parsed since if the tag has more than 1
	 * match the user will be asked to choose which one.
	 * When options are parsed the terminal subsystem has not been started -> crash
	 */
	if(tagp != (char*)NULL) {
		tcpys(vTEXT(playValStr), tagp);
		if(find_fun(vTEXT(playValStr), playText2, searchPat, NULL) == FALSE) {
			tagp = NULL;	// Not found
		} else {
			// Create a buffer
			BYTE*	bname = makename(ttext(playText2));
			BUFFER*	TagBp = bfind(bname, BFNCREATE|BFNACTIVE, 0);
			setFnames(TagBp, ttext(playText2));
			// Set this to inactive - so that it will be read in by swbuffer()
			TagBp->b_active = FALSE;
			TagBp->b_mode |= tagModes;	// Eg -v option might be in effect

			firstBufName = strmake_t(bname);// Tag buffer is the first one

			cleantag(searchPat);	// Convert from vi format
		}
	}

	/* If there are any files to read, read the first one!
	 * Beware: a command line @macro-file might have deleted main, or some
	 * other buffer - so get the first buffer by name as any saved BUFFER*
	 * might now be invalid, in which case the only buffer is probably the
	 * macro buffer
	 */
	bp = bfind(mainBname, BFNNONE, 0);
	if(firstBufName != NULL && (gflags & GFREAD)) {
		// TRUE should not be needed, but a macro might have deleted it
		BUFFER*	fbp = bfind(firstBufName, BFNCREATE|BFNACTIVE, 0);

		if( ! fbp)	// Coward's way out
			Die("Buffer %s cannot be found", firstBufName);

		swbuffer(fbp);			// Make current & read file

		chkmode(curbp);			// In particular how they interact

		/* Kill the buffer "main" as we are now using firstBufName
		 * But don't if that is the name of the first buffer.
		 * Also don't kill if "main" was a command line argument.
		 * This must be after the bfind() above else "main" might
		 * be magically recreated.
		 */
		if(bp != NULL && strcmp(mainBname, firstBufName) != 0 && !seenMain)
			zotbuf(bp, TRUE);

		bp = fbp;			// Buffer switched to

		ShowingImportantMessage = 1;	/* Don't overwrite [Read xx lines] message */
	} else {
		if(bp)		// If main exists
			chkmode(bp);
	}

#if	CTAGS
#define	OR_TAGFLAG	|| tagp != (char*)NULL
#else
#define	OR_TAGFLAG
#endif
	/* Deal with startup gotos and searches */
	if((bp->b_flag & BLINE_IN_MARKO) && (searchflag OR_TAGFLAG)) {
		update(FALSE);
		InitialError = "Cannot search and goto at the same time!";
		mlwrite("%s", InitialError);
	} else if(searchflag OR_TAGFLAG) {
		int	where = searchflag ? PTEND: PTBEG;

		// Caseless or not - check & set searchPat:
		if(searchflag)
			CheckCaselessSearch();

		if(scanner(searchPat, FORWARD, where) == FALSE) {
			update(FALSE);
#if	CTAGS
			if(tagp)
				tagerr(tagp);
			else
#endif
			{
				mlwrite("%ECan't find: %t", searchPat);
			}
		}
#if	CTAGS
		  else
			if(tagp)
				/* see comments in tag.c */
				moveForw(1);
#endif
	}

	/* setup to process commands */
        lastflag = 0;                           /* Fake last flags.     */

	// Paint the screen -- this will happen anyway
	//	updgar();
	//	sgarbf = TRUE;

	// This only if something interesting has not been written there:
	if(InitialError)
		mlwrite("%E""%h%s", InitialError);
	else	if( !ShowingImportantMessage)
			mlwrite(InitialMessage);

loop:	// **** Execute loop ****

	/* if a macro error is pending, or received SIGINT wait for a character */
	if(exec_error || IntSignalReceived) {
		int	oldcmd = discmd;

		discmd = TRUE;
		mlwrite("%E""%s. Press ? to see buffer %s, or press another key to continue",
			(IntSignalReceived ? "Interrupt" : "Error"),
			bmlNameMlHist);
		discmd = oldcmd;
		if(tgetc(MGC_NONE, 0) == '?')
			flag = 42;
		sgarbf = TRUE;
		update(FALSE);
		mlferase();
		exec_error = FALSE;
		IntSignalReceived = FALSE;
		kbdmode = STOP;

		// Test magic flag value that says done now't yet, ie error in startup file.
		// Auto open history buffer since interaction can be hard with a bad startup file.
		// Care: restricted mode, don't want to show something that the user should not see
		if(flag == 42 && !restflag &&
		   (bp = bfindCheck(bmlNameMlHist, BFNNONE)) && swbuffer(bp))
			updgar();
	}

#if	CALLED
	/* if we were called as a subroutine and want to leave, do so */
	if(eexitflag)
		return(eexitval);
#endif 

#if	0		/* What was this all about ? */
	/* execute the "command" macro...normally null */
	saveflag = lastflag;	/* preserve lastflag through this */
	(void) FuncExec(BindNullproc, FALSE, 1);
	lastflag = saveflag;
#endif

#if	PRSVMSG
	if(--ccounted == 0)
		mlerase();
#else
	if(mpresf != FALSE) {
		mlerase();
#if	CLRMSG
		if(c == ' ') {			/* ITS EMACS does this	*/
			update(FALSE);
			goto loop;
		}
#endif
	}
#endif

	// Note that the cursor is in whichever window is being edited
	CursorInMsg = 0;
	c = tgetc(MGC_MAP | MGC_NO_REFRESH, 0);
	CursorInMsg = 1;

	statbufInvalid = 1;	// Things might change while waiting for a key

	// Some window manipulation (changing what is curwp) doesn't always keep the current buffer pointer in track
	curbp = curwp->w_bufp;

	CurKey = getcmd(c, &flag, &count);

	/* Execute the command, record it's status.
	 * The moving to the next entry in command history was done in
	 * execute(), but the C compiler took the value of CurCmd before
	 * running execute() ... so the wrong lc_status was assigned!
	 */

	CurCmd->lc_status = FuncExec(getbind(CurKey, TRUE), flag, count);

#if	TRACEBUFFER
	{
		/* Hopefully useful in tracking the rare & hard to reproduce bugs.
		 * The line containing command names could get quite long.
		 */
		char	cbuf[30];

		sprintf(cbuf, " %#lx", (long)CurKey);
		lcats(traceCmdKeys, cbuf);

		lcats(traceCommands, "  ");
		if(flag) {
			sprintf(cbuf, "%ld ", (long)count);
			lcats(traceCommands, cbuf);
		}
		lcats(traceCommands, getbind(CurKey, TRUE)->n_name);
		sprintf(cbuf, "=%ld", (long)CurCmd->lc_status);
		lcats(traceCommands, cbuf);
	}
#endif

	/* The move to next entry is done after execute() to make it easier
	 * for Die() determine if any commands have been executed.
	 */
	if(++CurCmd >= CmdHistory + sizeof(CmdHistory) / sizeof(CmdHistory[0]))
		CurCmd = CmdHistory;

	cHECKeVERYcOMMAND();		// Bug hunting

	goto loop;
}

/* For vi compatability: '+nn' means goto line nn, also -g nn
 * +nn.cc means line nn, column cc.
 * +nn,cc & +nn:cc also for compatability with other editors.
 * Negative numbers count from bottom of the file
 * Nothing after the '+' move to last line.
 *
 * This does not do any error detection.
 *
 * Return: indication of line # or line # and col #
 *	Ie: BLINE_IN_MARKO or (BLINE_IN_MARKO | BCOL_IN_DOTO)
 */
short	ParseGline(const char* num)
{
	char*	s2;
	short	gotoflag = BLINE_IN_MARKO;

	gline = gcol = 0;

	if(!*num) {
		gline = -1;	// Empty is bottom
	} else {
		gline = atol(num);

		if((s2 = strchr(num, '.')) || (s2 = strchr(num, ',')) || (s2 = strchr(num, ':'))) {
			gcol = atol(s2 + 1);
			gotoflag = BLINE_IN_MARKO | BCOL_IN_DOTO;
		}
	}

	return(gotoflag);
}

/* **argv points to something like -o.
 * This needs a value, return a pointer to the value and increment *argvp
 * appropriately (NB argv will later be incremented).
 *
 * argvp	Pointer to argv - this might be incremented
 * oi		Index into *argv where the option character is.
 *		The point is to allow several after the leading '-'
 *
 * Bomb the program if there is an error.
 *
 * Since this will 'consume' the rest of the current argument,
 * what calls this should always 'goto argLoop'
 */
static	char*	GetOptarg(char*** argvp, int oi)
{
	char	opt;

	opt = (**argvp)[oi];	/* In case of error */

	if((**argvp)[oi + 1] != '\0')
		return(**argvp + oi + 1);

	(*argvp)++;	/* Next command line argument */

	if(**argvp == (char*)NULL)
		Die("Missing argument to '-%c'", opt);

	return(**argvp);
}

/* Parse the command line.
 * Bomb the program if there is something wrong.
 */
static	void	ParseOptions(char* argv[])
{
	char*	play;
	char*	play2;
	BYTE*	bname;		/* buffer name of file to read */
	BUFFER*	bp;		/* temp buffer pointer */
	int	i;
	char	opt;		/* The current option */
	VDESC	vDesc;		/* For setting variables */
	int	dupFiles = 0;	// Files in duplicate on the command line
	int	moreOpts = 1;	// Accept options starting: - or + or @
	short	gotoflag = 0;	// Line # & column # ?

	tcpys(playText2, "Duplicate files: ");

#if	CRYPT
	int	CryptFlag = FALSE;	/* encrypting on the way in? */
#endif

argLoop:
	for(argv++; *argv != (char*)NULL; argv++) {
		/* Process Switches */
		if(moreOpts && **argv == '-') {
			if( !strcmp(*argv, "--help")) {
				Usage(0);
				/*NOTREACHED*/
			}
			if( !strcmp(*argv, "--list-commands")) {
				ListCommands(FALSE);
				exit(0);
			}
			if( !strcmp(*argv, "--list-bindings")) {
				ListBindings = 1;
				break;
			}
			if( !strcmp(*argv, "--list-keys")) {
				ListKeys();
				exit(0);
			}
			if( !strcmp(*argv, "--version")) {
				printf("Version %s (%s) Compiled %s\n", Version, VERSION_DATE, CompileTime);
				exit(0);
			}
			if( !strcmp(*argv, "--")) {
				// End of options ?
				moreOpts = 0;
				continue;
			}

			int	oi;
			for(oi = 1; ; oi++) {
				switch(opt = (*argv)[oi]) {

				    case 'a':	/* process error file */
					errflag = TRUE;
					break;

#if	UNIX
				    case 'c':	/* turn off signal catching */
					sigcatch = FALSE;
					break;
#endif

				    case 'e':	/* -e for Edit file */
					CmdModes &= ~(MDVIEW);
					break;

				    case 'g':	/* -g for initial goto */
					gotoflag = ParseGline(GetOptarg(&argv, oi));
					goto argLoop;

				    case 'i':	/* Initialise a variable
						 * This can have several reasonable forms:
						 * -i%var=xx
						 * -i %var=xx
						 * -i%var xx
						 * -i %var xx
						 * $environmentVariables are also acceptable.
						 */
					play = GetOptarg(&argv, oi);

					if(*play != '$' && *play != '%')
						Die("Option -i must have environment variable ($var) or user variable (%%var), found '%s'",
							play);

					if((play2 = strchr(play, '=')))
						*play2++ = '\0';
					else if((play2 = *++argv) == (char*)NULL)
						Die("Missing value for variable %s", play);

					// OK Now have a name & value, set it:
					tcpys(playVar, play);
					splatSetText(play2);
					splatSetLen(strlen(play2));
					if(findvar(playVar, &vDesc, 0, VT_GLOBAL) == FALSE || !svar(&vDesc, &splatStrVAL))
						Die("Cannot add variable '%s' type %d", play, vDesc.v_type);

					goto argLoop;
#if	CRYPT
				    case 'k':	/* -k<key> for code key */
					tcpys(playCrypt, &((*argv)[2]));
					CryptFlag = tused(playCrypt) ? TRUE : FALSE;

					// Try to clear the key from the command line, think: ps /proc/*/cmdline
					memset(&((*argv)[2]), '\0', (size_t)tused(playCrypt));

					break;
#endif
#if	TERMDEBUG
				    case 'K':	/* Trace input keys & print termcap table
						 * 2 or more trace how output is generated.
						 */
					TermDebug++;
					if(TermDebFD == (FILE*)NULL) {
						// This is not explicitly closed - let exit() do that
						TermDebFD = fopen("DEBUG", "w+");
						setlinebuf(TermDebFD);
					}
					break;
#endif
				    case 'm':	/* Set Mode */
				    case 'M':	/* Unset Mode */
					/* Convert to upper case, only test that */
					play = mkupper(GetOptarg(&argv, oi));
					if((i = ModeMatch(&ModeBuf, play, (int)strlen(play),
							FORW, -1)) < 0)
						Die("Unrecognised mode: %s", play);

					// Set or unset the mode ?
					if(opt == 'm')
						CmdModes |= (short)(1 << i);
					else
						CmdModes &= (short)~(1 << i);
					goto argLoop;

				    case 'N':	/* Number window option to be set */
					DefaultWindowOpts |= WO_NUMBER;
					break;

				    case 'r':	/* -r restricted use */
					restflag = TRUE;
					break;

				    case 's':	/* -s for initial search string */
					searchflag = TRUE;
					// OOM check TODO
					play = GetOptarg(&argv, oi);
					tcpys(searchPat, play);
					addHistoryLine(bmlSrchHistp, play, AHL_NO_DUP);		// Put into search history
					bmlSrchHistp->b_dotp = lback(bmlSrchHistp->b_linep);	// Make default for next search
					goto argLoop;

				    case 'S':	/* -S - slow tty output. Want divisor: 2 .. 16 */
					if(OutputSlowness == 0)
						OutputSlowness = 1;
					else
						OutputSlowness *= 2;
					if(OutputSlowness > 32)
						OutputSlowness = 32;
					break;

#if	CTAGS
				    case 't':	/* -t for a tag, a la vi */
					play = GetOptarg(&argv, oi);

					if(tagp)
						Die("Only one tag at a time please '%s", play);

					tagp = play;
					tagModes = CmdModes;
					addHistoryLine(bmlTagHistp, tagp, AHL_NO_DUP);	// Put into tag history

					goto argLoop;	// Do rest of tagging on buffers opened
#endif	/* CTAGS */

				    case 'v':	/* -v for View File */
					CmdModes |= MDVIEW;
					break;

				    case 'x':	/* Explain */
					Usage(0);
					/*NOTREACHED*/

				    case '\0':
					if(oi != 1)	// End of -opts - look at next optarg
						goto argLoop;
					// DROP THROUGH
				    default:	/* unknown switch */
					fprintf(stderr, "%s: Unknown option '-%c'\n",
						progname, (*argv)[1]);
					Usage(2);
					/*NOTREACHED*/
				}
			}

			/* Go get next argument
			 * Can actually get here.
			 */
			// continue;
		}

		if(moreOpts && **argv == '@') {
			/* Run a macro file. Run it now, ie while parsing options.
			 * The terminal is still in cooked mode.
			 */

			/* Process Startup macros.
			 * Unless the name has a '.' or '/' search in macro path
			 */
			if(strchr(*argv, '/') || strchr(*argv, '.')) {
				if(dofile(*argv + 1) == TRUE)
					startflag = TRUE;
			} else {
				if(startup(*argv + 1) == TRUE)
					/* don't execute emacs.rc */
					startflag = TRUE;
			}

			continue;	/* Next cmd line option */
		}

		/* Process an input file */
#if	UNIX
		/* This is a horrid kludge to expand '~' characters
		 * passed to us from the command line by people not
		 * using C shells and the like.
		 */
		if(**argv == '~') {
			tsetused(playText, 0);
			if(insStrText(playText, 0, *argv, (LineLn)strlen(*argv) + 1, 0) >= 0) {
				if((play = exp_twiddle(playText, 0))) {
					// Prob only seen in message buffer
					mlerr(play, (char*)NULL);
				} else {
					if((play = strmake(ttext(playText))))
						*argv = play;
					// Don't report out of memory ... very unlikely to
					// happen here & if it does it will soon be obvious
				}
			}
		}
#endif

		/* For vi compatability: '+nn' means goto line nn
		 * Negative numbers count from bottom of the file
		 * Nothing after the '+' move to last line.
		 */
		if(moreOpts && **argv == '+') {
			gotoflag = ParseGline(*argv + 1);
			continue;
		}

		/* Before we create a buffer, check that the file is not the same file
		 * (same name or a link) as a previous one on the command line.
		 */
		char*	reason;
		if(alreadyHaveFile(*argv, &reason) != (BUFFER*)NULL) {
			tcats(playText2, *argv);
			tcats(playText2, reason);
			tcats(playText2, " ");
			dupFiles++;
			goto argLoop;
		}

		/* Set up a buffer for this file.
		 * Silly hardly-ever-happens bug, the code is prob not worth it:
		 * if the file is "main" (the buffer will exist even if no other does)
		 * then makename() will return "main<1>" - which looks nasty.
		 * So: instead of creating another buffer - just use it.
		 * But then: what if there is more than one "main" on the command
		 * line ... this should work, but there may still be a benign bug.
		 */
		if((bname = strrchr(*argv, '/')) != NULL)
			bname++;
		else
			bname = *argv;
		if(seenMain == 0 && strcmp(bname, mainBname) == 0)
			seenMain = 1;
		else
			bname = makename(*argv);

		if(firstBufName == NULL)	// Remember the first one
			firstBufName = strmake_t(bname);

		bp = bfind(bname, BFNCREATE, 0);
		bp->b_active = FALSE;	// Set this to inactive so that it is read later
		setFnames(bp, *argv);
		bp->b_mode = CmdModes;	// Remove gmode, added when buffer activated

		UpdateMetaData(bp, *argv);	// Store inode data

		// Desired line # ?
		// The sanity of the column is checked when the buffer is read in.
		if(gotoflag) {
			bp->b_toplinenum = gline;
			bp->b_doto = gcol;	// Might not be assigned, not a problem
			bp->b_flag = gotoflag;
			gotoflag = 0;
		}

#if	CRYPT
		if(CryptFlag) {
			/* -kxx option - copy key to buffer
			 */
			bp->b_mode |= MDCRYPT;
			cryptBuffer((char *)NULL, 0);

			cryptBuffer(ttext(playCrypt), tused(playCrypt));
			tcpyt(bp->b_key, playCrypt);
		}
#endif
	}

	/* We could clear out playCrypt ... but there is a nice (not munged) copy
	 * in playCrypt ... won't bother.
	 */

	/* Problems with files ?
	 * Complain, but this is so early that it will be overwritten with something else.
	 * Set InitialError which will see it replaced (probably).
	 * This relies on playText2 not being used for a bit.
	 */
	if(dupFiles) {
		mlwrite("%t", playText2);
		InitialError = ttext(playText2);
	}
}

/* Print a usage message & exit
 */
static	void	Usage(int exitc)
{
	char**	exp;

	for(exp = Explanation; *exp != (char*)NULL; exp++) {
		printf(*exp, progname);
		putchar('\n');
	}

	exit(exitc);
}

/* Catch nasty signals when running under UNIX unless the appropriate
 * option appeared on the command line.
 */
static	void	CatchSignals(void)
{
#if	UNIX
	if(!sigcatch)
		return;

	for(sigcatch = 1; /*void*/; sigcatch++) {
		if(sigcatch == SIGKILL
#ifdef	SIGCLD
			|| sigcatch == SIGCLD
#else
#ifdef	SIGCHLD		// For some BSDs
			|| sigcatch == SIGCHLD
#else
#error	UNIX but neither SIGCLD nor SIGCHLD defined
#endif
#endif	// SIGCLD
#ifdef	SIGSTOP				/* Can't catch me */
			|| sigcatch == SIGSTOP
#endif
#ifdef	SIGTTOU				/* Terminal output on BG job - default action is to stop
					 * which is what we want.
					 * Scenario is Emacs being accidentally backgrounded.
					 */
			|| sigcatch == SIGTTOU
#endif
#ifdef	SIGTSTP				/* USG as well as  BSD_UNIX these days */
			|| sigcatch == SIGTSTP || sigcatch == SIGCONT
#endif
#ifdef	SIGWINCH			/* Trapped in tty setup */
			|| sigcatch == SIGWINCH
#endif
#ifdef	SIGPROF			/* Profiling - compiler option -pg */
			|| sigcatch == SIGPROF
#endif
				)
				continue;

		if(signal(sigcatch, sigeat) == (SigType(*)(int))-1 &&
						errno == EINVAL)
			break;
	}

	/* It makes handling child processes much easier if these are ignored,
	 * these being the ones that the user may generate by bashing the keyboard.
	 * When running a child the terminal is sometimes put into cooked
	 * mode which means that signals could be generated by keystroke.
	 *
	 * IntSignalReceived is now set when SIGINT is received.
	 */
	// (void) signal(SIGINT,  SIG_IGN);
	(void) signal(SIGQUIT, SIG_IGN);

#endif
}

/* Odds & ends of initialisation.
 */
void	RestInit(void)
{
#if	MSDOS && (TURBO || LATTICE || AZTEC || MSC)
	shell = getenv("COMSPEC");
#endif

#if	UNIX
	/* Bourne shell is default, if the user is enought of a pervert to
	 * want the C shell, he should ask for it in the environment.
	 */
        if((shell = getenv("SHELL")) == (BYTE*)NULL || *shell == '\0')
		shell = "/bin/sh";

	/* It should have a meaningful name as shown by ps */
	if((shell_name = strrchr(shell, '/')) == (BYTE*)NULL)
		shell_name = shell;
	else
		shell_name++;
#endif
}

/* Initialize all of the buffers and windows. The buffer name is passed down
 * as an argument, because the main routine may have been told to read in a
 * file by default, and we want the buffer name to be right.
 */
void	edinit(char bname[])
{
	HistList*	hp;
	LINE***		lpp;
	TEXT***		tpp;

	// Empty list of procedure names:
	ProcNames = calloc(1, sizeof(NBIND));

	// Allocate dynamic LINEs.
	// Set the NUL flag & UTF-8 if its a global mode:
	for(lpp = mlLines; *lpp; lpp++) {
		**lpp = lalloc(0);
		lsetNUL(**lpp);

		if(gmode & MDUTF8)
			lsetUtf8(**lpp);
	}
	for(tpp = mlTexts; *tpp; tpp++) {
		**tpp = talloc(0);
		tsetNUL(**tpp);

		if(gmode & MDUTF8)
			tsetUtf8(**tpp);
	}
	searchPat = searchPat_e;	// Point to something valid & reasonable
	tcpys(errorLine, errorm);	// Line with error message
	gwd();				// Initialise cwDir $pwd

	curbp  = bfind(bname, BFNCREATE|BFNACTIVE, 0);	// First buffer
	wmessp = allocWindow(1, 0);			// Message window
	wheadp = allocWindow(1, 0);			// First editable window
	curwp  = wheadp;
	playValStr = valGet(V_STR);
	playValUndef = valGetUndef();
	playValChr = valGet(V_CHAR);

	/* Create the various history buffers & the list buffer */
	for(hp = HistoryBuffers; hp->hl_bufp; hp++) {
		if( !(*hp->hl_bufp = bfind(hp->hl_name, BFNCREATE|BFNACTIVE, BFINVS)))
			Die(OutOfMemory);

		(*hp->hl_bufp)->b_mode |= MDVIEW;
		(*hp->hl_bufp)->b_tabsize = 1;
	}

	// Forbidding the user from changing [Help Stack], [Kill Buffer] & [Eval Stack] saves many problems:
	bmlHelpStack->b_flag |= BIMMUTABLE;
	bmlKill->b_flag |= BIMMUTABLE;
	bmlEvalStack->b_flag |= BIMMUTABLE;

#if	TRACEBUFFER	// Debugging - records of how we got to where we are
	bmlTrace->b_flag |= BIMMUTABLE;
	addline("Input keys: ", bmlTrace, TRUE);
	traceKeys = lback(bmlTrace->b_linep);
	addline("Commands:",   bmlTrace, TRUE);
	traceCommands = lback(bmlTrace->b_linep);
	addline("Command Keys:",   bmlTrace, TRUE);
	traceCmdKeys = lback(bmlTrace->b_linep);
#endif

	/* Check for OOM - unlikely.
	 * Also put an empty line into the [Message Line].
	 */
	if(strsSubj == NULL || curbp == NULL || wheadp == NULL || wmessp == NULL ||
	   BufAppendLine(bmlPromptHistp) == NULL)
		Die(OutOfMemory);
	bmlPromptHistp->b_dotp = lforw(bmlPromptHistp->b_linep);

	/* Need to give the initial/head window a size else it will get killed
	 * by ZapZeroWidthWindows() - which may be called via vtinit().
	 */
	wheadp->w_ncolumns = MIN_WINDOW_COLS;

	iwCreate();				/* Non displayed information window */

	// Message window:
	ConnectBuffer(bmlPromptHistp, wmessp, 0);
	wmessp->w_wndp = wheadp;
	wmessp->w_opts = 0;			/* Must NOT be given DefaultWindowOpts */
	wmessp->w_lasttop = wmessp->w_toplinenum = 1;
	wmessp->w_attrib = WATT_NOSPLIT;

	/* Initialize window.
	 * This cannot be done as others are done in dup_window_on_split()
	 * since that expects curwp to point to something valid.
	 */
	ConnectBuffer(curbp, wheadp, 0);

	// Don't bother with the next 2 since term has not yet been initialised
	// This is fixed main() a little later
	// wp->w_ntrows = (char) (AllocLines - 2);/* "-1" for mode line.	*/
	// wp->w_ncolumns = AllocColumns - 1;
}

/* Return a pointer to the Window structure containing buffer "bp" if it is
 * onscreen, return NULL otherwise.
 */
Window*	onscreen(const BUFFER* bp)
{
	register Window*  wp;

	for(wp = wmessp; wp != (Window*) NULL; wp = wp->w_wndp)
		if(wp->w_bufp == bp)
			return(wp);

	return(NULL);
}

/* This function applies function "func" to all windows in TRACK mode
 * except the current window.
 * This used to be a buffer mode, more sense as a window option.
 * It also means that we can track different parts of the same buffer.
 *
 * Nothing is returned since "status" is set to the value of the command
 * applied to the current buffer (window).
 *
 * Statistics: number of Emacs commands executed is incremented for uses
 * of "func" in tracked windows.
 */
void	keeptrack(int (*const func)(int f, Cnt n), int f, Cnt n)
{
	Window* 	wp;
	BUFFER*		obufp = curbp;
	Window*		owinp = curwp;
	int		needUpdate = 0;

	/* Traverse windows. If a window exists then it is visible
	 */
	for(wp = wheadp; wp != (Window*) NULL; wp = wp->w_wndp)
		if(wp != owinp && (wp->w_opts & WO_TRACK)) {
			curwp = wp;
			curbp = wp->w_bufp;
			(void) (*func)(f, n);
			ST_commands++;
			needUpdate = 1;
		}

	// Always reset, regardless of if we did nothing
	curwp = owinp;
	curbp = obufp;

	// If any other windows needed tracking, check screen update:
	if(needUpdate)
		update(TRUE);
}

/* Execute the function pointed to by struct member Binding.
 * Arguments to function: f & n.
 *
 * Returns the command status. This is also copied to global cmdstatus ($status)
 */
int	FuncExec(const NBIND* Binding, int f, Cnt n)
{
	int		(*Function)(int f, Cnt n);
	int		JustAICpy;
	register int	status;
	int		FuncFlags;

	JustAICpy = JustAutoIndented;
	JustAutoIndented = FALSE;

	thisflag = 0;

	/* Record the command that we are executing + arguments.
	 * We might update the arguments shortly, so they will be saved again.
	 */
	CurCmd->lc_cmd = (NBIND*)Binding;
	CurCmd->lc_flag = f;
	CurCmd->lc_arg = n;
	CurCmd->lc_status = -1;		// Set status to 'being executed'

	FuncFlags = Binding->n_flag;

	/* If this would change the buffer, stop it now if we
	 * are in read only mode.
	 */
	if((FuncFlags & N_MODIFY) != 0) {
		if((curbp->b_mode & MDVIEW) != 0)
			return(rdonly(Binding));

		// OK: to change the buffer, but zap any procedure info
		if(curbp->b_procinfo != (Prinfo*)NULL)
			ClearProcInfo(curbp);
	}

	// Do not change an executing buffer
	if((FuncFlags & N_NOT_EXEC) && curbp->b_exec != 0) {
		mlwrite("%ENot when executing buffer");
		return(FALSE);
	}

	if(restflag && (FuncFlags & N_RESTRICTED))	/* don't allow this command if restricted */
		return(resterr(Binding));

	/* If ^M indented the last line zap white space.
	 * This approach works sort of OK, but there are some
	 * strange things, like <backspace> puts you onto the
	 * prev line ('cos we effectively move to SOL first).
	 * We really need something more key sensitive.
	 */
	if(JustAICpy == TRUE && (FuncFlags & N_AI_BEFORE) != 0) {
		/* Zap before we do whatever */
		ZapTrailingSpaces(curwp->w_dotp);
		lchange(WFHARD);
	}

	/* If the argument is unset should we supply one of 1 ? */
	if((FuncFlags & N_DEFAULT_1) && !f) {
		CurCmd->lc_flag = f = TRUE;
		CurCmd->lc_arg = n = 1;
	}

	/* Some commands are not allowed with hex option */
	if(curwp->w_opts & WO_HEX && FuncFlags & N_NOT_HEX) {
		mlwrite("%E""%B""%s: Command not allowed with Hex option", Binding->n_name);
		return(FALSE);
	}

	Function = Binding->n_func;

	/* If a reverse function is defined & the argument is -ve,
	 * do the reverse function.
	 * Doing it here means that the function code doesn't need to
	 * worry about the argument being -ve and so calling the reverse
	 * function itself.
	 */
	if(Binding->n_rfunc && f && n < 0) {
		n = 0 - n;
		Function = Binding->n_rfunc;
		FuncFlags = Binding->n_flag;
	}

	/* Is a negative argument allowed ? */
	if(f && n < 0 && (FuncFlags & N_FAIL_NEG_ARG)) {
		mlwrite("%E""%B""%s: Command cannot have negative argument", Binding->n_name);
		return(FALSE);
	}

	/* Is a zero argument allowed ? */
	if(f && n == 0 && (FuncFlags & N_FAIL_ZERO_ARG)) {
		mlwrite("%E""%B""%s: Command cannot have argument of zero", Binding->n_name);
		return(FALSE);
	}

	// Just incase we changed these above:
	CurCmd->lc_flag = f;
	CurCmd->lc_arg = n;

	// Run what we have:
	if(Function == NULL) {
		BUFFER*	bp;

		/* Find the buffer */
		if((bp = findProcedure(Binding->n_name)) == NULL) {
			mlwrite("%ENo such procedure '%s'", Binding->n_name);
	                return(FALSE);
		}

		// Set count:
		CmdFlag = f;
		CmdCount = n;
		// ANY as that will also accept no argument in !return
		VAL* retval = valGet(V_ANY);

		// Execute the procedure:
		status = dobuf(bp, &retval);

		valRelease(retval);
	} else
		status = (*Function)(f, n);

	ST_commands++;

	/* Zap after we insert new line (or whatever) so that we
	 * keep indented.
	 */
	if(JustAICpy == TRUE && (FuncFlags & N_AI_AFTER) != 0)
			ZapTrailingSpaces(lback(curwp->w_dotp));

	/* Do the business in track option.
	 */
	if(status == TRUE && (curwp->w_opts & WO_TRACK) &&
			(FuncFlags & N_TRACK) != 0)
		keeptrack(*Function, f, n);

	/* If the current buffer is in cursor sync mode, then sync
	 * all to curpw.
	 */
	if(curbp->b_flag & BSYNC_CURSOR)
		synccursor();

	lastflag = thisflag;

	cmdstatus = status;	// Assign $status

	return(status);
}

/* Insert the character in global CurKey into the buffer.
 *
 * BEWARE: lastkey is the last key pressed on the keyboard,
 * CurKey may be different when in a keyboard macro.
 *
 * This checks for auto save - if needed.
 *
 * Default count 1. Must be >= 0.
 *
 * Command:  self-insert
 * Bound to: Not in bind tables
 *
 * Return:
 *	TRUE
 *	FALSE
 */
int	SelfInsert(int f, Cnt n)
{
	int	suc;		// Insert success

	if(n <= 0)		// Will not be -ve
		return(TRUE);

	/* Hex edit option: */
	if(curwp->w_opts & WO_HEX)
		suc = insHexChar((UniChar)CurKey, n);
	else
		suc = insChar((UniChar)CurKey, n);

	if(suc) {
		/* The insert/overwrite worked.
		 * Check auto-save mode.
		 */
		if(curbp->b_mode & MDASAVE && --gacount <= 0) {
			/* And save the file if needed */
			upscreen(FALSE, 0);
			filesave(FALSE, 0);
			gacount = gasave;
		}

		lastflag = thisflag;
	}

	return(suc);
}

/* Unbound/unknown key sequence:
 *
 * Count not used. No default.
 *
 * Command:  unbound-key
 * Bound to: Not in bind tables
 *
 * Return FALSE
 */
int	UnboundKey(int f, Cnt n)
{
	char*	prompref;	/* Prompt prefix */
	char	keySeq[50];	/* Printable version of CurKey */

	// Produce a message saying that we don't know how to deal with it
	if(CurKey & (META | CTLX | SPEC | INTL))
		prompref = "sequence";
	else
		prompref = "key";

	cmdstr(CurKey, keySeq);	// The key made visible

	// Complain
	mlwrite("%E""%B[%s %s not bound]", prompref, keySeq);

	lastflag = 0;				// Fake last flags

        return(FALSE);
}


/* Window has hex option, insert a character - or rather insert a: nybble, 4 bits, 1/2 a byte.
 *
 * c must be a hex digit
 *
 * This will not be called in view mode.
 *
 * Default count is 1.
 *
 * Return TRUE if all OK.
 */
int	insHexChar(UniChar c, Cnt n)
{
	char*	hd;
	int	status = TRUE;

	if(Islower(c))
		c = toupper(c);

	// Check that it is valid:
	if( !(hd = strchr(hexdigits, c))) {
		TTbeep();
		return FALSE;
	}

	c = (UniChar)(hd - hexdigits);	// Convert to number: 0 .. 15


	while(--n >= 0 && status == TRUE) {
		/* Get a pointer to the byte that we will be changing
		 * If it is a virtual newline, just point to a newline,
		 * this change may make it something else.
		 */
		BYTE*	cp = lgetcaddr(curwp->w_dotp, curwp->w_doto);
		BYTE	nl = '\n';
		int	Eol = lused(curwp->w_dotp) == curwp->w_doto;

		if(Eol)	// End of line, fake the \n
			cp = &nl;

		if(curbp->b_mode & MDOVER) {
			/* Overwrite mode:
			 * PO = 0: replace top 4 bits; PO = 1
			 * PO = 1: replace bottom 4 bits; next char, PO = 0
			 */
			BYTE	newch = *cp;

			// Update the top/bottom 4 bits
			if(curwp->w_partoff == 0)
				newch = (BYTE)((c << 4) | (newch & 0xF));
			else
				newch = (BYTE)((c) | (newch & 0xF0));

			// Will that result in a good UTF8 sequence ?
			if( !utf8_CheckMangle(cp, newch, 0)) {
				status = FALSE;
				mlwrite("%E""%BCannot replace nyble: %s", utf8_perror());
				break;
			}

			*cp = newch;

			lchange(WFEDIT);

			if(Eol) {
				// We were editing a fake \n
				curwp->w_partoff = (char)(1 - curwp->w_partoff);

				if(*cp != '\n') {
					// The newline is no more; delete and replace with the new character
					if( !ldelete(1, KL_KILL_APPEND) || !linsert(1, *cp))
						return(FALSE);

					if(curwp->w_partoff == 1)
						moveBack(2);
				}
				else	// If ovwritten the 2nd (lower) 4 bits move on a character
					if(curwp->w_partoff == 0)
						moveForw(1);	// Next line
			} else	// We were not at end of line
				// If the character been transformed into a newline
				// Break the line here
				if(*cp == '\n') {
					// The only special recognition is newline
					if( !ldelete(1, KL_KILL_APPEND) || !lnewline(FALSE))
						return(FALSE);

					curwp->w_partoff = (char)(1 - curwp->w_partoff);
					if(curwp->w_partoff == 1)
						moveBack(1);

			} else	if(++curwp->w_partoff > 1) {
					curwp->w_partoff = 0;
					curwp->w_doto++;
			}
		} else {
			// Insert mode:

			// If inserting first 4 bits, create/insert a new character in the buffer
			if(curwp->w_partoff == 0) {
				BYTE	newch = (BYTE)(c << 4);

				// Is that a good start to UTF8 ?
				// In practice: a char 0..'~'
				if( !utf8_CheckMangle(cp, newch, 1)) {
					status = FALSE;
					mlwrite("%E""%BCannot insert: %s", utf8_perror());
					break;
				}

				linsert(1, newch);
				curwp->w_doto--;
				curwp->w_partoff = 1;
			} else {
				// Inserting 2nd 4 bits, OR them into the byte
				BYTE	newch = (BYTE)((*cp & 0xF0) | c);

				// Will that result in a good UTF8 sequence ?
				if( !utf8_CheckMangle(cp, newch, 0)) {
					status = FALSE;
					mlwrite("%E""%BCannot insert: %s", utf8_perror());
					break;
				}

				*cp = newch;

				// If that created a newline, put it in the buffer:
				if(*cp == '\n') {
					if( !ldelete(1, KL_KILL_APPEND) || !lnewline(FALSE))
						return(FALSE);
				}
				else	// Not newline, just move ons
					curwp->w_doto++;

				curwp->w_partoff = 0;
			}
			lchange(WFEDIT);
		}
	}

	return status;
}

/* Insert the character c into the buffer - n times
 *
 * This will not be called in view mode.
 *
 * c could be a Unicode character, insert it.
 *
 * Default count is 1.
 *
 * Return: TODO
 */
int	insChar(UniChar c, Cnt n)
{
        int	status;

        /* If a space was typed, fill column is defined, the argument
         * is non negative, wrap mode is enabled, and we are now past
         * fill column, perform word wrap.
         */
       	if(c == ' ' && (curwp->w_bufp->b_mode & MDWRAP) &&
		fillcol > 0 && n >= 0 && getdcol(FALSE) >= fillcol)
			FuncExec(BindWrapword, FALSE, 1);

	/* if we are in overwrite mode, not at eol,
	 * and next char is not a tab or we are at a tab stop,
	 * delete a char forword.
	 * BUG: Assumes that all characters are width 1. Fix TODO
	 */
	if(curwp->w_bufp->b_mode & MDOVER &&
	    curwp->w_doto < lused(curwp->w_dotp) &&
		(lgetc(curwp->w_dotp, curwp->w_doto) != '\t' ||
		 (curwp->w_doto) % Tabsize == (Tabsize - 1)))
			if( !ldelete(1L, KL_KILL_APPEND))
				return(FALSE);

	/* do the appropriate insertion */
	if(c == '}' && (curbp->b_mode & (MDCMOD | MDPHP)))
	       	status = insbrace(n, c);
	else if(c == '#' && (curbp->b_mode & MDCMOD))
		status = insHash();
	else
		status = linsert((BufChrs)n, c);

#if	CFENCE
	/* check for CMODE fence matching.
	 * Don't in macro replay, it just wastes time.
	 */
	if((c == '}' || c == ')' || c == ']') &&
		(curbp->b_mode & (MDCMOD | MDPHP)) && kbdmode != PLAY) {
#if	RDFENCE
		ShowMatchingBrace();
#else
		/* Search for matching fence stops at top-of-screen,
		 * so screen is never redrawn.
		 * Complain if we can't find it.
		 * first get the display update out there.
		 */
		update(FALSE);
		if(fmatch(c, 0))
			curwp->w_flag |= WFMOVE;
		else
			TTbeep();

#endif	/* ! RDFENCE */
	}
#endif	/* CFENCE */

	return(status);
}

/* Fancy quit command, as implemented by Norm. If the any buffer has
 * changed do a write on that buffer and exit Emacs, otherwise simply exit.
 *
 * This might have been called from Die() on memory allocation failure ...
 * so try to avoid a double/recursive fail on setting b_fname.
 * It doesn't matter if the same buffer is used for several buffers, also
 * it won't be freed on the way to exit.
 *
 * Default count unset.
 *
 * If the count is set:
 *	ignore buffer write errors - ie make sure that Emacs dies.
 *	used as exit code
 *
 * Command:  quick-exit
 * Bound to: M-z
 *
 * Return:
 *	TRUE  - said no to quitting as there are modified buffers
 *	FALSE - said no
 *	ABORT - bumped out of the question with a ^G.
 */
int	quickexit(int f, Cnt n)
{
	register BUFFER* bp;	/* scanning pointer to buffers */
	register int	status;
	register int	frigged = 0;

	for(bp = bheadp; bp != (BUFFER*) NULL; bp = bp->b_bufp) {

		if((bp->b_flag & BFCHG) == 0 || bp->b_flag & BFINVS)
			continue;

		/* If the buffer is filename-less, try to save it somewhere.
		 * Under unix we can easily save in the home directory by
		 * putting a twiddle at the start.
		 */
		if(bp->b_fname[0] == '\0') {

			frigged++;
#if	UNIX
			if(home)
				printToText(playText, "%s/%s", home, bp->b_bname);
			else
				tcpys(playText, bp->b_bname);
#else
			tcpys(playText, bp->b_bname);
#endif
			setFnames(bp, ttext(playText));
		}
		curbp = bp;		/* make that buffer cur	*/
		if(frigged) {
			mlwrite("%ISaving buffer %s as %s", bp->b_bname,
								bp->b_fnameClean);
			frigged = 0;
		} else
			mlwrite("%ISaving %s", bp->b_fnameClean);
		TTputc('\n');

		// Save the file, ignore errors if the count is set
		if((status = filesave(f, n)) != TRUE && f == FALSE)
			return(status);

		// If the count is set ignore any file write problems - mark as clean
		if(f)
			curbp->b_flag &= ~(BFCHG | BRENEWED);
	}

	quit(f, n);				/* conditionally quit	*/
	return(TRUE);
}

/* Quit command.
 *
 * If an argument, always quit. Otherwise confirm if a buffer has been changed
 * and not written out.
 *
 * Default count unset.
 *
 * Command:  exit-emacs
 * Bound to: C-x C-c   FN10
 *
 * Return: If this returns (ie Emacs does not exit) the value is
 *	the result of the prompt:
 *		FALSE - said no
 *		ABORT - bumped out of the question with a ^G.
 */
int	quit(int f, Cnt n)
{
        register int    s;

	if(wbuffs && savemany(TRUE, 1) == FALSE)
#if	MSDOS
		;
#else
		DoSleep(1);			/* time to see prompt	*/
#endif

	if(f != FALSE				/* Argument forces it.  */
	|| (s = anycb()) == 0			/* All buffers clean.	*/
						/* User says it's OK.	*/
	|| (s=mlyesno("%d modified buffers exist. Leave anyway [y/n]? ", s)) == TRUE) {

#if	FILOCK
		if(lockrel() != TRUE) {
			TTputc('\n');
			TTputc('\r');
			TTkclose();
			TTclose();
			exit(1);
		}
#endif

		vttidy();

#if	CLEAN
		/* Remove buffers.
		 * Bring this in when testing under valgrind, look at memory leaks.
		 * Don't leave in full time since it causes some messages to flash
		 * on the message line - does not look nice.
		 *
		 * There are a few things left still allocated at exit, these don't
		 * worry me: tgetent() is responsible for that.
		 *
		 * To run a check: switch this block on and compile with debugging and run:
		 *
		 * valgrind -v --log-file=VALGRIND --leak-check=full --show-reachable=yes ./me *.c
		 */
#undef	CLEAN
#define	CLEAN	1
		cexit(f ? (int)n : GOOD);
#endif

		exit(f ? (int)n : GOOD);
	}
	mlwrite("");
	return(s);
}

/* Begin a keyboard macro.
 *
 * Error if not at the top level in keyboard processing. Set up variables and
 * return.
 *
 * The key are stored in TEXT kbdMacro - this means that we don't
 * have to worry about a sequence becoming too long.
 *
 * The count is ignored. The default count is unset
 *
 * Command:  begin-macro
 * Bound to: C-x (
 *
 * Return:
 *	TRUE  recording of macro started
 *	FALSE macro alread being recorded
 */
/* ARGSUSED */
int	ctlxlp(int f, Cnt n)
{
	if(kbdmode != STOP) {
		mlwrite("%EMacro already active or being recorded");
		return(FALSE);
	}
	mlwrite("%IStart macro");
	tsetused(kbdMacro, 0);
	kbdmode = RECORD;

	/* update the mode lines of all the windows */
	upmode(NULL, WFMODE);

	return(TRUE);
}

/*
 * End keyboard macro.
 *
 * Check for the same limit conditions as the above routine. Set up the
 * variables and return to the caller.
 *
 * Note that this will be called when a keyboard macro is being played and
 * the end is reached. This is because the keys that end-macro is bound
 * to are recorded.
 *
 * The count is ignored. The default count is unset
 *
 * Command:  end-macro
 * Bound to: C-x )
 *
 * Return:
 *	TRUE  macro recording stopped
 *	FALSE macro was not being recorded
 */
/* ARGSUSED */
int	ctlxrp(int f, Cnt n)
{
        if(kbdmode == STOP) {
                mlwrite("%WMacro not active");
                return(FALSE);
        }
	if(kbdmode == RECORD) {
	        mlwrite("%IEnd macro");
	        kbdmode = STOP;
	}

	/* update the mode lines of all the windows */
	upmode(NULL, WFMODE);

        return(TRUE);
}

/* Execute a macro.
 * 
 * The command argument is the number of times to loop. Quit as soon as a
 * command gets an error.
 *
 * Default count 1. Must be >= 0
 *
 * Command:  execute-macro
 * Bound to: C-x e
 *
 * Return
 *	TRUE  Macro executed
 *	FALSE Macro being recorded
 */
int	ctlxe(int f, Cnt n)
{
	if(kbdmode != STOP) {
		mlwrite("%EMacro already active or being recorded");
		return(FALSE);
	}

	kbdrep = n;		/* remember how many times to execute */
	kbdmode = PLAY;		/* start us in play mode */

	/* Start at the beginning.
	 * Setting kbdptr is quite safe since the t_text in kbdMacro
	 * will not move - we will not be extending it while we
	 * play it.
	 * Cast through void* else clang will complain about increasing the
	 * alignment requirement. This is a valid comment, but what gets stored
	 * are KeyCodes - so that is OK as malloc() returns something maximally
	 * aligned.
	 */
	kbdptr = (KeyCode*)(void*)ttext(kbdMacro);

	return(TRUE);
}

/* Abort.
 * 
 * Beep the beeper. Kill off any keyboard macro, etc., that is in progress.
 * Sometimes called as a routine, to do general aborting of stuff.
 *
 * Output ABORT into terminal debugging. This is because this is an easy way
 * of getting something into the log file that can be searched for.
 *
 * The count is ignored. The default count is unset.
 *
 * Command:  abort-command
 * Bound to: C-g   C-x C-g   M-C-g
 *
 * Return: ABORT
 */
/* ARGSUSED */
int	ctrlg(int f, Cnt n)
{
	static	int	CalledJustNow;

	TERMOUTTRACE((TermDebFD, "\nABORT\n"))

	TTbeep();
	if(kbdmode == RECORD) {
		kbdmode = STOP;
		/* update the mode lines of all the windows */
		upmode(NULL, WFMODE);
	}

	mlwrite(MsgAborted);

#if	UNIX
	/* Try to deduce that we have lost the terminal - HUP without the signal.
	 * If we don't catch it we loop forever.
	 * Flush to see if this provokes a write error.
	 * Don't do this gratuitiously as we don't want to kill the edit for some
	 * transitory condition, so sleep a bit.
	 * This is done because error on reading the terminal results in abortc being returned.
	 */

	update(FALSE);
	if(TtyReadFailure || TtyWriteFailure) {
		char	Errm[100];
		if(TtyReadFailure)
			sprintf(Errm, "Term read fail: %s", strerror(TtyReadFailure));
		else
			sprintf(Errm, "Term writefail: %s", strerror(TtyWriteFailure));
		mlwrite("%E""%s", Errm);
		DoSleep(5);
		if(++CalledJustNow > 5)
			Die(Errm);
	}
	else	/* Reset consequtive error count */
		CalledJustNow = 0;
#endif

	return(ABORT);
}

/* tell the user that this command is illegal while we are in
 *  VIEW (read-only) mode
 */
int	rdonly(const NBIND* binding)
{
	mlwrite("%E""%B""%s: Command illegal in VIEW mode", binding->n_name);

	return(FALSE);
}

/* Command is restricted */
int	resterr(const NBIND* binding)
{
	mlwrite("%E""%B""%s: That command is RESTRICTED", binding->n_name);

	return(FALSE);
}

/* User function that does NOTHING
 *
 * Command:  nop
 * Bound to: FN11   FN12
 *
 * Evaluate any command line that it might have.
 * Not a problem if there is not one.
 *
 * The default count is not set.
 *
 * Return: TRUE
 *	   If the count is set: return that value.
 */
int	nullproc(int f, Cnt n)
{
	// Evaluate anything that might be on the line
	VAL*	tok = valGet(V_ANY);
	nextarg("Evaluate: ", &tok, MLEMPTYOK, (BUFFER*)NULL);
	valRelease(tok);

	if( !f)
		return(TRUE);

	return((int)n);
}

/* dummy function for binding to meta prefix
 */
int	metadummy(void)
{
	return(FALSE);
}

/* dummy function for binding to control-x prefix
 */
int	cex(void)
{
	return(FALSE);
}

/* dummy function for binding to universal-argument
 */
int	unarg(void)
{
	return(FALSE);
}

#if	RAMSIZE
/*	These routines will allow me to track memory usage by placing
 *	a layer on top of the standard system malloc() and free() calls.
 *	with this code defined, the environment variable, $RAM, will
 *	report on the number of bytes allocated via malloc.
 *
 *	with SHOWRAM defined, the number is also posted on the
 *	end of the bottom mode line and is updated whenever it is changed.
 */

#undef	malloc
#undef	free

char*	allocate(nbytes)	/* allocate nbytes and track */
	unsigned nbytes;	/* # of bytes to allocate */
{
	char *mp;	/* ptr returned from malloc */
	char *malloc();

	mp = malloc(nbytes);
	if(mp) {
		envram += nbytes;
#if	RAMSHOW
		dspram();
#endif
	}

	return(mp);
}

/* release malloced memory and track
 *
 * mp	chunk of RAM to release
 */
int	release(char* mp)
{
	unsigned *lp;	/* ptr to the long containing the block size */

	if(mp) {
		/* update amount of ram currently malloced */
		lp = ((unsigned *)mp) - 1;
		envram -= (long)*lp - 2;
		free(mp);
#if	RAMSHOW
		dspram();
#endif
	}
}

#if	RAMSHOW
int	dspram(void)	/* display the amount of RAM currently malloced */
{
	char mbuf[20];
	char *sp;

	TTmove(AllocLines - 2, 70);

	TTforg(COLOUR_WHITE);
	TTbacg(COLOUR_BLACK);

	sprintf(mbuf, "[%lu]", envram);
	sp = &mbuf[0];
	while (*sp)
		TTputc(*sp++);
	TTmove(AllocLines - 1, 0);
	movecursor(AllocLines - 1, 0);
}
#endif
#endif

/*	On some primitave operation systems, and when Emacs is used as
 *	a subprogram to a larger project, Emacs needs to de-alloc its
 *	own used memory.
 *
 * This can also be used as part of QA - run under valgrind to look for
 * memory leaks; so free up stuff to minimise what valgrind complains about.
 *
 * Termcap strings are not freed.
 *
 * status	return status of Emacs 
 */
#if	CLEAN
#if	CALLED
int	cexit(int status)
#else
void	cexit(int status)
#endif
{
	register BUFFER* bp;	/* buffer list pointer */
	register Window* wp;	/* window list pointer */
	char**		 cpp;	/* Play char** */
	Window		 fakeWindow;
	LINE***		lpp;
	TEXT***		tpp;
	VAL*		vp;
	NBIND*		bindp;

	// fprintf(stderr, "cexit()\n");

	/* If we don't do this then the last lines written to stdout
	 * will get lost.
	 */
	fflush(stdout);

#if	MAGIC	// Regular expressions
	rmcclear();
	mcclear();
#endif

	// Free table of procedures:
	for(bindp = ProcNames; bindp->n_name != NULL; bindp++) {
		free(bindp->n_name);
		if(bindp->n_desc != NULL)
			free(bindp->n_desc);
	}
	free(ProcNames);	// List of procedure names

	for(cpp = StartupFiles; *cpp; cpp++)
		free(*cpp);

	/* dealloc the user variables */
	varclean(uv_head);

	// Free values (stack)
	// Remove any NOFREE flag, then try to free value
	playValStr->v_flags &= (unsigned)~V_NOFREE;
	playValChr->v_flags &= (unsigned)~V_NOFREE;
	valRelease(playValStr);	// This will never be in the stack - until now
	valRelease(playValChr);
	for(vp = valhead; vp != (VAL*)NULL; ) {
		VAL*	valnext = vp->v_v.v_next;

		valFree(vp);
		free(vp);
		vp = valnext;
	}

	/* First clean up the windows.
	 * This code is not nice.
	 * Set wmessp & wheadp to NULL first otherwise utf8_ensurePosStart()
	 * tries to walk down a chain that has been freed.
	 */
	wp = wmessp;
	wmessp = wheadp = NULL;
	while(wp) {
		Window* tp = wp->w_wndp;
		freeWp(wp);
		wp = tp;
	}

	if(firstBufName)
		free(firstBufName);

	// Work in progress, expect to be broken on occasion:
	// freeWp(winfop);
	FreeLine(winfop->w_modeline);
	free(winfop->w_lineFlags);
	free(winfop);

	/* zotbuf() may recreate a buffer if the current window points to the bp being zapped.
	 * wheadp must be valid, the fake window will do.
	 * But this needs a width else ZapZeroWidthWindows() has a go at it.
	 */
	curwp = &fakeWindow;
	wheadp = &fakeWindow;
	memset(&fakeWindow, '\0', sizeof(fakeWindow));
	fakeWindow.w_ncolumns = MIN_WINDOW_COLS;

	/* Free the buffers */
	while((bp = bheadp)) {
		bp->b_nwnd = 0;
		bp->b_flag = 0;	/* don't say anything about a changed buffer! */
		bp->b_exec = 0;	// Else cannot go exit if this is an executing macro file
		zotbuf(bp, TRUE);
	}

	// Free dynamically allocated non buffer LINEs:
	for(lpp = mlLines; *lpp; lpp++) {
		// These LINEs were never intended to be freed, their
		// forward/backwards pointers are invalid.
		// Set them so that we can free the LINE.
		lforw(**lpp) = **lpp;
		lback(**lpp) = **lpp;

		FreeLine(**lpp);
	}
	// and TEXTs:
	for(tpp = mlTexts; *tpp; tpp++)
		FreeText(**tpp);

	searchPat = NULL;	// What it points to has been freed

	/* and the video buffers */
	vtfree();

	// Misc
	free(language);
	free(territory);
	free(TermName);
	free(utf8_modeReplace);
	free(suffixesC);
	free(suffixesPHP);

	// Hash tables
	hashFree(&hashKeyBindings);
	hashFree(&hashCmdNames);
	hashFree(&hashEnv);
	hashFree(&hashFunc);

#if	TERMDEBUG
	if(TermDebFD != (FILE*)NULL)
		fclose(TermDebFD);
#endif

#if	!(_BSD_SOURCE || _POSIX_C_SOURCE >= 200112L || _XOPEN_SOURCE >= 600)
	if(newPWD != NULL)
		free(newPWD);
#endif

	/* and now.. we leave [pick the return if we are a subprogram] */
#if	CALLED
	eexitflag = TRUE;	/* flag a program exit */
	eexitval = status;
	return(status);
#else
//#undef	exit
	_exit(status);
#endif
}
#endif

/* This gets called when we are hit by a signal - any signal.
 *
 * SIGPIPE - set the flag PipeWriteFail & return.
 *
 * Other signals: bring the editor to a screeching halt; save modified buffer
 * contents as safely as possible - avoid using stdio as malloc() may fail.
 */
#if	UNIX
#if	BSD_UNIX
void	sigeat(void)
#else
void	sigeat(int sig)
#endif
{
	static	int	buster;		// multiple call loop buster

	/* Got hit by writing to a dead pipe ?
	 * Rearm the signal handler, set a flag and return.
	 */
	if(sig == SIGPIPE) {
		PipeWriteFail = 1;
		signal(sig, sigeat);
		return;
	}

	/* This will not have been generated by the user pressing C-c at the keyboard,
	 * well - probably not.
	 * More likely it was sent from another terminal.
	 * Used to bust procedures, etc, that go on for too long.
	 */
	if(sig == SIGINT) {
		IntSignalReceived = TRUE;
		signal(sig, sigeat);
		return;
	}

	// Caught a signal when handling a signal
	// Crawl out as simply as possible
	if(buster) {
		char	b[40] = "\r\nCaught signal ";
		char	c[40] = "\r\nWhen dealing with signal ";
		mlputli(sig, 10, (CHAR*)strchr(b, '\0'));
		mlputli(buster, 10, (CHAR*)strchr(c, '\0'));

		ErrorWriteList(b, c, "\r\n", (char*)NULL);

		_exit(2);
	}
	buster = sig;

	// Mainly if debugging is being output
	fflush(stderr);

	/* Trap nasty signals under UNIX. Write files and quit.
	 * This is one way, it we come in here, we dont get out.
	 *
	 * The first thing to do is to ignore signals, this way if file
	 * writing causes SIGBUS we wont get called lots of times.
	 */
	signal(SIGHUP, SIG_IGN);
	signal(SIGBUS, SIG_IGN);
	signal(SIGILL, SIG_IGN);
	signal(SIGINT, SIG_IGN);
	signal(SIGQUIT, SIG_IGN);
	signal(SIGSEGV, SIG_IGN);
	signal(SIGTRAP, SIG_IGN);
#ifdef	SIGIOT
	signal(SIGIOT, SIG_IGN);
#endif
#ifdef	SIGEMT
	signal(SIGEMT, SIG_IGN);
#endif
#ifdef	SIGSYS
	signal(SIGSYS, SIG_IGN);
#endif
	
	// Kludge: we want a core dump out of this but not for all signals,
	// eg no need to core dump if got logged out.
	// Always dump on BSD as we don't know what signal we got hit with.
	// Do not dump if running as root.

	Die_a();

#if	BSD_UNIX
	DieShouldAbort = !suidRoot;

	ErrorWrite("\n*** caught signal: emergency quit ***\n");
#else
	// Specify the signals that we want a dump from:
	static	char	dumpSigs[] = {
		[SIGBUS]	= 1,
		[SIGSEGV]	= 1,
		[SIGILL]	= 1,
		[SIGABRT]	= 1,
		[SIGFPE]	= 1,
#ifdef	SIGSYS
		[SIGSYS]	= 1,
#endif
	};
	char*	sigName = strsignal(sig);
	if( !sigName)
		sigName = "unknown";

	DieShouldAbort = !suidRoot && sig < (int)sizeof(dumpSigs) && dumpSigs[sig];

	// Do the following, but avoid vfprintf()
	// Die("*** caught signal %d (%s) emergency quit ***", sig, sigName);

	char	numbuf[20];
	mlputli((long)sig, 10, numbuf);
	ErrorWriteList("\n*** caught signal ", numbuf, " (", sigName, ") emergency quit ***\n", (char*)NULL);

#endif	/*  BSD_UNIX */

	Die_b();
}
#endif	/* UNIX */

#if	MSDOS
/* Set an array of strings to lower case
 * Intended use is lowering argv[].
 */
void	arglower(char** argvec)
{
	char*	p;
	
	for(p = *++argvec; p != (char*)NULL; p = *++argvec)
		(void)mklower(p);
}
#endif

/* Check $LANG, set try_utf8 if it contains .utf8 or .utf-8
 * Case insensitve check.
 *
 * Also try to extract & set language & territory.
 */
void	Utf8Init(void)
{
	char*	lang = getenv("LANG");
	char*	dot;
	char*	us;
	char	DefLang[] = "en_GB.utf8";

	// Make it a string that can be assigned to via $utf8-replacement
	utf8_modeReplace = strmake_t(utf8_replace);

	if( !lang)
		lang = DefLang;

	if((dot = strchr(lang, '.')) != NULL) {
		try_utf8 = !cstrcmp(dot, ".utf8") || !cstrcmp(dot, ".utf-8");

		if(try_utf8)
			gmode |= MDUTF8;

		*dot = '\0';
	}

	/* Try and get values for language & territory
	 * Note that _territory might be missing.
	 */
	if((us = strchr(lang, '_')) != NULL)
		*us = '\0';

	language = strmake_t(lang);

	territory = strmake_t(us ? (us + 1) : "");

	// Repair damage to $LANG
	if(dot != NULL)
		*dot = '.';
	if(us != NULL)
		*us = '_';
}

/* Start the die process.
 *
 * Quick exit if we get into a Die() loop.
 * Close the terminal - back to cooked mode.
 */
void	Die_a(void)
{
	static	int	been_here;

	if(been_here++) {

		// Try to close the terminal - don't clear the screen
		if(been_here == 1)
			TTclose();

		ErrorWrite("\nError while processing an error\n");

		exit(2);
	}

	/* Warn the user.
	 */
	TTbeep();

	/* Come out of visual mode so that the user sees the messages below.
	 * Return to canonical mode, etc, so that the terminal works as expected.
	 */
	TTkclose();
	TTclose();
}

/* Quick & nasty print a message and kill the program.
 *
 * A \n is not needed to terminate the message.
 *
 * This will also save modified buffers and print the last
 * few lines of the message buffer.
 *
 * This has been chopped into 3, this is the part that prints the msg. This is
 * as Fedora 32 GCC 10 new option -fanalyzer pointed out that vfprintf() must
 * not be used in a signal handler and that Die() was call from sigeat().
 */
void	Die(BYTE* msg, ...)
{
	va_list ap;

	/* Try and get early error messages out.
	 * Hopefully this will not cause further errors.
	 */
	if(TTopened != TTfullyOpened)
		mlFlush();

	Die_a();

	// Output the message that is the argument
	va_start(ap, msg);
	fprintf(stderr, "%s: ", progname);
	vfprintf(stderr, msg, ap);
	va_end(ap);

	Die_b();
}

/* Print recent activity and die.
 *
 * Prints:
 * * last few lines in the message buffer
 * * last few commands executed
 *
 * Then saves buffers to .emacssave
 * Dies
 */
void	Die_b(void)
{
	BUFFER* bp;	/* scanning pointer to buffers */
	int	i;

	/* Write the last few lines of the message buffer,
	 * could contain something interesting.
	 */
	LINE*	lp = lback(bmlHistp->b_linep);
	char	mbEnd[] = "=========\n\n";

	// If there aren't any don't write
	if(lback(lp) != bmlHistp->b_linep) {
		int	lines = CRASHMSGLINES;

		// Work up from the bottom for a few lines:
		while(--lines > 0) {
			LINE*	prev = lback(lp);
			if(prev == bmlHistp->b_linep)
				break;
			lp = prev;
		}

		// Work back down to the last message show to the user:
		ErrorWrite("\n\nLast few editor messages\n=========\n");
		for( ; lp != bmlHistp->b_linep; lp = lforw(lp)) {
			IGNORE(write(2, ltext(lp), (size_t)lused(lp)))
			ErrorWrite("\n");
		}
		ErrorWrite(mbEnd);
	}

	// Now show the last few commands executed (if any):
	if(CmdHistory[0].lc_cmd != (NBIND*)NULL) {
		ErrorWrite("\nLast few editor commands\n=========\n");
		for(i = CMDSKEPT; --i >= 0; ) {
			char	msgbuf[50];
				// Next command, or at start wrap to CMDSKEPT'th previous command
			if(++CurCmd >= CmdHistory + sizeof(CmdHistory) / sizeof(CmdHistory[0]))
				CurCmd = CmdHistory;

			// Very early on we will have empty slots:
			if(CurCmd->lc_cmd == (NBIND*)NULL)
				continue;

			ErrorWrite(CurCmd->lc_cmd->n_name);
			if(CurCmd->lc_status == -1) {
				ErrorWrite(" did not complete");
			} else {
				mlputli((long)CurCmd->lc_status, 10, msgbuf);
				ErrorWriteList(" status=", msgbuf, (char*)NULL);
			}

			if(CurCmd->lc_flag) {
				mlputli((long)CurCmd->lc_arg, 10, msgbuf);
				ErrorWriteList(" count=", msgbuf, (char*)NULL);
			}
			ErrorWrite("\n");
		}
		ErrorWrite(mbEnd);
	}

	/* And do the business.
	 * Save visible modified buffers.
	 */
	for(bp = bheadp; bp != (BUFFER*) NULL; bp = bp->b_bufp) {

		if((bp->b_flag & BFCHG) == 0 || bp->b_flag & BFINVS)
			continue;

		bquicksave(bp);
	}

	// Kludge to get a core dump:
	if(DieShouldAbort) {
		// Say that we are about to abort
		// Shows that the quick saves completed
		ErrorWrite("Emacs aborting\n");

		abort();
	}

	ErrorWrite("\nEmacs dieing\n");

	_exit(2);

	__NOT_REACHED
}

/* Write the string msg to file descriptor 2 - stderr.
 * Use a low level write(), ie not stdio
 * because this is primarily designed to be use when things
 * have gone badly wrong and using stdio might make further errors.
 *
 * Do not append a newline after msg.
 */
void	ErrorWrite(const char* msg)
{
	IGNORE(write(2, msg, strlen(msg)))
}

/* Write a NULL terminated list of strings with ErrorWrite()
 */
void	ErrorWriteList(const char* msg, ...)
{
	va_list ap;
	va_start(ap, msg);

	while(msg != (char*)NULL) {
		ErrorWrite(msg);
		msg = va_arg(ap, char*);
	}

	va_end(ap);
}

/* end */
