/* Map characters on input from the tty.
 * ADDW SCCS: @(#)mapkey.c	1.26 09/14/20 22:57:02
 * 
 * This is not a general macro facility, but something to cope with keyboards
 * that have characters in the ``wrong'' place.
 */

#include	"estruct.h"
#include	"utf8.h"
#include	"edef.h"

/* Read a Unicode character from the keyboard or executing file/macro.
 * This can either be a literal character or something like U+123
 *
 * Return:
 *	the Unicode character
 *	UNISTR_END meaning:
 *	    mlprompt() failure or nothing entered
 *	    Bad Unicode character entered
 *	    C-g pressed
 */
UniChar	mapInterpKey(const char* prompt)
{
	UniChar	ch;

	if(mlprompt(prompt, MLSTRING | MLNOSPACE | MLNOMAP, bmlPromptHistp) != TRUE)
		return(UNISTR_END);

	/* Return the first character as long as the string
	 * does not start like: U+
	 */
	if((lgetc(bmlPromptHistp->b_dotp, 0) != 'u' && lgetc(bmlPromptHistp->b_dotp, 0) != 'U') || lgetc(bmlPromptHistp->b_dotp, 1) != '+')
		return(utf8_decode(ltext(bmlPromptHistp->b_dotp)));

	if(sscanf(lgetcaddr(bmlPromptHistp->b_dotp, 2), "%" UNICHAR_LMOD "x", &ch) != 1) {
		mlwrite("%WCould not interpret as Unicode character (hex) '%l'", bmlPromptHistp->b_dotp);
		return(UNISTR_END);
	}

	if( !utf8_isvalid(ch)) {
		mlwrite("%EError in %l: %s", bmlPromptHistp->b_dotp, utf8_perror());
		return UNISTR_END;
	}

	return(ch);
}

/* Read in a key mapping for all future tty input.
 *
 * command arguments [IGNORED]
 *
 * The count is ignored.
 *
 * Command:  map-key
 * Bound to: unbound
 *
 * Return
 *	TRUE  success
 *	FALSE bad Unicode character entered
 */
int	mapkey(int f, Cnt n)
{
	UniChar		from;		/* Char to go from */
	UniChar		to;		/* Char to go to */
	CharMap**	cmpp;

	/* Prompt the user to type in a key to bind */
	tcpys(playPrompt, ": map-key (press the key or U+Unicode_value, press ENTER); from ");
	if((from = mapInterpKey(ttext(playPrompt))) == UNISTR_END)
		return(FALSE);

	printToText(playPrompt, "%A""%C (U+%X) to ", from, (long)from);
	if((to = mapInterpKey(ttext(playPrompt))) == UNISTR_END)
		return(FALSE);

	// It is simpler to do this the long way round.
	// First delete any existing mapping:
	for(cmpp = &CharMapChain; *cmpp; cmpp = &((*cmpp)->chm_next))
		if((*cmpp)->chm_from == from) {
			CharMap*	cmp = *cmpp;
			*cmpp = (*cmpp)->chm_next;
			free((void*)cmp);
			break;
		}

	// Also if US-ASCII straight map:
	if(from < (UniChar)sizeof(keymap))
		keymap[from] = (BYTE)from;

	// If one of the characters won't fit into US-ASCII then it won't
	// fit into the bytemap. So: put something in the list.
	// Keep sorted.
	if(from >= (UniChar)sizeof(keymap) || to >= (UniChar)sizeof(keymap)) {
		CharMap*	cmp;

		if(!(cmp = malloc(sizeof(CharMap))))
			return(FALSE);	// maybe ought to Die

		cmp->chm_from = from;
		cmp->chm_to = to;

		for(cmpp = &CharMapChain; ; cmpp = &((*cmpp)->chm_next)) {
			if( !*cmpp || (*cmpp)->chm_from > from) {
				cmp->chm_next = *cmpp;
				*cmpp = cmp;
				break;
			}
		}
	} else {
		// A simple map:
		keymap[from] = (BYTE)to;
	}

	// Say what is happening, unless running a file:
	if( !clexec) {
		printToText(playPrompt, "%A""%C (U+%X)", to, (long)to);
		mlwrite("%I""%t", playPrompt);
	}

	return(TRUE);
}

/* Initialise the mapping table to be nul effect, ie everything maps to itself.
 */
void	mapinit(void)
{
	BYTE*		mp;
	unsigned int	i;	// Used to be type BYTE - but Minix clang generated bad code

	for(mp = keymap, i = 0; i < sizeof(keymap); )
		*mp++ = (BYTE)i++;
}

/* To be included in what is produced by list-buffers
 */
int	WriteInputKeyMappingToBuffer(BUFFER* buffer)
{
	int		ret = 0;
	BYTE*		mp;
	CharMap*	cmp;
	int		i;
	BYTE		line[100];	// Fixed size buffer - but quite safe

	ret |= !addline("", buffer, 1);
	ret |= !addline("Keyboard Input Key Mapping", buffer, 1);
	ret |= !addline("--------------------------", buffer, 1);

	/* Not as straightforwards as it might seem.
	 * The controls and Basic Latin need to be looked for both in keymap
	 * and in the CharMapChain.
	 */
	for(cmp = CharMapChain, mp = keymap, i = 0; i < (int)sizeof(keymap); i++) {
		UniChar	to = mp[i];

		if(cmp && cmp->chm_from == i) {
			to = cmp->chm_to;
			cmp = cmp->chm_next;
		}
		if(to != i) {
			sprintf(line, "    %c  -> %s", (char)i, utf8_encode(to));
			ret |= !addline(line, buffer, 1);
		}
	}

	// Now what is left from the list of mappings
	for( ; cmp; cmp = cmp->chm_next) {
		char	from[10];
		strcpy(from, utf8_encode(cmp->chm_from));	// Beware: static buffer

		sprintf(line, "    %s%s -> %s", from, (mk_wcwidth(cmp->chm_from) == 2 ? "" : " "),
			utf8_encode(cmp->chm_to));
		ret |= !addline(line, buffer, 1);
	}


	return( !ret);
}

/* end */
