/* New, Perl like, regular expressions
 *
 * This is pre-alpha code. As of Oct 2018 the regex compiler is
 * written, the matcher code has not been started.
 * For a description see: Design/regex
 *
 * ADDW SCCS: @(#)newRE.c	1.15 09/22/20 22:44:38
 *
 * Copyright (C) Alain D D Williams <addw@phcomp.co.uk>, 2018
 * This is released as free software.
 * Use as you wish, no guarantee for anything at all is given.
 */


#include	"estruct.h"
#include	"utf8.h"
#include	"edef.h"

/* Defining this TRUE results in a standalone program that will do regression
 * testing on the new regular expression code.
 * This will be made with the make target reTest
 * For more details look where this define is tested below.
 */
#ifndef	RE_TEST_PROGRAM
#define	RE_TEST_PROGRAM	0
#endif

/* Atom values, must fit in a BYTE, but might be followed
 * by some value bytes.
 */
#define	RE_START		0x00	// Start of line ^
#define	RE_END			0x01	// End of line $
#define	RE_STR			0x02	// String literal
#define	RE_VAR			0x03	// Variable
#define	RE_SOW			0x04	// Start of word \<
#define	RE_EOW			0x05	// End of word \>
#define	RE_WORDB		0x06	// Word boundary \b
#define	RE_ANY			0x07	// Any: .
#define	RE_PROP			0x08	// Unicode property
#define	RE_CLASS		0x09	// Character class [...]
#define	RE_LINE			0x0a	// Line: \n
#define	RE_REP			0x0b	// Repetition
#define	RE_ALT			0x0c	// Alternate: |
#define	RE_GRP			0x0d	// Group: (...)
#define	RE_REFG			0x0e	// Reference to group: \1
#define	RE_SET_GN		0x0f	// Set the number of the current group, ie next will be +1
#define	RE_OPT			0x10	// Option: (?i)
#define	RE_HINT			0x11	// Hint to run time
#define	RE_VERB_MASK		0x1f	// Mask for all of the above

/* Flags that also fit in that atom BYTE
 * The same bit can mean different things depending on the VERB.
 *
 * There is an assumption later that only RE_NOT sets the top bit,
 * otherwise lcatc() might break in UTF-8 mode.
 */
#define	RE_NOT			0x80	// Not flag

#define	RE_NON_GREEDY		0x40	// Non greedy flag on RE_REP
#define	RE_LEN			0x40	// Byte count on atom where length usually implicit
					// Recognised on RANGE, ALT
#define	RE_NO_CAPTURE		0x40	// RE_GRP does not capture
#define	RE_VAR_MATCH		0x40	// Something like $2, VAR

#define	RE_AEND			0x20	// Atom end, where Atom is GRP or ALT
#define	RE_POSSESSIVE		0x20	// Flag on RE_REP - exclusive with RE_NON_GREEDY

// Combinations of the above:

// Replacement codes -- AKA the atoms above
#define	REP_CAP			0x00	// Something captured
#define	REP_VAR			0x01	// Variable
#define	REP_STR			0x02	// Literal string
#define	REP_CNT			0x03	// Count of repetitions
#define	REP_LINE		0x04	// LINE - \n
#define	REP_HINT		0x05	// Hint

/* Atoms within a Class
 * Some are the same as atoms outside of classes, but keep separate.
 * Don't use atom value 0x00 - messes up the way shortHandExp[] is used (NUL term string)
 */
#define	REC_STR			0x01	// String literal
#define	REC_RANGE		0x02	// Character rang
#define	REC_PROP		0x03	// Unicode property
#define	REC_POSIX		0x04	// POSIX class
#define	REC_VERB_MASK		0x0f	// Mask for the above
#define	REC_NOT			0x80	// Not flag
#define	REC_LEN			0x40	// Length follows Class atom

// Hints
#define	RE_HINT_OPT		0x00	// Options, not just at the very start of the RE
#define	RE_HINT_LINE		0x01	// \n seen: at end of the RE or followed by a metacharacter
#define	RE_HINT_LINE_CH		0x02	// \n seen followed by non metacharacter character
#define	RE_HINT_BACKREFS 	0x03	// Highest numbered back reference, eg: \1
#define	RE_HINT_GROUPS		0x04	// Highest group seen
#define	RE_HINT_SETOPTS		0x05	// All options set

// Replacement hints
#define	REP_HINT_CAPS		0x00	// Maximum replacement seen


/* Return codes:
 * OK			The RE was compiled and is in re_comp
 * INCOMPLETE		The RE could be OK if there was more - this is needed
 *			for interactive RE searching
 * ERROR		Non specific error
 *
 * Anything other than OK and INCOMPLETE will prob just be treated as an
 * error and playError printed. They are, however, useful for testing
 * of this code.
 */
#define	RE_OK			0	// Compile OK
#define	RE_INCOMPLETE		1	// More input expected
#define	RE_ERROR		2	// Non specific error
#define	RE_BAD_CONTROL		3	// Bad \cX
#define	RE_BAD_CODEPOINT	4	// Bad \x{XXX}
#define	RE_CODEPOINT_RANGE	5	// Codepoint out of range
#define	RE_BAD_UTF8		6	// Invalid UTF-8 sequence
#define	RE_OOM			7	// Out of memory
#define	RE_CIRC_PLACE		8	// ^ not at start of RE
#define	RE_BAD_OPTION		9	// Unknown option (?x)
#define	RE_UNKNOWN_GRP_START	10	// Unknown at start of group (?X
#define	RE_NO_GRP_OPEN		11	// Missing group open. Seen ')' but no '('
#define	RE_INTERNAL		12	// Internal error - look at playError
#define	RE_NO_RANGE_START	13	// In a class, no char before '-', eg [--b]
#define	RE_RANGE_BAD		14	// Bad 'char' in range, eg [a-\s]
#define	RE_NOT_IMPLEMENTED	15	// Unimplemented feature
#define	RE_BAD_CLASS_ESCAPE	16	// Something like [\q]
#define	RE_RANGE_ORDER		17	// Bad range order, eg [9-0]
#define	RE_EMPTY_CLASS		18	// Empty class
#define	RE_MISSING_END_BRACE	19	// Missing '}' after '{a,b' repetition
#define	RE_CANNOT_REPEAT	20	// A repetition after something that cannot be repeated
#define	RE_NO_REP_AT_START	21	// Cannot have a repetition at the start of something - see playError
#define	RE_UNKNOWN_REFERENCE	22	// Unknown reference, eg \1 when there are no groups
#define	RE_UNKNOWN_ESCAPE	23	// Unknown escape, eg \q
#define	RE_UNKNOWN_PROPERTY	24	// Unknown Unicode property
#define	RE_BAD_POSIX_CLASS	25	// Bad POSIX class syntax or unknown POSIX class
#define	RE_BAD_REPETITION_ORDER	26	// {n,m} where n > m
#define	RE_TOO_MANY_GROUPS	27	// More than RE_GRP_MAX groups seen
#define	RE_DOLLAR_PLACE		28	// $ not at end of RE


/* Some flags about atoms:
 */
#define	RE_CAN_REP		0x01		// Can have repetition
#define	RE_REP_NEED_CONTENT	0x02		// Needs content to be able to repeat

// Indexed by RE_VERB_MASK
BYTE	AtomFlags[] = {
	[RE_START]=	0,					// ^
	[RE_END]=	0,					// $
	[RE_STR]=	RE_CAN_REP,				// String literal
	[RE_VAR]=	RE_CAN_REP,				// Variable
	[RE_SOW]=	0,					// Start of word \<
	[RE_EOW]=	0,					// End of word \>
	[RE_WORDB]=	0,					// Word boundary \b
	[RE_ANY]=	RE_CAN_REP,				// Any: .
	[RE_PROP]=	RE_CAN_REP,				// Unicode property
	[RE_CLASS]=	RE_CAN_REP,				// Character class [...]
	[RE_LINE]=	RE_CAN_REP,				// Line: \n
	[RE_REP]=	0,					// Repetition
	[RE_ALT]=	RE_REP_NEED_CONTENT,			// Alternate: |
	[RE_GRP]=	RE_CAN_REP | RE_REP_NEED_CONTENT,	// Group: (...)
	[RE_REFG]=	RE_CAN_REP,				// Reference to group: \1
//	[RE_RANGE]=	0,					// Character range, only within a CLASS
	[RE_OPT]=	0,					// Option: (?i)
	[RE_HINT]=	0,					// Hint to search engine
};

/* Names of atoms - messages
 * Indexed by RE_VERB_MASK
 */
char*	AtomNames[] = {
	[RE_START]=	"start of line",	// ^
	[RE_END]=	"end of line",		// $
	[RE_STR]=	"text",			// String literal
	[RE_VAR]=	"variable",		// Variable
	[RE_SOW]=	"start of word",	// Start of word \<
	[RE_EOW]=	"end of word",		// End of word \>
	[RE_WORDB]=	"word boundary",	// Word boundary \b
	[RE_ANY]=	"any character",	// Any: .
	[RE_PROP]=	"Unicode property",	// Unicode property
	[RE_CLASS]=	"character class",	// Character class [...]
	[RE_LINE]=	"line end",		// Line: \n
	[RE_REP]=	"repetition",		// Repetition
	[RE_ALT]=	"alternate",		// Alternate: |
	[RE_GRP]=	"group",		// Group: (...)
	[RE_REFG]=	"group reference",	// Reference to group: \1
//	[RE_RANGE]=	"char range",		// Character range, only within a CLASS
	[RE_OPT]=	"option",		// Option: (?i)
	[RE_HINT]=	"hint",			// Hint to search engine
};

typedef	struct	{
	TEXT*	re_comp;	// Compiled Regex
	TEXT*	re_rcomp;	// Compiled replacement
	int	re_groups;	// # Groups seen - at least 0 == everything matched
// error in playError
//	TEXT*	re_error;	// Description of error when compiling
} RegEx;

/* Scanning options
 * Then a string of the flags that the user can give for these options,
 * these MUST be in the order corresponding to the bit position.
 */
#define	RE_OPT_I		0x01	// case insensitive
#define	RE_OPT_E		0x02	// Equivalence
#define	RE_OPT_S		0x04	// Span spaces
#define	RE_OPT_DEFAULT		(RE_OPT_I | RE_OPT_E | RE_OPT_S)
#define	RE_OPT_EXACT_DEFAULT	0x00	// If EXACT mode or seen an upper char in RE
static	char*	optNames = "ies";

/* Escapes that yeild a character, processed by re_escChar()
 * Beware the ordering, see how this is used in re_escChar().
 */
static	char	EscapeToChar[] = "\\trfeacx";


/* Describe a frame.
 * A Frame is anything that contains something else, eg a GRP '(...)' or an ALT '|'.
 *
 * The Frame knows how long its contents are and what the OPTs are
 */
typedef	struct	{
	LineLn	f_start;	// Where the Frame started in output string
				// This will be the Atom verb byte
	BYTE	f_type;		// '(' or '|'
	int	f_flag;		// Frame flags, depends on f_type
				// RE_NO_CAPTURE
	int	f_opts;		// Options ... RE_OPT_???
	int	f_mopts;	// More options
#define	RE_MOPT_BRN	0x01	// Branch reset numbering
	int	f_gn_start;	// Group # at start of the frame
	int	f_gn_hwm;	// Group # high water mark
//	int	f_init_grp;	// Group # at start of the frame
//	int	f_end_grp;	// Group # at end of the frame
} Frame;
// RE_FRAMES > RE_GRP_MAX as frames used for '|' and some '(' are non capture
#define	RE_FRAMES	20	// # elements in frames array
#define	RE_GRP_MAX	255	// Highest group # allowable

static	int	reDebug;	// If true print stuff as we do things

/* Lookup value for POSIX character classes.
 * This will need to be reworked for full Unicode
 */
typedef	struct	{
	char*	pc_name;			// POSIX class name
	BYTE*	pc_expr;			// Class expression
} PCclass;

// Taken from wctype(3) & perlre(1) & locale(5), info: Classification of Characters
// The table below assumes ASCII
PCclass	rePOSIX[] = {
	{ "alnum", (BYTE[]){REC_RANGE | REC_LEN, (BYTE)6, 'a', 'z', 'A', 'Z', '0', '9'} },
	{ "alpha", (BYTE[]){REC_RANGE | REC_LEN, (BYTE)4, 'a', 'z', 'A', 'Z'} },
	{ "ascii", (BYTE[]){REC_RANGE | REC_LEN, (BYTE)2, '\0', '\x7f'} },
	{ "blank", (BYTE[]){REC_STR,             (BYTE)2, ' ', '\t'} },
	{ "cntrl", (BYTE[]){REC_RANGE | REC_LEN, (BYTE)4, '\0', '\x1f'} },
	{ "digit", (BYTE[]){REC_RANGE | REC_LEN, (BYTE)2, '0', '9'} },
	{ "graph", (BYTE[]){REC_RANGE | REC_LEN, (BYTE)2, '!', '~'} },
	{ "lower", (BYTE[]){REC_RANGE | REC_LEN, (BYTE)2, 'a', 'z'} },
	{ "print", (BYTE[]){REC_RANGE | REC_LEN, (BYTE)2, ' ', '~'} },
	// punct is any printing character that is not alphanumeric or a space character
	{ "punct", (BYTE[]){REC_RANGE | REC_LEN, (BYTE)8, '!', '/', ':', '@', '[', '`', '{', '~'} },
	{ "space", (BYTE[]){REC_STR,             (BYTE)6, ' ', '\f', '\n', '\r', '\t', '\v'} },
	{ "upper", (BYTE[]){REC_RANGE | REC_LEN, (BYTE)2, 'A', 'Z'} },
	{ "word",  (BYTE[]){REC_RANGE | REC_LEN, (BYTE)8, 'a', 'z', 'A', 'Z', '0', '9', '_', '_'} }, // From Perl
	{ "xdigit", (BYTE[]){REC_RANGE| REC_LEN, (BYTE)6, '0', '9', 'a', 'f', 'A', 'F'} },
	{ NULL }
};

/* Unicode propery names
 * This is an initial attempt for recognition, but does not do anything useful - yet.
 */
typedef	struct	{
	char*	nm_name;
	char*	nm_sname;
	int	nm_index;
} NameMap;
NameMap	PropNames[] = {
	{ "l",	"letter",			0},
	{ "lc",	"casedletter",			1},
	{ "lu",	"uppercaseletter",		2},
	{ "ll",	"lowercaseletter",		3},
	{ "lt",	"titlecaseletter",		4},
	{ "lm",	"modifierletter",		5},
	{ "lo",	"otherletter",			6},
	{ "m",	"mark",				7},
	{ "mn",	"nonspacingmark",		8},
	{ "mc",	"spacingmark",			9},
	{ "me",	"enclosingmark",		10},
	{ "n",	"number",			11},
	{ "nd",	"decimalnumber",		12},
	{ "nl",	"letternumber",			13},
	{ "no",	"othernumber",			14},
	{ NULL,	NULL,				0},
};


/* Append length encoded as UTF-8 to text.
 * Also good to append any number or Unicode character.
 */
void	reCatLen(TEXT* text, UniChar len)
{
	BYTE*	enc = utf8_encode(len);
	tcatm(text, enc, CHAR_LEN(*enc));
}

/* Insert the encoded character (or length) into text at offSet.
 */
void	reInsLen(TEXT* text, LineLn offSet, UniChar ch)
{
	BYTE*	enc = utf8_encode(ch);
	tinsm(text, offSet, enc, CHAR_LEN(*enc));
}

/* Read an escape that yields one character: \ctrfeax
 * If there is an error, assign to playError
 *
 * *bp points at the character after the \ and will be updated to point to
 * the character after the escape.
 * The character is *appended* to playText.
 *
 * Beware using strchr(), if the char being looked for is NUL
 * the NUL byte at the end of string will be found.
 *
 * Return one of: RE_OK RE_ERROR RE_INCOMPLETE
 */
int	re_escChar(RegEx* cR, BYTE**bp)
{
	BYTE*	p = *bp;
	BYTE	b;
	char*	fmt;
	char	curly;	// The '}' that follows Unicode #
	UniChar	uc;	// Unicode char
	UniChar	ucMax;	// Maximum value of \x{Unicode char}

	switch(*p++) {
	  case 'c':	// Control character
		b = *p;
		if(b == '\0') {
			tcpys(playError, "Incomplete control: \\c");
			return(RE_INCOMPLETE);
		}
		if(Isletter(b) || strchr("@;<=>?{|}~[\\]^_", b)) {
			b = ChrToCtrl(b);
			tcatm(playText, &b, 1);	// KV: could be '\0'
		} else {
			printToText(playError, "Invalid control: \\c%u", p);
			return(RE_BAD_CONTROL);
		}
		p++;
		break;
	  case 'x':	// code point: \xhh or one of: \x{hhh} \x{#dec} \x{ooct}
		if(isxdigit(*p) && p[1] == '\0') {
			printToText(playError, "Incomplete control: \\x%c", *p);
			return(RE_INCOMPLETE);
		}
		if(isxdigit(*p) && isxdigit(p[1])) {
			unsigned int	ui;
			sscanf(p, "%2x", &ui);
			uc = (UniChar)ui;
		} else {
			BYTE*	curlyAddr;

			if(*p != '{') {
				tcpys(playError, "\\x must be followed by {code-point} or XX (hex)");
				return(*p == '\0' ? RE_INCOMPLETE : RE_BAD_CODEPOINT);
			}
			if( !(curlyAddr = strchr(++p, '}'))) {
				tcpys(playError, "\\x must be followed by {code-point}");
				return(RE_INCOMPLETE);
			}
			fmt = *p == '#' ? ("#%" UNICHAR_LMOD "d%c") :
			      *p == 'o' ? ("o%" UNICHAR_LMOD "o%c") : ("%" UNICHAR_LMOD "x%c");
			curly = 'x';
			if(sscanf(p, fmt, &uc, &curly) < 2 || curly != '}') {
				tcpys(playError, "Code point should be like \\x{20ac}, \\x{#8364} or \\x{o20254}");
				return(curly == '\0' ? RE_INCOMPLETE : RE_BAD_CODEPOINT);
			}
			p = curlyAddr + 1;
		}
		ucMax = curbp->b_mode & MDUTF8 ? UTF8_MAX_CODE_POINT : UCHAR_MAX;
		if(uc > ucMax) {
			printToText(playError, "Code point must be no more than \\x{%x}", ucMax);
			return(RE_CODEPOINT_RANGE);
		}
		if(curbp->b_mode & MDUTF8) {
			reCatLen(playText, uc);
		} else {
			b = (BYTE)uc;
			tcatm(playText, &b, 1);	// KV: could be '\0'
		}
		break;
	  case '\0':
		tcpys(playError, "Expecting something after \\");
		return(RE_INCOMPLETE);
	  default:	// One of: \ t r f e a
		tcatc(playText, "\\\t\r\f\e\a"[strchr(EscapeToChar, p[-1]) - EscapeToChar]);
	}

	*bp = p;
	return(RE_OK);
}

/* Read a Unicode property.
 * prePtr	Pointer to pointer to input - the p or P in \px \p{xx}
 * reEnd	Pointer to end of input
 * pnp		Put map found here
 * AtomCode	Assign RE_PROP or RE_PROP|RE_NOT
 *
 * Return RE_OK or another code on error.
 */
int	reReadUniProp(BYTE** prePtr, BYTE* reEnd, NameMap** pnp, BYTE* AtomCode)
{
	char	pCode = (char)**prePtr;	// p or P
	BYTE	byte;
	char	nameBuf[50];		// What we look up, converted to lowercase, _ removed
	char	rNameBuf[80];		// nameBuf - but raw

	*AtomCode = (BYTE)(pCode == 'p' ? RE_PROP : RE_PROP | RE_NOT);

	// \pX or \p{prop}
	if(*prePtr >= reEnd) {
		tcpys(playError, "Incomplete Unicode Character Property");
		return(RE_INCOMPLETE);
	}
	if(*++*prePtr == '{') {
		// Read \p{xxx}
		int	i = 0;
		int	j = 0;

		// \p{^xxx} invert NOT:
		if((*prePtr)[1] == '^') {
			*AtomCode ^= RE_NOT;
			++*prePtr;
		}

		while(i < sizeof(nameBuf) - 1 && j < sizeof(rNameBuf) - 1) {
			if(*prePtr >= reEnd) {
				tcpys(playError, "Incomplete Unicode Character Property");
				return(RE_INCOMPLETE);
			}
			if(*++*prePtr == '}')
				break;
			byte = **prePtr;
			rNameBuf[j++] = (char)byte;

			// Ignore - & _ & spaces, force to lower case:
			if(byte == ' ' || byte == '_' || byte == '-')
				continue;
			if(Isupper(byte))
				byte = (BYTE)tolower(byte);

			nameBuf[i++] = (char)byte;
		}
		nameBuf[i]  = '\0';
		rNameBuf[j] = '\0';
	} else {
		// \px
		byte = **prePtr;
		nameBuf[0] = (char)byte;
		rNameBuf[1] = '\0';
		if(Isupper(byte))
			byte = (BYTE)tolower(byte);
		nameBuf[0] = (char)byte;
		nameBuf[1] = '\0';
	}
	(*prePtr)++;

	// Look up property name
	for(*pnp = PropNames; ; (*pnp)++) {
		if((*pnp)->nm_name == NULL) {
			printToText(playError, "Unknown Unicode property '%s'", rNameBuf);
			return(RE_UNKNOWN_PROPERTY);
		}
		// Found it ?
		if( !strcmp(nameBuf, (*pnp)->nm_name) || !strcmp(nameBuf, (*pnp)->nm_sname))
			break;
	}

	printToText(playError, "Unicode properies not yet implemented: \\%c", pCode);
	return(RE_NOT_IMPLEMENTED);

	// return(RE_OK);	// eventually
}

/*
 * re	Regular expression, a NUL terminated string.
 *	This might not be complete - not all of it as been typed (yet)
 *	It can contain a NUL (which does not terminate it), but must have a NUL after the end
 * cR	Compiled re is put here.
 *
 * Bugs:
 * *	^ must be the the first char in the RE, should allow it to be
 *	preceeded by options, eg: (?-i)^
 * *	groupsSeen should be within the frame, it can go down again in (?| ... )
 *
 * Return:
 *	0  OK
 *	1  Error in re, error messages in global playError
 *	2  Incomplete RE - a message will be in playError in case the
 *	   caller thinks that the re is complete.
 */
int	reCompile(TEXT* re, RegEx* cR)
{
	BYTE*	rePtr = tgetcaddr(re, 0);	// Pointer into re
	BYTE*	reEnd = tgetcaddr(re, tused(re)); // Address just beyond the end of re - might be a \0
	TEXT*	cRe;				// Compiled RE
	int	r_min, r_max;			// REP minimum & maximum
	BYTE	AtomCode;
	BYTE	byte;
	LineLn	frameStart;			// Start of the current frame
	LineLn	pos;
	// cAtomStart must be offset as TEXT could be realloc()ed
	LineLn	cAtomStart;			// Start of Compiled Atom
	LineLn	lastcAS = -1;			// Last value of cAtomStart
	BYTE*	iAtomStart;			// Where the current atom parsing started
	BYTE*	play;
	int	i, j, len;
// Remove groupsSeen, groupNesting - make part of frame
	int	frameCount = 0;			// Top frame used
	int	groupsSeen = 0;			// # Groups seen in their entirety for the purpose of \1, etc
	int	groupNesting = 0;		// # Group '(' seen but not closed
	int	maxRefSeen = 0;			// Max \n reference seen
// Check when the above are set

	int	seenNewline = 0;		// Seen a \n in the RE
	BYTE	endChar[UTF8_MAX_BYTES];	// Char after first \n
	int	endCharValid = 0;

	// Option manipulation
	int	seenUpper = 0;			// Seen an upper case character in the RE
	int	seenOption = 0;			// Saw an option setting - not at very start of RE
	int	initOptsSet = RE_OPT_DEFAULT;
	int	opts;				// Used when see (?opts)
	int	seenQ;				// Seen ? after (
	int	everSetOpt = 0;			// Every option set

	int	utf8 = curbp->b_mode & MDUTF8;

	Frame	frames[RE_FRAMES];
//	Frame*	cF;				// Current frame
//	Frame*	pF;				// Parent frame

	// Positions - escaped characters:
	static	BYTE posChars[] = "<>bB";
	// Corresponding Atom codes:
	static	BYTE posCodes[] = { RE_SOW, RE_EOW, RE_WORDB, RE_WORDB | RE_NOT };

	// List of meta characters:
	static	char	metaCharacters[]    = ".?+*[^$\\|{()";

	// List of chars made ordinary by prefixing with backslash:
	static	char	escMetaCharacters[] = ".?+*[]^$\\|{}()";	// Has ']}' not in metaCharacters

	static	char	repChars[] = "?+*{";		// Chars that are/start a REPetition

	// Shorthands for character classes
	static	BYTE	shortHands[] = "sSdDwW";	// In same order as table below
	static	BYTE	shortHandExp[][14] = {		// These are NUL terminated strings
		{ RE_CLASS, (BYTE)5, REC_STR,                       (BYTE)3, '\t', ' ', '\v', '\0'},		// \s
		{ RE_CLASS, (BYTE)5, REC_STR   | REC_NOT,           (BYTE)3, '\t', ' ', '\v', '\0'},		// \S
		{ RE_CLASS, (BYTE)3, REC_RANGE,                              '0', '9', '\0'},			// \d
		{ RE_CLASS, (BYTE)3, REC_RANGE | REC_NOT,                    '0', '9', '\0'},			// \D
		{ RE_CLASS, (BYTE)8, REC_RANGE |           REC_LEN, (BYTE)8, 'a', 'z', 'A', 'Z', '0', '9', '_', '_', '\0'},	// \w
		{ RE_CLASS, (BYTE)8, REC_RANGE | REC_NOT | REC_LEN, (BYTE)8, 'a', 'z', 'A', 'Z', '0', '9', '_', '_', '\0'},	// \W
	};
// TODO: u8 mode, \s includes more

	if(utf8 && !utf8_chkbuf(ttext(re), (size_t)tused(re))) {
		tcpys(playError, "Regex is corrupt: invalid UTF8");
		return(RE_BAD_UTF8);
	}

	/* This TEXT must not have the UTF-8 flag set
	 * otherwise UTF-8 checking will be made on non UTF-8 bytes.
	 */
	if(!cR->re_comp && (cR->re_comp = talloc(0)) == (TEXT*)NULL) {
		tcpys(playError, OutOfMemory);
		return(RE_OOM);
	}
	cRe = cR->re_comp;
	tsetempty(cRe);		// Clear out what might have been there

	// Implicit () round everything, group 0:
	frames[0].f_start = 0;
	frames[0].f_type = '(';
	frames[0].f_mopts = 0;
	frames[0].f_gn_start = 0;
	frames[0].f_gn_hwm = 0;
#define	RE_GRP_LEN	1
// was when group had # group after it 
//	tcatm(cRe, (BYTE[]){RE_GRP, (BYTE)0}, RE_GRP_LEN);	// Which is why nothing has been added if lused(cRe) == RE_GRP_LEN
	tcatm(cRe, (BYTE[]){RE_GRP}, RE_GRP_LEN);	// Which is why nothing has been added if lused(cRe) == RE_GRP_LEN

	/* Initial scan to look for an upper char
	 * The complication is that we ignore some of them that are part of, eg an escape
	 * This does a good enough job for this scan, but will accept errors
	 * (that should get picked up later).
	 */
	while(rePtr < reEnd) {
		byte = *rePtr++;

		if(byte == '(' && rePtr[1] == '?' && rePtr[2] == '#') {
			// Span to end of comment
			while(rePtr < reEnd && *rePtr++ != ')')
				;
			continue;
		}

		if(byte == '$' && rePtr[1] == '{') {
			// Span ${variable}
			while(rePtr < reEnd && *rePtr++ != '}')
				;
			continue;
		}

		if(byte == '\\') {
			// \px \Px \p{} \P{} \x{} \cx
			byte = *rePtr++;
			if(byte == 'c') {	// \cx
				rePtr++;
				continue;
			}
			if(((byte == 'p' || byte == 'P' || byte == 'x') && *rePtr == '{')) {
				while(rePtr < reEnd && *rePtr++ != '}')
					;
				continue;
			}
			if(byte == 'x')		// \xXX
				rePtr += 2;
			// Ignore \<anything-else> eg \S \px
			continue;
		}

		if(Isupper(byte)) {
			seenUpper = 1;
			break;
		}
	}

	// Initial options to not be caseless ?
	if((curbp->b_mode & MDEXACT) != 0 || seenUpper)
		initOptsSet = RE_OPT_EXACT_DEFAULT;
	frames[0].f_opts = initOptsSet;
if(seenUpper)printf("SCAN seenUpper\n");
//reDebug=1;

	rePtr = tgetcaddr(re, 0);

	while(rePtr < reEnd) {
		if(reDebug) {
			printf("Loop ch='%c' cRe.used=%d lastcAS=%ld", *rePtr, (int)tused(cRe), lastcAS);
			if(lastcAS != -1)
				printf(" Verb=%#2.2x '%s'", tgetch(cRe, lastcAS), AtomNames[tgetch(cRe, lastcAS) & RE_VERB_MASK]);
			putchar('\n');

			int	reo = 0;
			printf("Bytes:");
			while(reo < tused(cRe))
				printf(" %2.2x", tgetch(cRe, reo++));
			printf("\n");
		}

		cAtomStart = tused(cRe);
		iAtomStart = rePtr;

		// Be really careful using lcatc() below, if the value is > 0x7f
		// then it will be converted to something multibyte.
		switch(*rePtr++) {
		  case '^':	// Start of line
				// Test offset one as just did ++
			if(rePtr != tgetcaddr(re, 1)) {
				tcpys(playError, "^ not at start of regex");
				return(RE_CIRC_PLACE);
			}
			tcatc(cRe, RE_START);
			break;

		  case '$':	// End of line or Variable or $n
			if(rePtr == tgetcaddr(re, tused(re))) {
				tcatc(cRe, RE_END);
				break;
			}

			// Something follows the $

			if(Isdigit(*rePtr)) {
				// Something like $3
				int	grp = 0;

				while(Isdigit(*rePtr)) {
					grp = grp * 10 + (*rePtr - '0');

					// End of RE implied end of back reference
					if(++rePtr >= reEnd)
						break;
				}

				if(grp > RE_GRP_MAX) {
					printToText(playError, "No more than %d groups allowed, at $%d", RE_GRP_MAX, grp);
					return(RE_TOO_MANY_GROUPS);
				}

				tcatm(cRe, (BYTE[]){RE_VAR | RE_VAR_MATCH, (BYTE)grp}, 2);

				break;
			}

			if(*rePtr == '{') {
				// ${variable_name} or ${}

				if(++rePtr >= reEnd) {
					tcpys(playError, "Incomplete comment");
					return(RE_INCOMPLETE);
				}

				// Empty variable - ignored
				if(*rePtr == '}') {
					rePtr++;
					break;
				}

				// Code not complete
				tcpys(playError, "${variable} not parsed\n");
				return(RE_NOT_IMPLEMENTED);
			}

			tcpys(playError, "$ not at end of RE");
			return(RE_DOLLAR_PLACE);

		  case '\n':	// Newline
			tcatc(cRe, RE_LINE);
			if( !seenNewline && rePtr < reEnd && rePtr + CHAR_LEN(*rePtr) <= reEnd &&
			    *rePtr != '\0' && !strchr(metaCharacters, *rePtr)) {
				endCharValid = 1;
				UTF8_MOVE(endChar, rePtr);
			}
			seenNewline = 1;
			break;

		  case '.':	// Any character
			tcatc(cRe, RE_ANY);
			break;

		  case '|':	// Alternate
			frameStart = frames[frameCount].f_start;
if(0)printf("Seen | frameStart = %ld frameCount=%d type='%c' lused=%ld\n", frameStart, frameCount, frames[frameCount].f_type, tused(cRe));
// Look at combining this test with the one just below
			if(frames[frameCount].f_type == '(')
				frameStart++;

			// Insert the previous ALT and its length
			pos = tused(cRe);

			if(frames[frameCount].f_type != '|') {
				frames[frameCount + 1] = frames[frameCount];
				// New ALT frame, put an ALT at the start of playText
				frameCount++;
				frames[frameCount].f_type = '|';
				frames[frameCount].f_flag = frames[frameCount - 1].f_flag;	// Get MOPT_BRN
				frames[frameCount].f_gn_start = frames[frameCount - 1].f_gn_start;
				frames[frameCount].f_gn_hwm = frames[frameCount - 1].f_gn_hwm;

				tcpyc(playText, RE_ALT | RE_LEN);

				reCatLen(playText, (UniChar)(pos - frameStart));
if(0){int reo=0;
	printf("playText=");
	while(reo < tused(playText))
		printf(" %2.2x", tgetch(playText, reo++));
	printf("\n");
}
if(0){int reo=0;
	printf("cRe=");
	while(reo < tused(cRe))
		printf(" %2.2x", tgetch(cRe, reo++));
	printf("\n");
}
				tinst(cRe, frameStart, playText);


				frames[frameCount].f_start = tused(cRe);
				tcatc(cRe, (BYTE)(RE_ALT | RE_LEN));
				if(reDebug)printf("ALT insert first | at %ld f_start=%ld\n", frameStart, (long)frames[frameCount].f_start);
			} else {
				// Insert the length after the previous ALT
				reInsLen(cRe, frameStart + 1, (UniChar)(pos - frameStart - 1));

				// New ALT
				frames[frameCount].f_start = tused(cRe);
				tcatc(cRe, (BYTE)(RE_ALT | RE_LEN));
				if(reDebug)printf("ALT insert succ len %d at %ld new frame@ %ld\n", (UniChar)(pos - frameStart), frameStart+1, (long)frames[frameCount].f_start);

				// Branch reset numbering ?
				if(frames[frameCount].f_mopts & RE_MOPT_BRN) {
					groupNesting = frames[frameCount].f_gn_start;
					tcatm(cRe, (BYTE[]){(BYTE)RE_SET_GN, (BYTE)groupNesting}, 2);
				}
			}

			if(groupsSeen > frames[frameCount].f_gn_hwm)
				frames[frameCount].f_gn_hwm = groupsSeen;
			if((frames[frameCount].f_mopts & RE_MOPT_BRN) != 0)
				groupsSeen = frames[frameCount].f_gn_start;

			// Will put the length in when we know what it is
			cAtomStart = tused(cRe) - 1;
			if(reDebug)printf("ALT:: gnest=%d cAtomStart=%ld\n", frameCount, (long)cAtomStart);
			break;

		  case '(':	// Group, comment or options:
			seenQ = 0;
			opts = frames[frameCount].f_opts;
			tsetempty(playText);		// Put options string here
			if(*rePtr == '?') {
				rePtr++;
				seenQ = 1;

				// comment
				if(*rePtr == '#') {
					while(*++rePtr != ')')
						if(rePtr >= reEnd) {
							tcpys(playError, "Incomplete comment");
							return(RE_INCOMPLETE);
						}

					++rePtr;	// Eat the ')'

					// This does not exist in output, fudge start
					cAtomStart = lastcAS;
					break;
				}

				// Options
				// Accumulate in playText, might not want to apply
				byte = RE_OPT;
				while(Isletter(*rePtr) || *rePtr == '-') {

					if(rePtr >= reEnd) {
						tcpys(playError, "Incomplete option");
						return(RE_INCOMPLETE);
					}

					// - unsets all following options
					if(*rePtr == '-') {
						rePtr++;
						byte = RE_OPT | RE_NOT;
					}

					if(rePtr >= reEnd) {
						tcpys(playError, "Incomplete option");
						return(RE_INCOMPLETE);
					}

					// Check valid option
					if(*rePtr && (play = strchr(optNames, *rePtr)) != NULL) {
						tcatm(playText, (BYTE[]){byte, *rePtr}, 2);

						i = 1 << (play - (BYTE*)optNames);

						// Tweak current options
						if(byte & RE_NOT) {
							opts &= ~i;
						} else {
							opts |= i;
							everSetOpt |= i;
						}

						// Initial option ?
						if(tused(cRe) == RE_GRP_LEN) {
							// Manipulate initial options
							if(byte & RE_NOT)
								initOptsSet &= ~i;
							else
								initOptsSet |= i;
						}

						rePtr++;	// Eat the option character

						continue;
					}
					printToText(playError, "Unknown option '%u'", rePtr);
					return(RE_BAD_OPTION);

				}

				if(*rePtr == ')') {
					rePtr++;	// Eat the ')'

					frames[frameCount].f_opts = opts;

					// Not an initial option ?
					// Generate Atoms & note that a non-initial option was seen
					if(tused(cRe) != RE_GRP_LEN) {
						tcatt(cRe, playText);
						seenOption = 1;
					}

					break;
				}
			}

			// Start of a group:
			frameCount++;
			frames[frameCount].f_type = '(';
			frames[frameCount].f_flag = 0;
			frames[frameCount].f_opts = opts;	// Inherit (possibly changed) options
			frames[frameCount].f_mopts = 0;

			/* All sorts of strange stuff may come after (?
			 * But the only thing recognised here is ':' - not capture expression
			 */
			if(seenQ) {
				
				if(rePtr >= reEnd) {
					tcpys(playError, "Expecting something after (?");
					return(RE_INCOMPLETE);
				}

				switch(*rePtr) {
				  case ':':	// A non capturing sub expression ?
					rePtr++;	// Skip :
					frames[frameCount].f_flag = RE_NO_CAPTURE;

					/* Option specific to this group ?
					 * If not ensure that playText is empty for when it is appended to cRe below
					 */
					if(tused(playText) != 0) {
						if(tused(cRe) != RE_GRP_LEN)
							seenOption = 1;
						else
							tsetempty(playText);
					}

					break;
				  case '|':	// Branch reset
					frames[frameCount].f_mopts |= RE_MOPT_BRN;
					frames[frameCount].f_flag = RE_NO_CAPTURE;	// Does not capture
					groupNesting = frames[frameCount].f_gn_start;
					break;
				  default:
					printToText(playError, "Unknown sequence '(?%u'", rePtr);
					return(RE_UNKNOWN_GRP_START);
				}
			}

			// Set count of open/capturing groups:
			if((frames[frameCount].f_flag & RE_NO_CAPTURE) == 0)
				groupNesting++;
			frames[frameCount].f_gn_start = frames[frameCount].f_gn_hwm = groupNesting;

			frames[frameCount].f_start = tused(cRe);
			if(reDebug)printf("seen ( startpos=%ld\n", (long)frames[frameCount].f_start);
			// Start the group, we'll insert the length later when we know what it is
			tcatc(cRe, (BYTE)(RE_GRP | frames[frameCount].f_flag));

			// Add group specific options (if any):
			tcatt(cRe, playText);

			break;

		  case ')':	// Group:

			/* Close frames opened inside the group.
			 * This should only be ALTs.
			 * Frame 0 is a GRP - ie f_type == '('.
			 */
			while(frames[frameCount].f_type != '(') {
				if(frames[frameCount].f_type == '|') {
					// Insert the length after the previous ALT
					reInsLen(cRe, frames[frameCount].f_start + 1, (UniChar)(tused(cRe) - frames[frameCount].f_start - 1));

					// Now the ending ALT
					tcatc(cRe, (BYTE)(RE_ALT | RE_AEND));
				} else {
					printToText(playError, "BUG: at ')' seen frame '%c'", frames[frameCount].f_type);
					return(RE_INTERNAL);
				}
				--frameCount;
			}

			if(groupNesting < frames[frameCount].f_gn_start) {
				tcpys(playError, "Seen group close ')' but no matching '('");
				return(RE_NO_GRP_OPEN);
			}

			if(groupsSeen > frames[frameCount].f_gn_hwm)
				frames[frameCount].f_gn_hwm = groupsSeen;

			pos = tused(cRe);
			tcatc(cRe, (BYTE)(RE_GRP | RE_AEND));	// End the group

			// Branch reset numbering ?
			if(frames[frameCount].f_mopts & RE_MOPT_BRN) {
				groupNesting = frames[frameCount].f_gn_hwm;
				tcatm(cRe, (BYTE[]){(BYTE)RE_SET_GN, (BYTE)groupNesting}, 2);
			}

			// Insert length of the group after the GRP code at the start of the group:
			i = (frames[frameCount].f_flag & RE_NO_CAPTURE) == 0 ? RE_GRP_LEN : 1;
			reInsLen(cRe, frames[frameCount].f_start + i, (UniChar)(pos - frames[frameCount].f_start - 1));

			cAtomStart = frames[frameCount].f_start;
			if(reDebug)printf("end of ) cAtomStart=%ld ='%#2.2x' len=%"UNICHAR_LMOD"d startpos=%ld\n", (long)cAtomStart, tgetch(cRe, cAtomStart), (UniChar)(pos - frames[frameCount].f_start), (long)frames[frameCount].f_start);

			// Group captured ?
			if((frames[frameCount].f_flag & RE_NO_CAPTURE) == 0) {
				if(++groupsSeen > RE_GRP_MAX) {
					printToText(playError, "No more than %d groups allowed", RE_GRP_MAX);
					return(RE_TOO_MANY_GROUPS);
				}
			}

			--frameCount;

			if(frames[frameCount + 1].f_gn_hwm > frames[frameCount].f_gn_hwm)
				frames[frameCount].f_gn_hwm = frames[frameCount + 1].f_gn_hwm;
			groupsSeen = frames[frameCount].f_gn_hwm;

			break;
		  case '[':	// Class
			/* Use of playText? TEXTs
			 * playText	Used by re_escChar()
			 * playText2	Build up individual chars to be a STR
			 * playText3	Build up the class in playText3
			 * playText4	The last character read; which might be the first of a range
			 * playText5	The first of a range
			 */
			AtomCode = RE_CLASS;
			if(*rePtr == '^') {	// ^ is NOT CLASS
				AtomCode = RE_CLASS | RE_NOT;
				rePtr++;
			}
			tcatm(cRe, &AtomCode, 1);

			tsetempty(playText2);
			tsetempty(playText3);
			tsetempty(playText4);

			// A - or ] at the start don't have their normal meaning
			if(*rePtr == '-') {
				tcatc(playText2, '-');
				rePtr++;
			}
			if(*rePtr == ']') {
				tcatc(playText2, ']');
				rePtr++;
			}

			/* Hunt through looking for:
			 * * single characters
			 *   * literal
			 *   * \something-that-is-one-codepoint
			 * * a range which is 2 single characters separated by a '-'
			 *	the first will be, temporarily stashed in playText4.
			 * * a Unicode property - recognised/parsed only not acted
			 * * a class, eg: \s
			 * * POSIX classes: [:class:] [:^class:]
			 */
			enum {
				start,		// Expecting a character or shortcut
				gotC1,		// Seen the first char
				gotMinus	// Seen a - want another character
			} cState = start;

			while(*rePtr != ']') {
				if(rePtr >= reEnd + 1) {
					tcpys(playError, "Class started with '[' but ']' is missing");
					return(RE_INCOMPLETE);					
				}

				// [:a or [:^a - ie start of POSIX class ?
				if(*rePtr == '[' && cState == start && rePtr < reEnd + 3 && rePtr[1] == ':' &&
				   ((rePtr[2] == '^' && Isletter(rePtr[3])) || Isletter(rePtr[2]))) {
					char	className[10];

					// Looks like it
					rePtr += 2;
					int	not = 0;
					if(*rePtr == '^') {
						not = REC_NOT;
						rePtr++;
					}
					i = 0;
					while(Isletter(*rePtr) && i < sizeof(className) - 1) {
						if(rePtr >= reEnd + 1) {
							tcpys(playError, "POSIX Class started with '[:' but ':]' is missing");
							return(RE_INCOMPLETE);
						}

						className[i++] = (char)*rePtr++;
					}
					className[i] = '\0';
					if(rePtr[0] != ':' || rePtr[1] != ']') {
						tcpys(playError, "POSIX Class started with '[:' but ':]' not seen");
						return(RE_BAD_POSIX_CLASS);
					}
					rePtr += 2;	// Skip :]

					// Look up class name
					PCclass* pcp;
					for(pcp = rePOSIX; ; pcp++) {
						if(pcp->pc_name == NULL) {
							printToText(playError, "Unknown POSIX class '%s'", className);
							return(RE_BAD_POSIX_CLASS);
						}
						if( !strcmp(pcp->pc_name, className))
							break;
					}

					// Append POSIX & index to class:
					tcatm(playText3, (BYTE[]){REC_POSIX | REC_NOT, (BYTE)(pcp - rePOSIX)}, 2);
					continue;
				}

				// Looks like a range: a-z
				// But accept '-' as the second character in a range, eg: +--
				if(*rePtr == '-' && cState != gotMinus) {
					if(rePtr[1] == ']') {
						// Deal with ending '-]'
// upper
						tcatt(playText2, playText4);
						tcpyc(playText4, '-');
						rePtr++;
						continue;	// Which will stop the loop
					}

					// Found '-' it must be after (at least) one character
					if(cState == start) {
						tcpys(playError, "Range (-) but no character before it");
						return(RE_NO_RANGE_START);
					}

#if 0
					// Remove the last character from the STR (in playText2)
					// put into playText4
					play = tgetcaddr(playText2, tused(playText2) - 1);
					if(utf8) {
						UTF8_BACK_FIND_START(play);
						len = CHAR_LEN(*play);
					} else {
						play--;
						len = 1;
					}
					tcpym(playText4, play, len);
					tsetused(playText2, tused(playText2) - len);
#endif

					// Now set state to grab the second in the range
					rePtr++;
					cState = gotMinus;
					tcpyt(playText5, playText4);
					tsetempty(playText4);
//printf("copy to range '%s'\n", ttext(playText5));
					continue;
				}

				if(*rePtr == '\\') {
					rePtr++;
					if(rePtr >= reEnd) {
						tcpys(playError, "Escape within character class is incomplete");
						return(RE_INCOMPLETE);
					}

					// \s etc
					if(*rePtr && (play = strchr(shortHands, *rePtr)) != (BYTE*)NULL) {
						if(cState == gotMinus) {
							printToText(playError, "\\%c cannot be part of range", *rePtr);
							return(RE_RANGE_BAD);
						}

						/* Construct a CLASS within this class
						 * Simple optimisation, if the CLASS is not NOT - skip 2 bytes
						 * and just copy the RANGE
						 */
						play = shortHandExp[play - shortHands] + 2;
						tcats(playText3, play);

						rePtr++;
						cState = start;

						continue;
					}

					// with what follows, put a character into playText
					else if(*rePtr && strchr(EscapeToChar, *rePtr)) {
						// Standard escapes that yeild a character
						tsetempty(playText);
						if((j = re_escChar(cR, &rePtr)) != RE_OK)
							return(j);
//						tcpyt(playText4, playText);
					} else if(*rePtr == '-' || *rePtr == ']') {
						// \- and \]
						tcpyc(playText, *rePtr++);
					} else if(*rePtr == 'p' || *rePtr == 'P') {
						// Unicode property:
						// REC_PROP
						NameMap*pnp;
						if((j = reReadUniProp(&rePtr, reEnd, &pnp, &AtomCode)) != RE_OK)
							return(j);
					} else {
						printToText(playError, "Unrecognised escape in class: \\%u", rePtr);
						return(RE_BAD_CLASS_ESCAPE);
					}

				} else {

					// First check if it is an upper case character

					// Copy any last char to the growing STR
					if(Isupper(tgetch(playText4, 0)) && ((frames[frameCount].f_opts & RE_OPT_I) != 0)) {
						BYTE	up[UTF8_MAX_BYTES];
						up[0] = (BYTE)tolower(tgetch(playText4, 0));
						len = utf8 ? CHAR_LEN(up[0]) : 1;
						tcatm(playText2, up, len);
					} else {
						// Not an upper char or not in case insensitive mode
						tcatt(playText2, playText4);
					}

					// Copy the next character to playText
					len = utf8 ? CHAR_LEN(*rePtr) : 1;
					tcpym(playText, rePtr, len);
//printf("playline='%s'\n", ttext(playText));
					rePtr += len;
				}

				// If seen a '-' then create a RANGE
				if(cState == gotMinus) {
#if 0
					/* Don't do this. Let search work out case independence in classes
					 * This did not work anyway, consider: [@-_]
					 */
					if((frames[frameCount].f_opts & RE_OPT_I) != 0) {
						if(Isupper(tgetch(playText5, 0))) {
							BYTE	up[UTF8_MAX_BYTES];
							up[0] = (BYTE)tolower(tgetch(playText5, 0));
							len = utf8 ? CHAR_LEN(up[0]) : 1;
							tcpym(playText5, up, len);
						}
						if(Isupper(tgetch(playText, 0))) {
							BYTE	up[UTF8_MAX_BYTES];
							up[0] = (BYTE)tolower(tlgetch(playText, 0));
							len = utf8 ? CHAR_LEN(up[0]) : 1;
							tcpym(playText, up, len);
						}
						
					}
#endif
					// Sanity check:
					if(memcmp(ttext(playText5), ttext(playText), UTF8_MAX_BYTES) > 0) {
						printToText(playError, "Invalid character class range, bad order [%u-%u]", ttext(playText5), ttext(playText));
						return(RE_RANGE_ORDER);
					}

//printf("Create range '%s' - '%s' playline2 '%s' 3='%s' 4='%s'\n", ttext(playText5), ttext(playText), ttext(playText2), ttext(playText3), ttext(playText4));
					// Append the RANGE to the growing class
					tcatc(playText3, REC_RANGE);
					tcatt(playText3, playText5);
					tcatt(playText3, playText);
					tsetempty(playText4);		// Nothing there now
					tsetempty(playText5);		// Nothing there now
					tsetempty(playText);		// Nothing there now

//printf("after range 0='%s' 2='%s' 3='%s' 4='%s' 5='%s'\n", ttext(playText), ttext(playText2), ttext(playText3), ttext(playText4), ttext(playText5));
					cState = start;
					continue;
				}

				tcpyt(playText4, playText);
				tsetempty(playText);
				cState = gotC1;
			}

			rePtr++;	// Skip the ]

			// If we have any literal characters - append to the class
//printf("END range 0='%s' 2='%s' 3='%s' 4='%s' 5='%s'\n", ttext(playText), ttext(playText2), ttext(playText3), ttext(playText4), ttext(playText5));
			tcatt(playText2, playText4);
			if(tused(playText2)) {
				tcatc(playText3, REC_STR);
				play = utf8_encode((UniChar)tused(playText2));	// length
				tcatm(playText3, play, CHAR_LEN(*play));
				tcatt(playText3, playText2);
			}

			if(tused(playText3) == 0) {
				tcpys(playError, "Empty class: []");
				return(RE_EMPTY_CLASS);
			}

			// Insert the length of the CLASS
			play = utf8_encode((UniChar)tused(playText3));
			tcatm(cRe, play, CHAR_LEN(*play));
			tcatt(cRe, playText3);
			break;

		  case '?':	// REP 0,1
			r_min = 0;
			r_max = 1;
			goto rep_store;
		  case '+':	// REP 1,0
			r_min = 1;
			r_max = 0;
			goto rep_store;
		  case '*':	// REP 0,0
			r_min = 0;
			r_max = 0;
			goto rep_store;
		  case '{':	// REP r_min,r_max
			r_min = 0;	// Set defaults
			r_max = 0;

			// Read n after {
			if(Isdigit(*rePtr)) {
				// atoi() & strspn() are obvious here, but get silly warnings on strspn() with gcc
				r_min = (int)strtol(rePtr, (char**)&rePtr, 10);
				// {n} is the same as {n,n}
				if(*rePtr == '}')
					r_max = r_min;
			}

			// Read n after ,
			if(*rePtr == ',') {
				rePtr++;
				if(Isdigit(*rePtr)) {
					r_max = (int)strtol(rePtr, (char**)&rePtr, 10);
					// min must not be more than max
					if(r_min > r_max) {
						printToText(playError, "Bad repetition {%d,%d} as %d is > %d", r_min, r_max, r_min, r_max);
						return(RE_BAD_REPETITION_ORDER);
					}
				}
			}

			if(*rePtr != '}') {
				printToText(playError, "Missing '}' after %u", iAtomStart);
				return(rePtr >= reEnd ? RE_INCOMPLETE : RE_MISSING_END_BRACE);
			}
			rePtr++;	// Span the }

		rep_store:
			if(reDebug)printf("REP: prev is %#2.2x\n", tgetch(cRe, lastcAS));

			/* Check that the repetition is after something that can be repeated.
			 * Start of string - bit of a kludge, but it yields a sensible message.
			 */
			AtomCode = (BYTE)(lastcAS >= 0 ? tgetch(cRe, lastcAS) : RE_START);
			if((AtomFlags[AtomCode & RE_VERB_MASK] & RE_CAN_REP) == 0) {
				printToText(playError, "Cannot repeat %s", AtomNames[AtomCode & RE_VERB_MASK]);
				return(RE_CANNOT_REPEAT);
			}

			if(reDebug)printf("last=%s cRe used=%ld lastcAS=%d\n", AtomNames[AtomCode & RE_VERB_MASK], (long)tused(cRe), (int)lastcAS);

			if((AtomFlags[AtomCode & RE_VERB_MASK] & RE_REP_NEED_CONTENT) &&
			   tused(cRe) == lastcAS + 1) {
				printToText(playError, "Cannot have repetition at start of %s", AtomNames[AtomCode & RE_VERB_MASK]);
				return(RE_NO_REP_AT_START);
			}

			// Check if non-greedy:
			AtomCode = RE_REP;
			if(*rePtr == '?') {
				rePtr++;
				AtomCode = RE_REP | RE_NON_GREEDY;
			} else if(*rePtr == '+') {
				// Possessive repetition
				rePtr++;
				AtomCode = RE_REP | RE_POSSESSIVE;
			}

			// Set the group nesting if we are repeating a group, this could have
			// further internal groups that muck up numbering.
			// Don't do this if the last Atom was Group # set - will happen with repeated branch reset
			if((tgetch(cRe, lastcAS) & RE_VERB_MASK) == RE_GRP && (tgetch(cRe, tused(cRe) - 2) & RE_VERB_MASK) != RE_SET_GN)
				tcatm(cRe, (BYTE[]){(BYTE)RE_SET_GN, (BYTE)groupNesting}, 2);

			// Move to before the atom that repeats
			tcpym(playText, &AtomCode, 1);
			reCatLen(playText, r_min);
			reCatLen(playText, r_max);
			reCatLen(playText, (UniChar)(tused(cRe) - lastcAS));	// Length
			tinst(cRe, lastcAS, playText);

			cAtomStart = lastcAS;

			break;

		  case '\\':	// Escape something
			if(rePtr >= reEnd) {
				tcpys(playError, "Expecting something after \\");
				return(RE_INCOMPLETE);
			}

			// Process the ones that mark a place, cannot be part of a class
			if(*rePtr && (play = strchr(posChars, *rePtr)) != (BYTE*)NULL) {
				play = &posCodes[play - posChars];
				tcatm(cRe, play, 1);
				rePtr++;
				break;
			}

			// Unicode property:
			if(*rePtr == 'p' || *rePtr == 'P') {
				// RE_PROP
				NameMap*pnp;
				if((j = reReadUniProp(&rePtr, reEnd, &pnp, &AtomCode)) != RE_OK)
					return(j);
			}

			// Is it a meta character that can be made ordinary ?
			if(*rePtr && strchr(escMetaCharacters, *rePtr)) {
				// It might be nice to try to join it with preceeding/following
				// characters making a longer string. Next release, maybe
				tcatm(cRe, (BYTE[]){RE_STR, (BYTE)1, *rePtr++}, 3);
				break;
			}

			if(*rePtr == 'Q') {	// \Q is quote
				rePtr++;
				tsetempty(playText);

				// Copy everything until \E or end of input
				while(rePtr < reEnd && (rePtr[0] != '\\' || (rePtr + 1 >= reEnd ? 1 : rePtr[1] != 'E')))
					tcatm(playText, rePtr++, 1);

				// Generate a string if there was anything:
				if(tused(playText) > 0) {
					tcatm(cRe, (BYTE[]){RE_STR}, 1);
					reCatLen(cRe, (UniChar)tused(playText));
					tcatt(cRe, playText);
				}

				rePtr += 2;	// Might put it beyond reEnd, that is OK: immediate stop
				break;
			}

			if(*rePtr == 'n') {	// Newline
				tcatc(cRe, RE_LINE);
				rePtr++;
				if( !seenNewline && rePtr < reEnd && rePtr + CHAR_LEN(*rePtr) <= reEnd &&
				    *rePtr != '\0' && !strchr(metaCharacters, *rePtr)) {
					endCharValid = 1;
					UTF8_MOVE(endChar, rePtr);
				}
				seenNewline = 1;
				break;
			}

			if(Isdigit(*rePtr)) {	// Backreference to group. Eg \1 or \123
				int	grp = 0;

				while(Isdigit(*rePtr)) {
					grp = grp * 10 + (*rePtr - '0');

					// End of RE implied end of back reference
					if(++rePtr >= reEnd)
						break;
				}

				// Don't allow \0 or where the number refers to a group not yet complete
				if(grp > groupsSeen || grp == 0) {
					printToText(playError, "Unknown group reference \\%d", grp);
					return(RE_UNKNOWN_REFERENCE);
				}

				tcatm(cRe, (BYTE[]){RE_REFG, (BYTE)grp}, 2);
				if(grp > maxRefSeen)
					maxRefSeen = grp;

				break;
			}

			// All that is valid now is what could be found in a character class
			if(*rePtr && strchr(EscapeToChar, *rePtr)) {
				tsetempty(playText);
				if((j = re_escChar(cR, &rePtr)) != RE_OK)
					return(j);

				len = utf8 ? CHAR_LEN(tgetch(playText, 0)) : 1;
				tcatm(cRe, (BYTE[]){RE_STR, (BYTE)len}, 2);
				tcatm(cRe, tgetcaddr(playText, 0), len); 
				break;
			}

			// Things like \s
			if(*rePtr && (play = strchr(shortHands, *rePtr)) != (BYTE*)NULL) {
				tcats(cRe, shortHandExp[(int)(play - shortHands)]);
				rePtr++;
				break;
			}

			printToText(playError, "Unknown escape: \\%u", rePtr);
			return(RE_UNKNOWN_ESCAPE);

		  default:
			/* A character to add to a string
			 * Read the string, if the last is followed by a REP
			 * add it as a separate STR. Eg abc? attaches ? to the c
			 */
			tsetempty(playText);

			i = 0;		// Count of characters - not bytes
			rePtr--;	// Look at what brought us here

			// Scan while there are sequences that can be converted to characters
			while(rePtr < reEnd) {
				// \<something that resolves to a character>
				if(*rePtr == '\\') {
					if(rePtr >= reEnd) {
						tcpys(playError, "Expecting something after \\");
						return(RE_INCOMPLETE);
					}

					rePtr++;

					// Things like \t
					if(*rePtr && strchr(EscapeToChar, *rePtr)) {
						if((j = re_escChar(cR, &rePtr)) != RE_OK)
							return(j);
						i++;
						continue;
					}

					// Meta characters made ordiary
					if(*rePtr && strchr(escMetaCharacters, *rePtr)) {
						tcatm(playText, rePtr, 1);
						rePtr++;
						i++;
						continue;
					}

					// \<something not understoof above>
					// Process below, make the \ the current character again
					rePtr--;
				}

				// Metacharacters end the string as does a newline
				if((*rePtr && strchr(metaCharacters, *rePtr)) || *rePtr == '\n')
					break;

				// Append the character, copy the correct # bytes

				// First check if it is an upper case character
				if(Isupper(*rePtr) && ((frames[frameCount].f_opts & RE_OPT_I) != 0)) {
					BYTE	up[UTF8_MAX_BYTES];
					up[0] = (BYTE)tolower(*rePtr);
					len = utf8 ? CHAR_LEN(up[0]) : 1;
					tcatm(playText, up, len);
				} else {
					len = utf8 ? CHAR_LEN(*rePtr) : 1;
					tcatm(playText, rePtr, len);
				}
				rePtr += len;
				i++;
			}

			len = (int)tused(playText);	// Length of string
			j = 0;
			if(reDebug)printf("str len=%d\n", (int)len);

			/* If more than 1 character & followed by a REP
			 * split the last character into a separate STR
			 */
			if(i > 1 && *rePtr != '\0' && strchr(repChars, *rePtr)) {
				play = tgetcaddr(playText, tused(playText) - 1);
				len = 1;
				if(utf8) {
					UTF8_BACK_FIND_START(play);
					len = CHAR_LEN(*play);
				}

				tinsm(playText, (LineLn)(play - ttext(playText)), (BYTE[]){RE_STR, (BYTE)len}, 2);
				j = len + 2;

				// Length of string prior to the STR inserted
				len = (int)(play - ttext(playText));

				if(reDebug)printf("after seen REP len=%d j=%d\n", (int)len, j);
			}

			if(reDebug)printf("app str with .len=%ld cRe.len=%ld\n", (long)tused(playText)+2, (long)tused(cRe));

			// Put STR, append the length of the string then the string:
			tcatm(cRe, (BYTE[]){RE_STR}, 1);
			reCatLen(cRe, len);
			tcatt(cRe, playText);

			// If split off last char, point cAtomeStart at that last char
			if(j != 0)
				cAtomStart = tused(cRe) - j;
			break;
		}

		lastcAS = cAtomStart;
	}

	if(reDebug) {
		printf("Loop end @ %d lastcAS=%ld", (int)tused(cRe), (long)lastcAS);
		if(lastcAS != -1)
			printf(" Verb=%#2.2x '%s'", tgetch(cRe, lastcAS), AtomNames[tgetch(cRe, lastcAS) & RE_VERB_MASK]);
		putchar('\n');
	}

	// Output any ALTs that need ending
	while(frameCount > 0) {
		if(frames[frameCount].f_type == '|') {
			// The can auto close, so do so
			pos = tused(cRe);

			if(reDebug)printf("END gn=%d insert len %"UNICHAR_LMOD"d at %ld then ALT|END\n", frameCount, (UniChar)(pos - frames[frameCount].f_start), (long)frames[frameCount].f_start);
			// Insert the length after the previous ALT
			reInsLen(cRe, frames[frameCount].f_start + 1, (UniChar)(pos - frames[frameCount].f_start - 1));

			// Now the ending ALT
			tcatc(cRe, (BYTE)(RE_ALT | RE_AEND));

			frameCount--;
		} else {
			// Drop down & return incomplete
			break;
		}
	}

	if(frameCount > 0) {
		tcpys(playError, "Seen group start '(' but not enough matching ')'");
		return(RE_INCOMPLETE);
	}

	// Insert length of group 0 that encloses the entire RE, then end the group
	reInsLen(cRe, RE_GRP_LEN, (UniChar)(tused(cRe) - 1));
	tcatc(cRe, (BYTE)(RE_GRP | RE_AEND));

	/* At the very start of the cRe (before the GRP) we may want to
	 * insert options & hints
	 */
	tsetempty(playText);
if(1){
	// Set default option EXACT - if needed:
	if(initOptsSet == RE_OPT_EXACT_DEFAULT)
		tcatm(playText, (BYTE[]){RE_OPT, (BYTE)'E'}, 2);
	else	// Set the options that are left:
		for(i = 0; optNames[i] != '\0'; i++)
			if((initOptsSet & (1 << i)) != 0)
				tcatm(playText, (BYTE[]){RE_OPT, (BYTE)optNames[i]}, 2);

	if(seenOption)
		tcatm(playText, (BYTE[]){RE_HINT, RE_HINT_OPT}, 2);

	if(seenNewline) {
		if(endCharValid) {
			tcatm(playText, (BYTE[]){RE_HINT, RE_HINT_LINE_CH}, 2);
			tcatm(playText, (BYTE*)endChar, CHAR_LEN(endChar[0]));
		} else
			tcatm(playText, (BYTE[]){RE_HINT, RE_HINT_LINE}, 2);
	}

	if(maxRefSeen)
		tcatm(playText, (BYTE[]){RE_HINT, RE_HINT_BACKREFS, (BYTE)maxRefSeen}, 3);

	if(groupsSeen)
		tcatm(playText, (BYTE[]){RE_HINT, RE_HINT_GROUPS, (BYTE)groupsSeen}, 3);
	cR->re_groups = groupsSeen;

	if(everSetOpt)
		tcatm(playText, (BYTE[]){RE_HINT, RE_HINT_SETOPTS, (BYTE)everSetOpt}, 3);

	if(tused(playText) != 0)
		tinst(cRe, 0, playText);
}
	cR->re_comp = cRe;

	// Did we run out of memory at some stage ?
	if(textOOM) {
		tcpys(playError, OutOfMemory);	// This might, of course, fail
		return(RE_OOM);
	}

	return(RE_OK);
}

/* Compile a replacement string.
 *
 * The string is almost too simple to bother to compile, but
 * compilation lets us detect errors.
 *
 * Return:
 * RE_OK		OK
 * anything else	Error in rep, error messages in global playError
 *			Error numbers as reCompile()
 */
int	repCompile(TEXT* rep, RegEx* cRep)
{
	BYTE*	repPtr = tgetcaddr(rep, 0);	// Pointer into rep
	BYTE*	repEnd = tgetcaddr(rep, tused(rep)); // Address just beyond the end of rep - might be a \0
	TEXT*	cLine;				// Compiled REP
	int	utf8 = curbp->b_mode & MDUTF8;
	int	i, j;
	int	len;
	char*	dollar;
	int	maxCap = -1;			// Highest number of captured string


	static	char	escRepMetaCharacters[] = "&\\";	// Escape to get ordinary

	if(utf8 && !utf8_chkbuf(ttext(rep), (size_t)tused(rep))) {
		tcpys(playError, "Regex is corrupt: invalid UTF8");
		return(RE_BAD_UTF8);
	}

	/* This TEXT must not have the UTF-8 flag set
	 * otherwise UTF-8 checking will be made on non UTF-8 bytes.
	 */
	if(!cRep->re_rcomp && (cRep->re_rcomp = talloc(0)) == (TEXT*)NULL) {
		tcpys(playError, OutOfMemory);
		return(RE_OOM);
	}
	cLine = cRep->re_rcomp;
	tsetempty(cLine);	// Clear out what might have been there
	
	while(repPtr < repEnd) {
		switch(*repPtr++) {
		  case '&':	// Captured 0
			tcatm(cLine, (BYTE[]){REP_CAP, (BYTE)0}, 2);
			if(maxCap < 0)
				maxCap = 0;
			break;
		  case '\n':	// Line
			tcatc(cLine, REP_LINE);
			break;
		  case '\\':	// Escape
			if(repPtr >= repEnd) {
				tcpys(playError, "Expecting something after \\");
				return(RE_INCOMPLETE);
			}

			// Is it a meta character that can be made ordinary ?
			if(*repPtr && strchr(escRepMetaCharacters, *repPtr)) {
				// It might be nice to try to join it with preceeding/following
				// characters making a longer string. Next release, maybe
				tcatm(cLine, (BYTE[]){REP_STR, (BYTE)1, *repPtr++}, 3);
				break;
			}

			if(*repPtr == 'n') {	// Newline
				tcatm(cLine, (BYTE[]){REP_LINE}, 1);
				repPtr++;
				break;
			}

			if(*repPtr == '#') {	// Count
				tcatm(cLine, (BYTE[]){REP_CNT}, 1);
				repPtr++;
				break;
			}

			dollar = "";
			if(*repPtr == '$') {	// Something like \$1 or \${$variable}
				repPtr++;
				if( !Isdigit(*repPtr)) {
					// Not a backreference to group. Must be variable
					if(*repPtr != '{') {
						printToText(playError, "Bad escape '\\$%s'", repPtr);
						return(RE_UNKNOWN_ESCAPE);
					}

					if(++repPtr >= repEnd) {
						tcpys(playError, "Expecting something after \\${");
						return(RE_INCOMPLETE);
					}

					// \${} is allowed, substitutes nothing, use as separator
					if(*repPtr == '}') {
						repPtr++;
						break;
					}

					// Deal with variable
					printf("VARiables in replacement not yet available\n");
					break;
				}

				// Digit - drop to next if
				dollar = "$";
			}
			if(Isdigit(*repPtr)) {	// Backreference to group. Eg \1 and pick up \$1 from above
				int	grp = 0;

				while(Isdigit(*repPtr)) {
					grp = grp * 10 + (*repPtr - '0');

					// End of REP implied end of back reference
					if(++repPtr >= repEnd)
						break;
				}

				// Don't allow where the number refers to a group not seen
				if(grp > cRep->re_groups) {
					printToText(playError, "Unknown group reference \\%s%d", dollar, grp);
					return(RE_UNKNOWN_REFERENCE);
				}

				tcatm(cLine, (BYTE[]){REP_CAP, (BYTE)grp}, 2);
				if(maxCap < grp)
					maxCap = grp;
				break;
			}

			// All that is valid now are character class escapes
			if(*repPtr && strchr(EscapeToChar, *repPtr)) {
//printf("EscapeToChar %#2.2x\n", *repPtr);
				tsetempty(playText);
				if((i = re_escChar(cRep, &repPtr)) != RE_OK)
					return(i);

				len = utf8 ? CHAR_LEN(tgetch(playText, 0)) : 1;
				tcatm(cLine, (BYTE[]){REP_STR, (BYTE)len}, 2);
				tcatm(cLine, tgetcaddr(playText, 0), len); 
				break;
			}
//printf("unknown REP '%c'\n", *repPtr);
			printToText(playError, "Unknown replacement escape: \\%s%u", dollar, repPtr);
			return(RE_UNKNOWN_ESCAPE);

			break;
		  default:	// Literal characters
			/* A character to add to a string
			 */
			tsetempty(playText);

			i = 0;		// Count of characters - not bytes
			repPtr--;	// Look at what brought us here

			// Scan while there are sequences that can be converted to characters
			while(repPtr < repEnd) {
				// \<something that resolves to a character>
				if(*repPtr == '\\') {
					if(++repPtr >= repEnd) {
						tcpys(playError, "Expecting something after \\");
						return(RE_INCOMPLETE);
					}

					// Things like \t
					if(*repPtr && strchr(EscapeToChar, *repPtr)) {
//printf("S-EscapeToChar %#2.2x\n", *repPtr);
						// re_escChar() can return RE_INCOMPLETE
						// for this function that is an error
						if((j = re_escChar(cRep, &repPtr)) != RE_OK)
							return(j);
						i++;
						continue;
					}

					// Meta characters made ordinary
					if(*repPtr && strchr(escRepMetaCharacters, *repPtr)) {
						tcatm(playText, repPtr, 1);
						repPtr++;
						i++;
						continue;
					}

					// \<something not understoof above>
					// Process below, make the \ the current character again
					repPtr--;
				}

				// Metacharacters end the string as does a newline
				if((*repPtr && strchr(escRepMetaCharacters, *repPtr)) || *repPtr == '\n')
					break;

				// Append the character, copy the correct # bytes
				len = utf8 ? CHAR_LEN(*repPtr) : 1;
				tcatm(playText, repPtr, len);
				repPtr += len;
				i++;
			}

			len = (int)tused(playText);	// Length of string
			if(reDebug)printf("str len=%d\n", (int)len);

			// Put STR, append the length of the string then the string:
			tcatm(cLine, (BYTE[]){REP_STR}, 1);
			reCatLen(cLine, len);
			tcatt(cLine, playText);

			break;
		}
	}

	if(maxCap >= 0)
		tinsm(cLine, 0, (BYTE[]){REP_HINT, REP_HINT_CAPS, (BYTE)maxCap}, 3);

	// Did we run out of memory at some stage ?
	if(textOOM) {
		tcpys(playError, OutOfMemory);	// This might, of course, fail
		return(RE_OOM);
	}

	return(RE_OK);
}


#if	RE_TEST_PROGRAM
/* This results in an executable standalone program.
 *
 * It yanks in all of the other microemacs code (most of which is not used).
 * For this main() in main.c should be called emacs() to not clash with main()
 * below. The Makefile.orig has the target emacs.o that is built with the right
 * flags for this.
 *
 * If you add new features do NOT initially add new tests.
 * First run it and check that you have not broken anything by comparing retest_new.
 * Then add new tests for the new features.
 */

/* A debug/test routine that prints a compiled RE.
 */
void	rePrint(RegEx* cR)
{
	TEXT*	re = cR->re_comp;
	LineLn	reo = 0;
	BYTE	atom;
	int	len, len2;
	int	i, j;
	int	level = 0;		// Nesting level
	int	groupNum = -1;		// Calculated group #. -1 as there is group 0 round it all
	BYTE	atoms[30];		// What the current atom level is: ALT or GRP

	// Clear out:
	atoms[0] = 0;
	memcpy(atoms + 1, atoms, sizeof(atoms) - 1);

	printf("re compiled length %ld\n", (long)lused(re));
	printf("Bytes:");
	while(reo < lused(re))
		printf(" %2.2x", lgetc(re, reo++));
	printf("\n");

	reo = 0;
	atoms[0] = 0;	// Hmmm
	while(reo < lused(re)) {
		atom = lgetc(re, reo++);
//printf("a %#2.2x\n", atom);
		switch(atom & RE_VERB_MASK) {
		  case RE_START:	// ^
			printf("%*s", level * 4, "");
			printf("START ^\n");
			break;
		  case RE_END:		// $
			printf("%*s", level * 4, "");
			printf("END $\n");
			break;
		  case RE_STR:		// String literal
			len = (int)utf8_decode(lgetcaddr(re, reo++));
			printf("%*s", level * 4, "");
			printf("STR len %d '", len);
			while(--len >= 0)	// Not printf, screws on NUL byte
				putchar(lgetc(re, reo++));

			printf("'\n");
			break;
		  case RE_VAR:		// Variable
			printf("%*s", level * 4, "");
			printf("VAR");
			if(atom & RE_VAR_MATCH) {
				printf(" $%d", lgetc(re, reo++));
			} else {
				printf(" Not able to decode");
			}
			putchar('\n');
			break;
		  case RE_SOW:		// Start of word \<
			printf("%*s", level * 4, "");
			printf("START WORD \\<\n");
			break;
		  case RE_EOW:		// End of word \>
			printf("%*s", level * 4, "");
			printf("END WORD \\>\n");
			break;
		  case RE_WORDB:	// Word boundary \b
			printf("%*s", level * 4, "");
			printf(atom & RE_NOT ? "WORD BOUNDARY NOT \\B\n" : "WORD BOUNDARY \\b\n");
			break;
		  case RE_LINE:		// Line: \n
			printf("%*s", level * 4, "");
			printf("NEWLINE \\n\n");
			break;
		  case RE_PROP:		// Unicode property
			printf("%*s", level * 4, "");
			printf("PROP\n");
			break;
		  case RE_ANY:		// Any: .
			printf("%*s", level * 4, "");
			printf("ANY .\n");
			break;
		  case RE_CLASS:	// Character class [...]
			len = lgetc(re, reo++);
			char*	Not = atom & RE_NOT ? " NOT" : "";
			printf("%*s", level * 4, "");
			printf("CLASS%s len %d\n", Not, len);
			level++;
			atoms[level] = RE_CLASS;
			while(len > 0) {
				atom = lgetc(re, reo++);
				Not = atom & REC_NOT ? " NOT" : "";
				len--;
				switch(atom & REC_VERB_MASK) {
				  case REC_PROP:
					printf("%*s", level * 4, "");
					printf("  PROP%s - not implemented\n", Not);
					break;
				  case REC_STR:
					i = lgetc(re, reo++);
					len--;
					printf("%*s", level * 4, "");
					printf("  STR%s len %d '", Not, i);
					while(--i >= 0) {
						putchar(lgetc(re, reo++));
						len--;
					}
					printf("'\n");
					break;
				  case REC_RANGE:
					printf("%*s", level * 4, "");
					printf("  RANGE%s ", Not);
					if(atom & REC_LEN) {
						len2 = lgetc(re, reo++);
						len--;
						printf("LEN %d", len2);
					} else {
						len2 = CHAR_LEN(lgetc(re, reo));
						len2 += CHAR_LEN(lgetc(re, reo + len2));
						printf("IMPLIED LEN %d", len2);
					}
					len -= len2;
					while(len2 > 0) {
						printf(" '");
						j = CHAR_LEN(lgetc(re, reo));
						if(j == 0) {
							printf("BUG CHAR_LEN == 0\n");
							break;
						}
						while(--j >= 0) {
							putchar(lgetc(re, reo++));
							len2--;
						}
						printf("' '");
						j = CHAR_LEN(lgetc(re, reo));
						if(j == 0) {
							printf("BUG CHAR_LEN == 0\n");
							break;
						}
						while(--j >= 0) {
							putchar(lgetc(re, reo++));
							len2--;
						}
						printf("'");
					}
					printf("\n");
					break;
				  case REC_POSIX:
					printf("%*s", level * 4, "");
					i = lgetc(re, reo++);
					len--;
					printf("  POSIX%s index %d = ", Not, i);
					if(i < sizeof(rePOSIX) / sizeof(rePOSIX[0]))
						printf("'%s'\n", rePOSIX[i].pc_name);
					else
						printf("bad index\n");
					break;
				  default:
					printf("  unknown CLASS verb: %#2.2x length left %d\n", atom, len);
				}
			}
			level--;
			break;
		  case RE_REP:		// Repetition
			i = (int)utf8_decode(lgetcaddr(re, reo));
			reo += CHAR_LEN(lgetc(re, reo));
			j = (int)utf8_decode(lgetcaddr(re, reo));
			reo += CHAR_LEN(lgetc(re, reo));
			len = (int)utf8_decode(lgetcaddr(re, reo));
			reo += CHAR_LEN(lgetc(re, reo));
			printf("%*s", level * 4, "");
			printf("REP min=%d, max=%d len=%d", i, j, len);
			if(atom & RE_NON_GREEDY)
				printf(" non-greedy");
			if(atom & RE_POSSESSIVE)
				printf(" possessive");
			printf("\n");
			break;
		  case RE_ALT:		// Alternate: |
//printf("l=%d a[l]=%2.2x\n", level, atoms[level]);
			printf("%*s", level * 4, "");
			if(atoms[level] != RE_ALT)
				atoms[++level] = RE_ALT;
//printf("  l=%d a[l]=%2.2x\n", level, atoms[level]);
			printf("ALT |");
			if(atom & RE_LEN) {
				len = (int)utf8_decode(lgetcaddr(re, reo));
				reo += CHAR_LEN(lgetc(re, reo));
				printf(" len=%d", len);
			}
			if(atom & RE_AEND) {
				printf(" END");
				level--;
			}
			printf("\n");
			break;
		  case RE_GRP:		// Group: (...)
			if(atom & RE_AEND) {
				printf("%*s", level * 4, "");
				printf("END_GROUP\n");
				level--;
				break;
			}
			if((atom & RE_NO_CAPTURE) == 0)
				groupNum++;
			len = (int)utf8_decode(lgetcaddr(re, reo));
			reo += CHAR_LEN(lgetc(re, reo));
			printf("%*s", level * 4, "");
			printf("GROUP");
			level++;
			atoms[level] = RE_GRP;
			if(atom & RE_NO_CAPTURE)
				printf(" NO-CAPTURE");
			else
				printf(" %d", groupNum);	// Group #
			printf(" len=%d\n", len);
			break;
		  case RE_REFG:		// Reference to group: \1
			printf("%*s", level * 4, "");
			printf("REFERENCE %d\n", (int)lgetc(re, reo++));
			break;
		  case RE_OPT:		// Option: (?i)
			printf("%*s", level * 4, "");
			printf("Opt ");
			if(atom & RE_NOT)
				printf("-");
			printf("%c\n", lgetc(re, reo++));
			break;
		  case RE_HINT:		// Hint to RE engine
			printf("%*s", level * 4, "");
			printf("HINT ");
			switch(i = lgetc(re, reo++)) {
			  case RE_HINT_OPT:
				printf("options in RE");
				break;
			  case RE_HINT_LINE:
				printf("newline in RE");
				break;
			  case RE_HINT_LINE_CH:
				len = CHAR_LEN(lgetc(re, reo));
				printf("first newline followed by '%*.*s'", len, len, lgetcaddr(re, reo));
				reo += len;
				break;
			  case RE_HINT_BACKREFS:
				printf(" highest backreference: %d", lgetc(re, reo++));
				break;
			  case RE_HINT_GROUPS:
				printf(" groups: %d", lgetc(re, reo++));
				break;
			  case RE_HINT_SETOPTS:		// All options set
				printf(" all opts: %#2.2x", lgetc(re, reo++));
				break;
			  default:
				printf("unknown code %#2.2x", i);
			}
			putchar('\n');
			break;
		  case RE_SET_GN:	// Set current group #
			printf("%*s", level * 4, "");
			groupNum = (int)lgetc(re, reo++);
			printf("SET GROUP # %d\n", groupNum);
			break;
		  default:
			printf("%*s", level * 4, "");
			printf("Unknown atom verb %#2.2x at offset %ld\n", atom, reo - 1);
			break;
		}
	}
}

void	repPrint(RegEx* cR)
{
	TEXT*	re = cR->re_rcomp;
	LineLn	reo = 0;
	BYTE	atom;
	int	len;

	printf("re compiled length %ld\n", (long)lused(re));
	printf("Bytes:");
	while(reo < lused(re))
		printf(" %2.2x", lgetc(re, reo++));
	printf("\n");

	reo = 0;

	while(reo < lused(re)) {
		atom = lgetc(re, reo++);
//printf("a %#2.2x\n", atom);
		switch(atom) {
		  case REP_CAP:		// Something captured
			printf("CAP group %d\n", lgetc(re, reo++));
			break;
		  case REP_VAR:		// Variable
			printf("VAR \n");
			break;
		  case REP_STR:		// Literal string
			len = (int)utf8_decode(lgetcaddr(re, reo++));
			printf("STR len %d '", len);
			while(--len >= 0)	// Not printf, screws on NUL byte
				putchar(lgetc(re, reo++));

			printf("'\n");
			break;
		  case REP_CNT:		// Count of repetitions
			printf("COUNT\n");
			break;
		  case REP_LINE:	// LINE - \n
			printf("LINE\n");
			break;
		  case REP_HINT:
			atom = lgetc(re, reo++);
			switch(atom) {
			  case REP_HINT_CAPS:
				printf("HINT MAX CAP %d\n", lgetc(re, reo++));
				break;
			  default:
				printf("Unknown hint, code: %#2.2x\n", atom);
			}
			break;
		  default:
			printf("Unknown atom verb %#2.2x at offset %ld\n", atom, reo - 1);
			break;
		}
	}
}

/* Print a line to stream out.
 * The length (bytes) is printed followed by the line text.
 * If out is NULL - nothing is printed.
 *
 * To avoid problems on read/comparing it do a few transformations:
 * \     ->  \\
 * NL    ->  \n
 * CNTRL -> \Cx, eg 0x01 prints as \CA
 */
void	rePrintLine(FILE* out, TEXT* text)
{
	BYTE*	ch = ttext(text);
	LineLn	len = tused(text);

	if( !out)
		return;

	fprintf(out, "len %ld '", (long)len);

	for( ; --len >= 0; ch++) {
		if(*ch == '\\')
			fprintf(out, "\\\\");
		else if(*ch == '\n')
			fprintf(out, "\\n");
		else if(*ch < ' ')
			fprintf(out, "\\C%c", (*ch + '@'));
		else
			fputc(*ch, out);
	}

	fprintf(out, "'\n");
}

/* Span/ignore comment lines in the reference file
 *
 * Comment lines start with a '#'.
 */
void	reSpanComments(FILE* ref)
{
	int	c;

	while((c = fgetc(ref)) == '#') {
		while((c = fgetc(ref)) != '\n')
			if(c == EOF)
				return;
	}

	if(c != EOF)
		ungetc(c, ref);
}


/* Compare a line with what should be the same in the result file.
 * ref	Where to read from, if NULL do now't
 * cntx	The context - something that says something about what we are looking at
 * i	Test count
 * line	The line that we should compare equal to
 *
 * The point of comparison is to point out a difference so that it
 * can be checked. It does not try to be clever and diagnose a problem.
 *
 * This code is not bullet proof & does not recover well from errors in
 * the reference file.
 *
 * Return TRUE if we read the reference correctly and the result is as expected.
 * Also return TRUE if the reference file is not open.
 */
int	reCompareTest(FILE* ref, char* cntx, int i, TEXT* text)
{
	LineLn	len;
	LineLn	cmpLen;		// Length of string that is compared
	long	LongVar;	// For fscanf - LineLn could be int or long
	int	c, j;
	BYTE	b;
	int	res = TRUE;

	if( !ref)
		return(TRUE);

	reSpanComments(ref);

	if((j = fscanf(ref, "len %ld '", &LongVar)) != 1) {
		printf("!! test %d %s fail to parse length in reference file\n", i, cntx);
		return(FALSE);
	}
	cmpLen = len = (LineLn)LongVar;
	if(len != tused(text)) {
		printf("!! test %d %s reference length %ld generated %ld\n", i, cntx, (long)len, (long)tused(text));
		// Compare the smaller of the two:
		cmpLen = len > tused(text) ? tused(text) : len;
	}

	tsetempty(playText);
	tunsetUtf8(playText);

	while(--len >= 0) {
		c = fgetc(ref);
		if(c == EOF) {
			printf("!! test %d %s EOF found in reference\n", i, cntx);
			return(FALSE);
		}
		if(c == '\n') {
			printf("!! test %d %s unexpected newtext found in reference\n", i, cntx);
			return(FALSE);
		}
		if(c == '\\') {
			c = fgetc(ref);

			if(c == '\\')
				;
			else if(c == 'n')
				c = '\n';
			else if(c == 'C')
				c = fgetc(ref) & 0x1F;
			else {
				printf("!! test %d %s bad escape in reference file '%c'\n", i, cntx, c);
				res = FALSE;
			}
		}

		b = (BYTE)c;
		tcatm(playText, &b, 1);
	}

	// Eat "'\n"
	if((c = fgetc(ref)) != '\'' || (c = fgetc(ref)) != '\n') {
		printf("!! test %d %s expecting \"'\\n\" in reference file, found %#2.2x\n", i, cntx, c);
		res = FALSE;

		// Find end of line, give processing the next line a chance of working
		while((c = fgetc(ref)) != '\n' && c != EOF)
			;
	}

	if(textcmp(text, playText) != 0) {
		printf("!! test %d %s generated & reference strings differ\n", i, cntx);
		res = FALSE;
	}

	return(res);
}

/* Compare compile results
 * ref	Where to read from, if NULL do now't
 * fmt	a printf/sscanf format
 * i	Text number
 * ret	compile result expected
 *
 * Return TRUE if we read the reference correctly and the result is as expected.
 * Also return TRUE if the reference file is not open.
 */
int	reCompareRes(FILE* ref, char* fmt, int i, int ret)
{
	int	cnt, res, j;
	char	line[100];	// Buffer results line here

	if( !ref)
		return(TRUE);

	reSpanComments(ref);

	// Read the line into a buffer - so that it can be shown below if there is an error
	if( !fgets(line, sizeof(line), ref)) {
		printf("!! Cannot read reference line for test %d\n", i);
		return(FALSE);
	}

	if((j = sscanf(line, fmt, &cnt, &res)) != 2) {
		printf("!! Bad results line, test %d parsed %d item(s) read '%s'\n", i, j, line);
		return(FALSE);
	}

	if(i != cnt) {
		printf("!! Result compare, expected test %d found %d\n", i, cnt);
		return(FALSE);
	}

	if(ret != res) {
		printf("!! Result compare, test %d got result %d expected %d\n", i, ret, res);
		return(FALSE);
	}

	return(TRUE);
}

typedef	struct	{
	char*	ret_re;		// RE
	int	ret_relen;	// How long it is
	char*	ret_su;		// Subject string
	int	ret_sulen;	// How long it is
	char*	ret_ca;		// Something captured
	int	ret_calen;	// How long it is
	int	ret_grp;	// Which captured group is given in ret_ca
} reTest;
#define	RE_MTEST(re, subj, cap, grp)	{ re, sizeof(re) - 1, subj, sizeof(subj) - 1, cap, sizeof(cap) - 1, grp }

/* Array of lines to search, put into 'scan' buffer
 * Include the length, as can have the NUL byte in the line.
 *
 * Maybe I should read a file ... but this removes external dependencies.
 */
typedef	struct	{
	char*	st_text;	// String
	int	st_len;		// Bytes
} sText;
#define	RE_STEXT(str)	{ str, sizeof(str) - 1 }
sText	SearchText[] = {
	RE_STEXT("\"If\""),
	RE_STEXT(""),
	RE_STEXT("If you can keep your head when all about you"),
	RE_STEXT("Are losing theirs and blaming it on you,"),
	RE_STEXT("If you can trust yourself when all men doubt you"),
	RE_STEXT("But make allowance for their doubting too,"),
	RE_STEXT("If you can wait and not be tired by waiting,"),
	RE_STEXT("Or being lied about, don't deal in lies,"),
	RE_STEXT("Or being hated, don't give way to hating,"),
	RE_STEXT("And yet don't look too good, nor talk too wise:"),
	RE_STEXT(""),
	RE_STEXT("If you can dream--and not make dreams your master,"),
	RE_STEXT("If you can think--and not make thoughts your aim;"),
	RE_STEXT("If you can meet with Triumph and Disaster"),
	RE_STEXT("And treat those two impostors just the same;"),
	RE_STEXT("If you can bear to hear the truth you've spoken"),
	RE_STEXT("Twisted by knaves to make a trap for fools,"),
	RE_STEXT("Or watch the things you gave your life to, broken,"),
	RE_STEXT("And stoop and build 'em up with worn-out tools:"),
	RE_STEXT(""),
	RE_STEXT("If you can make one heap of all your winnings"),
	RE_STEXT("And risk it on one turn of pitch-and-toss,"),
	RE_STEXT("And lose, and start again at your beginnings"),
	RE_STEXT("And never breathe a word about your loss;"),
	RE_STEXT("If you can force your heart and nerve and sinew"),
	RE_STEXT("To serve your turn long after they are gone,"),
	RE_STEXT("And so hold on when there is nothing in you"),
	RE_STEXT("Except the Will which says to them: \"Hold on!\""),
	RE_STEXT(""),
	RE_STEXT("If you can talk with crowds and keep your virtue,"),
	RE_STEXT("Or walk with kings--nor lose the common touch,"),
	RE_STEXT("If neither foes nor loving friends can hurt you;"),
	RE_STEXT("If all men count with you, but none too much,"),
	RE_STEXT("If you can fill the unforgiving minute"),
	RE_STEXT("With sixty seconds' worth of distance run,"),
	RE_STEXT("Yours is the Earth and everything that's in it,"),
	RE_STEXT("And--which is more--you'll be a Man, my son!"),
	RE_STEXT(""),
	RE_STEXT("By Rudyard Kipling (1865-1936)."),
	{ (char*)NULL,	0 }
};

/* Create the buffer to scan
 * Then put some text into it
 */
BUFFER*	reMkScan(char* sbuf)
{
	BUFFER*	sbp;	// Scan buffer pointer
	sText*	slp;	// Scan line pointer - the lines that we are adding

	if( !(sbp = bfind(sbuf, BFNCREATE|BFNACTIVE, 0))) {
		printf("Error: Cannot create buffer %s\n", sbuf);
		return(NULL);
	}

	for(slp = SearchText; slp->st_text != (char*)NULL; slp++) {
		LINE*	ln = BufAppendLine(sbp);

		if( !ln || !lcpym(ln, slp->st_text, slp->st_len)) {
			printf("Error: out of memory while building %s\n", sbuf);
			return(NULL);
		}
	}

	return(sbp);
}


// Test a regex
typedef	struct	{
	char*	it_re;		// Regex to test
	int	it_relen;	// Length of the test - bytes
	int	it_res;		// Expected result
} iTest;
#define	RE_TEST(str, res)	{str, sizeof(str)-1, res}

// Test a replacement
typedef	struct	{
	char*	rt_re;		// Regex to set up the replacement
	char*	rt_rep;		// Replacement to test
	int	rt_relen;	// Length of the regex - bytes
	int	rt_replen;	// Length of the replacement - bytes
	int	rt_res;		// Expected result of replacement
} rTest;
#define	REP_TEST(re, rep, res)	{re, rep, sizeof(re)-1, sizeof(rep)-1, res}

iTest	tests[] = {
	RE_TEST("", RE_OK),		// Empty RE

	RE_TEST("abcde", RE_OK),	// string
	RE_TEST("abcd€e", RE_OK),	// string with Unicode
	RE_TEST("abc\0def", RE_OK),	// string with NUL byte

	// Positions
	RE_TEST("^abcde$", RE_OK),
	RE_TEST("\\<abc\\>", RE_OK),
	RE_TEST("\\Babc\\b", RE_OK),
	RE_TEST("\\babc\\B", RE_OK),

	// Any
	RE_TEST("abc.def", RE_OK),

	// Repetitions
	RE_TEST("^abcde?$", RE_OK),
	RE_TEST("^abcde??$", RE_OK),
	RE_TEST("a?bcd", RE_OK),		// Rep after 1 char
	RE_TEST("abc€?d", RE_OK),
	RE_TEST("abc*def", RE_OK),
	RE_TEST("abc*?def", RE_OK),		// Non greedy
	RE_TEST("abc*+def", RE_OK),		// Possessive
	RE_TEST("abc+def", RE_OK),
	RE_TEST("abc+?def", RE_OK),		// Non greedy
	RE_TEST("abc++def", RE_OK),		// Possessive
	RE_TEST("abc{2}def", RE_OK),
	RE_TEST("abc{2}?def", RE_OK),		// Non greedy
	RE_TEST("abc{2}+def", RE_OK),		// Possessive
	RE_TEST("abc{,2}def", RE_OK),
	RE_TEST("abc{,2}?def", RE_OK),		// Non greedy
	RE_TEST("abc{,2}+def", RE_OK),		// Possessive
	RE_TEST("abc{2,}def", RE_OK),
	RE_TEST("abc{2,}?def", RE_OK),		// Non greedy
	RE_TEST("abc{2,}+def", RE_OK),		// Possessive
	RE_TEST("abc{,}def", RE_OK),
	RE_TEST("abc{,}?def", RE_OK),		// Non greedy
	RE_TEST("abc{,}+def", RE_OK),		// Possessive
	RE_TEST("abc{2,4}def", RE_OK),
	RE_TEST("abc{2,4}?def", RE_OK),		// Non greedy
	RE_TEST("abc{2,4}+def", RE_OK),		// Possessive
	RE_TEST("€?abc", RE_OK),		// Rep on 1st char that is multibyte

	// Options
	RE_TEST("z(?i)abc(?-i)123", RE_OK),
	RE_TEST("(?i)abc(?-i)123", RE_OK),
	RE_TEST("a(?ies)b", RE_OK),
	RE_TEST("a(?i-es)b", RE_OK),		// Some set, some unset
	RE_TEST("(?-ies)fred", RE_OK),		// Opts all off at very start
	RE_TEST("(?-ies)fr(?i)ed", RE_OK),	// Opts all off at very start but later reset
	RE_TEST("(?i-es)fred", RE_OK),		// Opts some off at very start
	RE_TEST("bill(?-ies)fred", RE_OK),	// Opts off but not at very start
	RE_TEST("a(?)b", RE_OK),		// Empty option
	RE_TEST("abCd", RE_OK),			// Implies exact mode

	// Newline, test that the next character may be given as a hint
	RE_TEST("a\\nb", RE_OK),
	RE_TEST("a\n", RE_OK),
	RE_TEST("a\\n", RE_OK),
	RE_TEST("a\\n.b", RE_OK),
	RE_TEST("a\\n(b)", RE_OK),
	RE_TEST("a\\n+b", RE_OK),
	RE_TEST("a\\n\\\\b", RE_OK),
	RE_TEST("a\\n^b", RE_CIRC_PLACE),
	RE_TEST("a\\n[abc]b", RE_OK),
	RE_TEST("a\\n€b", RE_OK),
	// Duplicate the above but with a literal newline not \n
	RE_TEST("a\nb", RE_OK),
	RE_TEST("a\n", RE_OK),
	RE_TEST("a\n.b", RE_OK),
	RE_TEST("a\n(b)", RE_OK),
	RE_TEST("a\n+b", RE_OK),
	RE_TEST("a\n\\\\b", RE_OK),
	RE_TEST("a\n^b", RE_CIRC_PLACE),
	RE_TEST("a\n[abc]b", RE_OK),
	RE_TEST("a\n€b", RE_OK),

	// Escapes:
	RE_TEST("a:\\\\b:\\tc:\\rd:\\cae\\ad\\ef\\fh", RE_OK),			// \<special-character>
	RE_TEST("a\\.\\?\\+\\*\\[\\]\\^\\$\\\\\\|\\{\\}\\(\\)b", RE_OK),	// All metaCharacters (& a few) escaped
	RE_TEST("a\\c@b", RE_OK),		// Control - NUL
	RE_TEST("a\\cab", RE_OK),
	RE_TEST("a\\cAb", RE_OK),
	RE_TEST("a\\c|b", RE_OK),
	RE_TEST("a\\x22z", RE_OK),
	RE_TEST("a\\x2", RE_INCOMPLETE),
	RE_TEST("a\\x2zq", RE_BAD_CODEPOINT),	// Want 2 hex digits
	RE_TEST("a\\x{20ac}b", RE_OK),		// Euro
	RE_TEST("a\\x{#8364}b", RE_OK),		// Euro - again
	RE_TEST("a\\x{o20254}b", RE_OK),	// Euro - again
	RE_TEST("a\\x{0}b\\c@c", RE_OK),	// NUL 2 different ways
	// There is a difference (implementation) if the escape is not within a STR
	RE_TEST("\\c@b", RE_OK),
	RE_TEST("\\x{20ac}b", RE_OK),
	RE_TEST("\\x{#8364}b", RE_OK),
	RE_TEST("\\x{0}b", RE_OK),

	// Quoting
	RE_TEST("abc\\Q123\\Edef", RE_OK),
	RE_TEST("abc\\Q1+23\\Edef", RE_OK),
	RE_TEST("abc\\Q1(23\\Edef", RE_OK),
	RE_TEST("abc\\Q1\\x{abc}23\\Edef", RE_OK),
	RE_TEST("abc\\Q123\\Edef", RE_OK),
	RE_TEST("abc\\Q123", RE_OK),		// Missing \E (allowed)
	RE_TEST("a(bc\\Q123", RE_INCOMPLETE),	// Missing ) - error
	RE_TEST("abc\\Q123\\", RE_OK),		// Quoted string ends with \ - edge case
	RE_TEST("abc\\Q", RE_OK),		// Empty quoted string
	RE_TEST("abc\\Q\\E123", RE_OK),		// Empty quoted string

	// Comment
	RE_TEST("a(?#comment)b", RE_OK),

	// Groups
	RE_TEST("(abc)", RE_OK),
	RE_TEST("(\\.abc)", RE_OK),
	RE_TEST("(abc)(def)", RE_OK),
	RE_TEST("(abc)(def)(ijk)", RE_OK),
	RE_TEST("F(abc)J(def)K(ijk)L", RE_OK),
	RE_TEST("((a)bc)", RE_OK),
	RE_TEST("((a)(bc))", RE_OK),

	// References
	RE_TEST("(abc)\\1", RE_OK),
	RE_TEST("(abc)(.)(.)(.)(.)(.)(.)(.)(.)(.)(.)(.)(.)(.)\\12x", RE_OK),
	RE_TEST("(abc)\\123x", RE_OK),
	RE_TEST("(abc)\\1(?)23x", RE_OK),		// Group 1 followed by '23x'
	RE_TEST("(abc)\n(def)\\1\\2z", RE_OK),
	RE_TEST("(abc)A(def)\\1\\2z", RE_OK),
	RE_TEST("(abc)(def)\\1\\2z", RE_OK),
	RE_TEST("a()b\\1c", RE_OK),
	RE_TEST("a(b", RE_INCOMPLETE),
	RE_TEST("a)b", RE_NO_GRP_OPEN),
	RE_TEST("ab\\1cd", RE_UNKNOWN_REFERENCE),
	RE_TEST("a(?:bcde)", RE_OK),
	RE_TEST("a(?-i:bcde)f", RE_OK),			// No capture with options
	RE_TEST("a(?:(?-i)bcde)f", RE_OK),		// Same as test above
	RE_TEST("a(bc)(?:de)\\1", RE_OK),
	RE_TEST("a(bc)(?:de)\\1\\2", RE_UNKNOWN_REFERENCE),	// Unknown \2
	RE_TEST("a(bc)(?Xde)\\1", RE_BAD_OPTION),	// Unknown option X

	// Alternate
	RE_TEST("aA|bc", RE_OK),
	RE_TEST("cd|efg|ij", RE_OK),
	RE_TEST("cd|efg|ij|klmn", RE_OK),
	RE_TEST("ab(cd|efg|ij)kl", RE_OK),

	// Alternate groups
	RE_TEST("(ab)|(cd)", RE_OK),
	RE_TEST("(ab)|(cd)|(ef)", RE_OK),
	RE_TEST("(ab)|(cd)|(ef)|(gh)", RE_OK),
	RE_TEST("ST(ab)|(cd)|(ef)|(gh)EN", RE_OK),

	// Branch reset numbering
	RE_TEST("(?|(ab)|(cd))", RE_OK),
	RE_TEST("(?|AB(ab)12|34(cd))", RE_OK),
	RE_TEST("qwe(?|AB(ab)12|34(cd))rty", RE_OK),
	RE_TEST("qwe(?|AB(ab)(12)|34(cd))r(t)y", RE_OK),	// Different alts have different # groups

	// Short character classes
	RE_TEST("\\s", RE_OK),
	RE_TEST("\\S", RE_OK),
	RE_TEST("\\w", RE_OK),
	RE_TEST("\\W", RE_OK),
	RE_TEST("\\d", RE_OK),
	RE_TEST("\\D", RE_OK),

	// Character classes
	RE_TEST("[\\d]", RE_OK),
	RE_TEST("[\\d\\D\\s]", RE_OK),
	RE_TEST("[abc]", RE_OK),
	RE_TEST("[abc\\d]", RE_OK),
	RE_TEST("[abc\\defgh]", RE_OK),
	RE_TEST("[]abc]", RE_OK),
	RE_TEST("[-]abc]", RE_OK),
	RE_TEST("[^abc]", RE_OK),
	RE_TEST("[^]abc]", RE_OK),
	RE_TEST("[^-abc]", RE_OK),
	RE_TEST("[^-]abc]", RE_OK),
	RE_TEST("[abc-]", RE_OK),
	RE_TEST("[ab\\-c\\]d\\re\\tf\\eg\\ag\\\\h\\cai\\x{20ac}j]", RE_OK),
	RE_TEST("[abc\0def]", RE_OK),	// NUL in class
	RE_TEST("[abc\0-a]", RE_OK),	// NUL in range

	// Class ranges
	RE_TEST("[abc\\x{20ac}d]", RE_OK),
	RE_TEST("[a-c]", RE_OK),
	RE_TEST("[0-9]", RE_OK),
	RE_TEST("[9-0]", RE_RANGE_ORDER),	// bad range
	RE_TEST("[0-9a-f]", RE_OK),
	RE_TEST("[0-9a-£]", RE_OK),
	RE_TEST("[0-9£-\\x{20ac}]", RE_OK),
	RE_TEST("[qa-c\\dx]", RE_OK),
	RE_TEST("[^qa-c\\dx]", RE_OK),
	RE_TEST("[+--]", RE_OK),	// The range + to -

	RE_TEST("\\W", RE_OK),
	RE_TEST("\\W\\d", RE_OK),
	RE_TEST("[abc\\W\\ddef]", RE_OK),
	RE_TEST("abc[£-€0-9]def", RE_OK),
	RE_TEST("abc[^£-€0-9zxc\\s\\W]def", RE_OK),

	RE_TEST("abc[:ascii:]def", RE_OK),	// Not recognised as [[:ascii:]]
	RE_TEST("abc[[:ascii:]]def", RE_OK),
	RE_TEST("abc[[:^ascii:]]def", RE_OK),
	RE_TEST("abc[^[:^ascii:]]def", RE_OK),	// The double not means that ascii is expected
	RE_TEST("abc[[:ascii:][:xdigit:]]def", RE_OK),
	RE_TEST("abc[1-3[:ascii:]qwerty]def", RE_OK),
	RE_TEST("abc[^1-3[:ascii:]qwerty]def", RE_OK),


	// Repeats
	RE_TEST("a[0-9]{2,4}x", RE_OK),
	RE_TEST("abc{2,4}x", RE_OK),
	RE_TEST("a\\w{2,4}x", RE_OK),
	RE_TEST("a\\x{20ac}{2,4}x", RE_OK),
	RE_TEST("a(bcd){2,4}x", RE_OK),
	RE_TEST("a{0}", RE_OK),				// Not illegal - but pointless
	RE_TEST("a\\n+b", RE_OK),
	RE_TEST("^?abc", RE_CANNOT_REPEAT),		// Can't repeat ^
	RE_TEST("?abc", RE_CANNOT_REPEAT),		// Can't repeat start of line
	RE_TEST("*abc", RE_CANNOT_REPEAT),		// Can't repeat start of line
	RE_TEST("+abc", RE_CANNOT_REPEAT),		// Can't repeat start of line
	RE_TEST("{1,2}abc", RE_CANNOT_REPEAT),		// Can't repeat start of line
	RE_TEST("(?abc)", RE_BAD_OPTION),		// Bad option a
	RE_TEST("(??abc)", RE_UNKNOWN_GRP_START),	// Unknown (??
	RE_TEST("a???", RE_CANNOT_REPEAT),		// Can't repeat repeat
	RE_TEST("a?{1,2}", RE_CANNOT_REPEAT),		// Can't repeat repeat
	RE_TEST("abc(+def)", RE_NO_REP_AT_START),	// Can't have repeat at start of group
	RE_TEST("abc(*def)", RE_NO_REP_AT_START),	// Can't have repeat at start of group
	RE_TEST("abc|+def", RE_CANNOT_REPEAT),		// Can't repeat |

	// Repetitions & groups
	RE_TEST("a(bcd)+e", RE_OK),
	RE_TEST("a(bcd|(def{4}abc))+e", RE_OK),
	RE_TEST("a(bcd|(de[aeiou]{2,4}abc))+e", RE_OK),
	RE_TEST("a(bcd|(de([aeiou]){2,4}abc))+e", RE_OK),	// Repetition of group within repeated group
	RE_TEST("a(bcd|(de([aeiou]){2,4}a(b)+c))+e", RE_OK),	// Repetition of group within repeated group
	RE_TEST("a(?|bcd|(de([aeiou]){2,4}a(b)+c))+e", RE_OK),	// Repetition of group within repeated group
	RE_TEST("a(?|(de([aeiou]){2,4}a(b)+c)|bcd)+e", RE_OK),	// Repetition of group within repeated group
	RE_TEST("a(?|(de([aeiou]){2,4}a(b)+c)|a(x)+|bcd)+e", RE_OK),	// Repetition of group within repeated group

	// Variables
	RE_TEST("abc$2def", RE_OK),
	RE_TEST("abc$12def", RE_OK),			// Group 12
	RE_TEST("abc$1(?)2def", RE_OK),			// Group 1 followed by '2'
	RE_TEST("abc$1${}2def", RE_OK),			// Last group 1 followed by '2'
	RE_TEST("a(b)c\\1${}2def", RE_OK),		// Group 1 followed by '2'
	RE_TEST("${a_variable}", RE_NOT_IMPLEMENTED),	// Not yet

	// Upper test - should convert to lower case in the compiled RE
	RE_TEST("aBc", RE_OK),
	RE_TEST("abc", RE_OK),
	RE_TEST("Abc", RE_OK),
	RE_TEST("abC", RE_OK),
	RE_TEST("[abc]", RE_OK),
	RE_TEST("[abC]", RE_OK),
	RE_TEST("[Abc]", RE_OK),
	RE_TEST("(?i)Abc", RE_OK),
	RE_TEST("Abc(?i)xYz", RE_OK),
	RE_TEST("Abc((?i)xYz)pQr", RE_OK),
	RE_TEST("=(?i)[Abc]=", RE_OK),
	RE_TEST("=Abc(?i)[xYz]=", RE_OK),
	RE_TEST("=Abc((?i)[xYz])pQr=", RE_OK),
	RE_TEST("=Abc[X-Z]((?i)[xYzG-L](?-i)[JklA-D])[s-v][R-T][a-cG-L]pQr=", RE_OK),
	RE_TEST("=Abc[X-Z]((?i)[xYz][G-L](?-i)[Jkl][A-D])[s-v][R-T][a-cG-L]pQr=", RE_OK),

	// Posix code points
	// Ends up not implemented if all else works
	RE_TEST("ab[\\p{l123}]cd", RE_UNKNOWN_PROPERTY),
	RE_TEST("ab\\p{l321}cd", RE_UNKNOWN_PROPERTY),

	RE_TEST("ab[\\p{l}]cd", RE_NOT_IMPLEMENTED),
	RE_TEST("ab\\p{l}cd", RE_NOT_IMPLEMENTED),
	RE_TEST("ab\\p{Letter}cd", RE_NOT_IMPLEMENTED),
	RE_TEST("ab\\p{LETTER}cd", RE_NOT_IMPLEMENTED),
	RE_TEST("ab\\p{letter}cd", RE_NOT_IMPLEMENTED),
	RE_TEST("ab\\p{LowerCaseLetter}cd", RE_NOT_IMPLEMENTED),
	RE_TEST("ab\\p{Lower_Case_Letter}cd", RE_NOT_IMPLEMENTED),
	RE_TEST("ab\\p{lower_case_letter}cd", RE_NOT_IMPLEMENTED),
	RE_TEST("ab\\P{l}cd", RE_NOT_IMPLEMENTED),

	RE_TEST("ab\\p{^l}cd", RE_NOT_IMPLEMENTED),
	RE_TEST("ab\\P{^l}cd", RE_NOT_IMPLEMENTED),

	RE_TEST("ab[\\p{l", RE_INCOMPLETE),
	RE_TEST("ab\\p{l", RE_INCOMPLETE),

	RE_TEST("ab[\\P{not_a_property}]cd", RE_UNKNOWN_PROPERTY),

	{ NULL, 0, 0}
};

rTest	repTests[] = {
	REP_TEST(".", "abc", RE_OK),
	REP_TEST(".", "a\nc", RE_OK),
	REP_TEST(".", "a\\nc", RE_OK),
	REP_TEST(".", "a\\tc", RE_OK),
	REP_TEST(".", "a\\c@c", RE_OK),
	REP_TEST(".", "a\\cac", RE_OK),
	REP_TEST(".", "a\\ca\\t\\r\\\\A\\aF\\fE\\ec", RE_OK),
	REP_TEST(".", "a\\#c", RE_OK),
	REP_TEST(".", "a\\&c", RE_OK),
	REP_TEST(".", "a\\\\c", RE_OK),
	REP_TEST(".", "a\\c", RE_INCOMPLETE),		// Incomplete as there should be something after \c
	REP_TEST(".", "a\\q", RE_UNKNOWN_ESCAPE),	// should fail
	REP_TEST(".", "a\\x{20ac}c", RE_OK),
	REP_TEST(".", "a&c", RE_OK),
	REP_TEST(".", "a\\0c", RE_OK),
	REP_TEST(".", "a\\1c", RE_UNKNOWN_REFERENCE),	// should fail
	REP_TEST("(.)", "a\\1c", RE_OK),
	REP_TEST("(.)", "a\\12c", RE_UNKNOWN_REFERENCE),	// should fail
	REP_TEST("(.)", "a\\1\\${}c", RE_OK),			// Empty var end group number
	REP_TEST("(.)", "a\\$1c", RE_OK),
	REP_TEST("(.)", "a\\$12c", RE_UNKNOWN_REFERENCE),	// should fail
	REP_TEST("(.)", "a\\$1\\${}2c", RE_OK),			// Empty var end group number

	{ NULL, NULL, 0, 0, 0}
};

/* Bytecode is dumped to the new file and compared with the reference.
 * Flag up differences.
 */
#define	RETEST_NEW	"retest_new"	// Bytecode from this run written here
#define	RETEST_REF	"retest_ref"	// Compared to this - should be the same

/* Error messages that should be looked at all start '!! ',
 * ie these will be unexpected errors.
 */
int	main(int argc, char* argv[])
{
	RegEx	compRe;
	iTest*	res;
	rTest*	rep;
	int	i;
	int	ret;
	int	fails = 0;	// Count of unexpected results
	FILE*	newRes;		// Write new results here
	FILE*	oldRes;		// Read previous results from here
	char*	sbuf = "scan";	// Name of the buffer that contains text to scan
	BUFFER*	sbp;		// Main buffer pointer
	char	resBuf[1024];
	char	resMagic[] = "#retest";

	/* Get the program name, last component of path */
	if((progname = strrchr(argv[0], '/')) != NULL)
		progname++;
	else
		progname = argv[0];

	Utf8Init();
	edinit(mainBname);	/* Buffers, windows */
	varinit();		/* user variables */
	RestInit();		// Get shell name

	/* Open the file where we read previous results (ie reference) to
	 * compare with what we get and the file where we write the results
	 * that we get this time.
	 *
	 * It is not an error if these files cannot be opened, but say so.
	 */
	if((newRes = fopen(RETEST_NEW, "w")) == (FILE*)NULL)
		perror("!! Cannot write to " RETEST_NEW);
	if((oldRes = fopen(RETEST_REF, "r")) == (FILE*)NULL) {
		perror("!! Cannot read from " RETEST_REF);
		fails = 1;		// Deem this an error reported at program end
	}
	if(newRes) {
		fprintf(newRes, "%s Results of microemacs regular expression code tests\n", resMagic);
		fprintf(newRes, "# Do NOT edit this file by hand, it is generated by %s\n", progname);
	}
	if(oldRes) {
		if( !fgets(resBuf, sizeof(resBuf), oldRes)) {
			fprintf(stderr, "!! Cannot read from " RETEST_REF "\n");
			oldRes = (FILE*)NULL;	// Stop further read attempts
			fails = 1;		// Deem this an error reported at program end
		} else {
			// Check signature
			if(strncmp(resBuf, resMagic, strlen(resMagic)) != 0) {
				fprintf(stderr, "!! Bad signature at the start of " RETEST_REF "\n");
				oldRes = (FILE*)NULL;
			} else {
				// Read & ignore the second line of old/previous results
				fgets(resBuf, sizeof(resBuf), oldRes);
			}
		}
	}

	lsetNUL(searchPat_e);

	compRe.re_comp = NULL;
	compRe.re_rcomp = NULL;

	/* Test compiles of REs
	 */
	if(newRes)
		fprintf(newRes, "# Start of tests\n");

	for(res = tests, i = 0; res->it_re; res++, i++) {
		printf("\n\nTest %d re: '%s'\n", i, res->it_re);
		lcpym(searchPat_e, res->it_re, res->it_relen);
		ret = reCompile(searchPat_e, &compRe);
		printf("compile returned %d\n", ret);
		if(ret != res->it_res)
			printf("!! Unexpected result - expected %d\n", res->it_res);

		if(newRes)
			fprintf(newRes, "test %d result %d\n", i, ret);

		if( !reCompareRes(oldRes, "test %d result %d\n", i, ret))
			fails++;

		if(ret == RE_INCOMPLETE)
			printf("warning: re incomplete: %s\n", ttext(playError));
		else if(ret == RE_OK) {
			if(reDebug)	// As debug mode can generate a lot of lines
				printf("RE: '%s'\n", res->it_re);
			rePrint(&compRe);
			rePrintLine(newRes, compRe.re_comp);
			if( !reCompareTest(oldRes, "RE compile", i, compRe.re_comp))
				fails++;
		}
		else
			printf("error: %s\n", ttext(playError));
	}

	printf("\n\n\n*************\n\n\n");

	/* Test compiles of replacements and their REs
	 */
	if(newRes)
		fprintf(newRes, "# Start of rep tests\n");

	for(rep = repTests, i = 0; rep->rt_re; rep++, i++) {
		printf("\n\nRep Test %d re: '%s'  rep: '%s'\n", i, rep->rt_re, rep->rt_rep);

		// Compile the pattern:
		lcpym(searchPat_e, rep->rt_re, rep->rt_relen);
		ret = reCompile(searchPat_e, &compRe);
		printf("RE compile returned %d\n", ret);
		if(newRes)
			fprintf(newRes, "reptest %d result %d\n", i, ret);
		if( !reCompareRes(oldRes, "reptest %d result %d\n", i, ret))
			fails++;
		if(ret != RE_OK) {
			printf("RE compile error, code %d : %s\n", ret, ttext(playError));
			continue;
		}

		rePrintLine(newRes, compRe.re_comp);
		if( !reCompareTest(oldRes, "RE compile", i, compRe.re_comp))
			fails++;

		lcpym(searchRep, rep->rt_rep, rep->rt_replen);

		// Compile the replacement:
		ret = repCompile(searchRep, &compRe);
		printf("Rep compile returned %d\n", ret);
		if(newRes)
			fprintf(newRes, "represult %d result %d\n", i, ret);
		if( !reCompareRes(oldRes, "represult %d result %d\n", i, ret))
			fails++;
		if(ret != rep->rt_res)
			printf("!! Unexpected result - expected %d\n", rep->rt_res);

		if(ret == RE_OK) {
			if(reDebug)	// As debug mode can generate a lot of lines
				printf("re: '%s'  rep: '%s'\n", rep->rt_re, rep->rt_rep);
			repPrint(&compRe);
			rePrintLine(newRes, compRe.re_rcomp);
			if( !reCompareTest(oldRes, "RE compile", i, compRe.re_rcomp))
				fails++;
		} else {
			printf("Error: %s\n", ttext(playError));
		}
	}

	printf("\n\n\n*************\n\n\n");

	/* Test actual running of REs
	 * Check that it is doing the correct thing by looking at $1 (ret_grp)
	 */

	printf("\n\n\n*************\n\n\n");

	// Make & fill the scan buffer
	if( !(sbp = reMkScan(sbuf)))
		return(1);

	if(newRes)
		fprintf(newRes, "# End of reference\n");

	printf("\n\nNumber of unexpected results: %d\n", fails);

	return(fails ? 0 : 1);
}

#endif	// RE_TEST_PROGRAM
