/*
 * ADDW SCCS: @(#)prefix.c	1.29 01/22/20 20:33:17
 */

#include "estruct.h"
#include "utf8.h"
#include "edef.h"

#if	SHOWMODREL
char	prefvers[] = "@(#)prefix.c	1.29 01/22/20 20:33:17";
#endif

/* TODO Unicode
 * There is some '& 0x7f' and similar below.
 */

// All the following are global - otherwise too many function arguments
static	int	sign;			// Sign of aCount: +1 or -1
static	int	firstdigit;		// TRUE if waiting for first decimal digit of count
static	int	state;			// Has values:
#define	NODIGIT		-1		/* no current count */
#define	INDIGIT		0		/* there is a count */
#define	DIGILL		1		/* digits are illegal */
static	int	aFlag;			// Argument flag
static	Cnt	aCount;			// Argument count - always +ve

/* Return the next keyboard key.
 * This is a key that has been prefixed by M- or C-X or C-U
 * Avoid displaying the prefix, but do show it if the user hangs around
 * long enough -- global ddelay seconds.
 *
 * ch		is the current key so far - think the C-X prefix with other flags
 *		maybe ORed in. It is -1 if there is not a value to display.
 * *displayit	is true if ch is to be displayed in the command line
 *
 * After ddelay seconds a read without a prompt times out, the prompt is displayed,
 * *displayit is set true and a non-timeout read is done.
 *
 * This can be used with all of the prefixes. It replaces previous nastiness that
 * used alarm() & signal() to achieve the same thing; which is less desirable
 * these days as tgetc() can do a lot of work (eg re-reading a RENEW mode buffer)
 * during which a receipt of signal might make something barf. Also possible
 * as it is now possible to timeout a terminal read.
 */
KeyCode	getDelayedKey(KeyCode ch, int* displayit)
{
	KeyCode	c;
	long	readDelay;	// Microseconds delay in key read

	while(1) {

		if(ddelay != 0 && !*displayit) {
			readDelay = ddelay * MILLION;
		} else {
			readDelay = 0;			// Now delay reads forever
			CursorInMsg = 1;		// This gets reset in main() before next key read
			disp(ch);			// display char and count
		}

		/* MGC_NO_REFRESH else we cannot have a binding where the 2nd char is C-l or something
		 * It also means that we cannot fix screen corruption on multi key
		 * command - tough!
		 */
		c = tgetc(MGC_MAP | MGC_NO_REFRESH, readDelay);

		/* Got the key.
		 * Unset message line colouring and return the key
		 */
		if(c != KEY_TIMEOUT) {
			msgColourFrom = -1;
			return(c);
		}

		*displayit = 1;
	}
}

/* Get command key handling Meta/^X/^U prefixes
 * This is because keytab[] is bindings with 'strange characters' to create eg:
 * M-A or ^XA or ^X^A. These are all put into one binding table.
 *
 * What ought to happen is that M- or ^X ought to result in a state shift which means
 * that a different lookup table is used. The tables would, generally, be 128 entries big so
 * index directly in - fast.
 * This would also allow tables that modify others as prefix (ie if no found, default to another),
 * eg for command line edit.
 */
KeyCode	getcmd(KeyCode c, int* flag, Cnt* count)
{
	int		again;
	int		disp  = 0;
	PRETAB*		kp;

	// Default values for flag and count:
	aFlag = FALSE;
	aCount = 1;

	// Starting state:
	firstdigit = 1;
	sign = 1;

	// Initially leave the cursor wherever it is
	// It gets put in the message window if a reply is not received quickly enough
	CursorInMsg = 0;

	do {
		again = 0;

		/* Look to see if we have a prefix */
		for(kp = preftab; kp->pt_fun != NULL; kp++)
			if(c == kp->pt_code)
				c = (*kp->pt_fun)(c, &disp, &again);
	} while(again);

	if(disp)
		mlerase();

	*flag = aFlag;			// Argument flag
	*count = aCount * sign;		// Argument count

	return(c);
}

/* Append a representation of the character in "c" to "buf" ready for display.
 *
 * Considerably reduced from what it once was.
 * c will either be Escape, C-x or C-u
 * That is all that it needs to worry about.
 *
 * It is only called from one place, it prob ought to be put in-line there.
 */
void	dispc(BYTE* buf, KeyCode c)
{
	register BYTE*	p = strchr(buf, '\0');

	if(c == CharEscape) {
		*p++ = 'M';
		*p++ = '-';
	} else {
		*p++ = 'C';
		*p++ = '-';
#if	MODFOLD == 0
		*p++ = (BYTE)(c | 0x60);	// Control -> lower case
#else
		*p++ = (BYTE)ChrFromCtrl(c);	// Control -> upper case
#endif
	}

	*p = '\0';
}

/* Display the character in "c" with the count "aCcount" (if "aFlag" is true).
 * The character in "c" may be or'd with one of the prefix keys, ie it could
 * be M-z or something.
 *
 * c == -1 just display the current count
 *
 * There is still someexcessive complication here ... going slowly.
 */
void	disp(KeyCode c)
{
	BYTE		prompt[MAXCOL];

	prompt[0] = '\0';

	msgColourFrom = 0;	// The prompt is coloured

	if(aFlag)
		sprintf(prompt, "%ld ", (long)(sign * aCount));

	if(c != -1)
		dispc(prompt, c);

	/* Should use something faster (less complicated) than mlwrite here.
	 * Set that the highlightened length is the prompt so far.
	 */
	msgColourLen = (int)strlen(prompt);

	mlwrite("%s", prompt);
}

/* Arrive here when the (default) ^U key is pressed. Multiply arguments by
 * 4. Return the first character pressed that isnt a digit or the character
 * which brought us here.
 */
KeyCode	ctlu(KeyCode ch, int* displayit, int* reprocess)
{
	KeyCode 	c   = -1;

	aFlag      = TRUE;		/* we have a count, flag is set   */
	*reprocess = TRUE;		/* want to process char we return */

	// Multiply existing count
	aCount = aCount < (CNT_MAX / 4) ? aCount * 4 : 1;

	for(;;) {
		/* Get the next keystroke if we havent got one already.
		 */
		if(c == -1)
			c = getDelayedKey(ch, displayit);

		if(c == ch) {
			/* Repeat of the character which got us here,
			 * multiply the count by 4 and go around again.
			 *
			 * Try to be intellegent about handling large numbers.
			 */
			aCount = aCount < (CNT_MAX / 4) ? aCount * 4 : 1;

		} else if(c == '-') {
			/* Cope with signs. If positive, go negative, if
			 * negative then return the character.
			 */
			if(sign == 1)
				sign = -1;
			else
				break;

		} else if(Isdigit(c)) {
			/* Handle "^U n" to give "n" (GNU does this)
			 *
			 * This code ONLY works because we keep *count +ve.
			 */
			if(firstdigit) {
				aCount = c - '0';
				firstdigit = 0;
			} else	// Avoid a number that cannot be stored:
				aCount = aCount < (CNT_MAX / 10) ? aCount * 10 + (c - '0') : 1;
		} else
			/* Neither a ^U nor a digit, nor a '-'.
			 */
			break;
		
		c = -1;			/* we want another character */

		if(*displayit)
			disp(-1);	/* update display */
        }

	return(c);
}


/* Process the meta prefix. Return the first character which is not a digit
 * or'd with META.
 * Set *reprocess to 1 if we need to do something again.
 */
KeyCode	metaPrefix(KeyCode ch, int* displayit, int* reprocess)
{
	KeyCode	c = -1;

	state = NODIGIT;

	for(;;) {
		/* Get the next keystroke if we havent got one already.
		 */
		if(c == -1)
			c = getDelayedKey(ch, displayit);

		if(c == ch) {
			/* We've seen the character which got us here. A
			 * possible reason for this is that the user is
			 * entering the sequence, for example, M- 98 M- f
			 * 
			 * If there is a count then ok, if not then return.
			 * We must also stop accepting digits.
			 */
			if(state == INDIGIT)
				break;
			state = DIGILL;

		} else if(c == '-' && state != DIGILL) {
			/* Cope with signs. If positive, go negative, if
			 * negative then return the character.
			 */
			aFlag = TRUE;
			if(sign > 0)
				sign = -1;
			else
				break;
			state = INDIGIT;

		} else if(Isdigit(c) && state != DIGILL) {
			/* Handle digits.
			 */
			aFlag = TRUE;
			if(firstdigit) {
				aCount = c - '0';
				firstdigit = 0;
			} else
				aCount = aCount < (CNT_MAX / 10) ? aCount * 10 + (c - '0') : 1;

			state = INDIGIT;
		} else
			/* Neither a digit nor a '-'.
			 */
			break;
		
		c = -1;			/* we want another character */

		if(*displayit) {
			/* update display */
			if(state == INDIGIT)
				disp(-1);
			else
				disp(ch);
		}
        }

#if	MODFOLD
	/* Beware: c can have all sorts of things ORed in
	 * TODO Unicode - won't work
	 */
	if(Islower(c & 0xff))				/* Force to upper */
		c = toupper(c & 0xff);
#endif

	if(state == INDIGIT) {
		*reprocess = 1;
		return(c);
	}

	return(META | c);
}


/* Process the (default) ^X prefix. Return the first character typed
 * or'd with CTLX.
 */
KeyCode	ctlx(KeyCode ch, int* displayit, int* reprocess)
{
	KeyCode	c;

	c = getDelayedKey(ch, displayit);

	/* Beware: c can have all sorts of things ORed in
	 * TODO Unicode - bad mask
	 */
#if	MODFOLD
	if(Islower(c & 0xff))				/* Force to upper */
		c = toupper(c & 0xff);
#endif

	return(CTLX | c);
}

/* end */
