/*
 * ADDW SCCS: @(#)random.c	1.253 10/14/20 22:49:17
 */

/*
 * This file contains the command processing functions for a number of random
 * commands. There is no functional grouping here, for sure.
 */

#include	"estruct.h"
#include	"utf8.h"
#include	"edef.h"

#if	SHOWMODREL
#include	"emdmap.h"

char	randvers[] = "@(#)random.c	1.253 10/14/20 22:49:17";
#endif

/* Functions to check move in/out of a mode
 * bp - the BUFFER to check/set. This might/might-not be related to curwp.
 *	If it is NOT related, the function MUST NOT try to change what is
 *	displayed in any window.
 * set - true if setting the mode
 * global - true if setting the global flag - else buffer bp
 *
 * The function might set/unset other modes or check or modify buffer lines.
 *
 * Return true if OK to do the action.
 */
static	int (*ModeCheckFuns[])(BUFFER* bp, int set, int global, int f, Cnt n) = {
	NULL,			// WRAP
	c_modeset,		// CMODE
	NULL,			// EXACT
	view_modeset,		// VIEW
	NULL,			// OVER
	NULL,			// MAGIC
	NULL,			// CRYPT
	NULL,			// ASAVE
	indent_modeset,		// INDENT
	renew_modeset,		// RENEW
	NULL,			// NONL
	php_modeset,		// PHP
	utf8_modecheck,		// UTF8
	dos_modeset,		// DOS
	NULL,			// TABEXPAND
};


/* Set fill column to n.
 *
 * If the count is not set the current fill column is shown.
 *
 * The default count is not set. Count must be >= 0.
 *
 * Command:  set-fill-column
 * Bound to: C-x f
 *
 * Return: TRUE
 */
int	setfillcol(int f, Cnt n)
{
	if(f == TRUE)
		fillcol = (int)n;

	mlwrite("%IFill column is %d", fillcol);

	return(TRUE);
}

/* Display the current position of the cursor, in origin 1 X-Y coordinates,
 * the character that is under the cursor (in hex), and the fraction of the
 * text that is before the cursor.
 *
 * Position on the line is shown 3 ways:
 * + disp	Display position, taking the screen to be same width columns.
 * 		Wide characters cause this to be incremented by 2.
 *		Zero width characters display a pink square (U+25A1) and count as width 1.
 *		Tabs count as 1-7 positions.
 * + WP	Notional Word Processor count.
 * 		Zero width characters count as 0.
 *		Tabs count as 1-7 positions.
 *		All others count as 1.
 * + byte	The byte offset from the start of line. Multi byte (Unicode) characters count as 1 to 4
 * The counts all start at 1.
 *
 * The displayed column is not the current column, but the
 * column that would be used on an infinite width display.
 *
 * If the window has hex option, offsets/positions are displayed in hex.
 *
 * The count is ignored.
 *
 * Command:  buffer-position
 * Bound to: C-x =
 *
 * Return: TRUE
 */
/*ARGSUSED*/
int	showcpos(int f, Cnt n)
{
	/* In the descriptions below - be aware of the difference between bytes and characters */
	register LINE*	lp;		/* current line */
	BufChrs		numbytes;	/* # of bytes in buffer */
	BufLns		numlines;	/* # of lines in buffer */
	BufChrs		predbytes = 0;	/* # bytes preceding point */
	BufLns		predlines = 0;	/* # lines preceding point */
	BufLns		linesNar = -1;	/* # lines in narrow */
	BufLns		predNar = 0;	/* # lines pre "." in narrow */
	BufLns		startNar = 1;	/* # lines before narrow */
	int		inNar = 0;	/* In the narrow */
	BYTE*		curchar = NULL;	/* address of character under cursor - if not utf8 */
	BufChrs		predchars;	/* Characters before the point */
	BufChrs		numchars;	/* Characters in the buffer */
	int		ratio;		/* Bytes before/after the point */
	LineLn		savepos;	/* temp save for current offset */
	LineLn		dcol;		/* Display column */
	LineLn		decol;		/* column pos/end of current line */
	LineLn		wpcol;		/* Word processing notion of col */
	LineLn		wpecol;		/* End column */
	char*		chrInfo;	/* Info about current char */
	char*		chrWidth;	/* Info about its width */
	char*		FormatString;	/* To mlwrite() */
	char		printbuf[10];
	char		printbuf2[10];
	char		chrDec[20];	// Character as a decimal
	char		narrowInfo[40];	// If narrow: print of info here

	/* starting at the beginning of the buffer */
	lp = lforw(curbp->b_linep);

	/* start counting bytes, chars and lines */
	numbytes = 0;
	numlines = 0;
	predchars = numchars = 0;

	if(curbp->b_linep == curbp->b_btopline) {
		// Start counting about narrowing
		inNar = 1;
		linesNar = 0;
	}

	while(lp != curbp->b_linep) {

		/* if we are on the current line, record it */
		if(lp == curwp->w_dotp) {
			predlines = numlines;
			predbytes = numbytes + curwp->w_doto;

			if(lisUtf8(lp))
				predchars = numchars + (BufChrs)utf8_chrnlen(ltext(lp), (size_t)curwp->w_doto);
			else
				predchars = numchars + curwp->w_doto;

			if((curwp->w_doto) == lused(lp))
				curchar = NewlineString;
			else
				curchar = lgetcaddr(lp, curwp->w_doto);

			predNar = numlines - startNar + 1;	// How far "." in narrow
		}

		/* on to the next line */
		++numlines;
		numbytes += lused(lp) + 1;
		if(lisUtf8(lp))
			numchars += (BufChrs)utf8_chrnlen(ltext(lp), (size_t)lused(lp)) + 1;
		else
			numchars += lused(lp) + 1;

		// Start a narrowing not at the top buffer line
		if(lp == curbp->b_btopline) {
			startNar = numlines + 1;
			inNar = 1;
		}
		if(lp == curbp->b_abotline)	// End of narrowing
			inNar = 0;

		if(inNar)
			linesNar++;		// One more narrowed line

		lp = lforw(lp);
	}

	/* if at end of file, record it */
	if(curwp->w_dotp == curbp->b_linep) {
		predlines = numlines;
		predbytes = numbytes;
		predchars = numchars;
		curchar = NULL;
	}

	/* Get display & column and end-of-line column. */
	wpcol = getccol(FALSE);
	dcol = getdcol(0) + curwp->w_partoff;
	savepos = curwp->w_doto;
	curwp->w_doto = lused(curwp->w_dotp);
	wpecol = getccol(FALSE);
	decol = getdcol(0);
	curwp->w_doto = savepos;

	ratio = 0;		/* Ratio before dot. */
	if(numbytes != 0)
		ratio = (int)( (100L*predbytes) / numbytes );

	// l_utf8 is a hint that the line might have a UTF8 string on it.
	// It really means that it did at some point in the past.
	// Check that that is still the case (and fix it if not):
	if(lisUtf8(curwp->w_dotp)) {
		utf8_chkbuf(ltext(curwp->w_dotp), (size_t)lused(curwp->w_dotp));
		if(utf8_chars == lused(curwp->w_dotp))
			lunsetUtf8(curwp->w_dotp);
	}

	/* Summarize and report the info.
	 * Don't show char if at EOF, put a \n at line end.
	 * In UTF8 mode give the Unicode point & bytes that make it up.
	 *
	 * Complication: with the HEX option the current character could be inside a UTF-8 sequence.
	 * The routines used below assume that this is not possible.
	 * So: check UTF-8 mode & not line end then temporarily move to the start of the sequence:
	 * UTF8_BACK_FIND_START_NX() does nothing if already at the start.
	 */
	if((curbp->b_mode & MDUTF8) && curwp->w_doto != lused(curwp->w_dotp))
		UTF8_BACK_FIND_START_NX(lgetc(curwp->w_dotp, curwp->w_doto), curwp->w_doto);

	/* Compute as strings current char: as decimal, as hex/U+, width
	 * As strings because some might be the empty string
	 */
	if(curchar == NULL) {
		// End of file - don't show much
		chrDec[0] = '\0';
		chrInfo = "EOF";
		chrWidth = "";
	} else {
		// What gets shown depends on if UTF-8
		UniChar	c;

		if(curbp->b_mode & MDUTF8) {
			c = utf8_decode(curchar);
			chrInfo = utf8_CurrCharInfo();
			sprintf(printbuf2, " W=%d", mk_wcwidth(c));
			chrWidth = printbuf2;
		} else {
			c = *curchar;
			sprintf(printbuf, "0x%2.2x", c);
			chrInfo = printbuf;
			chrWidth = "";
		}
		sprintf(chrDec, "%ld =", (long)c);
	}

	curwp->w_doto = savepos;	// Reset

	/* Put a string in narrowInfo that describes the narrow:
	 * line of "." / lines in narrow @ start of narrow in buffer
	 */
	narrowInfo[0] ='\0';
	if(curbp->b_linep != curbp->b_btopline || curbp->b_linep != curbp->b_abotline)
		sprintf(narrowInfo, ((curwp->w_opts & WO_HEX) ? " <%lx/%lx@%lx>" : " <%ld/%ld@%ld>"),
			(long)(predNar + 1), (long)linesNar, (long)startNar);

	// Display offsets: base 10 or 16 ?
	FormatString = (curwp->w_opts & WO_HEX) ?
		"%sLine %X/%X""%s Col: disp %X/%X WP %X/%x byte %X/%X; Buf: chr %X/%X byte %X/%X %d%% (%s%s%s)" :
		"%sLine %D/%D""%s Col: disp %D/%D WP %D/%D byte %D/%D; Buf: chr %D/%D byte %D/%D %d%% (%s%s%s)";

	mlwrite(FormatString,
		(lisUtf8(curwp->w_dotp) ? "U" : ""),
		(long)(predlines+1), (long)numlines, narrowInfo,
		(long)dcol, (long)decol, (long)wpcol, (long)wpecol,
		curwp->w_doto, lused(curwp->w_dotp),
		predchars, numchars,
		predbytes, numbytes, ratio, chrDec, chrInfo, chrWidth);

	return(TRUE);
}

/* Get the current line number - counting from 1
 *
 * Counting down from w_toplinenum might seem a good idea, but if update() has
 * not happened recently it could be very wrong.
 */
BufLns	getcline(void)
{
	return(getlinenum(curbp, curwp->w_dotp));
}

#if	0
/* Return the number of lines in the current window/buffer
 * This function is not currently used.
 * A better way of doing this is: curbp->b_numLines
 */
BufLns	getNumLines(void)
{
	return(getlinenum(curbp, (LINE*)NULL));
}
#endif

/* Return the line number of a particular line.
 * The first line is numbered 1 (not 0).
 *
 * bp		buffer to get current line from
 * sline	line to search for
 *		This may be NULL - won't match so returns # lines in the buffer.
 *		If this is the line at the end of the buffer (b_linep)
 *		returns the number of buffer lines + 1
 */
LineLn 	getlinenum(const BUFFER* bp, const LINE* sline)
{
	LINE*	lp;		/* current line */
	LineLn	numlines;	/* # of lines before point */

	/* starting at the beginning of the buffer */
	lp = lforw(bp->b_linep);

	/* start counting lines */
	numlines = 1;
	while(1) {
		/* if we are on the current line, record it */
		if(lp == sline)
			break;

		lp = lforw(lp);

		// End of buffer ?
		if(lp == bp->b_linep) {
			/* If we were looking for the virtual line at the end of the
			 * buffer - count one more.
			 */
			if(lp == sline)
				++numlines;
			break;
		}
		++numlines;
	}

	/* and return the resulting count */
	return(numlines);
}

/* Return the number of tabs or spaces at the start of TEXT tp.
 *
 * Note that is bytes - NOT width, ie tabs are not expanded.
 */
LineLn	countWhiteBytes(const TEXT* tp)
{
	LineLn	count = 0;

	while(count < tused(tp) && (tgetch(tp, count) == ' ' || tgetch(tp, count) == '\t'))
		count++;

	return(count);
}

/* Find the position of the first non white character on the current line.
 *
 * This is the WP column (ie counting tabs as 8 -- or whatever).
 * The length, in bytes, of what there is before the word is
 * left in global WordColBytes, set to 0 if nothing.
 *
 * If the line is empty or only contains white space: return 0,
 * aka the position of the \n.
 */
LineLn	GetNonBlankCol(void)
{
	LineLn	NonWCol = 0;		/* 1st non white column */
	WordColBytes = countWhiteBytes(&ltxt(curwp->w_dotp));

	if(WordColBytes)
		NonWCol = getccolAtCol(FALSE, WordColBytes);	// Get position

	return(NonWCol);
}

/* Return the position of the start of first word on the current line.
 *
 * This is the WP column (ie counting tabs as 8 -- or whatever).
 * The length, in bytes, of what there is before the word is
 * left in global WordColBytes, set to 0 if nothing.
 *
 * If joinedPunct then punctuation joined is considered part of the word, ie:
 * any non alphanumeric characters immediately before it are considered part of
 * the word, but spaces are separators.
 *
 * Thus: "hello		- the " is part of the word
 *	 " hello	- the " is NOT part of the word
 *
 * A word character is defined as:
 * anything >= U+80 -- need something better at some stage
 * when: < U+80 Isalnum()
 *
 * If the line is empty or only contains white space: return -1
 * and set WordColBytes to 0 - a safe length to copy!
 *
 * This should, probably, only be used in indent mode.
 */
LineLn	GetWordCol(int joinedPunct)
{
	LineLn	WordCol = -1;
	LineLn	pos = 0;
	int	lastSpace = 1;	// Last char was white, deem line start with one
	LineLn	wstart = -1;	// Start of non word chars joined to the first word. -1 == unknown

	WordColBytes = 0;

	while(pos < lused(curwp->w_dotp)) {
		BYTE	ch = lgetc(curwp->w_dotp, pos);
		// fprintf(stderr, "GWC pos=%ld ch='%c' ls=%d ws=%ld\n", (long)pos, ch, lastSpace, (long)wstart);
		if(ch > UTF8_MAX_CP1 || Isalnum(ch)) {
			// Found start of word
			if(lastSpace || !joinedPunct)
				wstart = pos;

			// What is the column of the start of word ?
			WordCol = getccolAtCol(FALSE, wstart);

			WordColBytes = wstart;	// Note length before the word

			break;
		}

		// We are here because the current character is not a word character.

		lastSpace = (ch == ' ' || ch == '\t');

		// wstart being -1 means that we do not know the position of
		// the last non word non space character before the word
		if(lastSpace) {
			wstart = -1;
		} else {
			// Only note if we don't have a position, ie only note
			// the position of the first in a string of these
			if(wstart == -1)
				wstart = pos;
		}

		pos++;
	}

	return(WordCol);
}

/* Return current display column.
 *
 * Basically count characters, tabs take to next multiple of b_tabsize, everything
 * else is width 1 or 2.
 *
 * bflg	Stop at first non-blank this is TRUE.
 *
 * This is from the start of the line, which might be different from what is displayed
 * if w_leftoffset != 0.
 *
 * Note that not_utf8 is as much a reflection of the mode that the editor
 * is in as if the char itself is multibyte.
 */
LineLn	getccol(int bflg)
{
	LineLn	offset;
	int	save = curwp->w_opts & WO_META_ATTRIB;

	/* Find an offset to work from.
	 * Use doto - unless bflg in which case use the first non white character
	 * but not after doto.
	 */
	offset = bflg ? countWhiteBytes(&ltxt(curwp->w_dotp)) : curwp->w_doto;

	curwp->w_opts |= WO_META_ATTRIB;
	offset = getdiscol(curwp, curwp->w_dotp, offset, DisCol_WP);
	if( !save)
		curwp->w_opts &= ~WO_META_ATTRIB;

	return(offset);
}

/* Return display column for LINE position pos.
 *
 * bflg	Stop at first non-blank this is TRUE.
 *
 * Calls getccol() to do the dirty work.
 */
LineLn	getccolAtCol(int bflg, LineLn pos)
{
	LineLn	save = curwp->w_doto;
	LineLn	col;

	curwp->w_doto= pos;

	col = getccol(bflg);

	curwp->w_doto = save;

	return(col);
}

/* Return current display column - units: L.
 *
 * The difference from getccol() is that control & parity bit set chars count
 * for more than 1 'cos of the way that we display them - iff !AttribMetaCtrl (curwp->w_opts & WO_META_ATTRIB).
 *
 * Note that not_utf8 is as much a reflection of the mode that the editor
 * is in as if the char itself is multibyte.
 *
 * If Disp return Display characters regardless of Hex option.
 */
LineLn	getdcol(int Disp)
{
	return(getdiscol(curwp, curwp->w_dotp, curwp->w_doto,
		curwp->w_opts & WO_HEX ? DisCol_HEX : (curwp->w_opts & WO_META_ATTRIB ? DisCol_MetaChar : DisCol_Char)));
}

/* Set current cursor column to pos.
 *
 * Note that not_utf8 is as much a reflection of the mode that the editor
 * is in as if the char itself is multibyte.
 *
 * Return TRUE if this works.
 * It can fail if the current line is shorter than pos characters.
 */
int	setccol(LineLn pos)
{
	int	CountMode = curwp->w_opts & WO_HEX ? DisCol_HEX : (curwp->w_opts & WO_META_ATTRIB ? DisCol_MetaChar : DisCol_Char);

	/* set us at the new position */
	curwp->w_doto = getdoto(curwp, curwp->w_dotp, pos, CountMode);

	/* and tell whether we made it */
	return(getdiscol(curwp, curwp->w_dotp, curwp->w_doto, CountMode) >= pos);
}

/* cptr must be a pointer into the current line.
 * Return the column number of that character.
 * Return 0 if cptr isn't in the current line -- really an error.
 */
LineLn	GetColAtPtr(const BYTE* cptr)
{
	LineLn	indent;

	/* Sanity check */
	if(cptr < ltext(curwp->w_dotp) ||
	   cptr > lgetcaddr(curwp->w_dotp, lused(curwp->w_dotp)))
		return(0);

	indent = getccolAtCol(FALSE, cptr - ltext(curwp->w_dotp));

	return(indent);
}

/* Twiddle the two characters on either side of dot.
 *
 * If dot is at the end of the line twiddle the two characters before it.
 *
 * Return with an error if dot is at the beginning of line; it seems to be a
 * bit pointless to make this work. This fixes up a very common typo with a
 * single stroke.
 *
 * This always works within a line, so "WFEDIT" is good enough. Leave the
 * cursor stitting at the line end or the (first byte of) second character.
 *
 * This also works swapping nybbles with HEX option.
 *
 * The count is ignored. The default count is unset
 *
 * Command:  transpose-characters
 * Bound to: C-t
 *
 * Return:
 *	TRUE  on successful twiddle
 *	FALSE "." at start of line, cannot twiddle
 *	    Hex	"." on first nybble
 *		"." on end of line 0A
 *		In UTF-8 mode and twiddle would create bad UTF-8 character
 */
int	twiddle(int f, Cnt n)
{
	LINE*	dotp;
	LineLn	doto;	// offset in the line
	BYTE	cl;
	BYTE	cr;
	int	doto_at_end;

	dotp = curwp->w_dotp;
	doto = curwp->w_doto;
	doto_at_end = doto == lused(dotp);

	if(curwp->w_opts & WO_HEX) {
		LineLn	ol;				// Left byte offset
		LineLn	or;				// Right byte offset
		BYTE	ocl;				// Old character left
		BYTE	ocr;				// Old character right
		int	partoff = curwp->w_partoff;	// Operating on 2nd nybble

		// Can't twiddle first nybble on the line:
		if(doto == 0 && !partoff)
			return(FALSE);

		// If on the 2nd nybble of the \n (0A) line end, can't twiddle:
		if(doto_at_end && partoff)
			return(FALSE);

		ol = or = doto;		// Offsets to fiddle with

		// When at end of line, twiddle the previous byte
		if(doto_at_end) {
			partoff = 1;
			ol--;
		}

		// If not twiddle nybbles in one byte, left is previous byte
		if( !partoff)
			ol--;

		ocl = lgetc(dotp, ol);	// Get current values
		ocr = lgetc(dotp, or);

		if(partoff) {
			// Cursor on 2nd nybble in byte
			// Work on the left byte - ignore the right hand one
			lputc(dotp, ol, (BYTE)(((ocl & 0x0f) << 4) | ((ocl & 0xf0) >> 4)));
		} else {
			cl = (BYTE)((ocl & 0xf0) | ((ocr & 0xf0) >> 4));
			cr = (BYTE)((ocr & 0x0f) | ((ocl & 0x0f) << 4));

			lputc(dotp, ol, cl);
			lputc(dotp, or, cr);
		}

		// The easiest way of checking that we have not created a UTF8 monster is
		// to check the whole line.
		if(curbp->b_mode & MDUTF8) {
			if(!utf8_chkbuf(ltext(dotp), (size_t)lused(dotp))) {
				// Check failed - restore the bytes:

				lputc(dotp, ol, ocl);
				lputc(dotp, or, ocr);

				mlwrite("%W""%BLeave UTF-8 mode if you really want that twiddle");

				return(FALSE);
			}

			// Might have created a UTF8 character:
			if(lused(dotp) != utf8_chars)
				lsetUtf8(dotp);

			utf8_ensurePosStart(curbp, 0);	// Ensure all "." & marks are OK
		}

		lchange(WFEDIT);

		return(TRUE);
	}

	if(doto_at_end && --doto < 1)
		return(FALSE);	// Too short

	if(lisUtf8(curwp->w_dotp)) {
		/* Multibyte: neither is, one is, both are.
		 * The nice thing is that we end up with the same length in total.
		 */
		BYTE	tmp[UTF8_MAX_BYTES];	// Somewhere to hold a character
		BYTE*	c_high;			// Start of char to move.
		BYTE*	c_low;
		BYTE*	c_new_low;		// Where low is put
		int	len_low, len_high;

		// Find the start of the high (RH) char, get its length, move it to tmp
		// The find-start is only needed if we moved back a byte earlier (--doto)
		c_high = lgetcaddr(dotp, doto);
		UTF8_BACK_FIND_START(c_high);
		len_high = CHAR_LEN(*c_high);
		UTF8_MOVE(tmp, c_high);

		// Find the start of the low (LH) char, get its length
		c_low = c_high - 1;		// Character before
		if(c_low < ltext(dotp))		// Don't twiddle 1st char on the line
			return(FALSE);

		UTF8_BACK_FIND_START(c_low);
		len_low = CHAR_LEN(*c_low);

		// Determine where the low char moves to & move it
		c_new_low = c_low + len_high;
		UTF8_MOVE_UP(c_new_low, c_low);

		// Move the high char from tmp to where low was
		UTF8_MOVE(c_low, tmp);

		// If the cursor is after the last char on the line, it is still in the right place
		// as we have just shuffled characters around that are before it.
		// If it was on a character and we swapped, we might have swapped with
		// something of a different length (bytes), thus we are no longer sitting
		// at the start of the 2nd character.
		// Adjust "." and marks:
		if(!doto_at_end)
			FixOffsets(curwp->w_dotp, c_low - ltext(dotp), len_high - len_low, curbp, 0);
	} else {
		cr = lgetc(dotp, doto);
		if(--doto < 0)			// Don't twiddle 1st char on the line
			return(FALSE);

		cl = lgetc(dotp, doto);
		lputc(dotp, doto+0, cr);
		lputc(dotp, doto+1, cl);
	}

	lchange(WFEDIT);

	return(TRUE);
}

/* Read the next character & insert literally into the buffer.
 *
 * Or bits with it first - allow meta (0x80) - not in UTF8 mode
 * or if the character > 0xFF.
 * If the count is -ve, make it a control, not if the character > 0xFF.
 * Consider control '?' to be delete. Keyboad mapping is not performed.
 * NB newline splits lines as usual.
 *
 * Take care if we are in overstrike mode, we do a direct replacement - which
 * isn't quite what happens in non quote where each line is considered extended
 * with lots of blanks.
 *
 * act	A description of what is happening, used in prompt.
 *
 * Insert the character absolute(n) times.
 * Default arg should be 1.
 *
 * Nothing special needed for UTF8 mode - linsert() will do the right thing, just check U+0 - U+7f.
 *
 * Return:
 *	TRUE  success
 *	FALSE attempt to insert non character, eg left-arrow
 *	      attempt to make control of something not 0x00 to U+7F or 0xFF
 *	linsert() ldelete()
 */
int	InsSpec(int f, Cnt n, const char* act, int bits)
{
	UniChar	c;
	int	s;
	int	ctrl_it;

	/* Meta characters in UTF8 mode would really upset things.
	 * They would prob be seen as part of a multi byte character.
	 */
	if((curbp->b_mode & MDUTF8) && bits) {
		mlwrite("%EIn UTF-8 mode cannot make a meta character");
		return(FALSE);
	}

	ctrl_it = 0;
	if(n < 0) {
		n = 0 - n;
		ctrl_it = 1;
	}

	/* This results in there always being a count in the prompt, unusual, but ... */
	c = mlGetKey(MGC_MAP | MGC_LEAVEP | MGC_NO_REFRESH | MGC_MACARG, NULL, "%D %s%s : ", (long)n, (ctrl_it ? "Ctrl " : ""), act);	/* Go get it */

	mlerase();

	/* Test for arrow keys & the such:
	 */
	if((KeyCode_u)c & K_CODE_MASK) {
		mlwrite("%ECannot insert that");
		return(FALSE);
	}

	/* Making control: only 0x00 .. 0xFF; in UTF8 mode: 0x00 .. 0x7F
	 */
	if(ctrl_it && ((c & ((curbp->b_mode & MDUTF8) ? UTF8_MAX_CP1 : 0xFF)) != c)) {
		mlwrite("%ECannot make control of '%C'", c);
		return(FALSE);
	}

	/* Non ASCII characters are allowed, want to be able to insert pound, etc.
	 */

	if(n == 0)		/* Do nothing ? */
		return(TRUE);

	if(ctrl_it) {		/* Make control char ? */
		if(c == '?')	/* ^? -> delete */
			c = CharDelete;
		else
			c = ChrToCtrl(c);
	}
	c |= bits;		/* Add meta ? */

	/* Newline is always newline - starts a new line.
	 * Except in invisible buffers - which can contain \n in a line.
	 */
	if(c == '\n' && (curbp->b_flag & BFINVS) == 0) {
		do {
			s = ((curwp->w_bufp->b_mode & MDOVER) ?
				ldelete(1, KL_KILL_APPEND) : 1) && lnewline(0);
		} while(s==TRUE && --n);
		return(s);
	}

	return(((curwp->w_bufp->b_mode & MDOVER) ?
		ldelete(n, KL_KILL_APPEND) : 1) && linsert(n, c));
}

/* Quote the next character, and insert it into the buffer.
 *
 * All the characters are taken literally, with the exception of the newline,
 * which always has its line splitting meaning. The character is always read,
 * even if it is inserted 0 times, for regularity.
 *
 * Insert the character absolute(n) times. A count < 0 makes the character a
 * control.
 * The default count is 1.
 *
 * Command:  quote-character
 * Bound to: C-q   C-x q
 *
 * Return: as InsSpec()
 */
int	quote(int f, Cnt n)
{
	return(InsSpec(f, n, "Quote", 0));
}

/* Read the next character literally & or meta (0x80) to it.
 * If the count is negative, make the character control first.
 *
 * Insert the character absolute(n) times. A count < 0 makes the character a
 * control.
 * The default count is 1.
 *
 * Command:  meta-quote
 * Bound to: C-x C-[
 *
 * Return: as InsSpec()
 */
int	meta_quote(int f, Cnt n)
{
	return(InsSpec(f, n, "Meta quote", 0x80));
}

/* Insert a tab into the file.
 *
 * If the mode TABEXPAND simulate a tab with b_tabsize spaces,
 * if not set, use true tabs (\t).
 * This used to be part of tab().
 * In C mode tidy up the line, see if we can convert multiple spaces to a tab.
 * Only do this up to the first non white space, we don't want to put tabs into
 * strings.
 * In PHP mode - leave leading spaces as spaces - don't convert to tabs.
 *
 * Insert count tabs.
 * Default count 1. Must be >= 0.
 *
 * Command:  insert-tab
 * Bound to: TAB
 *
 * Return TRUE on success. Non true means out of memory.
 */
int     instab(int f, Cnt n)
{
	int	v;
	BYTE*	cp;
	BYTE*	dp;		/* Points to dot */
	BYTE*	rp;		/* Where we are replacing chars */

	/* Spaces for tabs.
	 * Insert at least one to the next tab stop, if more insert
	 * multiples of tab stop.
	 */
	if(curbp->b_mode & MDTABEXP) {
		if((v = linsert(curbp->b_tabsize - getccol(FALSE) % curbp->b_tabsize, ' ')) ==
							TRUE && n > 1)
			v = linsert(--n * curbp->b_tabsize, ' ');

		return(v);
	}

	/* Insert true tabs */
	if((v = linsert((BufChrs)n, '\t')) != TRUE || (curbp->b_mode & MDCMOD) == 0)
		return(v);

	/* C mode tabs.
	 * Try to tidy up the line a bit.
	 * See if we can replace some spaces by tabs.
	 * Since we have just inserted a tab, we are looking
	 * for spaces followed by a tab.
	 * Sometimes we just chuck the space, sometimes replace
	 * by a tab.
	 * Be conservative, don't do this if there are things other than
	 * space/tab before the point.
	 *
	 * Also do it if the only thing before "." is a comment continuation (cc).
	 * An irritation was where "." was at the 'A' on this line and
	 * TAB was pressed that would result in '\t * \t', change to '\t *\t'.
	 * Only if there was only \w* before the cc, do not tidy before the cc.
	 * Do this:
	 *	(\w*)(\*|\*\*|//)\w+ -> \1\2\t
	 */
	cp = ltext(curwp->w_dotp);
	rp = cp;	// Start replacing tabs from here
	dp = cp + curwp->w_doto;

	// Look before ".", is it just white space and maybe a cc ?
	for(; cp < dp; cp++)
		if(*cp != ' ' && *cp != '\t') {
			// Been here before ? Ie already seen one cc ?
			if(rp != ltext(curwp->w_dotp))
				return(TRUE);

			// Comment continuation is '*' or '**' or '//'
			// Cannot be a comment ?
			if(*cp != '*' && *cp != '/')
				return(TRUE);

			// cp[1] is valid as cp < dp
			if(cp[0] == '/') {
				if(cp[1] != '/')
					return(v);	// Not '//'
				cp += 2;
				rp = cp;
				continue;
			}

			// Have one '*'. Is there a second one ?
			rp = ++cp;
			if(cp[0] == '*')
				rp = ++cp;
		}

	/* Find which column "." (dp) is at (ie after the just inserted TAB).
	 * Remove spaces & TABs between dp & rp.
	 * Insert TABs until we reach that column.
	 *
	 * We will be deleting the TAB just inserted, which might be the only
	 * thing that we delete and is promptly replaced. Don't bother to detect
	 * this case.
	 */
	LineLn	dotCol = getccol(FALSE);
	BufChrs	chrsToDel = (BufChrs)(dp - rp);

	// Move to rp so that ldelete() can work
	curwp->w_doto -= chrsToDel;
	ldelete(chrsToDel, KL_DISCARD);

	while(getccol(FALSE) < dotCol)
		if(linsert(1, '\t') == FALSE)
			return(FALSE);

	return(TRUE);
}

/* Check if the cursor is not in the middle of a UTF-8 character sequence.
 * The point is that we don't want to insert something else in the middle one one.
 *
 * It is sometimes legitimate for "." to be in the middle of a UTF-8 character,
 * think: HEX option.
 *
 * Return TRUE if OK - ie "." is not in a bad place
 *	  FALSE "." is in the middle of a UTF-8 character
 */
int	CheckNotInUtf8(void)
{
	if(lisUtf8(curwp->w_dotp) && curwp->w_doto != lused(curwp->w_dotp) &&
			BYTE_IS_MB_CONT(lgetc(curwp->w_dotp, curwp->w_doto))) {
		mlwrite("%ECursor is inside a multi-byte character");
		return(FALSE);
	}

	return(TRUE);
}


/* Insert Unicode character.
 *
 * This will not be called in view mode.
 *
 * Insert count characters.
 * Default count 1. Must be >= 0.
 *
 * Command:  insert-unicode
 * Bound to: C-x u
 *
 * Return:
 *	TRUE  it worked
 *	FALSE "." not somewhere where a character can be inserted (eg Hex mode sometimes)
 *	      Value entered is not a good UTF-8 character
 *	      Nothing available at prompt
 *	insChar(), instab(), newline()
 *	ABORT C-g
 */
int	insUnicode(int f, Cnt n)
{
	UniChar	uc;	// Unicode char
	int	utf8 = curbp->b_mode & MDUTF8;
	char	gash;	// Ensure nothing follows Unicode #
	char*	fmt;	// Convertion format
	char*	fmtName;// Format name
	int	ret;	// Return value
	VAL*	u8code;	// UTF-8 code - entered by user

	// Check that cursor is somewhere valid
	if( !CheckNotInUtf8())
		return(FALSE);

	u8code = valGet(V_STR);

	if((ret = mlreply((utf8 ?
			    "Enter Unicode as U+hex, U+#decimal or U+ooctal  U+" :
			    "Max: FF as non UTF-8 buffer. U+"),
			   &u8code, MLINTEGER | MLNOSPACE, NULL)) != TRUE)
		goto off;

	// Get appropriate decimal/hex/octal format code depending on first letter being '#' or 'o'
	fmt = "%" UNICHAR_LMOD "x%c";
	fmtName = "hex";
	if(vtext(u8code)[0] == '#') {
		fmt = "#%" UNICHAR_LMOD "d%c";
		fmtName = "decimal";
	} else if(vtext(u8code)[0] == 'o') {
		fmt = "o%" UNICHAR_LMOD "o%c";
		fmtName = "octal";
	}

	if(sscanf(vtext(u8code), fmt, &uc, &gash) != 1) {
		mlwrite("%WCould not interpret as Unicode character (%s) 'U+%v'", fmtName, u8code);
		ret = FALSE;
		goto off;
	}

	if( !utf8 && uc > UCHAR_MAX) {
		mlwrite("%WU+%X is too large for non UTF-8 buffer", uc);
		ret = FALSE;
		goto off;
	}

	/* Process a couple of specials,
	 * but not in invisible buffers - tabs & newlines inserted literally.
	 */
	if((curbp->b_flag & BFINVS) == 0)
		switch(uc) {
		case '\t':
			return instab(f, n);
		case '\n':
			return newline(f, n);
		}

	// Is this legal ?
	if( !utf8_isvalid(uc)) {
		mlwrite("%EError in U+%v: %s", u8code, utf8_perror());
		ret = FALSE;
		goto off;
	}

	mlwrite("%IU+%X = '%C'", (long)uc, uc);

	ret = insChar(uc, n);		// Default count is 1

off:	valRelease(u8code);

	return(ret);
}

/* Set the global (ie default) tab width to the count.
 *
 * If not given a count or the count is zero, set the tabsize setting (ie reset
 * to default 8 with MDTABEXP switched on) and also the tabsize for the current
 * buffer.
 *
 * If the count is zero set the global values, not that of the current buffer.
 *
 * If given a +ve argument, set the tabsize to that.
 *
 * Part of the functionality has been split off into instab().
 *
 * The default count is not set. Must be >= 0
 *
 * Command:  handle-tab
 * Bound to: unbound
 *
 * Return:
 *	TRUE  it worked
 *	FALSE value entered greater than TAB_MAXIMUM
 */
int	tab(int f, Cnt n)
{
	/* Count of 0 means reset to default */
	if(f == FALSE || n == 0) {
		Tabsize = DEF_TAB;
		gmode &= ~MDTABEXP;

		if(f == FALSE) {
			curbp->b_tabsize = DEF_TAB;
			curbp->b_mode &= ~MDTABEXP;
		}

		return(TRUE);
	}

	if(n > TAB_MAXIMUM) {
		mlwrite("%E""%BTab size cannot be more than %d", TAB_MAXIMUM);

		return(FALSE);
	}

	/* n >= 0 */
	Tabsize = (short)n;
	gmode |= MDTABEXP;

	return(TRUE);
}

#if	AEDIT
/* Insert tabs in the current region.
 *
 * The count is ignored.
 *
 * Command:  entab-region
 * Bound to: unbound
 *
 * Return TRUE if OK.
 */
int	entabRegion(int f, Cnt n)
{
	return(tabMangleRegion(TRUE));
}

/* Remove tabs in the current region.
 *
 * The count is ignored.
 *
 * Command:  detab-region
 * Bound to: unbound
 *
 * Return TRUE if OK.
 */
int	detabRegion(int f, Cnt n)
{
	return(tabMangleRegion(FALSE));
}

/* Insert/remove tabs entab()/detab() the current region.
 * This does the work of entabRegion()/detabRegion()
 *
 * insTab - TRUE if to insert tabs (entab())
 *
 * "." and mark are left unmoved.
 *
 * Return TRUE if OK.
 */
int	tabMangleRegion(int insTab)
{
	int	i;
	REGION	reg;

	if((i = getregion(&reg)) != TRUE)
		return(i);

	pnmark0();		// So that we can put "." back where it started
	curwp->w_dotp = reg.r_linep;
	curwp->w_doto = reg.r_offset;

	i = tabMangle(insTab, reg.r_lines);

	retmark0(FALSE);

	return(i);
}

/* Change tabs to spaces for count lines.
 * 
 * Current line, if count that many lines, -ve means going up.
 *
 * Default count is 1.
 *
 * Command:  detab-line
 * Bound to: C-x C-d
 *
 * Return TRUE if OK.
 */
int	detab(int f, Cnt n)
{
	return tabMangle(FALSE, n);
}

/* Change spaces to tabs where posible.
 * 
 * If there is white space at the end of a line, leave it there (maybe entabbing it).
 *
 * Current line, if argument that many lines, -ve means going up.
 *
 * Will not insert a tab if it only replaces one space, unless there was a tab there already.
 * White space at end of line will be left - but entabbed.
 *
 * Almost complete rewrite for UTF8 - it was occasionally buggy anyway.
 *
 * Default count is 1.
 *
 * Command:  entab-line
 * Bound to: C-x C-e
 *
 * Return TRUE if OK.
 */
int	entab(int f, Cnt n)
{
	return tabMangle(TRUE, n);
}

/* This does the real work of entab() and detab().
 *
 * This works by removing tabs & spaces and then putting them back in.
 * detab works by only inserting spaces.
 *
 * insTab - TRUE if to insert tabs (entab())
 *
 * Count is the number of lines, -ve means going up, zero just do the current line and
 * leave "." on that line.
 * 
 * BUG: when the number of adjacent space/tab exceeds what can be store in spaceTab (int).
 *
 * Return TRUE if OK.
 */
int	tabMangle(int insTab, Cnt n)
{
	int	inc;				// increment to next line [sgn(n)]
	int	didSomething = 0;		// True if this changed the buffer
	int	ret = FALSE;			// Return value

	/* loop thru entabbing n lines */
	inc = ((n > 0) ? 1 : -1);
	while(1) {
		LINE*	lp = curwp->w_dotp;	// Current line
		int	width = 0;		// Width of current char - init to keep compiler happy
		int	spaceTab = 0;		// # space & tab chars
		int	WasTab = 0;		// Ate a tab - so replace it
		LineLn	ccol;			// Current cursor column
		UniChar	cchar;			// Current character
		UniChar	pchar = 0;		// Previous char (only value of interest is ' ')
		int	wantChange = 0;		// A change is needed ? Avoid doing work that results in no net change

		curwp->w_doto = 0;		/* start at the beginning */
		ccol = 0;			// ditto
		LineLn	scol = 0;		// position of last non white space

		// Beware diplay column is not same as byte position
		do {
			// Get the current character or '\n' at EOL
			// Try to avoid UTF8 work if we can
			// Compute the width of the character.
			if(curwp->w_doto < lused(lp)) {
				cchar = lgetc(lp, curwp->w_doto);

				if( !lisUtf8(lp) || CHAR_LEN(cchar) <= 1) {
					width = (cchar >= ' ' && cchar <= '~') || cchar >= 0x80 ? 1 : 0;
				} else {
					cchar = utf8_decode(lgetcaddr(lp, curwp->w_doto));
					width = mk_wcwidth(cchar);
					if(width < 0)
						width = 0;
				}

			} else {
				// Use this to flush out any space/tabs
				cchar = '\n';
			}

			switch(cchar) {
			case '\t':
				// Eat space/tab, just note how much room they take (change ccol)
				// and how many characters this sequence is.
				ccol = (ccol / curbp->b_tabsize + 1) * curbp->b_tabsize;
				spaceTab++;
				WasTab = 1;
				/* If removing tabs - wantChange
				 * If inserting tabs and we find a space before a tab - replace by tab-tab
				 */
				if( !insTab || pchar == ' ')
					wantChange = didSomething = 1;
				moveForw(1);
				break;

			case ' ':
				spaceTab++;
				ccol++;
				/* Only possible change if inserting tabs.
				 * Then only if this space is at a tab stop and preceeded by a space,
				 * ie don't replace single spaces.
				 */
				if(insTab && pchar == ' ' && (ccol % curbp->b_tabsize) == 0)
					wantChange = didSomething = 1;
				moveForw(1);
				break;

			default:
				// A non space/tab. Check if there have been any space/tab, if so
				// delete them and insert the proper combination of tabs & spaces.
				// Leave a possible tab as a space if it does no more than a space,
				// because we do not want to insert tabs between words in a sentence/code;
				// this is one of these things that can always be shown to be wrong.
				if(wantChange) {
					moveBack(spaceTab);

					if( !ldelete(spaceTab, KL_DISCARD))
						goto off;

					LineLn	t_ccol = ccol / curbp->b_tabsize;	// tabs before current col
					LineLn	t_scol = scol / curbp->b_tabsize;	// tabs to last non white char

					// Inserting tabs and will a tab take us closer to ccol ?
					if(insTab && t_ccol != t_scol &&
					    // Tab to be at least one or replacing one
					    (WasTab || t_ccol * curbp->b_tabsize > scol + 1)) {
						// Can use tabs to move from scol
						if( !linsert(t_ccol - t_scol, '\t'))
							goto off;
						scol = t_ccol * curbp->b_tabsize;
					}

					// Use spaces to go the rest of the way from scol
					if(ccol > scol && !linsert(ccol - scol, ' '))
						goto off;
				}

				// Reached end of line, entabbed trailing space
				// Go grab next line
				if(cchar == '\n')
					break;

				WasTab = 0;
				spaceTab = 0;
				wantChange = 0;
				// Ignore complaint about uninitialised width
				scol = (ccol += width);
				if(moveForw(1) == FALSE)
					goto off;
			}

			pchar = cchar;

		} while(cchar != '\n');

		/* Done all the lines that we need to ?
		 * The text for 0 is twice as if the count is zero
		 * one line is changed and "." remains on that line.
		 */
		if(n == 0)
			break;

		/* Advance/or back to the next line.
		 * Stop if can't move.
		 */
		if( !signedForwline(inc))
			break;

		n -= inc;
		if(n == 0)
			break;
	}

	ret = TRUE;

off:	// Come here on failure

	curwp->w_doto = 0;	/* to the begining of the line */
	thisflag &= ~CFCPCN;	/* flag that this resets the goal column */
	if(didSomething)
		lchange(WFEDIT);/* yes, we have made at least an edit */

	return(ret);
}

/* Set so that we are at column 'target', add tabs/spaces or zap them.
 * If PHPMODE don't insert tabs.
 *
 * Return TRUE if OK
 */
int	SetToCol(LineLn target)
{
	while(target != getccol(FALSE)) {
		if(target < getccol(FALSE))	/* on doit detruire des caracteres */
			while(getccol(FALSE) > target) {
				if(moveBack(1) == FALSE ||
				   ldelete(1, KL_DISCARD) == FALSE)
				return(FALSE);
			}
		else {				/* on doit en inserer */
			/* Insert tabs as far as we can */
			if( ! (curbp->b_mode & MDPHP))
				while(target / curbp->b_tabsize * curbp->b_tabsize > getccol(FALSE))
					if( !linsert(1, '\t'))
						return(FALSE);
			if( !linsert(target - getccol(FALSE), ' '))
				return(FALSE);
		}
	}

	return(TRUE);
}

/* Throw away trailing white space on line.
 *
 * Don't put the white space into a yank buffer.
 *
 * The line might not have association with any particular buffer,
 * thus lchange() is not called.
 */
void	ZapTrailingSpaces(LINE* line)
{
	LineLn	len;
	BYTE*	lbuf;
	BYTE	ch;
	Window*	wp;

	len = lused(line);
	lbuf = lgetcaddr(line, len - 1);

	while(len > 0) {
		if((ch = *lbuf--) != ' ' && ch != '\t')
			break;
		len--;
	}
	lsetused(line, len);

	/* Find and patch up all windows where we have left the mark or point
	 * out in no man's land - ie zapped the character.
	 */
	for(wp = wmessp; wp != (Window*)NULL; wp = wp->w_wndp) {
		if(wp->w_dotp == line && wp->w_doto > len)
			wp->w_doto = len;

		if(wp->w_markp == line && wp->w_marko > len)
			wp->w_marko = len;
	}

	// Alpha marked line ?
	// Adjust the line offset if needs be:
	if(lmarked(line)) {
		AMARK*	amp;
		for(amp = curbp->b_amark; amp; amp = amp->m_next)
			if(amp->m_pline == line && amp->m_oline >= len)
				amp->m_oline = len;
	}

	// Ditto numeric marked
	if(lisNmarked(line)) {
		NMARK*	nmp;

		for(wp = wmessp; wp != NULL; wp = wp->w_wndp) {
			if(wp->w_bufp != curbp)
				continue;

			for(nmp = wp->w_nmarks; nmp < wp->w_nmarks + NUM_NUM_MARKS; nmp++)
				if(nmp->n_pline == line && nmp->n_oline >= len)
					nmp->n_oline = len;
		}
	}
}

/* Trim trailing whitespace for the next n lines.
 *
 * The count is ignored.
 *
 * Command:  trim-region
 * Bound to: unbound
 *
 * Return TRUE if OK
 */
int	trimRegion(int f, Cnt n)
{
	int	i;
	REGION	reg;

	if((i = getregion(&reg)) != TRUE)
		return(i);

	pnmark0();		// So that we can put "." back where it started
	curwp->w_dotp = reg.r_linep;
	curwp->w_doto = reg.r_offset;

	i = trim(TRUE, reg.r_lines);

	retmark0(FALSE);

	return(i);
}

/* Trim trailing whitespace from the point to eol.
 *
 * Only mark the buffer as changed if we actually do something.
 *
 * It is tempting to just reset the line length (as one version of this code did),
 * the trouble with that is that it could leave a mark beyond the
 * end of line. ldelete() checks that.
 *
 * Deleted white space is not put into the Kill Buffer.
 *
 * Continue for count lines with "." set to the start of line.
 * Negative count: move up the buffer.
 *
 * Default argument is 1.
 *
 * Command:  trim-line
 * Bound to: C-x C-t
 *
 * Return TRUE if OK
 */
int	trim(int f, Cnt n)
{
	LINE*	lp;	/* current line pointer */
	LineLn	offset;	/* original line offset position */
	LineLn	length;	/* current length */
	int	inc;	/* increment to next line [sgn(n)] */

	/* loop thru trimming n lines */
	inc = ((n > 0) ? 1 : -1);

	while(n) {
		lp = curwp->w_dotp;		/* find current line text */
		offset = curwp->w_doto;		/* save original offset */
		length = lused(lp);		/* find current length */

		/* trim the current line */
		while(length > offset) {
			if(lgetc(lp, length-1) != ' ' &&
			   lgetc(lp, length-1) != '\t')
				break;
			length--;
		}

		if(length != lused(lp)) {
			curwp->w_doto = length;
			ldelete(lused(lp) - length, KL_DISCARD);
		}

		curwp->w_doto = 0;	// Start at the start of subsequent lines

		/* Advance/or back to the next line.
		 * Stop if can't move.
		 */
		if( !signedForwline(inc))
			break;
		n -= inc;
	}

	thisflag &= ~CFCPCN;	/* flag that this resets the goal column */

	return(TRUE);
}
#endif	// AEDIT

/* Open up some blank space. The basic plan is to insert a bunch of newlines,
 *
 * and then back up over them. Everything is done by the subcommand
 * procerssors. They even handle the looping.
 *
 * Default count 1. Must be >= 0.
 *
 * Command:  open-line
 * Bound to: C-o
 *
 * Return:
 *	TRUE  on success
 *	lnewline() moveBack()
 */
int	openline(int f, Cnt n)
{
	register Cnt	i;

	i = n;					/* Insert newlines. */

	while(--i >= 0)
		if(lnewline(1) == FALSE)
			return(FALSE);

	return(moveBack(n));		/* Then back up overtop of them all. */
}

/* Insert a newline.
 *
 * If we are in CMODE or PHPMODE, do automatic indentation as specified.
 *
 * WRAP mode: either trim or wrap as appropriate.
 *
 * Default arg is 1. Must be >= 0
 *
 * Command:  newline
 * Bound to: RET
 *
 * Return:
 *	TRUE  success
 *	cinsert() newlineAndIndent() lnewline()
 */
int	newline(int f, Cnt n)
{
	int	s;		// Status

	/* If we are in I mode and this is a default <NL> */
	if(curbp->b_mode & MDINDEN)
		return(newlineAndIndent(f, n));	/* Treat ^M as ^J */

	/* C mode ? */
	if(curbp->b_mode & (MDCMOD | MDPHP)) {
		for(; --n >= 0;) {
			if((s = cinsert()) != TRUE)
				return(s);

			/* That will have trimmed the line,
			 * so no trailing white space on either
			 * the old or new (current) lines.
			 */
		}

		return(TRUE);
	}

	// A couple of things to check/do in WRAP mode:
	if((curwp->w_bufp->b_mode & MDWRAP)) {
		if(fillcol > 0 && getccol(FALSE) > fillcol) {
			/* If a fill column is defined and we are past the fill
			 * column, perform word wrap
			 */
			FuncExec(BindWrapword, FALSE, 1);
		} else {
			// Trim SPC/TAB before the NL being inserted
			while(curwp->w_doto > 0 &&
			      (lgetc(curwp->w_dotp, curwp->w_doto - 1) == ' ' || lgetc(curwp->w_dotp, curwp->w_doto - 1) == '\t')) {
				curwp->w_doto--;
				ldelete(1, KL_DISCARD);
			      }
		}
	}

	/* Insert some lines.
	 * Force newline if this is at start of line.
	 */
	while(--n >= 0)
		if((s = lnewline(curwp->w_doto == 0)) != TRUE)
			return(s);

	return(TRUE);
}

/* Backspace 'till we either reach the start of the line, or we reach a
 * non white space character.
 *
 * This might leave w_doto being negative - thus invalid.
 */
void	SkipBackOverSpace(void)
{
	register int	ch;	/* last character before input */

	while(curwp->w_doto > 0 &&
		((ch = lgetc(curwp->w_dotp, curwp->w_doto)) == ' ' ||
							   ch == '\t'))
		curwp->w_doto--;
}

/* Return TRUE if the word in str is found with the last character at dot.
 * Return FALSE if it isn't.
 * Check that there isn't an alphanumeric immediately before it.
 */
int	BackWordMatch(const char* str)
{
	LineLn	len;

	len = (LineLn)strlen(str);

	if(curwp->w_doto < len - 1)
		return(FALSE);

	/* Check the char before -- if there is one */
	if(curwp->w_doto >= len &&
			Isalnum(lgetc(curwp->w_dotp, curwp->w_doto - len)))
		return(FALSE);

	return(strncmp(str, lgetcaddr(curwp->w_dotp, curwp->w_doto - (len - 1)),
							(size_t)len) == 0);
}

/* If we find 'if', 'for', 'while', 'foreach', 'until', 'unless', 'elseif', 'elsif', ...
 * before the matching R paren of a L paren that we *may* be on,
 * return the start column (indent) of that keyword.
 * If we find an 'else' return the indent of that keyword.
 * If we don't find any of these return -1.
 *
 * If a keyword is found, set global CKeyBytes to where it starts in the line.
 *
 * Return value is screen (word processor) position, CKeyBytes is byte in LINE.
 *
 * The current position if left unchanged.
 *
 * Interesting is where to start looking in the current line.
 *
 * More keywords than in C, also influenced by perl.
 */
LineLn	CKeyMatch(LineLn Interesting)
{
	// Keywords that we recognise:
	static	const char* const CKeyWords[] = {
			"elseif",
			"elsif",
			"for",
			"foreach",
			"if",
			"unless",
			"until",
			"while",
			(char*)NULL
		};

	// A different set of keywords that we recognise.
	// Ones that do not have (xx) after them:
	static	const char* const CKeyWords2[] = {
			"else",
			"do",
			(char*)NULL
		};

	LINE*	oldlp;
	LineLn	oldoff;
	LineLn	target;	/* Column that we end up at - match to */

	/* Note where we are */
	oldlp = curwp->w_dotp;
	oldoff = curwp->w_doto;

	target = -1;		/* Error return */

	curwp->w_doto = Interesting;

	if(curwp->w_doto < 0)
		curwp->w_doto = 0;

	// Might be at end of line:
	if(curwp->w_doto >= lused(curwp->w_dotp))
		curwp->w_doto = lused(curwp->w_dotp) - 1;

	SkipBackOverSpace();

	/* If there is something left on the line - see what */
	if(curwp->w_doto > 0) {
		const char* const* wordp = NULL;

		if(lgetc(curwp->w_dotp, curwp->w_doto) != ')') {
			wordp = CKeyWords2;
		} else {
			/* Found ')', find the match & see if it of interest */
			if(fmatch(')', 0)) {
				curwp->w_doto--;
				SkipBackOverSpace();
				wordp = CKeyWords;
			}
		}

		if(wordp) {	// If we have a list of words to scan for
			// Scan language keywords
			for( ; *wordp; wordp++)
				if(BackWordMatch(*wordp)) {
					// Length of the current word minus 1
					int	lenm1 = (int)strlen(*wordp) - 1;
					target = getccol(FALSE) - lenm1;
					CKeyBytes = curwp->w_doto - lenm1;
					break;
				}
		}
	}

	/* fprintf(stderr, "CKeyMatch -> %ld\n", (long)target); */

	/* Return to where we started out from */
	curwp->w_dotp = oldlp;
	curwp->w_doto = oldoff;

	return(target);
}

/* Insert a newline and indentation for C/PHP mode.
 *
 * This function is a morass of ad-hockery - not pretty.
 *
 * Return TRUE if OK.
 */
int	cinsert(void)
{
	BYTE*	cptr;		/* string pointer into text to copy */
	BYTE*	Intp;		// Point to interesting character
	int	bracef;		/* was there a brace at the end of line? */
	int	in_comment;	/* 1=In a comment part of a line,
				 * 2=line starts with unterminated comment */
	int	comment_starts;	/* Comment starts the line */
	int	was_in_comment;	/* At the start of the line */
	int	double_star;	/* Comment line starts '** ' */
	int	in_str;		/* Within a string */
	int	in_white;	/* In white space at the start of line */
	BYTE	last_ch;	// Last char while scanning
	BYTE*	dotp;		/* Position of '.' in the buffer */
	LineLn	indent;		/* How far indented */
	LineLn	indent2;	/* How far indented in comment */
	int	short_cut;	// See below for comments on this
	LineLn	j;
	LineLn	cindent;	/* Comment indent */
	int 	SeenNonWhiteInComment = 0;
	int	VirtualSpace;	/* Virtual space at end of line */
	LineLn	wantIndent;	// Position of first non-white on the line - WP columns, not bytes
	LineLn	keyWordIndent;	// Indent due to keyword
	int	insCpp = 0;	// Insert a C++ comment needed
	BYTE	quote = 0;	// The character that delimits the current string

	/* Find out how far we are indented */
	cindent = indent = getccol(TRUE);
	indent2 = getccol(FALSE);

	/* Get the current column so that we can return to it later.
	 * Work out if the cursor is in white space at the start of line
	 * before non white space.
	 */
	wantIndent = GetNonBlankCol();
	in_white = curwp->w_doto < WordColBytes;

	cptr = ltext(curwp->w_dotp);			// Start of line
	dotp = lgetcaddr(curwp->w_dotp, curwp->w_doto);	// Address of "."

	/* Quick test: if the cursor is on the '/' of a C style or C++ style
	 * comment start insert a newline and indent to the column that we were
	 * at.
	 */
	if(curwp->w_doto + 2 <= lused(curwp->w_dotp) &&
	   dotp[0] == '/' && (dotp[1] == '*' || dotp[1] == '/')) {
		if( !lnewline(1))
			return(FALSE);

		ZapTrailingSpaces(lback(curwp->w_dotp));

		return(SetToCol(indent2));
	}

	/* Make cptr point to the first non white space on the line or
	 * where the cursor is (if this line in blank).
	 * Try & deduce if we are in a comment, we decide this if the line
	 * starts '/\* ' or '/\*\t' or ' * ' or ' *\t' or ' *\n', these
	 * being what will be found in a comment block.
	 * Make a reasonable guess.
	 */
	while(cptr < dotp && (*cptr == ' ' || *cptr == '\t'))
		cptr++;


	in_comment = 0;
	was_in_comment = 0;
	comment_starts = 0;
	double_star = 0;

	/* If the line starts '* ' or '** ' followed by white then last is '/'
	 * we auto end the comment. short_cut records state:
	 * 0 - last seen is not lead in for auto comment end
	 * 1 - last seen is '*' followed by white space
	 * 2 - last seen is '/' which is separated from '*' by white space
	 */
	short_cut = 0;


	/* OK, comment here ?
	 * This is complicated because of all the slight variants of start of
	 * comment that we recognise.
	 * Also if we see end-of-comment, note that we were in a comment (so
	 * that we back off by 1 space).
	 *
	 * Beware buffer overruns: easy to do if "." is near/at line end,
	 * "." could be on virtual '\n'.
	 *
	 * cptr is currently at the first non white space before "."
	 */
	if(cptr <= dotp + 2 && cptr < lbend(curwp->w_dotp)) {
		// cptr[0] is valid

		if(cptr + 1 == lbend(curwp->w_dotp)) {
			// Is first & only char on the line a '*' ?
			in_comment = *cptr == '*';
		} else {
			// cptr[1] is valid

			/* Recognise '/\*' */
			if(*cptr == '/' && cptr[1] == '*') {
				in_comment = 2;
				comment_starts = 1;

				/* Check for line starting '/\*' and NOT white space
				 * This is on the border of being considered malformed.
				 *
				 * Take care overrun if last on line is '/\*'.
				 */
				if(cptr + 2 < lbend(curwp->w_dotp) && cptr[2] != ' ' && cptr[2] != '\t')
					cindent = getccolAtCol(FALSE,
						(LineLn)(cptr - ltext(curwp->w_dotp)) + 1);

			}

			/* Recognise '*\/' */
			if(*cptr == '*' && cptr[1] == '/')
				was_in_comment = 1;

			/* Line start with an '* ' or is '*' ?
			 * Beware: try not to class the next line as comment:
			 *	*foo = 1
			 * But if the user types:
			 *	* foo = 1
			 * we do recognise the line as a comment,
			 * but the idiot user deserves what they got!
			 *
			 * The test immediately below for space or tab is to distinguish
			 * from a ** comment.
			 */
			if(*cptr == '*' && (cptr[1] == ' ' || cptr[1] == '\t' ||
							cptr == dotp - 1)) {
				in_comment = 1;
				short_cut = 1;	// If first following non white is '/' - end comment
			}

			/* Line start with an '** ' or is '**' ?
			 * As above, do not class the following as a comment:
			 *	**foo = 1
			 * but this for white space test is done below.
			 */
			if(cptr <= dotp + 2 && cptr[0] == '*' && cptr[1] == '*') {
				in_comment = 1;
				double_star = 1;

				// If next is white: then first following non white is '/' - end comment
				if(cptr <= dotp + 3 && (cptr[2] == ' ' || cptr[2] == '\t'))
					short_cut = 1;
			}
		}
	}

	/* Identify any white space after the start of comment, this is
	 * so that we can reproduce that level of indent within the comment
	 * on the next line.
	 * If the line ends (ie dotp position) with the '*', then act as if
	 * there was a virtual space immediately after it, if we don't then
	 * coming back to an old comment doesn't work as expected.
	 */
	VirtualSpace = cptr + in_comment + double_star == dotp;

	if(in_comment && (( (VirtualSpace ||
			     cptr[in_comment + double_star] == ' ' ||
			     cptr[in_comment + double_star] == '\t' ||
			     cptr == dotp - 1 - double_star)) ||
	     cptr == dotp + in_comment + double_star)) {
		/* Start of a new comment, add 1 to get leading space */
		if(*cptr == '/' && !doublestar)
			cindent++;

		/* Find out how far the indent is after the start of
		 * comment - reproduce that.
		 */
		cptr += in_comment;

		cptr += double_star;	// Span the second * in **

		while(cptr < dotp && (*cptr == ' ' || *cptr == '\t'))
			cptr++;

		/* Turn the byte offset into a column # */
		indent2 = GetColAtPtr(cptr - VirtualSpace) + VirtualSpace;

		/* Min one space after '*' indent2 == indent) */
		if(cptr[-1] == '*')
			indent2++;

		was_in_comment = in_comment = 1;
	}
	else	/* Well, it turned out not to be a comment after all */
		in_comment = 0;

	/* This code, strictly speaking, isn't correct, but it is better than
	 * just testing the previous character to be an open brace.
	 * The trouble with that is that we miss the brace if it is followed
	 * by a comment. The touble with this code is that we don't look
	 * further back than the start of the current line, so any brace that
	 * we find is commented out - but we don't realise that. In practice
	 * this won't happen often.
	 */
	in_str = 0;
	last_ch = '\0';
	bracef = 0;

	/* We aren't interested in what is *after* dot, it could well
	 * be a close comment, to notice it would mean we don't insert
	 * the comment continuation.
	 * To make life more interesting there could be an unterminated comment
	 * after an open brace, in that case we need to indent the comment
	 * properly, when the comment is eventlually terminated we won't get
	 * the indent level correct, ho, hum.
	 */
	Intp = cptr;	// point to last white space before ".", do not advance past comments
	for(; cptr < dotp; last_ch = *cptr++) {

		switch(*cptr) {
		  case '*':	/* Start of comment continuation line ? */
			SeenNonWhiteInComment = 1;
			if(in_comment) {
				// End of comment ?
				if(cptr + 1 < dotp && cptr[1] == '/') {
					in_comment = 0;
					cptr++;	// Eat the '/'
				}
			}
			break;
		  case '{':	/* Indent level ? */
			SeenNonWhiteInComment = 1;
			bracef = 1;	/* Don't count, note */
			break;
		  case '}':	/* Indent level ? */
			SeenNonWhiteInComment = 1;
			bracef = 0;	/* Can't go -ve */
			break;
		  case '\t':
		  case ' ':	/* Hmmm, may be '* /' at line end */
			if(!SeenNonWhiteInComment)
				indent2 = GetColAtPtr(cptr + 1);

			// Possible start of '*   /' - transform to end of comment
			if(last_ch == '*')
				short_cut = 1;

			continue;	// So don't reset short_cut below

		  case '"':	/* Strings */
		  case '\'':
			SeenNonWhiteInComment = 1;
			if(in_comment)
				break;

			int	bsl = 0;	// Last char in string was a backslash
			in_str = 1;
			quote = *cptr;
			while(++cptr < dotp) {
				if(*cptr == '\\') {
					bsl = 1 - bsl;
					continue;
				}

				// End string if got matching quote not after a '\'
				if(*cptr == quote && !bsl) {
					in_str = 0;
					cptr++;	// Eat the quote
					Intp = cptr;
					break;	// End of string
				}

				bsl = 0;
			}
			// Either ending as end of string or reached "."
			break;

		  case '/':
			// Nothing beyond the '/' ?
			 if(++cptr >= dotp) {

				Intp = cptr - 1;

				// The trailing '/' of '*  /' - set to transform
				if(short_cut == 1) {
					short_cut = 2;
					goto EndOfLine;
				}
				break;
			}

			// Check for // comment
			if(*cptr == '/') {
				cptr--;	// Backspace over the second /
				goto EndOfLine;
			}

			// Check for \* comment
			if(*cptr == '*') {

				BYTE*	t = cptr;	// Save position of start of comment
				in_comment = 1;
				// Eat until end of comment
				while(++cptr + 1 < dotp) {

					if(cptr[0] == '*' && cptr[1] == '/') {
						in_comment = 0;
						short_cut = 0;
						cptr += 2;	// Skip '*' & '/'
						break;
					}
				}

				/* Run out of line while in comment ?
				 * Set cindent to the start of this comment position.
				 */
				if(in_comment) {
					cindent = getccolAtCol(FALSE, (LineLn)(t - ltext(curwp->w_dotp)));
					indent2 = cindent + 2;
					comment_starts = 1;
				}
				continue;
			}
			// DROP THROUGH
		  default:
			SeenNonWhiteInComment = 1;
			Intp = cptr;
			break;
		}

		/* Don't always get here */
		short_cut = 0;
	}

	// Come here if we find a C++ style comment or end of short cut ending
EndOfLine: ;

	/* Autoindent if we find certain C keywords.
	 * Look for 'if', 'for' & 'while' before the matching R paren.
	 */
	keyWordIndent = -1;	// No indent due to keyword
	if(!in_comment && !in_str && (keyWordIndent = CKeyMatch((LineLn)(Intp - ltext(curwp->w_dotp)))) != -1)
		indent = (keyWordIndent + curbp->b_tabsize + 1) / curbp->b_tabsize * curbp->b_tabsize;

	/* Recognised '* /' at line end, swap the ' ' & '/' to make and end
	 * of comment, the space gets zapped later.
	 * NB there could be several space/tab between the '*' & '/'.
	 */
	if(short_cut == 2) {
		in_comment = 0;
		*--cptr = ' ';		/* Zap '/' */

		/* Find '*' (first non white going backwards) */
		while(--cptr > ltext(curwp->w_dotp) &&
				(*cptr == ' ' || *cptr == '\t'))
			;

		/* If comment lines start '**' replace second '*' with  '/' */
		if(double_star)
			cptr--;

		cptr[1] = '/';
	}

	/* If we started the line in a comment, & are not longer in comment,
	 * we need to backspace one on indent.
	 * If the comment was the first non white on the line, keep the indent.
	 */
	if(in_comment == 0 && was_in_comment && indent > 0 &&
			!comment_starts && !double_star)
		indent--;

	/* And one more tab for a brace.
	 * Note, that we just noted that we saw a { after we saw any },
	 * anything else would be horrendous style & deserve whatever we do.
	 * Just in case some pervert wants a non 8 tab indent, look at b_tabsize.
	 * 	indent = (indent + 8) & ~7;
	 * In working out where to indent to, try to find the matching indent
	 * of the start of the statement, if we can't take the indent for
	 * this line.
	 * b_tabsize is a bit of a misnomer here - it is the extra indent on {}
	 */
	if(bracef) {
		/* If we don't already have a target indent from a keyword
		 * get one by trying to find a matching close brace.
		 */
		if(keyWordIndent == -1) {
			if((j = MatchIndent('}')) != -1)
				indent = j;

			indent = (indent + curbp->b_tabsize) / curbp->b_tabsize * curbp->b_tabsize;
		} else	// Use the indent of the keyword
			j = keyWordIndent;


		/* Look forwards to a possible match of closing brace.
		 * If it is of the same indent as the start of the current
		 * statement then it is probable that we have just inserted a
		 * newline at the start of an old compound statement (brace).
		 * In this case we don't want to insert the closing brace.
		 *
		 * Unfortunately consider the following ("." at X).
		 * The first non blank column is not where the keyword ('else')
		 * is, so we will try to insert another close brace.
		 *
		 *	if(xx) {
		 *		yy;
		 *	} else { X
		 *		zz;
		 *	}
		 */
		if(j != -1) {
			LINE*	currlp = curwp->w_dotp;
			LineLn	curroff = curwp->w_doto--;

			if(fmatch('{', 1) == TRUE) {

				if(GetNonBlankCol() == j) {
					bracef = 0;
				} else {
					/* White is a kludge to evade a bug in gcc (GCC) 4.4.7 20120313 (Red Hat 4.4.7-23)
					 * options: -O2 -Wall -Wno-pointer-sign -Wconversion
					 * If the literal string is put as an argument to strspn()
					 * it results in the error:
					 * "conversion to ‘long unsigned int’ from ‘int’ may change the sign of the result"
					 */
					char	White[] = " \t";

					/* Was there a keyword ?
					 * Look see if at WordColBytes there is a close brace
					 * followed by white space until we get to the keyword
					 * which is at offset CKeyPosition.
					 *
					 * GetNonBlankCol() leaves the byte offset in WordColBytes.
					 * CKeyMatch() leaves the byte offset in CKeyBytes.
					 */
					if(keyWordIndent != -1 &&
					   lgetc(currlp, WordColBytes) == '}' &&
					   WordColBytes < CKeyBytes &&
					   strspn(lgetcaddr(currlp, WordColBytes + 1), White) ==
					           (size_t)(CKeyBytes - WordColBytes - 1))
							bracef = 0;

				}
			}

			// lsetNUL(curwp->w_dotp);
			// fprintf(stderr, " bracef=%d in_comment=%d comment_starts=%d in_white=%d cindent=%ld doto=%ld '%s'\n", bracef, in_comment, comment_starts, in_white, (long)cindent, (long)curwp->w_doto, ltext(curwp->w_dotp));

			/* If "." was in a C++ comment then any text following must also
			 * be commented out. Set insCpp to do that and that the level
			 * of indent should be the start of the C++ comment.
			 *
			 * Global variable wasInCppComment is set to that indent.
			 */
			if(wasInCppComment) {
				insCpp = 1;
				indent = wasInCppComment;
			}

			// in_comment should be true - don't set it. Set indent to where the // starts

			// Restore to where we were before we moved by fmatch()
			curwp->w_dotp = currlp;
			curwp->w_doto = curroff;
			// fprintf(stderr, "restore . to %ld in '%s'\n", (long)curroff, ltext(currlp));
		}
	}

	/* Put in the newline */
	if(lnewline(0) == FALSE)
		return(FALSE);

	/* Trim the line that we just left & the line that we are on.
	 * Trailing spaces are a nuisance, I almost never want them.
	 */
	ZapTrailingSpaces(lback(curwp->w_dotp));
	ZapTrailingSpaces(curwp->w_dotp);

	/* Seen the start of a C multi-line comment ?
	 * Insert the close comment on an extra line below.
	 * However: if it looks like the next line is part of this comment, don't do it;
	 * it could be that something was appended to the comment first line.
	 * Look at the next line, span white space, if the first character is '*'
	 * (maybe the next is also '*') then followed by white space or end of line
	 * assume that the next line is part of this comment.
	 * Also don't insert if the next line is end of comment.
	 */
	if(in_comment && comment_starts) {
		LINE*	nl = lforw(curwp->w_dotp);
		LineLn	off = 0;
		BYTE	ch;
		int	insEnd = 1;	// Insert end of comment

		// Span leading white space
		while(off < lused(nl) && ((ch = lgetc(nl, off)) == ' ' || ch == '\t'))
			off++;

		// '*' next ?
		if(off < lused(nl) && lgetc(nl, off) == '*') {
			off++;
			// Span another '*' if there is one
			if(off < lused(nl) && lgetc(nl, off) == '*')
				off++;

			// End of line or white space or end of comment ?
			if(off == lused(nl) || ((ch = lgetc(nl, off)) == ' ' || ch == '\t' || ch == '/')) {
				insEnd = 0;
			}
		}

		/* If there is an end of comment later on the line - inserting another
		 * one would be silly. Look for one.
		 */
		for(nl = curwp->w_dotp, off = curwp->w_doto ; off < lused(nl) - 1; off++)
			if(lgetc(nl, off) == '*' && lgetc(nl, off + 1) == '/') {
				insEnd = 0;
				break;
			}

		if(insEnd) {
			// Insert */ on a line of its own:

			if(lnewline(0) == FALSE || SetToCol(cindent) == FALSE ||
			   linstr("*/") == FALSE)
				return(FALSE);

			moveUp(1);
		}
	}

	/* Unless "." is in the white space at the start of line, (before
	 * whatever non-white there is after) recreate the indent at
	 * the start of the line just inserted.
	 * This will set the indent if the cursor is at end of line,
	 * so repeated NEWLINE results in the cursor staying at whatever indent.
	 */
	if(in_white) {
		BYTE	b;		// A byte

		// Throw away the mix of spaces & tabs:
		while((b = lgetc(curwp->w_dotp, curwp->w_doto)) == ' ' || b == '\t')
			ldelete(1, KL_DISCARD);

		// Replace by something uniform:
		if( !SetToCol(wantIndent))
			return(FALSE);

		// Put the cursor in the column that it was when we entered this function:
		curwp->w_doto = getdoto(curwp, curwp->w_dotp, indent2, DisCol_WP);

	} else {
		if( !SetToCol(in_comment ? cindent : indent))
			return(FALSE);

		/* Just indented "." - but what is "." was in a // comment,
		 * we will have the comment text after "." and not commented
		 * out. So insert a C++ comment so that the text is still a comment.
		 * BUT: if there was nothing on the line after "." - no need.
		 */
		if(insCpp && curwp->w_doto < lused(curwp->w_dotp))
			linstr("// ");	// The indent before // might be a bit wrong
	}

	/* If we are still in a comment append the '*' to the line, and
	 * whatever is needed to get to the comment indent.
	 */
	if(in_comment) {
		if( !linsert(1, '*'))
			return(FALSE);
		if(double_star || (comment_starts && doublestar))
			if( !linsert(1, '*'))
				return(FALSE);
		if( !SetToCol(indent2))
			return(FALSE);
	}

	/* If there was a '{' at the end of the line, insert the
	 * closing '}' on a new line & put the cursor back at the indented
	 * position - so that the body can be typed without further ado.
	 * Note that by now a line (properly indented) has been inserted.
	 * However, there are times we don't want to do this if we are
	 * inserting a line into the start of an existing block, this will be
	 * one of two cases (cursor position when the user hit return shown
	 * by 'X'):
	 * a	if(bool) {X
	 *		foo();
	 *	}
	 * b	if(bool) {X
	 *	}
	 * Case (a) we detect by noting that the indent of 'foo' is >= indent of
	 * the line that we have just inserted.
	 * Case (b) we detect by inserting the line with the '}' and noting that
	 * it is the same as the line following - so we zap the '}' line that we
	 * just inserted.
	 */
#if 0
	if(bracef) {
		/* Check for case (a) */
		moveForw(1);
		j = GetNonBlankCol();
		moveBack(1);
		if(indent <= j)
			bracef = 0;
	}
#endif

	if(bracef && lnewline(0) == TRUE && insbrace(1, '}') == TRUE) {
#if 0
		/* Check for case (b), if so zap what we just inserted */
		if( !linecmp(curwp->w_dotp, lforw(curwp->w_dotp)))
			if( !ldelete(lused(curwp->w_dotp) + 1, KL_DISCARD))
				return(FALSE);
#endif

		/* Put the cursor back back after inserted white space */
		curwp->w_dotp = lback(curwp->w_dotp);
		curwp->w_doto = lused(curwp->w_dotp);
	}

	/* Note so that we can keep the line blank if nothing is put onto it */
	JustAutoIndented = 1;

	return(TRUE);
}

/* Find the indent level of the start of line of the matching opening brace.
 *
 * If, before the opening brace there is a close parenthesis
 * (possibly preceeded by white space) find the opening parenthesis.
 * This is so that we find the start of (eg) an 'if' that may be
 * spread over more than one line, with subsequent lines indented
 * by more than the 'if' itself. We want to make the close brace line
 * up with the 'if', not the line with the open brace.
 *
 * Return the indent or -1 if there was an error (can't find match).
 *
 * c is brace to match to (always } for now).
 */
LineLn	MatchIndent(int c)
{
	LINE*	oldlp;
	LineLn	oldoff;
	LineLn	target;	/* Column that we end up at - match to */

	/* Note where we are */
	oldlp = curwp->w_dotp;
	oldoff = curwp->w_doto;

	/* Find the matching brace. If we can. */
	if(!fmatch(c, c == '}'))
		return(-1);

	/* Got the matching brace, is there a ')' just before it ?
	 * If there is find the line containing the opening parenthesis.
	 * We don't actually care if this succeeds, 'cos if it doesn't
	 * we'll just indent to the start of the line that we are on.
	 */
	curwp->w_doto--;
	SkipBackOverSpace();

	if(curwp->w_doto > 0 && lgetc(curwp->w_dotp, curwp->w_doto) == ')')
		fmatch(')', 0);

	target = GetNonBlankCol();

	curwp->w_dotp = oldlp;
	curwp->w_doto = oldoff;

	return(target);
}

#if	NBRACE
/* Insert a '}' in C or PHP mode
 *
 * n	repeat count
 * c	brace to insert (always } for now)
 *
 * If it is only white space before "." that is removed, replaced by: indent to
 * match the level of indent of the if/whatever of the opening brace.
 *
 * Trailing white space is removed.
 *
 * This does not show the matching opening brace, that is done in insChar();
 *
 * This is not a command.
 *
 * Return:
 *	TRUE if OK
 *	FALSE if cannot insert to the indent
 *	linsert()
 */
int	insbrace(Cnt n, UniChar c)
{
	BYTE	ch;	/* last character before input */
	int	i;
	LineLn	offset;
	LineLn	target;	/* column brace should go after */

	/* If we aren't at the beginning of the line...
	 * scan to see if all space before this is white space.
	 * This also stops us doing something special in a comment - if
	 * approved style.
	 */
	if(curwp->w_doto != 0) {
		for(offset = curwp->w_doto - 1; offset >= 0; --offset) {
			ch = lgetc(curwp->w_dotp, offset);
			if(ch != ' ' && ch != '\t')
				return(linsert(n, c));
		}

		// Zap that white space
		// We'll replace with a nice set of tabs with SetToCol() below:
		offset = curwp->w_doto;
		curwp->w_doto = 0;
		ldelete(offset, KL_DISCARD);
	}

	/* Insert/zap to get to the right column */
	if((target = MatchIndent(c)) >= 0 && !SetToCol(target))
		return(FALSE);

	/* and insert the required brace(s), zap trailing spaces if OK */
	if((i = linsert((BufChrs)n, c)) == TRUE) {
		ZapTrailingSpaces(curwp->w_dotp);
		lchange(WFHARD);
	}

	return(i);
}
#else
/* Insert a brace into the text here...we are in CMODE
 */
int	insbrace(Cnt n, UniChar c)
{
	BYTE	ch;	/* last character before input */
	LineLn	i;
	LineLn	target;	/* column brace should go after */

	/* if we are at the beginning of the line, no go */
	if(curwp->w_doto == 0)
		return(linsert((BufChrs)n, c));

	/* scan to see if all space before this is white space */
	for(i = curwp->w_doto - 1; i >= 0; --i) {
		ch = lgetc(curwp->w_dotp, i);
		if(ch != ' ' && ch != '\t')
			return(linsert((BufChrs)n, c));
	}

	/* delete back first */
	target = getccol(FALSE);	/* calc where we will delete to */
	target -= 1;
	target -= target % curbp->b_tabsize;
	while(getccol(FALSE) > target)
		backdel(TRUE, 1);

	/* and insert the required brace(s) */
	return(linsert((BufChrs)n, c));
}
#endif

/* Insert a # into the text here...we are in CMODE or PHPMODE
 *
 * Return:
 *	TRUE  on success
 *	FALSE if SetToCol() fails
 *	linsert()
 */
int	insHash(void)
{
	int	ch;	/* last character before input */
	LineLn	i;

	/* if we are at the beginning of the line, no go */
	if(curwp->w_doto == 0)
		return(linsert(1, '#'));

	/* scan to see if all space before this is white space */
	for(i = curwp->w_doto - 1; i >= 0; --i) {
		ch = lgetc(curwp->w_dotp, i);
		if(ch != ' ' && ch != '\t')
			return(linsert(1, '#'));
	}

	/* Delete back first */
	if( !SetToCol(0))
		return(FALSE);

	/* and insert the required pound */
	return(linsert(1, '#'));
}

/* Delete blank lines around dot.
 *
 * This is badly named, it deleted empty lines - ie lines not containing any
 * characters. This now deletes lines that are empty or only contain spaces
 * and/or tabs. The original behaviour is obtained by giving a non zero count.
 *
 * What this command does depends if dot is sitting on a blank line. If dot is
 * sitting on a blank line, this command deletes all the blank lines above and
 * below the current line. If it is sitting on a non blank line then it deletes
 * all of the blank lines after the line.
 *
 * Any count is ignored.
 *
 * Command:  delete-blank-lines
 * Bound to: C-x C-o
 *
 * Return:
 *	TRUE  on success
 *	ldelete()
 */
int	deblank(int f, Cnt n)
{
	LINE*	lp1;
	LINE*	lp2;
	BufLns	nld;	// # characters (' ', '\t', '\n') to delete

	if(!f)
		n = 0;

	/* We we are sitting on a blank line, move up to find non blank
	 * Non zero count: line length must be 0 (ie empty)
	 * Zero count:	   line length must be same as leading white bytes
	 */
	lp1 = curwp->w_dotp;
	while(lused(lp1) == (n ? 0 : countWhiteBytes(&ltxt(lp1))) && (lp2 = lback(lp1)) != curbp->b_btopline)
		lp1 = lp2;

	/* Work down until we find a non blank/empty line.
	 * Count the number of characters.
	 */
	lp2 = lp1;
	nld = 0;
	while((lp2 = lforw(lp2)) != curbp->b_abotline && lused(lp2) == (n ? 0 : countWhiteBytes(&ltxt(lp2))))
		nld += 1 + (n ? 0 : lused(lp2));

	if(nld == 0)
		return(TRUE);

	curwp->w_dotp = lforw(lp1);
	curwp->w_doto = 0;

	return(ldelete(nld, KL_DISCARD));
}

/* Insert a newline, then copy the tabs and spaces at the start of the previous line.
 * This might not be optimal, but sidestep the issue of entab() or not.
 *
 * White space auto just inserted is removed from the previous line, ie avoid
 * creating lines that just contain white space.
 * JustAutoIndented records if white space is auto inserted.
 *
 * The starting line will be trimmed.
 * If "." is inside the leading spaces at the start of the line, then it is
 * as if count newlines are inserted before the current line, ie "." is placed
 * at the same position on the new line as it hadbefore.
 *
 * Return TRUE if all ok. Return FALSE if one of the subcomands failed.
 *
 * Also used in INDENT mode with C-M. (See newlineMaybeIndent())
 *
 * Default count is 1. Must be >= 0.
 *
 * Command:  newline-and-indent
 * Bound to: unbound   -- might be bound to C-j
 *
 * Return: TODO
 */
int	newlineAndIndent(int f, Cnt n)
{
	LINE*	startLine = curwp->w_dotp;
	LineLn	white = countWhiteBytes(&ltxt(startLine));
	LineLn	iDoto = curwp->w_doto;

	/* If there is alpha on the line the indent is whatever there
	 * is before the first alpha. If there are no alpha, the indent
	 * is what there is before the first non white space.
	 * The point about this is that it will copy things like email
	 * quoting at line start.
	 *
	 * There is, what will prob be an unusual case, where "." is near
	 * the start of line, after non white space but before alpha. In this
	 * case only copy the initial white space.
	 * Hopefully this will be the action of least surprise for the user.
	 */
	//fprintf(stderr, "nAI white=%ld\n", (long)white);
	if(GetWordCol(1) >= 0 && curwp->w_doto > WordColBytes)
		white = WordColBytes;

	//fprintf(stderr, "nAI now white=%ld doto=%ld WCB=%ld\n", (long)white, (long)curwp->w_doto, WordColBytes);

	while(n >= 1) {
		if(lnewline(TRUE) == FALSE)
			return(FALSE);
		if(n == 1) {
			BYTE	ch;

			curwp->w_doto = 0;
			// After the last line, insert the white space at the start of startLine
			if(insStrLine(curwp->w_dotp, 0, ltext(startLine), white, 1) < 0)
				return(FALSE);

			FixOffsets(curwp->w_dotp, 0, white, curbp, 0);

			curwp->w_doto = white;

			while(curwp->w_doto < lused(curwp->w_dotp) &&
			   ((ch = lgetc(curwp->w_dotp, curwp->w_doto)) == ' ' || ch == '\t'))
				ldelete(1, KL_DISCARD);

			if(curwp->w_doto > iDoto)
				curwp->w_doto = iDoto;

			// Made a change
			ZapTrailingSpaces(startLine);

			// Note if we inserted white space at the line start
			JustAutoIndented = white != 0;

			break;
		}

		n--;
	}

	return(TRUE);
}

/* Insert count lines at "."; unless in INDENT mode indent the last inserted line.
 * 
 * In INDENT (and C & PHP) mode: inserts count new lines before "."
 * If not in INDENT mode this acts like newline-and-indent.
 *
 * The point is that in INDENT mode an indent is always done by newline and
 * newline-and-indent, so it is hard to just start a new line.
 *
 * Return TRUE if all ok. Return FALSE if one of the subcomands failed.
 *
 * Default count 1. Must be >= 0.
 * Normally bound to "C-J" as an alternative to newlineAndIndent().
 *
 * Command:  newline-and-maybe-indent
 * Bound to: C-j
 *
 * Return: TODO
 */
int	newlineMaybeIndent(int f, Cnt n)
{
	int	s;		// Status

	if((curbp->b_mode & (MDINDEN | MDCMOD | MDPHP)) == 0)
		return(newlineAndIndent(f, n));

	while(--n >= 0)
		if((s = lnewline(curwp->w_doto == 0)) != TRUE)
			return(s);
	return(TRUE);
}

/* Starting "." insert tab tabs at the start of lines lines.
 * lines == 0 - just do the current line.
 *
 * Do not indent empty lines or lines starting '#' in C mode
 *
 * If the buffer is in mode MDTABEXP - insert the spaces to the
 * equivalent width of tabs. NB: each buffer has its own tab width.
 *
 * Return:
 *	TRUE  on success.
 *	FALSE run out of lines (end of buffer/narrow)
 *	      run out of memory
 */
int	indentLns(Cnt tabs, BufLns lines) {
	int	s;
	char	tab = '\t';	// What we insert

	if(curbp->b_mode & MDTABEXP) {
		tab = ' ';
		tabs *= curbp->b_tabsize;
	}

	/* loop thru indenting n lines */
	while(1) {
		if(EonLine(curwp->w_dotp))
			return(FALSE);

		// Only indent lines that have some content, but not
		// lines starting '#' in CMODE
		if(lused(curwp->w_dotp) != 0 &&
		   !((curbp->b_mode & MDCMOD) && (lgetc(curwp->w_dotp, 0) == '#'))) {

			curwp->w_doto = 0;	/* start at the beginning */

			/* shift current line using spaces */
			if((s = linsert((BufChrs)tabs, tab)) != TRUE)
				return(s);
		}

		if(lines == 0)
			break;

		/* advance/or back to the next line */
		if(moveDown(1) == FALSE)
			return(FALSE);

		if(--lines == 0)
			break;
	}

	return(TRUE);
}

/* Indent a region n tab-stops or equivalent size.
 *
 * Do not indent empty lines or lines starting '#' in C mode
 *
 * Default count 1. Must be >= 0.
 *
 * Command:  indent-region
 * Bound to: M-)
 *
 * Return TRUE on success.
 */
int	indentRegion(int f, Cnt n)
{
	int	i;
	REGION	reg;

	if((i = getregion(&reg)) != TRUE)
		return(i);

	pnmark0();		// So that we can put "." back where it started
	curwp->w_dotp = reg.r_linep;
	curwp->w_doto = reg.r_offset;

	i = indentLns(n, reg.r_lines);

	retmark0(FALSE);

	return(i);
}

/* Indent n lines by the buffer's tab size
 * n == 0 - just do the current line.
 *
 * Default count 1. Must be >= 0.
 *
 * Command:  indent-line
 * Bound to: M-]
 *
 * Return TRUE on success.
 */
int	indentLines(int f, Cnt n)
{
	int	s = indentLns(1, (BufLns)n);

	curwp->w_doto = 0;	// Return here seems to cause least surprise

	return(s);
}

/* undent/unindent/outdent lines lines by removing up to tabs TABs at the
 * start of every line.
 *
 * tabs	    The number of tabs to remove
 * lines    The number of lines to change
 *	    lines == 0 - just do the current line.
 *
 * Return:
 *	TRUE  on success
 *	FALSE run out of lines (end of buffer/narrow)
 *	      run out of memory
 */
int	unindentLns(Cnt tabs, BufLns lines)
{
	int	s;
	LineLn	spaces = tabs * curbp->b_tabsize;

	/* loop thru indenting n lines */
	while(1) {
		LineLn	lcol = 0;	// Line column - ie position, tabs expanded
		LineLn	doto = 0;	// Byte offset in the line

		curwp->w_doto = 0;	/* start at the beginning */

		if(EonLine(curwp->w_dotp))
			return(FALSE);

		/* Delete up to spaces ' ' at the start of the line.
		 * A tab is one byte but multiple space positions, it always
		 * ends on a multiple of b_tabsize, so setting lcol is safe.
		 * Doing it this way results in one call to ldelete() to zap
		 * all the white space that we remove.
		 */
		while(lcol < spaces && lused(curwp->w_dotp) > doto) {
			if(lgetc(curwp->w_dotp, doto) == ' ')
				lcol++;
			else if(lgetc(curwp->w_dotp, doto) == '\t')
				lcol = (lcol / curbp->b_tabsize + 1) * curbp->b_tabsize;
			else
				break;	// Neither space nor tab

			doto++;
		}

		// Any to delete ?
		if(doto > 0 && (s = ldelete((BufChrs)doto, KL_DISCARD)) != TRUE)
			return(s);

		if(lines == 0)
			break;

		/* advance/or back to the next line */
		if(moveDown(1) == FALSE)
			return(FALSE);

		if(--lines == 0)
			break;
	}

	return(TRUE);
}

/* Unindent a region n tab-stops or equivelent size.
 *
 * Default count 1. Must be >= 0.
 *
 * Command:  unindent-region
 * Bound to: M-(
 *
 * Return TRUE on success.
 */
int	unidentRegion(int f, Cnt n)
{
	int	i;
	REGION	reg;

	if((i = getregion(&reg)) != TRUE)
		return(i);

	pnmark0();		// So that we can put "." back where it started
	curwp->w_dotp = reg.r_linep;
	curwp->w_doto = reg.r_offset;

	i = unindentLns(n, reg.r_lines);

	retmark0(FALSE);

	return(i);
}

/* Unindent n lines by the buffer's tab size
 * n == 0 - just do the current line.
 *
 * Default count 1. Must be >= 0.
 *
 * Command:  unindent-line
 * Bound to: M-[
 *
 * Return TRUE on success.
 */
int	unindentLines(int f, Cnt n)
{
	return(unindentLns(1, (BufLns)n));
}

/* Return the deduced start-of-line comment string to use for the current buffer.
 * The returned string is non writable.
 * The returned string will end with a space.
 *
 * CMODE - use '// ' (good for C99 & later -- ie any modern compiler).
 * Not CMODE - look at the first line of the buffer, assume that that is a comment,
 * where the commentary is separated from the comment string by a space or tab; eg:
 *	# This program does ...
 * If the first line starts '#!' assume that it is a script magic shell line
 * so look at the second line instead.
 * Blank lines will be ignored.
 *
 * Leading white space on a potential comment string will be spanned.
 *
 * PHPMODE used to be treated the same was as CMODE, but PHPMODE is often used
 * for things like Perl programs (which do not use '//' for comments) - so do
 * not do this any more.
 *
 * If we deduce C multi-line ('/ *') - change to '// '.
 * If we see '<?php ' at the start of file, then change to '// '. Not all
 * of these will have a .php suffix.
 *
 * If the buffer only contains 0 lines (or 1 if first is #!) then this might return "".
 * If the buffer only contains blank lines this will return " " - but stupid user.
 *
 * This will work for many languages, but not all eg: CSS, XML, HTML, Pascal,
 * Haskell
 */
char*	getCommentStr(void)
{
	char*		commentStr;	// The comment string that is returned
	static	char	cmtBuf[MAX_COMMENT_LEN + 2];

	if(curbp->b_mode & (MDCMOD)) {
		commentStr = "// ";	// Use this as comment string
	} else {
		LineLn	len;
		char*	play;

		// Look at the first line of buffer, skip if it starts #! something:
		LINE*	firstLn = lforw(curbp->b_linep);
		if(lused(firstLn) >= 2 && memcmp(ltext(firstLn), "#!", 2) == 0)
			firstLn = lforw(firstLn);

		// Skip over blank lines:
		for( ; lforw(firstLn) != curbp->b_linep; firstLn = lforw(firstLn))
			if(countWhiteBytes(&ltxt(firstLn)) != lused(firstLn))
				break;

		// (Almost) empty buffer:
		if(firstLn == curbp->b_linep)
			return("");

		// Find first non white char:
		play = lgetcaddr(firstLn, countWhiteBytes(&ltxt(firstLn)));

		// len will be the max length of comment string:
		len = lused(firstLn) - countWhiteBytes(&ltxt(firstLn));
		if(len > sizeof(cmtBuf) - 2)
			len = sizeof(cmtBuf) - 2;

		// Copy the start of the first line
		memcpy(cmtBuf, play, (size_t)len);
		cmtBuf[len] = '\0';

		// If there is a space (or tab) in that, truncate after the space/tab
		if((play = strchr(cmtBuf, ' ')) != (char*)NULL)
			play[1] = '\0';
		if((play = strchr(cmtBuf, '\t')) != (char*)NULL) {
			play[0] = ' ';		// Tab replaced by space
			play[1] = '\0';
		}
		if( !strchr(cmtBuf, ' '))	// Ensure a space at the end of comment
			strcat(cmtBuf, " ");

		commentStr = cmtBuf;

		/* If it looks like a C style multi-line comment, make it C++ style
		 * Ditto if this looks like PHP source
		 */
		if( !strcmp(cmtBuf, "/* ") || !strcmp(cmtBuf, "<?php "))
			commentStr = "// ";
	}

	return(commentStr);
}

/* Un/comment the current region.
 *
 * This works for languages where comment characters at the start of
 * line make a comment to the end of line. This will not insert beginning/end
 * comment style like this C block comment.
 *
 * If the count is set: remove the comment at the start of each line.
 * Note *start of line* - ie this undoes what it did.
 *
 * Default count unset.
 *
 * Command:  comment-region
 * Bound to: M-C
 *
 * Return TRUE on success.
 */
int	commentRegion(int f, Cnt n)
{
	int	i;
	REGION	reg;
	char*	commentStr = getCommentStr();	// The comment string that is inserted
	size_t	cmtLen = strlen(commentStr);

	if((i = getregion(&reg)) != TRUE)
		return(i);

	pnmark0();		// So that we can put "." back where it started
	curwp->w_dotp = reg.r_linep;
	curwp->w_doto = reg.r_offset;

	/* loop thru indenting n lines */
	while(1) {
		curwp->w_doto = 0;	/* start at the beginning */

		if( !f) {
			// Insert the comment string at start of line
			if((i = linstr(commentStr)) != TRUE)
				return(i);
		} else {
			// Delete the comment string at start of line
			if(lused(curwp->w_dotp) >= (LineLn)cmtLen &&
			   memcmp(ltext(curwp->w_dotp), commentStr, cmtLen) == 0 &&
			   (i = ldelete((BufChrs)cmtLen, KL_DISCARD)) != TRUE)
				return(i);
		}

		if(reg.r_lines == 0)
			break;

		/* advance/or back to the next line */
		moveDown(1);

		if(--reg.r_lines == 0)
			break;
	}

	retmark0(FALSE);

	return(TRUE);
}

/* Delete forward. This is real easy, because the basic delete routine does
 * all of the work. Watches for negative arguments, and does the right thing.
 * If any argument is present, it kills rather than deletes, to prevent loss
 * of text if typed with a big argument.
 *
 * Default cound 1. Must be >= 0.
 *
 * Command:  delete-next-character
 * Bound to: C-d   delete
 *
 * Return: TODO
 */
int	forwdel(int f, Cnt n)
{
	// If the window has the HEX option we might not be at the character start.
	// Move there:
	if(lisUtf8(curwp->w_dotp) && curwp->w_doto < lused(curwp->w_dotp))
		UTF8_BACK_FIND_START_NX(lgetc(curwp->w_dotp, curwp->w_doto), curwp->w_doto);

	/* Always make ldelete save the deleted stuff in a kill buffer
	 * even if only one character.
	 */
	return(ldelete(n, KL_KILL_APPEND));
}

/* Delete backwards.
 *
 * This is quite easy too, because it's all done with other functions. Just
 * move the cursor back, and delete forwards. Like delete forward, this
 * actually does a kill if presented with an argument.
 *
 * This might produce a top of message if it cannot move back
 *
 * Default count 1. Must be >= 0.
 *
 * Command:  delete-previous-character
 * Bound to: C-h   DEL
 *
 * Return: TODO
 *	FALSE fail to move back (top of buffer/narrow)
 */
int	backdel(int f, Cnt n)
{
	int	s = TRUE;
	short	wo_hex = curwp->w_opts & WO_HEX;

	curwp->w_opts &= ~WO_HEX;	// temporarily remove HEX option

	// If the window has the HEX option we might be in the middle of a character.
	// In that case move to the next so that we delete the character that we were on
	while(lisUtf8(curwp->w_dotp) && curwp->w_doto < lused(curwp->w_dotp) && BYTE_IS_MB_CONT(lgetc(curwp->w_dotp, curwp->w_doto)))
		curwp->w_doto++;

	curwp->w_partoff = 0;		/* Ensure that we delete the character before */

	while(--n >= 0)
		if((s = backchar(TRUE, 1)) != TRUE ||
		   (s = ldelete(1, KL_KILL_START)) != TRUE)
			break;

	curwp->w_partoff = 0;		/* Ensure that with hex option the cursor ends up on first hex digit */

	curwp->w_opts |= wo_hex;	// restore HEX option (if it was set)

	return(s);
}

/* Kill text.
 *
 * If called without an argument, it kills from dot to the end of the line,
 * unless it is at the end of the line, when it kills the newline. The
 * exception to this rule occurs when the cursor is at the start of a blank
 * line (one with only spaces and tabs on it), when the newline gets killed too.
 *
 * If called with an argument of 0, it kills from the start of the line to dot.
 * If called with a positive argument, it kills from dot forward over that
 * number of newlines. If called with a negative argument it kills backwards
 * that number of newlines.
 *
 * Default count unset.
 *
 * Command:  kill-to-end-of-line
 * Bound to: C-k
 *
 * Return: TODO
 */
int	killtext(int f, Cnt n)
{
	LINE*		nextp;
	LineLn		chunk;		/* Number of characters to kill */
	LineLn		off;		/* offset within current line	*/

	curwp->w_partoff = 0;

	if(f == FALSE) {
		// ldelete() deletes characters
		// Find how many from dot to end of line
		if(lisUtf8(curwp->w_dotp))
			chunk = (LineLn)utf8_chrnlen(lgetcaddr(curwp->w_dotp, curwp->w_doto), (size_t)(lused(curwp->w_dotp) - curwp->w_doto));
		else
			chunk = lused(curwp->w_dotp) - curwp->w_doto;

		/* If we are at start of line and the line consists of spaces
		 * and tabs, eat the newline at the end of it too.
		 *
		 * This makes sure that blank lines are eaten.
		 */
		if(curwp->w_doto == 0) {
			for(off = 0; off < chunk; off++)
				if(lgetc(curwp->w_dotp, off) != '\t' &&
				   lgetc(curwp->w_dotp, off) != ' ')
					break;
			if(off == chunk)
				chunk++;/* eat newline too 	*/
		} else {
			/* if we're not at the start of a line and we're at the
			 * end of the line, then eat the newline too
			 */
			if(chunk == 0)
				chunk++;
		}

	} else if(n == 0) {
		if(lisUtf8(curwp->w_dotp))
			chunk = (LineLn)utf8_chrnlen(lgetcaddr(curwp->w_dotp, 0), (size_t)curwp->w_doto);
		else
			chunk = curwp->w_doto;

		curwp->w_doto = 0;

	} else if(n > 0) {
		/* Find out how many characters to the end of the current line.
		 * The cursor might not be at the start of it.
		 */
		nextp = curwp->w_dotp;
		off = curwp->w_doto;
		chunk = 0;

		/* Add in the number of characters in the number of lines that are to be killed.
		 */
		while(--n >= 0) {
			// Stop if the kill if it would take us beyond the end of the buffer/narrow
			if(EonLineB(curbp, nextp))
				break;

			if(lisUtf8(nextp))
				chunk += (LineLn)utf8_chrnlen(lgetcaddr(nextp, off), (size_t)(lused(nextp) - off));
			else
				chunk += lused(nextp) - off;

			chunk++;	// For the newline

			// Start of next line
			nextp = lforw(nextp);
			off = 0;
		}
	} else {
		/* Negative count.
		 * Find out how many characters to the start of the current line.
		 * The cursor might not be at the start of it.
		 */
		if(lisUtf8(curwp->w_dotp))
			chunk = (LineLn)utf8_chrnlen(ltext(curwp->w_dotp), (size_t)(curwp->w_doto));
		else
			chunk = curwp->w_doto;

		nextp = curwp->w_dotp;

		/* Add in the number of characters in the remaining number of lines that are to be killed.
		 */
		while(++n <= 0) {
			LINE*	prev = lback(nextp);

			// Stop the kill if it would take us beyond the start of the buffer/narrow
			if(SonLineB(curbp, prev))
				break;

			nextp = prev;
			chunk++;	// For the newline

			if(lisUtf8(nextp))
				chunk += (LineLn)utf8_chrnlen(ltext(nextp), (size_t)lused(nextp));
			else
				chunk += lused(nextp);
		}

		curwp->w_dotp = nextp;
		curwp->w_doto = 0;
	}

	// Killing zero bytes - must be one end or the other
	if(chunk == 0) {
		mlwrite("%EEnd of buffer");
		return(FALSE);
	}

	// Delete the number of characters that we worked out above:
	return(ldelete(chunk, KL_KILL_APPEND));
}

/* Prompt and set an editor mode.
 *
 * Used to give a 0 or neg argument, kill the mode.
 * To only need setMode() mapped, not delMode().
 * Not used, now pass f, n to the adjust function.
 *
 * Not setmode() to avoid clash with BSD file permission manipulation function.
 *
 * The count is passed to the adjust function - which might look at it.
 * Default count unset.
 *
 * Command:  add-mode
 * Bound to: C-x m
 *
 * Return: TODO
 */
int	setMode(int f, Cnt n)
{
	return(adjustmode(TRUE, FALSE, f, n));
}

/* Prompt and delete an editor mode.
 *
 * The count is passed to the adjust function - which might look at it.
 * delMode() not delmode() to keep setMode() company
 *
 * Default count unset.
 *
 * Command:  delete-mode
 * Bound to: C-x C-m
 *
 * Return: TODO
 */
int	delMode(int f, Cnt n)
{
	return(adjustmode(FALSE, FALSE, f, n));
}

/* Prompt and set a global editor mode.
 *
 * setgMode() not setgmode() to keep setMode() company
 *
 * The count is passed to the adjust function - which might look at it.
 * The default count is unset.
 *
 * Command:  add-global-mode
 * Bound to: M-m
 *
 * Return: TODO
 */
int	setgMode(int f, Cnt n)
{
	return(adjustmode(TRUE, TRUE, f, n));
}

/* Prompt and delete a global editor mode.
 *
 * delgMode() not delgmode() to keep setMode() company
 *
 * The count is passed to the adjust function - which might look at it.
 * The default count is unset.
 * 
 * Command:  delete-global-mode
 * Bound to: M-C-m
 *
 * Return: TODO
 */
int	delgMode(int f, Cnt n)
{
	return(adjustmode(FALSE, TRUE, f, n));
}

/* Change the editor mode status.
 *
 * kind;	true = set,		false = delete
 * global;	true = global flag,	false = current buffer flag
 *
 * Return TRUE if OK.
 *	  FALSE nothing available at prompt
 *	  ABORT C-g
 */
int	adjustmode(int kind, int global, int f, Cnt n)
{
	register int	i;		/* loop index */
	char		prompt[50];	/* string to prompt user with */
	int*		modep;		/* Which mode flags we are changing */
	int		ret = TRUE;	// What this returns
	VAL*		mode;		// Mode - entered by user

	mode = valGet(V_STR);

	/* build the proper prompt string */
	strcpy(prompt, global ? "Global mode to " : "Mode to ");

	strcat(prompt, (kind == TRUE) ? "add: " : "delete: ");

	/* prompt the user and get an answer */
	if((ret = mlreply(prompt, &mode, MLUPPER | MLXMODE, NULL)) != TRUE)
		goto off;

	modep = global ? &gmode : &curwp->w_bufp->b_mode;

	// ? shows the current modes
	if(vtext(mode)[0] == '?') {
		if(vtext(mode)[1] == '?')
			global = TRUE;

		ShowModes((global ? "Global modes" : "Buffer modes"), &ModeBuf, *modep);
		goto off;
	}

	/* First, is it a mode that we know about ? */
	if((i = ModeMatch(&ModeBuf, vtext(mode), (int)tused(vTEXT(mode)), FORW, -1)) >= 0) {
		/* Before we set it, check that it is OK to do it.
		 * Any checking function is responsible for an appropriate error message.
		 */
		int (*modeCheck)(BUFFER*, int, int, int, Cnt) = ModeCheckFuns[i];

		mlerase();	// Erase the junk - before mode function which might write something

		if(modeCheck && !((*modeCheck)(curbp, kind, global, f, n))) {
			// Might have done something - eg modified the buffer
			upmode(NULL, WFMODE);

			ret = FALSE;
			goto off;
		}

		/* finding a match, we process it */
		if(kind == TRUE)
			*modep |=  (1 << i);
		else
			*modep &= ~(1 << i);
			
		/* display new mode line */
//		if(global == 0)
			upmode(NULL, WFMODE);

		goto off;
	}

	/* next, test it against the colors we know */
	for(i = 0; i < NCOLORS; i++) {
		if(strncmp(vtext(mode), cname[i], (size_t)tused(vTEXT(mode))) == 0) {
			/* finding the match, we set the color */
			if(global)
				gfcolor = i;
#if	LEGACY_DISPLAY
			else
				curwp->w_fcolor = (char)i;
#endif
			mlerase();
			goto off;
		}
	}

	mlwrite("%WNo such mode: %v", mode);
	ret = FALSE;

	// Unified return - to release mode
off:	valRelease(mode);

	return(ret);
}

/* Check that buffer modes newModes can be set on curbp, if OK
 * un/set them.
 * This may set and unset many modes all at once.
 *
 * Where newModes differs from the current b_mode run the
 * checking function - if there is one.
 *
 * If the modes do not differ - do nothing.
 *
 * Return TRUE  all tests are passed.
 *	  FALSE a test failed
 */
int	checkNewModes(BUFFER* bp, int newModes)
{
	int	oldModes = bp->b_mode;
	int	changed = newModes ^ oldModes;	// The bits that are to be changed
	int	bitPos = 0;

	for( ; changed != 0; bitPos++) {
		int	bitMask = 1 << bitPos;

		// This bit not changed ? Look at next bit
		if((changed & bitMask) == 0)
			continue;

		changed &= ~bitMask;	// Remove from those changed

		// Get checking function
		int (*modeCheck)(BUFFER*, int, int, int, Cnt) = ModeCheckFuns[bitPos];

		int	kind = newModes & bitMask;
		if(modeCheck && !((*modeCheck)(bp, kind, 0, FALSE, 0)))
			return(FALSE);

		// kind is true if setting the mode
		if(kind)
			bp->b_mode |=  bitMask;
		else
			bp->b_mode &= ~bitMask;
	}

	return(TRUE);
}

/* Check the modes of a buffer and if they are consistent.
 *
 * We are checking the modes that are already set, not will be set. Thus not
 * needed to set if OK, but must unset if a problem.
 *
 * Do things like: if CMODE is set then make sure that INDENT mode is not set.
 *
 * bp might not be the one displayed by curwp
 *
 * Beware: this will be called in early startup before curwp and curbp are assigned.
 * Don't do anything else we'll just crash on NULL dereference.
 */
void	chkmode(BUFFER* bp)
{
	int	i;

	// Early call - do nothing.
	if(curwp == NULL || curbp == NULL)
		return;

	// Check all possible bits in b_mode
	for(i = 0; i < (int)(sizeof(ModeCheckFuns) / sizeof(ModeCheckFuns[0])); i++) {
		int	bit = (1 << i);		// What bit ?

		// Not set - nothing to check
		if((bp->b_mode & bit) == 0)
			continue;

		// Is there a checking function ?
		if( !ModeCheckFuns[i])
			continue;		// No

		if( !(*ModeCheckFuns[i])(bp, TRUE, FALSE, FALSE, 0))
			bp->b_mode &= ~bit;	// Unset - the mode is not liked
	}
}

/* Flip/toggle/reverse the insert state of a buffer.
 * 
 * This is called overwrite elsewhere - but by default this is bound to the key marked: Insert
 *
 * If the count is set
 * * zero     argument reverses the overwrite state.
 * * positive argument turns overwrite mode on.
 * * negative argument turns overwrite mode off.
 *
 * The default count is not set.
 *
 * Command:  toggle-insert-mode
 * Bound to: insert
 *
 * Return: TRUE
 */
int	ToggleInsertMode(int f, Cnt n)
{
	if(f && n)
		if(n > 0)
			curbp->b_mode |= MDOVER;
		else
			curbp->b_mode &= ~(MDOVER);
	else
		curbp->b_mode ^= MDOVER;

	upmode(curbp, WFMODE);

	return(TRUE);
}


/* Called before un/setting VIEW mode on the current buffer - curbp.
 * 
 * bp	  the BUFFER to check/set.
 * set	  true if setting the mode
 * global true if setting the global flag - else the current buffer
 *
 * Do not allow unsetting of the flag if the buffer has RENEW mode.
 *
 * Do not allow unsetting of the flag if the buffer is immutable.
 *
 * The count is ignored.
 *
 * Return true as it is OK to do the action.
 */
int	view_modeset(BUFFER* bp, int set, int global, int f, Cnt n)
{
	if(global)		// We only worry about specific buffers.
		return(TRUE);

	if( !set && (bp->b_mode & MDRENEW)) {
		mlwrite("%ERENEW mode is set, cannot unset VIEW mode");
		return(FALSE);
	}

	if( !set && (bp->b_flag & BIMMUTABLE)) {
		mlwrite("%EImmutable buffer, cannot unset VIEW mode");
		return(FALSE);
	}

	return(TRUE);
}


/* This function simply clears the message line, mainly for macro usage.
 *
 * The count is ignored.
 *
 * Command:  clear-message-line
 * Bound to: unbound
 *
 * Return: TRUE
 */
/* ARGSUSED */
int	clrmes(int f, Cnt n)
{
	mlferase();
	return(TRUE);
}

/* This function writes a string on the message line mainly for macro usage.
 *
 * phcl change:
 * -  if count is >= 0 write to message line
 *
 * The default count is 0.
 *
 * Command:  write-message
 * Bound to: unbound
 *
 * Return:
 *	TRUE  on success
 *	FALSE nothing available at prompt
 *	ABORT C-g
 */
int	writemsg(int f, Cnt n)
{
	int	status;
	int	oldcmd;	/* original command display flag */

	if((status = mlprompt("Message to write: ", MLSTRING, bmlHistp)) != TRUE)
		return(status);

	if( !f)
		n = 0;

	/* write the message out
	 * Can't use mlforce() since that always writes to both history & message line.
	 * So we need to temporarily force discmd.
	 */
	oldcmd = discmd;	/* save the discmd value */
	discmd = TRUE;		/* and turn display on */

	/* The message is already in the history buffer,
	 * does it need to go to the message line as well ?
	 * Beware: defeat SCCS
	 */
	if(n >= 0) {
		// Ensure that the message is written regardless of $early-log
		int	le = EarlyLog;
		EarlyLog = TRUE;

		mlwrite("%F""%h%l", bmlHistp->b_dotp);

		EarlyLog = le;
	}

	discmd = oldcmd;	/* and restore the original setting */

	return(TRUE);
}

#if	CFENCE

/* Return the length of line, before any one-line comment.
 * A one-line comment is a C++ comment or comment starting with a #
 * Eg return the length before // in the line below:
 *	a = b;	// do something
 *
 * Recognise the file specific one-line comment string, but only when
 * preceeded by only white space and followed by space, tab or end of line.
 *
 * This is for use in fmatch() when scanning backwards.
 *
 * This does not cope well if the start of line is in a C style comment
 * that was started on a previous line. We do not know that and assume
 * that we are NOT in a C style comment.
 *
 * Scan from the start of the line -- which is easier than backwards
 */
LineLn	nonCommentLength(const LINE* line)
{
	LineLn	lineLength = lused(line);
	LineLn	pos;
	int	InComment = 0;		// In a C style comment
	int	InCppComment = 0;	// In a C++ style comment
	int	InHashLine = 0;		// In a #define/#if or shell style comment
	int	white = 1;		// Only seen white space so far
	BYTE	c;
	BYTE	lastc = '\0';
	char*	bufCmtStr = getCommentStr();	// Buffer specific start of line comment string
	LineLn	cmtLen;				// How long - excluding trailing space
	cmtLen = (LineLn)strlen(bufCmtStr) - 1;

	for(pos = 0; pos < lineLength; ) {
		if(white) {
			/* Check if seen start of buffer specific one-line comment
			 * We might actually be checking for this again below, eg
			 * '//' in a C file, tough.
			 *
			 * Do not compare the space, but insist on space, TAB or
			 * end of line following.
			 */
			if(lineLength - pos >= cmtLen &&
			   memcmp(bufCmtStr, lgetcaddr(line, pos), (size_t)cmtLen) == 0 &&
			   (lineLength - pos == cmtLen ||
				// Comment does not end the line, so check the char after it
			    (lgetc(line, pos + cmtLen) == ' ' ||
			     lgetc(line, pos + cmtLen) == CharTab))) {
				lineLength = pos;
				break;
			}
		}

		c = lgetc(line, pos++);
		switch(c) {

		   case '\'':	// Unify string/char processing
		   case '"':	if( !InComment) {
					BYTE	StringDelim = c;	// Delimiter of current string

					while(pos < lineLength) {
						c = lgetc(line, pos++);
						if(c == '\\') {
							pos++;
							continue;
						}
						if(c == StringDelim)
							break;
					}
				}
				break;

			// \ no longer special as strings done properly

		   case '#':	if(white)
					InHashLine = 1;
				break;

		   case '/':
				if(InComment) {
					if(lastc == '*')
						InComment = 0;
					break;
				}
				if(lastc == '/')
					InCppComment = 1;
				break;

		   case '*':
				if(lastc == '/')
					InComment = 1;
				break;

		   case ' ': case '\t':
				// Avoid setting white
				lastc = c;
				continue;
		   // Ignore anything else - except note non white
		}

		white = 0;

		// We ignore everything after # or // comment:
		if(InHashLine || InCppComment) {
			lineLength = pos - 1;
			break;
		}

		lastc = c;
	}

	return(lineLength);
}

/* Close fences are matched against their partners, and if
 * on screen the cursor briefly lights there.
 * This is now also used as a way of finding the opening line of an if/while/...
 * so that we can work out where to place a closing brace - or if to place
 * one at all.
 *
 * This can't be bound by the user.
 * Return TRUE/FALSE to indicate success.
 *
 * ch is the bracket to match against.
 * Special value -2 is unmatchable - hence we search the whole file
 * in the forward direction. This is useful for FindCError().
 *
 * If we failed 'cos there was something wrong with the C program
 * (eg unmatched quote), set FailAtDotp & FailAtDoto.
 * If we fail & FailAtDotp is NULL it is 'cos we didn't start on a fence.
 * This should really only be done going forwards, going back doesn't always
 * get it quite right - it is a harder problem.
 *
 * This worked properly when going forwards, but going backwards had problems
 * with some strings (involving '\') and C++ style comments.
 *
 * Strings must not span lines - otherwise strange things might happen.
 * But if a line ends in \ then span a string to the next/previous line.
 *
 * Recognise C++ style comments differently - they end with line end.
 * matchBeforeCpp will only be set when looking for a '{' and when we
 * want to know the indent for the line that is being inserted.
 * Global wasInCppComment is set to the start column of such a comment if "."
 * is in a C++ comment.
 *
 * Recoginse #if/#define & similar, also end with the line and effects
 * must not go beyond the line end. This is also useful for languages where
 * comments are introduced by a '#'. Call this a sh (shell) comment.
 * But we only recognise a # if it is at line start (first non white space),
 * # can be used by CPP in stringizing - this means that a trailing comment
 * introduced by a # will not be recognised as a comment.
 *
 * This is more complicated than I would like, partly because "." can start
 * inside a comment - when a match should happen but restricted to be
 * within that comment.
 * "." can also start within a string - we match brackets within that string.
 *
 * Nah - this code is just ugly!
 * It has had several updates/rewrites - still ugly.
 */
int	fmatch(int ch, int matchBeforeCpp)
{
	BYTE	c;		/* current character in scan	*/
	int	count;		/* current fence level count	*/
	char	ofence;		/* open/other fence		*/
	int	sdir;		/* search direction		*/
	LINE*	oldlp;		/* original line pointer	*/
	LineLn	oldoff;		/* and offset			*/
	LineLn	lineLength;	// Length of current line
	LineLn	lineStart = 0;	// Where the current line is considered to start

	int	InString;	// Double quoted character/string
	LineLn	StringStart = 0;// Where the string started - if InString
	int	InComment;	/* In a C style comment */
	int	LastCh;		/* Last char that we looked at */
	int	white = 1;	// Only seen white space so far this line
	int	StartedInComment = 0;	// At start "." is within a C or C++/sh comment
	int	ThisLineOnly = 0;	// Do not scan another line
	int	StartedInString = 0;
	BYTE	StartingCh;
	BYTE	strDelim = 0;	// String delimiter: single/double quote
	char*	bufCmtStr = getCommentStr();	// Buffer specific deduced start of line comment string
	LineLn	cmtLen;				// How long - excluding trailing space
	cmtLen = (LineLn)strlen(bufCmtStr) - 1;

	// Are we in C or PHP mode -- it is about the comments:
	int	c_php_mode = (curbp->b_mode & (MDCMOD | MDPHP)) != 0;

	wasInCppComment = 0;	// If we find a C++ comment this might get set later

	/* save the original cursor position */
	oldlp = curwp->w_dotp;
	oldoff = curwp->w_doto;

	FailAtDotp = (LINE*)NULL;

	/* setup proper matching fence */
	switch(ch) {
		case '(': ofence = ')'; sdir = FORWARD; break;
		case '{': ofence = '}'; sdir = FORWARD; break;
		case '[': ofence = ']'; sdir = FORWARD; break;
		case '<': ofence = '>'; sdir = FORWARD; break;
		case ')': ofence = '('; sdir = REVERSE; break;
		case '}': ofence = '{'; sdir = REVERSE; break;
		case ']': ofence = '['; sdir = REVERSE; break;
		case '>': ofence = '<'; sdir = REVERSE; break;
		case -2:  ofence = -2;  sdir = FORWARD; break;
		default: /* Didn't start at right place */
			return(FALSE);
	}

	//if(sdir == REVERSE) fprintf(stderr, "\nofence='%c' sdir=%d doto=%ld\n", ofence, sdir, (long)oldoff);

	/* If we are not sitting on the fence character - set count to 1 so that we try to hunt
	 * for it's match.
	 */
	StartingCh = (BYTE)(curwp->w_doto < lused(curwp->w_dotp) ? lgetc(curwp->w_dotp, curwp->w_doto) : '\n');
	count = StartingCh == ch ? 0 : 1;

	/* Not sure why I need this, but put in after removing backchar()
	 * Needed to ensure than an inserted close paren is displayed.
	 */
	curwp->w_flag |= WFMOVE;
	lineLength = lused(curwp->w_dotp);

	/* We get things wrong if we start in a char const, string or comment.
	 * So we fail once we discover that we are in a char const or string
	 * and hit newline.
	 */
	InString = InComment = 0;
	LastCh = '\0';

	/* Purpose:
	 * Determine starting state: if we are in a comment or not ? Are we in a string/char-constant or not ?
	 * If not no other action/setting needed: any comment that we come across will
	 * be handled by the code below.
	 * pos is the scanning position.
	 *
	 * If in C++/sh comment - ie "." after pos:
	 * 	set thislineonly, linelength to lused(.), linestart to pos (comment start)
	 * Deduced commans are treated as sh comments.
	 *
	 * If in C comment:
	 *     forward - if have end on this line
	 *     		set thislineonly & linelength to comment end
	 *		else do nothing, allow to move to next line but set startedInCcomment
	 *     backward - if have start on this line
	 *     		set thislineonly & linestart to comment start
	 *		else do nothing, allow to move to next line but set startedInCcomment
	 * Scan from start of line:
	 *     - If meet a C++/sh comment - stop scanning. It is possible that we are in
	 *     a multi line C comment (and thus it should not start a comment), but we would
	 *     need to look back (maybe many) previous lines to know this.
	 *     It is also possible that there might be a C comment after the C++/sh comment,
	 *     don't notice this specially.
	 *     Just accept that we might be wrong.
	 *     - If meet a C start/end comment continue scanning - it might change.
	 *     There are 3 states WRT C comments: unknown, in, not-in. Transition from
	 *     unknown will tell us about the status of "." if it is before transition.
	 *     - strings/chars: treat the same (AKA javascript).
	 *     Assume that they do not span lines. They function to protect comment sequences.
	 */

	// This used to be C/PHP mode only but deduced comments means that we need to always do this:
	if(1) {
		LineLn	pos = 0;
		int	InHashLine = 0;		// In a #define/#if, shell style or deduced comment
		int	InCppComment = 0;	// In a C++ style comment
		LineLn	CcommentStart = 0;	// init to keep compiler quiet
		enum {unknown, in, out} InCcomment = unknown;

		// Set to -1 as flag that we have not yet seen "."
		InString = -1;

		// fprintf(stderr, "pre: ll=%ld .=%ld\n", (long)lused(curwp->w_dotp), (long)curwp->w_doto);

		while(pos < lused(curwp->w_dotp) && pos <= curwp->w_doto) {

			// Line start with deduced start-of-line comment ?
			if(white && cmtLen > 0 && pos + cmtLen <= curwp->w_doto &&
			   memcmp(bufCmtStr, lgetcaddr(curwp->w_dotp, pos), (size_t)cmtLen) == 0 &&
			   (pos < curwp->w_doto ||
			    (lgetc(curwp->w_dotp, pos + cmtLen) == ' ' ||
			     lgetc(curwp->w_dotp, pos + cmtLen) == CharTab)))
				ThisLineOnly = InHashLine = 1;

			c = lgetc(curwp->w_dotp, pos++);

			// if(sdir == REVERSE)fprintf(stderr, "c='%c' LastCh='%c' pos=%ld w=%d InCppComment=%d InCcomment=%d InComment=%d ccstrt=%ld InString=%d InStr=%d\n", c, LastCh, (long)(pos-1), white, InCppComment, InCcomment, InComment, (long)CcommentStart, InString, InStr);

			switch(c) {
			   case '\'':
			   case '"':	if(InCcomment != in) {
						/* Note some things about the string.
						 * Usefull when we meet the start of the string
						 */
						strDelim = c;
						InString = 1;
						StringStart = pos - 1;

						// Look for the end of the string
						while(pos < lused(curwp->w_dotp) && pos <= curwp->w_doto) {
							c = lgetc(curwp->w_dotp, pos++);
							if(c == strDelim) {
								InString = 0;
								break;
							}
							// See a \ ?
							if(c == '\\' && pos < lused(curwp->w_dotp)) {
								pos++;	// Eat whatever is next
							}
						}
					}

					break;

			   // case '\\':	// \ only special in strings, dealed with above	

			   case '#':
					if( ! c_php_mode)
						break;

					if(white)	// white implies not in a string or comment
						InHashLine = 1;

					break;
			   case '/':
					if( ! c_php_mode)
						break;

					// End of C comment if last char was '*'
					if(InCcomment) {
						if(LastCh == '*')
							InCcomment = unknown;
						break;
					}

					if(LastCh == '/') {
						// Start of C++ comment
						InCppComment = 1;

						// Record where the comment started
						// This is so wrong!
						wasInCppComment = getccolAtCol(FALSE, pos - 2);
						break;
					}

					// End of C comment
					if(LastCh == '*') {
						if(InCcomment == in && curwp->w_doto >= CcommentStart && curwp->w_doto <= pos) {
							// "." is inside the comment
							InComment = 1;
							lineStart = CcommentStart;
							lineLength = pos;
							ThisLineOnly = 1;
						}

						if(InCcomment == unknown && curwp->w_doto <= pos) {
							// Ending comment containing "." started on a previous line
							InComment = 1;
							lineLength = pos;
							ThisLineOnly = 1;
						}

						InCcomment = out;
					}

					break;
			   case '*':
					if( ! c_php_mode)
						break;

					// Start of C style comment - if the last char was '/'
					if(LastCh == '/') {
						// Start of C comment, note where it starts
						InCcomment = in;
						CcommentStart = pos;
						break;
					}

					break;
			   case ' ': case '\t':
					// Avoid setting white false
					LastCh = c;
					continue;
			   // Ignore anything else - except note non white
			}

			white = 0;

			if((InHashLine || InCppComment)) {
				// if(sdir == REVERSE) fprintf(stderr, "InHashLine=%d InCppComment=%d\n", InHashLine, InCppComment);
				if(sdir == REVERSE && matchBeforeCpp) {
					curwp->w_doto = pos - 1;
					InCppComment = 0;
				}

				if(curwp->w_doto >= pos) {
					lineStart = pos;
					ThisLineOnly = 1;	// Cursor in a 1 line comment, don't move to another line
				}

				// lsetNUL(curwp->w_dotp);
				// if(sdir == REVERSE) fprintf(stderr, "seenC ls=%ld ThisLineOnly=%d line='%s'\n", (long)lineStart, ThisLineOnly, ltext(curwp->w_dotp));

				break;
			}

			/* "." is inside a C comment - we have just come across the end
			 * of that comment. Stop here as we do not match anything
			 * outside of the comment that we started within.
			 */
			if(InComment)
				break;

			LastCh = c;
		}

		// Not seen a C comment ? Then we are probably outside of one
		if(InCcomment == unknown)
			InCcomment = out;

		/* Seen the start of a C comment but not the end ?
		 * We must be in that comment - do not reverse scan this line before the start of that comment
		 */
		if(InCcomment == in && curwp->w_doto >= CcommentStart && !InComment) {
			//if(sdir == REVERSE)fprintf(stderr, "InCcomment=%d !InComment=%d\n", InCcomment, !InComment);
			InComment = 1;
			lineStart = CcommentStart + 2; // After the '/*'
			if(sdir == REVERSE)
				ThisLineOnly = 1;
		}

		/* If we started in a comment then we are allowed to notice
		 * new ch and ofence as we meet them.
		 */
		StartedInComment = InComment || InHashLine || InCppComment;

		// fprintf(stderr, "END pos=%ld w=%d ic+=%d InCcomment=%d InComment=%d ccstrt=%ld InString=%d ls=%ld ll=%ld ThisLineOnly=%d StartedInComment=%d\n",
			// (long)pos, white, InCppComment, InCcomment, InComment, CcommentStart, InString, (long)lineStart, (long)lineLength, ThisLineOnly, StartedInComment);

		if(InString == -1)
			InString = 0;

		StartedInString = InString;
		if(StartedInString) {
			ThisLineOnly = 1;
			if(StartingCh)
				count = 1;
		}

		LastCh = '\0';
	}

	/* Lift restriction about not going back very far if "." started
	 * in a C++ comment and we have been asked to do this.
	 */
	if(matchBeforeCpp)
		ThisLineOnly = 0;

	/* Scan until we find it, or reach the end of file */
	for(;;) {
		c = (BYTE)((curwp->w_doto < lineLength) ?
		    lgetc(curwp->w_dotp, curwp->w_doto) : '\n');

		/* A peek at the next char makes things much easier when
		 * going backwards (in which case next means previous in the line)
		 * for when we want to know if a quote is escaped by backslash.
		 */
		BYTE	NextCh;
		if(sdir == FORWARD) {
			NextCh = (BYTE)((curwp->w_doto + 1 < lineLength) ? lgetc(curwp->w_dotp, curwp->w_doto + 1) : '\n');
		} else {
			/* Backwards scanning.
			 * See comments below as to what is happening here
			 * In effect we are eliminating pairs of backslashes
			 */
			while(1) {
				BYTE	nextNext;

				NextCh = (BYTE)((curwp->w_doto - 1 >= lineStart) ? lgetc(curwp->w_dotp, curwp->w_doto - 1) : '\n');
				nextNext = (BYTE)((curwp->w_doto - 2 >= lineStart) ? lgetc(curwp->w_dotp, curwp->w_doto - 2) : '\n');

				if(NextCh != '\\' || nextNext != '\\')
					break;

				curwp->w_doto -= 2;
			}
		}

		/* This was just for C/PHP but things like shell scripts, CSS files, ...
		 * also benefit from this processing - so just do it everywhere.
		 * So, the comments are a bit more general than just C/PHP.
		 *
		 * In C mode, take care of comments, strings & char constants (S&CC).
		 * Braces in these are effectively hidden.
		 * **** Assume correct C syntax/programs. ****
		 *
		 * Strings: code & comments rewritten yet again.
		 * 
		 * Coping with a ["'] at start of line is more difficult & we
		 * don't get it always right with trailing '\'.
		 * When it comes to testing this, this code itself is superb
		 * at finding bugs.
		 * Do the same for PHP, good enough.
		 *
		 * Why the brace ? Well, it was if(CMODE)
		 */
		{
			int	wasSpace = 0;

			// Resume string scanning - if that is what we were doing
			if(InString)
				goto ScanString;

			switch(c) {

			case '\'':
			case '"':	/* String */
				/* If going forwards, eat to end of string;
				 *   if see a \ eat the next whatever it is,
				 *   it might be a \ or quote.
				 * If going backwards, eat to start of string;
				 *   if see a \ or quote, look forwards to see if there
				 *   is a \ before it that is not made ordinary by a
				 *   preceeding \.
				 *
				 * As far as we are concerned: strings end at end of line,
				 * unless there is a \ to continue it.
				 */
				strDelim = c;

		ScanString: 		;	// Come here to resume string scanning
				InString = 0;	// Don't come here again

				if(InComment)	// Ignore quotes within comments
					break;

				if(sdir == FORWARD) {
					while(1) {
						curwp->w_doto++;
						if(curwp->w_doto >= lineLength)
							goto LineEnd;
						c = lgetc(curwp->w_dotp, curwp->w_doto);

						if(StartedInString) {
							// Match brackets within the string that "." started in
							if(c == ch)
								++count;
							if(c == ofence)
								--count;

							/* Got to a match ? */
							if(count == 0)
								return(TRUE);
						}

						if(c == '\\') {
							curwp->w_doto++;	// Just eat whatever is next

							if(curwp->w_doto >= lused(curwp->w_dotp)) {
								/* Backslash is the last char on the line.
								 * Continue the string on the next line.
								 */
								curwp->w_dotp = lforw(curwp->w_dotp);
								curwp->w_doto = -1;
								lineLength = lused(curwp->w_dotp);
								if(curwp->w_dotp == curbp->b_abotline)
									goto Fail;	// Hit end of buffer
							}
							continue;
						}

						if(c == strDelim)
							break;		// Maybe

					}
					StartedInString = 0;
					InString = 0;
					break;	// Reached end of string
				}

				// sdir == REVERSE
				while(1) {
					curwp->w_doto--;
					if(curwp->w_doto < 0) {
						// If the previous line ends with a backslash, continue there
						LINE*	prev = lback(curwp->w_dotp);

						if(prev == curbp->b_btopline ||
						   lused(prev) <= 0 ||
						   lgetc(prev, lused(prev) - 1) != '\\')
							goto LineEnd;
// Really ought to match pairs of backslash
						curwp->w_dotp = prev;
						curwp->w_doto = lused(prev) - 1;
					}
					c = lgetc(curwp->w_dotp, curwp->w_doto);

					if(StartedInString) {
						// Match brackets within the string that "." started in
						if(c == ch)
							++count;
						if(c == ofence)
							--count;

						/* Got to a match ? */
						if(count == 0)
							return(TRUE);
					}

					// If not \ or quote then the current character is not interesting:
					if(c == '\\') {
						// Eliminate a string of backslashes
						while(curwp->w_doto > 1 && lgetc(curwp->w_dotp, curwp->w_doto - 1) == '\\') {
							curwp->w_doto--;
						}
						continue;
					}

					// Possible end of string ?
					if(c != strDelim)
						continue;

					/* Found a quote.
					 * Eliminate pairs of preceeding backslashes
					 * If what is left before the quote is a single backslash
					 * then the quote does not end the string
					 */
					while(curwp->w_doto > 2 &&
					      lgetc(curwp->w_dotp, curwp->w_doto - 1) == '\\' &&
					      lgetc(curwp->w_dotp, curwp->w_doto - 2) == '\\')
						curwp->w_doto -= 2;	// Backup over backslash pair

					// If we have a backslash left then it quotes/hides the end of string quote
					if(curwp->w_doto > 1 &&
					   lgetc(curwp->w_dotp, curwp->w_doto - 1) == '\\') {
						curwp->w_doto--;
						continue;
					}

					/* Reached the quote that we got to earlier without
					 * finding a match - fail
					 */
					if(StringStart == curwp->w_doto && StartedInString)
						goto Fail;

					StartedInString = 0;
					InString = 0;
					break;	// End of string
				}
				break;

			// Used to look at backslash, not needed now that the string code eats to end of string


			case '/':
				if( ! c_php_mode)
					break;

				/* End of C style comment ?
				 * Note how it is end of comment independently of direction.
				 */
				if(!InString && LastCh == '*') {
					// If we did not know that we were in a comment ... then
					// we prob started in one - don't look beyond here
					if( !InComment)
						goto Fail;

					InComment = 0;
				}

				/* C++ style comment ?
				 * Ignore the rest of the line.
				 */
				if(!InComment && !InString &&
				   sdir == FORWARD && LastCh == '/') {
					if(ThisLineOnly)
						goto Fail;

					curwp->w_doto = lineLength;
					LastCh = '\n';
					goto NextCh;
				}
				break;

			case '#':	// Start of #directive or # comment ?
				if(ThisLineOnly)
					goto Fail;

				if(sdir == REVERSE || (!white || InComment || InString))
					break;

				// Jump to end of line
				curwp->w_doto = lineLength;
				LastCh = '\n';
				goto NextCh;

			case '*':	/* Start of comment ? */
				if( ! c_php_mode)
					break;

				if(LastCh == '/')
					InComment = 1;
				break;

			case ' ': case '\t':
				wasSpace = 1;
				break;

			case '\n':	/* Strings/Chars not over newline */
					/* If going backwards defer the test
					 * 'till we get to the line.
					 */
LineEnd:			c = '\n';
				if(ThisLineOnly)
					goto Fail;

				white = 1;
				wasSpace = 1;
				StartedInString = 0;	// Do not match brackets within strings any more

				if(InComment)
					break;

				// Going backwards, we continue if we find ourselves
				// in a string or char

				// Processing for reverse search
				// Find C++ comment - put "." just before it
				if(sdir == REVERSE)
					lineLength = nonCommentLength(curwp->w_dotp);

				break;
			}

			// If seen anything other than white space - clear white space flag
			if(!wasSpace)
				white = 0;
		}

		LastCh = c;

		/* More or less nested brackets ?
		 */
		// if(sdir == REVERSE) fprintf(stderr, "MATCH '%c' == '%c' InString=%d count=%d\n", c, ch, InString, count);
		if(StartedInComment || !InComment) {
			if(c == ch)
				++count;
			if(c == ofence)
				--count;
		}

		/* Got to a match ? */
		if(count == 0)
			return(TRUE);


NextCh:		/* Next character */

		/* Before we try to go anywhere, have we hit the end ?
		 * Return to where we started.
		 */
		if(boundry(curwp->w_dotp, curwp->w_doto, sdir)) {
			if(0) {	/* Dubious */
Fail:				FailAtDotp = curwp->w_dotp;
				FailAtDoto = curwp->w_doto;
			}

			curwp->w_dotp = oldlp;
			curwp->w_doto = oldoff;

			return(FALSE);
		}

		int	ChangedLine = 0;	// Move up/down in the lines below

		if(sdir == REVERSE) {
			if(--curwp->w_doto < lineStart) {
				curwp->w_dotp = lback(curwp->w_dotp);
				lineLength = curwp->w_doto = lused(curwp->w_dotp);
				lineStart = 0;
				ChangedLine = 1;
			}
		} else {
			if(++curwp->w_doto > lineLength) {
				curwp->w_dotp = lforw(curwp->w_dotp);
				lineLength = lused(curwp->w_dotp);
				lineStart = curwp->w_doto = 0;
				ChangedLine = 1;
			}
		}

		// Hit end of buffer ?
		if(curwp->w_dotp == curbp->b_abotline)
			goto Fail;	// Sometimes these things need to be done

		if(ChangedLine && cmtLen > 0) {
			/* Just started a new line.
			 * Do a quick scan to see if it starts with a deduced start-of-line comment.
			 */
			LineLn	pos = 0;

			while(pos < lused(curwp->w_dotp) - cmtLen) {
				// Skip white space:
				if(lgetc(curwp->w_dotp, pos) == ' ' ||
				   lgetc(curwp->w_dotp, pos) == CharTab) {
					pos++;
					continue;
				}

				/* Deduced comment ?
				 * Set such that the next char is \n
				 */
				if(memcmp(bufCmtStr, lgetcaddr(curwp->w_dotp, pos), (size_t)cmtLen) == 0 &&
				   (pos == lused(curwp->w_dotp) - cmtLen ||
				    (lgetc(curwp->w_dotp, pos + cmtLen) == ' ' ||
				     lgetc(curwp->w_dotp, pos + cmtLen) == CharTab)))
					curwp->w_doto = sdir == REVERSE ? -1 : lineLength;

				break;
			}
		}

	}
}

/* Attempt to find an error in a C program.
 * 
 * We do this by trying to paren match something unmatchable. This runs the
 * fmatch() routine 'till it either hits EOF or finds an unmatched quote.
 *
 * Leave the cursor there. Searches from the point & forward.
 * Crude but useful.
 *
 * The count is ignored.
 *
 * Command:  find-c-error
 * Bound to: M-}
 *
 * Return:
 *	TRUE  no errors seen
 *	FALSE some problem
 */
int	FindCError(int f, Cnt n)
{
	pnmark1();

	/* No point in testing the return value - it won't be able to
	 * match it.
	 */
	fmatch(-2, 0);

	/* Have we got a failure point - ie didn't just hit EOF ? */
	if(FailAtDotp) {
		TTbeep();	/* Found an error */
		curwp->w_dotp = FailAtDotp;
		curwp->w_doto = FailAtDoto;
		return(FALSE);
	}

	mlwrite("%INo errors found");

	return(TRUE);
}

/* Show where the matching brace is. Move the cursor there, sleep a bit &
 * move it back to where it was.
 *
 * This is called just after the character has been inserted, so the point is
 * currently just after it.
 * This is not called in replay mode.
 * If the matching fence is off-screen, the screen is redrawn.
 * Don't use getfence to get back to where we are now - it isn't perfect.
 * Sleep tpause ($tpause) seconds.
 *
 * Remove any WindowMargin - this might help avoid a screen refresh due to
 * left/right scroll if the fence is near the screen edge. The refresh will
 * still happen if the fence is off the screen.
 */
void	ShowMatchingBrace(void)
{
	NMARK*	oldLine;	/* original line */
	int	oldMargin;	/* Preserve WindowMargin so can restore */

	/* Use numeric mark 11 as that will have its value
	 * restored if checkage() kicks in when we do the update()
	 * below - which could trash the buffer.
	 */
	oldLine = &curwp->w_nmarks[10];
	l_changeNmark(oldLine, curwp->w_dotp, curwp->w_doto);	// Save the current place

	moveBack(1);

	if(getfence(FALSE, 0)) {
		if(tpause > 0
#if	TYPEAH
				/* If there is something waiting to be read don't pause,
				 * it just causes annoyance. This will often be the case
				 * when text is pasted using the mouse or similar.
				 */
				&& typahead(TRUE) == FALSE
#endif
				) {
			oldMargin = WindowMargin;
			WindowMargin = 0;
			update(FALSE);
			DoSleep((unsigned)tpause);
			WindowMargin = oldMargin;

		}
	}
	else
		TTbeep();

	// Put "." back where it was
	curwp->w_dotp = oldLine->n_pline;
	curwp->w_doto = oldLine->n_oline;
	l_changeNmark(oldLine, NULL, 0);

	curwp->w_flag |= WFMOVE;

	/* No need to update, it'll be done sometime before the user
	 * types something again.
	 */
}
#endif	// CFENCE


/* Move to the beginning or end of a C function by searching backwards
 * for "<NL>{<NL>" or forwards for "<NL>}<NL>"
 * Used in getfence()
 *
 * Return: scanner() TODO
 */
int	movefunc(int f, Cnt n)
{
	register int	dirn = (f && n > 0) ? FORWARD : REVERSE;
	register int	status;

	tcpys(playText, dirn == FORWARD ? "\n}\n" : "\n{\n");

	if((status = scanner(playText, dirn, PTBEG)) != 0)
		/*
		 * Looks a bit silly if we end up on the newline before.
		 */
		moveForw(1);

	return(status);
}

/* Ask for and insert n copies of the string into the current buffer at ".".
 *
 * Default count 1. Must be >= 0.
 *
 * Command:  insert-string
 * Bound to: unbound
 *
 * Return: TODO
 */
/* ARGSUSED */
int	istring(int f, Cnt n)
{
	register int status;	/* status return code */

	/* ask for string to insert */
	status = mlprompt("String to insert<META>: ", MLSTRING | MLEOLESC, bmlValHistp);
	if(status != TRUE)
		return(status);

	/* insert it */
	while(--n >= 0 && ((status = linTEXT(&ltxt(bmlValHistp->b_dotp)))) != 0)
		;

	return(status);
}

/* Ask for and overwite a string into the current buffer at ".".
 *
 * Default count 1. Must be >= 0.
 *
 * Command:  overwrite-string
 * Bound to: unbound
 *
 * Return: TODO
 */
int	ovstring(int f, Cnt n)
{
	register int status;	/* status return code */

	/* ask for string to insert */
	if((status = mlprompt("String to overwrite<META>: ", MLSTRING | MLEOLESC, bmlValHistp))!= TRUE)
		return(status);

	/* insert it */
	while((--n >= 0) && ((status = lover(bmlValHistp->b_dotp))) == TRUE)
		;

	return(status);
}

/* The cursor is moved to a matching fence - to match the character under ".".
 *
 * If the character isn't a brace/bracket but we are in C mode, move to the
 * start/end of the function (start if no or -ve count, end if +ve count).
 * The count is ignored unless we move to the start/end function.
 *
 * The default count is not set.
 *
 * Command:  goto-matching-fence
 * Bound to: M-C-f   M-C-i
 *
 * Return: TODO
 */
int	getfence(int f, Cnt n)
{
	BYTE	ch;	/* fence type to match against */
	int	ret;
	CHAR*	msg;	// Message on error

	pnmark0();

	/* get the current character */
	ch = (BYTE)(curwp->w_doto < lused(curwp->w_dotp) ? lgetc(curwp->w_dotp, curwp->w_doto) : '\n');

	/* Check if we are on a bracket that we can match to.
	 * If not, abort unless C mode.
	 */
	if(!strchr("{}[]()<>", ch)) {
		/* if we're not in CMODE or PHP then barf */
		if(!(curbp->b_mode & (MDCMOD | MDPHP)))
			return(FALSE);

		/* else go to the beginning/end of the function */
		ret = movefunc(f, n);
		msg = (f && n > 0) ? "%EEnd of function not found" : "%EStart of function not found";
	} else {
		ret = fmatch(ch, 0);
		msg = "%EMatch to %c not found";
	}

	/* If we moved - rename mark-0 to mark-1
	 * otherwise say what went wrong.
	 */
	if(ret == TRUE)
		pushNmark();
	else
		mlwrite(msg, ch);

	return(ret);
}

/* Functions for moving the cursor to a chosen character on the current line,
 * to the right or left of the current cursor position.
 *
 * two user callable functions are provided:
 *	cscan()		look for the current character.
 *	getcscan()	get a character and call cscan() to look for it.
 *
 * the non-user-callable function doscan() does the work.
 *
 * Return TRUE if the character is found.
 */

static	UniChar	scan_char = -1;	// Invalid char to start. Allow for 0 being valid - is can hunt for U+0 ('\0')
static	int	scan_dir  = 1;

int	doscan(void)
{
	UniChar	c;
	LineLn	col    = curwp->w_doto;		/* start column */
	LineLn	length = lused(curwp->w_dotp);	/* line length */
	int	utf8_line = lisUtf8(curwp->w_dotp);

	c = -1;	// In case we don't actually start scanning - as already at a line end

	// col is a byte offset. Same as character offset with non-multibyte character

	if(scan_dir > 0) {
		/*
		 * Search to the right of the current cursor position.
		 * The end point is the end of the line.
		 */
		if(utf8_line) {
			BYTE*	ub = lgetcaddr(curwp->w_dotp, col);

			while(c != scan_char && col < length - 1) {
				col += CHAR_LEN(*ub);
				ub = lgetcaddr(curwp->w_dotp, col);
				c = utf8_decode(ub);
			}
		} else {
			while(c != scan_char && col < length - 1)
				c = lgetc(curwp->w_dotp, ++col);

		}
	} else {
		/*
		 * Search to the left of the current cursor position.
		 * The end point is the beginning of the line.
		 */
		if(utf8_line) {
			BYTE*	ub = lgetcaddr(curwp->w_dotp, col);

			/* Scanning back over UTF8 is more complicated.
			 * Need to keep going over continuation bytes
			 */
			while(c != scan_char && ub > lgetcaddr(curwp->w_dotp, 0)) {
				ub--;
				UTF8_BACK_FIND_START(ub);

				c = utf8_decode(ub);
			}
			col = ub - lgetcaddr(curwp->w_dotp, 0);
		} else {
			while(c != scan_char && col > 0)
				c = lgetc(curwp->w_dotp, --col);
		}
	}

	if(c != scan_char) {
		TTbeep();
		return(FALSE);
	}

	curwp->w_doto = col;
	return(TRUE);
}

/* Search for a previously set character, forwards or backwards (use doscan()
 * to do the work).
 *
 * Moan if there is no character to search for.
 * Reverse the remembered direction if we are called with an argument.
 *
 * If the count is set reverse the (remembered) direction of scan.
 *
 * Command:  goto-character
 * Bound to: C-^
 *
 * Return TRUE if the character is found.
 */
/*ARGSUSED*/
int	cscan(int f, Cnt n)
{
	if(scan_char == -1) {
		mlwrite("%WNo character");
		return(FALSE);
	}

	if(f)
		scan_dir = -scan_dir;

	return(doscan());
}


/* Get a charcter from the keyboard and use cscan to move forwards to it.
 *
 * If the count is set: scan backwards.
 * the default count is not set.
 *
 * Remember the direction of scan for goto-character (cscan())
 *
 * Command:  get-and-goto-character
 * Bound to: C-x C-^
 *
 * Return TRUE if the character is found.
 */
int	getcscan(int f, Cnt n)
{
	UniChar	c;

	if((c = mlGetKey(MGC_MAP | MGC_LEAVEP | MGC_MACARG, NULL, "Press character to scan for: ")) == abortc) {
		mlwrite(MsgAborted);
		TTbeep();
		return(FALSE);
	}

	mlerase();

	/* This used to be a char, now long so that it can hold a Unicode point */
	scan_char = c;

	/* Set the direction, ie forwards is default, if arg then go back.
	 */
	if(f)
		scan_dir = -1;
	else
		scan_dir = 1;

	return(cscan(FALSE, n));
}

/* Place an alpha mark at ".".
 *
 * This involves searching down the mark list and tacking a new mark list
 * element on the end.
 *
 * The argument is ignored.
 *
 * Command:  place-alpha-mark
 * Bound to: C-x C-a
 *
 * Return: TODO
 */
int	pamark(int f, Cnt n)
{
	register UniChar	c;
	register AMARK*	p    = curbp->b_amark;
	register AMARK*	last = curbp->b_amark;

	if((c = mlGetKey(MGC_MAP | MGC_MACARG, NULL, "place mark: ")) == abortc)
		return(ABORT);

	/* Check that its an alphabetic character */
	if(!Isletter(c)) {
		mlwrite("%E""%BInvalid mark name");
		return(FALSE);
	}

	/* convert to lowercase */
	if(Isupper(c))
		c = tolower(c);

	while(p != (AMARK*)NULL && p->m_name != (char)c) {
		last = p;
		p = p->m_next;
	}

	if(p != (AMARK*) NULL) {
		/* Unset that this line is marked - take care, it may be marked
		 * more than once, so scan to check.
		 */
		AMARK*	ap;
		LINE*	lp = p->m_pline;
		char*	what = "deleted";

		p->m_pline = (LINE*)NULL;
		if(lp != (LINE*)NULL) {
			lunsetMark(lp);
			what = "existing";
			for(ap = curbp->b_amark; ap != (AMARK*)NULL; ap = ap->m_next)
				if(ap->m_pline == lp)
					lsetMark(lp);
			}

		mlwrite("%IOverwriting %s mark %c", what, c);
	} else {
		/* Make new & put at end of chain */
		if((p = (AMARK*) malloc(sizeof(AMARK))) == (AMARK*) NULL)
			return(FALSE);

		p->m_next = (AMARK*) NULL;

		if(curbp->b_amark == (AMARK*) NULL)
			curbp->b_amark = p;
		else
			last->m_next   = p;

		mlwrite("%ICreating mark %c", c);
	}

	p->m_name  = (BYTE)c;
	p->m_pline = curwp->w_dotp;
	p->m_oline = curwp->w_doto;
	lsetMark(curwp->w_dotp);		/* mark the line as marked */

	return(TRUE);
}

/* Return to an alphabetic mark.
 *
 * Ie make the current line and the current offset equal to the line and offset
 * stored in the mark element with the required name.
 *
 * Alpha marks are specific to the buffer not the window.
 * One window could have HEX option; this could be at undesirable positions
 * in a non HEX UTF8 window. Check for this and force doto to the start of
 * the character.
 *
 * If a digit is entered, return to that numeric mark.
 * Numeric marks are specific to the window and not the buffer.
 *
 * The numeric mark returned to is remembered (even if it did not exist).
 * When the user enters
 * '+'	return to the remembered mark + 1
 * '-'	return to the remembered mark - 1
 * This will wrap to keep within 0..9
 *
 * The count is ignored.
 *
 * Command:  return-alpha-mark
 * Bound to: C-x a
 *
 * Return TODO
 *	FALSE The mark is not in the narrow
 */
int	retamark(int f, Cnt n)
{
	AMARK*	p = curbp->b_amark;
	int	c;
	BYTE	allmarks[2 * 26 + 1];			/* record of the marks	*/
	BYTE	markName[2];				// So can show mark name
	int	i = 0;
	static	char	lastMark = '0';			// Last numeric mark returned to

	if((c = mlGetKey(MGC_MAP | MGC_LEAVEP | MGC_MACARG, NULL, "goto mark: ")) == abortc)
		return(ABORT);

	// Deal with mark number '-' or '+':
	if(c == '-') {
		if(--lastMark < '0')
			lastMark = '9';
		c = lastMark;
	}

	if(c == '+') {
		if(++lastMark > '9')
			lastMark = '0';
		c = lastMark;
	}

	/* Check that its an alphanumeric character.
	 * Also accept '?' - which will not be found but generates a list of marks.
	 */
	if(!Isalnum(c) && c != '?') {
		mlwrite("%E""%BInvalid mark name '%c'", c);
		return(FALSE);
	}

	if(Isdigit(c)) {
		/* Numeric mark.
		 * First find where it is referenced in the window's numeric mark array:
		 */
		int	mNum = (curwp->w_nmarkOff - (c - '0') + 10) % 10;
		NMARK*	nm = &curwp->w_nmarks[mNum];

		lastMark = (char)c;		// Remember last - might have been entered at prompt

		if(nm->n_pline) {
			/* Move to the numbered mark, leave mark 0 at the old location.
			 * Do in 2 steps in case we are returning to mark 0.
			 */
			LINE*	nl = nm->n_pline;
			LineLn	no = nm->n_oline;
			int	dotOnMark = 0;		// True if "." on a numberic mark LINE

			// Are we allowed to go there ?
			if( ! isInNarrow(nl, "%WMark %c is not in narrow", c))
				return(FALSE);

			/* Do not set mark 0 if "." is on a numeric marked LINE.
			 * Do not check the offset in the line, just the line.
			 *
			 * This means that you can bounce around exploring the
			 * numeric marks without loosing the mark 0 set when
			 * you first started this.
			 *
			 * Only look at the 10 numeric marks.
			 */
			for(nm = curwp->w_nmarks; nm < curwp->w_nmarks + 10; nm++)
				if(curwp->w_dotp == nm->n_pline)
					dotOnMark = 1;
			if( !dotOnMark)
				pnmark0();

			curwp->w_dotp = nl;
			curwp->w_doto = no;
		} else {
			mlwrite("%WNumeric mark %c does not exist", c);
			return(FALSE);
		}
	}
	else {
		/* Alpha mark.
		 * Convert to lowercase
		 */
		if(Isupper(c))
			c = tolower(c);

		while(p != (AMARK*)NULL && p->m_name != (char)c) {
			if(p->m_pline) {	/* Only show where the line exists */
				allmarks[i++] = p->m_name;
				allmarks[i++] = ' ';
			}
			p = p->m_next;
		}
		allmarks[i] = '\0';

		if(p == (AMARK*)NULL) {
			// Did not find the mark requested

			if(allmarks[0] == '\0')
				mlwrite("%WNo marks in the buffer");
			else
				// But there are marks, show the list of them:
				if(c == '?')
					mlwrite("%WMarks in this buffer: %s", allmarks);
				else
					mlwrite("%WMark %c does not exist. Marks in this buffer: %s", c, allmarks);
			return(FALSE);
		}

		if(p->m_pline == (LINE*)NULL) {
			mlwrite("%EReferenced line has been deleted");
			return(FALSE);
		}

		// Are we allowed to go there ?
		if( ! isInNarrow(p->m_pline, "%WMark %c is not in narrow", c))
			return(FALSE);

		/* Do the business.
		 * First push mark 1 here in case a return is wanted.
		 */
		pnmark1();
		curwp->w_dotp  = p->m_pline;
		curwp->w_doto  = p->m_oline;
	}

	curwp->w_flag |= WFMOVE;

	// UTF8 buffer and non hex window: ensure at start of the character.
	// Being at the virtual \n is OK.
	// This should be entirely redundant if everything else works correctly.
	if((curbp->b_mode & MDUTF8) != 0 && (curwp->w_opts & WO_HEX) == 0 && lused(curwp->w_dotp) != curwp->w_doto)
		UTF8_BACK_FIND_START_NX(lgetc(curwp->w_dotp, curwp->w_doto), curwp->w_doto);

	/* Show the mark name just entered and returned to.
	 * This leaves an extra space before it - fix if *really* bored.
	 *
	 * Don't do this if running in a macro - it leaves numbers on the message line.
	 */
	if(clexec == FALSE) {
		markName[0] = (BYTE)c;
		markName[1] = '\0';
		mlSappend(markName);
	}

	return(TRUE);
}

/* Place numeric mark 0 at "." in the current window.
 * 
 * Replace whatever is currently mark 0 - if any.
 */
void	pnmark0(void)
{
	NMARK*	nm = &curwp->w_nmarks[(int)(curwp->w_nmarkOff)];

	l_changeNmark(nm, curwp->w_dotp, curwp->w_doto);
}

/* Return "." to mark 0 - if it is set.
 *
 * swap	If true place mark 0 where "." is before it was moved.
 *	Ie swap places of "." and mark 0
 *
 * Return:
 *	TRUE  Success - "." at where mark 0 was
 *	FALSE Failed - mark 0 is not set
 */
int	retmark0(int swap)
{
	NMARK*	nm = &curwp->w_nmarks[(int)(curwp->w_nmarkOff)];
	LINE*	oldDotp = curwp->w_dotp;
	LineLn	oldDoto = curwp->w_doto;

	if(nm->n_pline == (LINE*)NULL)
		return(FALSE);

	// nm->n_pline is already L_NMARKed - no need to set again
	curwp->w_dotp = nm->n_pline;
	curwp->w_doto = nm->n_oline;

	if(swap) {
		nm->n_pline = oldDotp;
		nm->n_oline = oldDoto;
		lsetNMark(nm->n_pline);
		l_checkNmark(curwp->w_dotp);	// Might no longer be L_NMARKed
	}

	return(TRUE);
}

/* Rename numeric mark 0 as mark 1.
 * All the others are renamed, 1 to 2, ...
 * Numeric mark 9 is discarded.
 *
 * BEWARE: w_nmarkOff is incremented, so the numeric marks are actually
 * stored backwards in w_nmarks[]. Ie mark3 mark2 mark1 mark0 mark9 ...
 */
void	pushNmark(void)
{
	NMARK*	nm;
	LINE*	line9;
	char	nmarkOff = curwp->w_nmarkOff;

	// Roll on one:
	curwp->w_nmarkOff = ++nmarkOff % 10;

	// The current numeric 0 was the old mark 9
	// Set that it is no longer referenced here
	nm = &curwp->w_nmarks[(int)(curwp->w_nmarkOff)];
	line9 = nm->n_pline;
	nm->n_pline = (LINE*)NULL;

	if(line9 != (LINE*)NULL)
		l_checkNmark(line9);
}

/* Place numeric mark 1 at ".", rotate the others.
 * Mark 0 is lost, old mark 9 is lost.
 */
void	pnmark1(void)
{
	pnmark0();
	pushNmark();
}

/* Show alpha and numeric marks that are in parts of buffers that
 * are currently visible.
 *
 * There may be more than one mark at any one position.
 * Alpha marks will be shown in preferance to numeric marks.
 * The alpha mark that get shown is which ever one gets shown last.
 * Lower numbered numeric marks are deemed more interesting than higher numbered ones.
 *
 * Prompt to hide the marks, unless zero marks are shown (non visible),
 * do not do this if the count is set.
 *
 * Do not show the internal numeric marks.
 *
 * The default count is unset.
 *
 * Command:  show-marks
 * Bound to: C-x M
 *
 * Return TRUE.
 */
int	showMarks(int f, Cnt n)
{
	Window*	wp;
	int	shown = 0;	// # marks shown

	/* displayChar() does an ephemeral update, so ensure that all
	 * is up to date else the ephmeral ones will get lost.
	 */
	update(TRUE);

	// All windows:
	for(wp = wheadp; wp; wp = wp->w_wndp) {
		AMARK*	amp = wp->w_bufp->b_amark;
		NMARK*	nmp = wp->w_nmarks + wp->w_nmarkOff + 1;	// Set at mark 9
		int	m;		// Mark #

		// All numeric marks. '0' is last as it is most interesting
		for(m = 9; m >= 0; m--) {
			// nmp might be out of bounds on entry - fixed here
			if(nmp > wp->w_nmarks + 9)	// Wrap when hit the end
				nmp = wp->w_nmarks;

			if(nmp->n_pline != (LINE*)NULL &&
			   placeChar(wp, nmp->n_pline, nmp->n_oline, (char)(m + '0'), PEN_MARK))
				shown++;

			nmp++;			// Previous mark (++ as stored backwards)
		}

		// All alpha marks
		for( ; amp != (AMARK*)NULL; amp = amp->m_next)
			if(placeChar(wp, amp->m_pline, amp->m_oline, (char)amp->m_name, PEN_MARK))
				shown++;
	}

	// Unless we have an argument, prompt to undo what we just did
	if( ! f) {
		if(shown == 0) {
			mlwrite("%INo marks shown");
		} else {
			mlGetKey(MGC_NONE, NULL, "Hide the %d marks shown: ", shown);
			upmode((BUFFER*)NULL, WFHARD);
		}
	}

	return(TRUE);
}

#if	SHOWMODREL		// Show module release numbers
#define		PADLEN		15

/* Print out the version strings for each module.
 *
 * The default count is not set.
 *
 * Command:  show-module-release
 * Bound to: unbound
 *
 * Return: TODO
 */
int	showmods(int f, Cnt n)
{
	int		i;
	long		llen;
	char		line[128];
	static	BUFFER*	bmodp = (BUFFER*) NULL;

	if(bmodp == (BUFFER*) NULL)
		bmodp = bfind("[Versions]", BFNCREATE|BFNACTIVE, BFINVS);

	bmodp->b_flag &= ~BFCHG;		 /* Don't complain!		*/
	bmodp->b_mode = 0;			 /* Clear the modes		*/
	if((i = bclear(bmodp)) != TRUE)		 /* Blow old text away		*/
		return(i);

	bmodp->b_fname[0] = '\0';
	bmodp->b_fnameClean[0] = '\0';

	for(i = 0; i < nmodmapent; i++) {

		strcpy(line, modmap[i].mod_name);/* copy in module name		*/
		llen = strlen(modmap[i].mod_name);

		for(/*void*/; llen < PADLEN; llen++)
			line[llen] = ' ';	 /* pad it out			*/
		line[llen] = '\0';		 /* null terminate		*/

		strcat(line, modmap[i].mod_vers);/* copy in module version #	*/

		if(addline(line, bmodp, 1) == FALSE)/* add the line		*/
			return(FALSE);
	}

	return(showlist(bmodp));		 /* show it			*/
}
#endif	// SHOWMODREL

/* Reverse overlapping copy.
 *
 * Ie use this when a byte is to be inserted in the middle of a string.
 *
 * This differs from memmove() in that it can be used to propagate a
 * byte that is set.
 *
 * To be clear: the characters end up in the origial order.
 * It is the order in which they are copied that is reversed.
 */ 
void	memrcpy(char* to, const char* from, size_t len)
{
	to += len;
	from += len;

	while(len-- > 0)	// Beware: len is unsigned, cannot go: --len>= 0
		*--to = *--from;
}

/* Overlapped memory copyer.
 *
 * memcpy isn't guaranteed to work with overlapped copies.
 *
 * This differs from memmove() in that it can be used to propagate a
 * byte that is set.
 */ 
void	memocpy(char* to, const char* from, size_t len)
{
	while(len-- > 0)	// Beware: len is unsigned, cannot go: --len>= 0
		*to++ = *from++;
}

/* Reverse the order of len bytes starting at mem.
 *
 * This is BYTE reverse, not UTF-8 characters.
 *
 * Return mem.
 */
BYTE*	memrev(BYTE* mem, size_t len)
{
	BYTE*	s = mem;
	BYTE*	e = mem + len - 1;
	BYTE	ch;

	// Reverse the bytes:
	while(e > s) {
		ch   = *e;
		*e-- = *s;
		*s++ = ch;
	}

	return(mem);
}

/* Reverse the order of NUL terminated string str.
 *
 * This is BYTE reverse, not UTF-8 characters.
 *
 * Return str.
 */
BYTE*	strrev(BYTE* str)
{
	return(memrev(str, strlen(str)));
}

/* rvstrcpy -- Reverse string copy.
 *
 * Ie copy to rvstr a string with the characters in reverse order.
 *
 * This does not check the size of what it is copying into
 *
 * This does the same as:
 *	strcpy(rvstr, str);
 *	strrev(rvstr);
 *
 * Should use: utf8_strrev()
 */
void	rvstrcpy(BYTE* rvstr, const BYTE* str)
{
	size_t	i;

	str += (i = strlen(str));

	while(i-- > 0)
		*rvstr++ = *--str;

	*rvstr = '\0';
}

/* Overlapped string copy.
 *
 * Result is NUL terminated.
 */
char*	strocpy(char* to, const char* from)
{
	size_t	len = strlen(from);
	memmove(to, from, len);
	to[len] = '\0';

	return(to);
}

/* Insert the string at from into to.
 *
 * Do not copy the NUL byte.
 *
 * Return to.
 */
char*	strinsstr(char* to, const char* from)
{
	return(memcpy(to, from, strlen(from)));
}

/* Make a copy of 'string' somewhere that we've malloc'd.
 *
 * NO error checking.
 *
 * Return: a pointer to the mallocd string.
 *	   NULL on out of memory.
 */
BYTE*	strmake(const BYTE* string)
{
	BYTE*	p;

	if((p = malloc(strlen(string) + 1)) != (BYTE*) NULL)
		strcpy(p, string);
	return(p);
}

/* Strmake but aborts if OOM
 */
BYTE*	strmake_t(const BYTE* string)
{
	BYTE*	s;
	if((s = strmake(string)))
		return(s);
	Die(MsgOutOfMemoryAlloc, "string");
}

/* Copy a string and store at *str.
 *
 * If *str contains something - free it *after* - in case: *str == string
 * Die() on out of memory.
 * 
 * Return what is allocated.
 */
BYTE*	strfmake(BYTE** str, const BYTE* string)
{
	BYTE*	oldstr = *str;

	*str = strmake_t(string);

	if(oldstr)
		free((void*)oldstr);

	return(*str);
}

/* Truncate a string so that it is at most len bytes long.
 *
 * This means that the buffer must be at least len+1 - for the \0
 * If str is NULL do nothing.
 */
void	strtrunc(BYTE* str, int len)
{
	if(str && (int)strlen(str) >= len)
		str[len] = '\0';
}

/* Return a malloced copy of string, NUL terminated.
 *
 * string is made UTF8 safe.
 * Any bad characters are replaced by UTF8_REPLACEMENT.
 *
 * Return NULL on out of memory.
 */
BYTE*	utf8_strmake(const BYTE* string)
{
	BYTE*	p;

	if(tcpys(strmakeTmp, string) == FALSE)
		return(NULL);

	if((p = malloc((size_t)tused(strmakeTmp) + 1)) == (BYTE*) NULL)
		return(NULL);

	memcpy(p, ttext(strmakeTmp), (size_t)tused(strmakeTmp));
	p[tused(strmakeTmp)] = '\0';

	return(p);
}

/* utf8_strmake but aborts if OOM
 */
BYTE*	utf8_strmake_t(const BYTE* string)
{
	BYTE*	s;

	if((s = utf8_strmake(string)))
		return(s);

	Die(MsgOutOfMemoryAlloc, "string");
}

/* Copy a string and store at *str.
 *
 * If *str contains something - free it *after* - in case: *str == string
 * Return what is allocated.
 * Die() on out of memory.
 */
BYTE*	utf8_strfmake(BYTE** str, const BYTE* string)
{
	BYTE*	oldstr = *str;

	*str = utf8_strmake_t(string);

	if(oldstr)
		free((void*)oldstr);

	return(*str);
}

/* Compare n bytes at a & b in a case insensitive manner.
 *
 * Case means byte (== ASCII) comparisons
 *
 * Return 0 if the two are the same
 *	  b - the BYTE difference if they differ.
 *	  The difference is the difference before they are case folded for comparison.
 */
int	memcasecmp(void* a, void* b, size_t n)
{
	for( ; n-- > 0; a = (void*)((char*)a + 1), b = (void*)((char*)b + 1)) {
		BYTE	ca = (BYTE)tolower(*(BYTE*)a);
		BYTE	cb = (BYTE)tolower(*(BYTE*)b);
		if(ca != cb)
			return(*(BYTE*)a - *(BYTE*)b);
	}

	return(0);
}

/* Case-independent string comparison.
 *
 * Return 0 if the strings are the same
 *	  the difference between the first characters that are different.
 */
int	cstrcmp(const BYTE* bfile, const BYTE* name)
{
	for(;; bfile++, name++)
		if(*bfile == '\0' || *name == '\0' ||
				tolower(*bfile) != tolower(*name))
			return(*bfile - *name);
}

/* bytecopy:	copy a string...with length restrictions
 *		always NUL terminate
 *
 * dst		destination of copied string
 * src		source
 * maxlen	maximum length - dst must be maxlen + 1
 *
 * Return dst
 */
char*	bytecopy(char* dst, const char* src, size_t maxlen)
{
	char*	dptr;			/* ptr into dst */

	dptr = dst;
	while((maxlen-- > 0) && *src)
		*dptr++ = *src++;
	*dptr = '\0';

	return(dst);
}

#if	0	// This is in the GNU C library
/* Find needle within haystack.
 * AKA strstr() but NUL safe.
 *
 * haystack	Where to start searching
 * haylen	How many bytes at haystack
 * needle	Start of search bytes
 * nlen		How many search bytes
 *
 * The search is exact bytes: no case folding, no recognition of UTF-8
 *
 * Return:
 *	Address of first occurence of needle in haystack
 *	NULL needle was not found
 */
BYTE*	memmem(BYTE* haystack, LineLn haylen, BYTE* needle, LineLn needlelen)
{
	for( ; haylen > 0; haystack++, haylen--) {
		if(*haystack != *needle)
			continue;
		if(haylen < needlelen)
			break;
		if(memcmp(haystack, needle, (size_t)needlelen) == 0)
			return(haystack);
	}

	return((BYTE*)NULL);
}
#endif

/*****		Compiler specific Library functions	****/

#if	MWC86 && MSDOS
int	movmem(source, dest, size)
	char*	source;	/* mem location to move memory from */
	char*	dest;	/* memory location to move text to */
	int	size;	/* number of bytes to move */
{
	register int i;

	for(i=0; i < size; i++)
		*dest++ = *source++;
}
#endif

#if	(MSDOS || AMIGA || ST520) && !MSC
/*	strncpy:	copy a string...with length restrictions
 *			ALWAYS null terminate
 */
char*	strncpy(dst, src, maxlen)
	char*	dst;	/* destination of copied string */
	char*	src;	/* source */
	int	maxlen;	/* maximum length */
{
	char *dptr;	/* ptr into dst */

	dptr = dst;
	while (*src && (maxlen-- > 0))
		*dptr++ = *src++;
	*dptr = 0;

	return(dst);
}
#endif

/* Look to see if name has one of the suffixes
 *
 * Return TRUE if a suffix at the end of name is found in suffixes.
 */
int	hasSuffix(char* name, char* suffixes)
{
	char*	dot;
	size_t	sufLen;

	// printf("name='%s' suffixes='%s\n", name, suffixes);
	if((dot = strrchr(name, '.')) == NULL)
		return(FALSE);

	sufLen = strlen(dot);
	// printf("suffix='%s' len=%d\n", dot, (int)sufLen);

	while((suffixes = strchr(suffixes, '.')) != NULL) {
		char*	sEnd;
		size_t	sl;

		if((sEnd = strchr(suffixes, ' ')) != NULL)
			sl = (size_t)(sEnd - suffixes);
		else
			sl = strlen(suffixes);

		if(sl == sufLen && memcasecmp(suffixes, dot, sl) == 0)
			return(TRUE);

		suffixes += sl;
	}

	return(FALSE);
}

/* Quick, dirty, unbuffered - ring the bell - for debug use
 */
void	bel(void)
{
	char	BelStr[] = "\a";

	// Ignore RC, casting to void no longer works with GCC :-(
	IGNORE(write(1, BelStr, sizeof(BelStr) - 1))
}

/* Call this instead of sleep().
 * Does a few things that might be needed if the editor sleeps a bit.
 *
 * * Flush the terminal
 * * Prevent next typahead() optimiation
 */
void	DoSleep(unsigned int seconds)
{
	DoTypahead = 1;

	TTflush();

	sleep(seconds);
}

/* end */
