/*
 * ADDW SCCS: @(#)region.c	1.32 04/27/20 00:04:47
 */

/* The routines in this file
 * deal with the region, that magic space
 * between "." and mark. Some functions are
 * commands. Some functions are just for
 * internal use.
 */
#include	"estruct.h"
#include	"utf8.h"
#include	"edef.h"

#if	SHOWMODREL
char	regnvers[] = "@(#)region.c	1.32 04/27/20 00:04:47";
#endif

#if	MEGAMAX && ST520
overlay "region"
#endif

/* Kill the region.
 *
 * Ask "getregion" to figure out the bounds of the region. Move "." to the
 * start, and kill the characters.
 *
 * Any count is ignored.
 *
 * Command:  kill-region
 * Bound to: C-w
 *
 * Return:
 *	getregion()
 *	ldelete()
 */
int	killregion(int f, Cnt n)
{
	REGION		region;
	int		status;

	if((status = getregion(&region)) != TRUE)
		return(status);

	curwp->w_dotp = region.r_linep;
	curwp->w_doto = region.r_offset;
	status = ldelete(region.r_size, KL_KILL_APPEND);

	if(kill_lcount)
		mlwrite("%I""%D line%s killed", (long)kill_lcount,
						kill_lcount == 1 ? "" : "s");

	return(status);
}

/* Copy all of the characters in the region to the kill buffer. Don't move dot
 * at all. This is a bit like a kill region followed by a yank.
 *
 * If the region is empty - create an empty entry in the yank buffer.
 *
 * Copying a region's text will be seen by the user as a distinct, discrete act.
 * So ensure that the copied text goes into a separate yank item in the kill
 * buffer. To do this we clear the CFDELETE flag.
 *
 * Any count is ignored.
 *
 * Command:  copy-region
 * Bound to: M-w
 *
 * Return true if all OK.
 *	TODO
 */
int	copyregion(int f, Cnt n)
{
	register LINE*	linep;
	register LineLn	loffs;
	register int	s;
	register BufLns	line_count;			/* count lines copied	*/
	REGION		region;
	int		utf8 = (curbp->b_mode & MDUTF8) != 0;	// UTF8 enabled buffer ?

	line_count = 0;

	if((s = getregion(&region)) != TRUE)
		return(s);

	linep = region.r_linep;			/* Current line.	*/
	loffs = region.r_offset;		/* Current offset.	*/

	// Stop our killed text being merged with previously killed text:
	lastflag &= ~CFDELETE;
	thisflag &= ~CFDELETE;

	// Zero sized region: insert empty in the yank buffer
	if(region.r_size == 0) {
		s = kNewline();
		mlwrite("%IEmpty region copied");
		return(s);
	}

	while(region.r_size) {
		if (loffs == lused(linep)) {	/* End of line.		*/
			line_count++;
			if((s = kinsert(NewlineString, 0)) != TRUE)
				return (s);

			linep = lforw(linep);
			loffs = 0;
			region.r_size--;
		} else {			/* Middle of line.	*/
			if((s = kinsert(lgetcaddr(linep, loffs), 0)) != TRUE)
				return (s);

			loffs += utf8 ? CHAR_LEN(lgetc(linep, loffs)) : 1;
			region.r_size--;
		}
	}

	// Stop our killed text being merged with subsequent killed text:
	thisflag &= ~CFDELETE;

	if(line_count == 0)
		line_count++;	// Part of one line copied

	mlwrite("%I""%D line%s copied", (long)line_count, line_count == 1 ? "" : "s");

	return (TRUE);
}

/* Lower case region.
 *
 * Zap all of the upper case characters in the region to lower case. Use the
 * region code to set the limits. Scan the buffer, doing the changes. Call
 * "lchange" to ensure that redisplay is done in all buffers.
 *
 * The count ignored.
 *
 * Command:  case-region-lower
 * Bound to: C-x C-l
 *
 * Return: TODO
 */
int	lowerregion(int f, Cnt n)
{
	LINE*	linep;
	LineLn	loffs;
	int	c;
	int	s;
	int	didChange = 0;	// True if we trimmed at least one character
	REGION		region;

	if ((s = getregion(&region)) != TRUE)
		return (s);

	linep = region.r_linep;
	loffs = region.r_offset;
	while (region.r_size--) {

		if (loffs == lused(linep)) {
			linep = lforw(linep);
			loffs = 0;
		} else {
			if(lisUtf8(linep)) {
				// TODO lower non ascii chars
				c = lgetc(linep, loffs);
				if (Isupper(c)) {
					lputc(linep, loffs, (BYTE)(tolower(c)));
					didChange = 1;
				}
				loffs += CHAR_LEN(lgetc(linep, loffs));
			} else {
				c = lgetc(linep, loffs);
				if (Isupper(c)) {
					lputc(linep, loffs, (BYTE)(tolower(c)));
					didChange = 1;
				}
				++loffs;
			}
		}
	}

	if(didChange)
		lchange(WFHARD);

	return (TRUE);
}

/* Upper case region.
 *
 * Zap all of the lower case characters in the region to upper case. Use the
 * region code to set the limits. Scan the buffer, doing the changes. Call
 * "lchange" to ensure that redisplay is done in all buffers.
 *
 * The count ignored.
 * 
 * Command:  case-region-upper
 * Bound to: C-x C-u
 *
 * Return: TODO
 */
int	upperregion(int f, Cnt n)
{
	LINE*	linep;
	LineLn	loffs;
	int	c;
	int	s;
	int	didChange = 0;	// True if we trimmed at least one character
	REGION		region;

	if((s = getregion(&region)) != TRUE)
		return (s);

	linep = region.r_linep;
	loffs = region.r_offset;
	while (region.r_size--) {

		if(loffs == lused(linep)) {
			linep = lforw(linep);
			loffs = 0;
		} else {
			if(lisUtf8(linep)) {
				// TODO upper non ascii chars
				c = lgetc(linep, loffs);
				if (Islower(c)) {
					lputc(linep, loffs, (BYTE)(toupper(c)));
					didChange = 1;
				}
				loffs += CHAR_LEN(lgetc(linep, loffs));
			} else {
				c = lgetc(linep, loffs);
				if (Islower(c)) {
					lputc(linep, loffs, (BYTE)(toupper(c)));
					didChange = 1;
				}
				++loffs;
			}
		}
	}

	if(didChange)
		lchange(WFHARD);

	return(TRUE);
}

/* This routine figures out the bounds of the region in the current window, and
 * fills in the fields of the "REGION" structure pointed to by "rp".
 *
 * Because the dot and mark are usually very close together, we scan outward
 * from dot looking for mark. This should save time.
 *
 * Also validate that the region is legal:
 * - a mark exists
 * - in UTF8 mode that the ends are at character ends,
 *   not part way through - easy in window with HEX option.
 *
 * if rp is NULL - just do checking that getregion() would work.
 *
 * If the mark/"." that is the bottom line of the region, don't count
 * the bottom line.
 *
 * This looks in the narrow, not the buffer.
 *
 * Return
 *	TRUE
 *	FALSE  No mark in the window
 *	       "." or mark inside a multi byte character
 *	       Mark not found (should not happen)
 *
 * Callers of this routine should be prepared to get
 * an "ABORT" status; we might make this have the
 * conform thing later.
 */
int	getregion(REGION* rp)
{
	LINE*		flp;
	LINE*		blp;
	BufChrs		fsize;	// Size looking forwards - characters
	BufChrs		bsize;	// Size looking backwards - characters

	// sizes include 1 for the virtual \n at the end of each line

	if(curwp->w_markp == NULL) {
		mlwrite("%WNo mark set in this window");
		return (FALSE);
	}

	// Fail if "." is inside a multi byte character
	if( !CheckNotInUtf8())
		return(FALSE);

	/* Do not remove this test, neither try to fix by moving the mark
	 * (or "." above) - it is sometimes legitimate, eg: HEX option.
	 */
	if(lisUtf8(curwp->w_markp) && curwp->w_marko != lused(curwp->w_markp) &&
			BYTE_IS_MB_CONT(lgetc(curwp->w_markp, curwp->w_marko))) {
		mlwrite("%EMark is inside a multi-byte character");
		return(FALSE);
	}

	if(rp == NULL)		// Just checking if it would work
		return(TRUE);

	rp->r_lines = 1;	// Minimum of 1 line in the region

	// Mark & dot on the same line:
	if(curwp->w_dotp == curwp->w_markp) {
		rp->r_linep = curwp->w_dotp;
		rp->r_endLinep = curwp->w_dotp;
		// Set the offset to the one (dot/mark) that is nearer the start of line
		if(curwp->w_doto < curwp->w_marko) {
			rp->r_offset = curwp->w_doto;
			rp->r_endOffset = curwp->w_marko;
			rp->r_size = (BufChrs)(curwp->w_marko - curwp->w_doto);
		} else {
			rp->r_offset = curwp->w_marko;
			rp->r_endOffset = curwp->w_doto;
			rp->r_size = (BufChrs)(curwp->w_doto - curwp->w_marko);
		}

		// UTF8 mode: convert bytes to characters:
		if(lisUtf8(curwp->w_dotp))
			rp->r_size = (BufChrs)utf8_chrnlen(lgetcaddr(rp->r_linep, rp->r_offset), (size_t)rp->r_size);

		return(TRUE);
	}

	// Mark & dot on different lines:
	flp = blp = curwp->w_dotp;
	if(lisUtf8(flp)) {
		bsize = (BufChrs)utf8_chrnlen(lgetcaddr(blp, 0), (size_t)curwp->w_doto);
		fsize = (BufChrs)utf8_chrnlen(lgetcaddr(flp, curwp->w_doto), (size_t)(lused(flp) - curwp->w_doto)) + 1;
	} else {
		bsize = (BufChrs)curwp->w_doto;
		fsize = (BufChrs)(lused(flp) - curwp->w_doto + 1);
	}

	// Scan forwards & backwards from the dot at the same time - a line at a time each.
	while(flp != curbp->b_abotline || lback(blp) != curbp->b_btopline) {
		rp->r_lines++;

		if(flp != curbp->b_abotline) {
			flp = lforw(flp);
			if(flp == curwp->w_markp) {
				rp->r_linep = curwp->w_dotp;
				rp->r_offset = curwp->w_doto;
				rp->r_endLinep = curwp->w_markp;
				rp->r_endOffset = curwp->w_marko;
				rp->r_size = fsize + (lisUtf8(flp) ? (BufChrs)utf8_chrnlen(lgetcaddr(flp, 0), (size_t)curwp->w_marko) : (BufChrs)curwp->w_marko);

				// Mark is at start of bottom line, don't count the line:
				if(curwp->w_marko == 0)
					rp->r_lines--;

				return (TRUE);
			}
			fsize += (lisUtf8(flp) ? (BufChrs)utf8_chrnlen(lgetcaddr(flp, 0), (size_t)lused(flp)) : (BufChrs)lused(flp)) + 1;
		}
		if(lback(blp) != curbp->b_btopline) {
			blp = lback(blp);
			bsize += (lisUtf8(blp) ? (BufChrs)utf8_chrnlen(lgetcaddr(blp, 0), (size_t)lused(blp)) : (BufChrs)lused(blp)) + 1;
			if(blp == curwp->w_markp) {
				rp->r_linep = blp;
				rp->r_offset = curwp->w_marko;
				rp->r_endLinep = curwp->w_dotp;
				rp->r_endOffset = curwp->w_doto;
				rp->r_size = bsize - (lisUtf8(blp) ? (BufChrs)utf8_chrnlen(lgetcaddr(blp, 0), (size_t)curwp->w_marko) : (BufChrs)curwp->w_marko);

				// "." is at start of bottom line, don't count the line:
				if(curwp->w_doto == 0)
					rp->r_lines--;

				return (TRUE);
			}
		}
	}

	mlwrite("%EBug: lost mark");

	return (FALSE);
}

/* Narrow the buffer to lines in the current region.
 *
 * Set in BUFFER: b_btopline and b_abotline.
 * This is set for whole lines only:
 * b_btopline	The line before the first in the region.
 * b_abotline	The line after the last line in the region,
 *		if the region ends at the start of a line then that line is used.
 *
 * This succeeds even if the buffer was already narrowed, ie it
 * is made even narrower.
 *
 * Any count is ignored.
 *
 * Command:  narrow-to-region
 * Bound to: C-x <
 *
 * Return:
 *	TRUE  if the buffer is narrowed.
 *	FALSE buffer not narrowed as there is not a region
 *	      or the buffer is an internal/invisible buffer
 */
int	narrow(int f, Cnt n)
{
	int	s;
	REGION	region;

	if((curbp->b_flag & BFINVS) != 0) {
		mlwrite("%ECannot narrow this buffer");
		return(FALSE);
	}

	if((s = getregion(&region)) != TRUE)
		return(s);

	// Top line is simple: the line before the first line of region
	curbp->b_btopline = lback(region.r_linep);

	// Ending line is start of line, unless part way in the line: take next line
	curbp->b_abotline = region.r_endOffset == 0 ? region.r_endLinep : lforw(region.r_endLinep);

	upmode(curbp, WFHARD | WFFORCE | WFMODE);

	return(TRUE);
}

/* Widen the lines the commands act on to the whole buffer.
 *
 * If the buffer was narrowed more than once - this still widens it to all
 * lines in the buffer.
 *
 * This succeeds even if the buffer was not narrowed.
 * ME 5 this is called widen-from-region --- but the region is no longer relevant.
 * GNU this is widen, thus the synonym.
 *
 * Any count is ignored.
 *
 * Command:  widen-from-narrow   widen
 * Bound to: C-x >
 *
 * Return: TRUE - always.
 */
int	widen(int f, Cnt n)
{
	curbp->b_btopline = curbp->b_linep;
	curbp->b_abotline = curbp->b_linep;

	upmode(curbp, WFHARD | WFFORCE | WFMODE);

	return(TRUE);
}

/* Check if a line is in curbp's narrow.
 *
 * line		The line to check
 * notMessage	Pass this to mlwrite() on fail
 *		If NULL ignore
 *
 * Return:
 *	TRUE  the line is in the narrow or no narrow set
 *	FALSE the line is not in the narrow
 */
int	isInNarrow(LINE* line, char* notMessage, ...)
{
	LINE*	lp;		// For scanning the buffer
	int	inNarrow;	// Scan in the narrow

	// Checking a large buffer can be expensive
	// If a narrow is not set: quick return
	if(curbp->b_btopline == curbp->b_linep && curbp->b_abotline == curbp->b_linep)
		return(TRUE);

	// The narrow might start on the top line
	inNarrow = curbp->b_linep == curbp->b_btopline;

	for(lp = lforw(curbp->b_linep); ; lp = lforw(lp)) {
		// Found what we are looking for:
		if(lp == line) {
			if(inNarrow)
				return(TRUE);
			break;	// Generate message
		}

		// Next line will be in narrow
		if(lp == curbp->b_btopline)
			inNarrow = 1;

		// End of narrow - not seen line:
		if(lp == curbp->b_abotline)
			break;

		// Test after test above in case line is last in buffer
		if(lp == curbp->b_linep)
			break;
	}

	// Fail - maybe generate a message
	if(notMessage != NULL) {
		va_list	ap;
		va_start(ap, notMessage);
		msgWrite(notMessage, ap);
		va_end(ap);
	}

	return(FALSE);
}

/* end */
