/* Test code for MicroEmacs
 * What this is testing depends on whatever is currently being developed.
 * It is envisaged that test code not be removed, but be conditionally compiled
 * out - in case more testing of whatever is needed.
 *
 * ADDW SCCS: @(#)test.c	1.11 08/12/20 17:24:28
 */
#include	"estruct.h"
#include	"utf8.h"
#include	"edef.h"

/* Master control for testing, none of it unless this is true
 *
 * This is called via the command test-emacs
 * The command does not exist unless TEST_HOOK
 *
 * Enable the individual tests each with its own #define
 * The function names all start test_ and they all take f, n arguments.
 */
#define	TEST_HASH	1

#if	TEST_HOOK
#if	SHOWMODREL
#include	"emdmap.h"

char	testvers[] = "@(#)test.c	1.11 08/12/20 17:24:28";
#endif

#define	TEST_OUTPUT_FILE	"test_output"
FILE*	test_fd;		// Opened/append when test_emacs() called

#endif	// TEST_HOOK


#if	TEST_HOOK && TEST_HASH
int	test_hash(int f, Cnt n)
{
	NBIND*	cp;
	int	max = 0;
	int	i;
	static	int	maxCnt[40];	// static for 0 initialise

	// Test string hashing - use command names
	for(cp = CmdNames; cp->n_name; cp++) {
		ST_val	before = ST_hashProbes;

		hashNx	nx = hashLookup(&hashCmdNames, &cp->n_name);
		int	probes = (int)(ST_hashProbes - before);
		fprintf(test_fd, "%-35s %3.3d %2.2d\n", cp->n_name, nx, probes);

		if(probes < 40)
			maxCnt[probes - 1]++;
		if(max < probes)
			max = probes;
	}

	for(i = 1; i <= max; i++) {
		fprintf(test_fd, "%d\t%d\n", i, maxCnt[i - 1]);
		maxCnt[i - 1] = 0;
	}

	// Test integer hashing - use key binding
	KEYTAB*	kp;
	max = 0;

	for(kp = keytab; kp->k_u.k_ufp; kp++) {
		ST_val	before = ST_hashProbes;

		hashNx	nx = hashLookup(&hashKeyBindings, &kp->k_code);
		int	probes = (int)(ST_hashProbes - before);
		fprintf(test_fd, "%10x %3.3d\t%2.2d\n", kp->k_code, nx, probes);

		if(probes < 40)
			maxCnt[probes - 1]++;
		if(max < probes)
			max = probes;
	}

	// Printable character - will be entered as text
	KeyCode	Char;
	for(Char = ' '; Char <= '~'; Char++) {
		ST_val	before = ST_hashProbes;

		hashNx	nx = hashLookup(&hashKeyBindings, &Char);
		int	probes = (int)(ST_hashProbes - before);
		fprintf(test_fd, "%10c %3.3d\t%2.2d\n", Char, nx, probes);

		if(probes < 40)
			maxCnt[probes - 1]++;
		if(max < probes)
			max = probes;
	}

	for(i = 1; i <= max; i++) {
		fprintf(test_fd, "%d\t%d\n", i, maxCnt[i - 1]);
		maxCnt[i - 1] = 0;
	}

	return(0);
}
#endif	// TEST_HASH



#if	TEST_HOOK
int	test_emacs(int f, Cnt n)
{
	if((test_fd = fopen(TEST_OUTPUT_FILE, "a")) == NULL) {
		mlwrite("%BCannot open %s as %e", TEST_OUTPUT_FILE, errno);
		return(FALSE);
	}

#if	TEST_HASH
	test_hash(f, n);
#endif

	fclose(test_fd);

	return(0);
}

#endif	// TEST_HOOK


#if	CHECK_EVERY
// Something in which to store page line # setup
typedef	struct	{
	LineLn	ce_top;		// Top line #
	LineLn	ce_dot;		// Line # for "."
	int	ce_winch;	// Window changed from last entry
} CE_page;
CE_page	ce_pages[CMDSKEPT];	// Same number as CmdHistory
CE_page*ce_CurPage = ce_pages;	// Current page
Window*	ce_window;		// Window of last page

/* Attempt to find a low frequency bug - where the symptoms can be detected
 * but it is not known what command/sequence-of-commands causes it.
 * On detection: an error message is shown, the bell rung and the last few
 * command names put into the message buffer.
 *
 * Regard this as a template: you need to write the symptom detection code,
 * the evidence of the error and just use the command name logging code.
 *
 * Leave old test code in #if-ed out - might be useful to steal from later.
 *
 * This was used in main.c, called after the command has been executed & CurCmd incremented.
 */
void	CheckEveryCommand(void)
{
	LASTCMD*cc = CurCmd;
	int	i;
	int	error = 0;	// True if error found
	char	b[500];

	/* Try to find the bug where line numbers are wrong.
	 * This only seems to happen in the current window.
	 * Numbers are calculated from the top line on the screen: w_toplinenum
	 * What this does is to check the value by counting from b_linep
	 *
	 * Update the screen first else get false alarms on things like find-tag.
	 */
	update(0);

#if 0	// Old test code
	// Don't check message line
	if(curwp != wmessp) {
		LINE*	lp;
		LineLn	lnum;
		CE_page*cp;

		// Record "." and top values:
		if(++ce_CurPage > ce_pages + sizeof(ce_pages) / sizeof(ce_pages[0]))
			ce_CurPage = ce_pages;
		for(lnum = 1, lp = lforw(curwp->w_bufp->b_linep);
				lp != curwp->w_bufp->b_linep && lp != curwp->w_dotp; lp = lforw(lp))
			lnum++;		// Calculate line# of "."
		ce_CurPage->ce_top = curwp->w_toplinenum;
		ce_CurPage->ce_dot = lnum;
		ce_CurPage->ce_winch = ce_window != curwp;	// Has window changed ?
		ce_window = curwp;

		for(lnum = 1, lp = lforw(curwp->w_bufp->b_linep);
				lp != curwp->w_bufp->b_linep && lp != curwp->w_topLinep; lp = lforw(lp))
			lnum++;		// Calculate line# of top

		// If we hit end of buffer - ignore
		if(curwp->w_toplinenum != lnum && lp != curwp->w_bufp->b_linep) {
			mlwrite("%E""%B""%F""Bad top line buf:'%s' #: wp:%D != count:%D", curwp->w_bufp->b_fnameClean, curwp->w_toplinenum, lnum);
			error = 1;

			// Show top,"." line # history
			strcpy(b, "top:. ");
			cp = ce_CurPage;

			for(i = CMDSKEPT; --i >= 0; ) {
				// Next command, or at start wrap to CMDSKEPT'th previous command
				if(++cp >= ce_pages + sizeof(ce_pages) / sizeof(ce_pages[0]))
					cp = ce_pages;

				// Very early on we will have empty slots:
				if(cp->ce_top == 0)
					continue;

				sprintf(strchr(b, '\0'), " %ld:%ld", cp->ce_top, cp->ce_dot);
				if(cp->ce_winch)
					strcat(b, ":winch");
			}

			mlwrite("%#%s", b);
		}
	}
#endif	// Old test code


#if 0	// Old test code
	/* The problem seems to be numeric marks remaining when the corresponding LINEs
	 * have been removed.
	 */
	Window*	wp;

	for(wp = wmessp, i = 0; wp != NULL; wp = wp->w_wndp, i++) {
		if( ! wp->w_flag)	// Not changed - assume OK
			continue;

		NMARK*	nm;

		int	show=0;
rescan:;
		if(show)
			mlwrite("%#ERROR window %d buffer %s", i, wp->w_bufp->b_bname);
		// I suspect that this code in l_unMark() could be a problem
		// Window not showing the current buffer ?
		// if(wp->w_bufp != curbp)
		//	continue;
		nm = wp->w_nmarks;

		LineLn	clno = 0;		// Line # of "."
		LineLn	lno = 0;		// Probe line #. Starts line 1
		LineLn	mlno = 0;		// Mark line#

		for( ; nm < wp->w_nmarks + 10; nm++) {
			LINE*	line;		// Probe/scan LINEs in buffer
			LineLn	seenLno = 0;	// The line # it was seen at


			// if(show)
			//	mlwrite("%#Mark %d n_line=%p", (nm - wp->w_nmarks), nm->n_pline);

			if( !nm->n_pline) {	// Not a mark
				if(show)
					mlwrite("%#Mark %d unset", (nm - wp->w_nmarks));
				continue;
			}

			// Buffer end can be marked, ignore it
			// if(nm->n_pline == wp->w_bufp->b_linep)
			//	continue;

			/* Scan whole buffer
			 * including b_linep as it is possible for a mark to end up there
			 */
			for(line = wp->w_bufp->b_linep, lno = 1;
			    line != wp->w_bufp->b_linep || lno == 1; line = lforw(line), lno++) {
				if(wp->w_dotp == line)	// "." ?
					clno = lno;
				if(wp->w_markp == line)	// mark ?
					mlno = lno;
				if(nm->n_pline == line) {
					seenLno = lno;
					// Continue scanning to get clno
				}
			}
				
			if(seenLno == 0) {
				// Rescan to show all marked line addresses
				if(!show){show=1;mlwrite("%#");goto rescan;}

				// Marked line not seen in the buffer - whoops!
				mlwrite("%W""%B""#mark %d mark=%p UNSEEN",
					(nm - wp->w_nmarks), nm->n_pline);
				error = 1;
				nm->n_pline = NULL;	// Don't complain twice
			} else
				if(show)
					mlwrite("%W""%B""#mark %d mark=%p",
						(nm - wp->w_nmarks), nm->n_pline);

		}
		if(show)
			mlwrite("%#.: %d/%d #=%d/%d .@=%p; mark=%d/%d #=%d mark@=%p END=%p",
				wp->w_doto, lused(wp->w_dotp), clno, lno, wp->w_dotp,
				(wp->w_markp?wp->w_marko:-1), (wp->w_markp?lused(wp->w_markp):-1), mlno, wp->w_markp,
				wp->w_bufp->b_linep);

	}
#endif	// Old test code

#if 0	// Old test code
	/* The number of windows that a buffer is in (b_nwnd) is wrong,
	 * in particular it never reaches zero, so is 2 when it is being displayed once.
	 */
	BUFFER*	bp;
	for(bp = bheadp; bp; bp = bp->b_bufp) {
		int	disCnt = 0;
		if(strcmp(bp->b_bname, "[Info Window]") == 0)	// Ignore this
			continue;
		for(wp = wmessp; wp; wp = wp->w_wndp)
			if(wp->w_bufp == bp)
				disCnt++;
		if(disCnt == bp->b_nwnd)
			continue;
		mlwrite("%FBuffer '%s' displayed %d times but believes %d",
			bp->b_bname, disCnt, bp->b_nwnd);
		error = 1;
	}

	int	lines = 1;	// 1 for command line

	for(wp = wheadp; wp; wp = wp->w_wndp) {
		if(wp->w_leftcol != 0)
			return;	// Too complicated
		lines += wp->w_ntrows + 1;
	}

	if(lines == AllocLines)
		return;

	// Whoops!
	mlwrite("%E""%BWindow line count error. sum(w) %d != AllocLines %d", lines, AllocLines);
	b[0] = '\0';
	for(wp = wheadp; wp; wp = wp->w_wndp)
		sprintf(strchr(b, '\0'), " t=%d #=%d", wp->w_toprow, wp->w_ntrows + 1);
	mlwrite("%#Windows:%s", b);
#endif	// Old test code


	if(error == 0)
		return;

	strcpy(b, "Last few editor commands: ");
	for(i = CMDSKEPT; --i >= 0; ) {
		// Next command, or at start wrap to CMDSKEPT'th previous command
		if(++cc >= CmdHistory + sizeof(CmdHistory) / sizeof(CmdHistory[0]))
			cc = CmdHistory;

		// Very early on we will have empty slots:
		if(cc->lc_cmd == (NBIND*)NULL)
			continue;

		strcat(b, cc->lc_cmd->n_name);

		if(cc->lc_flag)
			sprintf(strchr(b, 0), " count=%ld", (long)cc->lc_arg);

		if(cc->lc_status == -1)
			strcat(b, " did not complete");
		else
			sprintf(strchr(b, 0), " status=%d", cc->lc_status);

		strcat(b, "; ");
	}
	mlwrite("%#%s", b);
}

void	DumpBind(void)
{
	KEYTAB*	ktp;		/* pointer into the command table */
	NBIND*	nptr;		/* pointer into the name binding table */
	FILE*	fp = fopen("KEYS", "w");

	for(ktp = keytab; ktp->k_u.k_ubp != NULL; ++ktp) {
		fprintf(fp, "\t{%ld", (long)ktp->k_code);

		for(nptr = CmdNames; nptr->n_name != NULL; ++nptr) {
			if(ktp->k_u.k_ubp == nptr) {
				fprintf(fp, ", \"%s\"", nptr->n_name);
				break;
			}
		}
		fprintf(fp, "},\n");
	}
	fclose(fp);
}

#endif
