/*
 * ADDW SCCS: @(#)utf8.c	1.64 09/22/20 17:23:55
 */

#define	UTF8_EXTERN	/* Cause UTF8 vars to be defined */

#include	"estruct.h"
#include	"utf8.h"
#include	"edef.h"

#if	SHOWMODREL
char	utf8vers[] = "@(#)utf8.c	1.64 09/22/20 17:23:55";
#endif

char	utf8_msgbuf[200];	// Error & other messages here - for mode line & info display

BYTE	utf8_replace[] = {0xEF, 0xBF, 0xBD, 0};	// UTF8_REPLACEMENT as a string
BYTE*	utf8_modeReplace;	// What to use in utf8_modecheck(), utf8_replace if not set. NUL terminated

static	UniChar	code_point;	// Character U+xx values put here

/* Brief error messages - UTF8ERR_XXX
 * Check utf8_setmsg() if these are changed.
 */
static	char*	utf8_errlist[] = {
	"no error",				// UTF8ERR_NO_ERROR
	"truncated",				// UTF8ERR_TOO_SHORT
	"over long",				// UTF8ERR_OVER_LONG
	"illegal code point",			// UTF8ERR_INV_CODE_POINT
	"> 4bytes",				// UTF8ERR_TOO_MANY_BYTES
	"bad cont byte",			// UTF8ERR_BAD_CONT_BYTE
	"unexpected continuation byte",		// UTF8ERR_UNEXP_CONT_BYTE
	"Cannot start a UTF-8 character here",	// UTF8ERR_NOT_HERE
	"Makes different length character",	// UTF8ERR_DIFF_LEN
};

/* Index into this using the 1st byte of a UTF8 character to get
 * the length of the character in bytes.
 * This gives results for 5 & 6 byte encoded characters - these
 * are no longer part of the standard.
 * The continuation bytes have values 0.
 * Note 7 in the last 2. Not legal, a value to distinguish from continuation bytes.
 * Any value 5 & greater is not legal UTF8.
 */
unsigned char	utf8_charLen[256] = {
	/* U+00 - U+7F first byte: 0xxxxxxx */
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
	/* Continuation, first byte: 10xxxxxx */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
	/* U+0080 - U+07FF first byte: 110xxxxx */
	2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
	2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
	/* U+0800 - U+FFFF first byte: 1110xxxx */
	3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3, 3,
	/* U+10000 - U+1FFFFF first byte: 11110xxx */
	4, 4, 4, 4, 4, 4, 4, 4,
	/* U+200000 - U+3FFFFFF first byte: 111110xx */
				5, 5, 5, 5,
	/* U+4000000 - U+7FFFFFFF first byte: 1111110x */
					    6, 6,
	/* Not used, first byte: 1111111x */
						  7, 7
};

/* Clear the error message buffer
 */
void	utf8_errclear(void)
{
	utf8_msgbuf[0] = '\0';
}

/* Set details of an error message to be used (much) later.
 * The idea is that we scan for error but only report the first one.
 * Argument is the line number of the error, other stuff we know.
 *
 * Record info about this first error:
 * + error
 * + line number
 * + character position in line
 * + byte position in line
 *
 * Return the current error message - which might not relate to the current line!
 * This message counts with the start of line being position 1 (line positions in
 * messages to the user start at 1) - but internally we count from 0.
 *
 * This must be called as immediately after an error - else vars may be overwritten
 */
char*	utf8_setmsg(BufLns linenumber)
{
	char	buf[40];
	// Error, record first one only
	if(utf8_msgbuf[0] == '\0') {
		strcpy(buf, utf8_errlist[utf8_errno]);
		if(utf8_errno == UTF8ERR_INV_CODE_POINT)
			sprintf(strchr(buf, '\0'), " U+%lx", (long)code_point);
		sprintf(utf8_msgbuf, "UTF-8 error (%s) bytes in error %ld. At line %ld, char %ld, byte %ld. ",
			buf, (long)utf8_errLen, (long)linenumber, (long)utf8_chars + 1, (long)(utf8_errOffset + 1));
		NoteErr.u8errno = utf8_errno;
		NoteErr.errLen = utf8_errLen;
		NoteErr.Lineno = linenumber;
		NoteErr.CharPos = utf8_chars;
		NoteErr.Offset = utf8_errOffset;
		NoteErr.codepoint = code_point;
	}

	return(utf8_msgbuf);
}

/* Return a string value for an error
 */
char*	utf8_perror(void)
{
	return(utf8_errlist[utf8_errno]);
}

/* Return the current error message - might be the empty string if there is not an error.
 *
 * If fixup - fix the current buffer.
 * What fixing means is: if there is an error
 * + remove the MDUTF8 flag on the current buffer
 * + remove l_utf8 on all lines in the buffer
 */
char*	utf8_errmsg(int fixup)
{
	if(utf8_msgbuf[0] && fixup) {
		LINE*	lp;

		curbp->b_mode &= ~MDUTF8;

		for(lp = lforw(curbp->b_linep);lp != curbp->b_linep; lp = lforw(lp))
			lunsetUtf8(lp);
	}

	return(utf8_msgbuf);
}


/* Return TRUE if uc is a valid code point.
 * Note that invalid is as encoded in UTF-8.
 * Set utf8_errno on error/fail.
 *
 * See: Corrigendum #9: Clarification About Noncharacters
 * http://www.unicode.org/versions/corrigendum9.html
 * This *does* allow noncharacters to be interchanged (ie in files),
 * but the semantics are not defined.
 * Noncharacters consist of the values U+nFFFE and U+nFFFF (where n is from 0 to 0x10) and the values U+FDD0..U+FDEF.
 * Code fixed to allow non characters: July 2017 (not that it ever disallowed all of them).
 */
int	utf8_isvalid(UniChar uc)
{
	// Throw up hands if not a valid code point:
	if(uc < 0 || uc > UTF8_MAX_CODE_POINT ||
	  (uc >= UTF8_B1_LOW_SH && uc <= UTF8_B1_HIGH_SH)) {		// Surrogate halves used in UTF-16
		utf8_errno = UTF8ERR_INV_CODE_POINT;
		return 0;
	}

	return 1;
}

/* Encode character uc as a UTF8 string.
 * Return the string - in a static buffer.
 *
 * This will encode any value, there is not any checking done - ie checking should
 * be done before calling this.
 *
 * Note that this will be NUL terminated, and if the input was U+0
 * the string returned will be the empty string - followed by a NUL byte.
 *
 * This will encode UniChars > U+10FFFF if UTF8_MAX_CODE_POINT is set to
 * another maximum.
 * With the standard UTF8_MAX_CODE_POINT this will actually encode to UTF8_MAX_CP4,
 * if passed a value greater than that it gets it wrong and it might be an
 * invalid UTF-8 string.
 */
BYTE*	utf8_encode(UniChar uc)
{
	static	BYTE	ret[UTF8_MAX_BYTES + 1];		// What we return
	BYTE*		s = ret;

	if(uc <= UTF8_MAX_CP1)
		*s++ = (BYTE) uc;
	else   if(uc <= UTF8_MAX_CP2) {
		*s++ = (BYTE)( uc >>  6 | UTF8_B1_BITS2);
		*s++ = (BYTE)((uc       & UTF8_Bn_VMASK) | UTF8_Bn_TPVALUE);
	} else if(uc <= UTF8_MAX_CP3) {
		*s++ = (BYTE)( uc >> 12 | UTF8_B1_BITS3);
		*s++ = (BYTE)((uc >>  6 & UTF8_Bn_VMASK) | UTF8_Bn_TPVALUE);
		*s++ = (BYTE)((uc       & UTF8_Bn_VMASK) | UTF8_Bn_TPVALUE);
#if	UTF8_MAX_CODE_POINT > UTF8_MAX_CP4
	} else if(uc <= UTF8_MAX_CP4) {
#else
	} else {
#endif
		*s++ = (BYTE)( uc >> 18 | UTF8_B1_BITS4);
		*s++ = (BYTE)((uc >> 12 & UTF8_Bn_VMASK) | UTF8_Bn_TPVALUE);
		*s++ = (BYTE)((uc >>  6 & UTF8_Bn_VMASK) | UTF8_Bn_TPVALUE);
		*s++ = (BYTE)((uc       & UTF8_Bn_VMASK) | UTF8_Bn_TPVALUE);
#if	UTF8_MAX_CODE_POINT > UTF8_MAX_CP4
	// This code will probably never be compiled in.
	} else if(uc <= UTF8_MAX_CP5) {
		*s++ = (BYTE)( uc >> 24 | UTF8_B1_BITS5);
		*s++ = (BYTE)((uc >> 18 & UTF8_Bn_VMASK) | UTF8_Bn_TPVALUE);
		*s++ = (BYTE)((uc >> 12 & UTF8_Bn_VMASK) | UTF8_Bn_TPVALUE);
		*s++ = (BYTE)((uc >>  6 & UTF8_Bn_VMASK) | UTF8_Bn_TPVALUE);
		*s++ = (BYTE)((uc       & UTF8_Bn_VMASK) | UTF8_Bn_TPVALUE);
	} else {
		*s++ = (BYTE)( uc >> 30 | UTF8_B1_BITS6);
		*s++ = (BYTE)((uc >> 24 & UTF8_Bn_VMASK) | UTF8_Bn_TPVALUE);
		*s++ = (BYTE)((uc >> 18 & UTF8_Bn_VMASK) | UTF8_Bn_TPVALUE);
		*s++ = (BYTE)((uc >> 12 & UTF8_Bn_VMASK) | UTF8_Bn_TPVALUE);
		*s++ = (BYTE)((uc >>  6 & UTF8_Bn_VMASK) | UTF8_Bn_TPVALUE);
		*s++ = (BYTE)((uc       & UTF8_Bn_VMASK) | UTF8_Bn_TPVALUE);
#endif
	}

	*s = '\0';

	return(ret);
}

#if	0
/* Move the cursor back to the start of the multibyte character
 * that it is sitting on.
 * Does nothing if not on a multibyte character.
 * QV: UTF8_BACK_FIND_START()
 */
void	utf8_backToStart(void)
{
// This is not used - use UTF8_BACK_FIND_START_NX();
	while(curwp->w_doto > 0 && BYTE_IS_MB_CONT(lgetc(curwp->w_dotp, (curwp->w_doto)))
		curwp->w_doto--;
}
#endif

/* Return a string of information about the current character on the current line.
 * Return in a static buffer - utf8_msgbuf.
 *
 * Return a string like: 'U+xxxx [#XX #XX #XX]'
 */
char*	utf8_CurrCharInfo(void)
{
	BYTE*	currch;
	int	ch_bytes;	// Bytes of UTF8
	int	xwidth;		// Width to print U+

	/* End of file - no notional character */
	if(curwp->w_dotp == curbp->b_linep) {
		utf8_msgbuf[0] = '\0';
		return(utf8_msgbuf);
	}

	// Find a character - get its address
	// At end of line, magic up a newline
	if(curwp->w_doto == lused(curwp->w_dotp))
		currch = NewlineString;
	else
		currch = lgetcaddr(curwp->w_dotp, curwp->w_doto);

	// Decode it to code_point
	ch_bytes = CHAR_LEN(*currch);
	utf8_chkbuf(currch, (size_t)ch_bytes);
	xwidth = code_point <= 0xFF ? 2 : code_point <= 0xFFFF ? 4 : 6;
	sprintf(utf8_msgbuf, "U+%*.*lX [", xwidth, xwidth, (long)code_point);
	while(--ch_bytes >= 0)
		sprintf(strchr(utf8_msgbuf, '\0'), "#%2.2x ", *currch++);
	strchr(utf8_msgbuf, '\0')[-1] = ']';	// Blat last space with close square brackets

	return(utf8_msgbuf);
}

/* Return a decoded UTF8 code point that is at str.
 * This does no error checking to be fast.
 * Should be OK if the buffer has been verified to only contain good UTF8 characters
 *
 * Will return -1 if the first byte is not one that can start a UTF8 sequence.
 * If you are unsure of the origin of the sequence you should also check that
 * the value returned is not greater than UTF8_MAX_CODE_POINT.
 */
UniChar	utf8_decode(const BYTE* str)
{
	/* The long way round that checks that the character is valid.
	 *	return utf8_chkbuf(str, CHAR_LEN(*str)) ? code_point : -1;
	 */

	switch(CHAR_LEN(*str)) {
	case 1:
		return(*str);
	case 2:
		return(((str[0] & UTF8_B1_MASK2) <<  6) |  (str[1] & UTF8_Bn_VMASK));
	case 3:
		return(((str[0] & UTF8_B1_MASK3) << 12) | ((str[1] & UTF8_Bn_VMASK) <<  6) |  (str[2] & UTF8_Bn_VMASK));
	case 4:
		return(((str[0] & UTF8_B1_MASK4) << 18) | ((str[1] & UTF8_Bn_VMASK) << 12) | ((str[2] & UTF8_Bn_VMASK) <<  6) |  (str[3] & UTF8_Bn_VMASK));
#if	UTF8_MAX_CODE_POINT > UTF8_MAX_CP4
	case 5:
		return(((str[0] & UTF8_B1_MASK5) << 24) | ((str[1] & UTF8_Bn_VMASK) << 18) | ((str[2] & UTF8_Bn_VMASK) << 12) | ((str[3] & UTF8_Bn_VMASK) <<  6) |  (str[4] & UTF8_Bn_VMASK));
#endif
#if	UTF8_MAX_CODE_POINT > UTF8_MAX_CP5
	case 6:
		return(((str[0] & UTF8_B1_MASK6) << 30) | ((str[1] & UTF8_Bn_VMASK) << 24) | ((str[2] & UTF8_Bn_VMASK) << 18) | ((str[3] & UTF8_Bn_VMASK) << 12) | ((str[4] & UTF8_Bn_VMASK) << 6) | (str[5] & UTF8_Bn_VMASK));
#endif
	}

	// Should never get here
	return(-1);
}

/* Return true if str is a valid UTF8 string
 *
 * This will accept as correct a non UTF8 (ie plain ASCII) string.
 */
int	utf8_check(const BYTE* str)
{
	return utf8_chkbuf(str, strlen(str));
}

/* Return true if the len bytes at str form a valid UTF8 string.
 * If there is an error:
 * + utf8_errOffset: contains the byte offset of the start of the character in error
 * + utf8_errno: is an error code. There could be multiple things wrong with the character
 *	(eg over long and invalid code point) which one is set is not defined.
 * + utf8_errLen: contains how many bytes are in error. This might be low
 *   if there is a long sequence of bad bytes
 *
 * code_point is assigned the current character value
 *
 * Global utf8_chars is set to the number of characters (not bytes) in the string,
 * if there is an error it is the number processed before the error.
 *
 * This will accept as correct a non UTF8 (ie plain ASCII) string.
 * You can tell if there was a multibyte character in the line: if utf8_chars != len
 */
#define	LoopUnroll	1	/* Unroll loops to make this faster */

int	utf8_chkbuf(const BYTE* str, size_t len)
{
#if	LoopUnroll == 0
	int		len_byte;	/* First byte is length byte */
	UniChar		min_cp;		/* Min code point */
	int		data_mask;
	int		length;		/* Length of UTF8 sequence - 1 */
	int		i;
#endif

	const BYTE*	start_str = str;

	utf8_chars = 0;

	//fprintf(stderr, "given '%s' len %ld\n", str, (long)len);
	while(len-- > 0) {

#if	LoopUnroll

		switch(CHAR_LEN(*str)) {
		case 1:
			code_point = *str++;	// 1 byte US-ASCII
			break;
		case 2:
			// Enough left in input ?
			if(len < 1)
				goto too_short;

			// Subsequent byte continuation byte ?
			if( !BYTE_IS_MB_CONT(str[1]))
				goto bad_cont;

			code_point = ((str[0] & UTF8_B1_MASK2) << 6) | (str[1] & UTF8_Bn_VMASK);

			if(code_point < UTF8_MIN_CP2)
				goto over_long;

			// Do NOT fiddle with these until the above checks are done.
			// To do so will break setting of utf8_errXXX
			len--;
			str += 2;
			break;
		case 3:
			if(len < 2)
				goto too_short;

			if( !BYTE_IS_MB_CONT(str[1]) || !BYTE_IS_MB_CONT(str[2]))
				goto bad_cont;

			code_point = ((str[0] & UTF8_B1_MASK3) << 12) | ((str[1] & UTF8_Bn_VMASK) << 6) | (str[2] & UTF8_Bn_VMASK);

			if(code_point < UTF8_MIN_CP3)
				goto over_long;

			// Check the high and low surrogate halves used by UTF-16
			if((code_point >= UTF8_B1_LOW_SH && code_point <= UTF8_B1_HIGH_SH) || code_point == UTF8_SWAP_BOM || code_point == UTF8_ILL_FFFF)
				goto invalid_codepoint;

			len -= 2;
			str += 3;
			break;
		case 4:
			if(len < 3)
				goto too_short;

			if( !BYTE_IS_MB_CONT(str[1]) || !BYTE_IS_MB_CONT(str[2]) || !BYTE_IS_MB_CONT(str[3]))
				goto bad_cont;

			code_point = ((str[0] & UTF8_B1_MASK4) << 18) | ((str[1] & UTF8_Bn_VMASK) << 12) | ((str[2] & UTF8_Bn_VMASK) << 6) | (str[3] & UTF8_Bn_VMASK);

			if(code_point < UTF8_MIN_CP4)
				goto over_long;

			// UTF8 max
			if(code_point > UTF8_MAX_CODE_POINT)
				goto invalid_codepoint;

			len -= 3;
			str += 4;
			break;
#if	UTF8_MAX_CODE_POINT > UTF8_MAX_CP4
		case 5:
			// Non standard UTF-8 coding
			if(len < 4)
				goto too_short;

			if( !BYTE_IS_MB_CONT(str[1]) || !BYTE_IS_MB_CONT(str[2]) || !BYTE_IS_MB_CONT(str[3]) || !BYTE_IS_MB_CONT(str[4]))
				goto bad_cont;

			code_point = ((str[0] & UTF8_B1_MASK5) << 24) | ((str[1] & UTF8_Bn_VMASK) << 18) | ((str[2] & UTF8_Bn_VMASK) << 12) | ((str[3] & UTF8_Bn_VMASK) << 6) | (str[4] & UTF8_Bn_VMASK);

			if(code_point < UTF8_MIN_CP5)
				goto over_long;

			// UTF8 max
			if(code_point > UTF8_MAX_CODE_POINT)
				goto invalid_codepoint;

			len -= 4;
			str += 6;
			break;
#endif
#if	UTF8_MAX_CODE_POINT > UTF8_MAX_CP5
		case 6:
			if(len < 5)
				goto too_short;

			if( !BYTE_IS_MB_CONT(str[1]) || !BYTE_IS_MB_CONT(str[2]) || !BYTE_IS_MB_CONT(str[3]) || !BYTE_IS_MB_CONT(str[4]) || !BYTE_IS_MB_CONT(str[5]))
				goto bad_cont;

			code_point = ((str[0] & UTF8_B1_MASK6) << 30) | ((str[1] & UTF8_Bn_VMASK) << 24) | ((str[2] & UTF8_Bn_VMASK) << 18) | ((str[3] & UTF8_Bn_VMASK) << 12) | ((str[4] & UTF8_Bn_VMASK) << 6) | (str[5] & UTF8_Bn_VMASK);

			if(code_point < UTF8_MIN_CP6)
				goto over_long;

			// UTF8 max
			if(code_point > UTF8_MAX_CODE_POINT)
				goto invalid_codepoint;

			len -= 5;
			str += 6;
			break;
#endif
		default:
			// Something wrong:
			utf8_errLen = 1;
			utf8_errOffset = str - start_str;

			/* A valid first byte will have the top 2 bits (0xC0) set.
			 * Many 'random' bytes will not, complaining about sequence that is too long
			 * but is one byte long will cause confusion.
			 * Try to distinguish the case when just the top bit is set.
			 */
			if((*str & UTF8_Bn_MASK) == UTF8_Bn_MASK) {

				// Could be because the 1st byte is continuation - hopefully not.
				utf8_errno = UTF8ERR_TOO_MANY_BYTES;

				// How long ? Prob 1, but slurp up to 3 continuation bytes
				while(utf8_errLen < UTF8_MAX_BYTES && (int)len >= utf8_errLen && BYTE_IS_MB_CONT(str[utf8_errLen]))
					utf8_errLen++;
			} else {
				// This looks like a continuation byte, but it is not after a start byte
				// and is thus unexpected.
				utf8_errno = UTF8ERR_UNEXP_CONT_BYTE;
			}

			return 0;
		}

		utf8_chars++;

#else
		// I believe that this works, it did once but might have been broken since

		if(*str < UTF8_B1_LEN1) {	// Common case first
			code_point = *str++;	// 1 byte US-ASCII
			utf8_chars++;
			continue;
		}

		//fprintf(stderr, "Found '%s' len=%ld\n", str, (long)len);
		len_byte = *str++;

		// Work out length - number of following bytes needed
		// NB: length is # following bytes, not length of sequence

		if(len_byte < UTF8_B1_LEN2) {
			length = 1;
			min_cp = UTF8_MIN_CP2;
			data_mask = UTF8_B1_MASK2;
		} else if(len_byte < UTF8_B1_LEN3) {
			length = 2;
			min_cp = UTF8_MIN_CP3;
			data_mask = UTF8_B1_MASK3;
		} else if(len_byte < UTF8_B1_LEN4) {
			length = 3;
			min_cp = UTF8_MIN_CP4;
			data_mask = UTF8_B1_MASK4;
#if	UTF8_MAX_CODE_POINT > UTF8_MAX_CP4
		} else if(len_byte < UTF8_B1_LEN5) {
			length = 4;
			min_cp = UTF8_MIN_CP5;
			data_mask = UTF8_B1_MASK5;
		} else if(len_byte < UTF8_B1_LEN6) {
			length = 5;
			min_cp = UTF8_MIN_CP6;
			data_mask = UTF8_B1_MASK6;
#endif
		} else {	// Too long sequence
			utf8_errno = UTF8ERR_TOO_MANY_BYTES;
			utf8_errOffset = str - start_str;

			// How long ? Prob 1, but slurp up to 3 (UTF8_MAX_BYTES - 1) continuation bytes
			utf8_errLen = 1;
			while(utf8_errLen < UTF8_MAX_BYTES && len >= utf8_errLen && BYTE_IS_MB_CONT(str[utf8_errLen]))
				utf8_errLen++;

			return 0;
		}

		// Enough bytes left ?
		if(length > len) {
			utf8_errno = UTF8ERR_TOO_SHORT;
			utf8_errOffset = str - start_str;
			utf8_errLen = (int)len;

			return 0;
		}

		code_point = len_byte & data_mask;
		//fprintf(stderr, "length=%d len_byte %#2.2x code_point %#2.2x\n", length, len_byte, code_point);

		// Check subsequent bytes start properly
		for(i = 0; i < length; i++) {

			if((str[i] & UTF8_Bn_MASK) != UTF8_Bn_TPVALUE) {
				utf8_errno = UTF8ERR_BAD_CONT_BYTE;
				utf8_errOffset = str - start_str;
				utf8_errLen = length + 1;

				return 0;
			}
			code_point = (code_point << 6) | (str[i] & UTF8_Bn_VMASK)b;
		}

		// Overlong encoding ?
		if(code_point < min_cp) {
			utf8_errno = UTF8ERR_OVER_LONG;
			utf8_errOffset = str - start_str;
			utf8_errLen = length + 1;

			return 0;
		}

		// Increment by # continuation bytes:
		len -= length;
		str += length;
#endif
	}

	return(1);

#if	LoopUnroll

	// Error exits:
too_short:
	utf8_errLen = (int)len;
	utf8_errno = UTF8ERR_TOO_SHORT;
	utf8_errOffset = str - start_str;

	return 0;

bad_cont:
	/* How long the bad sequence is is an interesting problem.
	 * Must not include the non-continuation byte, it is prob the
	 * start of something else.
	 * Cannot be here with a UTF8 char of 1 byte length.
	 */
	utf8_errLen = 1;
	if(BYTE_IS_MB_CONT(str[1]))
		utf8_errLen = BYTE_IS_MB_CONT(str[2]) ? 3 : 2;

	utf8_errno = UTF8ERR_BAD_CONT_BYTE;
	utf8_errOffset = str - start_str;

	return 0;

over_long:
	utf8_errLen = CHAR_LEN(*str);
	utf8_errno = UTF8ERR_OVER_LONG;
	utf8_errOffset = str - start_str;
	return 0;

invalid_codepoint:
	utf8_errLen = CHAR_LEN(*str);
	utf8_errno = UTF8ERR_INV_CODE_POINT;
	utf8_errOffset = str - start_str;

	return 0;

#endif
}

/* Check that placing newch at *pos will not mangle a UTF8 character.
 *
 * If start - newch is a new character, ie start of a new sequence.
 * This must be a 7 bit ASCII character and what is at pos must not be
 * a continuation byte.
 *
 * On error return FALSE and utf8_errno will be set appropriately.
 *
 * This assumes that what is at pos is legal UTF8, so can move forwards & back.
 *
 * The buffer that this is in is that of curwp.
 *
 * Return TRUE if OK.
 */
int	utf8_CheckMangle(const BYTE* pos, BYTE newch, int start)
{
	const BYTE*	sp;			// Start of sequence
	BYTE		db[UTF8_MAX_BYTES + 1];	// Decode buffer

	// Not UTF8 mode: anything goes
	if((curbp->b_mode & MDUTF8) == 0)
		return TRUE;

	// New character: must not be UTF8
	if(start) {

		if(BYTE_IS_MB_CONT(*pos)) {
			utf8_errno = UTF8ERR_UNEXP_CONT_BYTE;
			return(FALSE);
		}

		if(CHAR_LEN(newch) != 1) {
			utf8_errno = UTF8ERR_NOT_HERE;
			return(FALSE);
		}

		utf8_errno = UTF8ERR_NO_ERROR;
		return(TRUE);
	}

	// Find the start of the sequence
	for(sp = pos; BYTE_IS_MB_CONT(*sp) ; sp--)
		;

	/* If the first char is being replaced and it is not the start of
	 * a sequence of the same length - fail.
	 * This also catches replacing a non-UTF8 character with a UTF8 one.
	 */
	if(sp == pos && CHAR_LEN(*pos) != CHAR_LEN(newch)) {
		utf8_errno = UTF8ERR_DIFF_LEN;
		return(FALSE);
	}

	// Create the new sequence
	memcpy(db, sp, CHAR_LEN(*sp));
	db[pos - sp] = newch;

	// Check and return what is found:
	return utf8_chkbuf(db, CHAR_LEN(*sp));
}

/* Check that a buffer contains valid UTF-8
 * Called before un/setting UTF-8 mode.
 * bp - the BUFFER to check/set.
 * set - true if setting the mode
 * global - true if setting the global flag - else the current buffer
 * If flag (f) is set, replace any bad character found:
 * * if the sequence is an over long encoding, replace with the shortest version
 * * n > 0  Fix up to n bad encodings. If not over long replace with utf8_modeReplace or UTF8_REPLACEMENT
 *	    Stop at n+1-th bad encoding or end of buffer (moving to UTF8 mode)
 * * n < 0  Equivalent to n == infinity, ie fix all.
 * * n == 0 Fix up all instances of over long encoding only,
 *	but count and report other errors.
 *	Return false if there are other errors - since we can't move to UTF8 mode.
 *
 * When setting utf8 on a buffer this checks that all characters in the current buffer are utf8 valid.
 *
 * If no flag & there is something invalid, put the cursor there and generate a message.
 *
 * A magic value of f == -2 is treated as 0 (ie unset) but suppresses a message on
 * errors being found and placing the cursor there. This prob ought to be done differently.
 * If bp is not associated with curwp - also suppress.
 *
 * If a fix is made a message is shown.
 *
 * Return
 *	TRUE   The buffer contains valid UTF-8
 *	       Could be as setting global flag
 *	       utf8_errno will also be set to 0 if everything works.
 *	FALSE  Out of memory when inserting a replacement
 *	       There was an error - replacement not asked for
 */
int	utf8_modecheck(BUFFER* bp, int set, int global, int f, Cnt n)
{
	LINE*	lp;
	int	ret = TRUE;
	int	lineno = 1;
	int	was_f = f;		// Record if f was set at some point
	long	replacements = 0;	// How many we do
	long	otherErrors = 0;	// Count of non over-long errors (n==0)
	LineLn	lineOffset = 0;		// Where to start looking in the line

	int	MsgAndMoveDot = 1;	// On error when flag is unset

	if(f == -2 || curwp->w_bufp != bp) {
		// Yuck!
		f = 0;
		MsgAndMoveDot = 0;
	}

	utf8_errno = UTF8ERR_NO_ERROR;

	// No prob setting a global 'desire' to utf8
	if(global)
		return(TRUE);

	// Setting:
	// Scan all lines
	if(set) {
		utf8_errclear();

		for(lp = lforw(curbp->b_linep); lp != curbp->b_linep; ) {

			if(utf8_chkbuf((BYTE*)ltext(lp) + lineOffset, (size_t)(lused(lp) - lineOffset))) {
				/* If did not start at the start of line, rescan the whole line.
				 * This is so that utf8_chars is a correct count of the whole line.
				 */
				if(lineOffset != 0)
					utf8_chkbuf((BYTE*)ltext(lp), (size_t)lused(lp));

				// Line OK, set flag of utf8 chars on the line (if characters != bytes)
				lunsetUtf8(lp);
				if(utf8_chars != lused(lp))
					lsetUtf8(lp);

				// Only move forwards if it checks out OK
				lp = lforw(lp);
				lineno++;
				lineOffset = 0;

				continue;
			}

			if(f) {
				/* Try to fix an error up.
				 */
				BYTE*	bad = ltext(lp) + lineOffset + utf8_errOffset;	// Where bad bytes start
				BYTE*	replace = NULL;				// What to replace with
				int	replen;					// # bytes at replace
				LineLn	remaining;				// # bytes after what we touch


				if(utf8_errno == UTF8ERR_OVER_LONG) {
					/* Over long encoded character
					 * It will decode properly, just get a new encoding
					 * Just overwrite with the correct (shorter) sequence
					 * and copy the rest down to fill the gap.
					 *
					 * Unfortunately: it might not encode - if it is an invalid code point.
					 * Thus need to drop back to replacing it.
					 * BUT: if we are only allowed to fix - drop through & error later
					 */
					UniChar	ch = utf8_decode(bad);
					replace = utf8_encode(ch);
				}

				// Unless we have a replacement for an over long encoding:
				if( !replace) {
					if(n == 0) {
						/* Fixing up instances of over long encoding.
						 * Ignore other errors, but count them.
						 * Set offset so that we can see what else is on this line.
						 */
						lineOffset += utf8_errOffset + (LineLn)utf8_errLen;
						otherErrors++;

						continue;
					}

					// Might have a specific replace string:
					replace = utf8_modeReplace ? utf8_modeReplace : utf8_replace;
				}

				// We have a replacement.
				// Trouble is that the replacement could be longer than the original

				remaining = lused(lp) - lineOffset - utf8_errOffset - (LineLn)utf8_errLen;

				replen = (int)strlen(replace);	// Replacement length
				if(replen > utf8_errLen) {
					// Inserting a longer sequence
					// Ensure enough room:
					if( !lineMoreBytes(lp, lused(lp) + (LineLn)(replen - utf8_errLen)))
						return(FALSE);

					bad = lgetcaddr(lp, lineOffset + utf8_errOffset);	// Might have reallocated the line

					// Copy reverse, overlapping to open up space for the replacement:
					memmove(bad + replen, bad + utf8_errLen, (size_t)remaining);

				} else if(replen < utf8_errLen) {
					// Copy bytes after the bad sequence down:
					memmove(bad + replen, bad + utf8_errLen, (size_t)remaining);
				}

				memcpy(bad, replace, (size_t)replen);

				// Adjust bytes on line:
				lsetused(lp, lused(lp) - (LineLn)(utf8_errLen - replen));

				// Fix "." & marks:
				FixOffsets(lp, (LineLn)(bad - ltext(lp)), (LineLn)(replen - utf8_errLen), curbp, 0);

				replacements++;
				curbp->b_flag |= BFCHG;

				// Reached the max number of fixes to make ?
				if(n > 0 && replacements >= n)
					f = FALSE;

				// Check the line again
				continue;
			}

			// Check failed, no form of replacement:
			ret = FALSE;
			// Unset UTF-8 flag below to ensure that elsewhere doesn't try UTF-8 stuff on the line

			if(MsgAndMoveDot) {
				// Say something wrong and put "." there:
				mlwrite("%E""%s (After %D fixups)", utf8_setmsg(lineno), replacements);

				// Put the cursor at the error:
				curwp->w_dotp = lp;
				curwp->w_doto = utf8_errOffset;
				curwp->w_flag |= WFMOVE;
			}
			break;
		}

		// Was that successful ? :
		if(ret) {
			if(was_f && MsgAndMoveDot)
				mlwrite("%I""%D %sfixups made for UTF-8, %D other errors", replacements,
					(n == 0 ? "over long encoding " : ""),
					otherErrors);

			// If no other errors, return OK, else drop through to remove the l_uft8 flag
			if( !otherErrors) {
				// Ensure that "." and marks (all windows) are not misplaced
				utf8_ensurePosStart(curbp, 1);

				return(TRUE);
			}

			// There are other errors, set to fail but first drop down to remove l_utf8 flag
			ret = FALSE;
		}
	}

	// Remove the l_utf8 flag on all lines
	// Either because an error was found or we are unsetting
	for(lp = lforw(curbp->b_linep); lp != curbp->b_linep; lp = lforw(lp)) {
		lunsetUtf8(lp);
	}

	return(ret);
}

/* Check buffer bp for UTF-8 correctness, set the mode if possible.
 * Return true if the buffer checked OK.
 *
 * Return
 *	TRUE   The buffer contains valid UTF-8
 *	       utf8_errno will also be set to 0 if everything works.
 *	FALSE  Out of memory when inserting a replacement
 *	       There was an error - replacement not asked for
 */
int	utf8_bufmodecheck(BUFFER* bp)
{
	int	ret;

	if((ret = utf8_modecheck(bp, 1, 0, 0, 0)))
		bp->b_mode |= MDUTF8;

	return(ret);
}

/* Some fiddling could leave "." or marks in the middle of a multi byte character
 * Check for this in the current buffer and place them at the start of a  MB char.
 * Windows has HEX option - leave alone.
 *
 * Also check curbp: in case it is not connected to a window - can get this
 * when reading in (or something).
 *
 * Silent return if the current buffer is not in UTF8 mode unless EvenIfNotU8.
 *
 * NB: cursor can be beyond the line buffer - at the virtual \n.
 */
void	utf8_ensurePosStart(BUFFER* bp, int EvenIfNotU8)
{
	Window*	wp;
	AMARK*	amp;

	if((bp->b_mode & MDUTF8) == 0 && !EvenIfNotU8)
		return;

	// If that worked: ensure that "." and marks (all windows) are
	// not on a continuation byte
	for(wp = wheadp; wp != NULL; wp = wp->w_wndp) {
		// Ignore if not current buffer.
		// Also allow HEX display to be sinful:
		if(wp->w_bufp != bp || (wp->w_opts & WO_HEX))
			continue;

		// "."
		if(lused(wp->w_dotp) != wp->w_doto)
			UTF8_BACK_FIND_START_NX(lgetc(wp->w_dotp, wp->w_doto), wp->w_doto);

		// Mark - if set
		if(wp->w_markp && lused(wp->w_markp) != wp->w_marko)
			UTF8_BACK_FIND_START_NX(lgetc(wp->w_markp, wp->w_marko), wp->w_marko);
	}

	if(lused(bp->b_dotp) != bp->b_doto)
		UTF8_BACK_FIND_START_NX(lgetc(bp->b_dotp, bp->b_doto), bp->b_doto);

	if(bp->b_markp && lused(bp->b_markp) != bp->b_marko)
		UTF8_BACK_FIND_START_NX(lgetc(bp->b_markp, bp->b_marko), bp->b_marko);

	// Check alpha marks:
	for(amp = bp->b_amark; amp; amp = amp->m_next)
		if(amp->m_pline && lused(amp->m_pline) != amp->m_oline)
			UTF8_BACK_FIND_START_NX(lgetc(amp->m_pline, amp->m_oline), amp->m_oline);
}

/* Return the number of characters in the buffer at str.
 * There are n bytes in the buffer
 */
size_t	utf8_chrnlen(const BYTE* str, size_t n)
{
	size_t		num = 0;
	const BYTE*	end = str + n;

	for( ; str < end; str += CHAR_LEN(*str))
		num++;

	return(num);
}

/* Find out how long *max UTF8 characters are in bytes.
 *
 * Span over *max characters and return the number of bytes spanned.
 *
 * available_bytes is the number of bytes available.
 *
 * This gives the number of bytes in the *max characters.
 *
 * *max is updated with the number of characters actually found/skipped,
 * this may be less than original *max if there are not enough in available_bytes.
 * (this will be no more than the original value of *max).
 *
 * This does not see the differences in on screen character widths.
 */
size_t	utf8_maxlen(const BYTE* str, size_t available_bytes, size_t* max)
{
	size_t	chrs = 0;	// Chars moved over
	size_t	bytes = 0;	// Bytes moved over
	size_t	toSkip = *max;	// How many chars to skip to reach the target

	// In the test below beware that available_bytes might be unsigned,
	// thus --available_bytes might give silly results
	while(toSkip > chrs) {
		size_t	len = CHAR_LEN(*str);

		if(bytes + len > available_bytes)
			break;

		bytes += len;
		str += len;
		chrs++;
	}

	*max = chrs;

	return(bytes);
}

/* Truncate a string (NUL terminate) so that it is at most len *characters* long.
 * This means that the buffer must be at least len+1 - for the \0
 * Quite what this means is interesting since characters can have different lengths.
 *
 * Return true (1).
 * If str is NULL do nothing, return 0.
 * If reach end of string before len characters have been found, return 0.
 */
int	utf8_strtrunc(BYTE* str, size_t len)
{
	if( !str)
		return(0);

	// Beware: not --len - might be unsigned
	while(len-- > 0) {
		if( !*str)	// Early end of string ?
			return(0);

		str += CHAR_LEN(*str);
	}

	*str = '\0';

	return(1);
}

/* Copy at most len UTF-8 characters from from to to.
 * from is NUL terminated
 * The buffer at to must be large enough.
 *
 * Return true (1).
 */
int	utf8_strncpy(BYTE* to, BYTE* from, size_t len)
{
	// Beware: not --len - might be unsigned
	while(len-- > 0 && *from != '\0') {
		*to++ = *from++;
		while(BYTE_IS_MB_CONT(*from))
			*to++ = *from++;
	}

	*to = '\0';

	return(1);
}

/* Return number of UTF8 character in NUL terminated strings.
 */
size_t	utf8_strlen(const BYTE* str)
{
	size_t	len = 0;

	while(*str) {
		len++;
		str += CHAR_LEN(*str);
	}

	return(len);
}

/* Reverse a NUL terminated string UTF8 encoded NUL terminated string.
 * Return the reversed string.
 */
BYTE*	utf8_strrev(BYTE* str)
{
	return(utf8_memrev(str, strlen(str)));
}

/* Reverse the UTF-8 characters at str.
 * len is the length in bytes - not characters.
 *
 * This will reverse the order of bytes within UTF8 characters,
 * so have a second pass re-reversing the bytes in each multibyte character.
 *
 * Return str
 */
BYTE*	utf8_memrev(BYTE* str, size_t len)
{
	BYTE*	s = str;
	BYTE*	e = str + len - 1;
	BYTE	ch;

	// Reverse the bytes:
	while(e > s) {
		ch   = *e;
		*e-- = *s;
		*s++ = ch;
	}

	// Start at the RH end, move backwards.
	// The first byte that we come across is encodes the length.
	e = str + len - 1;
	while(e > str) {
		switch(CHAR_LEN(*e)) {
		    case 1:	// just move back
			e--;
			break;
		    case 2:	// swap 2
			e   -= 2;
			ch   = e[1];
			e[1] = e[2];
			e[2] = ch;
			break;
		    case 3:	// don't need to swap the middle one
			e   -= 3;
			ch   = e[1];
			e[1] = e[3];
			e[3] = ch;
			break;
		    case 4:	// swap 2 pairs
			e   -= 4;
			ch   = e[1];
			e[1] = e[4];
			e[4] = ch;
			ch   = e[2];
			e[2] = e[3];
			e[3] = ch;
			break;
#if	UTF8_MAX_CODE_POINT > UTF8_MAX_CP4
		    case 5:	// don't need to swap the middle one
			e   -= 5;
			ch   = e[1];
			e[1] = e[5];
			e[5] = ch;
			ch   = e[2];
			e[2] = e[4];
			e[4] = ch;
			break;
#endif
#if	UTF8_MAX_CODE_POINT > UTF8_MAX_CP5
		    case 6:	// swap 3 pairs
			e   -= 6;
			ch   = e[1];
			e[1] = e[6];
			e[6] = ch;
			ch   = e[2];
			e[2] = e[5];
			e[5] = ch;
			ch   = e[3];
			e[3] = e[4];
			e[4] = ch;
			break;
#endif
		}
	}

	return(str);
}

/* Compare 2 UTF-8 character sequences at c1 and c2
 *
 * Note that this compared the UTF-8 encoded characters.
 *
 * Assume that they are valid UTF-8 sequences.
 *
 * Return
 *	TRUE if they are the same
 *	FALSE if they differ
 */
int	utf8_cheq(const BYTE* c1, const BYTE* c2)
{
	if(CHAR_LEN(*c1) != CHAR_LEN(*c2))
		return(FALSE);

	return(memcmp(c1, c2, CHAR_LEN(*c1)) == 0);
}

/* end */
