/*
 * ADDW SCCS: @(#)window.c	1.149 09/29/20 16:10:00
 */

/*
 * Window management. Some of the functions are internal, and some are
 * attached to keys that the user actually types.
 */

#include        "estruct.h"
#include	"utf8.h"
#include	"edef.h"

#if	SHOWMODREL
char	windovers[] = "@(#)window.c	1.149 09/29/20 16:10:00";
#endif


/* Dimensions of the window map. Largest of the last 2 settings.
 * See comments near setScreenSizeForMap() & BuildWindowMap()
 */
//#define	WINDOW_MAP_COLUMNS	(AllocColumns + 2)	// # columns in a map
//#define	WINDOW_MAP_LINES	(AllocLines + 3)	// # lines in a map
#define	WINDOW_MAP_COLUMNS	window_map_columns	// # columns in a map
#define	WINDOW_MAP_LINES	window_map_lines	// # lines in a map
static	int	window_map_columns;			// Value to use, ie allocate and index into
static	int	window_map_lines;
// The previous set value
// [0] - the last one set
// [1] - the one before that
static	int	window_map_columns_prev[2];
static	int	window_map_lines_prev[2];

// Values in cells in the window map:
#define	WINDOW_MAP_SC_EDGE		0		// Screen edge
#define	WINDOW_MAP_SC_NULL		1		// Nothing fills this place
#define	WINDOW_MAP_MSG_LINE		2		// Message line
#define	WINDOW_MAP_INFO			3		// Info window
#define	WINDOW_MAP_START		4		// Real window numbers start here

// Width of a window, ignoring whatever there is in it (text, line numbers) - not ornamentation
#define	WINDOW_WIDTH(wp)	((wp)->w_ncolumns)


/* These define:
 * * directions in which to move an edge
 * * an edge - named after which direction it is from the center of the window.
 */
#define	DIRECTION_UP	0
#define	DIRECTION_DOWN	1
#define	DIRECTION_LEFT	2
#define	DIRECTION_RIGHT	3
#define	DIRECTION_BAD	-1
// static	char*	DirectionNames[] = { "up",  "down",   "left", "right" };
static	char*	EdgeNames[] =      { "top", "bottom", "left", "right" };

/* Convert a direction indicating keystroke to DIRECTION_xxx
 * Keys are: u/t, d/b, l, r & arrows
 * Return DIRECTION_BAD if no match.
 *
 * Up and Top are synonyms, as are Down and Bottom.
 *
 * These are fixed keys, maybe they ought to be bindable - tough!
 */
int	GetDirection(KeyCode ch)
{
	switch(ch) {
	case 'u': case UTC_UP: case 't':
		return DIRECTION_UP;
	case 'd': case UTC_DOWN: case 'b':
		return DIRECTION_DOWN;
	case 'l': case UTC_LEFT:
		return DIRECTION_LEFT;
	case 'r': case UTC_RIGHT:
		return DIRECTION_RIGHT;
	default: return DIRECTION_BAD;
	}
}

/* Return the window number in Direction from curwp.
 * Direction is DIRECTION_???
 * Value returned might be something like WINDOW_MAP_MSG_LINE - which is not a valid window.
 *
 * There might be more than one window in the Direction, return the window that is in
 * the same row/column as the cursor.
 */
int	WindowNumberInDirection(int Direction, const MapCell* map)
{
	int	i;
	int	rowNo = curwp->w_disrow - curwp->w_toprow;	// cursor row within window
	int	colNo = curwp->w_discol - curwp->w_leftcol;	// cursor col within window

	// These 2 checks should never actually be triggered, but just in case ...
	if(rowNo < 0 || rowNo >= curwp->w_ntrows)
		rowNo = 0;
	if(colNo < 0 || colNo >= curwp->w_ncolumns)
		colNo = 0;

	switch(Direction) {
	case DIRECTION_UP:
		return map[curwp->w_toprow * WINDOW_MAP_COLUMNS + curwp->w_leftcol + colNo + 1];

	case DIRECTION_DOWN:
		return map[curwp->w_toprow * WINDOW_MAP_COLUMNS + curwp->w_leftcol + colNo + 1 + (curwp->w_ntrows + 2) * WINDOW_MAP_COLUMNS];

	case DIRECTION_LEFT:
		// Find a cell in the window at LH of the window:
		i = (curwp->w_toprow + 1 + rowNo) * WINDOW_MAP_COLUMNS + curwp->w_leftcol;

		// If not a leftmost window on the screen there will be a vertical separation - so look one column further
		if(curwp->w_leftcol)
			i--;
		// Could have also done this by checking if the winNo == WINDOW_MAP_SC_NULL

		return map[i];

	case DIRECTION_RIGHT:
		// Find a cell in window at RH of the window:
		i = (curwp->w_toprow + 1 + rowNo) * WINDOW_MAP_COLUMNS + curwp->w_ncolumns + curwp->w_leftcol + 1;

		// If not at RH of the screen, one more for vertical separation
		if(curwp->w_ncolumns + curwp->w_leftcol < AllocColumns)
			i++;

		return map[i];

	}
	/* NOTREACHED */
	return(0);	// Keep compiler quiet
}


/* Fill a rectangle of map cells with Value
 * The rectangle starts at Origin & is within a larger rectangle.
 * Fill in Lines lines of length LineLength.
 * Only part of each line is to be filled in, this is Columns wide.
 */
void	FillMapRectangle(MapCell* Origin, int Lines, int Columns, int LineLength, MapCell Value)
{
	while(--Lines >= 0) {
		int	col = Columns;
		while(--col >= 0)
			*Origin++ = Value;
		Origin += LineLength - Columns;
	}
}

/* Return an array of bytes representing the positions in rows/columns of the screen.
 * Each byte contains the number of the window that occupies that place.
 * Numbering starts at WINDOW_MAP_START and is incremented in the order that the window is
 * found in the linked list.
 *
 * There is an extra row and column all the way round the screen, these cells
 * have the value WINDOW_MAP_SC_EDGE. This means that code does not need to
 * worry too much about the array boundaries.
 *
 * Cells that do not belong to a window are given the value WINDOW_MAP_SC_NULL.
 * The message line is given the value WINDOW_MAP_MSG_LINE.
 *
 * The array returned is malloced and it is the responsibility of the caller to free it.
 *
 *
 *
 * That all appears nice and simple, but it all falls over in FixWindowSizes().
 * This is because we created the map sized at the new screen geometry and then filled in
 * a map using windows that fitted into the old window. OK if the new screen is larger than
 * the old one, but if the old windows are larger than the new screen we scribble all
 * over memory beyond what we malloced.
 *
 * This will happen when we need to resize: either on return from suspend (^z), or on
 * receipt of WINCH, or on return from running a sub command.
 * ****in which case WINCH is delivered - check.
 *
 * So: the map geometry needs to be based on the current size or the previous, which ever
 * is the larger - note things like one could be larger in columns and the other in lines.
 *
 * As far as everything else is concerned use: WINDOW_MAP_COLUMNS & WINDOW_MAP_LINES.
 */
MapCell*BuildWindowMap(void)
{
	int	mapLines = WINDOW_MAP_LINES;	// Lines in map array
	int	lines = mapLines - 2;		// Lines on screen
	int	mapCols = WINDOW_MAP_COLUMNS;	// Columns in map array
	int	columns = mapCols - 2;		// Columns on screen
	MapCell*map = calloc((size_t)mapLines, (size_t)mapCols * sizeof(MapCell));
	MapCell	i;
	Window*	wp;

	if( !map)
		Die(MsgOutOfMemoryWhile, "building Window Map");

	/* No need to fill in the edge (top, bottom & sides) with 0 (== WINDOW_MAP_START)
	 * since calloc() will have done that.
	 *
	 * Fill in the central part (where windows live) with WINDOW_MAP_SC_NULL.
	 * Start with position 1 (pos 0 is edge) on line 1.
	 */
	FillMapRectangle(map + mapCols + 1, lines, columns, mapCols, WINDOW_MAP_SC_NULL);

	// Message line: fill with WINDOW_MAP_MSG_LINE
	FillMapRectangle(map + lines * mapCols + 1, 1, columns, mapCols, WINDOW_MAP_MSG_LINE);

	/* Run down the windows, number their cells.
	 * wmessp is dealt with on the line above.
	 */
	for(i = WINDOW_MAP_START, wp = wheadp; wp; i++, wp = wp->w_wndp) {
		FillWindowMap(map, wp, i);
	}

	// DEBUG
	// MapPrint(map);

	return(map);
}

/* Fill in the window map for an individual window.
 * This is broken out of BuildWindowMap() as, later, when windows are expanded
 * to fill empty space - the map must be updated to avoid a later
 * window being given the same space.
 */
void	FillWindowMap(MapCell* map, Window* wp, MapCell WindowNum)
{
	int	mapCols = WINDOW_MAP_COLUMNS;		// Columns in map array

	// How determine if the window had any ornamentation - ie col to the right ?
	FillMapRectangle(map + ((wp->w_toprow + 1) * mapCols + wp->w_leftcol + 1),
	    wp->w_ntrows + 1, WINDOW_WIDTH(wp), mapCols, WindowNum);
}


/* Set window_map_columns & window_map_lines to be the largest of the current and
 * previous window sizes.
 * lines & columns are the new/latest sizes.
 * The sizes should be the actual number of lines and columns - without reserving anything.
 *
 * Add 2 to columns since we want WINDOW_MAP_SC_EDGE at each end
 * Add 2 to lines since we want WINDOW_MAP_SC_EDGE at the top and bottom
 */
void	setScreenSizeForMap(int lines, int columns)
{
	window_map_lines_prev[1]   = window_map_lines_prev[0];
	window_map_columns_prev[1] = window_map_columns_prev[0];

	window_map_lines_prev[0]   = lines;
	window_map_columns_prev[0] = columns;

	// Work out which is larger of each:
	window_map_lines   = (window_map_lines_prev[0]   > window_map_lines_prev[1]   ? window_map_lines_prev[0]   : window_map_lines_prev[1])   + 2;
	window_map_columns = (window_map_columns_prev[0] > window_map_columns_prev[1] ? window_map_columns_prev[0] : window_map_columns_prev[1]) + 2;
}

/* Return a pointer to the window numbered n
 * Return NULL if not found.
 *
 * n cannot be type MapCell as it might be unsigned char, one of the first
 * things that we do is to subtract from it - which might make it -ve, we don't
 * want it to be 255!
 */
Window*	FindWinByMapno(int n)
{
	Window*	wp;

	if((n -= WINDOW_MAP_START) < 0)
		return(NULL);

	for(wp = wheadp; wp; wp = wp->w_wndp)
		if(n-- == 0)
			return(wp);
	return(wp);
}

/* Return the window number for window wp, counting from 0.
 * Return -1 if it was not found.
 * You will probably have done SortWindows() before this.
 */
int	WindowNumber(Window* wp)
{
	Window*	win;
	int	winNo;

	for(win = wheadp, winNo = 0; win; win = win->w_wndp, winNo++)
		if(win == wp)
			return(winNo);

	return(-1);

}

/* Print the map to the file WMAP
 * Debug use only.
 */
void	MapPrint(MapCell* map)
{
	int	i;

	FILE *f = fopen("WMAP", "a");

	if(!f) {
		perror("WMAP open");
		return;
	}

	fprintf(f,"lines=%d cols=%d\n", window_map_lines, window_map_columns);

	for(i = 0; i < WINDOW_MAP_LINES; i++) {
		fwrite(map + i * WINDOW_MAP_COLUMNS, 1, (size_t)WINDOW_MAP_COLUMNS, f);
		fputs("\n", f);
	}
	fclose(f);
}

/* Free a window pointer.
 * Deal with saving dot and marks if this is the last window onto a buffer.
 *
 * Decrement count of number of windows.
 *
 * This DOES NOT remove it from the chain of window pointers.
 */
void	freeWp(Window* wp)
{
	DisconnectBuffer(wp);
	FreeLine(wp->w_modeline);
	free((void *) wp->w_lineFlags);
	free((void *) wp);
	NumberOfWindows--;
}

/* Allocate a Window that will show ntrows of text,
 * message line contains mslen bytes.
 * NB: ntrows is displayed text rows, this count does not include the mode line.
 *
 * Increment count of number of windows.
 *
 * Initialise the struct members to something vaguely sensible, although many
 * will probably be quickly changed.
 *
 * Return the pointer or NULL on failure.
 */
Window*	allocWindow(int ntrows, LineLn mslen)
{
	Window*	wp;

	if( !(wp = (Window*)calloc(1, sizeof(Window))))
		return((Window*) NULL);

	if(((wp->w_modeline  = lalloc(mslen)) == NULL) ||
	    (wp->w_lineFlags = calloc((size_t)(ntrows + 1), sizeof(Lflags))) == NULL) {
		if(wp->w_modeline)
			FreeLine(wp->w_modeline);
		free((void*)wp);
		return((Window*) NULL);
	}

	NumberOfWindows++;

	wp->w_ntrows = ntrows;

	wp->w_opts = DefaultWindowOpts;
	wp->w_lasttop = wp->w_toplinenum = 1;
	wp->w_attrib = WATT_MODELINE;

#if	LEGACY_DISPLAY
	/* initalize colors to global defaults */
	wp->w_fcolor = (char) gfcolor;
	wp->w_bcolor = (char) gbcolor;
#endif

	return(wp);
}

/* Change the number of rows that a window displays on the screen.
 * The number of lines is the number of text lines, ie does NOT include
 * the window's mode line.
 *
 * This primarily resizes w_lineFlags; update w_ntrows.
 *
 * Return TRUE on success - FALSE on failure to allocate memory.
 */
int	changeWindowRows(Window* wp, int ntrows)
{
	// lineFlags includes the mode line
	if(!(wp->w_lineFlags = realloc(wp->w_lineFlags, (size_t)(ntrows + 1) * sizeof(Lflags))))
		return(FALSE);

	wp->w_ntrows = ntrows;

	return(TRUE);
}

/* Find a window and remove it from the linked list. Then free it
 *
 * Return a pointer to the next in the list. Useful for traversing all
 * windows and removing some.
 */
Window*	findAndFreeWp(Window* wp)
{
	Window**lnp = &wheadp;
	Window*	swp;

	// Find what points to this window:
	for(swp = wheadp; swp != wp; lnp = &swp->w_wndp, swp = swp->w_wndp)
		;

	*lnp = wp->w_wndp;		// Remove from linked list

	freeWp(wp);

	wmessp->w_wndp = wheadp;	// Head window might have changed

	CheckCursorHexSync();

	return *lnp;
}

/* Remove all windows with zero width (w_ncolumns)
 * Zero width may be set as a way of marking it invalid.
 */
void	ZapZeroWidthWindows(void)
{
	Window*	wp;

	for(wp = wheadp; wp; )
		if(wp->w_ncolumns <= 0)
			wp = findAndFreeWp(wp);
		else
			wp = wp->w_wndp;
}

/* The current buffer might be displayed more than once.
 * Ensure that all LINE pointers are to something valid,
 * do this by giving them the same values as has curwp.
 *
 * This can be used after drastic changes have been made
 * to the buffer - eg re-reading it from disk.
 *
 * This is similar to ConnectBuffer() & ReconnectWindowsToBuffer()
 * but that synces to the buffer line, etc, which might not have been
 * recently saved. This also does not fiddle with the top
 * window line - hopefully preventing a reframe.
 *
 * This should use ConnectBuffer()
 */
void	WindowsSync(void)
{
	Window*	wp;

	/* No point if the buffer is only displayed once.
	 * This will generally be the case.
	 */
	if(curbp->b_nwnd == 1)
		return;

	for(wp = wheadp; wp; wp = wp->w_wndp)
		if(wp->w_bufp == curbp && wp != curwp) {
			wp->w_topLinep = curwp->w_topLinep;
			wp->w_dotp = curwp->w_dotp;
			wp->w_doto = curwp->w_doto;
			wp->w_markp = curwp->w_markp;
			wp->w_marko = curwp->w_marko;

			wp->w_nmarkOff = curwp->w_nmarkOff;	// Numeric marks
			memcpy((void*)wp->w_nmarks, (void*)curwp->w_nmarks, sizeof(wp->w_nmarks));

			// Try to avoid reframe of current window:
			wp->w_leftoffset = 0;
			wp->w_flag |= WFMODE|WFHARD|WFFORCE|WFNWID;
		}
}

/* Check if any of the windows open on the same buffer have
 * the WO_HEX flag set.
 * If so set the BSYNC_CURSOR flag on the buffer.
 * The purpose is to keep the cursor on the same displayed characters
 * in all windows on the buffer - one is hex, presumably another is not hex.
 *
 * This is done for all windows. It might be called after window deletion in
 * which case we may be removing links to a window that has gone to the fjords.
 *
 * Ignore wmessp - what it shows could be viewed in hex elsewhere, but syncing
 * would just cause confusion.
 */
void	CheckCursorHexSync(void)
{
	Window* ws;
	Window*	wp;

	// Clear all the flags
	for(ws = wheadp; ws; ws = ws->w_wndp)
		ws->w_bufp->b_flag &= ~BSYNC_CURSOR;

	// Find and count the number of windows & how many with hex option
	for(ws = wheadp; ws; ws = ws->w_wndp)
		for(wp = ws; wp; wp = wp->w_wndp)
			if(ws->w_bufp == wp->w_bufp && ws->w_opts & WO_HEX) {
				ws->w_bufp->b_flag |= BSYNC_CURSOR;
				wp->w_bufp->b_flag |= BSYNC_CURSOR;
			}
}

/* It looks like the same buffer is being viewed in more than one window.
 * Also one of the windows with the hex option.
 * This is only called when curbp's BSYNC_CURSOR is set.
 * Synchronise the position of "." in all these windows to where it is in curwp.
 * Do not synch the mark, each window will have separate marks.
 */
void	synccursor(void)
{
	BUFFER*	bp = curwp->w_bufp;
	Window* wp;

	for(wp = wheadp; wp; wp = wp->w_wndp)
		// If current buffer, but not current window
		if(wp->w_bufp == bp && wp != curwp) {
			// Has it changed line ?
			if(wp->w_dotp != curwp->w_dotp) {
				wp->w_flag |= WFMOVE;
				wp->w_dotp = curwp->w_dotp;
			}

			// Unconditionally set "." and it's partoff
			wp->w_doto = curwp->w_doto;

			// If buffer is UTF8 & window not hex ensure that the cursor
			// is at the start of a multibyte character. Virtual \n is OK:
			if((bp->b_mode & MDUTF8) != 0 && (wp->w_opts & WO_HEX) == 0 && lused(wp->w_dotp) != wp->w_doto)
				UTF8_BACK_FIND_START_NX(lgetc(wp->w_dotp, wp->w_doto), wp->w_doto);

			// partoff is 0 unless both from and to are in HEX option:
			wp->w_partoff = (char)((wp->w_opts & curwp->w_opts & WO_HEX) ? curwp->w_partoff : 0);

			// This causes more work than is often needed, but not often with hex option:
			wp->w_flag |= WFEDIT;
		}
}

/* Does the work of AddWindowOpt() & DelWindowOpt().
 * Prompt the user for an option name and adds/removes it from the current window.
 *
 * If add - add (set) the option, else remove (unset) it.
 * 
 * Return:
 *	TRUE  - the option is added or removed
 *	FALSE - bad option
 *	ABORT C-g
 */
int	SetWindowOption(int add)
{
	char*	prompt;				/* String to prompt user with */
	int	i;
	int	oldopts = curwp->w_opts;

	prompt = add ? "Window option to add: " : "Window option to remove: ";

	/* What does the user want ? */
	if((i = mlreply(prompt, &playValStr, MLUPPER | MLWOPTION, NULL)) != TRUE)
		return(i);

	// ? shows the current window options
	if(vtext(playValStr)[0] == '?') {
		char*	explan = "Window options";
		short	wopts = curwp->w_opts;

		if(vtext(playValStr)[1] == '?') {
			explan = "Default window options";
			wopts = DefaultWindowOpts;
		}

		ShowModes(explan, &WinOpt, wopts);
		return(TRUE);
	}

	if((i = ModeMatch(&WinOpt, vtext(playValStr), (int)tused(vTEXT(playValStr)), FORW, -1)) < 0) {
		mlwrite("%INo such window option!");
		return(FALSE);
	}

	if(add)
		curwp->w_opts |= (short)(  1 << i);
	else
		curwp->w_opts &= (short)(~(1 << i));

	// Some of these are not needed for some options, just easier to do them all:
	CheckCursorHexSync();

	// If hex display changed, set part offsets to 0:
	if((curwp->w_opts ^ oldopts) & WO_HEX) {
		curwp->w_partoff = 0;
		curwp->w_markpartoff = 0;
		curwp->w_lastdismarkpo = 0;

		// Coming out of HEX option ? Ensure that ".", etc, are properly at character start
		if((curwp->w_opts & WO_HEX) == 0)
			utf8_ensurePosStart(curbp, 0);
	}

	// In case there was space reserved for line number:
	if((curwp->w_opts & WO_NUMBER) == 0) {
		curwp->w_lefttextcol = curwp->w_leftcol;
		curwp->w_numwidth = 0;
	}

	// Force window redisplay, check everything:
	curwp->w_flag |= WFHARD | WFMODE | WFNWID;

	return TRUE;
}

/* Add window specific modes.
 *
 * Except that with windows they are called options.
 * window-menu - command 'o'
 *
 * The count is ignored.
 *
 * Command:  add-window-option
 * Bound to: M-o
 *
 * Return:
 *	TRUE  - the option is added
 *	FALSE - bad option
 *	ABORT C-g
 */
int	AddWindowOpt(int f, Cnt n)
{
	return(SetWindowOption(TRUE));
}

/* Delete window specific modes.
 *
 * The count is ignored.
 *
 * Command:  delete-window-option
 * Bound to: M-C-o
 *
 * Return:
 *	TRUE  - the option is removed
 *	FALSE - bad option
 *	ABORT C-g
 */
int	DelWindowOpt(int f, Cnt n)
{
	return(SetWindowOption(FALSE));
}

/* Reposition dot in the current window to line "n".
 *
 * If the argument is positive, it is that line counting from the top of the window.
 *
 * If it is negative it is that line counting from the bottom.
 *
 * If it is 0 the window is centered (this is what the standard
 * redisplay code does).
 *
 * With no argument act as if it is 0.
 * The default count is not set.
 *
 * Command:  redraw-display
 * Bound to: M-C-l
 *
 * Return: TRUE
 */
int	reposition(int f, Cnt n)
{
	if(f == FALSE || n == 0)
		n = curwp->w_ntrows / 2;

	curwp->w_force = (char)n;
	curwp->w_flag |= WFFORCE;

	return(TRUE);
}

/* Refresh the screen.
 *
 * With no argument, it just does the refresh.
 * With an argument it recenters "." in the current window.
 *
 * The default count is not set.
 *
 * Command:  clear-and-redraw
 * Bound to: C-l
 *
 * Return: TRUE
 */
int	refreshScreen(int f, Cnt n)
{
	if(f == FALSE)
		sgarbf = TRUE;
	else {
		curwp->w_force = 0;             /* Center dot. */
		curwp->w_flag |= WFFORCE;
	}

	return(TRUE);
}

/* The command make the next window (next => down the screen) the current
 * window. There are no real errors, although the command does nothing if
 * there is only 1 window on the screen.
 *
 * With multiple columns, the definition of next is more interesting:
 * basically column by column, top to bottom.
 *
 * With an argument this command finds the <n>th window from the top.
 * If n is negative: count from the bottom instead of the top.
 *
 * The default count is not set.
 *
 * Command:  next-window
 * Bound to: C-x C-n
 *
 * Return:
 *	TRUE  - window changed
 *	FALSE - the window asked for does not exist - ie not that many on the screen.
 */
int	nextwind(int f, Cnt n)
{
	register Window* wp;
	register int	 nwindows;		/* total number of windows */

	if(f) {

		/* first count the # of windows */
		wp = wheadp;
		nwindows = 1;
		while(wp->w_wndp != NULL) {
			nwindows++;
			wp = wp->w_wndp;
		}

		/* if the argument is negative, it is the nth window
		   from the bottom of the screen			*/
		if(n < 0)
			n = nwindows + n + 1;

		/* if an argument, give them that window from the top */
		if(n > 0 && n <= nwindows) {
			wp = wheadp;
			while(--n)
				wp = wp->w_wndp;
		} else {
			mlwrite("%WWindow number out of range");
			return(FALSE);
		}
	} else
		if((wp = curwp->w_wndp) == NULL)
			wp = wheadp;

	curwp = wp;
	curbp = wp->w_bufp;
	upmode(NULL, WFMODE);

	return(TRUE);
}

/* This command makes the previous window (previous => up the screen) the
 * current window. There arn't any errors, although the command does not do a
 * lot if there is 1 window.
 *
 * With an argument this command finds the <n>th window from the bottom.
 * If n is negative: count from the top instead of the bottom.
 *
 * The default count is not set.
 *
 * Command:  previous-window
 * Bound to: C-x p
 *
 * Return:
 *	TRUE  - window changed
 *	FALSE - the window asked for does not exist - ie not that many on the screen.
 */
int	prevwind(int f, Cnt n)
{
	register Window* wp1;
	register Window* wp2;

	/* if we have an argument, we mean the nth window from the bottom */
	if(f)
		return(nextwind(f, -n));

	wp1 = wheadp;
	wp2 = curwp;

	// If there is only 1 window, do nothing:
	if(wp1 == wp2)
		wp2 = NULL;

	while(wp1->w_wndp != wp2)
		wp1 = wp1->w_wndp;

	curwp = wp1;
	curbp = wp1->w_bufp;
	upmode(NULL, WFMODE);
	return(TRUE);
}

/* Move the current window down by "arg" lines.
 * Recompute the top line in the window.
 *
 * The move up and move down code is almost completely the same; most of the
 * work has to do with reframing the window, and picking a new dot. We share
 * the code by having "move down" just be an interface to "move up". Magic.
 *
 * The default count is 1. Will be >= 0.
 *
 * Command:  move-window-down
 * Bound to: C-x C-n
 * 
 * Return:
 *	TRUE  moved window up/down by all n lines
 *	FALSE maybe moved a bit, but not by all of them
 */
int	mvdnwind(int f, Cnt n)
{
	return(mvupwind(TRUE, -n));
}

/* Move the current window up by "arg" lines.
 * Recompute the new top line of the window.
 *
 * This command tries to not move "."; it moves the frame.
 *
 * Look to see if "." is still in the window. If it is, you win.
 * If it isn't, then move "." the minimum so that it is visible,
 * ie "." will end up on the top/bottom window line.
 * 
 * The default count is 1.
 * Negative count means move down.
 *
 * This just changes the text displayed in the window, it does not change the
 * position of the window on the screen.
 *
 * So: if move up 1: the top line in the window becomes the 2nd and the previously
 * unseen line above it is shown as the top line.
 *
 * Command:  move-window-up
 * Bound to: C-x C-p  FN4
 *
 * Return:
 *	TRUE  moved window up/down by all n lines
 *	FALSE maybe moved a bit, but not by all of them
 */
int	mvupwind(int f, Cnt n)
{
	register LINE* lp;
	register int i;
	int		up = n > 0;

	// Find line columns. Do before EoB test to set line move in thisflag
	curgoalSet();

	lp = curwp->w_topLinep;

	// Move the top line up/down by n lines:
	if(up) {
		while(n && !SonLine(lback(lp))) {
			lp = lback(lp);
			curwp->w_toplinenum--;
			n--;
		}
	} else {
		while(n && !EonLine(lp)) {
			lp = lforw(lp);
			curwp->w_toplinenum++;
			n++;
		}
	}

	curwp->w_topLinep = lp;
	curwp->w_flag |= WFHARD|WFNWID;	/* Mode line is OK. */

	// Is "." still visible ?
	for(i = 0; i < curwp->w_ntrows; ++i) {
		if(lp == curwp->w_dotp)
			goto ret;		// Yes, leave it alone
		if(SonLine(lp))
			break;
		lp = lforw(lp);
	}

	// Place "." - it has either disappeared off the top or bottom
	if(up) {
		curwp->w_dotp  = lback(lp);
	} else {
		// on the new top line
		curwp->w_dotp  = curwp->w_topLinep;
	}
	curwp->w_doto = getgoal(curwp->w_dotp);

ret:	// Return - compute return status & generate message on fail
	if(n == 0)
		return(TRUE);

	if(up)
		sayTopOf();
	else
		sayEndOf();

	return(FALSE);
}

/* Move the window's view onto the buffer several columns right.
 * Move by 1 / WINDOW_SEVERAL_FRACTION
 *
 * Command name: move-window-several-right
 *
 * Move argument times the fraction.
 *
 * The default count is 1. Will be >= 0.
 *
 * Command:  move-window-several-right
 * Bound to: FN17
 *
 * Return: TRUE
 */
int	moveRightWinSev(int f, Cnt n)
{
	n *= (int)((float)curwp->w_ncolumns / WINDOW_SEVERAL_FRACTION + 0.5F);

	/* In hex option moveRightWind() multiplies by 3
	 * Counteract that to prevent the move being too far.
	 */
	if(curwp->w_opts & WO_HEX)
		n /= 3;

	return(moveRightWind(f, n));
}

/* Move the window's view onto the buffer several columns left.
 * Move by 1 / WINDOW_SEVERAL_FRACTION
 *
 * Command name: move-window-several-left
 *
 * Move argument times the fraction.
 *
 * The default count is 1. Will be >= 0.
 *
 * Command:  move-window-several-left
 * Bound to: FN18
 *
 * Return: TRUE
 */
int	moveLeftWinSev(int f, Cnt n)
{
	return(moveRightWinSev(f, -n));
}

/* Move the window's view onto the buffer several lines down.
 * Move by 1 / WINDOW_SEVERAL_FRACTION
 *
 * Command name: move-window-several-down
 *
 * Move argument times the fraction.
 *
 * The default count is 1. Will be >= 0.
 *
 * Command:  move-window-several-down
 * Bound to: FN15
 *
 * Return:
 *	TRUE  moved window up/down by all n window fractions
 *	FALSE maybe moved a bit, but not by all of them
 */
int	moveDownWinSev(int f, Cnt n)
{
	return(moveUpWinSev(f, -n));
}

/* Move the window's view onto the buffer several lines up.
 * Move by 1 / WINDOW_SEVERAL_FRACTION
 *
 * Command name: move-window-several-up
 *
 * Move argument times the fraction.
 *
 * The default count is 1. Will be >= 0.
 *
 * Command:  move-window-several-up
 * Bound to: FN16
 *
 * Return:
 *	TRUE  moved window up/down by all n window fractions
 *	FALSE maybe moved a bit, but not by all of them
 */
int	moveUpWinSev(int f, Cnt n)
{
	n *= (int)((float)curwp->w_ntrows / WINDOW_SEVERAL_FRACTION + 0.5F);

	return(mvupwind(f, n));
}

/* Move the current window right by count columns.
 *
 * This just changes the text displayed in the window, it does not change the
 * position of the window on the screen.
 *
 * So: if move left 1: the left most column becomes the 2nd column and the
 * previously unseen column to the left is shown as the left most column.
 *
 * Available as command move-window-right
 *
 * This function does the real work.
 *
 * The difficult bit is ensuring that the window is not immediately moved back again.
 * This will happen if "." is too close to the left/right of the window.
 *
 * This allows the user to set the left offset anywhere as long as text remains visible
 * in the window - ie left offset cannot be more than the longest line.
 *
 * Arrange that a movement up/down (page up/down arrow up/down) does not change the
 * offset - thus a file with occasional long lines can be easily inspected.
 *
 * Note that we do not want the maximum left offset to be tied to the length of the
 * current line, but instead the longest that can be seen at the moment in the window,
 * ie some of the lines above and below may be longer.
 * This is because it is irritating to have to move "." onto a nearby longer line
 * to be able to get the desired shift.
 * Don't get this problem with mvupwind()
 *
 * Because we set w_leftoffset and CFCPCN updatedotpos() will not compute a
 * left offset that keeps "." visible. We might need to move "." so that it
 * remains within what is visible of a window.
 *
 * The default count is 1. Will be >= 0.
 *
 * Command:  move-window-right
 * Bound to: FN5
 *
 * Return: TRUE
 */
int	moveRightWind(int f, Cnt n)
{
	LineLn	longest = 0;		// Longest visible line - display units
	int	lines;			// # lines in the window to scan
	LINE*	scan;			// Scan point in the window
	LINE*	dotp = curwp->w_dotp;
	LineLn	savepos = curwp->w_doto;

	// Scan visible lines to determine the longest
	for(scan = curwp->w_topLinep, lines = curwp->w_ntrows; lines && scan; --lines, scan = lforw(scan)) {
		curwp->w_dotp = scan;
		curwp->w_doto = lused(scan);
		LineLn	len = getdcol(1);
		if(len > longest)
			longest = len;
	}

	// Restore "." to as before the scan
	curwp->w_dotp = dotp;
	curwp->w_doto = savepos;

	// Hex window option: move multiples of 'xx '
	if(curwp->w_opts & WO_HEX)
		n *= 3;

	// Change the offset, which cannot become negative
	curwp->w_leftoffset += n;
	if(curwp->w_leftoffset >= longest)
		curwp->w_leftoffset = longest - 1;

	if(curwp->w_leftoffset < 0)
		curwp->w_leftoffset = 0;

	/* This is to stop updatedotpos() from fiddling with the offset that has been worked out above.
	 */
	thisflag |= CFCPCN;

	/* Now work out a w_doto that fits within the left/right bounds of the window.
	 * This is iterative but the adjustment (n) is not likely to be large.
	 */
	while(1) {
		int	pos = (int)(getdcol(0) - curwp->w_leftoffset);

		// Off the LH side of the window ?
		if(pos < 0) {
			// Stop if at end of line, don't want to move to the next line:
			if(curwp->w_doto >= lused(curwp->w_dotp))
				break;

			moveForw(1);	// Move right a character on the line
			continue;
		}

		// Off the RH side of the window ?
		if(pos >= curwp->w_ncolumns) {
			moveBack(1);
			continue;
		}

		// Somewhere in the window. Stop.
		break;
	}

	/* Just as move-window-down/up changes the current and top lines,
	 * this also changes the left offset and current column.
	 */
	curwp->w_goalCol = getdcol(0);
	curwp->w_goalLeftoffset = curwp->w_leftoffset;

	// We scrolled, need to refresh everything in the window:
	curwp->w_flag  |= WFMODE | WFHARD | WFMOVE;

	return TRUE;
}

/* Move the current window leftt by "arg" columns.
 *
 * See moveRightWind() - this does the same but in moving right rather than left.
 *
 * The default count is 1. Will be >= 0.
 *
 * Available as command move-window-leftt
 *
 * Command:  move-window-left
 * Bound to: FN6
 *
 * Return: TRUE
 */
int	moveLeftWind(int f, Cnt n)
{
	return(moveRightWind(TRUE, -n));
}

/* This command makes the current window the only window on the screen.
 *
 * Try to set the framing so that "." does not have to move on the
 * display. Some care has to be taken to keep the values of dot and mark in
 * the buffer structures right if the destruction of a window makes a buffer
 * become undisplayed.
 *
 * wmessp does not count as a window for this purpose.
 *
 * The count is ignored.
 *
 * Command:  delete-other-windows
 * Bound to: C-x 1
 *
 * Return: TRUE
 */
int	onlywind(int f, Cnt n)
{
	register Window*wp;
	register LINE*	lp;
	register int	i;

	// ?? Merge the next 2 loops into one that traverses everything, skipping curwp; end setting wheadp = curwp
	while(wheadp != curwp) {
		wp = wheadp;
		wheadp = wp->w_wndp;
		wmessp->w_wndp = wheadp;	// Message is always before wheadp

		freeWp(wp);
	}

	while(curwp->w_wndp != NULL) {
		wp = curwp->w_wndp;
		curwp->w_wndp = wp->w_wndp;

		freeWp(wp);
	}

	lp = curwp->w_topLinep;
	i  = curwp->w_toprow;

	while(i != 0 && lback(lp) != curbp->b_btopline) {
		--i;
		lp = lback(lp);
		curwp->w_toplinenum--;
	}

	curwp->w_toprow = 0;
	if( !changeWindowRows(curwp, AllocLines - 2))
		return(FALSE);

	curwp->w_leftcol = 0;
	curwp->w_lefttextcol = curwp->w_numwidth;	// Same space for line numbers
	curwp->w_ncolumns = AllocColumns;
	curwp->w_topLinep	= lp;
	curwp->w_flag  |= WFMODE|WFHARD|WFNWID;

	CheckCursorHexSync();

	return(TRUE);
}

/* Choose another buffer and swap to it.
 * Use when killing a window and there is not another ones.
 * Do not swap to an invisible buffer.
 *
 * If there is not a sitable buffer: find "main" (creating if needs be).
 */
int	ChoseAnotherBuffer(void)
{
	BUFFER*	bp;

	/*
	 * Steam down the buffer list for a suitable buffer. At this
	 * stage we're not interested in invisible ones.
	 */
	for(bp = bheadp; bp != (BUFFER*) NULL; bp = bp->b_bufp)
		if(bp == curbp || bp->b_flag&BFINVS)
			continue;
		else
			break;
		
	/* Didnt find a buffer that time, create a main one - just
	 * like if the editor is opened without a file.
	 */
	if(bp == (BUFFER*) NULL &&
	   (bp = bfind(mainBname, BFNCREATE|BFNACTIVE, 0)) == (BUFFER*)NULL)
		return(FALSE);

	if(swbuffer(bp) != TRUE) {
		CheckCursorHexSync();
		return(FALSE);
	}

	CheckCursorHexSync();

	return(TRUE);
}

/* Delete the current window, placing its space in the window above,
 * or, if it is the top window, the window below. Bound to C-X 0.
 *
 * Give the space that the window uses to other windows.
 *
 * We can delete a window if we can find one of it's non-screen edges that
 * abuts the entire edge of one or more other windows. We allocate the freed
 * up space to those windows. If there is a choice, give space to windows in
 * the order: above, below, left, right.
 *
 * A -> B & E
 * B -> E & C - not D
 * C -> E - not F
 * D -> F - not B
 * E -> C
 * F -> D - not C
 *
 * ******************************************
 * *         *                        *     *
 * *         *        B               *  D  *
 * *         *                        *     *
 * *         ********************************
 * *    A    *                 *      *     *
 * *         *        E        *  C   *  F  *
 * *         *                 *      *     *
 * *         *                 *      *     *
 * ******************************************
 *
 * But this is not always possible to kill a window since it might result in
 * non rectangular windows.
 * Eg it is not possible to reallocate the space belonging to M below:
 *
 * ************************************
 * *         *                        *
 * *         *        J               *
 * *         *                        *
 * *         **************************
 * *    I    *                 *      *
 * *         *        M        *      *
 * *         *                 *      *
 * *         *                 *  K   *
 * *****************************      *
 * *                           *      *
 * *                  L        *      *
 * *                           *      *
 * ************************************
 *
 * However: if the window being deleted is at a screen edge it is always
 * possible to give the space to one or more windows, this is what
 * happens in FixWindowSizes() below.
 *
 * If the window is the only one displayed then another buffer will
 * be displayed - this will not kill a buffer.
 *
 * The count is ignored.
 * 
 * Command:  delete-window
 * Bound to: C-x 0
 *
 * Return TRUE  if the window is killed.
 * 	  FALSE if the geography does not allow the window to be killed.
 *		when only one window and "main" cannot be created or the buffer displayed.
 */
/* ARGSUSED */
int	delwind(int f, Cnt n)
{
	MapCell*	map;		/* Map of windows on the screen */
	int		deleted;	/* True if something deleted */

	/* If there is only one window on the screen: error and return.
	 */
	if(wheadp->w_wndp == NULL)
		return(ChoseAnotherBuffer());

	SortWindows();
	map = BuildWindowMap();

	/* Look to see if we can give the screen real estate away in various directions.
	 * Apply some tests, if it fails then try the next direction (ie edge)
	 */

	/* Look at row above the top, next test if:
	 * * screen edge
	 * * Top Left: the window there extends further to the left
	 * * Top Right: the window there extends further to the right
	 *
	 * Look at row below the bottom, same tests as row above top, but to the row below.
	 *
	 * Look at column to the left of left edge, next test if:
	 * * screen edge
	 * * Top Left: the window there extends further up
	 * * Bottom Left: the window there extends further down
	 *
	 * Look at column to the right of the right edge, same tests as column to the left, but to the column right.
	 *
	 * If we get here then the window space cannot be given away, so:
	 * fill all of the window's space with spaces (inc mode line ?).
	 * mark as a NULL window
	 *
	 * If a window is has its space given away, iterate over all NULL windows to try
	 * to remove them. If one is removed restart the scan for NULL windows.
	 *
	 * Example: in the above, if M is NULL and J is deleted, allocate space in 2 steps:
	 * * K & M extended upwards
	 * * I extended right into M
	 * So: NULL space is treated in the same way as other window space as far
	 * as window borders are concerned, ie merging areas. But if NULL is left
	 * then attempts are then made to merge the NULL space into other windows.
	 */

	if(DelMergeWindow(map, DIRECTION_UP)   &&
	   DelMergeWindow(map, DIRECTION_DOWN) &&
	   DelMergeWindow(map, DIRECTION_LEFT) &&
	   DelMergeWindow(map, DIRECTION_RIGHT)) {
		// Cannot give space away, mark NULL, fill with spaces
		mlwrite("%WCannot give window away");
		deleted = FALSE;

		/* It would be nice to try and give parts of a window
		 * away if we cannot give the whole thing away.
		 * TODO
		 */
	} else {
		deleted = TRUE;
		mlwrite("%IWindow zapped");
	}

	free(map);

	CheckCursorHexSync();

	// Fill the space of the zapped window, other window borders may have changed
	upmode((BUFFER*)NULL, WFHARD | WFMODE);

	return(deleted);
}

// map, horis/vert, top/left, 
// horis/vert saying if to test next row/column, true means test next row for horiz line
// top/left is true meaning test row/column to top/left, ie subtract 1 row/column #
// Operates on curwp
// 
/* Delete the curent window, giving its space away to the windows that abut it along edge.
 *
 * map is a screen map of size (AllocLines x AllocColumns)
 *
 * Get the window number in the direction of the nominated edge, - an emphemeral number
 * specific to the map's construction.
 * Find the window that that corresponds to
 *
 * Return:
 *	0	worked
 *	non 0	failed (EnlargeWindXXX error code)
 */
int	DelMergeWindow(MapCell* map, int EdgeToMove)
{
	int	i;
	int	res = -1;	// Init to prevent spurious warning
	Window*	JoiningWp;

	// Which window ?
	if( !(JoiningWp = FindWinByMapno(WindowNumberInDirection(EdgeToMove, map))))
		return 2;

	// Join up/down or left/right; find window height or width
	switch(EdgeToMove) {
	case DIRECTION_UP:
	case DIRECTION_DOWN:
		i = curwp->w_ntrows + 1;	// +1 for mode line
		res = EnlargeWindUpDown(-i, map, 1, EdgeToMove == DIRECTION_DOWN, 0);

		break;

	case DIRECTION_LEFT:
	case DIRECTION_RIGHT:
		// **** need to add 1 if there is a separator to swallow
		i = WINDOW_WIDTH(curwp);
		if(EdgeToMove == DIRECTION_LEFT && curwp->w_ncolumns + curwp->w_leftcol <= AllocColumns)
			i++;

		if(EdgeToMove == DIRECTION_RIGHT )
			i++;

		res = EnlargeWindLeftRight(-i, map, 1, EdgeToMove == DIRECTION_RIGHT, 0);
		break;
	}

	// If that worked: make the window moved to the current window:
	if(res == 0)
		curwp = JoiningWp;

	CheckCursorHexSync();

	return res;
}

/* The code below works but is not used.
 * This is because it works by expanding the window in a particular direction.
 * If that were to expand onto more than one other window - ExlargeWindXXX() does not work.
 */
#if	0
int	DelMergeWindow(MapCell* map, int EdgeToMove)
{
	int	i;
	
	// Which window ?
	// Find the one in the direction of the edge.
	// The subseqent calls to EnlargeWindxxx() are with respect to this window,
	// thus the notion of edge is reversed
	Window*	JoiningWp;
	if( !(JoiningWp = FindWinByMapno(WindowNumberInDirection(EdgeToMove, map))))
		// No window in that direction
		return FALSE;

	Window*	saveCurwp = curwp;
	curwp = JoiningWp;
	// Join up/down or left/right; find window height or width
	switch(EdgeToMove) {
	case DIRECTION_UP:
	case DIRECTION_DOWN:
		i = saveCurwp->w_ntrows + 1;	// +1 for mode line

		if(EnlargeWindUpDown(i, map, 1, EdgeToMove == DIRECTION_UP, 0)) {
			curwp=saveCurwp;
			return(FALSE);
		}
		return(EdgeToMove + 1);

	case DIRECTION_LEFT:
	case DIRECTION_RIGHT:
		// **** need to add 1 if there is a separator to swallow
		i = WINDOW_WIDTH(saveCurwp) + 1;

		if(EnlargeWindLeftRight(i, map, 1, EdgeToMove == DIRECTION_LEFT, 0)) {
			curwp=saveCurwp;
			return(FALSE);
		}
		return(EdgeToMove + 1);
	}
	/*NOTREACHED*/
}
#endif

/* Sort the windows so that starting from wheadp we move top to bottom and left to right.
 * Do a simple bubble sort - speed is not an issue, the number of windows
 * will never be really large.
 *
 * wmessp must not be part of this.
 *
 * Return true if changed the order.
 */
int	SortWindows(void)
{
	int	swapped;
	int	did_swap = 0;

	do {
		Window**	lnp = &wheadp;		// Last next pointer - points to what took us to cwp
		Window*		cwp = wheadp;		// Current window
		Window*		nwp = wheadp->w_wndp;	// Next window

		swapped = 0;

		/* Swap if:
		 * current is to the right of next (compare w_leftcol)
		 * column the same and current is below next (w_toprow)
		 */
		for( ; nwp; lnp = &cwp->w_wndp, cwp = nwp, nwp = nwp->w_wndp) {
			if(cwp->w_leftcol > nwp->w_leftcol ||
			   (cwp->w_leftcol == nwp->w_leftcol && cwp->w_toprow > nwp->w_toprow)) {
				cwp->w_wndp = nwp->w_wndp;
				nwp->w_wndp = cwp;
				*lnp = nwp;
				
				swapped = 1;
				did_swap = 1;
			}
		}

	} while(swapped);

	wmessp->w_wndp = wheadp;	// In case we changed the top window

	return(did_swap);
}

/* Duplicate a window. Use in splitting. Displays the same file/buffer.
 * The window options are copied.
 *
 * The one other place where a Window is made is in edinit().
 *
 * Return:
 *	Window* pointer of new window created
 *	NULL  - max # windows reached
 *		out of memory
 */
Window*	dup_window_on_split(void)
{
	register Window*wp;

	/* Should not really happen, but you never know ...
	 */
	if(curwp->w_attrib & WATT_NOSPLIT) {
		mlwrite("%WThe window cannot be split");
		return((Window*)NULL);
	}

	/* The maximim number of windows is governed by the maximum number that we can map.
	 * Although unlikely to exceed, it is possible as a window can occupy as little
	 * as 4 character positions on a screen (2 if on the left hand edge).
	 * Some twit will try to exceed this.
	 */
	if(NumberOfWindows >= MAX_WINDOW_MAP - (WINDOW_MAP_START - 1)) {
		mlwrite("%WMaximum number of windows reached");
		return((Window*)NULL);
	}

	if( !(wp = allocWindow(curwp->w_ntrows, lused(curwp->w_modeline)))) {
		mlwrite(MsgOutOfMemoryWhile, "splitting window");
		return((Window*)NULL);
	}

	++curbp->b_nwnd;			// Displayed once more

	// Copy struct members over - do so in the order declared - makes checking easier
	// Some were done by allocWindow()
	wp->w_bufp  = curbp;
	wp->w_topLinep = curwp->w_topLinep;
	wp->w_dotp  = curwp->w_dotp;
	wp->w_doto  = curwp->w_doto;
	wp->w_markp = curwp->w_markp;
	wp->w_marko = curwp->w_marko;

	curwp->w_lastdismarkp = (LINE*)NULL;	// Force redisplay
	// So that unmarkmark() does not erroniously think that it has moved
	curwp->w_lastdismarko = 0;

	lcpyl(wp->w_modeline, curwp->w_modeline);

	wp->w_toplinenum = curwp->w_toplinenum;
	wp->w_lasttop = curwp->w_lasttop;
	wp->w_lastnumLines= curwp->w_lastnumLines;
	wp->w_leftoffset = curwp->w_leftoffset;
	wp->w_curcol = curwp->w_curcol;
	wp->w_goalCol = curwp->w_goalCol;
	wp->w_goalLeftoffset = curwp->w_goalLeftoffset;

	memcpy(wp->w_nmarks, curwp->w_nmarks, sizeof(wp->w_nmarks));

	wp->w_leftcol = curwp->w_leftcol;
	wp->w_lefttextcol = curwp->w_lefttextcol;
	wp->w_numwidth = curwp->w_numwidth;
	wp->w_ncolumns = curwp->w_ncolumns;
	wp->w_disrow = curwp->w_disrow;
	wp->w_discol = curwp->w_discol;
	wp->w_lastcurcol = wp->w_lastcurcol;
	wp->w_lastcurlen = curwp->w_lastcurlen;
	wp->w_toprow = curwp->w_toprow;
	wp->w_opts = curwp->w_opts;
	wp->w_cursattr = curwp->w_cursattr;
	wp->w_partoff  = 0;
	wp->w_markpartoff = 0;
	wp->w_lastdismarkpo = curwp->w_lastdismarkpo;
	wp->w_force = 1;
	wp->w_flag  = WFNWID;
	wp->w_attrib = curwp->w_attrib;
	wp->w_nmarkOff = curwp->w_nmarkOff;

	CheckCursorHexSync();

	return(wp);
}

/* Split the current window either horizontally or vertically.
 *
 * Split the current window horizontally if wider than $pref-col-split
 *
 * < 3 lines cannot split vertically
 *
 * Any count is passed to splitwindhoriz() or splitwindvert() TODO
 *
 * The default count is not set.
 *
 * Command:  split-window
 * Bound to: unbound
 *
 * Return: TODO
 */
int	splitwind(int f, Cnt n)
{
	if(curwp->w_ntrows < (MIN_WINDOW_LINES * 2) - 1 ||
	   curwp->w_ncolumns > PrefSplitCols)
		return(splitwindhoriz(f, n));

	return(splitwindvert(f, n));
}

/* Split the current window vertically - top bottom.
 * A window smaller than 3 lines cannot be split.
 *
 * "." will end up in part of the screen (upper/lower window) that it was in
 * before this command.
 *
 * An argument of 1 (or not set) forces "." into the upper window,
 * an argument of not 1 forces "." to the lower window.
 *
 * The default count is not set.
 *
 * The only other error that is possible is a "malloc" failure allocating the
 * structure for the new window.
 *
 * Command:  split-current-window   split-window-vertically
 * Bound to: unbound
 *
 * Return TRUE on success.
 */
int	splitwindvert(int f, Cnt n)
{
	Window*	wp;
	LINE*	lp;
	int	ntru;	// New upper lines
	int	ntrl;	// New lower lines
	int	ntrd;	// Line # of "." in window
	Window*	wp1;
	Window*	wp2;
	int	ret = TRUE;

	if(curwp->w_ntrows < (MIN_WINDOW_LINES * 2) - 1) {
		// Can split 3 line as the mode line is not included in w_ntrows.
		mlwrite("%WCannot split a %d line window", curwp->w_ntrows);
		return(FALSE);
	}

	if( !(wp = dup_window_on_split()))
		return(FALSE);

	ntru = (curwp->w_ntrows-1) / 2;		/* Upper size		*/
	ntrl = (curwp->w_ntrows-1) - ntru;	/* Lower size		*/

	// Find the number of window line that contains "."
	lp = curwp->w_topLinep;
	ntrd = 0;
	while(lp != curwp->w_dotp) {
		++ntrd;
		lp = lforw(lp);
	}

	lp = curwp->w_topLinep;
	if(((f == FALSE) && (ntrd <= ntru)) || ((f == TRUE) && (n == 1))) {
		/* Old is upper window. */

		if(ntrd == ntru) {		/* Hit mode line.	*/
			lp = lforw(lp);
			curwp->w_toplinenum++;
		}

		wp->w_wndp = curwp->w_wndp;	// Stitch in to chain of windows
		curwp->w_wndp = wp;

		wp->w_toprow = curwp->w_toprow + ntru + 1;	// Set size & screen position of new window
		ret = changeWindowRows(wp, ntrl);	// Set new size
		if( !changeWindowRows(curwp, ntru))	// Set new size
			ret = FALSE;

	} else {
		/* Old is lower window	*/

		// Stitch into place before curwp:
		wp1 = NULL;
		wp2 = wheadp;
		while(wp2 != curwp) {
			wp1 = wp2;
			wp2 = wp2->w_wndp;
		}
		if(wp1 == NULL) {
			wheadp = wp;
			wmessp->w_wndp = wheadp;	// Message is always before wheadp
		} else {
			wp1->w_wndp = wp;
		}
		wp->w_wndp   = curwp;

		ret = changeWindowRows(wp, ntru);
		++ntru;				/* Mode line.		*/
		curwp->w_toprow = curwp->w_toprow + ntru;
		if( !changeWindowRows(curwp, ntrl))
			ret = FALSE;

		// I think that this is supposed to keep the displayed lines in the same
		// place on the screen.
		if(ntrd > ntru)
			while(ntru--) {
				lp = lforw(lp);
				curwp->w_toplinenum++;
			}
	}

	curwp->w_topLinep = lp;			/* Adjust the top lines */
	wp->w_topLinep = lp;			/* if necessary.	*/
	wp->w_toplinenum = curwp->w_toplinenum;
	curwp->w_flag |= WFMODE|WFHARD|WFFORCE|WFNWID;
	wp->w_flag    |= WFMODE|WFHARD|WFNWID;
	upmode(NULL, WFMODE);			/* All, the above 2 not enough */

	SortWindows();

	CheckCursorHexSync();

	return(ret);
}

/* Split current window left/right into two windows side by side.
 *
 * The window is split evenly to create the new ones.
 * If the count is set the new window gets argument columns.
 * Negative argument means the old window gets -n columns.
 * The selected window always becomes the right one - which (visually)
 * is the new one.
 *
 * Window size (n) includes the width of the divider column.
 *
 * The default count is unset.
 *
 * Command:  split-horizontally
 * Bound to: C-x 3
 *
 * Return:
 *	TRUE  - on success
 *	FALSE - a window would be too small
 *		max # windows reached
 *		out of memory
 */
int	splitwindhoriz(int f, Cnt n)
{
	int	left_size;	// Columns for left hand window
	int	right_size;	// Columns for right hand window
	int	cur_size;
        Window*	wp;

	cur_size = WINDOW_WIDTH(curwp);

	// Compute new left hand window size
	if(f)
		left_size = n > 0 ? (int)n : (cur_size + (int)n);
	else
		left_size = cur_size / 2;

	// New right hand window gets the rest
	right_size = cur_size - left_size;

	/* Small subtlety here.
	 * Out of the space allocated to the LH window we need to grab a column for
	 * the separator. Ensure that we have enough room.
	 */
	if(left_size <= MIN_WINDOW_COLS || right_size < MIN_WINDOW_COLS) {
		mlwrite("%WCannot make a window less than %d wide", MIN_WINDOW_COLS);
		return(FALSE);
	}

	if( !(wp = dup_window_on_split()))
		return(FALSE);

	wp->w_wndp = curwp->w_wndp;
	curwp->w_wndp = wp;

	/* Work out where it goes on the screen. The new window is on the right.
	 * Row numbers are the same, just columns are different.
	 * The number of columns might not be the same, either because that is what was
	 * requested or integer division by 2 leaves one 1 column wider.
	 *
	 * Leave w_doto the same - this might result in scrolling - ie change in w_leftoffset
	 */
	wp->w_leftcol += left_size;
	wp->w_lefttextcol += left_size;
	wp->w_ncolumns = right_size;

	// One column left for separator
	curwp->w_ncolumns = left_size - 1;

        curwp->w_flag |= WFMODE|WFHARD;
        wp->w_flag    |= WFMODE|WFHARD;

	// Make the current window the one just created:
	curwp = wp;
	// curbp already points to the right place - since we are using the same buffer

	SortWindows();

	CheckCursorHexSync();

	return(TRUE);
}

/* Find a buffer into another window, split the current window left/right
 *
 * The count is passed to swbufferPrompt() TODO
 * The default count is unset.
 *
 * Command:  split-horizontally-and-find-buffer
 * Bound to: C-x 3
 *
 * Return TRUE  - success
 *	  FALSE - ??
 */
int	splitwindhorizbuf(int f, Cnt n)
{
	return(swbufferPrompt(f, n, splitwindhoriz));
}

/* Enlarge the current window (curwp) top edge in the up direction by n.
 *
 * If n is -ve the window is shrunk by -n lines.
 * The count n includes the mode line.
 *
 * Do leave the window from which we steal lines with enough to operate.
 * Similarly the window cannot be shrunk by more lines than it has.
 *
 * If there are more than 1 windows above, then steal the same number of lines
 * all of them.
 *
 * If obliterate - take all lines from the window(s) that are being stolen from
 * (including the mode line) then delete the window.
 * This routine trusts that if obliterate that n is the correct number to completely
 * eat the windows being obliterated.
 *
 * If bottom then expand using the bottom edge. Variables 'topXX' will contain
 * the value at the bottom instead.
 *
 * If TestMovable return 0 if the edge is movable, else return a value below.
 * Do not do anything.
 * In the way that this is currently used the same effect could be had by setting n
 * to 0 - but checking if arbitrary changes would work could be useful.
 *
 * Return:
 * 	0 success
 *	1 window end points do not match this windows end points
 *	2 cannot expand in that direction (ie screen edge)
 *	3 assert error - should not happen
 *	< 0-max-1 (ie -ve number) # lines that could be taken (or window size if shrinking)
 *	-1 out of memory
 */
int	EnlargeWindUpDown(int n, MapCell* map, int obliterate, int bottom, int TestMovable)
{
	Window*	wp;
	int	topLeftLoc;
	MapCell	tlWin;
	int	maxLines = n;
	int	lastWinNo;
	int	i;

	/* Do not leave a window with fewer than a minimum of lines.
	 * If we are obliterating it - take the mode line as well -- thus -ve reserve.
	 */
	int	reserveLines = obliterate ? -1 : MIN_WINDOW_TEXT_LINES;

	// Get location of a cell in last line of the window above:
	topLeftLoc = curwp->w_toprow * WINDOW_MAP_COLUMNS + curwp->w_leftcol + 1;

	if(bottom)	// Maybe make that top line of the window below:
		topLeftLoc += (curwp->w_ntrows + 2) * WINDOW_MAP_COLUMNS;

	tlWin = map[topLeftLoc];

#if 0
	// Debug
	map[topLeftLoc]+= '@';
	map[topLeftLoc - 1]+= '@';
	map[topLeftLoc + curwp->w_ncolumns - 1]+= '@';
	map[topLeftLoc + curwp->w_ncolumns]+= '@';
	MapPrint(map);
	return 3;
#endif

	// Screen edge or bottom of screen message line - can't go there:
	if(tlWin == WINDOW_MAP_SC_EDGE || tlWin == WINDOW_MAP_MSG_LINE)
		return(2);

	// Check that left/right edges of window(s) above align with ours:
	if(tlWin == map[topLeftLoc - 1] || map[topLeftLoc + curwp->w_ncolumns - 1] == map[topLeftLoc + curwp->w_ncolumns])
		return(1);

	if(n == 0)	// Not strictly necessary, but makes me feel happier
		return(0);

	if(n > 0) {
		// Making the window bigger.
		// Check that all of the windows above can take the hit of n lines stolen.
		// Check every column across - we don't know how it will change.
		lastWinNo = 0;
		for(i = 0; i < WINDOW_WIDTH(curwp); i++) {
			// Already checked this one ?
			if(lastWinNo == map[topLeftLoc + i])
				continue;
			lastWinNo = map[topLeftLoc + i];

			// NULL is what we get between side by side windows, this is OK to eat
			if(lastWinNo == WINDOW_MAP_SC_NULL)
				continue;

			wp = FindWinByMapno(lastWinNo);
			if(!wp) {
				// Really an assert error
				printf("EUD wp = NULL i=%d lwn=%d tLL=%d  ", i, lastWinNo, topLeftLoc);
				map[topLeftLoc] = '*';
				MapPrint(map);
				return(3);
			}

			// Is it enough, if not have we seen less ?
			if(n > wp->w_ntrows - reserveLines)
				if(wp->w_ntrows - reserveLines < maxLines)
					maxLines = wp->w_ntrows - reserveLines;
		}

		// Can't because not enough lines ?
		if(maxLines != n)
			return(-maxLines - 1);

	} else {
		// Shrinking the window.
		// Just check that the amount that we have been asked to shrink it by
		// will leave the minimum required for a window.
		if(-n > curwp->w_ntrows - reserveLines)
			return( -(curwp->w_ntrows - reserveLines + 1));

	}

	if(TestMovable)	// Did the caller just want to know if movement was possible ?
		return(0);

	// All of the windows above have rows that we can take.
	// Trim them all down by n rows
	lastWinNo = 0;
	for(i = 0; i < WINDOW_WIDTH(curwp); i++) {
		// Already modified this one ?
		if(lastWinNo == map[topLeftLoc + i])
			continue;
		lastWinNo = map[topLeftLoc + i];

		if(lastWinNo == WINDOW_MAP_SC_NULL)
			continue;

		wp = FindWinByMapno(lastWinNo);
		if(!wp) {
			// Really an assert error
			printf("EUD wp = NULL i=%d lwn=%d tLL=%d  ", i, lastWinNo, topLeftLoc);
			map[topLeftLoc] = '*';
			MapPrint(map);
			return(3);
		}

		if( !changeWindowRows(wp, wp->w_ntrows - n))
			return(-1);	// Eeek!

		if(bottom)	// bottom edge: the windows stolen from start lower on the screen
			wp->w_toprow = wp->w_toprow + n;
		wp->w_flag |= WFMODE|WFHARD;

		if(obliterate && n > 0) {
			// The number of rows will not be zero as we ate the mode line as well.
			// Not -ve since w_ntrows is unsigned.
			// Set columns to zero as a flag.
			// Can't free now as it would screw the map.
			wp->w_ncolumns = 0;
		}
		// Need to ensure that the cursor is still displayed - steal code
	}

	// Find windows that we obliterated:
	if(obliterate)
		ZapZeroWidthWindows();

	if(obliterate && n < 0) {
		findAndFreeWp(curwp);
	} else {
		// Increase the rows of the current window by n
		if( !changeWindowRows(curwp, curwp->w_ntrows + n))
			return(-1);	// Eeek!

		curwp->w_flag |= WFMODE|WFHARD;
		if( !bottom)
			curwp->w_toprow = curwp->w_toprow - n;
	}

	return(0);
}

/* Make a window wider/narrower.
 *
 * API as EnlargeWindUpDown() except that the orientation is different.
 *
 * Return:
 *	0 - success
 *	????
 *	2 - screen edge met
 */
int	EnlargeWindLeftRight(int n, MapCell* map, int obliterate, int right, int TestMovable)
{
	Window*	wp;
	int	topLeftLoc;
	MapCell	tlWin;
	int	maxColumns = n;
	int	lastWinNo;
	int	i;
	int	width;

	/* Do not leave a window with fewer than a minimum of columns.
	 * If we are obliterating it - take the separator column as well -- thus -ve reserve.
	 */
	int	reserveColumns = obliterate ? 0 : MIN_WINDOW_COLS;

	// Getting horizontal cell addresses is made more complicated because there will be
	// a vertical separator between windows that are horizontally side by side (left/right).

	// Get location of a cell in rightmost column of the window to the left
	topLeftLoc = (curwp->w_toprow + 1) * WINDOW_MAP_COLUMNS + curwp->w_leftcol;

	width = WINDOW_WIDTH(curwp);
	if(right) {
		// Maybe make that leftmost column of the window to the right:
		topLeftLoc += width + 1;
		// If not at RH of the screen, one more for vertical separation
		if(curwp->w_ncolumns + curwp->w_leftcol < AllocColumns) {
			topLeftLoc++;
			width++;

			// If we have a left separator - reserve its space
			if( !obliterate && n < 0)
				reserveColumns++;

		}
	} else {
		// If that is a vertical separator, go back over it, increment width
		if(curwp->w_leftcol) {
			topLeftLoc--;
			width++;
		}
		if(curwp->w_leftcol && ! obliterate)
			reserveColumns++;
	}

	tlWin = map[topLeftLoc];

#if 0
	// Debug
	map[topLeftLoc]+= '@';
	map[topLeftLoc + width + 1]+= '@';
	map[topLeftLoc + curwp->w_ntrows * WINDOW_MAP_COLUMNS]+= '@';
	map[topLeftLoc + width + curwp->w_ntrows * WINDOW_MAP_COLUMNS + 1]+= '@';
	MapPrint(map);
	return 3;
#endif

	// Screen edge - can't go there:
	if(tlWin == WINDOW_MAP_SC_EDGE)
		return(2);

	// Check that top/bottom edges of window(s) to the left align with ours:
	if(tlWin == map[topLeftLoc - WINDOW_MAP_COLUMNS] || map[topLeftLoc + curwp->w_ntrows * WINDOW_MAP_COLUMNS] == map[topLeftLoc + (curwp->w_ntrows + 1) * WINDOW_MAP_COLUMNS])
		return(1);

	if(n == 0)	// Not strictly necessary, but makes me feel happier
		return(0);

	if(n > 0) {
		// Making the window bigger.
		// Check that all of the windows to the left can take the hit of n columns stolen.
		// Check every row down - we don't know how it will change.
		lastWinNo = 0;
		for(i = 0; i < curwp->w_ntrows + 1; i++) {
			int	thisWidth;

			// Already checked this one ?
			if(lastWinNo == map[topLeftLoc + i * WINDOW_MAP_COLUMNS])
				continue;
			lastWinNo = map[topLeftLoc + i * WINDOW_MAP_COLUMNS];

			wp = FindWinByMapno(lastWinNo);
			if(!wp) {
				// Really an assert error
				printf("ELR wp = NULL i=%d lwn=%d tLL=%d  ", i, lastWinNo, topLeftLoc);
				map[topLeftLoc] = '*';
				MapPrint(map);
				return(3);
			}

			// Is it enough, if not have we seen less ?
			thisWidth = WINDOW_WIDTH(wp);

			if( !right)	// We have a separator to eat
				thisWidth++;

			if(obliterate && right)
				thisWidth++;

			if(n > thisWidth - reserveColumns)
				if(thisWidth - reserveColumns < maxColumns)
					maxColumns = thisWidth - reserveColumns;
		}

		// Can't because not enough columns ?
		if(maxColumns != n)
			return(-maxColumns - 1);
	} else {
		// Shrinking the window.
		// Just check that the amount that we have been asked to shrink it by
		// will leave the minimum required for a window.
		if(-n > width - reserveColumns)
			return( -(width - reserveColumns + 1));
	}

	if(TestMovable)	// Did the caller just want to know if movement was possible ?
		return(0);

	// All of the windows to the left have columns that we can take.
	// Trim them all down by n columns
	lastWinNo = 0;
	for(i = 0; i < curwp->w_ntrows + 1; i++) {
		// Already modified this one ?
		if(lastWinNo == map[topLeftLoc + i * WINDOW_MAP_COLUMNS])
			continue;
		lastWinNo = map[topLeftLoc + i * WINDOW_MAP_COLUMNS];

		// This should not happen. Test stops compiler warnings
		if((wp = FindWinByMapno(lastWinNo)) == NULL)
			return(3);

		// Adjust width
		wp->w_ncolumns -= n;

		// If this ends up leaving the last column of the window just before the
		// last screen column then extend to cover that.
		// If we don't do this then after some splits and joins we end up with multiple dividers on the RH side
		// This will only happen with obliterate - else min width protection applies.
		// Obliterate always with -ve n
		// Only needed when joining right
		if(right && obliterate && wp->w_leftcol + wp->w_ncolumns == AllocColumns - 1)
			wp->w_ncolumns++;

		if(right) {	// Going right: the windows stolen from start further to the right on the screen
			wp->w_leftcol += n;
			wp->w_lefttextcol += n;

			// Just as we can end up with a column on the right, so we can
			// lose the 0th column of the screen. Fix that:
			if(wp->w_leftcol == 1) {
				wp->w_leftcol--;
				wp->w_lefttextcol--;
				wp->w_ncolumns++;
			}

		}
		wp->w_flag |= WFMODE|WFHARD;

		if(obliterate && n > 0) {
			// Set columns to zero as a flag that this window is toast.
			// Can't free now as it would screw the map.
			wp->w_ncolumns = 0;
		}
	}
	// Find windows that we obliterated:
	if(obliterate)
		ZapZeroWidthWindows();

	if(obliterate && n < 0) {
		// Throw the current window away
		findAndFreeWp(curwp);
	} else {
		// Increase the columns of the current window by n
		curwp->w_ncolumns += n;
		curwp->w_flag |= WFMODE|WFHARD;
		if( !right) {
			curwp->w_leftcol -= n;
			curwp->w_lefttextcol -= n;
		}
	}

	return(0);
}

/* Argument is the return value from EnlargeWindLeftRight() or EnlargeWindUpDown()
 * Print an error message if this is not 0
 * Return the argument.
 */
int	EnlargeWindMsg(int res)
{
	// Returns 0 on OK
	if(res) {
		if(res < 0)
			mlwrite("%WCould only expand by %d", -(res - 1));
		else
			mlwrite(res == 1 ? "%WWindow ends do not match those of this window" :
					   "%WCannot expand beyond screen edge");
	}

	return res;
}

/* Move the top edge of the window by n
 *
 * Default count 1. May be -ve.
 *
 * Command:  window-top-edge-move
 * Bound to: unbound
 *
 * Return:
 * 	0 success
 *	1 window end points do not match this windows end points
 *	2 cannot expand in that direction (ie screen edge)
 *	< 0-max-1 # lines that could be taken (or window size if shrinking)
 *	-1 out of memory
 */
int	WindowMoveTop(int f, Cnt n)
{
	int	res;
	MapCell*map = BuildWindowMap();

	res = EnlargeWindUpDown((int)n, map, 0, 0, 0);

	free(map);

	return res;
}

/* Move the bottom edge of the window by n.
 *
 * Default count 1. May be -ve.
 *
 * Command:  window-bottom-edge-move
 * Bound to: unbound
 *
 * Return:
 * 	0 success
 *	1 window end points do not match this windows end points
 *	2 cannot expand in that direction (ie screen edge)
 *	< 0-max-1 # lines that could be taken (or window size if shrinking)
 *	-1 out of memory
 */
int	WindowMoveBottom(int f, Cnt n)
{
	int	res;
	MapCell*map = BuildWindowMap();

	res = EnlargeWindUpDown((int)n, map, 0, 1, 0);

	free(map);

	return res;
}

/* Move the left edge of the window by n
 *
 * Default count 1. May be -ve.
 *
 * Command:  window-left-edge-move
 * Bound to: unbound
 *
 * Return:
 * 	0 success
 *	???
 */
int	WindowMoveLeft(int f, Cnt n)
{
	int	res;
	MapCell*map = BuildWindowMap();

	res = EnlargeWindLeftRight((int)n, map, 0, 0, 0);

	free(map);

	return res;
}

/* Move the right edge of the window by n
 *
 * Default count 1. May be -ve.
 *
 * Command:  window-right-edge-move
 * Bound to: unbound
 *
 * Return:
 * 	0 success
 *	???
 */
int	WindowMoveRight(int f, Cnt n)
{
	int	res;
	MapCell*map = BuildWindowMap();

	res = EnlargeWindLeftRight((int)n, map, 0, 1, 0);

	free(map);

	return res;
}

/* Enlarge the current window.
 *
 * Find the window that loses space, this is the window above unless there is
 * not one there in which case use the window below.
 *
 * Make sure the found window is big enough. If so, hack the window
 * descriptions, and ask redisplay to do all the hard work. You don't just set
 * "force reframe" because dot would move.
 *
 * Default count 1.
 *
 * Command:  grow-window
 * Bound to: C-x CIRC  C-x z
 *
 * Return:
 * 	TRUE  - success
 *	FALSE - There is only one window - a message will say this
 *		Cannot make window expand beyond screen edge
 *	This is not right TODO
 */
int	enlargewind(int f, Cnt n)
{
	register int	i, j;
	MapCell*	map;		/* Map of windows on the screen */

	if(wheadp->w_wndp == NULL) {
		mlwrite("%WOnly one window");
		return(FALSE);
	}

	map = BuildWindowMap();

	if( !(i = EnlargeWindUpDown((int)n, map, 0, 0, 0)) ||	// Try up
	    !(j = EnlargeWindUpDown((int)n, map, 0, 1, 0))) {	// Try down
		free(map);
		SortWindows();
		return(TRUE);
	}

	// Failed, try to print a sensible error message
	if(i < 0 && j < 0)
		mlwrite("%WCould only expand by %d lines", -(i > j ? j : i));
	else
		mlwrite(i == 1 ? "%WWindow ends do not match those of this window" :
				 "%WCannot expand beyond screen edge");

	free(map);

	return(FALSE);
}

/* Shrink the current window.
 *
 * Find the window that gains space. Hack at the window descriptions. Ask the
 * redisplay to do all the hard work.
 *
 * Default count 1.
 *
 * Command:  shrink-window
 * Bound to: C-x C-z
 *
 * Return:
 * 	0 success
 *	???
 */
int	shrinkwind(int f, Cnt n)
{
	return(enlargewind(f, -n));
}

/* Move to the window above the current one.
 *
 * The count is ignored.
 *
 * Command:  move-to-window-up
 * Bound to: M-up-arrow
 *
 * Return:
 *	TRUE  success
 *	FALSE on error
 */
int	MoveToWindowUp(int f, Cnt n)
{
	return(MoveToWindow(DIRECTION_UP));
}

/* Move to the window below the current one.
 * 
 * The count is ignored.
 *
 * Command:  move-to-window-down
 * Bound to: M-up-arrow
 *
 * Return:
 *	TRUE  success
 *	FALSE on error
 */
int	MoveToWindowDown(int f, Cnt n)
{
	return(MoveToWindow(DIRECTION_DOWN));
}

/* Move to the window to the left of the current one.
 *
 * The count is ignored.
 *
 * Command:  move-to-window-left
 * Bound to: M-left-arrow
 *
 * Return:
 *	TRUE  success
 *	FALSE on error
 */
int	MoveToWindowLeft(int f, Cnt n)
{
	return(MoveToWindow(DIRECTION_LEFT));
}

/* Move to the window to the right of the current one.
 * 
 * The count is ignored.
 *
 * Command:  move-to-window-right
 * Bound to: M-right-arrow
 *
 * Return:
 *	TRUE  success
 *	FALSE on error
 */
int	MoveToWindowRight(int f, Cnt n)
{
	return(MoveToWindow(DIRECTION_RIGHT));
}

/* Move to the window in Direction from the current one.
 *
 * Direction is one of: DIRECTION_UP, DIRECTION_DOWN, DIRECTION_LEFT, DIRECTION_RIGHT
 *
 * Return:
 *	TRUE  success
 *	FALSE on error
 */
int	MoveToWindow(int Direction)
{
	int		winNo;		// Number of window to move to
	MapCell*	map;		// Map of windows on the screen
	Window*		win;
	int		ret = TRUE;

	SortWindows();
	map = BuildWindowMap();
	winNo = WindowNumberInDirection(Direction, map);

	/* If we find a window - switch to it
	 * Refresh all mode lines & window borders, two will be different.
	 */
	if((win = FindWinByMapno(winNo)) != (Window*)NULL) {
		curwp = win;
		upmode(NULL, WFMODE);
	} else {
		// Nothing there:
		ret = FALSE;
	}

	free(map);

	return(ret);
}

/* Interactive command to manipulate windows.
 * Commands:
 * e	- move window edge, then choose edge, then move it
 * m	- move text within window
 * c	- close window
 * k	- kill buffer (if not used elsewhere) then close window
 * s	- split window, then 'h' or 'v'
 * j	- join window, then choose edge to join onto whatever is in that direction
 * n	- next window
 * b	- next buffer
 * p	- previous window
 * o	- set/unset window option
 * q ^g	- quit
 * ^u	- count multiply by 4
 * 0-9	- enter count
 * -	- negate count
 *
 * The count is ignored.
 *
 * Command:  window-menu
 * Bound to: C-x w
 *
 * Returns: TRUE
 */
// The possible menus/modes for resize()
#define	WIN_MANIP_MAIN		0	// Main menu
#define	WIN_MANIP_EDGE		1	// Choose edge
#define	WIN_MANIP_EDGE_MOVE	2	// Move an edge
#define	WIN_MANIP_SPLIT		3	// Split window horizontally/vertically
#define	WIN_MANIP_JOIN		4	// Join along an edge
#define	WIN_MANIP_MOVE		5	// Move text within window
#define	WIN_MANIP_OPTION	6	// Set/Unset window option

// Structure of info on the modes:
typedef	struct	{
	int	wm_flag;
#define	WM_DIRN			0x01	// Parse input for a direction
#define	WM_GOOD_DIRN		0x02	// Error if the direction is not good
#define	WM_SET_DIR		0x04	// Set moveDirection else set EdgeToMove
	char*	wm_prompt;		// Prompt string, NULL means in msg[]
	char*	wm_help;		// Help message
} wm_mode;

/* Info on the modes
 * Ordered by WIN_MANIP_xxx
 */
static	wm_mode	ModeInfo[] = {
	{ WM_DIRN| WM_SET_DIR,			"? for help",
						"Split, Join, Move, Edge, Next/Prev, Up/Down/Left/Right/arrow, Opt, Close, Kill, Buffer, Quit" },
	{ WM_DIRN|WM_GOOD_DIRN,			"Edge for window move. Choose edge: ",
						"Set edge by Top/Bottom/Left/Right or arrows. Or Quit" },
	{ WM_DIRN|WM_GOOD_DIRN|WM_SET_DIR,	NULL,
						"Move edge by Top/Bottom/Left/Right or arrows. Or Quit" },
	{ 0,					"Split: ",
						"Split: H or V for Horizontally or Vertically. Or Quit" },
	{ WM_DIRN|WM_GOOD_DIRN,			"Join on edge: ",
						"Give edge to join on by Top/Bottom/Left/Right or arrows. Or Quit" },
	{ WM_DIRN|WM_GOOD_DIRN,			"Move text in window. Choose direction: ",
						"Move text within window by Top/Bottom/Left/Right or arrows. Or Quit" },
	{ 0,					"Option Set/Unset: ",
						"Window options: S or U to Set or Unset. Or Quit" },
};

int	windowManipulate(int f, Cnt n)
{
	BYTE		msg[MAXCOL] = "";
	int		winNo;				// Current window number
	Window*		win;
	UniChar		ch;				// Read from keyboard
	int		cmdMode = WIN_MANIP_MAIN;	// Where we enter a sub command
	int		moveDirection = DIRECTION_UP;	// A DIRECTION_xx value
	int		EdgeToMove = DIRECTION_UP;	// A DIRECTION_xx value
	int		i = 0;				// Init to keep compiler quiet
	int		flag;
	char*		Help = NULL;			// Show as help - if not NULL
	MapCell*	map;				// Map of windows on the screen
	int		countMl = mlmsgcount;		// Detect (error) message
	int		Alarm = 0;			// Set true to sound the alarm
	int		Arg = FALSE;
	int		Count = 0;

	char*		mp;				// Points to message - in prompt

	while(1) {
		if( !Arg)
			Count = 0;

		SortWindows();	// Prob overkill

		// What window # is current ?
		for(win = wheadp, winNo = 0; win; win = win->w_wndp, winNo++)
			if(win == curwp)
				break;

		// Set message pointer, if NULL then in msg:
		if( !(mp = ModeInfo[cmdMode].wm_prompt))
			mp = msg;

		// Generate a prompt if there has not been an (error) message
		if(countMl == mlmsgcount) {
			if(Help) {
				mlwrite("%s", Help);
				Help = NULL;
			} else if(Arg) {
				mlwrite("%ICount %d %s", Count, mp);
			} else {
				// Info about the window & some message
				// The size shown is the size of the text area, ie exclude mode and any separator line
				mlwrite("%IWindow %d size: %dx%d@%dx%d >>%d %s %s", winNo, WINDOW_WIDTH(curwp), curwp->w_ntrows,
					curwp->w_toprow, curwp->w_leftcol, curwp->w_leftoffset,
					curwp->w_bufp->b_bname, mp);
			}
		}

		countMl = mlmsgcount;

		if(Alarm) {
			TTbeep();
			Alarm = 0;
		}
		
		upmode(NULL, WFMODE);
		update(0);
		msglFlush();

		ch = tgetc(MGC_MAP, 0);
		// printf(">%#x<", ch);

		// Some window manipulation (changing what is curwp) here doesn't always keep the current buffer pointer in track
		curbp = curwp->w_bufp;

		if(ch == 'q' || ch == abortc) {
			if(cmdMode == WIN_MANIP_MAIN) {
				mlwrite("");
				return(TRUE);
			}
			cmdMode = WIN_MANIP_MAIN;
			Arg = FALSE;
			continue;
		}

		if(ch == '\x1a') {	// Suspend: ^z
			bktoshell(FALSE, 0);
			cmdMode = WIN_MANIP_MAIN;
			Arg = FALSE;
			continue;
		}

		if(ch == ('u' ^ 0x60)) {
			Count = Arg ? Count * 4 : 4;
			Arg = TRUE;
			continue;
		}

		if(Isdigit(ch)) {
			Count = (Arg ? Count * 10 : 0) + (ch - '0');
			Arg = TRUE;
			continue;
		}

		if(ch == '-') {
			if( !Arg)
				Count = 1;
			Count *= -1;
			Arg = TRUE;
			continue;
		}

		/* Some generic processing for the modes.
		 * Process help (?) and parse/verify a direction that might be entered.
		 */
		moveDirection = DIRECTION_UP;	// Set to anything valid (not BAD)
		// Parse a direction ?
		if((flag = ModeInfo[cmdMode].wm_flag) & WM_DIRN)
			*((flag & WM_SET_DIR) ? &moveDirection : &EdgeToMove) = i = GetDirection(ch);

		// Did the direction have to be a good one ?
		// Ignore uninitialised i
		if((flag & WM_GOOD_DIRN) && i == DIRECTION_BAD) {
			Alarm = 1;
			ch = '?';
			// Drop down to give help
		}

		if(ch == '?') {
			Help = ModeInfo[cmdMode].wm_help;
			Arg = FALSE;
			continue;
		}

		map = BuildWindowMap();

		switch(cmdMode) {
		case WIN_MANIP_MAIN:
			switch(ch) {
			case 'b':	// Next buffer
				nextbuffer(FALSE, 1);
				break;
			case 'e':	// Move window edge
				cmdMode = WIN_MANIP_EDGE;
				break;
			case 'm':	// Move text within window
				cmdMode = WIN_MANIP_MOVE;
				break;
			case 's':	// Split windows
				cmdMode = WIN_MANIP_SPLIT;
				break;
			case 'j':	// Join windows
				cmdMode = WIN_MANIP_JOIN;
				break;
			case 'k':	// Kill buffer, then close window
				if(curwp->w_bufp->b_nwnd > 1)
					delwind(TRUE, TRUE);
				else
					zotbuf(curwp->w_bufp, 0);
				break;
			case 'c':	// Close window
					// If the last window, should do nothing
				delwind(0, 0);
				break;
			case 'o':	// Set/Unset window option
				cmdMode = WIN_MANIP_OPTION;
				break;
			case 'n':	// Next
				nextwind(FALSE, 0);
				break;
			case 'p':	// Previous
				prevwind(FALSE, 0);
				break;
			default:// Could be direction
				// Move to window in the parsed direction
				// Look at the map get fnd out what lies in that direction
				if(moveDirection == DIRECTION_BAD) {
					Alarm = 1;
					break;
				}
				if( !MoveToWindow(moveDirection))
					Alarm = 1;	// Nothing there

				break;
			}
			break;

		case WIN_MANIP_EDGE:
			// Key is the name of the edge that is wanted to be moved
			// Is it possible to move that edge ?
			// Choose the up/down or left/right move function
			// Call it with a zero move to see if it will work
			// Generate an appropriate message (if error)
			if(EnlargeWindMsg(
				((EdgeToMove == DIRECTION_UP || EdgeToMove == DIRECTION_DOWN) ? EnlargeWindUpDown : EnlargeWindLeftRight)
					((Arg ? Count : 1), map, 0, (EdgeToMove == DIRECTION_DOWN || EdgeToMove == DIRECTION_RIGHT), 1))) {
				Alarm = 1;
				cmdMode = WIN_MANIP_MAIN;
				break;
			}

			sprintf(msg, "Move %s window edge: ", EdgeNames[EdgeToMove]);
			cmdMode = WIN_MANIP_EDGE_MOVE;
			break;

		case WIN_MANIP_EDGE_MOVE:
			// Move an edge; key is the direction to move that edge
			i = Arg ? Count : 1;
			switch(EdgeToMove) {
			case DIRECTION_UP:
			case DIRECTION_DOWN:
				if(moveDirection != DIRECTION_UP && moveDirection != DIRECTION_DOWN) {
					mlwrite("%WUp or down only");
					Alarm = 1;
					break;
				}
				if(EnlargeWindUpDown(moveDirection == EdgeToMove ? i : -i, map, 0, EdgeToMove == DIRECTION_DOWN, 0))
					Alarm = 1;

				break;
			case DIRECTION_LEFT:
			case DIRECTION_RIGHT:
				if(moveDirection != DIRECTION_LEFT && moveDirection != DIRECTION_RIGHT) {
					mlwrite("%WLeft or right only");
					Alarm = 1;
					break;
				}
				if(EnlargeWindLeftRight(moveDirection == EdgeToMove ? i : -i, map, 0, EdgeToMove == DIRECTION_RIGHT, 0))
					Alarm = 1;

				break;
			}
			break;

		case WIN_MANIP_SPLIT:
			// Split a window horizontally or vertically
			switch(ch) {
			case 'v':
				// Wants a buffer name -- change to not need
				splitwind(Arg, Count);
				cmdMode = WIN_MANIP_MAIN;
				break;
			case 'h':
				if(splitwindhoriz(Arg, Count))
					countMl = mlmsgcount;
				cmdMode = WIN_MANIP_MAIN;
				break;
			default:
				mlwrite("%Wh or v only");
				Alarm = 1;
				break;
			}
			break;

		case WIN_MANIP_JOIN:
			// Join along an edge; key is the edge to join on
			// This works by expanding the window on the other side of the edge
			// to consume all of the space of curwp.
			// This might fail if that window does not have a clean edge with the current window.
			// Enter command mode immediately afterwards:
			cmdMode = WIN_MANIP_MAIN;

			if(EnlargeWindMsg(DelMergeWindow(map, EdgeToMove)))
				Alarm = 1;

			break;

		case WIN_MANIP_MOVE:
			// Move text within window:
			switch(EdgeToMove) {
			case DIRECTION_UP:
				mvdnwind(Arg, Count);
				break;
			case DIRECTION_DOWN:
				mvupwind(Arg, Count);
				break;
			case DIRECTION_LEFT:
				moveRightWind(Arg, Count);
				break;
			case DIRECTION_RIGHT:
				moveLeftWind(Arg, Count);
				break;
			}
			break;

		case WIN_MANIP_OPTION:
			// Set/Unset window option
			if(ch != 's' && ch != 'u') {
				mlwrite("%Ws or u only");
				Alarm = 1;
				break;
			}
			SetWindowOption(ch == 's');
			cmdMode = WIN_MANIP_MAIN;
			break;

		}

		// Actions above could mean that this is now wrong:
		free(map);

		Arg = FALSE;
	}
}

/* Resize the current window to the requested size.
 *
 * The default count is unset.
 * If the count is unset this returns FALSE
 *
 * Command:  resize-window
 * bound to: unbound
 *
 * Return - needs documenting - TODO
 *	FALSE count is unset
 */
int	resize(int f, Cnt n)
{
	int clines;	/* current # of lines in window */
	
	/* Must have a non-default argument, else ignore call */
	if(f == FALSE)
		return(FALSE);

	/* find out what to do */
	clines = curwp->w_ntrows;

	/* already the right size? */
	if(clines == n)
		return(TRUE);

	return(enlargewind(TRUE, n - clines));
}

/* Pick a window for a pop-up. Split the screen if there is only one window.
 *
 * Pick the uppermost window that isn't the current window. An LRU algorithm
 * might be better. Return a pointer, or NULL on error.
 *
 * This could now split left/right. If so return the rightmost one
 */
Window*	wpopup(void)
{
	register Window* wp = curwp;

	if(wheadp->w_wndp == NULL) {			// Only 1 window
		int	width = curwp->w_ncolumns;

		if(splitwind(FALSE, 0) == FALSE)	// and it won't split
			return(NULL);

		/* Different width means split left/right
		 * splitwindhoriz() will make curwp the rightmost one
		 */
		if(width != curwp->w_ncolumns)
			return(curwp);
	}

	/* Find window to use
	 * First one that is not the current one.
	 */
	wp = wheadp;
	while(wp->w_wndp != NULL && wp == curwp)
		wp = wp->w_wndp;

	return(wp);
}

/* scroll the next window up (back) a page.
 *
 * Fails backpage() if no movement happens.
 *
 * Default count 1. Must be >= 0.
 *
 * Command:  scroll-next-up
 * Bound to: M-C-z
 *
 * Return TRUE if moved, even if not n times (where n > 1)
 *	  FALSE when did not - at start of buffer
 */
int	scrnxtup(int f, Cnt n)
{
	int	ret;

	nextwind(FALSE, 1);
	ret = backpage(f, n);
	prevwind(FALSE, 1);

	return(ret);
}

/* scroll the next window down (forward) a page.
 *
 * Fails forwpage() if no movement happens.
 *
 * Default count 1. Must be >= 0.
 *
 * Command:  scroll-next-down
 * Bound to: M-C-v
 *
 * Return TRUE if moved, even if not n times (where n > 1)
 *	  FALSE when did not - at end of buffer
 */
int	scrnxtdw(int f, Cnt n)
{
	int	ret;

	nextwind(FALSE, 1);
	ret = forwpage(f, n);
	prevwind(FALSE, 1);

	return(ret);
}

/* Save ptr to current window.
 * It does not matter if swindow end up pointing to a window that becomes
 * deallocated. It will only ever be used by comparing it to a list of valid
 * windows - so if it contains garbage it will just not find anything.
 *
 * The count is ignored.
 *
 * Command:  save-window
 * Bound to: unbound
 *
 * Returns: TRUE
 */
int	savewnd(int f, Cnt n)
{
	swindow = curwp;
	return(TRUE);
}

/* Restore the saved screen
 *
 * The count is ignored.
 * 
 * Command:  restore-window
 * Bound to: unbound
 *
 * Return TRUE  - on success
 *	  FALSE - the window does not exist
 */
int	restwnd(int f, Cnt n)
{
	register Window* wp;

	/* find the window */
	wp = wheadp;
	while(wp != NULL) {
		if(wp == swindow) {
			curwp = wp;
			curbp = wp->w_bufp;
			upmode(NULL, WFMODE);
			return(TRUE);
		}
		wp = wp->w_wndp;
	}

	mlwrite("%WNo such window exists");
	return(FALSE);
}

/* Resize the screen - ie set the number of lines to the count, re-writing the screen.
 *
 * If no count, assume largest possible.
 *
 * If window size can be determined via WINCH signal then we do not want this.
 *
 * The default count is not set.
 *
 * Command:  change-screen-size
 * Bound to: unbound
 *
 * Return TRUE  - on success
 *	  FALSE - failure bad size or out of memory.
 */
int	newsize(int f, Cnt n)
{
#if	WINSIZE
	mlwrite("%ESetting window lines/height not implemented");
	return(FALSE);
#else
	Window* wp;	/* current window being examined */
	Window* nextwp; /* next window to scan */
	Window* lastwp; /* last window scanned */
	int lastline;	/* screen line of last line of current window */

	/* if the command defaults, assume the largest */
	if(f == FALSE)
		n = term.t_mrow;	// Default is max PHYSICALLY possible

        /* make sure it's in range */
	if(n < MIN_SCREEN_LINES || n > term.t_mrow) {
		mlwrite("%WScreen size out of range");
		return(FALSE);
	}

	if(AllocLines == n)		// No change
		return(TRUE);

	AllocLines = n;

	/* screen is garbage */
	sgarbf = TRUE;

	return(FixWindowSizes(n, AllocColumns));
#endif
}

/* Resize the screen, re-writing the screen.
 *
 * If window size can be determined via WINCH signal then we do not want this.
 *
 * The default count is not set.
 *
 * Command:  change-screen-width
 * Bound to: unbound
 *
 * Return TRUE on success
 *	  FALSE - failure will be bad size or out of memory.
 */
int	newwidth(int f, Cnt n)
{
#if	WINSIZE
	mlwrite("%ESetting window width not implemented");
	return(FALSE);
#else
	register Window* wp;

	/* if the command defaults, assume the largest */
	if(f == FALSE)
		n = term.t_mcol;	// Default is max PHYSICALLY possible

	/* make sure it's in range */
	if(n < MIN_SCREEN_COLS || n > term.t_mcol) {
		mlwrite("%WScreen width out of range");
		return(FALSE);
	}

	if(AllocColumns == n)		// No change
		return(TRUE);

	/* otherwise, just re-width it (no big deal) */
	AllocColumns = n;	// Reset screen size to *use*
	term.t_margin = n / 10;	// These 2 no longer used
	term.t_scrsiz = n - (term.t_margin * 2);

	/* screen is garbage */
	sgarbf = TRUE;

	return(FixWindowSizes(AllocLines, n));
#endif
}

/* Get screen offset of current line in current window.
 */
int	getwpos(void)
{
	register int   sline;	/* screen line from top of window */
	register LINE* lp;	/* scannile line pointer */

	/* search down the line we want */
	lp = curwp->w_topLinep;
	sline = 1;
	while(lp != curwp->w_dotp) {
		++sline;
		lp = lforw(lp);
	}

	/* and return the value */
	return(sline);
}

/* Check window sizes.
 *
 * Check that they all still fit on the screen, this really of use if the
 * screen size has changed, but is called more often.
 */
void	CheckSizes(void)
{
	FixWindowSizes(AllocLines, AllocColumns);
}

/* Fix the sizes of windows to fit in the physical screen.
 *
 * lines is the new # lines for the screen.
 * columns is the new # columns.
 * This was stolen from newsize().
 *
 * For each window fix up w_ncolumns & w_ntrows.
 *
 * It is quite possible that no change is actually done.
 * We don't know what the old size might have been - if it is different.
 * However: there is an existing arrangement of windows that does (or did) fit within
 * a window and they are organised in the way that the user wants. They also
 * all fit together, ie a map can be made.
 *
 * This goes round and kills windows that will not fit in the new size
 * even after resizing to minimum window sizes.
 * The windows that are killed are those on the right and bottom, the ones killed
 * should always be at screen edge and so will always be killable.
 *
 * Then it find unused space and extends windows to fill the space.
 * See comments at delwind().
 */
void	FixWindowSizes(int lines, int columns)
{
	Window*	wp;
	MapCell*map;
	Window*	saveCurwp;

	// Ensure that we don't fall over, make map big enough
	setScreenSizeForMap(lines, columns);

	/* Traverse all windows and mark for killing those that, even after resizing
	 * will not fit in the screen.
	 * Mark them by making the width (w_ncolumns) zero.
	 */
	for(wp = wheadp; wp != NULL; wp = wp->w_wndp) {
		if(wp->w_leftcol + MIN_WINDOW_COLS > AllocColumns ||
		   wp->w_toprow + MIN_WINDOW_LINES >= AllocLines) {
			wp->w_ncolumns = 0;

			// If we will kill the current window - make the first window current
			if(wp == curwp)
				curwp = wheadp;
		}
	}

	/* Deal with the message line, it is always on the bottom line and may
	 * be width one less than the physical width -- writing to the last
	 * position on the screen makes some terminals scroll.
	 * 
	 * Reset position on the line to 0 - keep things simple.
	 */
	wmessp->w_ncolumns = AllocColumns - 1;

	wmessp->w_disrow = AllocLines - 1;
	wmessp->w_toprow = AllocLines - 1;
	wmessp->w_lastcurcol = wmessp->w_discol = 0;

	// Get rid of what we just marked for killing:
	ZapZeroWidthWindows();

	// The above might have killed windows, in which case there is space to be used.
	// The screen might have been made bigger, in which case there is space to be used.
	// Produce a map, extend every window that has NULL to the right or underneath it.
	// Beware: every window with a window to the right of it has NULL in the first
	// column to it's right.
	// We might also have windows that are too wide/deep that need shrinking.
	SortWindows();

	map = BuildWindowMap();

	// That will have right hand & bottom borders of the largest recent screen.
	// Within that will be the windows at their sizes that were good for the last
	// screen size - which might not be fight for the current screen size.
	// Write borders for the current screen so that we can do edge detection, etc.
	// These might be the borders that it already has - don't try to be clever.
	FillMapRectangle(map + (AllocLines-1) * (window_map_columns) + 1, 1, window_map_columns-2, window_map_columns, WINDOW_MAP_MSG_LINE);
	FillMapRectangle(map + (AllocLines  ) * (window_map_columns) + 1, 1, window_map_columns-2, window_map_columns, WINDOW_MAP_SC_EDGE);

	// RH border
	FillMapRectangle(map + AllocColumns + 1, window_map_lines, 1, window_map_columns, WINDOW_MAP_SC_EDGE);

	// Some functions operate on curwp, so save current value so that we can restore
	saveCurwp = curwp;

	for(curwp = wheadp; curwp != NULL; curwp = curwp->w_wndp) {
		int	rightPos = curwp->w_leftcol + WINDOW_WIDTH(curwp);

		// Adjust: too wide/narrow. Might adjust by zero
		// When too narrow extend if there is nothing to the right
		if(rightPos >= AllocColumns ||
		   !FindWinByMapno(WindowNumberInDirection(DIRECTION_RIGHT, map)))
			curwp->w_ncolumns -= rightPos - AllocColumns;

		// Now the depth (# lines)
		// +1 for mode line
		int	bottomPos = curwp->w_toprow + curwp->w_ntrows + 1;
		if(bottomPos >= AllocLines ||
		   !FindWinByMapno(WindowNumberInDirection(DIRECTION_DOWN, map))) {
			changeWindowRows(curwp, curwp->w_ntrows - (bottomPos - AllocLines + 1));
		}
	}

	curwp = saveCurwp;

	// Just in case the head got changed. This should be redundant
	wmessp->w_wndp = wheadp;

	free(map);
}

/* Are we to calculate wordwrap position from how big the screen is ?
 * Leave a margin, but take care not to go -ve.
 */
void	CheckAutoWidth(void)
{
	if(auto_width) {
		fillcol = AllocColumns - 5;
		if(fillcol < 10)
			fillcol = 10;
	}
}

/* Find the bottom line in a window.
 *
 * Return a pointer to the bottom line displayed.
 * If there is not a displayed line return NULL - because the
 * end of buffer was reached.
 *
 * The buffer header line will not match the last line.
 */
LINE*	WinBottomLine(Window* wp)
{
	int	nlines = wp->w_ntrows;
	LINE*	linep = wp->w_topLinep;

	while(--nlines > 0) {
		if(EofLineB(wp->w_bufp, linep)) {
			linep = (LINE*)NULL;
			break;
		}

		linep = lforw(linep);
	}

	return(linep);
}

/* The Information Window. A popup for extra information.
 *
 * This is a window that is not, usually, displayed. When it is to be
 * displayed it is placed such that it overlaps other windows. It is
 * rendered last by update(), this gives it the appearance
 * of being on top of other windows. The idea is that shortly after (timeout
 * or keystroke) it will no longer be rendered and so disappears.
 *
 * Notes on how this special window works:
 *
 * It is only ever associated with buffer binfop ([Info Window]).
 * "." is always the first character on the first line.
 * w_toplinenum is 1 if the window is to be displayed, otherwise 0.
 * The window will be displayed with a border on all 4 edges
 * The window does not count towards NumberOfWindows
 */

/* Create the info window.
 * This will be positioned when it is needed.
 */
void	iwCreate(void)
{
	if((winfop = allocWindow(1, 0)) == NULL)
		Die(MsgOutOfMemoryAlloc, "info window");	/* Should never happen */

// put binfop into UTF8 mode
	winfop->w_bufp = binfop;
	
	binfop->b_nwnd = 1;
	winfop->w_doto = 0;
	winfop->w_toplinenum = 0;
	winfop->w_attrib = WATT_NOSPLIT | WATT_BORDER | WATT_MODESTATIC;

}

/* Cause the info window to not be displayed
 */
void	iwHide(void)
{
	winfop->w_toplinenum = 0;

	/* Need to mark as needing display those lines that are uncovered
	 * by this window no longer being displayed.
	 */
}

/* Empty out the info window.
 * First hides it.
 */
void	iwClear(void)
{
	iwHide();

	bclear(binfop);
}

/* Cause the info window to be displayed.
 * n and return value as for iw_display()
 * except that n must not be negative.
 */
int	iwShow(int n)
{
	// Work out best position for the window ?
// 	if(n == 0) {
// 		;
// 	}

	winfop->w_toplinenum = 1;

	return(n);
}

/* User accessable control of the information window
 *
 * If f is false, display the window as low as possible, ie so that
 * the last line is just above the message line at the bottom of the screen.
 * If the window is too long to be entirely displayed, the first line
 * will be displayed at the top of the screen.
 * At some stage (maybe) implement a mechanism for scroll up/down
 *
 * If f is set:
 * n < 0 Hide the window
 * n = 0 As for f being false (above)
 * n > 0 Place the first line on line n of the screen
 * The screen line number is the line of the top border of the window,
 * the top line is numbered 1.
 *
 * Return FALSE or the line number of the top line.
 *
 * Command: info-window-display
 * It probably does not make much sense to bind this to anything, but could be
 * used in macros.
 */
int	iwDisplay(int f, Cnt n)
{
	if( !f)
		n = 0;

	// No longer show the window ?
	if(n < 0) {
		iwHide();
		return(TRUE);
	}

	return(iwShow((int)n));
}

/* The info window will be shown (n=0) (if not already)
 * Wait for a keystroke to dismiss the window.
 * Interpret: arrow/page up/down; ^z, ^l.
 * Dismiss on: space, return/enter, ^G, Escape
 *
 * timeout	seconds after which the window will be hidden
 *		<= 0 -- no timeout
 *
 * TODO: timeout
 */
void	iwPrompt(long timeout)
{
	UniChar		ch;		// Read from keyboard

	if( !winfop->w_toplinenum)
		iwShow(0);

	while(1) {
		upmode(NULL, WFMODE);

		ch = mlGetKey(MGC_NONE, NULL, NULL);
		// printf(">%#x<", ch);

		if(ch == 'q' || ch == abortc || ch == ' ' /* || ch == ESCAPE ???? */

)
			break;

		if(ch == '\x0c') {	// Refresh screen: \f or ^l
			ReDraw();
			continue;
		}

		if(ch == '\x1a') {	// Suspend: ^z
			bktoshell(FALSE, 0);
			continue;
		}

		// arrow & page keys
	}

	iwHide();
}


/* end */
