/*
 * ADDW SCCS: @(#)word.c	1.50 07/31/20 09:44:22
 */

/*
 * The routines in this file implement commands that work word or a
 * paragraph at a time.  There are all sorts of word mode commands.  If I
 * do any sentence mode commands, they are likely to be put in this file. 
 */

#include	"estruct.h"
#include	"utf8.h"
#include	"edef.h"

#if	SHOWMODREL
char	wordvers[] = "@(#)word.c	1.50 07/31/20 09:44:22";
#endif

// TODO utf8


/* Check if what follows "." is not unjoined punctuation.
 * (Because if so we don't want to wrap before it).
 * If it is not return TRUE. "." will not be moved.
 *
 * &&& xxx	the &&& is unjoined punctuation
 * &&&xxx	the &&& is not unjoined punctuation
 *
 * If it is unjoined punctuation then move "."
 * back (span over SPC & TAB) & to the start of the previous word & return FALSE.
 * movedBy will be set to the number of characters that have been
 * moved over to get to the start of that word.
 *
 * If there is no previous word (before line start) return TRUE
 * (which is a lie, but still ...) and "." will not have been moved.
 *
 * Do not move off the current line.
 */
int	NotUnJoinedPunctuation(LineLn wp, LineLn* movedBy)
{
	LineLn	dotoStart = curwp->w_doto;	// So that we can get back there
	LineLn	doto = curwp->w_doto;
	BYTE	c;

	*movedBy = 0;

	// fprintf(stderr, "NotUnJoinedPunctuation() doto=%ld ('%c') .used=%ld wp=%ld\n",
		// (long)curwp->w_doto, lgetc(curwp->w_dotp, curwp->w_doto), (long)lused(curwp->w_dotp), wp);
	// "." is on SPC or TAB, move on
	while(doto < wp && (
	      (c = lgetc(curwp->w_dotp, doto)) == ' ' || c == '\t'))
		doto++;

	// fprintf(stderr, "init to %ld '%c'\n", (long)doto, c);
	// Run out of line ?
	if(doto >= lused(curwp->w_dotp)) {
		return(TRUE);
	}

	int	seenPunct = 0;
	// Look at the next word, determine if there is an alpha before space
	// fprintf(stderr, "look at next word starting @ %ld\n", (long)doto);
	while(1) {
		if(doto > wp) {
			// Must not look further
			if(!seenPunct)
				return(TRUE);
			break;
		}
		c = lgetc(curwp->w_dotp, doto);
		// fprintf(stderr, "'%c' @ %ld\n", c, (long)doto);
		if(c > UTF8_MAX_CP1 || Isalnum(c)) {
			// Yes, it is not unjoined puctuation
			// fprintf(stderr, "Seen alnum '%c' at %ld\n", c, (long)doto);
			return(TRUE);
		}
		if(c == ' ' || c == '\t') {
			// end of word
			if(!seenPunct)
				return(TRUE);
			break;
		}

		// If here then we have seen some punctuation
		seenPunct = 1;

		doto++;
	}

	// Here as we must move back to word before dotoStart
	// fprintf(stderr, "end punct look @ %ld, back to dotoStart %ld seenPunct=%d\n", (long)doto, (long)dotoStart, seenPunct);
	doto = dotoStart;

	int	seenAlpha = 0;	// End up moved to start of word ?

	// Look for the start of the previous word
	// First span spaces
	while(doto > 0 && (
	      (c = lgetc(curwp->w_dotp, doto)) == ' ' || c == '\t'))
		doto--;

	// fprintf(stderr, "back over white to %ld\n", (long)doto);
	// Scan back over next 'word'
	while(doto > 0 && (
	      (c = lgetc(curwp->w_dotp, doto)) != ' ' && c != '\t')) {
		if(c > UTF8_MAX_CP1 || Isalnum(c))
			seenAlpha = 1;
		doto--;
	}

	/* If the previous word had an alpha in it them move "." back.
	 * If it did not then do nothing - which might result in non-alpha
	 * at the start of line ... but we try to cope with common cases!
	 */
	if(seenAlpha) {
		curwp->w_doto = doto;

		// How far did we move back ?
		// Note that this is needed in characters - not bytes
		*movedBy = (curbp->b_mode & MDUTF8) ?
			(LineLn)utf8_chrnlen(lgetcaddr(curwp->w_dotp, doto), (size_t)(dotoStart - doto)) :
			(dotoStart - doto);
	}
	// fprintf(stderr, "Ret doto=%ld seenA=%d\n", (long)curwp->w_doto, seenAlpha);

	return(TRUE);

}

/* Break the line before the current word.
 *
 * Find the start of the word containing "." and start a new line there. "." remains
 * at the same place in the word, the word will be at start of line.
 *
 * If "." is at start of line, a new line is inserted before the last word of the
 * previous line.
 *
 * If the word is already at the start of line and "." is not at start of line:
 * the newline is inserted after the word. This differs from other microemacs that
 * insert the newline after the end of line.
 *
 * In INDENT mode, the indent of the current line is preserved and white space
 * at line start is spanned.
 *
 * The count is ignored.
 *
 * Command:  wrap-word
 * Bound to: unbound
 *
 * Returns
 *	TRUE on success
 *	FALSE on errors
 *	    hit end of buffer/narrow
 *	    newlineAndIndent() ldelete() linsert()
 */
int	wrapword(int f, Cnt n)
{
	int	cnt;		/* size of word wrapped to next line */
	BYTE	c;		/* character temporary */
	LineLn	LnStart = 0;	/* Where line start is deemed to start - bytes */
	LineLn	movedBy = 0;	/* How much */

	if(EonLine(curwp->w_dotp))	// End of narrow ?
		return(FALSE);

	// lsetNUL(curwp->w_dotp);
	// fprintf(stderr, "wrapword() doto=%ld '%s'\n", (long)curwp->w_doto, lgetcaddr(curwp->w_dotp, curwp->w_doto));
	/* Indent complication:
	 * White space at line start is considered to be line start - ie span it.
	 */
	if(curbp->b_mode & MDINDEN) {
		LnStart = countWhiteBytes(&ltxt(curwp->w_dotp));
		if(curwp->w_doto <= LnStart)
			curwp->w_doto = 0;
	}

	/* backup from the <NL> 1 char */
	if(!backchar(1, 1))
		return(FALSE);

	/* Back up until we aren't in a word,
	 * make sure there is a break in the line.
	 */
	cnt = 0;

	LineLn wp = curwp->w_doto;		// Word position
	while(1) {
		// fprintf(stderr, "LOOP: '%c' @ %ld\n", lgetc(curwp->w_dotp, curwp->w_doto), (long)curwp->w_doto);
		if(((c = lgetc(curwp->w_dotp, curwp->w_doto)) == ' ') ||
						(c == '\t')) {
			// Somewhere where we could potentially break the line
			// In indent mode ensure that we don't break before non joined punctuation
			if((curbp->b_mode & MDINDEN) == 0)
				break;

			/* If what immediately follows is just punctuation, search back
			 * to the start of the word before.
			 */
			// int n = NotUnJoinedPunctuation(wp, &movedBy);
			// fprintf(stderr, "NotUnJoinedPunctuation() -> %d\n", n);
			// if(n)
			if(NotUnJoinedPunctuation(wp, &movedBy))
				break;
		}

		cnt++;

		if(!backchar(1, 1))
			return(FALSE);

		// Go back over line end, reset notion of line start:
		if(curwp->w_doto >= lused(curwp->w_dotp))
			LnStart = 0;

		/* If we make it to the beginning of line, find the end
		 * of the current word so that the line is broken there.
		 * This behaviour differs from other microemacs.
		 * This leaves cnt negative.
		 */
		if(curwp->w_doto <= LnStart) {
			// Indent mode, keep going back, but not beyond line start:
			if((curbp->b_mode & MDINDEN) && curwp->w_doto > 0)
				continue;

			if(forwchar(TRUE, cnt) == FALSE)
				return(FALSE);

			for(cnt = 0; (c = lgetc(curwp->w_dotp, curwp->w_doto)) != ' ' && c != '\t'; cnt--)
				if(forwchar(TRUE, 1) == FALSE)
					return(FALSE);

			break;
		}
	}

	/* If we are in INDENT mode, insert the indent of the previous line
	 * (ie the one that we just left) before the word. Only if this is
	 * less than the wrap position.
	 * Not INDENT mode: just insert the newline.
	 */
	if(curbp->b_mode & MDINDEN) {
		if( ! newlineAndIndent(TRUE, 1))
			return(FALSE);
	} else {
		LINE*	startLine = curwp->w_dotp;

		if( !ldelete(1, KL_DISCARD))
			return(FALSE);

		if(!lnewline(0))
			return(FALSE);

		ZapTrailingSpaces(startLine);
	}

	cnt += (int)movedBy;

	/* And back to where we were in the word */
	if(cnt < 0) {
		if(backchar(TRUE, -cnt) == FALSE)
			return(FALSE);
	} else {
		if(forwchar(TRUE, cnt) == FALSE)
			return(FALSE);
	}

	return(TRUE);
}

/* Move the cursor backward by "n" words.
 *
 * All of the details of motion are performed by the "backchar" and "forwchar"
 * routines. Error if you try to move beyond the buffers.
 *
 * Default count 1. Will be >= 0.
 *
 * Command:  previous-word
 * Bound to: M-b
 *
 * Return:
 *	TRUE  moved back a word
 *	FALSE hit end of buffer
 */
int	backword(int f, Cnt n)
{
	if(backchar(TRUE, 1) == FALSE)
		return(FALSE);

	while(--n >= 0) {
		while(inword() == FALSE) {
			if(backchar(TRUE, 1) == FALSE)
				return(FALSE);
		}

		while(inword() != FALSE) {
			if(backchar(TRUE, 1) == FALSE)
				return(FALSE);
		}
	}

	return(forwchar(TRUE, 1));
}

/* Move the cursor forward by the specified number of words.
 *
 * All of the motion is done by "forwchar". Error if you try and move beyond
 * the buffer's end.
 *
 * Default count 1. Will be >= 0
 *
 * Command:  next-word
 * Bound to: M-f
 *
 * Return:
 *	TRUE  moved forward a word
 *	FALSE hit end of buffer
 */
int	forwword(int f, Cnt n)
{
	while(--n >= 0) {
		while(inword() == TRUE) {
			if(forwchar(TRUE, 1) == FALSE)
				return(FALSE);
		}

		while(inword() == FALSE) {
			if(forwchar(TRUE, 1) == FALSE)
				return(FALSE);
		}
	}

	return(TRUE);
}

/* Move the cursor forward by the specified number of words.
 *
 * As you move, convert any characters to upper case. Error if you try and move
 * beyond the end of the buffer.
 *
 * Handle multibyte cased characters. TODO
 *
 * Default count 1. Will be >= 0.
 *
 * Command:  case-word-upper
 * Bound to: M-u
 *
 * Return:
 *	TRUE  success
 *	FALSE hit end of buffer
 */
int	upperword(int f, Cnt n)
{
	BYTE	c;

	while(--n >= 0) {
		while(inword() == FALSE) {
			if(forwchar(TRUE, 1) == FALSE)
				return(FALSE);
		}

		while(inword() != FALSE) {
			c = lgetc(curwp->w_dotp, curwp->w_doto);

			if(Islower(c)) {
				c = (BYTE)toupper(c);
				lputc(curwp->w_dotp, curwp->w_doto, c);
				lchange(WFHARD);
			}

			if(forwchar(TRUE, 1) == FALSE)
				return(FALSE);
		}
	}

	return(TRUE);
}

/* Move the cursor forward by the specified number of words.
 *
 * As you move convert characters to lower case. Error if you try and move over
 * the end of the buffer.
 *
 * Handle multibyte cased characters. TODO
 *
 * Default count 1. Will be >= 0.
 *
 * Command:  case-word-lower
 * Bound to: M-l
 *
 * Return:
 *	TRUE  success
 *	FALSE hit end of buffer
 */
int	lowerword(int f, Cnt n)
{
	BYTE	c;

	while(--n >= 0) {
		while(inword() == FALSE) {
			if(forwchar(TRUE, 1) == FALSE)
				return(FALSE);
		}

		while(inword() != FALSE) {
			c = lgetc(curwp->w_dotp, curwp->w_doto);
			if(Isupper(c)) {
				c = (BYTE)tolower(c);
				lputc(curwp->w_dotp, curwp->w_doto, c);
				lchange(WFHARD);
			}

			if(forwchar(TRUE, 1) == FALSE)
				return(FALSE);
		}
	}

	return(TRUE);
}

/* Invert the case of the character under the cursor and move on.
 * 
 * This is only valid for a -> z and A -> Z. If we're on something else
 * then dont do anything.
 *
 * Handle multibyte cased characters. TODO
 *
 * Default count 1. Will be >= 0.
 *
 * Command:  change-char-case
 * Bound to: C-]
 *
 * Return:
 *	TRUE  success
 *	FALSE hit end of buffer
 */
int	invert(int f, Cnt n)
{
	BYTE	c;

	while(--n >= 0) {
		c = lgetc(curwp->w_dotp, curwp->w_doto);

		/* Only worry about upper and lower case characters.
		 */
		if(Isletter(c)) {

			if(Islower(c))
				c = (BYTE)toupper(c);
			else if(Isupper(c))
				c = (BYTE)tolower(c);

			lputc(curwp->w_dotp, curwp->w_doto, c);
			lchange(WFHARD);
		}

		if(forwchar(TRUE, 1) == FALSE)
			return(FALSE);
	}
	return(TRUE);
}

/* Move the cursor forward by the specified number of words.
 *
 * As you move convert the first character of the word to upper case, and
 * subsequent characters to lower case. Error if you try and move past the end
 * of the buffer.
 *
 * Handle multibyte cased characters. TODO
 *
 * Default count 1. Will be >= 0.
 *
 * Command:  case-word-capitalize
 * Bound to: M-c
 *
 * Return:
 *	TRUE  success
 *	FALSE hit end of buffer
 */
int	capword(int f, Cnt n)
{
	BYTE	c;

	while(--n >= 0) {
		while(inword() == FALSE) {
			if(forwchar(TRUE, 1) == FALSE)
				return(FALSE);
		}

		if(inword() != FALSE) {
			c = lgetc(curwp->w_dotp, curwp->w_doto);
			if(Islower(c)) {
				c = (BYTE)toupper(c);
				lputc(curwp->w_dotp, curwp->w_doto, c);
				lchange(WFHARD);
			}

			if(forwchar(TRUE, 1) == FALSE)
				return(FALSE);

			while(inword() != FALSE) {
				c = lgetc(curwp->w_dotp, curwp->w_doto);
				if(Isupper(c)) {
					c = (BYTE)tolower(c);
					lputc(curwp->w_dotp, curwp->w_doto, c);
					lchange(WFHARD);
				}

				if(forwchar(TRUE, 1) == FALSE)
					return(FALSE);
			}
		}
	}

	return(TRUE);
}

/* Kill forward by "n" words.
 *
 * Remember the location of dot. Move forward by the right number of words. Put
 * dot back where it was and issue the kill command for the right number of
 * characters.
 *
 * Default count 1. Will be >= 0.
 *
 * Command:  delete-next-word
 * Bound to: M-d
 *
 * Return:
 *	TRUE  success
 *	FALSE hit end of buffer
 *	ldelete()
 */
int	delfword(int f, Cnt n)
{
	LINE*	dotp;
	LineLn	doto;
	int	c;
	LineLn	size;

	/* save the current cursor position */
	dotp = curwp->w_dotp;
	doto = curwp->w_doto;

	/* figure out how many characters to give the axe */
	size = 0;

	/* get us into a word.... */
	while(inword() == FALSE) {
		if(forwchar(TRUE, 1) == FALSE)
			return(FALSE);
		++size;
	}

	if(n == 0) {
		/* skip one word, no whitespace! */
		while(inword() == TRUE) {
			if(forwchar(TRUE, 1) == FALSE)
				return(FALSE);
			++size;
		}
	} else {
		/* skip n words.... */
		while(--n >= 0) {

			/* if we are at EOL; skip to the beginning of the next */
			while(curwp->w_doto == lused(curwp->w_dotp)) {
				if(forwchar(TRUE, 1) == FALSE)
					return(FALSE);
				++size;
			}

			/* move forward till we are at the end of the word */
			while(inword() == TRUE) {
				if(forwchar(TRUE, 1) == FALSE)
					return(FALSE);
				++size;
			}

			/* if there are more words, skip the interword stuff */
			if(n != 0)
				while(inword() == FALSE) {
					if(forwchar(TRUE, 1) == FALSE)
						return(FALSE);
					++size;
				}
		}

		/* skip whitespace and newlines */
		while((curwp->w_doto == lused(curwp->w_dotp)) ||
			((c = lgetc(curwp->w_dotp, curwp->w_doto)) == ' ') ||
			(c == '\t')) {
				if(forwchar(TRUE, 1) == FALSE)
					break;
				++size;
		}
	}

	/* restore the original position and delete the words */
	curwp->w_dotp = dotp;
	curwp->w_doto = doto;

	return(ldelete(size, KL_KILL_APPEND));
}

/* Kill backwards by "n" words.
 *
 * Move backwards by the desired number of words, counting the characters. When
 * dot is finally moved to its resting place, fire off the kill command.
 *
 * Default count 1. Will be >= 0.
 *
 * Command:  delete-previous-word
 * Bound to: M-C-h   M-DEL
 *
 * Return:
 *	TRUE  success
 *	FALSE hit end of buffer
 *	ldelete()
 */
int	delbword(int f, Cnt n)
{
	long	size;
	LINE*	lp = curwp->w_dotp;
	LineLn	doto = curwp->w_doto;

	if(backchar(TRUE, 1) == FALSE)
		return(FALSE);

	size = 0;

	while(--n >= 0) {
		/* Skip back, until we find a word to swat.
		 */
		while(inword() != TRUE) {
			if(backchar(TRUE, 1) == FALSE)
				return(FALSE);
			++size;
		}

		/*
		 * Now we've found the word, swat it.
		 */
		while(inword() == TRUE) {
			if(backchar(TRUE, 1) == FALSE)
				goto kludge;
			++size;
		}
	}

kludge:
	curwp->w_dotp = lp;
	curwp->w_doto = doto;

	/* Use backdel() rather than ldeleted() so that deleted characters are put
	 * into the correct place, and correct order, in the yank buffer.
	 * BUG: fails, integer overflow. Not a real concern on modern machines (int >= 32 bit)
	 * The cast is left in to emphasise this.
	 */
	return(backdel(TRUE, (int)size));
}

/* Return TRUE if the character at dot is a character that is considered to be
 * part of a word.
 *
 * The word character list is hard coded. Should be setable.
 */
int	inword(void)
{
	register UniChar c;

	if(curwp->w_doto == lused(curwp->w_dotp))
		return(FALSE);

	c = lgetc(curwp->w_dotp, curwp->w_doto);

	/* If we're in C Mode, we really want most things to count as a
	 * word, so its easier to say what we dont want.
	 */
	if(curbp->b_mode & (MDCMOD | MDPHP))
		switch(c) {
		  case ' ':
		  case '\t':
		  case '\'':
		  case '"':
		  case ']':
		  case '[':
		  case '|':
		  case '(':
		  case ')':
		  case ',':
		  case '&':
			return(FALSE);
		  default:
			return(TRUE);
		}

	if(Isalnum(c))
		return(TRUE);

	return(FALSE);
}

#if	WORDPRO
/* Insert indent at the start of the current line, indent is in playText.
 * We will have just moved to the current line, ie "." is at the start of line.
 * Leave "." after the indent.
 *
 * Either the line is already indented and we just span to Indent.
 *
 * If the line is indented but does not start with the same indent bytes, replace
 * those bytes by what there is in playText.
 *
 * Indent	Column position that we should end up at. Beware: this could
 *		involve tab calculations.
 * From2	Indent came from the 2nd line, ie not a case where the first line
 *		is indented and subsequent lines are not.
 *		Thus we need to span the Indent.
 *
 * Return:
 *	TRUE  on all OK
 *	FALSE insStrLine() ldelete() TODO
 */
int	indentFromPlay(int From2, LineLn Indent)
{
	// Either move over the indent that is there, or insert it
	if(From2) {
		// How many characters make up the start of line indent ?
		LineLn	iChars = getdoto(curwp, curwp->w_dotp, Indent, DisCol_MetaChar);

		// Indent sequence the same ?
		if(lused(curwp->w_dotp) >= tused(playText) &&
		   memcmp(ltext(curwp->w_dotp), ttext(playText), (size_t)tused(playText)) == 0) {

			curwp->w_doto = iChars;
			return(TRUE);
		}

		// Delete the differing indent sequence so that it can be replaced:
		if(ldelete(iChars, KL_DISCARD) == FALSE)
			return(FALSE);
	}

	// Insert the contents of playText before ".":
	if(insStrLine(curwp->w_dotp, 0, ttext(playText), tused(playText), 0) < 0)
		return(FALSE);

	FixOffsets(curwp->w_dotp, 0, tused(playText), curbp, 1);

	lchange(WFEDIT);

	curwp->w_doto = tused(playText);

	return(TRUE);
}

/* Fill a paragraph - make lines about the same length.
 *
 * Examine the LINEs but only make a change if needed. This makes it more
 * complicated than it would otherwise be but means that the buffer is not
 * marked modified if the paragraph does not need formatting.
 *
 * Look at this as checking that a paragraph is correcty formatted, only make
 * a change if needs be. The assumption is that it was once correct and that
 * a few words might have been added/removed.
 * The exception is in INDENT mode, an indented first line will have its indent
 * given to the remaining lines in the paragraph -- EXCEPT if there is a
 * second line in which case use the indent from the second line.
 *
 * Note that some UTF8 characters have width 2.
 * Zero/partial width characters: given width of 1 - as usual
 * Interaction with the INDENT or C mode:
 * * If unset
 *   * subsequent lines will not be indented
 *   * the paragraph is deemed ended by the first line that starts with a space
 * * If set
 *   * Subsequent lines will be indented
 *     * should this be to the indent of the first (or only) line or the indent of the second line
 *       Indenting to the second line allows first lines of paragraphs to be of different 'hang' to later ones.
 *   * Leading space on subsequent lines will be spanned (ignored)
 *     This allows the command to be rerun with no change happening
 *
 * If the first line does not start with alpanumeric the (space trimmed) sequence is remembered
 * If subsequent lines start with this string then wrapping
 * The 2nd line may have a different sequence, in which case this will be used as
 * the test for that (and subsequent) lines.
 * If, after this sequence, there is EOL - this is taken as the end of paragraph.
 * Initial indent is only replaced if words have to be moved to/from the
 * previous line.
 *
 * If the last line is pure non alphanum then it is left as it is.
 *
 * Any character >= U+80 is currently considered alphanumeric.
 * Words are deemed separated by white space (space or tab).
 * Any tab seen as a word separator is replaced by a space.
 * Multiple spaces are replaced by a single space.
 * After a sentence end ('.', '!', '?') two spaces are placed,
 * unless $french-spacing (default TRUE) where one is placed.
 * https://en.wikipedia.org/wiki/Sentence_spacing
 *
 * Eg:
 *	// A line of comment
 *	// another line of comment
 *
 *     / * A C language comment
 *	 * A second line
 *	 *
 *	 * Another paragraph is the C comment
 *	 * /
 *
 * If CMODE should it look at some of the formatting after the comment lead in ?
 *
 * Problem with 'line start with non-alpha' ... consider a paragraph that is a quote
 * and so starts with a " or ' or similar. We would end up duplicating it on every line.
 * Maybe what the 2nd line starts with is key - which just leave the problem
 * of what to do if filling just one long line... maybe start subsequent lines with
 * nothing - the point is that, if this is wrong, it is easy to put something
 * there and reformat.
 *
 * Bug: in indent mode after a reformat this can go up to the start of the para reformatted
 * Happens when 1st line has indent, subsequent do not so are given an indent;
 * this indent then makes them the same indent as a line below which then/thus
 * becomes part of the paragraph.
 *
 * Return:
 *	TRUE  Success
 *	FALSE
 *	    No fill column set
 *	    Out of memory
 */
int	fill1para(void)
{
	LINE*	eopline;		// The last line in the paragraph
	int	utf8 = curbp->b_mode & MDUTF8;

	LineLn	wantIndent = 0;		// The desired indent - columns
	int	indentFrom2 = 0;	// Indent comes from 2nd line - ie all para lines indented

	LineLn	bpos;
	LineLn	lcol;			// Column on the line - work out if need to wrap

	int	didTab2Space = 0;	// For WFEDIT flag
	int	firstWord = 1;		// Current word is first word on the line

	LineLn	wch = 0;		// # chars in word being considered
	LineLn	wwid = 0;		// width of word being considered - units columns
	LineLn	sn = 0;			// # Spaces needed after word
	LineLn	sa = 0;			// # Spaces actual after word
	LineLn	nwch = 0;		// # chars in next word being considered
	LineLn	nwwid = 0;		// width of next word being considered - units columns
	LineLn	nsn = 0;		// # Spaces needed after next word
	LineLn	nsa = 1;		// # Spaces actual after next word


	if(fillcol == 0) {	/* no fill column set */
		mlwrite("%WNo fill column set");
		return(FALSE);
	}

	// Record the pointer to the last line in the paragraph
	gotoeop(FALSE, 1);
	eopline = curwp->w_dotp;

	/* and back top the beginning of the paragraph.
	 * The indent to use is that on the second line, but if there is not
	 * a second line - use the first line.
	 * If the buffer is not in indent mode: set the desired indent to 0.
	 */
	gotobop(FALSE, 1);

	/* Shortly a copy of any indent ends up here.
	 * The # bytes is the line length.
	 */
	tsetused(playText, 0);

	if(curbp->b_mode & (MDINDEN | MDCMOD | MDPHP)) {
		// Indent mode ... work a little harder to find an indent
		// Look at the first line:
		TEXT*	indentText = &ltxt(curwp->w_dotp);	// The text from which we take the indent
		wantIndent = GetWordCol(1);
		LineLn	indentBytes = WordColBytes;	// Bytes of the desired indent

		// This is where we start, so record that first line position in the
		// variables that we use as we go forwards
		lcol = wantIndent;
		bpos = WordColBytes;

		// If there is a second line, look at that:
		if(curwp->w_dotp != eopline) {
			curwp->w_dotp = lforw(curwp->w_dotp);
			LineLn	wi = GetWordCol(1);

			/* Take the 2nd line's indent now that we know that it exists.
			 * This allows the first line to have a different indent
			 * from the rest of the paragraph, eg a '*' bullet on
			 * line 1 and spaces (but matching indent) on the rest.
			 */
			wantIndent = wi;
			indentFrom2 = 1;
				indentText = &ltxt(curwp->w_dotp);

			indentBytes = WordColBytes;

			curwp->w_dotp = lback(curwp->w_dotp);

		} else if(curbp->b_mode & (MDCMOD | MDPHP)) {
			/* This is a 1 line paragraph. Avoid screwing up a long C comment
			 * by starting continuation lines with '/ *' instead of ' *'.
			 * Look for '/ *' at this line start, if so create an indent line
			 * that is correct.
			 */
			LineLn	wh = countWhiteBytes(indentText);

			if(tused(indentText) >= wh + 2 &&
			   tgetch(indentText, wh) == '/' && tgetch(indentText, wh + 1) == '*') {
				/* Line starts like this comment line does */
				if(tcpyt(playText2, indentText) != TRUE)
					return(FALSE);
				tputc(playText2, wh, (doublestar ? '*' : ' '));
				indentText = playText2;
			}
		}

		if(wantIndent > fillcol)
			wantIndent = fillcol;

		/* Copy the indent to playText.
		 * If we make changes the chances are that what is currently at
		 * the start of the 2nd line might get corrupted.
		 * Safest to grab a copy.
		 */
		if(tcpym(playText, ttext(indentText), indentBytes) != TRUE)
			return(FALSE);
	} else {
		// Non indent
		bpos = countWhiteBytes(&ltxt(curwp->w_dotp));
		lcol = GetNonBlankCol();
		// We only look at space & tab so don't need to worry about UTF-8 - phew!
	}

	/* Start at the first non white space on the first line.
	 * Then read a word at a time, wrapping it if needs be.
	 * One nice thing is that we just put one space after every word
	 * (unless the word ends '.') so we don't need to worry about tab
	 * expansion - except in the initial indent.
	 */

	curwp->w_doto = bpos;
	// lcol is the column

	// Until end of para, process lines
	while(1) {
		int	endSentence = 0;				// Last word ends '.!?'
		LineLn	odoto;
		int	eopl = curwp->w_dotp == eopline;		// Started out on last line in para
		int	neopl = lforw(curwp->w_dotp) == eopline;	// Next line is last line in para

		BYTE	ch = 0;						// Current byte
		nwch = 0;
		nwwid = 0;
		nsa = 0;
		nsn = 1;

		// Move on if the next char is not white:
span_word:
		while(curwp->w_doto < lused(curwp->w_dotp) &&
		   (ch = lgetc(curwp->w_dotp, curwp->w_doto)) != ' ' && ch != '\t') {

			endSentence = ch == '.' || ch == '!' || ch == '?';

			// Increment byte, character and column positions:
			if(utf8) {
				// Get how many columns wide the character is:
				int	width = mk_wcwidth(utf8_decode(lgetcaddr(curwp->w_dotp, curwp->w_doto)));
				if(width <= 0)
					width = 1;
				nwwid += width;
				curwp->w_doto += CHAR_LEN(ch);
			} else {
				// All bytes are width 1
				curwp->w_doto++;
				nwwid++;
			}
			nwch++;
		}

		// One space needed except, maybe, after sentence
		nsn = 1 + (endSentence ? !FrenchSpacing : 0);


		/* Somehow we are at a word's end; either line end of tab/space.
		 * Look to see if, after any spaces, there is something that we should
		 * consider part of this word -- ie some non alphas followed by space.
		 */
		while(curwp->w_doto < lused(curwp->w_dotp)) {
			if((ch = lgetc(curwp->w_dotp, curwp->w_doto)) == '\t') {
				/* If we see a tab: change it to a space
				 * Always do this even if the tab actually results in the
				 * correct spacing - as it might after a '.'.
				 */
				lputc(curwp->w_dotp, curwp->w_doto, ' ');
				if( !didTab2Space) {
					lchange(WFEDIT);
					didTab2Space = 1;
				}

				ch = ' ';
			}

			if(ch == ' ') {
				// If we already have enough spaces, delete this one
				// Need to ldelete in case a mark is here
				if(nsa >= nsn) {
					ldelete(1, KL_DISCARD);
					continue;
				}

				nsa++;
				curwp->w_doto++;
				continue;
			}

			/* A non blank - but still on the line.
			 * If non alphas followed by space ?
			 */
			if(ch > UTF8_MAX_CP1 || Isalnum(ch)) {
				/* Found start of word without non alpha before
				 */
				break;
			}

			/* Span over sequence of punctuation chars to see what
			 * follows it.
			 * This is about avoiding wrapping a punctuation character to the
			 * start of the next line - which could mess up start/end of paragraph
			 * detection.
			 *
			 * Beware just stopping at end of line - that is as much of a
			 * break as is space/tab.
			 */
			odoto = curwp->w_doto;

			while(++curwp->w_doto <= lused(curwp->w_dotp)) {
				if(curwp->w_doto == lused(curwp->w_dotp)) {
					ch = ' ';	// Virtual space at end of line
				} else {	/* Get what is really there.
						 * If it is a tab - replace with a space.
						 */
					if((ch = lgetc(curwp->w_dotp, curwp->w_doto)) == '\t') {
						lputc(curwp->w_dotp, curwp->w_doto, ' ');
						if( !didTab2Space) {
							lchange(WFEDIT);
							didTab2Space = 1;
						}

						ch = ' ';
					}
				}

				if(ch == ' ') {
					// Punct followed by space
					// Extend the previous word to here
					nwch += nsa + curwp->w_doto - odoto;
					nwwid += nsa + curwp->w_doto - odoto;
					nsa = 0;

					// Continue scanning the next word
					goto span_word;
				}
				if(ch > UTF8_MAX_CP1 || Isalnum(ch)) {
					// Punct next to word, OK to break before
					break;
				}
			}

			curwp->w_doto = odoto;	// move back over punct

			break;
		}

		// w_doto is after the space(s) after the new word, maybe beyond line end

		// This is not nice - there are lots of corner cases.
		// Some of the code below might never be executed.

		// End of line, nothing follows on the line
		if(nwch == 0) {
			// End of paragraph ?
			if(eopl) {
				break;
			}

			// Peek at the next line, get the width of the first word on the next line.
			// Join it unless the first word would make the current line too long,
			// in which case move to the next line:
			LineLn	nlwwid = 0;
			LineLn	savDoto = curwp->w_doto;

			curwp->w_dotp = lforw(curwp->w_dotp);

			// We know that this will work else we would have hit end of para
			// DisCol_MetaChar or what ?
			// indentChrs is the number of characters (prob == bytes) that make up the indent
			LineLn	indentChrs = curwp->w_doto = getdoto(curwp, curwp->w_dotp, wantIndent, DisCol_MetaChar);

			while(curwp->w_doto < lused(curwp->w_dotp) &&
			      (ch = lgetc(curwp->w_dotp, curwp->w_doto)) != ' ' && ch != '\t') {

				// Increment byte, character and column positions:
				if(utf8) {
					// Get how many columns wide the character is:
					int	width = mk_wcwidth(utf8_decode(lgetcaddr(curwp->w_dotp, curwp->w_doto)));
					if(width <= 0)
						width = 1;
					nlwwid += width;
					curwp->w_doto += CHAR_LEN(ch);
				} else {
					// All bytes are width 1
					curwp->w_doto++;
					nlwwid++;
				}
			}

			// Return to whence we came:
			curwp->w_dotp = lback(curwp->w_dotp);
			curwp->w_doto = savDoto;

			// First word on the next line will fit on this line -- join the line
			if(lcol + wwid + sn + nlwwid < fillcol) {

				if(wantIndent && indentFrom2) {
					// Remove the indent from the next line
					// Move to it, delete, come back here
					LineLn	cdoto = curwp->w_doto;
					curwp->w_doto = 0;
					curwp->w_dotp = lforw(curwp->w_dotp);

					ldelete(indentChrs, KL_DISCARD);

					// Return to where we were
					curwp->w_doto = cdoto;
					curwp->w_dotp = lback(curwp->w_dotp);
				}
				ldelete(1, KL_DISCARD);
				if(!linsert(sn - sa, ' '))
					return(FALSE);
				sa = sn;

				if(neopl)	// if joined/destroyed eopline, make this the eopline
					eopline = curwp->w_dotp;
				continue;
			}

			// This line is full: move to the next line

			// Remove trailing spaces
			if(sa) {
				moveBack(sa);
				ldelete(sa, KL_DISCARD);
			}

			moveForw(1);	// Moves to the next line

			if( !indentFromPlay(indentFrom2, wantIndent))
				return(FALSE);

			// Reset to as initial conditions - start a first line
			wwid = 0;
			wch = 0;
			sa = 0;
			sn = 0;
			firstWord = 1;
			lcol = wantIndent;

			continue;
		}

		// Next word will fit on the line after current + spaces
		if(lcol + wwid + sn + nwwid < fillcol) {

			if(sa < sn) {
				// Insert extra space after full stop
				moveBack(nwch + nsa);
				if(!linsert(sn - sa, ' '))
					return(FALSE);
				moveForw(nwch + nsa);
			}

			lcol += wwid + sn;
			wwid = nwwid;
			wch = nwch;
			sa = nsa;
			sn = nsn;
			firstWord = 0;

			continue;
		}

		// Current word will fit on the line, but break after it:
		if(lcol + wwid < fillcol) {

			if( (lcol + wwid + sn + nwwid == fillcol && nsa == 0)) {
				// can do nothing else than break if first word

				// Remove trailing spaces
				if(nsa) {
					moveBack(nsa);
					ldelete(nsa, KL_DISCARD);
				}

				// Don't move beyond the para
				if( !eopl) {
					moveForw(1);
					nwwid = nwch = nsn = nsa = 0;

					// Either move over the indent that is there, or insert it
					if( !indentFromPlay(indentFrom2, wantIndent))
						return(FALSE);
				}

				wwid = nwwid;
				wch = nwch;
				sa = nsa;
				sn = nsn;
				firstWord = 1;	// First word on the line
				lcol = wantIndent;

				continue;
			}
			if(firstWord) {
				// Trim spaces after the word - remove trailing spaces
				if(nsa) {
					moveBack(nsa);
					ldelete(nsa, KL_DISCARD);
				}

				// If there is anything left break; else move onto next line
				if(curwp->w_doto < lused(curwp->w_dotp)) {

					if(!lnewline(0))
						return(FALSE);
					// Insert indent
					if(wantIndent) {
						if(insStrLine(curwp->w_dotp, 0, ttext(playText), tused(playText), 0) < 0)
							return(FALSE);
						FixOffsets(curwp->w_dotp, 0, tused(playText), curbp, 1);
						lchange(WFEDIT);
						curwp->w_doto = tused(playText);
					}
				} else {
					// Don't move beyond the para
					if( !eopl) {
						moveForw(1);
						nwwid = nwch = nsn = nsa = 0;

						// Either move over the indent that is there, or insert it
						if( !indentFromPlay(indentFrom2, wantIndent))
							return(FALSE);
					}
				}
			} else {
				// Move back over the new word and insert a newline before it:
				moveBack(nwch + sa + nsa);

				if(!lnewline(0))
					return(FALSE);
				// Remove spaces after the last word
				ldelete(sa, KL_DISCARD);

				// Insert indent
				if(wantIndent) {
					if(insStrLine(curwp->w_dotp, 0, ttext(playText), tused(playText), 0) < 0)
						return(FALSE);
					FixOffsets(curwp->w_dotp, 0, tused(playText), curbp, 1);
					lchange(WFEDIT);
					curwp->w_doto = tused(playText);
				}

				// Back to scanning point
				moveForw(nwch + nsa);
			}

			if(eopl)	// if split eopline, move to new one
				eopline = curwp->w_dotp;

			wwid = nwwid;
			wch = nwch;
			sa = nsa;
			sn = nsn;
			firstWord = 1;	// First word on the line
			lcol = wantIndent;

			continue;
		}


		// The current word will not fit on the line,
		// move to next line if that is all that there is on the line
		if(firstWord) {
			if(curwp->w_dotp == eopline) {
				break;
			}

			moveForw(1);
			if(neopl)	// if joined/destroyed eopline, make this the eopline
				eopline = curwp->w_dotp;

			curwp->w_doto = getdoto(curwp, curwp->w_dotp, wantIndent, DisCol_MetaChar);
			wwid = 0;
			wch = 0;
			sa = 0;
			sn = 0;
			firstWord = 1;
			lcol = wantIndent;

			continue;
		}

		// Is this reached ?

		// The current word will not fit on the line, break before it
		moveBack(wch);
		if(!lnewline(0))
			return(FALSE);

		moveForw(wch);
		// Insert indent
		if(wantIndent) {
			if(insStrLine(curwp->w_dotp, 0, ttext(playText), tused(playText), 0) < 0)
				return(FALSE);
			FixOffsets(curwp->w_dotp, 0, tused(playText), curbp, 1);
			lchange(WFEDIT);
			curwp->w_doto = tused(playText);
		}
		if(neopl)	// if joined/destroyed eopline, make this the eopline
			eopline = curwp->w_dotp;
	}


	// Done filling the paragraph.

	// Trim trailing spaces:
	curwp->w_doto = lused(curwp->w_dotp);
	while(moveBack(1) && lgetc(curwp->w_dotp, curwp->w_doto) == ' ')
		if(ldelete(1, KL_DISCARD) == FALSE)
			return(FALSE);

	// Move to the last alpha on the line, end and then backward until we are in a word
	curwp->w_doto = lused(curwp->w_dotp);
	while(moveBack(1) && !inword())
		;

	// Which took us back one too far
	moveForw(1);

	return(TRUE);
}

/* Fill n paragraphs.
 *
 * Default count 1. Will be >= 0.
 *
 * Command:  fill-paragraph
 * Bound to: M-q
 *
 * Return:
 *	TRUE  Success
 *	FALSE
 *	    No film column set
 *	    Out of memory
 */
int	fillpara(int f, Cnt n)
{
	int	ret = TRUE;

	pnmark1();

	while(--n >= 0 && (ret = fill1para()))
		;

	return(ret);
}

/* Delete n paragraphs starting with the current one.
 *
 * n	# of paras to delete
 *
 * Default count 1. Will be >= 0.
 *
 * Command:  kill-paragraph
 * Bound to: M-C-w
 *
 * Return:
 *	TRUE  success
 *	FALSE hit buffer end
 *	      out of memory
 */
int	killpara(int f, Cnt n)
{
	register int	status;		/* returned status of functions */

	while(--n >= 0) {		/* for each paragraph to delete */

		/* mark out the end and beginning of the para to delete */
		if(gotoeop(FALSE, 1) == FALSE)
			return(FALSE);

		/* set the mark here */
		curwp->w_markp = curwp->w_dotp;
		curwp->w_marko = curwp->w_doto;

		/* go to the beginning of the paragraph */
		if(gotobop(FALSE, 1) == FALSE)
			return(FALSE);
		curwp->w_doto = 0;	/* force us to the beginning of line */

		/* and delete it */
		if((status = killregion(FALSE, 1)) != TRUE)
			return(status);

		/* and clean up the 2 extra lines */
		if( !ldelete(2, KL_KILL_APPEND))
			return(FALSE);
	}

	return(TRUE);
}


/* Count the # of words in the marked region,
 * along with average word sizes, # of chars, etc, and report on them.
 *
 * The count is ignored.
 *
 * Command:  count-words
 * Bound to: M-C-c
 *
 * Return:
 *	TRUE  success
 *	FALSE there is not a region
 */
int	wordcount(int f, Cnt n)
{
	LINE*	lp;		/* current line to scan */
	LineLn	offset;		/* BYTE offset of current char to scan */
	long	size;		/* size of region left to count */
	UniChar	ch;		/* current character to scan */
	int	wordflag;	/* are we in a word now? */
	int	lastword;	/* were we just in a word? */
	BufChrs	nwords;		/* total # of words */
	BufChrs	nchars;		/* total number of chars */
	BufChrs	letters;	/* total # chars in words (letters + digits)*/
	BufChrs	bytes;		/* total # bytes */
	BufLns	nlines;		/* total number of lines in region */
	int	avgch;		/* average number of chars/word */
	int	avgletters;	/* Average letters/word */
	int	status;		/* status return code */
	REGION	region;		/* region to look at */

	/* make sure we have a region to count */
	if((status = getregion(&region)) != TRUE)
		return(status);

	lp = region.r_linep;
	offset = region.r_offset;
	size = region.r_size;	// Size in characters

	/* count up things */
	lastword = FALSE;
	nchars = 0L;
	nwords = 0L;
	letters = 0L;
	bytes = 0L;
	nlines = 0;
	while(size > 0) {
		/* get the current character */
		if(offset == lused(lp)) {	/* end of line */
			ch = '\n';
			lp = lforw(lp);
			offset = 0;
			++nlines;
			--size;
			++bytes;
		} else {
			if(lisUtf8(lp)) {
				BYTE*	a = lgetcaddr(lp, offset);
				ch = utf8_decode(a);
				offset += CHAR_LEN(*a);
				bytes += CHAR_LEN(*a);
			} else {
				ch = lgetc(lp, offset);
				++offset;
				++bytes;
			}
			--size;
		}
		++nchars;

		/* What is it ?
		 * If it is a word char, add to the count of characters
		 * that make up words.
		 * If the last char was a in word & the current one isn't
		 * count another word.
		 * TODO utf8 - better classify characters
		 */
		wordflag = Isalnum(ch);
		if(wordflag == TRUE) {
			letters++;
			if(lastword == FALSE)
				++nwords;
		}
		lastword = wordflag;
	}

	/* and report on the info */
	if(nwords > 0L) {
		avgletters = (int)((100L * letters) / nwords);
		avgch = (int)((100L * nchars) / nwords);
	} else
		avgletters = avgch = 0;

	mlwrite("%IWords %D Chars %D Bytes %D Lines %d Avg chars/word %f letters/word %f",
		(long)nwords, (long)nchars, (long)bytes, (long)nlines, avgch, avgletters);

	return(TRUE);
}
#endif

/* end */
