/* Encode UniCode characters in different widths.
 * The idea is to produce over long encoding - for testing.
 *
 * Copyright: Alain D D Williams <addw@phcomp.co.uk> April 2016
 *
 * This program ignores its arguments.
 */


#include <stdio.h>
#include <string.h>

// Stuff stolen from estruct.h to get this to compile:
typedef	long	UniChar;		// Storage for Unicode code points
#define	CHAR	unsigned char		// Legacy, do not use
#define	BYTE	unsigned char		// For underlying storage

#include	"utf8.h"

/* Encode character uc as a UTF8 string - with specified length.
 * Return the string - in a static buffer.
 * Return NULL on error, utf8_errno will be set.
 *
 * Note that this will be NUL terminated, and if the input was U+0
 * the string returned will be the empty string - followed by a NUL byte.
 *
 * This will quite happily encode invalid code points.
 */
BYTE*	utf8_len_encode(UniChar uc, int length)
{
	static	BYTE	ret[UTF8_MAX_BYTES + 1];	// What we return
	int		min;				// Min length to encode
	BYTE*		s = ret;

	/* Work out minimum length needed, generating too long encoding
	 * is one thing - but minimum sizes need to be respected.
	 */
	if(uc <= UTF8_MAX_CP1)
		min = 1;
	else   if(uc <= UTF8_MAX_CP2)
		min = 2;
	else if(uc <= UTF8_MAX_CP3)
		min = 3;
	else
		min = 4;

	if(min > length) {
		fprintf(stderr, "Cannot encode U+%X in %d bytes, need %d\n",
			uc, length, min);
		return NULL;
	}

	if(length == 1)
		*s++ = (char) uc;
	else   if(length == 2) {
		*s++ = (char)( uc >>  6 | UTF8_B1_BITS2);
		*s++ = (char)((uc       & UTF8_Bn_VMASK) | UTF8_Bn_TPVALUE);
	} else if(length == 3) {
		*s++ = (char)( uc >> 12 | UTF8_B1_BITS3);
		*s++ = (char)((uc >>  6 & UTF8_Bn_VMASK) | UTF8_Bn_TPVALUE);
		*s++ = (char)((uc       & UTF8_Bn_VMASK) | UTF8_Bn_TPVALUE);
	} else {
		*s++ = (char)( uc >> 18 | UTF8_B1_BITS4);
		*s++ = (char)((uc >> 12 & UTF8_Bn_VMASK) | UTF8_Bn_TPVALUE);
		*s++ = (char)((uc >>  6 & UTF8_Bn_VMASK) | UTF8_Bn_TPVALUE);
		*s++ = (char)((uc       & UTF8_Bn_VMASK) | UTF8_Bn_TPVALUE);
	}

	*s = '\0';

	return(ret);
}


// The characters to encode, terminated by -1
UniChar	chars[] = {
	0xA3,		// Pound
	0x20AC,		// Euro
	-1
};

int	main(int argc, char* argv[])
{
	UniChar* ap;

	for(ap = chars; *ap != -1; ap++) {
		BYTE*	seq;
		int	len;

		for(len = 2; len <= 4; len++) {
			if(seq = utf8_len_encode(*ap, len)) {
				int	i;
				printf("Char U+%X '%s' len %d :", *ap, seq, len);
				for(i = 0; i < len; i++)
					printf(" %2.2x", seq[i]);
				printf("\n");
			}
		}
	}

	return 0;
}
